#include "widgetData.h"
#include "widgetPlot.h"
#include <QLabel>
#include <QTime>
#include <QFileDialog>
#include <QProcess>

widgetData::widgetData(widgetPlot *plotWidget, QWidget *parent) :
  QWidget(parent)
{
  // I/O stuff
  plotWidgetLoc = plotWidget;

  // Layout
  widgetDataLayout             = new QGridLayout();
  this                         ->setLayout(widgetDataLayout);

  // Buttons
  loadDataButton               = new QPushButton("Load data", this);
  reloadDataButton             = new QPushButton("Reload",    this);
  fitDataButton                = new QPushButton("Fit data",  this);
  widgetDataLayout             ->addWidget(loadDataButton,         0, 0, 1, 2);
  widgetDataLayout             ->addWidget(reloadDataButton,       0, 2, 1, 2);
  widgetDataLayout             ->addWidget(fitDataButton,          0, 4, 1, 2);
  reloadDataButton             ->setEnabled(false);

  // Connect buttons
  connect(loadDataButton,  SIGNAL(clicked()),this,SLOT(loadData()));
  connect(reloadDataButton,SIGNAL(clicked()),this,SLOT(reloadData()));
  connect(fitDataButton,   SIGNAL(clicked()),this,SLOT(fitData()));
  
  // Poles
  QLabel *zeroOrderPoleLabel   = new QLabel("Pole 0",this);
  QLabel *firstOrderPoleLabel  = new QLabel("Pole 1",this);
  QLabel *secondOrderPoleLabel = new QLabel("Pole 2",this);
  zeroOrderPoleSpinBox         = new QSpinBox(this);
  firstOrderPoleSpinBox        = new QSpinBox(this);
  secondOrderPoleSpinBox       = new QSpinBox(this);
  widgetDataLayout             ->addWidget(zeroOrderPoleLabel,     1, 0);
  widgetDataLayout             ->addWidget(zeroOrderPoleSpinBox,   1, 1);
  widgetDataLayout             ->addWidget(firstOrderPoleLabel,    1, 2);
  widgetDataLayout             ->addWidget(firstOrderPoleSpinBox,  1, 3);
  widgetDataLayout             ->addWidget(secondOrderPoleLabel,   1, 4);
  widgetDataLayout             ->addWidget(secondOrderPoleSpinBox, 1, 5);

  // Connect poles
  connect(zeroOrderPoleSpinBox,  SIGNAL(valueChanged(int)),this,SLOT(activateFit()));
  connect(firstOrderPoleSpinBox, SIGNAL(valueChanged(int)),this,SLOT(activateFit()));
  connect(secondOrderPoleSpinBox,SIGNAL(valueChanged(int)),this,SLOT(activateFit()));

  // Small output term
  termTextEdit            = new QPlainTextEdit(this);
  termTextEdit            ->setReadOnly(true);
  widgetDataLayout        ->addWidget(termTextEdit,                2, 0, 2, 6);

  // Set default paramters
  setDefault();
}

widgetData::~widgetData()
{

}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for default parameters /////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::setDefault()
{
  loadDataButton               ->setEnabled(true);
  fitDataButton                ->setEnabled(false);  
  zeroOrderPoleSpinBox         ->setEnabled(false);
  firstOrderPoleSpinBox        ->setEnabled(false);
  secondOrderPoleSpinBox       ->setEnabled(false);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*/////////////////////////// Slot for writing in termBox ///////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::writeToTermBox(const QString &text)
{
  termTextEdit->appendPlainText(QTime::currentTime().toString()+" "+text);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////// Slot for loading data files ////////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::loadData()
{
  // Get filename
  QString file;
  file = QFileDialog::getOpenFileName(this,tr("Load data file"), QDir::currentPath());
  if(file != NULL){

    // Set filename
    filename = file;

    // Write in term
    writeToTermBox("Loaded data file: "+filename);
    
    // Activate poles
    activatePoles();

    // Plot data
    plotWidgetLoc->plotData(filename,true);
  } else {
    // Write in term
    writeToTermBox("No data file provided");
  }

  // Activate reloadDataButton
  if (!(reloadDataButton->isEnabled())) {
    reloadDataButton->setEnabled(true);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////// Slot for reloading last data file //////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::reloadData()
{
  // Write in term
  writeToTermBox("Loaded data file: "+filename);
  
  // Activate poles
  activatePoles();
  
  // Plot data
  plotWidgetLoc->plotData(filename,false);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for poles activation ///////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::activatePoles()
{
  if (!(zeroOrderPoleSpinBox->isEnabled())) {
    zeroOrderPoleSpinBox         ->setEnabled(true);
  }
  if (!(firstOrderPoleSpinBox->isEnabled())) {
    firstOrderPoleSpinBox        ->setEnabled(true);
  }
  if (!(secondOrderPoleSpinBox->isEnabled())) {
    secondOrderPoleSpinBox       ->setEnabled(true);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*/////////////////////// Slot for fitting activation ///////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::activateFit()
{
  int zero = zeroOrderPoleSpinBox  ->value();
  int one  = firstOrderPoleSpinBox ->value();
  int two  = secondOrderPoleSpinBox->value();

  if (zero+one+two > 0) {
    fitDataButton->setEnabled(true);
  } else {
    fitDataButton->setEnabled(false);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////// Slot for fitting data files ////////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetData::fitData()
{
  int zero = zeroOrderPoleSpinBox  ->value();
  int one  = firstOrderPoleSpinBox ->value();
  int two  = secondOrderPoleSpinBox->value();

  writeToTermBox("Fitting "+filename+" data file using:");
  writeToTermBox(QString::number(zero)+" 0th order pole");
  writeToTermBox(QString::number(one)+" 1st order pole(s)");
  writeToTermBox(QString::number(two)+" 2nd order pole(s)");

  QStringList options;
  options << filename << QString::number(zero) << QString::number(one) << QString::number(two);
  process = new QProcess(this);
  process->start("fit_dispersion_data", options);
  process->waitForFinished(-1);
  QString out = process->readAllStandardOutput();
  writeToTermBox(out);
  QString file=filename+"_samples";
  plotWidgetLoc->plotFit(file);
}
