#include "widgetMaterial.h"
#include "widgetPlot.h"
#include <QLabel>
#include <QTime>
#include <QFileDialog>
#include <QProcess>
#include <QDirIterator>
#include <QInputDialog>

widgetMaterial::widgetMaterial(widgetPlot *plotWidget, QWidget *parent) :
  QWidget(parent)
{
  // I/O stuff
  plotWidgetLoc = plotWidget;

  // Path for materials
  QString appPath = QCoreApplication::applicationDirPath();
  materialPath = appPath.append("/../../data/materials/");
  currPath = QDir::currentPath();

  // Layout
  widgetMaterialLayout            = new QGridLayout();
  this                            ->setLayout(widgetMaterialLayout);

  // Buttons
  plotDataButton                  = new QPushButton("Plot data", this);
  plotFitButton                   = new QPushButton("Plot fit", this);
  plotBothButton                  = new QPushButton("Plot both", this);
  deleteMaterialButton            = new QPushButton("Delete from library", this);
  
  widgetMaterialLayout            ->addWidget(plotDataButton,          0, 1, 1, 1);
  widgetMaterialLayout            ->addWidget(plotFitButton,         1, 1, 1, 1);
  widgetMaterialLayout            ->addWidget(plotBothButton,          2, 1, 1, 1);
  widgetMaterialLayout            ->addWidget(deleteMaterialButton,        3, 1, 1, 1);

  // Connect buttons
  connect(plotDataButton,SIGNAL(clicked()),this,SLOT(plotData()));
  connect(plotFitButton, SIGNAL(clicked()),this,SLOT(plotFit()));
  connect(plotBothButton,SIGNAL(clicked()),this,SLOT(plotBoth()));

  // List of all materials in data repository
  materialList                    = new QListWidget(this);
  widgetMaterialLayout            ->addWidget(materialList, 0, 0, 6, 1);

  // Load existing materials from materials repository
  loadAllMaterials();

  // Disactivate buttons
  plotDataButton                  ->setEnabled(false);
  plotFitButton                   ->setEnabled(false);
  plotBothButton                  ->setEnabled(false);
  deleteMaterialButton            ->setEnabled(false);

  // Connect selection changes in materialList to selectedMaterial slot
  connect(materialList,SIGNAL(currentItemChanged(QListWidgetItem*,QListWidgetItem*)),this,SLOT(selectedMaterial(QListWidgetItem*)));
  
  // Connect saveSelectedButton from widgetPlot to saveSelected slot
  connect(plotWidgetLoc->saveSelectedButton,SIGNAL(clicked()),this,SLOT(saveSelected()));

  // Connect deleteMaterialButton to deleteMaterial slot
  connect(deleteMaterialButton,SIGNAL(clicked()),this,SLOT(deleteMaterial()));
}

widgetMaterial::~widgetMaterial()
{

}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*/////////////////////// Slot for loading existing materials ///////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::loadAllMaterials()
{
  // Deselect graphs
  plotWidgetLoc->deselectGraphs();

  // Clear list
  materialList->clear();

  // Clear file paths
  dataFile.clear();
  fitFile.clear();

  // Load materials
  QDirIterator it(materialPath, QDir::Dirs | QDir::NoDotAndDotDot);
  while (it.hasNext()) {
    it.next();
    fileName = it.fileName();
    materialList->addItem(fileName);
  }
  // Reset origin path
  //  QDir::setCurrent(currPath);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*/////////////////////// Slot for handling selected material ///////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::selectedMaterial(QListWidgetItem *item)
{
  if (item) {
    // Activate buttons if necessary
    if (!(plotDataButton->isEnabled())) {
      plotDataButton->setEnabled(true);
    }
    if (!(plotFitButton->isEnabled())) {
      plotFitButton->setEnabled(true);
    }
    if (!(plotBothButton->isEnabled())) {
      plotBothButton->setEnabled(true);
    }
    if (!(deleteMaterialButton->isEnabled())) {
      deleteMaterialButton->setEnabled(true);
    }
    
    // Fill filenames so plotting buttons point to the right files
    dataFile = item->text();
    dataFile.append("/"+dataFile+"_data");
    dataFile.prepend(materialPath);
    fitFile = item->text();
    fitFile.append("/"+fitFile+"_fit");
    fitFile.prepend(materialPath);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////////////// Slot for plotting data /////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::plotData()
{
  plotWidgetLoc->plotData(dataFile, true);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////////////// Slot for plotting model ////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::plotFit()
{
  plotWidgetLoc->plotFit(fitFile);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////////////// Slot for plotting both /////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::plotBoth()
{
  plotWidgetLoc->plotData(dataFile, true);
  plotWidgetLoc->plotFit(fitFile);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*////////////////////////////// Slot for deleting data /////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::deleteMaterial()
{
  // Get current item
  QListWidgetItem *currentItem = materialList->currentItem();

  // Remove files
  QString file = currentItem->text();
  QString command = "rm -r "+file.prepend(materialPath);
  system(qPrintable(command));

  // Reload list
  loadAllMaterials();
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*/////////////////Slot for saving selected material from plot //////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetMaterial::saveSelected()
{
  // Open dialog to get the name of the material
  QString defaultName = plotWidgetLoc->realFileTable[plotWidgetLoc->currentRealGraph].baseName();
  bool ok;
  materialName = new QDialog(this);
  QString name = QInputDialog::getText(this, tr("QInputDialog::getText()"),
				       tr("Material Name:"), QLineEdit::Normal,
				       defaultName, &ok);
  // If name is ok
  if (ok && !name.isEmpty()) {
    // Get the address of the fit file
    QString address = plotWidgetLoc->realFileTable[plotWidgetLoc->currentRealGraph].absoluteFilePath();
    
    // Create material directory
    QString command = "mkdir "+materialPath+name;
    system(qPrintable(command));      
    
    // Copy fit file
    command = "cp "+address+" "+materialPath+name+"/"+name+"_fit";
    system(qPrintable(command));
    
    // Get the address of the data file
    address = address.split("_samples").first();
    
    // Copy data file
    command = "cp "+address+" "+materialPath+name+"/"+name+"_data";
    system(qPrintable(command));
    
    // Get the address of the model file
    address.append("_model");
    
    // Copy model file
    command = "cp "+address+" "+materialPath+name+"/"+name+"_model";
    system(qPrintable(command));
    
    // Reload materials list
    loadAllMaterials();
  }
}
