#include <QLabel>
#include <QFileDialog>
#include <QTextStream>
#include "qcustomplot.h"
#include "widgetPlot.h"

widgetPlot::widgetPlot(QWidget *parent) :
  QWidget(parent)
{
  // Layout
  widgetPlotLayout                = new QGridLayout();
  this                            ->setLayout(widgetPlotLayout);

  // Prepare plots for real and imaginary parts
  realPlot = new QCustomPlot(this);
  imagPlot = new QCustomPlot(this);
  realPlot->xAxis->setLabel("Omega (THz)");
  realPlot->yAxis->setLabel("eps_r");
  realPlot->legend->setVisible(true);
  realPlot->setInteractions(QCP::iRangeZoom | QCP::iSelectPlottables | QCP::iSelectLegend);

  imagPlot->xAxis->setLabel("Omega (THz)");
  imagPlot->yAxis->setLabel("eps_i");
  imagPlot->legend->setVisible(true);
  imagPlot->setInteractions(QCP::iRangeZoom | QCP::iSelectPlottables | QCP::iSelectLegend);

  // Add plots to layout
  widgetPlotLayout->addWidget(realPlot, 0, 0, 1, 4);
  widgetPlotLayout->addWidget(imagPlot, 2, 0, 1, 4);

  // Buttons
  removeSelectedButton         = new QPushButton("Remove selected", this);
  removeAllButton              = new QPushButton("Remove all",      this);
  rescaleButton                = new QPushButton("Rescale",         this);
  saveSelectedButton           = new QPushButton("Save selected",   this);
  widgetPlotLayout             ->addWidget(removeSelectedButton,    1, 0, 1, 1);
  widgetPlotLayout             ->addWidget(removeAllButton,         1, 1, 1, 1);
  widgetPlotLayout             ->addWidget(saveSelectedButton,      1, 2, 1, 1);
  widgetPlotLayout             ->addWidget(rescaleButton,           1, 3, 1, 1);
  removeSelectedButton         ->setEnabled(false);
  removeAllButton              ->setEnabled(false);
  rescaleButton                ->setEnabled(false);
  saveSelectedButton           ->setEnabled(false);

  // Connect buttons
  // Note that saveSelectedButton is connected in widgetMaterial widget
  connect(removeSelectedButton,SIGNAL(clicked()),this,SLOT(removeSelectedGraph()));
  connect(removeAllButton,     SIGNAL(clicked()),this,SLOT(removeAllGraphs()));
  connect(rescaleButton,       SIGNAL(clicked()),this,SLOT(rescaleGraphs()));

  // Connect curve selection and deselection to selectPlottable and modifySelection slots
  currentRealGraph = NULL;
  currentImagGraph = NULL;
  connect(realPlot, SIGNAL(selectionChangedByUser()),                           this,SLOT(deselectGraphs()));
  connect(imagPlot, SIGNAL(selectionChangedByUser()),                           this,SLOT(deselectGraphs()));
  connect(realPlot, SIGNAL(plottableClick(QCPAbstractPlottable*, QMouseEvent*)),this,SLOT(selectGraphs(QCPAbstractPlottable*)));
  connect(imagPlot, SIGNAL(plottableClick(QCPAbstractPlottable*, QMouseEvent*)),this,SLOT(selectGraphs(QCPAbstractPlottable*)));  

  // Connect legend selection with selectLegend slot
  connect(realPlot, SIGNAL(legendClick(QCPLegend*, QCPAbstractLegendItem*, QMouseEvent*)), this, SLOT(selectLegend(QCPLegend*, QCPAbstractLegendItem*)));
  connect(imagPlot, SIGNAL(legendClick(QCPLegend*, QCPAbstractLegendItem*, QMouseEvent*)), this, SLOT(selectLegend(QCPLegend*, QCPAbstractLegendItem*)));
}  

widgetPlot::~widgetPlot()
{

}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for plotting data and fit //////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::plotDataAndFit(QString filename)
{
  plotData(filename,true);
  plotFit(filename);
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*//////////////////////////// Slot for plotting data ///////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::plotData(QString dataFilename, bool doClearGraph)
{
  // Clear graphs if new data is loaded
  if (doClearGraph) {
    realPlot->clearGraphs();
    imagPlot->clearGraphs();
  }

  // Format data
  QString command = "awk '{$1=$1}1' OFS=\"\t\" "+dataFilename+" > my_file && mv my_file "+dataFilename;
  system(qPrintable(command));

  // Read data file
  QFile loadFile(dataFilename);
  if (loadFile.open(QFile::ReadOnly | QFile::Truncate)) {

    QTextStream in(&loadFile);
    QString readQString;
    QStringList lstLine;
    omega.clear();
    epsReal.clear();
    epsImag.clear();
    
    // Layers
    while(!in.atEnd()) {
      readQString = in.readLine();
      lstLine = readQString.split("\t");
      omega.append(lstLine[0].toDouble());
      epsReal.append(lstLine[1].toDouble());
      epsImag.append(lstLine[2].toDouble());
    }
  }
  
  // Add to real plot
  realPlot->addGraph();
  realFileTable[realPlot->graph()] = QFileInfo (dataFilename);
  realPlot->graph()->setData(omega, epsReal);
  realPlot->graph()->setName(realFileTable[realPlot->graph()].baseName());
  realPlot->graph()->setPen(QPen(Qt::blue));
  realPlot->graph()->rescaleAxes();
  realLegendItem = realPlot->legend->itemWithPlottable(realPlot->graph());
  realLegendItem->setSelectable(true);
  realPlot->replot();

  // Add to imaginary plot
  imagPlot->addGraph();
  imagFileTable[imagPlot->graph()] = QFileInfo (dataFilename);
  imagPlot->graph()->setData(omega, epsImag);
  imagPlot->graph()->setName(imagFileTable[imagPlot->graph()].baseName());
  imagPlot->graph()->setPen(QPen(Qt::blue));
  imagPlot->graph()->rescaleAxes();
  imagLegendItem = imagPlot->legend->itemWithPlottable(imagPlot->graph());
  imagLegendItem->setSelectable(true);
  imagPlot->replot();

  // Fill typeOfGraph
  realTypeOfGraph[realPlot->graph()] = "data";
  imagTypeOfGraph[imagPlot->graph()] = "data";

  // Fill graphTable
  graphTable[realPlot->graph()] = imagPlot->graph();
  graphTable[imagPlot->graph()] = realPlot->graph();

  // Enable removeAll button if not already done
  if (!(removeAllButton->isEnabled())) {
    removeAllButton->setEnabled(true);
  }

  // Enable rescale button if not already done
  if (!(rescaleButton->isEnabled())) {
    rescaleButton->setEnabled(true);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////////// Slot for plotting fit ///////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::plotFit(QString fitFilename)
{
  // Format data
  QString command = "awk '{$1=$1}1' OFS=\"\t\" "+fitFilename+" > my_file && mv my_file "+fitFilename;
  system(qPrintable(command));

  // Read data file
  QFile loadFile(fitFilename);
  if (loadFile.open(QFile::ReadOnly | QFile::Truncate)) {
    QTextStream in(&loadFile);
    QString readQString;
    QStringList lstLine;
    omega.clear();
    epsReal.clear();
    epsImag.clear();
    
    // Layers
    while(!in.atEnd()) {
      readQString = in.readLine();
      lstLine = readQString.split("\t");
      omega.append(lstLine[0].toDouble());
      epsReal.append(lstLine[1].toDouble());
      epsImag.append(lstLine[2].toDouble());
    }
  }
  
  // Add to real plot
  realPlot->addGraph();
  realFileTable[realPlot->graph()] = QFileInfo (fitFilename);
  realPlot->graph()->setData(omega, epsReal);
  realPlot->graph()->setName(realFileTable[realPlot->graph()].baseName());
  realPlot->graph()->setPen(QPen(Qt::red));
  realPlot->graph()->rescaleAxes();
  realPlot->replot();

  // Add to imaginary plot
  imagPlot->addGraph();
  imagFileTable[imagPlot->graph()] = QFileInfo (fitFilename);
  imagPlot->graph()->setData(omega, epsImag);
  imagPlot->graph()->setName(imagFileTable[imagPlot->graph()].baseName());
  imagPlot->graph()->setPen(QPen(Qt::red));
  imagPlot->graph()->rescaleAxes();
  imagPlot->replot();

  // Fill typeOfGraph
  realTypeOfGraph[realPlot->graph()] = "fit";
  imagTypeOfGraph[imagPlot->graph()] = "fit";

  // Fill graphTable
  graphTable[realPlot->graph()] = imagPlot->graph();
  graphTable[imagPlot->graph()] = realPlot->graph();

  // Enable removeAll button if not already done
  if (!(removeAllButton->isEnabled())) {
    removeAllButton->setEnabled(true);
  }

  // Enable rescale button if not already done
  if (!(rescaleButton->isEnabled())) {
    rescaleButton->setEnabled(true);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for plottable selection ////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::selectGraphs(QCPAbstractPlottable* graph)
{  
  // Select related graph in other plot, depends on which was selected initially
  if (graph->parent() == realPlot) {
    currentRealGraph  = graph;
    currentRealGraph  ->setSelected(true);
    currentImagGraph  = graphTable[graph];
    currentImagGraph  ->setSelected(true);
    realLegendItem    = realPlot->legend->itemWithPlottable(currentRealGraph);
    imagLegendItem    = imagPlot->legend->itemWithPlottable(currentImagGraph);
    realLegendItem    ->setSelected(true);
    imagLegendItem    ->setSelected(true);
    imagPlot          ->replot();
  }

  if (graph->parent() == imagPlot) {
    currentImagGraph  = graph;
    currentImagGraph  ->setSelected(true);
    currentRealGraph  = graphTable[graph];
    currentRealGraph  ->setSelected(true);
    realLegendItem    = realPlot->legend->itemWithPlottable(currentRealGraph);
    imagLegendItem    = imagPlot->legend->itemWithPlottable(currentImagGraph);
    realLegendItem    ->setSelected(true);
    imagLegendItem    ->setSelected(true);
    realPlot          ->replot();
  }

  // Activate buttons
  removeSelectedButton->setEnabled(true);
  if (realTypeOfGraph[currentRealGraph] == "fit") {
    saveSelectedButton->setEnabled(true);
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for deselecting graphs /////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::deselectGraphs()
{  
  if (!(currentRealGraph == NULL) && !(currentImagGraph == NULL)) {    
    // Desactivate buttons
    removeSelectedButton->setEnabled(false);
    saveSelectedButton  ->setEnabled(false);

    // Deselect all
    currentRealGraph    ->setSelected(false);
    currentImagGraph    ->setSelected(false);
    realLegendItem      ->setSelected(false);
    imagLegendItem      ->setSelected(false);
    realPlot            ->replot();
    imagPlot            ->replot();

    // Reset current values
    currentRealGraph    = NULL;
    currentImagGraph    = NULL;
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*//////////////////////////// Slot for legend selection ////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::selectLegend(QCPLegend* legend, QCPAbstractLegendItem* legendItem)
{  
  // Cast legendItem  
  // I do not understand this cast, but it is shown in the examples of QCP package...
  QCPPlottableLegendItem *legendPlottable = qobject_cast<QCPPlottableLegendItem *>(legendItem);
  
  // Apply selection
  if (legendItem->selected()) {
      selectGraphs(legendPlottable->plottable());
  }
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for removing selected graph ////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::removeSelectedGraph()
{ 
  // Remove graph and replot
  realPlot            ->removePlottable(currentRealGraph);
  imagPlot            ->removePlottable(currentImagGraph);
  realPlot            ->replot();
  imagPlot            ->replot();

  // Disable selection-related buttons
  removeSelectedButton->setEnabled(false);
  saveSelectedButton  ->setEnabled(false);

  // Disable removeAllButton if necessary
  if (realPlot->graphCount() == 0) {
    removeAllButton   ->setEnabled(false);
    rescaleButton     ->setEnabled(false);
  }

  // Reset current values
  currentRealGraph    = NULL;
  currentImagGraph    = NULL;
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for removing all graphs ////////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::removeAllGraphs()
{  
  // Remove graphs and replot
  realPlot            ->clearGraphs();
  imagPlot            ->clearGraphs();
  realPlot            ->replot();
  imagPlot            ->replot();

  // Disable buttons
  removeSelectedButton->setEnabled(false);
  saveSelectedButton  ->setEnabled(false);
  removeAllButton     ->setEnabled(false);
  rescaleButton       ->setEnabled(false);

  // Reset current values
  currentRealGraph    = NULL;
  currentImagGraph    = NULL;
}

/*///////////////////////////////////////////////////////////////////////////////////*/
/*///////////////////////// Slot for plotting data and fit //////////////////////////*/
/*///////////////////////////////////////////////////////////////////////////////////*/
void widgetPlot::rescaleGraphs()
{
  realPlot->graph(0)->rescaleAxes(false);
  realPlot->replot();
  imagPlot->graph(0)->rescaleAxes(false);
  imagPlot->replot();
  
  if (realPlot->graphCount() >= 2) {
    for (int iPlot=1; iPlot<=realPlot->graphCount()-1; iPlot++) {
      realPlot->graph(iPlot)->rescaleAxes(true);
      realPlot->replot();
      imagPlot->graph(iPlot)->rescaleAxes(true);
      imagPlot->replot();
    }
  }
}
