PROGRAM CONVERT_CRITICAL_POINTS_TO_GENDISP
  !----------------------------------------------------------------------
  !---- Convert critical points model to generalized dispersive model
  !---- The mandatory input format for critical points model is :
  !---- 
  !
  !---- IN
  !---- dataFile : file containing critical points data
  !                the mandatory data format for a CP files is:
  !                eps
  !                0.0  0.0  omegaD  gammaD  (Drude pole)
  !                A    phi  omega   gamma   (CP pole)
  !                A    phi  omega   gamma   (CP pole)
  !
  !---- OUT
  !  
  !----------------------------------------------------------------------
  !---- Shared modules
  !----------------------------------------------------------------------
  USE IO_FILES  
  USE DISPERSION_FIT
  USE MATERIALS_STRUCTURES
  USE MATH_PHYS_CONSTANTS
  USE ISO_C_BINDING  
  !----------------------------------------------------------------------  
  IMPLICIT NONE
  !----------------------------------------------------------------------
  !---- Shared variable
  !----------------------------------------------------------------------
  INTEGER(C_INT)                              :: nArg, nPole, nCols, nRows, &
       iPole, nUnit, iOmega
  INTEGER(C_INT),                   PARAMETER :: nSamples = 500

  CHARACTER(LEN=100, KIND=C_CHAR)             :: dataFile, outputFile

  REAL(C_DOUBLE)                              :: epsCP, epsGD, A, phi, &
       omega, gamma, c, d, e, f, dOmega
  REAL(C_DOUBLE),                   PARAMETER :: fMin = 0.3  ! in PHz
  REAL(C_DOUBLE),                   PARAMETER :: fMax = 1.50 ! in PHz
  REAL(C_DOUBLE)                              :: omegaMin, omegaMax
  REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: criticalPts, genDisp
  !----------------------------------------------------------------------
  !---- Get arguments
  !----------------------------------------------------------------------
  nArg = COMMAND_ARGUMENT_COUNT()
  IF (nArg /= 1) THEN
     WRITE(6,'(a)') 'You did not provide the right number of arguments'
     WRITE(6,'(a)') 'Please provide an experimental data file'
     STOP
  ENDIF
  !
  CALL GETARG(1,dataFile)
  !----------------------------------------------------------------------
  !---- Testing for bad input
  !----------------------------------------------------------------------
  !
  !----------------------------------------------------------------------
  !---- Read input file and convert
  !----------------------------------------------------------------------
  CALL GET_SIZE_OF_FILE(TRIM(dataFile), nRows, nCols)
  nPole = nRows - 1
  ALLOCATE(criticalPts(4,nPole))
  ALLOCATE(genDisp    (4,nPole))
  criticalPts(:,:) = 0.0
  genDisp(:,:)     = 0.0
  !
  ALLOCATE(Mat(1), MatDisp(1))
  ALLOCATE(MatDisp(1)%pole0(0))
  ALLOCATE(MatDisp(1)%pole1(0))
  ALLOCATE(MatDisp(1)%pole2(nPole))
  Mat(1)%iDispMat = 1
  MatDisp(1)%iMat = 1
  !
  CALL GET_NEW_FILE_UNIT(nUnit)
  OPEN(UNIT=nUnit, FILE=TRIM(dataFile))
  !
  READ(nUnit,*) epsCP
  Mat(1)%eps(:) = epsCP
  !
  DO iPole=1,nPole
     READ(nUnit,*) criticalPts(:,iPole)
     !
     IF (iPole == 1) THEN ! First pole is Drude
        c     = criticalPts(3,iPole)**2
        d     = 0.0
        e     = 0.0
        f     = criticalPts(4,iPole)
     ELSE                 ! Other poles are CP
        A     = criticalPts(1,iPole)
        phi   = criticalPts(2,iPole)
        omega = criticalPts(3,iPole)
        gamma = criticalPts(4,iPole)
        !
        c     = 2.0*A*omega*(omega*COS(phi) - gamma*SIN(phi))
        d     = 2.0*A*omega*SIN(-phi)
        e     = gamma**2 + omega**2
        f     = 2.0*gamma
     ENDIF
     !
     MatDisp(1)%pole2(iPole)%c = c*1E-30 ! Convert in PHz**2
     MatDisp(1)%pole2(iPole)%d = d*1E-15 ! Convert in PHz
     MatDisp(1)%pole2(iPole)%e = e*1E-30 ! Convert in PHz**2
     MatDisp(1)%pole2(iPole)%f = f*1E-15 ! Convert in PHz
  ENDDO
  !----------------------------------------------------------------------
  !---- Output for Data_Disp
  !----------------------------------------------------------------------
  CALL GET_NEW_FILE_UNIT(nUnit)
  OPEN(UNIT=nUnit, FILE=TRIM(dataFile)//'_model')
  WRITE(nUnit,'(i1,a,i1)' ) 0,' ',nPole
  WRITE(nUnit,'(2(e15.7))') Mat(1)%eps(1), 0.0
  !
  DO iPole=1,nPole
     c = MatDisp(1)%pole2(iPole)%c
     d = MatDisp(1)%pole2(iPole)%d
     e = MatDisp(1)%pole2(iPole)%e
     f = MatDisp(1)%pole2(iPole)%f
     WRITE(nUnit,'(4(e15.7))') c*1E12, d*1E6, e*1E12, f*1E6 ! Convert from PHz to GHz     
  ENDDO
  CLOSE(nUnit)
  !----------------------------------------------------------------------
  !---- Write output files
  !----------------------------------------------------------------------
  outputFile = TRIM(dataFile)//'_samples'
  CALL GET_NEW_FILE_UNIT(nUnit)
  omegaMin = 2.0*Pi*fMin
  omegaMax = 2.0*Pi*fMax
  CALL WRITE_PERMITTIVITY(outputFile, nUnit, 1, omegaMin, omegaMax, nSamples)
  !----------------------------------------------------------------------
END PROGRAM CONVERT_CRITICAL_POINTS_TO_GENDISP
