PROGRAM FIT_DISPERSION_DATA
  !----------------------------------------------------------------------
  !---- Program for fitting experimental dispersion data on generalized model
  !
  !---- IN
  !---- dataFile      : File containing the experimental dispersion data
  !----                 The mandatory data format is:
  !----                 omega(THz) Re(eps) Im(eps)
  !
  !---- OUT
  !---- *_disp_param  : File containing the generalized dispersion parameters
  !----                 ready to use in the DG solver
  !---- disp_quality  : A quality measurement for the fit
  !---- *_real_plot   : File containing a sample of the fitted 
  !----                 real part of the permittivty
  !---- *_imag_plot   : File containing a sample of the fitted 
  !----                 imaginary part of the permittivty
  !----------------------------------------------------------------------
  !---- Shared modules
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  USE DISPERSION_FIT
  USE IO_FILES
  USE MATERIALS_STRUCTURES
  !----------------------------------------------------------------------  
  IMPLICIT NONE
  !----------------------------------------------------------------------
  !---- Shared variable
  !----------------------------------------------------------------------
  INTEGER(C_INT)                            :: nArg, nPole0, nPole1, &
       nPole2, iPole, iSol, nUnit
  INTEGER(C_INT),                 PARAMETER :: nSamples = 200

  CHARACTER(LEN=100, KIND=C_CHAR)           :: dataFile, outputFile, arg
  CHARACTER(LEN=1,   KIND=C_CHAR)           :: cPole0, cPole1, cPole2

  REAL(C_DOUBLE)                            :: errorR, errorI, c, d, e, f
  REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: omega, epsR, epsI, sol
  !----------------------------------------------------------------------
  !---- Get arguments
  !----------------------------------------------------------------------
  nArg = COMMAND_ARGUMENT_COUNT()
  IF (nArg /= 4) THEN
     WRITE(6,'(a)') 'You did not provide the right number of arguments'
     WRITE(6,'(a)') 'Please provide:'
     WRITE(6,'(a)') '      - An experimental data file'
     WRITE(6,'(a)') '      - A number of zero-order pole (zero or one)'
     WRITE(6,'(a)') '      - A number of first-order poles'
     WRITE(6,'(a)') '      - A number of second-order poles'
     STOP
  ENDIF
  !
  CALL GETARG(1,dataFile)
  CALL GETARG(2,cPole0)
  CALL GETARG(3,cPole1)
  CALL GETARG(4,cPole2)
  !
  CALL CHAR_TO_INT(cPole0,nPole0)
  CALL CHAR_TO_INT(cPole1,nPole1)
  CALL CHAR_TO_INT(cPole2,nPole2)
  !----------------------------------------------------------------------
  !---- Testing for bad input
  !----------------------------------------------------------------------
  IF (nPole0 .GT. 1) THEN
     WRITE(6,'(a)') 'Number of zero-order pole cannot be higher than 1'
     WRITE(6,'(a)') 'I will set the number of zero-order pole to 1 for you'
     nPole0 = 1
  ENDIF
  !
  IF (nPole0+nPole1+nPole2 .EQ. 0) THEN
     WRITE(6,'(a)') 'No dispersive poles detected'
     WRITE(6,'(a)') 'Program will now exit'
     STOP
  ENDIF
  !----------------------------------------------------------------------
  !---- Allocate and initialize the material data structure
  !----------------------------------------------------------------------
  ALLOCATE(Mat(0:1), MatDisp(1))
  ALLOCATE(MatDisp(1)%pole0(nPole0))
  ALLOCATE(MatDisp(1)%pole1(nPole1))
  ALLOCATE(MatDisp(1)%pole2(nPole2))
  !
  CALL INIT_MAT_STRUCT(Mat(0))  
  CALL INIT_MAT_STRUCT(Mat(1))
  CALL INIT_MATDISP_STRUCT(MatDisp(1))
  !
  !---- Correspondance entre parties dispersive et non-dispersive
  Mat(1)%iDispMat = 1
  MatDisp(1)%iMat = 1
  !----------------------------------------------------------------------
  !---- Read data file
  !----------------------------------------------------------------------
  CALL READ_EXPERIMENTAL_DISPERSION_DATA(dataFile, omega, epsR, epsI)
  !----------------------------------------------------------------------
  !---- Process simulated annealing
  !----------------------------------------------------------------------
  CALL SIMULATED_ANNEALING(omega, epsR, epsI, nPole0, nPole1, nPole2, sol)
  !----------------------------------------------------------------------
  !---- Fill Mat and MatDisp with fit results
  !----------------------------------------------------------------------
  iSol                             = 1
  Mat(1)%eps(:)                    = sol(iSol)
  iSol                             = iSol + 1
  !
  !---- Zero order poles
  DO iPole=1,nPole0
     MatDisp(1)%pole0(iPole)%sigma = sol(iSol)
     iSol                          = iSol + 1
  ENDDO
  !
  !---- First order poles
  DO iPole=1,nPole1
     MatDisp(1)%pole1(iPole)%a     = sol(iSol)
     MatDisp(1)%pole1(iPole)%b     = sol(iSol+1)
     iSol                          = iSol + 2
  ENDDO
  !
  !---- Second order poles
  DO iPole=1,nPole2
     MatDisp(1)%pole2(iPole)%c     = sol(iSol)
     MatDisp(1)%pole2(iPole)%d     = sol(iSol+1)
     MatDisp(1)%pole2(iPole)%e     = sol(iSol+2)
     MatDisp(1)%pole2(iPole)%f     = sol(iSol+3)
     iSol                          = iSol + 4
  ENDDO
  !----------------------------------------------------------------------
  !---- Output for dataDiogenes
  !----------------------------------------------------------------------
  CALL GET_NEW_FILE_UNIT(nUnit)  
  OPEN(UNIT=nUnit, FILE=TRIM(dataFile)//'_model')
  !
  WRITE(nUnit,'(i1,a,i1)') nPole1,' ',nPole2
  IF (nPole0 == 0) THEN
     WRITE(nUnit,'(2(e15.7))') Mat(1)%eps(1), 0.0
  ELSE
     WRITE(nUnit,'(2(e15.7))') Mat(1)%eps(1), MatDisp(1)%pole0(1)%sigma*1E15 ! Convert to Hz
  ENDIF
  DO iPole=1,nPole2
     c = MatDisp(1)%pole2(iPole)%c
     d = MatDisp(1)%pole2(iPole)%d
     e = MatDisp(1)%pole2(iPole)%e
     f = MatDisp(1)%pole2(iPole)%f
     WRITE(nUnit,'(4(e15.7))') c*1E30, d*1E15, e*1E30, f*1E15 ! Convert from PHz to Hz
  ENDDO
  CLOSE(nUnit)
  !----------------------------------------------------------------------
  !---- Compute quality of the fit
  !----------------------------------------------------------------------
  CALL QUALITY_DISPERSION_FIT(omega, epsR, epsI, errorR, errorI)
  !
  WRITE(6,'(a49)')       '!------------------------------------------------'
  WRITE(6,'(a20)')       'Dispersion model   :'
  WRITE(6,'(a20,i7)')    '  Zero-order poles :', nPole0
  WRITE(6,'(a20,i7)')    ' First-order poles :', nPole1
  WRITE(6,'(a20,i7)')    'Second-order poles :', nPole2
  WRITE(6,'(a20)')       'Fit quality        :'
  WRITE(6,'(a20,e15.7)') '         Real part :', errorR
  WRITE(6,'(a20,e15.7)') '    Imaginary part :', errorI
  WRITE(6,'(a49)')       '!------------------------------------------------'
  !----------------------------------------------------------------------
  !---- Write output files
  !----------------------------------------------------------------------
  outputFile = TRIM(dataFile)//'_samples'
  CALL WRITE_PERMITTIVITY(outputFile, nUnit, 1, omega(1), omega(SIZE(omega)), nSamples)
  !----------------------------------------------------------------------
END PROGRAM FIT_DISPERSION_DATA
