MODULE IO_FIELDS

  !====================================================================
  ! All the routines related to writing the fields for visualization
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Array containing the order of each cell, for Raw mesh format output
  !--------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: orderArray
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_EM_FIELDS
    !------------------------------------------------------------------
    !---- Write the 3D solution
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE NUMERICAL_TYPES
    USE RAW_MESH
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write mesh and solution
    !------------------------------------------------------------------
    n3dView = n3dView + 1
    !
    CALL BUILD_ORDER_ARRAY
    IF(n3dView == 1) CALL WRITE_RAW_MESH(orderArray)
    CALL WRITE_RAW_SOLUTION
    !------------------------------------------------------------------
    DEALLOCATE(orderArray)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_EM_FIELDS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_ORDER_ARRAY
    !------------------------------------------------------------------
    !---- Allocate and build the order array for Raw mesh format output
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, order
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write mesh
    !------------------------------------------------------------------
    ALLOCATE(orderArray(Mesh%nCell))
    orderArray(:) = 0
    !
    DO iCell=1,Mesh%nCell
       IF (ALLOCATED(dgEM))  CALL HACK_P0_VIEW_DG ( dgEM(iCell))
       IF (ALLOCATED(hdgEM)) CALL HACK_P0_VIEW_HDG(hdgEM(iCell))
       !
       IF (ALLOCATED(dgEM))  orderArray(iCell) =  dgEM(iCell)%order
       IF (ALLOCATED(hdgEM)) orderArray(iCell) = hdgEM(iCell)%order
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_ORDER_ARRAY

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE HACK_P0_VIEW_DG(numCell)
    !----------------------------------------------------------------------
    !---- Hack the P0 configuration for DG cell
    !
    !---- IN
    !---- numCell : numerical element
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE MESH_CELLS
    USE IO_FILES
    USE MESH_ELEMENTS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    TYPE(numerical_element_dg), INTENT(INOUT) :: numCell
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3)              :: Eloc, Hloc
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    IF (numCell%order /= 0) RETURN
    !----------------------------------------------------------------------
    !---- Hack P0
    !----------------------------------------------------------------------
    Eloc(:) = numCell%E%vec(1,:)
    Hloc(:) = numCell%H%vec(1,:)
    !
    DEALLOCATE(numCell%E%vec, numCell%H%vec)
    ALLOCATE(numCell%E%vec(4,3))
    ALLOCATE(numCell%H%vec(4,3))
    !
    numCell%E%vec(1,:) = Eloc(:)
    numCell%E%vec(2,:) = Eloc(:)
    numCell%E%vec(3,:) = Eloc(:)
    numCell%E%vec(4,:) = Eloc(:)
    numCell%H%vec(1,:) = Hloc(:)
    numCell%H%vec(2,:) = Hloc(:)
    numCell%H%vec(3,:) = Hloc(:)
    numCell%H%vec(4,:) = Hloc(:)
    !
    numCell%order      = 1
    numCell%nDofV      = 4
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE HACK_P0_VIEW_DG

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE HACK_P0_VIEW_HDG(numCell)
    !----------------------------------------------------------------------
    !---- Hack the P0 configuration for HDG cell
    !
    !---- IN
    !---- numCell : HDG numerical element
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE MESH_CELLS
    USE IO_FILES
    USE MESH_ELEMENTS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    TYPE(numerical_element_hdg), INTENT(INOUT) :: numCell
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3)    :: Eloc, Hloc
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    IF (numCell%order /= 0) RETURN
    !----------------------------------------------------------------------
    !---- Hack P0
    !----------------------------------------------------------------------
    Eloc(:) = numCell%E%vec(1,:)
    Hloc(:) = numCell%H%vec(1,:)
    !
    DEALLOCATE(numCell%E%vec, numCell%H%vec)
    ALLOCATE(numCell%E%vec(4,3))
    ALLOCATE(numCell%H%vec(4,3))
    !
    numCell%E%vec(1,:) = Eloc(:)
    numCell%E%vec(2,:) = Eloc(:)
    numCell%E%vec(3,:) = Eloc(:)
    numCell%E%vec(4,:) = Eloc(:)
    numCell%H%vec(1,:) = Hloc(:)
    numCell%H%vec(2,:) = Hloc(:)
    numCell%H%vec(3,:) = Hloc(:)
    numCell%H%vec(4,:) = Hloc(:)
    !
    numCell%order      = 1
    numCell%nDofV      = 4
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE HACK_P0_VIEW_HDG

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE IO_FIELDS
