MODULE IO_FILES

  !====================================================================
  ! Module containing general I/O subroutines for files and arrays
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------
  
  !--------------------------------------------------------------------
  ! Annoying character stuff
  !--------------------------------------------------------------------
  CHARACTER(KIND=C_CHAR), PARAMETER :: TABCHAR = ACHAR(9)
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_SIZE_OF_FILE(datafile, n_rows, n_cols)
    !------------------------------------------------------------------
    !---- Determine the number of rows and columns in a file
    !
    !---- IN
    !---- datafile : file containing data
    !
    !---- OUT
    !---- n_rows   : number of rows in datafile
    !---- n_cols   : number of cols in datafile
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: datafile
    !
    INTEGER(C_INT),                INTENT(OUT) :: n_rows, n_cols
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                            :: file_exists

    INTEGER(C_INT), PARAMETER                  :: maxLength = 1000000
    INTEGER(C_INT)                             :: nUnit, error, i, j

    CHARACTER(LEN=maxLength, KIND=C_CHAR)      :: string

    REAL(C_DOUBLE)                             :: value
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE  :: line
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    INQUIRE(FILE=TRIM(datafile), EXIST=file_exists)
    IF (.NOT. file_exists) THEN
       WRITE(6, '(a)') 'Error in routine GET_SIZE_OF_FILE'
       WRITE(6, '(a)') 'File ', TRIM(datafile), ' not found'
    ENDIF
    !------------------------------------------------------------------
    !---- Get size
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    !
    !---- Initialize data and open file
    n_rows = 0
    n_cols = 0
    !
    OPEN(UNIT=nUnit, FILE=TRIM(datafile))
    !
    !---- Determine number of columns
    READ(nUnit, '(a)') string
    CALL GET_NUMBER_OF_DOUBLE_IN_STRING(string, n_cols)
    !
    REWIND(nUnit)
    !
    !---- Determine number of rows
    ALLOCATE(line(n_cols))
    !
    DO
       READ(nUnit, *, IOSTAT=error) line(:)
       IF (error /= 0) EXIT
       n_rows = n_rows + 1
    ENDDO
    !
    CLOSE(nUnit)
    DEALLOCATE(line)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_SIZE_OF_FILE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHAR_TO_INT(char, int)
    !------------------------------------------------------------------
    !---- Conversion of a CHAR to an INT
    !
    !---- IN
    !---- char : character to convert
    !
    !---- OUT			
    !---- int  : converted integer
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: char

    INTEGER(C_INT),                INTENT(OUT) :: int
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                             :: error
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (INDEX(char,'.') > 0) THEN
       WRITE(6, '(a)') 'Error in routine CHAR_TO_INT'
       WRITE(6, '(a)') 'Input character seems not to be integer-like'
    ENDIF
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    READ(char, *, IOSTAT=error) int
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !---- Tried to put something that is not an int in an int
    IF (error == 5010) int = 0
    !------------------------------------------------------------------
  END SUBROUTINE CHAR_TO_INT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INT_TO_CHAR(int, char)
    !------------------------------------------------------------------
    !---- Conversion of an INT to a CHAR
    !
    !---- IN
    !---- int   : integer to convert
    !
    !---- OUT			
    !---- char  : converted character
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                INTENT(IN)  :: int

    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(OUT) :: char
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    WRITE(char, '(i0)') int

    char = TRIM(char)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INT_TO_CHAR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHAR_TO_DOUBLE(char, dble)
    !------------------------------------------------------------------
    !---- Conversion of a CHAR to a DOUBLE
    !
    !---- IN
    !---- char  : character to convert
    !
    !---- OUT			
    !---- dble  : converted double
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: char
    !
    REAL(C_DOUBLE), INTENT(OUT) :: dble
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    READ(char, *) dble
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHAR_TO_DOUBLE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE REMOVE_EXTENSION(readname, extension)
    !------------------------------------------------------------------
    !---- Removes the extension in a file name
    !
    !---- IN
    !---- readname  : input file name
    !
    !---- OUT
    !---- readname  : modified in the process
    !---- extension : extension of the file
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readname, extension
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                   :: size
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check for the existence of a dot in the name, and split
    !---- into name and extension
    !------------------------------------------------------------------
    size = SCAN(ADJUSTL(readname), '.', .TRUE.)
    IF (size == 0) THEN
       WRITE(6,'(a)') 'Error in REMOVE_EXTENSION'
       WRITE(6,'(a)') 'The file name you provided does not contain any extension'
       CALL TILT
    ENDIF
    !
    extension = readname(size:)
    readname  = readname(1:size-1)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REMOVE_EXTENSION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE REWIND_FILE_AND_FIND(fileUnit, tag, iRead)
    !------------------------------------------------------------------
    !---- Find appropriate tag in file
    !
    !---- IN
    !---- fileUnit : file unit for reading
    !---- tag      : tag to find
    !
    !---- OUT
    !---- iRead    : IO status at the end of reading
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR), INTENT(IN)  :: tag

    INTEGER(C_INT),                  INTENT(IN)  :: fileUnit
    INTEGER(C_INT),                  INTENT(OUT) :: iRead
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR)              :: readChain
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read and find
    !------------------------------------------------------------------
    readChain = ""
    REWIND(fileUnit)
    !
    DO WHILE (INDEX(readChain,TRIM(tag))==0)
       READ(fileUnit,'(a)',IOSTAT=iRead) readChain
       IF (iRead < 0) EXIT
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REWIND_FILE_AND_FIND

  !====================================================================
  !====================================================================
  !====================================================================
 
  SUBROUTINE GET_NEW_FILE_UNIT(fileUnit)
    !------------------------------------------------------------------
    !---- Provide a new, unused file unit
    !
    !---- IN
    !---- fileUnit : file unit
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: fileUnit
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT),        SAVE :: nUnit = 10
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Provide new value
    !------------------------------------------------------------------
    nUnit    = nUnit + 1
    fileUnit = nUnit
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_NEW_FILE_UNIT

  !====================================================================
  !====================================================================
  !====================================================================
 
  SUBROUTINE DOES_FILE_EXIST(fileName, fileExists)
    !------------------------------------------------------------------
    !---- Check that file exists
    !
    !---- IN
    !---- fileName   : file name
    !
    !---- OUT
    !---- fileExists : TRUE if file was found
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: fileName

    LOGICAL(C_BOOL),               INTENT(OUT) :: fileExists
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                             :: fileSize
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check that file exists
    !------------------------------------------------------------------
    fileExists = .FALSE.
    INQUIRE(FILE=TRIM(fileName), SIZE=fileSize, EXIST=fileExists)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DOES_FILE_EXIST

  !====================================================================
  !====================================================================
  !====================================================================
 
  SUBROUTINE GET_FILE_SIZE_IN_BYTES(fileName, fileSize)
    !------------------------------------------------------------------
    !---- Check that file exists
    !
    !---- IN
    !---- fileName : file name
    !
    !---- OUT
    !---- fileSize : size of file in bytes
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: fileName

    INTEGER(C_INT),                INTENT(OUT) :: fileSize
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check file size
    !------------------------------------------------------------------
    fileSize = 0
    INQUIRE(FILE=TRIM(fileName), SIZE=fileSize)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_FILE_SIZE_IN_BYTES

  !====================================================================
  !====================================================================
  !====================================================================
 
  SUBROUTINE REMOVE_FILE(fileName)
    !------------------------------------------------------------------
    !---- Check that file exists
    !
    !---- IN
    !---- fileName   : file name
    !
    !---- OUT
    !---- fileExists : TRUE if file was found
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: fileName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR)          :: rmCommand
    
    LOGICAL(C_BOOL)                           :: fileExists    
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Remove file if it exists
    !------------------------------------------------------------------
    CALL DOES_FILE_EXIST(fileName, fileExists)
    IF (fileExists) THEN
       rmCommand = 'rm '//TRIM(fileName)//';'
       CALL SYSTEM(TRIM(rmCommand))
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REMOVE_FILE

  !====================================================================
  !====================================================================
  !====================================================================
 
  SUBROUTINE BUILD_SUBMESH_NAME(readName, iDomain, subName)
    !------------------------------------------------------------------
    !---- Build a submesh name
    !
    !---- IN
    !---- readName : input mesh file name
    !---- iDomain  : index of the subdomain
    !
    !---- OUT
    !---- subName  : subMesh name
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName, subName

    INTEGER(C_INT)                   :: iDomain
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: path, name, cDomain
    
    INTEGER(C_INT)                   :: length
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Modify name by checking for path symbols "/"
    !------------------------------------------------------------------
    CALL INT_TO_CHAR(iDomain,cDomain)
    length  = SCAN(ADJUSTL(readName), '/', .TRUE.)
    path    = readName(1:length)
    name    = readName(length+1:LEN(readName))
    subName = TRIM(path)//'subMesh_'//TRIM(name)//'_'//TRIM(cDomain)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_SUBMESH_NAME

  !====================================================================
  !====================================================================
  !====================================================================
 
  SUBROUTINE BUILD_SUBINFO_NAME(readName, iDomain, subName)
    !------------------------------------------------------------------
    !---- Build a subinfo name
    !
    !---- IN
    !---- readName : input mesh file name
    !---- iDomain  : index of the subdomain
    !
    !---- OUT
    !---- subName  : subInfo name
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName, subName

    INTEGER(C_INT)                   :: iDomain
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: path, name, cDomain
    
    INTEGER(C_INT)                   :: length
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Modify name by checking for path symbols "/"
    !------------------------------------------------------------------
    CALL INT_TO_CHAR(iDomain,cDomain)
    length  = SCAN(ADJUSTL(readName), '/', .TRUE.)
    path    = readName(1:length)
    name    = readName(length+1:LEN(readName))
    subName = TRIM(path)//'subInfo_'//TRIM(name)//'_'//TRIM(cDomain)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_SUBINFO_NAME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COULD_NOT_FIND_TAG(tag)
    !------------------------------------------------------------------
    !---- Print error message when tag could not be found in config file
    !
    !---- IN
    !---- tag : tag not found
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR), INTENT(IN) :: tag
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print error message
    !------------------------------------------------------------------
    WRITE(6, '(a)') 'I could not find the '//TRIM(tag)//' tag'
    CALL TILT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COULD_NOT_FIND_TAG

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_NUMBER_OF_DOUBLE_IN_STRING(string, nDouble)
    !------------------------------------------------------------------
    !---- Determine the # of double in a string of double read
    !---- from an array of unknown size
    !
    !---- IN
    !---- string  : string to read
    !
    !---- OUT
    !---- nDouble : # of double in the string
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: string

    INTEGER(C_INT),                INTENT(OUT) :: nDouble
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                             :: strLength, iDble, &
         jDble, error

    REAL(C_DOUBLE)                             :: valDble
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine the # of double in string
    !------------------------------------------------------------------
    nDouble   = 0
    strLength = LEN_TRIM(string) + 1
    !
    DO iDble=1,strLength
       READ(string, *, IOSTAT=error) (valDble, jDble=1,iDble)
       !
       IF (error /= 0) THEN
          nDouble = iDble - 1
          RETURN
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_NUMBER_OF_DOUBLE_IN_STRING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_NUMBER_OF_INT_IN_STRING(string, nInt)
    !------------------------------------------------------------------
    !---- Determine the # of int in a string of int read
    !---- from an array of unknown size
    !
    !---- IN
    !---- string : string to read
    !
    !---- OUT
    !---- nInt   : # of int in the string
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: string

    INTEGER(C_INT),                INTENT(OUT) :: nInt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                             :: strLength, valInt, &
         iInt, jInt, error
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine the # of int in string
    !------------------------------------------------------------------
    nInt      = 0
    strLength = LEN_TRIM(string)
    !
    DO iInt=1,strLength
       READ(string, *, IOSTAT=error) (valInt, jInt=1,iInt)
       !
       IF (error /= 0) THEN
          nInt = iInt - 1
          RETURN
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_NUMBER_OF_INT_IN_STRING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_AN_INT(char, isAnInt)
    !------------------------------------------------------------------
    !---- Test if what is in char fits in an int
    !
    !---- IN
    !---- char    : character to convert
    !
    !---- OUT			
    !---- isAnInt : TRUE if what is in char is an int
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: char

    LOGICAL(C_BOOL),               INTENT(OUT) :: isAnInt
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                             :: int, error
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    isAnInt = .TRUE.
    READ(char, *, IOSTAT=error) int
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (error == 5010) isAnInt = .FALSE.
    !------------------------------------------------------------------
  END SUBROUTINE IS_AN_INT

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE IO_FILES
