MODULE IO_TESTS

  !====================================================================
  ! Module containing I/O subroutines for tests
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Test tolerances
  !--------------------------------------------------------------------
  REAL(C_DOUBLE) :: seqTol = 0.05
  REAL(C_DOUBLE) :: parTol = 0.00001
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_TEST_FILE(fileName, testName, testNumber)
    !------------------------------------------------------------------
    !---- Check that file exists
    !
    !---- IN
    !---- fileName   : file name
    !---- testName   : test name
    !---- testNumber : test number
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(INOUT) :: fileName
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(INOUT) :: testName
    
    INTEGER(C_INT),                   INTENT(INOUT) :: testNumber
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                 :: fileExists

    INTEGER(C_INT)                                  :: fileSize
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check file
    !------------------------------------------------------------------
    CALL DOES_FILE_EXIST(fileName, fileExists)
    IF (.NOT. fileExists) THEN
       WRITE(6,'(3a)') 'File ',TRIM(fileName),' was not found'
       CALL PRINT_TEST_ERROR_MESSAGE(testName, testNumber)
    ENDIF       
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_TEST_FILE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_TEST_ERROR_MESSAGE(testName, testNumber)
    !------------------------------------------------------------------
    !---- Print error message
    !
    !---- IN
    !---- testName   : test name
    !---- testNumber : test number
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(IN) :: testName
    
    INTEGER(C_INT),                   INTENT(IN) :: testNumber
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print error message
    !------------------------------------------------------------------
    WRITE(6,'(3a)')   'Test for program ',TRIM(testName),' failed'
    WRITE(6,'(a,i0)') 'Failure occured for test #',testNumber
    CALL EXIT(1)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_TEST_ERROR_MESSAGE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_DIOGENES_ERROR_MESSAGE(caseName, testNumber, refError, maxError)
    !------------------------------------------------------------------
    !---- Print diogenes error message
    !
    !---- IN
    !---- caseName   : case name
    !---- testNumber : test number
    !---- refError   : reference error
    !---- maxError   : maximal computed error
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(IN) :: caseName
    
    INTEGER(C_INT),                   INTENT(IN) :: testNumber

    REAL(C_DOUBLE), OPTIONAL,         INTENT(IN) :: refError, maxError
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print error message
    !------------------------------------------------------------------
    WRITE(6,'(3a)')         'Test for case ',TRIM(caseName),' failed'
    !
    IF ((PRESENT(refError)) .AND. (PRESENT(maxError))) THEN
       WRITE(6,'(a,e12.5)') 'Reference error is :', refError
       WRITE(6,'(a,e12.5)') 'Computed  error is :', maxError
    ENDIF
    !
    WRITE(6,'(a,i0)')       'Failure occured for test #',testNumber
    CALL EXIT(1)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_DIOGENES_ERROR_MESSAGE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CLEAN_DIOGENES_TEST
    !------------------------------------------------------------------
    !---- Clean repository after tests
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_APPS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Clean repository
    !------------------------------------------------------------------
    CALL CLEAN_DIOGENES_SUB
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CLEAN_DIOGENES_TEST

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MOVE_TO_REPO(repo)
    !------------------------------------------------------------------
    !---- Move to repository and copy apps
    !
    !---- IN
    !---- repo : repository path
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(IN) :: repo
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Move to repo
    !------------------------------------------------------------------
    CALL CHDIR(TRIM(repo))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MOVE_TO_REPO

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_MAX_ERROR(errorFile, maxError)
    !------------------------------------------------------------------
    !---- Get maximal error from error file
    !
    !---- IN
    !---- errorFile : name of file to read
    !
    !---- OUT
    !---- maxError  : maximal error level
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(IN)  :: errorFile

    REAL(C_DOUBLE),                   INTENT(OUT) :: maxError
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                                :: nUnit, iRead

    REAL(C_DOUBLE)                                :: error, time
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get maximal error
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    !
    OPEN(UNIT=nUnit, FILE=TRIM(errorFile))
    maxError = 0.0d0
    DO 
       READ(nUnit,*,IOSTAT=iRead) time, error
       maxError = MAX(maxError, error)
       IF (iRead < 0) EXIT
    ENDDO
    CLOSE(nUnit)  
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_MAX_ERROR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_ERROR_FROM_FILES(compFile, compCol, refFile, refCol, errFile)
    !------------------------------------------------------------------
    !---- Read two files, compute error and write it in third file
    !
    !---- IN
    !---- compFile : file with computed data
    !---- compCol  : column for computed data in related file
    !---- refFile  : file with reference data
    !---- refCol   : column for reference data in related file    
    !---- errFile  : file containing the computed error    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR), INTENT(IN) :: compFile, refFile, errFile

    INTEGER(C_INT),                   INTENT(IN) :: compCol, refCol
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                               :: nUnitComp, nUnitRef, &
         nUnitErr, iRead

    REAL(C_DOUBLE)                               :: error
    REAL(C_DOUBLE), DIMENSION(compCol)           :: compData
    REAL(C_DOUBLE), DIMENSION(refCol)            :: refData
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get maximal error
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnitComp)
    CALL GET_NEW_FILE_UNIT(nUnitRef)
    CALL GET_NEW_FILE_UNIT(nUnitErr)
    !
    OPEN(UNIT=nUnitComp, FILE=TRIM(compFile))
    OPEN(UNIT=nUnitRef,  FILE=TRIM(refFile ))
    OPEN(UNIT=nUnitErr,  FILE=TRIM(errFile ))
    !
    DO 
       READ(nUnitComp,*,IOSTAT=iRead) compData(:)
       READ(nUnitRef, *,IOSTAT=iRead) refData(:)
       !
       error = ABS(refData(refCol) - compData(compCol))
       WRITE(nUnitErr,*) compData(1), error
       !
       IF (iRead < 0) EXIT
    ENDDO
    !
    CLOSE(nUnitComp)
    CLOSE(nUnitRef)
    CLOSE(nUnitErr)  
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_ERROR_FROM_FILES
  
  !====================================================================
  !====================================================================
  !====================================================================

END MODULE IO_TESTS
