MODULE ARRAY_OPERATIONS

  !======================================================================
  ! Module containing the subroutines related to array operations
  !======================================================================

  !----------------------------------------------------------------------
  ! Shared modules
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  !----------------------------------------------------------------------
  IMPLICIT NONE
  !----------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Interface for overloading array duplicates finding
  !----------------------------------------------------------------------
  INTERFACE FIND_ARRAY_DUPLICATES
     MODULE PROCEDURE FIND_ARRAY_DUPLICATES_INT, FIND_ARRAY_DUPLICATES_DOUBLE
  END INTERFACE FIND_ARRAY_DUPLICATES
  !----------------------------------------------------------------------

CONTAINS

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE FIND_ARRAY_DUPLICATES_INT(array, nRows, sorted, nSorted)
    !--------------------------------------------------------------------
    !---- Finds duplicate rows in an array of integers
    !---- Warnings: - the output list "sorted" is an indirection list
    !               - the array is modified in the process
    !
    !---- IN
    !---- array   : Original array
    !---- nRows   : Original number of rows in array
    !
    !---- OUT
    !---- array   : Array is modified in the process
    !---- sorted  : Indirection list of size nRows
    !----           Use as sorted(oldIndex) = newIndex
    !---- nSorted : Number of unique elements
    !-------------------------------------------------------------------
    !---- Modules
    !-------------------------------------------------------------------
    USE SORT
    !-------------------------------------------------------------------
    !---- I/O variables
    !-------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)    :: nRows
    INTEGER(C_INT),               INTENT(OUT)   :: nSorted
    INTEGER(C_INT), DIMENSION(:), INTENT(OUT)   :: sorted
    INTEGER(C_INT), DIMENSION(:), INTENT(INOUT) :: array
    !-------------------------------------------------------------------
    !---- Local variables
    !-------------------------------------------------------------------
    INTEGER(C_INT)  :: i, j, k, iRow, duplicate, oldIndex

    LOGICAL(C_BOOL) :: isduplicate, keepLooking, isEqual

    INTEGER(C_INT)  :: locArray(2,nRows)
    !-------------------------------------------------------------------
    !---- Testing for bad input
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
    !---- Initializations
    !-------------------------------------------------------------------
    nSorted   = 0
    sorted(:) = 0
    !
    DO iRow=1,nRows
       locArray(1,iRow) = array(iRow)
       locArray(2,iRow) = iRow
    ENDDO
    !
    array(:) = 0.0d0
    !-------------------------------------------------------------------
    !---- Find duplicates and create indirection
    !-------------------------------------------------------------------
    !---- Lexicographic sort of array
    CALL SORT_ARRAY(2,nRows,locArray)
    !
    iRow    = 1
    nSorted = 0
    DO WHILE(iRow <= nRows)
       !---- Initialize loop variables
       keepLooking      = .TRUE.
       nSorted          = nSorted + 1
       array(nSorted)   = locArray(1,iRow)
       oldIndex         = locArray(2,iRow)
       sorted(oldIndex) = nSorted
       duplicate        = 1
       !
       !---- Exit if that was the last element (no possible duplicates)
       IF (iRow == nRows) EXIT
       !
       !---- Loop over rows after iRow until we no longer find duplicates
       DO WHILE (keepLooking)
          !
          !---- If next explored row is last row, we must exit after this iteration
          IF (iRow + duplicate == nRows) keepLooking = .FALSE.
          !
          !---- Check if next explored row is indeed a duplicate
          isEqual = (locArray(1,iRow) == locArray(1,iRow+duplicate))
          !
          IF (isEqual) THEN ! It is a duplicate, fill indirection list
             oldIndex         = locArray(2,iRow+duplicate)
             sorted(oldIndex) = nSorted
             duplicate        = duplicate + 1
          ELSE              ! It is not a duplicate, don't look further
             keepLooking      = .FALSE.
          ENDIF
       ENDDO
       !
       !---- Progress until next unexplored row
       iRow = iRow + duplicate
    ENDDO
    !-------------------------------------------------------------------
    !---- Testing for errors
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
  END SUBROUTINE FIND_ARRAY_DUPLICATES_INT

  !=====================================================================
  !=====================================================================
  !=====================================================================

  SUBROUTINE FIND_ARRAY_DUPLICATES_DOUBLE(array, nRows, nCols, tol, sorted, nSorted)
    !-------------------------------------------------------------------
    !---- Finds duplicate rows in an array of double
    !---- Warnings: - the first coordinate of array is num_col
    !               - the output list "sorted" is an indirection list
    !               - the array is modified in the process
    !
    !---- IN
    !---- array   : Original array
    !---- nRows   : Original number of rows in array
    !---- nCols   : Number of columns in array
    !
    !---- OUT
    !---- array   : Array is modified in the process
    !---- sorted  : Indirection list of size nRows
    !----           Use as sorted(oldIndex) = newIndex
    !---- nSorted : Number of unique elements
    !-------------------------------------------------------------------
    !---- Modules
    !-------------------------------------------------------------------
    USE SORT
    !-------------------------------------------------------------------
    !---- I/O variables
    !-------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(IN)    :: nRows, nCols
    INTEGER(C_INT),                 INTENT(OUT)   :: nSorted
    INTEGER(C_INT), DIMENSION(:),   INTENT(OUT)   :: sorted

    REAL(C_DOUBLE),                 INTENT(IN)    :: tol
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(INOUT) :: array
    !-------------------------------------------------------------------
    !---- Local variables
    !-------------------------------------------------------------------
    INTEGER(C_INT)  :: i, j, k, iRow, duplicate, oldIndex

    LOGICAL(C_BOOL) :: isduplicate, keepLooking, isEqual

    REAL(C_DOUBLE)  :: locArray(nCols+1,nRows)
    !-------------------------------------------------------------------
    !---- Testing for bad input
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
    !---- Initializations
    !-------------------------------------------------------------------
    nSorted   = 0
    sorted(:) = 0
    !
    DO iRow=1,nRows
       locArray(1:nCols,iRow) = array(1:nCols,iRow)
       locArray(nCols+1,iRow) = DBLE(iRow)
    ENDDO
    !
    array(:,:) = 0.0d0
    !-------------------------------------------------------------------
    !---- Find duplicates and create indirection
    !-------------------------------------------------------------------
    !---- Lexicographic sort of array
    CALL SORT_ARRAY(nCols+1,nRows,locArray)
    !
    iRow    = 1
    nSorted = 0
    DO WHILE(iRow <= nRows)
       !---- Initialize loop variables
       keepLooking            = .TRUE.
       nSorted                = nSorted + 1
       array(1:nCols,nSorted) = locArray(1:nCols,iRow)
       oldIndex               = FLOOR(locArray(nCols+1,iRow))
       sorted(oldIndex)       = nSorted
       duplicate              = 1
       !
       !---- Exit if that was the last element (no possible duplicates)
       IF (iRow == nRows) EXIT
       !
       !---- Loop over rows after iRow until we no longer find duplicates
       DO WHILE (keepLooking)
          !
          !---- If next explored row is last row, we must exit after this iteration
          IF (iRow + duplicate == nRows) keepLooking = .FALSE.
          !
          !---- Check if next explored row is indeed a duplicate
          isEqual = ALL(ABS(locArray(1:nCols,iRow)-locArray(1:nCols,iRow+duplicate)) < tol)
          !
          IF (isEqual) THEN ! It is a duplicate, fill indirection list
             oldIndex         = FLOOR(locArray(nCols+1,iRow+duplicate))
             sorted(oldIndex) = nSorted
             duplicate        = duplicate + 1
          ELSE              ! It is not a duplicate, don't look further
             keepLooking      = .FALSE.
          ENDIF
       ENDDO
       !
       !---- Progress until next unexplored row
       iRow = iRow + duplicate
    ENDDO
    !-------------------------------------------------------------------
    !---- Testing for errors
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
  END SUBROUTINE FIND_ARRAY_DUPLICATES_DOUBLE

  !=====================================================================
  !=====================================================================
  !=====================================================================

  SUBROUTINE FIND_INT_IN_ARRAY(array, size, val, isFound)
    !-------------------------------------------------------------------
    !---- Try to find an int in an array
    !
    !---- IN
    !---- array   : original array
    !---- size    : size of array
    !---- val     : value to look for
    !
    !---- OUT
    !---- isFound : TRUE if val was found
    !-------------------------------------------------------------------
    !---- Modules
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
    !---- I/O variables
    !-------------------------------------------------------------------
    LOGICAL(C_BOOL),              INTENT(OUT) :: isFound

    INTEGER(C_INT),               INTENT(IN)  :: size, val
    INTEGER(C_INT), DIMENSION(:), INTENT(IN)  :: array
    !-------------------------------------------------------------------
    !---- Local variables
    !-------------------------------------------------------------------
    INTEGER(C_INT)                            :: iVal
    !-------------------------------------------------------------------
    !---- Testing for bad input
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
    !---- Find val in array
    !-------------------------------------------------------------------
    isFound = .FALSE.
    !
    !---- Look for value
    DO iVal=1,size
       IF (val == array(iVal)) THEN
          isFound = .TRUE.
          RETURN
       ENDIF
    ENDDO
    !-------------------------------------------------------------------
    !---- Testing for errors
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
  END SUBROUTINE FIND_INT_IN_ARRAY

  !=====================================================================
  !=====================================================================
  !=====================================================================

  SUBROUTINE EXTEND_INTEGER_ARRAY(intArray, intArraySize, intBlockSize)
    !-------------------------------------------------------------------
    !---- Add a new slot to intArray
    !
    !---- IN
    !---- intArray     : array of integers
    !---- intArraySize : number of elements in intArray
    !---- intBlockSize : size of integer block
    !
    !---- OUT
    !---- intArray     : modified in the process
    !---- intArraySize : modified in the process
    !
    !-------------------------------------------------------------------
    !---- Modules
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
    !---- I/O stuff
    !-------------------------------------------------------------------
    INTEGER(C_INT),             INTENT(INOUT) :: intArraySize
    INTEGER(C_INT),             INTENT(IN)    :: intBlockSize
    
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: intArray
    !-------------------------------------------------------------------
    !---- Local stuff
    !-------------------------------------------------------------------
    LOGICAL(C_BOOL)                           :: isArrayFull, isAllocated

    INTEGER(C_INT)                            :: iInt
    
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: tempInt
    !-------------------------------------------------------------------
    !---- Testing for bad input
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
    !---- Check if cell array is full
    !-------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(intArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(intArray) < intArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !-------------------------------------------------------------------
    !---- Extend array if necessary
    !-------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempInt)) DEALLOCATE(tempInt)
          ALLOCATE(tempInt(intArraySize))
          tempInt = intArray
          DEALLOCATE(intArray)
       ENDIF
       !
       ALLOCATE(intArray(intArraySize + intBlockSize))
       intArray(:) = 0
       !
       intArray(1:intArraySize) = tempInt(1:intArraySize)
       IF (ALLOCATED(tempInt)) DEALLOCATE(tempInt)
    ENDIF
    !
    intArraySize = intArraySize + 1
    !-------------------------------------------------------------------
    !---- Testing for errors
    !-------------------------------------------------------------------
    !
    !-------------------------------------------------------------------
  END SUBROUTINE EXTEND_INTEGER_ARRAY

  !=====================================================================
  !=====================================================================
  !=====================================================================

END MODULE ARRAY_OPERATIONS
