MODULE BASES_LAGRANGE_HDG_MATRICES

  !====================================================================
  ! HDG matrices with Lagrange basis
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE BASES_LAGRANGE_ORDER_TO_NDOF
  USE BASES_LAGRANGE_POLYNOMIALS
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE BASES_LAGRANGE_MATRICES_HYBRID_HDG_E(iFace, orderHex, coorHex)
    !------------------------------------------------------------------
    !---- Computation of hybrid surfacic matrix E
    !---- E_ij := Int_face ( phi[i].psi[j) ) ds
    !---- face is a triangular face
    !---- phi[i] in the space Q_pe[hex_e] and psi[j] in Q_pe[hex_e]
    !
    !---- N.B. For HDG method we compute two surfacic matrices, the 
    !          matrix E and the matrix F.
    !          For a complete description of E refer to HORSE 
    !          documentation: '/horse/documents/HDGMinHORSE.pdf' 
    !          in 'https://gitlab.inria.fr:Nachos/horse.git'. 
    !
    !---- IN
    !---- iFace      : index of the hybrid face in 'MeshFacHyb' structure
    !                  (cf mesh_faces.f90 module)
    !---- orderHex   : interpolation order of the attached hexahedron
    !---- coorHex    : coordinates of vertices of the hexahedron
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !-----------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_VERTICES
    USE QUADRATURES_2D
    USE REFERENCE_CELLS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(IN) :: iFace, orderHex

    REAL(C_DOUBLE), DIMENSION(3,8), INTENT(IN) :: coorHex
    !------------------------------------------------------------------
    !---- Local variable
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, i, iPerm, j, jPerm, k, nDofHex, nDofQua, &
         permHex, xPlan, yPlan, zPlan 

    REAL(C_DOUBLE) :: faceRef, h, val
    REAL(C_DOUBLE) :: xHex, yHex, zHex
    REAL(C_DOUBLE) :: xi,   yi,   zi
    REAL(C_DOUBLE) :: xtmp, ytmp, ztmp

    REAL(C_DOUBLE), DIMENSION(3)                :: crossProd2, crossProd3, crossProd4
    REAL(C_DOUBLE), DIMENSION(3,3)              :: coorFace
    REAL(C_DOUBLE), DIMENSION(3)                :: ridge2_1, ridge3_1, ridge4_1
    REAL(C_DOUBLE), DIMENSION(3,8)              :: coorHref
    REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: phi, psi
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: mat
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Some coefficients and initializations
    !------------------------------------------------------------------
    h = 0.0000001d0
    !
    !---- Coordinates of the reference hexahedron
    coorHref(1:3,1) = Href_Vert1
    coorHref(1:3,2) = Href_Vert2
    coorHref(1:3,3) = Href_Vert3
    coorHref(1:3,4) = Href_Vert4
    coorHref(1:3,5) = Href_Vert5
    coorHref(1:3,6) = Href_Vert6
    coorHref(1:3,7) = Href_Vert7
    coorHref(1:3,8) = Href_Vert8
    !
    !---- Coordinates of the hybrid face
    faceNum = MeshFacHyb(iFace)%faceNum
    DO i=1,3
       coorFace(:,i) = MeshVert(MeshFac(faceNum)%vertex(i))%coor(:)
    ENDDO
    !
    !---- Find orientation of the hybrid face
    faceRef = 0
    IF(  (ABS(coorFace(1,1)-coorFace(1,2)) < h)   .AND.                                     &
         (ABS(coorFace(1,1)-coorFace(1,3)) < h) ) THEN
       xPlan   =  1
       faceRef = -1 + 2*(coorFace(1,1)-coorHex(1,1))/(coorHex(1,7)-coorHex(1,1))
    ELSE
       xPlan   =  0
    ENDIF
    IF(  (ABS(coorFace(2,1)-coorFace(2,2)) < h)   .AND.                                     &
         (ABS(coorFace(2,1)-coorFace(2,3)) < h) ) THEN
       yPlan   =  1
       faceRef = -1 + 2*(coorFace(2,1)-coorHex(2,1))/(coorHex(2,7)-coorHex(2,1))
    ELSE
       yPlan   =  0
    ENDIF
    IF(  (ABS(coorFace(3,1)-coorFace(3,2)) < h)   .AND.                                     &
         (ABS(coorFace(3,1)-coorFace(3,3)) < h) ) THEN
       zPlan   =  1
       faceRef = -1 + 2*(coorFace(3,1)-coorHex(3,1))/(coorHex(3,7)-coorHex(3,1))
    ELSE
       zPlan   =  0
    ENDIF
    !
    !---- Compute the number of DoFs
    CALL ORDER_TO_NDOFV_LAGRANGE(orderHex, 'HEX1',nDofHex)
    CALL ORDER_TO_NDOFS_LAGRANGE(orderHex, 'QUA1',nDofQua)
    !------------------------------------------------------------------
    !---- Hybrid surfacic matrix E 
    !------------------------------------------------------------------
    mSurfHyb(iFace)%ndof1 = nDofQua
    ALLOCATE(mSurfHyb(iFace)%tab(nDofQua,nDofQua))
    mSurfHyb(iFace)%tab   = 0.d0
    !
    !---- Gauss points and weights on the reference triangle
    CALL QUADRATURES_DUNAVANT_2D(nDofQua)
    ALLOCATE(mat(nDofHex, nDofHex))
    mat = 0.d0
    DO k=1,nDofHex
       DO j=1,nDofHex
          val = 0.0d0
          DO i=1,nq2d(orderHex)
             xi   = faceRef*xPlan + qx2d(i)*(1-xPlan)
             yi   = faceRef*yPlan + qx2d(i)*xPlan + qy2d(i)*zPlan
             zi   = faceRef*zPlan + qy2d(i)*(1-zPlan)
             !
             xtmp = coorFace(1,1)*xPlan                                                     &
                  + (coorFace(1,1)*(1.0-xi-zi) + coorFace(1,2)*xi + coorFace(1,3)*zi)*yPlan &
                  + (coorFace(1,1)*(1.0-xi-yi) + coorFace(1,2)*xi + coorFace(1,3)*yi)*zPlan
             ytmp = coorFace(2,1)*yPlan                                                     &
                  + (coorFace(2,1)*(1.0-yi-zi) + coorFace(2,2)*yi + coorFace(2,3)*zi)*xPlan &
                  + (coorFace(2,1)*(1.0-xi-yi) + coorFace(2,2)*xi + coorFace(2,3)*yi)*zPlan
             ztmp = coorFace(3,1)*zPlan                                                     &
                  + (coorFace(3,1)*(1.0-yi-zi) + coorFace(3,2)*yi + coorFace(3,3)*zi)*xPlan &
                  + (coorFace(3,1)*(1.0-xi-zi) + coorFace(3,2)*xi + coorFace(3,3)*zi)*yPlan
             !
             xHex = xi*xPlan                                                                &
                  + (-1 + 2*(xtmp-coorHex(1,1))/(coorHex(1,7)-coorHex(1,1)))*(1-xPlan)
             yHex = yi*yPlan                                                                &
                  + (-1 + 2*(ytmp-coorHex(2,1))/(coorHex(2,7)-coorHex(2,1)))*(1-yPlan)
             zHex = zi*zPlan                                                                &
                  + (-1 + 2*(ztmp-coorHex(3,1))/(coorHex(3,7)-coorHex(3,1)))*(1-zPlan)
             !
             CALL BASES_LAGRANGE_POLY_HEX(coorHref, xHex, yHex, zHex, orderHex, phi)
             CALL BASES_LAGRANGE_POLY_HEX(coorHref, xHex, yHex, zHex, orderHex, psi)
             !
             val  = val + phi(k)*psi(j)*qw2d(i)
          ENDDO
          mat(k,j) = val
       ENDDO
    ENDDO
    !
    IF (MeshCell(MeshFac(faceNum)%neighb(1))%cellType == 'TET1') THEN
       permHex = MeshFac(faceNum)%perm(2)
    ELSE
       permHex = MeshFac(faceNum)%perm(1)
    ENDIF
    !
    DO j=1,nDofQua
       jPerm = iPermDofHex(orderHex)%tab(j,permHex)
       DO i=1,nDofQua
          iPerm                    = iPermDofHex(orderHex)%tab(i,permHex)
          mSurfHyb(iFace)%tab(i,j) = mat(iPerm,jPerm)*2.0d0
       ENDDO
    ENDDO
    !
    IF (ALLOCATED(qx2d)) DEALLOCATE(qx2d)
    IF (ALLOCATED(qy2d)) DEALLOCATE(qy2d)
    IF (ALLOCATED(qw2d)) DEALLOCATE(qw2d)
    IF (ALLOCATED(nq2d)) DEALLOCATE(nq2d)
    IF (ALLOCATED(mat))  DEALLOCATE(mat)
    IF (ALLOCATED(phi))  DEALLOCATE(phi)
    IF (ALLOCATED(psi))  DEALLOCATE(psi)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_MATRICES_HYBRID_HDG_E

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE BASES_LAGRANGE_MATRICES_HYBRID_HDG_F(iFace,                &
       orderHex, coorHex, measureTet, coorTet, orderFace)
    !------------------------------------------------------------------
    !---- Computation of hybrid surfacic matrix F
    !---- F_ij := Int_face ( phi[i].psi[j) ) ds
    !---- face is a triangular face
    !---- phi[i] in the space Q_pe[hex_e] and psi[j] in P_pf[face]
    !
    !---- N.B. For HDG method we compute two surfacic matrices, the 
    !          matrix E and the matrix F.
    !          For a complete description of F refer to HORSE 
    !          documentation: '/horse/documents/HDGMinHORSE.pdf' 
    !          in 'https://gitlab.inria.fr:Nachos/horse.git'. 
    !
    !---- IN
    !---- iFace      : index of the hybrid face in 'MeshFacHyb' structure
    !                  (cf mesh_faces.f90 module)
    !---- orderHex   : interpolation order of the attached hexahedron
    !---- coorHex    : coordinates of vertices of the hexahedron
    !---- measureTet : measure of the tetrahedron
    !---- coorTet    : coordinates of vertices of the tetrahedron
    !---- orderFace  : interpolation order of the hybrid face
    !                  (optional, only usefull for HDG method)
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !-----------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_VERTICES
    USE QUADRATURES_2D
    USE REFERENCE_CELLS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(IN) :: iFace, orderHex
    INTEGER(C_INT),                 INTENT(IN) :: orderFace

    REAL(C_DOUBLE),                 INTENT(IN) :: measureTet

    REAL(C_DOUBLE), DIMENSION(3,4), INTENT(IN) :: coorTet
    REAL(C_DOUBLE), DIMENSION(3,8), INTENT(IN) :: coorHex
    !------------------------------------------------------------------
    !---- Local variable
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, i, iPerm, j, jPerm, k,                  &
         nDofFace, nDofHex, nDofTet, nDofQua, permHex, permTet,        &
         xPlan, yPlan, zPlan 

    REAL(C_DOUBLE) :: faceRef, h, val
    REAL(C_DOUBLE) :: xHex, yHex, zHex
    REAL(C_DOUBLE) :: xi,   yi,   zi
    REAL(C_DOUBLE) :: xTet, yTet, zTet
    REAL(C_DOUBLE) :: xtmp, ytmp, ztmp

    REAL(C_DOUBLE), DIMENSION(3)                :: crossProd2, crossProd3, crossProd4
    REAL(C_DOUBLE), DIMENSION(3,3)              :: coorFace
    REAL(C_DOUBLE), DIMENSION(3)                :: ridge2_1, ridge3_1, ridge4_1
    REAL(C_DOUBLE), DIMENSION(3,8)              :: coorHref
    REAL(C_DOUBLE), DIMENSION(3,4)              :: coorTref
    REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: phi, psi
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: mat
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Some coefficients and initializations
    !------------------------------------------------------------------
    h = 0.0000001d0
    !
    !---- Coordinates of the reference tetrahedon
    coorTref(1:3,1) = Tref_Vert1
    coorTref(1:3,2) = Tref_Vert2
    coorTref(1:3,3) = Tref_Vert3
    coorTref(1:3,4) = Tref_Vert4
    !
    !---- Coordinates of the reference hexahedron
    coorHref(1:3,1) = Href_Vert1
    coorHref(1:3,2) = Href_Vert2
    coorHref(1:3,3) = Href_Vert3
    coorHref(1:3,4) = Href_Vert4
    coorHref(1:3,5) = Href_Vert5
    coorHref(1:3,6) = Href_Vert6
    coorHref(1:3,7) = Href_Vert7
    coorHref(1:3,8) = Href_Vert8
    !
    !---- Coordinates of the hybrid face
    faceNum = MeshFacHyb(iFace)%faceNum
    DO i=1,3
       coorFace(:,i) = MeshVert(MeshFac(faceNum)%vertex(i))%coor(:)
    ENDDO
    !
    !---- Find orientation of the hybrid face
    faceRef = 0
    IF(  (ABS(coorFace(1,1)-coorFace(1,2)) < h)   .AND.                                     &
         (ABS(coorFace(1,1)-coorFace(1,3)) < h) ) THEN
       xPlan   =  1
       faceRef = -1 + 2*(coorFace(1,1)-coorHex(1,1))/(coorHex(1,7)-coorHex(1,1))
    ELSE
       xPlan   =  0
    ENDIF
    IF(  (ABS(coorFace(2,1)-coorFace(2,2)) < h)   .AND.                                     &
         (ABS(coorFace(2,1)-coorFace(2,3)) < h) ) THEN
       yPlan   =  1
       faceRef = -1 + 2*(coorFace(2,1)-coorHex(2,1))/(coorHex(2,7)-coorHex(2,1))
    ELSE
       yPlan   =  0
    ENDIF
    IF(  (ABS(coorFace(3,1)-coorFace(3,2)) < h)   .AND.                                     &
         (ABS(coorFace(3,1)-coorFace(3,3)) < h) ) THEN
       zPlan   =  1
       faceRef = -1 + 2*(coorFace(3,1)-coorHex(3,1))/(coorHex(3,7)-coorHex(3,1))
    ELSE
       zPlan   =  0
    ENDIF
    !
    !---- Build ridge vectors
    ridge2_1(:) = coorTet(:,2) - coorTet(:,1)
    ridge3_1(:) = coorTet(:,3) - coorTet(:,1)
    ridge4_1(:) = coorTet(:,4) - coorTet(:,1)
    !
    !---- Compute cross product
    CALL CROSS_PROD(-ridge4_1,  ridge3_1, crossProd2)
    CALL CROSS_PROD(-ridge4_1,  ridge2_1, crossProd3)
    CALL CROSS_PROD(-ridge3_1,  ridge2_1, crossProd4)
    !
    !---- Compute the number of DoFs
    CALL ORDER_TO_NDOFV_LAGRANGE(orderHex, 'HEX1',nDofHex)
    CALL ORDER_TO_NDOFS_LAGRANGE(orderHex, 'QUA1',nDofQua)
    CALL ORDER_TO_NDOFV_LAGRANGE(orderFace,'TET1',nDofTet)
    CALL ORDER_TO_NDOFS_LAGRANGE(orderFace,'TRI1',nDofFace)
    !------------------------------------------------------------------
    !---- Hybrid surfacic matrix F
    !------------------------------------------------------------------
    mSurfHybF(iFace)%ndof1 = nDofQua
    mSurfHybF(iFace)%ndof2 = nDofFace
    ALLOCATE(mSurfHybF(iFace)%tab(nDofQua,nDofFace))
    mSurfHybF(iFace)%tab   = 0.d0
    !
    !---- Gauss points and weights on the reference triangle
    CALL QUADRATURES_DUNAVANT_2D(nDofQua)
    ALLOCATE(mat(nDofHex,nDofTet))
    mat = 0.d0
    DO k=1,nDofHex
       DO j=1,nDofTet
          val = 0.d0
          DO i=1,nq2d(orderHex)
             xi   = faceRef*xPlan + qx2d(i)*(1-xPlan)
             yi   = faceRef*yPlan + qx2d(i)*xPlan + qy2d(i)*zPlan
             zi   = faceRef*zPlan + qy2d(i)*(1-zPlan)
             !
             xtmp = coorFace(1,1)*xPlan                                                     &
                  + (coorFace(1,1)*(1.0-xi-zi) + coorFace(1,2)*xi + coorFace(1,3)*zi)*yPlan &
                  + (coorFace(1,1)*(1.0-xi-yi) + coorFace(1,2)*xi + coorFace(1,3)*yi)*zPlan
             ytmp = coorFace(2,1)*yPlan                                                     &
                  + (coorFace(2,1)*(1.0-yi-zi) + coorFace(2,2)*yi + coorFace(2,3)*zi)*xPlan &
                  + (coorFace(2,1)*(1.0-xi-yi) + coorFace(2,2)*xi + coorFace(2,3)*yi)*zPlan
             ztmp = coorFace(3,1)*zPlan                                                     &
                  + (coorFace(3,1)*(1.0-yi-zi) + coorFace(3,2)*yi + coorFace(3,3)*zi)*xPlan &
                  + (coorFace(3,1)*(1.0-xi-zi) + coorFace(3,2)*xi + coorFace(3,3)*zi)*yPlan
             !
             xTet = (crossProd2(1)*(xtmp-coorTet(1,1)) + crossProd2(2)*(ytmp-coorTet(2,1))  &
                  + crossProd2(3)*(ztmp-coorTet(3,1)))/(6.0*measureTet)
             yTet =-(crossProd3(1)*(xtmp-coorTet(1,1)) + crossProd3(2)*(ytmp-coorTet(2,1))  & 
                  + crossProd3(3)*(ztmp-coorTet(3,1)))/(6.0*measureTet)
             zTet = (crossProd4(1)*(xtmp-coorTet(1,1)) + crossProd4(2)*(ytmp-coorTet(2,1))  & 
                  + crossProd4(3)*(ztmp-coorTet(3,1)))/(6.0*measureTet)
             !
             xHex = xi*xPlan                                                                &
                  + (-1 + 2*(xtmp-coorHex(1,1))/(coorHex(1,7)-coorHex(1,1)))*(1-xPlan)
             yHex = yi*yPlan                                                                &
                  + (-1 + 2*(ytmp-coorHex(2,1))/(coorHex(2,7)-coorHex(2,1)))*(1-yPlan)
             zHex = zi*zPlan                                                                &
                  + (-1 + 2*(ztmp-coorHex(3,1))/(coorHex(3,7)-coorHex(3,1)))*(1-zPlan)
             !
             CALL BASES_LAGRANGE_POLY_HEX(coorHref, xHex, yHex, zHex, orderHex,  phi)
             CALL BASES_LAGRANGE_POLY_TET(coorTref, xTet, yTet, zTet, orderFace, psi)
             !
             val  = val + phi(k)*psi(j)*qw2d(i)
          ENDDO
          mat(k,j) = val
       ENDDO
    ENDDO
    !
    IF (MeshCell(MeshFac(faceNum)%neighb(1))%cellType == 'TET1') THEN
       permHex = MeshFac(faceNum)%perm(2)
       permTet = MeshFac(faceNum)%perm(1)
    ELSE
       permHex = MeshFac(faceNum)%perm(1)
       permTet = MeshFac(faceNum)%perm(2)
    ENDIF
    !
    DO j=1,nDofFace
       jPerm = iPermDofTet(orderFace)%tab(j,permTet)
       DO i=1,nDofQua
          iPerm                     = iPermDofHex(orderHex)%tab(i,permHex)
          mSurfHybF(iFace)%tab(i,j) = mat(iPerm,jPerm)*2.0d0
       ENDDO
    ENDDO
    !
    IF (ALLOCATED(qx2d)) DEALLOCATE(qx2d)
    IF (ALLOCATED(qy2d)) DEALLOCATE(qy2d)
    IF (ALLOCATED(qw2d)) DEALLOCATE(qw2d)
    IF (ALLOCATED(nq2d)) DEALLOCATE(nq2d)
    IF (ALLOCATED(mat))  DEALLOCATE(mat)
    IF (ALLOCATED(phi))  DEALLOCATE(phi)
    IF (ALLOCATED(psi))  DEALLOCATE(psi)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_MATRICES_HYBRID_HDG_F

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE BASES_LAGRANGE_HDG_MATRICES


