MODULE BASES_LAGRANGE_POLYNOMIALS

  !====================================================================
  ! Polynomial functions for Lagrange basis
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE BASES_LAGRANGE_ORDER_TO_NDOF
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------
  
CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BASES_LAGRANGE_POLY(celltype, verts, xq, yq, zq, deg, phi)
    !------------------------------------------------------------------
    !---- Generic call to basis functions with a cell type
    !
    !---- IN
    !---- cellType : cell type
    !---- vtk      : tetrahedron vertices
    !---- *q       : coordinates of the point
    !---- deg      : polynomial degree
    !
    !---- OUT
    !---- phi      : vector of lagrange polynomials evaluated at given point
    !    
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR)                             :: cellType
    
    INTEGER(C_INT),                            INTENT(INOUT)  :: deg

    REAL(C_DOUBLE),                            INTENT(INOUT)  :: xq, yq, zq
    REAL(C_DOUBLE), DIMENSION(:,:),            INTENT(INOUT)  :: verts
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE                 :: phi
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call right subroutine
    !------------------------------------------------------------------
    SELECT CASE(cellType)
    CASE('TET','TET1','TET2','TET3','TET4')
       CALL BASES_LAGRANGE_POLY_TET(verts, xq, yq, zq, deg, phi)
    CASE('HEX','HEX1','HEX2')
       CALL BASES_LAGRANGE_POLY_HEX(verts, xq, yq, zq, deg, phi)
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_POLY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BASES_LAGRANGE_POLY_TET(vtk, xq, yq, zq, deg, phi)
    !------------------------------------------------------------------
    !---- Computation of Lagrange polynomials value at at given point
    !
    !---- IN
    !---- vtk : tetrahedron vertices
    !---- *q  : coordinates of the point
    !---- deg : polynomial degree
    !
    !---- OUT
    !---- phi : vector of lagrange polynomials evaluated at given point
    !    
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                            INTENT(IN)  :: deg

    REAL(C_DOUBLE),                            INTENT(IN)  :: xq, yq, zq
    REAL(C_DOUBLE), DIMENSION(3,4),            INTENT(IN)  :: vtk
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE, INTENT(OUT) :: phi
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                                         :: d0, q3, &
         q6, u1, u2, u3, u4
    REAL(C_DOUBLE), DIMENSION(3)                           :: v1, v2, &
         v3, v4, qi
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (ALLOCATED(phi)) DEALLOCATE(phi)
    !------------------------------------------------------------------
    !---- Initialize a few things
    !------------------------------------------------------------------
    q3 = 1.d0/3.d0
    q6 = 1.d0/6.d0
    !
    !---- Coordinates of the vertices of the tetrahedron containing the point
    v1(:) = vtk(:,1)
    v2(:) = vtk(:,2)
    v3(:) = vtk(:,3)
    v4(:) = vtk(:,4)
    !
    !---- Coordinates of the point
    qi(1) = xq
    qi(2) = yq
    qi(3) = zq
    !
    !---- Compute the distance of the point to the barycentric coordinates
    d0 = 1.d0/SPECIAL_DET_4X4(v1, v2, v3, v4)
    u1 = SPECIAL_DET_4X4(qi, v2, v3, v4)*d0
    u2 = SPECIAL_DET_4X4(v1, qi, v3, v4)*d0
    u3 = SPECIAL_DET_4X4(v1, v2, qi, v4)*d0
    u4 = SPECIAL_DET_4X4(v1, v2, v3, qi)*d0
    !------------------------------------------------------------------
    !---- Compute values of Lagrange polynomials
    !------------------------------------------------------------------
    SELECT CASE (deg)
    CASE (0)
       !---- Zero order
       ALLOCATE(phi(1))
       !
       phi(1) = 1.0d0
    CASE (1)
       !---- First order
       ALLOCATE(phi(4))
       !
       phi(1) = u1
       phi(2) = u2
       phi(3) = u3
       phi(4) = u4
    CASE (2)
       !---- Second order
       ALLOCATE(phi(10))  
       !
       phi(1)  = u1*(2.d0*u1 - 1.d0)
       phi(2)  = u2*(2.d0*u2 - 1.d0)
       phi(3)  = u3*(2.d0*u3 - 1.d0)
       phi(4)  = u4*(2.d0*u4 - 1.d0)
       phi(5)  = 4.d0*u1*u2
       phi(6)  = 4.d0*u2*u3
       phi(7)  = 4.d0*u1*u3
       phi(8)  = 4.d0*u2*u4
       phi(9)  = 4.d0*u1*u4
       phi(10) = 4.d0*u3*u4
    CASE (3)
       !---- Third order
       ALLOCATE(phi(20))
       !
       phi(1)  = u1*(3.d0*u1 - 1.d0)*(3.d0*u1 - 2.d0)*0.5d0
       phi(2)  = u2*(3.d0*u2 - 1.d0)*(3.d0*u2 - 2.d0)*0.5d0
       phi(3)  = u3*(3.d0*u3 - 1.d0)*(3.d0*u3 - 2.d0)*0.5d0
       phi(4)  = u4*(3.d0*u4 - 1.d0)*(3.d0*u4 - 2.d0)*0.5d0
       phi(5)  = 9.d0*u1*u2*(3.d0*u1 - 1.d0)*0.5d0
       phi(6)  = 9.d0*u1*u2*(3.d0*u2 - 1.d0)*0.5d0
       phi(7)  = 9.d0*u2*u3*(3.d0*u2 - 1.d0)*0.5d0
       phi(8)  = 9.d0*u2*u3*(3.d0*u3 - 1.d0)*0.5d0
       phi(9)  = 9.d0*u1*u3*(3.d0*u3 - 1.d0)*0.5d0
       phi(10) = 9.d0*u1*u3*(3.d0*u1 - 1.d0)*0.5d0
       phi(11) = 27.d0*u1*u2*u3
       phi(12) = 9.d0*u2*u4*(3.d0*u4 - 1.d0)*0.5d0
       phi(13) = 9.d0*u2*u4*(3.d0*u2 - 1.d0)*0.5d0
       phi(14) = 9.d0*u1*u4*(3.d0*u1 - 1.d0)*0.5d0
       phi(15) = 9.d0*u1*u4*(3.d0*u4 - 1.d0)*0.5d0
       phi(16) = 27.d0*u1*u2*u4
       phi(17) = 9.d0*u3*u4*(3.d0*u3 - 1.d0)*0.5d0
       phi(18) = 9.d0*u3*u4*(3.d0*u4 - 1.d0)*0.5d0
       phi(19) = 27.d0*u1*u3*u4
       phi(20) = 27.d0*u2*u3*u4
    CASE (4)
       !---- Fourth order
       ALLOCATE(phi(35))
       !
       phi(1)  = u1*(4.d0*u1 - 1.d0)*(4.d0*u1 - 2.d0)*(4.d0*u1 - 3.d0)*q6
       phi(2)  = u2*(4.d0*u2 - 1.d0)*(4.d0*u2 - 2.d0)*(4.d0*u2 - 3.d0)*q6
       phi(3)  = u3*(4.d0*u3 - 1.d0)*(4.d0*u3 - 2.d0)*(4.d0*u3 - 3.d0)*q6
       phi(4)  = u4*(4.d0*u4 - 1.d0)*(4.d0*u4 - 2.d0)*(4.d0*u4 - 3.d0)*q6
       phi(5)  = 8.d0*u1*u2*(4.d0*u1 - 1.d0)*(4.d0*u1 - 2.d0)*q3
       phi(6)  = 4.d0*u1*u2*(4.d0*u1 - 1.d0)*(4.d0*u2 - 1.d0)
       phi(7)  = 8.d0*u1*u2*(4.d0*u2 - 1.d0)*(4.d0*u2 - 2.d0)*q3
       phi(8)  = 8.d0*u2*u3*(4.d0*u2 - 1.d0)*(4.d0*u2 - 2.d0)*q3
       phi(9)  = 4.d0*u2*u3*(4.d0*u2 - 1.d0)*(4.d0*u3 - 1.d0)
       phi(10) = 8.d0*u2*u3*(4.d0*u3 - 1.d0)*(4.d0*u3 - 2.d0)*q3
       phi(11) = 8.d0*u1*u3*(4.d0*u3 - 1.d0)*(4.d0*u3 - 2.d0)*q3
       phi(12) = 4.d0*u1*u3*(4.d0*u1 - 1.d0)*(4.d0*u3 - 1.d0)
       phi(13) = 8.d0*u1*u3*(4.d0*u1 - 1.d0)*(4.d0*u1 - 2.d0)*q3
       phi(14) = 32.d0*u1*u2*u3*(4.d0*u1 - 1.d0)
       phi(15) = 32.d0*u1*u2*u3*(4.d0*u2 - 1.d0)
       phi(16) = 32.d0*u1*u2*u3*(4.d0*u3 - 1.d0)
       phi(17) = 8.d0*u2*u4*(4.d0*u4 - 1.d0)*(4.d0*u4 - 2.d0)*q3
       phi(18) = 4.d0*u2*u4*(4.d0*u2 - 1.d0)*(4.d0*u4 - 1.d0)
       phi(19) = 8.d0*u2*u4*(4.d0*u2 - 1.d0)*(4.d0*u2 - 2.d0)*q3
       phi(20) = 8.d0*u1*u4*(4.d0*u1 - 1.d0)*(4.d0*u1 - 2.d0)*q3
       phi(21) = 4.d0*u1*u4*(4.d0*u1 - 1.d0)*(4.d0*u4 - 1.d0)
       phi(22) = 8.d0*u1*u4*(4.d0*u4 - 1.d0)*(4.d0*u4 - 2.d0)*q3
       phi(23) = 32.d0*u1*u2*u4*(4.d0*u1 - 1.d0)
       phi(24) = 32.d0*u1*u2*u4*(4.d0*u2 - 1.d0)
       phi(25) = 32.d0*u1*u2*u4*(4.d0*u4 - 1.d0)
       phi(26) = 8.d0*u3*u4*(4.d0*u3 - 1.d0)*(4.d0*u3 - 2.d0)*q3
       phi(27) = 4.d0*u3*u4*(4.d0*u3 - 1.d0)*(4.d0*u4 - 1.d0)
       phi(28) = 8.d0*u3*u4*(4.d0*u4 - 1.d0)*(4.d0*u4 - 2.d0)*q3
       phi(29) = 32.d0*u1*u3*u4*(4.d0*u1 - 1.d0)
       phi(30) = 32.d0*u1*u3*u4*(4.d0*u3 - 1.d0)
       phi(31) = 32.d0*u1*u3*u4*(4.d0*u4 - 1.d0)
       phi(32) = 32.d0*u2*u3*u4*(4.d0*u2 - 1.d0)
       phi(33) = 32.d0*u2*u3*u4*(4.d0*u3 - 1.d0)
       phi(34) = 32.d0*u2*u3*u4*(4.d0*u4 - 1.d0)
       phi(35) = 256.d0*u1*u2*u3*u4
    END SELECT
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_POLY_TET

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BASES_LAGRANGE_POLY_HEX(vhk, xq, yq, zq, deg, phi)
    !------------------------------------------------------------------
    !---- Computation of Lagrange polynomials value at at given point 
    !
    !---- IN
    !---- vtk : hexahedron vertices
    !---- *q  : coordinates of the point
    !---- deg : polynomial degree
    !
    !---- OUT
    !---- phi : vector of lagrange polynomials evaluated at given point
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                            INTENT(IN)  :: deg

    REAL(C_DOUBLE),                            INTENT(IN)  :: xq, yq, zq
    REAL(C_DOUBLE), DIMENSION(3,8),            INTENT(IN)  :: vhk
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE, INTENT(OUT) :: phi
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                                         :: q2, q4, q8, q16, q48
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (ALLOCATED(phi)) DEALLOCATE(phi)
    !------------------------------------------------------------------
    !---- Initialize a few things
    !------------------------------------------------------------------
    !---- Compute denominators and factors
    q2  = 1.d0 / 2.d0
    q4  = 1.d0 / 4.d0
    q8  = 1.d0 / 8.d0
    q16 = 1.d0 / 16.d0
    q48 = 1.d0 / 48.d0
    !------------------------------------------------------------------
    !---- Compute values of Lagrange polynomials
    !------------------------------------------------------------------
    SELECT CASE (deg)
    CASE (0)
       !---- Zero order
       ALLOCATE(phi(1))
       !
       phi(1) = 1.0d0
    CASE (1)
       !---- First order
       ALLOCATE(phi(8))
       !
       phi(1) = q8*(1-xq)*(1-yq)*(1-zq)
       phi(2) = q8*(1+xq)*(1-yq)*(1-zq)
       phi(3) = q8*(1+xq)*(1+yq)*(1-zq)
       phi(4) = q8*(1-xq)*(1+yq)*(1-zq)
       phi(5) = q8*(1-xq)*(1-yq)*(1+zq)
       phi(6) = q8*(1+xq)*(1-yq)*(1+zq)
       phi(7) = q8*(1+xq)*(1+yq)*(1+zq)
       phi(8) = q8*(1-xq)*(1+yq)*(1+zq)
    CASE (2)
       !---- Second order
       ALLOCATE(phi(27))  
       !
       phi(1)  = q8*xq*(xq-1)*yq*(yq-1)*zq*(zq-1)
       phi(2)  = q8*xq*(xq+1)*yq*(yq-1)*zq*(zq-1)
       phi(3)  = q8*xq*(xq+1)*yq*(yq+1)*zq*(zq-1)
       phi(4)  = q8*xq*(xq-1)*yq*(yq+1)*zq*(zq-1)
       phi(5)  = q8*xq*(xq-1)*yq*(yq-1)*zq*(zq+1)
       phi(6)  = q8*xq*(xq+1)*yq*(yq-1)*zq*(zq+1)
       phi(7)  = q8*xq*(xq+1)*yq*(yq+1)*zq*(zq+1)
       phi(8)  = q8*xq*(xq-1)*yq*(yq+1)*zq*(zq+1)
       phi(9)  = q4*(1-xq**2)*yq*(yq-1)*zq*(zq-1)
       phi(10) = q4*xq*(xq+1)*(1-yq**2)*zq*(zq-1)
       phi(11)  = q4*(1-xq**2)*yq*(yq+1)*zq*(zq-1)
       phi(12)  = q4*xq*(xq-1)*(1-yq**2)*zq*(zq-1)
       phi(13)  = q2*(1-xq**2)*(1-yq**2)*zq*(zq-1)
       phi(14)  = q4*(1-xq**2)*yq*(yq-1)*zq*(zq+1)
       phi(15)  = q4*xq*(1+xq)*(1-yq**2)*zq*(zq+1)
       phi(16)  = q4*(1-xq**2)*yq*(1+yq)*zq*(zq+1)
       phi(17)  = q4*xq*(xq-1)*(1-yq**2)*zq*(zq+1)
       phi(18)  = q2*(1-xq**2)*(1-yq**2)*zq*(zq+1)
       phi(19)  = q4*xq*(xq-1)*yq*(yq-1)*(1-zq**2)
       phi(20)  = q4*xq*(xq-1)*yq*(yq+1)*(1-zq**2)
       phi(21)  = q2*xq*(xq-1)*(1-yq**2)*(1-zq**2)
       phi(22)  = q4*xq*(xq+1)*yq*(yq-1)*(1-zq**2)
       phi(23)  = q4*xq*(xq+1)*yq*(yq+1)*(1-zq**2)
       phi(24)  = q2*xq*(xq+1)*(1-yq**2)*(1-zq**2)
       phi(25)  = q2*(1-xq**2)*yq*(yq-1)*(1-zq**2)
       phi(26)  = q2*(1-xq**2)*yq*(yq+1)*(1-zq**2)
       phi(27)  = (1-xq**2)*(1-yq**2)*(1-zq**2)
    CASE (3)
       !---- Third order
       ALLOCATE(phi(64))
       !
       phi(1) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(2) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q48)*(3*yq+1d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(3) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q48*(3*yq+3d0)     &
            *(3*yq+1d0)*(3*yq-1d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(4) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(5) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(6) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q48)*(3*yq+1d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(7) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q48*(3*yq+3d0)     &
            *(3*yq+1d0)*(3*yq-1d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(8) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(9) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(10) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)  &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(11) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(12) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(13) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q48*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-1d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(14) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(15) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(16) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(17) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(18) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(19) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(20) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*(-q48)*(3*zq+1d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(21) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)  &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(22) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(23) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(24) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-1d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(25) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(26) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)  &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(27) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(28) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(29) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q48*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-1d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(30) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(31) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(32) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(33) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(34) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(35) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(36) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*q16*(3*zq+3d0)*(3*zq-1d0)*(3*zq-3d0)
       phi(37) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)  &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(38) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(39) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(40) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-1d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(41) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(42) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)  &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(43) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(44) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(45) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q48*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-1d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(46) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(47) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(48) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(49) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(50) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(51) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(52) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*(-q16)*(3*zq+3d0)*(3*zq+1d0)*(3*zq-3d0)
       phi(53) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(54) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q48)*(3*yq+1d0)  &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(55) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(56) = q48*(3*xq+3d0)*(3*xq+1d0)*(3*xq-1d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(57) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q48*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-1d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(58) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q48*(3*yq+3d0)    &
            *(3*yq+1d0)*(3*yq-1d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(59) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(60) = (-q48)*(3*xq+1d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(61) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*q16*(3*yq+3d0)    &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(62) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*q16*(3*yq+3d0)   &
            *(3*yq-1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(63) = (-q16)*(3*xq+3d0)*(3*xq+1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)  &
            *(3*yq+1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
       phi(64) = q16*(3*xq+3d0)*(3*xq-1d0)*(3*xq-3d0)*(-q16)*(3*yq+3d0)   &
            *(3*yq+1d0)*(3*yq-3d0)*q48*(3*zq+3d0)*(3*zq+1d0)*(3*zq-1d0)
    CASE (4)
       !---- Fourth order
       ALLOCATE(phi(125))
       !
       phi(1) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(2) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(3) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(4) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(5) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(6) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(7) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(8) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(9) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(10) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(11) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(12) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(13) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(14) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(15) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(16) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(17) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(18) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(19) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(20) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(21) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(22) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(23) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(24) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(25) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(26) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(27) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(28) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(29) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(2*zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(30) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(31) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(32) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(33) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(34) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(35) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(36) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(37) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(38) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(39) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(40) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(41) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(42) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(43) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(44) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(45) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(46) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(47) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(48) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(49) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(50) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(51) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(52) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(53) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(54) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*zq*(2*zq-1d0)*(zq-1d0)
       phi(55) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(56) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(57) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(58) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(59) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(60) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(61) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(62) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(63) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(64) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(65) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(66) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(67) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(68) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(69) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(70) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(71) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(72) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(73) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(74) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(75) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(76) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(77) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(78) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(79) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(zq+1d0)*(2*zq+1d0)*(2*zq-1d0)*(zq-1d0)
       phi(80) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(81) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(82) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(83) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(84) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(85) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(86) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(87) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(88) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(89) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(90) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(91) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(92) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(93) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(94) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(95) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(96) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(97) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(98) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(99) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(100) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(101) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(102) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(103) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(104) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(-8./6.)*(zq+1d0)*(2*zq+1d0)*zq*(zq-1d0)
       phi(105) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(106) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(107) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(2*yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(108) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(109) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(110) = (1./6.)*(xq+1d0)*(2*xq+1d0)*xq*(2*xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(111) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(112) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(113) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(1./6.)*(yq+1d0)*(2*yq+1d0)*yq*(2*yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(114) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(115) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(116) = (1./6.)*(2*xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(117) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(118) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(119) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*yq*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(120) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(121) = (-8./6.)*(xq+1d0)*(2*xq+1d0)*xq*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(122) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(123) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(-8./6.)*(yq+1d0)*(2*yq+1d0)*yq*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(124) = (-8./6.)*(xq+1d0)*xq*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
       phi(125) = (xq+1d0)*(2*xq+1d0)*(2*xq-1d0)*(xq-1d0)	 	 	 &
            *(yq+1d0)*(2*yq+1d0)*(2*yq-1d0)*(yq-1d0)	 	 &
            *(1./6.)*(zq+1d0)*(2*zq+1d0)*zq*(2*zq-1d0)
    END SELECT
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_POLY_HEX

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE BASES_LAGRANGE_POLYNOMIALS


