MODULE BASES_LAGRANGE_PROJECTIONS

  !====================================================================
  ! Vector field projections with Lagrange basis
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE BASES_LAGRANGE_DOF
  USE BASES_LAGRANGE_POLYNOMIALS
  USE BASES_LAGRANGE_ORDER_TO_NDOF
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROJECT_VECTOR_FIELD(vecField, cellType, oldOrder, &
       newOrder, newNDof)
    !------------------------------------------------------------------
    !---- Project a vector field from a polynomial degree to another
    !
    !---- IN
    !---- vecField : vector field to project
    !---- oldOrder : order of vector field at input
    !---- newOrder : order of vector field at output
    !---- cellType : type of cell
    !
    !---- OUT
    !---- vecField : modified in the process
    !---- newNDof  : new number of dof
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=4), INTENT(IN)    :: cellType
    INTEGER(C_INT),                INTENT(IN)    :: oldOrder, newOrder
    INTEGER(C_INT),                INTENT(OUT)   :: newNDof
    
    TYPE(real_vector_field),       INTENT(INOUT) :: vecField
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Select proper routine
    !------------------------------------------------------------------
    SELECT CASE (cellType)
    CASE('TET1','TET2','TET3','TET4')
       CALL PROJECT_VECTOR_FIELD_TET(vecField, cellType, oldOrder, &
       newOrder, newNDof)
    CASE('HEX1','HEX2','HEX3','HEX4')
       CALL PROJECT_VECTOR_FIELD_HEX(vecField, cellType, oldOrder, &
       newOrder, newNDof)
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROJECT_VECTOR_FIELD

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROJECT_VECTOR_FIELD_TET(vecField, cellType, oldOrder, &
       newOrder, newNDof)
    !------------------------------------------------------------------
    !---- Project a vector field from a polynomial degree to another
    !---- on a tetrahedron
    !
    !---- IN
    !---- vecField : vector field to project
    !---- oldOrder : order of vector field at input
    !---- newOrder : order of vector field at output
    !---- cellType : type of cell
    !
    !---- OUT
    !---- vecField : modified in the process
    !---- newNDof  : new number of dof
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=4), INTENT(IN)    :: cellType
    
    INTEGER(C_INT),                INTENT(IN)    :: oldOrder, newOrder
    INTEGER(C_INT),                INTENT(OUT)   :: newNDof
    
    TYPE(real_vector_field),       INTENT(INOUT) :: vecField
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                              :: doReallocate
    INTEGER(C_INT)                               :: oldNDof
    
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE  :: projFields
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !---- Check if all projection structures are allocated
    IF (.NOT. ALLOCATED(xDofTet))  CALL BASES_LAGRANGE_DOF_TET
    IF (.NOT. ALLOCATED(mProjTet)) CALL BASES_LAGRANGE_PROJECTION_MATRICES_TET
    !------------------------------------------------------------------
    !---- Project
    !------------------------------------------------------------------
    !---- Get # of d.o.f.
    CALL ORDER_TO_NDOFV_LAGRANGE(oldOrder, cellType, oldNDof)
    CALL ORDER_TO_NDOFV_LAGRANGE(newOrder, cellType, newNDof)    
    !
    !---- Re-allocate local array to right size
    doReallocate = .FALSE.
    IF (.NOT. ALLOCATED(projFields))                                 doReallocate = .TRUE.
    IF (ALLOCATED(projFields) .AND. (SIZE(projFields,2) /= newNDof)) doReallocate = .TRUE.
    IF (doReallocate) ALLOCATE(projFields(3,newNDof))
    !
    !---- Multiply by projection matrix
    CALL MATVECMUL(mProjTet(newOrder,oldOrder)%tab, &
         vecField%vec(1,:), newNDof, oldNDof, projFields(1,:))
    CALL MATVECMUL(mProjTet(newOrder,oldOrder)%tab, &
         vecField%vec(2,:), newNDof, oldNDof, projFields(2,:))
    CALL MATVECMUL(mProjTet(newOrder,oldOrder)%tab, &
         vecField%vec(3,:), newNDof, oldNDof, projFields(3,:))
    
    !---- Modify view element
    DEALLOCATE(vecField%vec)
    ALLOCATE(vecField%vec(3,newNDof))
    vecField%vec(:,:) = projFields(:,:)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROJECT_VECTOR_FIELD_TET

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROJECT_VECTOR_FIELD_HEX(vecField, cellType, oldOrder, &
       newOrder, newNDof)
    !------------------------------------------------------------------
    !---- Project a vector field from a polynomial degree to another
    !---- on a hexahedron
    !
    !---- IN
    !---- vecField : vector field to project
    !---- oldOrder : order of vector field at input
    !---- newOrder : order of vector field at output
    !---- cellType : type of cell
    !
    !---- OUT
    !---- vecField : modified in the process
    !---- newNDof  : new number of dof
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=4), INTENT(IN)    :: cellType

    INTEGER(C_INT),                INTENT(IN)    :: oldOrder, newOrder
    INTEGER(C_INT),                INTENT(OUT)   :: newNDof
    
    TYPE(real_vector_field),       INTENT(INOUT) :: vecField
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                               :: oldNDof
    
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE  :: projFields
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !---- Check if all projection structures are allocated
    IF (.NOT. ALLOCATED(xDofHex))  CALL BASES_LAGRANGE_DOF_HEX
    IF (.NOT. ALLOCATED(mProjHex)) CALL BASES_LAGRANGE_PROJECTION_MATRICES_HEX
    !------------------------------------------------------------------
    !---- Project
    !------------------------------------------------------------------
    !---- Get # of d.o.f.
    CALL ORDER_TO_NDOFV_LAGRANGE(oldOrder, cellType, oldNDof)
    CALL ORDER_TO_NDOFV_LAGRANGE(newOrder, cellType, newNDof)    
    !
    !---- Re-allocate local array to right size
    ALLOCATE(projFields(3,newNDof))
    projFields(:,:) = 0.0
    !
    !---- Multiply by projection matrix
    CALL MATVECMUL(mProjHex(newOrder,oldOrder)%tab, &
         vecField%vec(1,:), newNDof, oldNDof, projFields(1,:))
    CALL MATVECMUL(mProjHex(newOrder,oldOrder)%tab, &
         vecField%vec(2,:), newNDof, oldNDof, projFields(2,:))
    CALL MATVECMUL(mProjHex(newOrder,oldOrder)%tab, &
         vecField%vec(3,:), newNDof, oldNDof, projFields(3,:))
    
    !---- Modify view element
    DEALLOCATE(vecField%vec)
    ALLOCATE(vecField%vec(3,newNDof))
    vecField%vec(:,:) = projFields(:,:)
    DEALLOCATE(projFields)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROJECT_VECTOR_FIELD_HEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BASES_LAGRANGE_PROJECTION_MATRICES_TET
    !------------------------------------------------------------------
    !---- Computation of projection matrices on the reference tetrahedron
    !
    !---- IN
    !
    !---- OUT
    !    
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE QUADRATURES_2D
    USE QUADRATURES_3D
    USE MATRICES_STRUCTURES
    USE REFERENCE_CELLS
    USE MESH_ELEMENTS_DOFS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variable
    !------------------------------------------------------------------
    INTEGER(C_INT)                              :: deg, nDofV, deg1, deg2, &
         nDofV1, nDofV2, j, k

    REAL(C_DOUBLE), DIMENSION(3,4)              :: vtk
    REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: phi, psi
    REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- A few coefficients and initializations
    !------------------------------------------------------------------
    !---- Vertices of the reference tetrahedron
    vtk(1:3,1) = Tref_Vert1
    vtk(1:3,2) = Tref_Vert2
    vtk(1:3,3) = Tref_Vert3
    vtk(1:3,4) = Tref_Vert4
    !------------------------------------------------------------------
    !---- Allocate matrices
    !------------------------------------------------------------------
    !---- Allocate objects
    ALLOCATE(mProjTet(0:4,0:4))
    !
    !---- Allocate inner arrays for surfacic matrix
    DO deg1=0,4
       CALL ORDER_TO_NDOFV_LAGRANGE(deg1,'TET1',nDofV1)
       !
       DO deg2=0,4
          CALL ORDER_TO_NDOFV_LAGRANGE(deg2,'TET1',nDofV2)
          !
          mProjTet(deg1,deg2)%ndof1 = nDofV1
          mProjTet(deg1,deg2)%ndof2 = nDofV2
          !
          ALLOCATE(mProjTet(deg1,deg2)%tab(nDofV1,nDofV2))
          mProjTet(deg1,deg2)%tab = 0.d0
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Computation of the projection matrices on the reference tetrahedron
    !------------------------------------------------------------------
    DO deg1=0,4
       CALL ORDER_TO_NDOFV_LAGRANGE(deg1,'TET1',nDofV1)
       !
       ALLOCATE(dofX(nDofV1), dofY(nDofV1), dofZ(nDofV1))
       CALL GET_TET_DOF(Tref_Vert1, Tref_Vert2, Tref_Vert3, Tref_Vert4, &
            nDofV1, deg1, dofX, dofY, dofZ)
       !
       DO deg2=0,4
          CALL ORDER_TO_NDOFV_LAGRANGE(deg2,'TET1',nDofV2)
          !
          DO j=1,nDofV1
             CALL BASES_LAGRANGE_POLY_TET(vtk,dofX(j),dofY(j),dofZ(j),deg1,phi)
             CALL BASES_LAGRANGE_POLY_TET(vtk,dofX(j),dofY(j),dofZ(j),deg2,psi)
             !
             DO k=1,nDofV2
                mProjTet(deg1,deg2)%tab(j,k) = phi(j)*psi(k)
             ENDDO
          ENDDO
          !
          IF (ALLOCATED(phi))  DEALLOCATE(phi)
          IF (ALLOCATED(psi))  DEALLOCATE(psi)
       ENDDO
       !
       IF (ALLOCATED(dofX)) DEALLOCATE(dofX)
       IF (ALLOCATED(dofY)) DEALLOCATE(dofY)
       IF (ALLOCATED(dofZ)) DEALLOCATE(dofZ)
    ENDDO
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_PROJECTION_MATRICES_TET

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BASES_LAGRANGE_PROJECTION_MATRICES_HEX
    !------------------------------------------------------------------
    !---- Computation of projection matrices on the reference hexahedron
    !
    !---- IN
    !
    !---- OUT
    !    
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE QUADRATURES_2D
    USE QUADRATURES_3D
    USE MATRICES_STRUCTURES
    USE REFERENCE_CELLS
    USE MESH_ELEMENTS_DOFS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variable
    !------------------------------------------------------------------
    INTEGER(C_INT)                              :: deg, nDofV, deg1, deg2, &
         nDofV1, nDofV2, j, k

    REAL(C_DOUBLE), DIMENSION(3,8)              :: vhk
    REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: phi, psi
    REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- A few coefficients and initializations
    !------------------------------------------------------------------
    !---- Vertices of the reference hexahedron
    vhk(1:3,1) = Href_Vert1
    vhk(1:3,2) = Href_Vert2
    vhk(1:3,3) = Href_Vert3
    vhk(1:3,4) = Href_Vert4
    vhk(1:3,5) = Href_Vert5
    vhk(1:3,6) = Href_Vert6
    vhk(1:3,7) = Href_Vert7
    vhk(1:3,8) = Href_Vert8
    !------------------------------------------------------------------
    !---- Allocate matrices
    !------------------------------------------------------------------
    !---- Allocate objects
    ALLOCATE(mProjHex(0:4,0:4))
    !
    !---- Allocate inner arrays for surfacic matrix
    DO deg1=0,4
       CALL ORDER_TO_NDOFV_LAGRANGE(deg1,'HEX1',nDofV1)
       !
       DO deg2=0,4
          CALL ORDER_TO_NDOFV_LAGRANGE(deg2,'HEX1',nDofV2)
          !
          mProjHex(deg1,deg2)%ndof1 = nDofV1
          mProjHex(deg1,deg2)%ndof2 = nDofV2
          !
          ALLOCATE(mProjHex(deg1,deg2)%tab(nDofV1,nDofV2))
          mProjHex(deg1,deg2)%tab = 0.d0
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Computation of the projection matrices on the reference tetrahedron
    !------------------------------------------------------------------
    DO deg1=0,4
       CALL ORDER_TO_NDOFV_LAGRANGE(deg1,'HEX1',nDofV1)
       !
       ALLOCATE(dofX(nDofV1), dofY(nDofV1), dofZ(nDofV1))
       CALL GET_HEX_DOF(Href_Vert1, Href_Vert2, Href_Vert3, &
            Href_Vert4, Href_Vert5, Href_Vert6, Href_Vert7, &
            Href_Vert8, nDofV1, deg1, dofX, dofY, dofZ)
       !
       DO deg2=0,4
          CALL ORDER_TO_NDOFV_LAGRANGE(deg2,'HEX1',nDofV2)
          !
          DO j=1,nDofV1
             CALL BASES_LAGRANGE_POLY_HEX(vhk,dofX(j),dofY(j),dofZ(j),deg1,phi)
             CALL BASES_LAGRANGE_POLY_HEX(vhk,dofX(j),dofY(j),dofZ(j),deg2,psi)
             !
             DO k=1,nDofV2
                mProjHex(deg1,deg2)%tab(j,k) = phi(j)*psi(k)
             ENDDO
          ENDDO
          !
          IF (ALLOCATED(phi))  DEALLOCATE(phi)
          IF (ALLOCATED(psi))  DEALLOCATE(psi)
       ENDDO
       !
       IF (ALLOCATED(dofX)) DEALLOCATE(dofX)
       IF (ALLOCATED(dofY)) DEALLOCATE(dofY)
       IF (ALLOCATED(dofZ)) DEALLOCATE(dofZ)
    ENDDO
    !------------------------------------------------------------------
  END SUBROUTINE BASES_LAGRANGE_PROJECTION_MATRICES_HEX  

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE BASES_LAGRANGE_PROJECTIONS


