MODULE CUBIC_CAVITY

  !====================================================================
  ! Routines related to the cubic cavity case
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! CUBIC CAVITY DATA ('CC')
  ! computeErrCC     ==> TRUE enables the computation of the overall L2 error
  !                   (see Observable data)
  ! modeNumberCC(:)  ==> defines the cavity mode to compute
  ! *sizeCC          ==> lateral sizes of the cavity
  ! omegaCC          ==> mode pulsation
  ! k*CC             ==> waveNumbers
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)              :: computeErrCC

  INTEGER(C_INT), DIMENSION(3) :: modeNumberCC

  REAL(C_DOUBLE)               :: xSizeCC, ySizeCC, zSizeCC, omegaCC, &
       kxCC, kyCC, kzCC
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_CUBIC_CAVITY
    !------------------------------------------------------------------
    !---- Pre-processing for cubic cavity case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MATH_PHYS_CONSTANTS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: nx, ny, nz
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    nx = modeNumberCC(1)
    ny = modeNumberCC(2)
    nz = modeNumberCC(3)
    !
    IF ((nx <= 0) .OR. (ny <= 0) .OR. (nz <= 0)) THEN
       IF(idWrite) WRITE(6, '(a)') 'Error in routine INIT_CUBIC_CAVITY'
       IF(idWrite) WRITE(6, '(a)') 'The mode number you provided is not allowed'
       IF(idWrite) WRITE(6, '(a)') 'Mode numbers should be positive integers'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Determine dimensions of the box
    !------------------------------------------------------------------
    xSizeCC = globalMesh%coorMax(1) - globalMesh%coorMin(1)
    ySizeCC = globalMesh%coorMax(2) - globalMesh%coorMin(2)
    zSizeCC = globalMesh%coorMax(3) - globalMesh%coorMin(3)
    !------------------------------------------------------------------
    !---- Initialize data
    !------------------------------------------------------------------
    omegaCC = Pi*SQRT((nx/xSizeCC)**2 + (ny/ySizeCC)**2 + (nz/zSizeCC)**2)
    kxCC    = Pi*nx/xSizeCC
    kyCC    = Pi*ny/ySizeCC
    kzCC    = Pi*nz/zSizeCC
    !------------------------------------------------------------------
    !---- Print sum-up
    !------------------------------------------------------------------
    IF(idWrite) WRITE(6, '(a)')        'Source informations'
    IF(idWrite) WRITE(6, 10)
    IF(idWrite) WRITE(6, '(a)')        'Running case: cubic cavity mode'
    IF(idWrite) WRITE(6, '(a,3i7)')    'Mode number is: ',nx, ny, nz
    IF(idWrite) WRITE(6, '(a,3f11.3)') 'Cavity dimensions (in m.) are: ',xSizeCC, ySizeCC, zSizeCC
    IF(idWrite) WRITE(6, '(a,f11.3)')  'Omega in GHz is: ',omegaCC
    IF(idWrite) WRITE(6, 10)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_CUBIC_CAVITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_CUBIC_CAVITY_TIME(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the cubic cavity mode at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_CUBIC_CAVITY(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_CUBIC_CAVITY_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_CUBIC_CAVITY(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Solution of the cubic cavity mode at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: ct, st, iomegaCC, &
         cx, cy, cz, sx, sy, sz
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    iomegaCC = 1.0/omegaCC
    !
    cx = COS(x*kxCC)
    cy = COS(y*kyCC)
    cz = COS(z*kzCC)
    sx = SIN(x*kxCC)
    sy = SIN(y*kyCC)
    sz = SIN(z*kzCC)
    !
    ct = COS(omegaCC*time)
    st = SIN(omegaCC*time)
    !
    solE(1) =-cx*sy*sz*ct
    solE(2) = 0.0d0
    solE(3) = sx*sy*cz*ct
    solH(1) =-sx*cy*cz*st*kyCC*iomegaCC
    solH(2) = cx*sy*cz*st*(kxCC+kzCC)*iomegaCC
    solH(3) =-cx*cy*sz*st*kyCC*iomegaCC
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_CUBIC_CAVITY

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE CUBIC_CAVITY
