MODULE CUBIC_CAVITY_NL

  !====================================================================
  ! Routines related to the nonlocal cubic cavity case
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! NON-LOCAL CUBIC CAVITY DATA ('NL')
  ! computeErrNL     ==> TRUE enables the computation of the overall L2 error
  !                   (see Observable data)
  ! modeNumberNL(:)  ==> defines the cavity mode to compute
  ! *sizeNL          ==> lateral sizes of the cavity
  ! omegaNL          ==> mode pulsation
  ! k*NL             ==> waveNumbers
  ! betaNL           ==> beta parameter for non local model
  ! paramNL          ==> parameter with default value 1.0, it is put to
  !                      0.0 in the case of the non-local cavity case
  !                      to decouple (E,H) from (J,Q)
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)              :: computeErrNL

  INTEGER(C_INT), DIMENSION(3) :: modeNumberNL

  REAL(C_DOUBLE)               :: paramNL = 1.0
  REAL(C_DOUBLE)               :: xSizeNL, ySizeNL, zSizeNL, omegaNL, &
       kxNL, kyNL, kzNL, betaNL
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_CUBIC_CAVITY_NL
    !------------------------------------------------------------------
    !---- Pre-processing for nonlocal cubic cavity case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MATERIALS_STRUCTURES
    USE MATH_PHYS_CONSTANTS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: nx, ny, nz
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    nx = modeNumberNL(1)
    ny = modeNumberNL(2)
    nz = modeNumberNL(3)
    !
    IF ((nx <= 0) .OR. (ny <= 0) .OR. (nz <= 0)) THEN
       IF(idWrite) WRITE(6, '(a)') 'Error in routine INIT_CUBIC_CAVITY_NL'
       IF(idWrite) WRITE(6, '(a)') 'The mode number you provided is not allowed'
       IF(idWrite) WRITE(6, '(a)') 'Mode numbers should be positive integers'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Determine dimensions of the box
    !------------------------------------------------------------------
    xSizeNL = globalMesh%coorMax(1) - globalMesh%coorMin(1)
    ySizeNL = globalMesh%coorMax(2) - globalMesh%coorMin(2)
    zSizeNL = globalMesh%coorMax(3) - globalMesh%coorMin(3)
    !------------------------------------------------------------------
    !---- Initialize data
    !---- paramNL is put to 0.0 to decouple (E,H) from (J,Q)
    !------------------------------------------------------------------
    betaNL  = MatDisp(1)%poleNL(1)%beta
    omegaNL = betaNL*Pi*SQRT((nx/xSizeNL)**2 + (ny/ySizeNL)**2 + (nz/zSizeNL)**2)
    kxNL    = Pi*nx/xSizeNL
    kyNL    = Pi*ny/ySizeNL
    kzNL    = Pi*nz/zSizeNL
    paramNL = 0.0
    !------------------------------------------------------------------
    !---- Print sum-up
    !------------------------------------------------------------------
    IF(idWrite) WRITE(6, '(a)')        'Source informations'
    IF(idWrite) WRITE(6, 10)
    IF(idWrite) WRITE(6, '(a)')        'Running case: nonlocal cubic cavity mode '
    IF(idWrite) WRITE(6, '(a,3i7)')    'Mode number: ',nx, ny, nz
    IF(idWrite) WRITE(6, '(a,3e11.3)') 'Cavity dimensions in meters: ',xSizeNL, ySizeNL, zSizeNL
    IF(idWrite) WRITE(6, '(a,e11.3)')  'Omega in GHz: ',omegaNL
    IF(idWrite) WRITE(6, 10)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_CUBIC_CAVITY_NL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_CUBIC_CAVITY_NL_TIME(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the cubic cavity mode at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell        
    !---- time  : physical time for solution
    !
    !---- OUT
    !---- J*    : current field solution vectors
    !---- Q     : charge density field solution 
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE)               :: solQ
    REAL(C_DOUBLE), DIMENSION(3) :: solJ
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_CUBIC_CAVITY_NL(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solJ, solQ)
       !
       JxV(iDof) = solJ(1)
       JyV(iDof) = solJ(2)
       JzV(iDof) = solJ(3)
       QV (iDof) = solQ
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_CUBIC_CAVITY_NL_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_CUBIC_CAVITY_NL(time, x, y, z, solJ, solQ)
    !------------------------------------------------------------------
    !---- Solution of the cubic cavity mode at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for J and Q
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE),               INTENT(OUT) :: solQ
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solJ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: ct, st, iomegaNL, &
         cx, cy, cz, sx, sy, sz
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    iomegaNL = 1.0/omegaNL
    !
    cx = COS(x*kxNL)
    cy = COS(y*kyNL)
    cz = COS(z*kzNL)
    sx = SIN(x*kxNL)
    sy = SIN(y*kyNL)
    sz = SIN(z*kzNL)
    !
    ct = COS(omegaNL*time)
    st = SIN(omegaNL*time)
    !
    solJ(1) = -betaNL**2*iomegaNL*kxNL*sx*cy*cz*ct
    solJ(2) = -betaNL**2*iomegaNL*kyNL*cx*sy*cz*ct
    solJ(3) = -betaNL**2*iomegaNL*kzNL*cx*cy*sz*ct
    solQ    =                          cx*cy*cz*st
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_CUBIC_CAVITY_NL

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE CUBIC_CAVITY_NL
