MODULE CUBIC_MESH

  !====================================================================
  ! Everything related to regular cubic mesh generation
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! General data
  !
  ! meshFile           ==> name of the output file
  ! cellType           ==> type of cells 
  ! volumicTag         ==> tag for volumic cells
  ! cubeDiscretization ==> # of elements per direction
  ! plusBoundaryTags   ==> tags for x+, y+ and z+ faces
  ! minusBoundaryTags  ==> tags for x-, y- and z- faces
  ! cubeOrigin         ==> coordinates of the cube origin
  ! cubeSize           ==> length of the cube sides
  ! stepSize           ==> length of each element
  !--------------------------------------------------------------------
  CHARACTER(LEN=1000, KIND=C_CHAR) :: meshFile
  CHARACTER(LEN=3,    KIND=C_CHAR) :: cellType

  INTEGER(C_INT)                   :: volumicTag
  INTEGER(C_INT), DIMENSION(3)     :: cubeDiscretization, plusBoundaryTags, &
       minusBoundaryTags

  REAL(C_DOUBLE), DIMENSION(3)     :: cubeOrigin, cubeSize, stepSize
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_CONFIG_CUBIC_MESH(configFile)
    !------------------------------------------------------------------
    !---- This routine reads cubic mesh config file
    !
    !---- IN
    !---- dataFile : file containing cubic mesh parameters
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: configFile
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR)          :: extension
    CHARACTER(LEN=100, KIND=C_CHAR)           :: tag
    
    INTEGER(C_INT)                            :: nUnit, iRead, iPlane, iPt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Open file
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(configFile))
    !------------------------------------------------------------------
    !---- Read MeshFile
    !------------------------------------------------------------------
    tag = '#MeshFile'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) meshFile
    !
    CALL REMOVE_EXTENSION(meshFile, extension)
    Mesh%name      = TRIM(meshFile)
    Mesh%extension = TRIM(extension)
    !------------------------------------------------------------------
    !---- Read CellType
    !------------------------------------------------------------------
    tag = '#CellType'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) cellType
    !------------------------------------------------------------------
    !---- Read CubeOrigin
    !------------------------------------------------------------------
    tag = '#CubeOrigin'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) cubeOrigin(:)
    !------------------------------------------------------------------
    !---- Read CubeSize
    !------------------------------------------------------------------
    tag = '#CubeSize'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) cubeSize(:)
    !------------------------------------------------------------------
    !---- Read CubeDiscretization
    !------------------------------------------------------------------
    tag = '#CubeDiscretization'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) cubeDiscretization(:)
    !------------------------------------------------------------------
    !---- Read BoundaryTags
    !------------------------------------------------------------------
    tag = '#BoundaryTags'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) THEN
       READ(nUnit,*) minusBoundaryTags(1), plusBoundaryTags(1)
       READ(nUnit,*) minusBoundaryTags(2), plusBoundaryTags(2)
       READ(nUnit,*) minusBoundaryTags(3), plusBoundaryTags(3)
    ENDIF
    !------------------------------------------------------------------
    !---- Read VolumicTag
    !------------------------------------------------------------------
    tag = '#VolumicTag'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) volumicTag
    !------------------------------------------------------------------
    !---- Close file
    !------------------------------------------------------------------
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_CONFIG_CUBIC_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_CUBE_VERTICES_AND_CELLS
    !------------------------------------------------------------------
    !---- This routine builds the vertices and cells of the cubic mesh
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_CELLS
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iX, iY, iZ
    INTEGER(C_INT), DIMENSION(4) :: tetBuffer
    INTEGER(C_INT), DIMENSION(8) :: vertIndex, hexBuffer

    REAL(C_DOUBLE), DIMENSION(3) :: pt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute preliminary data
    !------------------------------------------------------------------
    stepSize(:) = cubeSize(:)/cubeDiscretization(:)
    !------------------------------------------------------------------
    !---- Build cube vertices and cells
    !------------------------------------------------------------------
    DO iX=1,cubeDiscretization(1)
       DO iY=1,cubeDiscretization(2)
          DO iZ=1,cubeDiscretization(3)
             !---- First vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX-1)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY-1)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ-1)
             CALL ADD_VERTEX(pt)
             vertIndex(1) = Mesh%nVert
             !
             !---- Second vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY-1)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ-1)
             CALL ADD_VERTEX(pt)
             vertIndex(2) = Mesh%nVert
             !
             !---- Third vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX-1)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ-1)
             CALL ADD_VERTEX(pt)
             vertIndex(3) = Mesh%nVert
             !
             !---- Fourth vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ-1)
             CALL ADD_VERTEX(pt)
             vertIndex(4) = Mesh%nVert
             !
             !---- Fifth vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX-1)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY-1)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ)
             CALL ADD_VERTEX(pt)
             vertIndex(5) = Mesh%nVert
             !
             !---- Sixth vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY-1)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ)
             CALL ADD_VERTEX(pt)
             vertIndex(6) = Mesh%nVert
             !
             !---- Seventh vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX-1)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ)
             CALL ADD_VERTEX(pt)
             vertIndex(7) = Mesh%nVert
             !
             !---- Eighth vertex
             pt(1) = cubeOrigin(1) + stepSize(1)*(iX)
             pt(2) = cubeOrigin(2) + stepSize(2)*(iY)
             pt(3) = cubeOrigin(3) + stepSize(3)*(iZ)
             CALL ADD_VERTEX(pt)
             vertIndex(8) = Mesh%nVert
             !
             !---- Cells
             SELECT CASE(cellType)
             CASE('TET')
                !---- First tetrahedron
                tetBuffer(1) = vertIndex(2)
                tetBuffer(2) = vertIndex(4)
                tetBuffer(3) = vertIndex(3)
                tetBuffer(4) = vertIndex(6)
                CALL ADD_CELL(tetBuffer(:), volumicTag, cellType)
                !
                !---- Second tetrahedron
                tetBuffer(1) = vertIndex(2)
                tetBuffer(2) = vertIndex(3)
                tetBuffer(3) = vertIndex(5)
                tetBuffer(4) = vertIndex(6)
                CALL ADD_CELL(tetBuffer(:), volumicTag, cellType)
                !
                !---- Third tetrahedron
                tetBuffer(1) = vertIndex(1)
                tetBuffer(2) = vertIndex(2)
                tetBuffer(3) = vertIndex(3)
                tetBuffer(4) = vertIndex(5)
                CALL ADD_CELL(tetBuffer(:), volumicTag, cellType)
                !
                !---- Fourth tetrahedron
                tetBuffer(1) = vertIndex(7)
                tetBuffer(2) = vertIndex(4)
                tetBuffer(3) = vertIndex(6)
                tetBuffer(4) = vertIndex(3)
                CALL ADD_CELL(tetBuffer(:), volumicTag, cellType)
                !
                !---- Fifth tetrahedron
                tetBuffer(1) = vertIndex(7)
                tetBuffer(2) = vertIndex(3)
                tetBuffer(3) = vertIndex(6)
                tetBuffer(4) = vertIndex(5)
                CALL ADD_CELL(tetBuffer(:), volumicTag, cellType)
                !
                !---- Sixth tetrahedron
                tetBuffer(1) = vertIndex(8)
                tetBuffer(2) = vertIndex(4)
                tetBuffer(3) = vertIndex(6)
                tetBuffer(4) = vertIndex(7)
                CALL ADD_CELL(tetBuffer(:), volumicTag, cellType)
             CASE('HEX')
                hexBuffer(1) = vertIndex(1)
                hexBuffer(2) = vertIndex(2)
                hexBuffer(3) = vertIndex(4)
                hexBuffer(4) = vertIndex(3)
                hexBuffer(5) = vertIndex(5)
                hexBuffer(6) = vertIndex(6)
                hexBuffer(7) = vertIndex(8)
                hexBuffer(8) = vertIndex(7)
                CALL ADD_CELL(hexBuffer(:), volumicTag, cellType)
             END SELECT
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished building cube vertices and cells'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_CUBE_VERTICES_AND_CELLS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_CUBE_FACES
    !------------------------------------------------------------------
    !---- This routine builds the faces of the cubic mesh
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE MESH_VERTICES
    USE MESH_STATS
    USE MESH_ELEMENTS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=3, KIND=C_CHAR) :: faceType
    
    LOGICAL(C_BOOL)               :: isInFace
    
    INTEGER(C_INT)                :: iCell, iFace, iVert, faceNum, &
         nVerts, nLinVerts, nHoVerts, geoOrder
    INTEGER(C_INT), DIMENSION(4)  :: vFace

    REAL(C_DOUBLE), DIMENSION(4)  :: xFace, yFace, zFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build cube faces
    !------------------------------------------------------------------
    CALL CELL_TYPE_TO_FACE_TYPE(cellType, faceType)
    CALL FACE_TYPE_TO_VERTS(faceType, nVerts, nLinVerts, nHoVerts, geoOrder)
    CALL COMPUTE_MESH_BOUNDING_BOX
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    !
    DO iCell=1,Mesh%nCell
       DO iFace=1,MeshCell(iCell)%nFaces
          !---- Retrieve face vertices coordinates
          CALL GET_FACE_LINEAR_VERTICES_FROM_CELL(iCell, iFace, vFace(1:nLinVerts))
          !
          xFace(:) = 0.0
          yFace(:) = 0.0
          zFace(:) = 0.0
          !
          DO iVert=1,nLinVerts
             xFace(iVert) = MeshVert(vFace(iVert))%coor(1)
             yFace(iVert) = MeshVert(vFace(iVert))%coor(2)
             zFace(iVert) = MeshVert(vFace(iVert))%coor(3)
          ENDDO
          !
          !---- Face x=xMin
          isInFace = ALL(ABS(xFace(1:nLinVerts)-Mesh%coorMin(1)) < Mesh%tolerance1D)
          IF (isInFace) THEN
             CALL ADD_FACE(vFace(1:nLinVerts), minusBoundaryTags(1), faceType, faceNum)
             CYCLE
          ENDIF
          !
          !---- Face x=xMax
          isInFace = ALL(ABS(xFace(1:nLinVerts)-Mesh%coorMax(1)) < Mesh%tolerance1D)
          IF (isInFace) THEN
             CALL ADD_FACE(vFace(1:nLinVerts), plusBoundaryTags(1), faceType, faceNum)
             CYCLE
          ENDIF          
          !
          !---- Face y=yMin
          isInFace = ALL(ABS(yFace(1:nLinVerts)-Mesh%coorMin(2)) < Mesh%tolerance1D)
          IF (isInFace) THEN
             CALL ADD_FACE(vFace(1:nLinVerts), minusBoundaryTags(2), faceType, faceNum)
             CYCLE
          ENDIF          
          !
          !---- Face y=yMax
          isInFace = ALL(ABS(yFace(1:nLinVerts)-Mesh%coorMax(2)) < Mesh%tolerance1D)
          IF (isInFace) THEN
             CALL ADD_FACE(vFace(1:nLinVerts), plusBoundaryTags(2), faceType, faceNum)
             CYCLE
          ENDIF          
          !
          !---- Face z=zMin
          isInFace = ALL(ABS(zFace(1:nLinVerts)-Mesh%coorMin(3)) < Mesh%tolerance1D)
          IF (isInFace) THEN
             CALL ADD_FACE(vFace(1:nLinVerts), minusBoundaryTags(3), faceType, faceNum)
             CYCLE
          ENDIF          
          !
          !---- Face z=zMax
          isInFace = ALL(ABS(zFace(1:nLinVerts)-Mesh%coorMax(3)) < Mesh%tolerance1D)
          IF (isInFace) THEN
             CALL ADD_FACE(vFace(1:nLinVerts), plusBoundaryTags(3), faceType, faceNum)
             CYCLE
          ENDIF          
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished building cube faces'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_CUBE_FACES
  
  !====================================================================
  !====================================================================
  !====================================================================

END MODULE CUBIC_MESH
