MODULE DIPOLE

  !====================================================================
  ! Routines related to the dipole case
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! DIPOLE WITH GAUSSIAN TIME DEPENDENCE DATA ('DI')
  !--------------------------------------------------------------------
  ! computeErrDI     ==> TRUE enables the computation of the overall L2 error
  ! ampliDI          ==> Amplitude for dipole (V.m-1)
  ! retardDI         ==> Retardation for dipole (seconds)
  ! freqMinDI        ==> Minimal frequency for dipole (Hz)
  ! freqMaxDI        ==> Maximal frequency for dipole (Hz)
  ! freqDI           ==> Mean of freqMinDI and freqMaxDI
  ! omegaMinDI       ==> Minimal angular frequency for dipole (Hz)
  ! omegaMaxDI       ==> Maximal angular frequency for dipole (Hz)
  ! omegaDI          ==> Mean of omegaMinDI and omegaMaxDI
  ! sigmaDI          ==> Control parameter for the gaussian exponential,
  !                      computed from freqMinDI and freqMaxDI
  ! posDI(:)         ==> position of the dipole
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)               :: computeErrDI
  
  REAL(C_DOUBLE)                :: ampliDI, retardDI, sigmaDI, freqMinDI, &
       freqMaxDI, freqDI, omegaMinDI, omegaMaxDI, omegaDI
  REAL(C_DOUBLE), DIMENSION(3)  :: posDI
  
  !--------------------------------------------------------------------
  ! Threshold for the computation of the adequate gaussian parameter
  !--------------------------------------------------------------------
  REAL(C_DOUBLE), PARAMETER :: epsGaussian = 1.0e-1
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_DIPOLE
    !------------------------------------------------------------------
    !---- Pre-processing for dipole case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    USE SPECIAL_FUNCTIONS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute missing data
    !------------------------------------------------------------------
    CALL FREQUENCY_TO_INVERSE_METERS(freqMinDI)
    CALL FREQUENCY_TO_INVERSE_METERS(freqMaxDI)
    freqDI     = 0.5*(freqMinDI+freqMaxDI)
    omegaDI    = 2.0*Pi*freqDI
    omegaMinDI = 2.0*Pi*freqMinDI
    omegaMaxDI = 2.0*Pi*freqMaxDI
    sigmaDI    = SQRT(2.0*LOG(1.0/epsGaussian))/ABS(omegaDI-omegaMinDI)
    CALL SECONDS_TO_METERS(retardDI)
    !
    !---- Test if retardDI is sufficiently large
    IF (retardDI < 5.0*sigmaDI) THEN
       IF(idWrite) WRITE(6, '(a)')       'WARNING from PREPROCESS_PLANE_WAVE'
       IF(idWrite) WRITE(6, '(a)')       'The value you took for retardDI might be too small'
       IF(idWrite) WRITE(6, '(a)')       'A minimal value is 5*sigmaDI'
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Your value is ', retardDI/c0
       IF(idWrite) WRITE(6, '(a,e12.5)') '5*sigmaDI  is ', 5.0*sigmaDI/c0
    ENDIF
    !------------------------------------------------------------------
    !---- Test if the center of the dipole is inside the TF/SF interface
    !------------------------------------------------------------------
    !
    ! TO DO
    ! 
    !------------------------------------------------------------------
    !---- Handle the TF and SF zones
    !------------------------------------------------------------------
    !
    ! TO DO
    !
    !------------------------------------------------------------------
    !---- Print sum-up
    !------------------------------------------------------------------
    IF(idWrite) WRITE(6, '(a)')       'Source informations'
    IF(idWrite) WRITE(6, 10)
    IF(idWrite) WRITE(6, '(a)')       'Running case: dipole '
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Central frequency in Hz   : ',freqDI*c0
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Central pulsation in Hz   : ',omegaDI*c0
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Min frequency in Hz       : ',freqMinDI*c0
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Max frequency in Hz       : ',freqMaxDI*c0
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Min pulsation in Hz       : ',omegaMinDI*c0
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Max pulsation in Hz       : ',omegaMaxDI*c0    
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Retardation in seconds    : ',retardDI/c0
    IF(idWrite) WRITE(6, '(a,e12.5)') 'Gaussian value in seconds : ',sigmaDI/c0
    IF(idWrite) WRITE(6, 10)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_DIPOLE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_DIPOLE_TIME(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the plane wave at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell        
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_DIPOLE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_DIPOLE_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FACE_SOL_DIPOLE_TIME(iFace, nDofS, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the dipole at the dof of a face
    !
    !---- IN
    !---- iFace : number of the cell
    !---- nDofS : number of dof in the face
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell        
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iFace, nDofS, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_FACE_DOF(iFace, order, nDofS, nDofV, dofXS(1:nDofS), dofYS(1:nDofS), dofZS(1:nDofS))
    !
    DO iDof=1,nDofS
       CALL TIME_SOL_DIPOLE(time, dofXS(iDof), dofYS(iDof), dofZS(iDof), solE, solH)
       !
       ExS(iDof) = solE(1)
       EyS(iDof) = solE(2)
       EzS(iDof) = solE(3)
       HxS(iDof) = solH(1)
       HyS(iDof) = solH(2)
       HzS(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FACE_SOL_DIPOLE_TIME
  
  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_DIPOLE(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the dipole at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: iSigmaDI2, rad, iRad1, &
         iRad2, val, cosT, sinT, gauss, p, pp, ppp, eR, eT, hP, xx, yy, zz
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    iSigmaDI2 = (1.0/sigmaDI)**2
    xx        = x - posDI(1)
    yy        = y - posDI(2)
    zz        = z - posDI(3)
    rad       = SQRT(xx**2 + yy**2 + zz**2)
    iRad1     = 1.0/rad
    iRad2     = iRad1**2
    !
    val       = time - retardDI - rad
    cosT      = COS(omegaDI*val)
    sinT      = SIN(omegaDI*val)
    gauss     = EXP(-(val**2)/(2.0*sigmaDI**2))
    !
    p         = sinT*gauss
    pp        = (omegaDI*cosT - val*sinT*iSigmaDI2)*gauss
    ppp       = (-2.0*omegaDI*val*cosT*iSigmaDI2 &
              + ((val*iSigmaDI2)**2 - omegaDI**2 - iSigmaDI2)*sinT)*gauss
    !
    eR        = (2.0*iRad1/(4.0*Pi))*(      pp*iRad1 + p*iRad2)
    eT        = (    iRad1/(4.0*Pi))*(ppp + pp*iRad1 + p*iRad2)
    hP        = (    iRad1/(4.0*Pi))*(ppp + pp*iRad1          )
    !
    solE(1)   = xx*zz*iRad2*(eR + eT)
    solE(2)   = yy*zz*iRad2*(eR + eT)
    solE(3)   = (iRad2*zz**2)*eR - (xx**2 + yy**2)*iRad2*eT
    solH(1)   =-yy*iRad1*hP
    solH(2)   = xx*iRad1*hP
    solH(3)   = 0.0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_DIPOLE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE DIPOLE
