MODULE DISPERSION_FIT

  !====================================================================
  ! Module containing the subroutines related to array operations
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_EXPERIMENTAL_DISPERSION_DATA(datafile, omega, eps_r, eps_i)
    !------------------------------------------------------------------
    !---- Read experimental dispersion data from file
    !---- The mandatory data format is:
    !---- omega(THz) Re(eps) Im(eps)
    !
    !---- IN
    !---- datafile : file containing experimental data
    !
    !---- OUT
    !---- omega    : array of omega of interest
    !---- eps_r    : array of real permittivity parts
    !---- eps_i    : array of imaginary permittivity parts
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR)           :: datafile
    !
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: omega, eps_r, eps_i
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: n_unit, n_rows, &
         n_cols, i_row
    !
    LOGICAL(C_BOOL)                           :: fileExists
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    CALL DOES_FILE_EXIST(dataFile, fileExists)
    IF (.NOT. fileExists) THEN
       WRITE(6, '(a)') 'Error in routine READ_EXPERIMENTAL_DISPERSION_DATA'
       WRITE(6, '(a)') 'File ',TRIM(datafile),' not found'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Determine size
    !------------------------------------------------------------------
    n_unit = 1
    CALL GET_SIZE_OF_FILE(datafile, n_rows, n_cols)
    ALLOCATE(omega(n_rows), eps_r(n_rows), eps_i(n_rows))
    !------------------------------------------------------------------
    !---- Read file
    !------------------------------------------------------------------
    OPEN(UNIT=n_unit, FILE=TRIM(datafile))
    DO i_row=1,n_rows
       READ(n_unit,*) omega(i_row), eps_r(i_row), eps_i(i_row)
    ENDDO
    !
    CLOSE(n_unit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_EXPERIMENTAL_DISPERSION_DATA

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE QUALITY_DISPERSION_FIT(omega, eps_r, eps_i, error_r, error_i)
    !------------------------------------------------------------------
    !---- Computes a quality measure of a dispersion fit
    !---- The error returned by this routine is the mean absolute error
    !---- per sample, computed over all samples
    !
    !---- IN
    !---- omega    : array of omega of interest
    !---- eps_r    : array of real permittivity parts
    !---- eps_i    : array of imaginary permittivity parts
    !
    !---- OUT
    !---- error_r  : error on real part
    !---- error_i  : error on imaginary part
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATERIALS_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: error_r, error_i
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: omega, eps_r, eps_i
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: n_samples, i_sample
    !
    REAL(C_DOUBLE) :: omega_min, omega_max, loc_omega, loc_error_r, &
         loc_error_i, loc_eps_r, loc_eps_i
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine size and initialize stuff
    !------------------------------------------------------------------
    n_samples = SIZE(omega)
    omega_min = omega(1)
    omega_max = omega(n_samples)
    !
    error_r   = 0.0
    error_i   = 0.0
    !------------------------------------------------------------------
    !---- Compute error
    !------------------------------------------------------------------
    DO i_sample=1,n_samples
       loc_error_r = 0.0
       loc_error_i = 0.0
       !
       CALL EVAL_PERMITTIVITY(1, omega(i_sample), loc_eps_r, loc_eps_i)
       !
       loc_error_r = ABS(eps_r(i_sample) - loc_eps_r)
       loc_error_i = ABS(eps_i(i_sample) - loc_eps_i)
       !
       error_r     = error_r + loc_error_r
       error_i     = error_i + loc_error_i
    ENDDO
    !
    error_r = error_r/n_samples
    error_i = error_i/n_samples
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE QUALITY_DISPERSION_FIT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SIMULATED_ANNEALING(omega, eps_r, eps_i, npole0, npole1, &
       npole2, X)
    !------------------------------------------------------------------
    !---- Perform simulated annealing optimization for permittivity function
    !
    !---- IN
    !---- omega      : array of omega of interest (in THz)
    !---- eps_r      : array of real permittivity parts
    !---- eps_i      : array of imaginary permittivity parts
    !---- npole0     : number of zero-order poles
    !---- npole1     : number of first-order poles
    !---- npole2     : number of second-order poles
    !
    !---- OUT
    !---- X          : vecteur solution
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: npole0, npole1, npole2
    !
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: omega, eps_r, eps_i
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iseed1, iseed2, ns, nt, maxevl, iprint, I, J, lines, &
         LL, IER, N, nacc, nfcnev, nobds, neps, NR
    !
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: nacp
    !
    REAL(C_DOUBLE) :: eps, RT, T, fopt
    !
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: X, VM, LB, UB, C, xopt, XP
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: fstar
    !
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Parameters
    !------------------------------------------------------------------
    mymax  = .FALSE.
    eps    = 1.0e-10
    RT     = 0.99
    iseed1 = 1
    iseed2 = 2
    ns     = 1
    nt     = 200
    maxevl = 1000000000
    iprint = 0
    T      = 1.0
    NR     = SIZE(omega)
    !------------------------------------------------------------------
    !---- Allocate and initial values
    !------------------------------------------------------------------
    N    = 1 + npole0 + 2*npole1 + 4*npole2
    neps = 4
    ALLOCATE(X(N), VM(N), LB(N), UB(N), C(N), xopt(N), XP(N))
    ALLOCATE(fstar(neps))
    ALLOCATE(nacp(N))
    !
    DO I=1,N
       X(I)  = 0.0
       VM(I) = 1.0
       LB(I) = 0.0
       UB(I) = 1.0e3
       X(I)  = 0.0
       C(I)  = 1.0
    ENDDO
    !
    !---- Initial guess and lower bound on eps_infinity
    X(1)  = 1.0
    LB(1) = 1.0
    !------------------------------------------------------------------
    !---- Calling main routine
    !------------------------------------------------------------------
    CALL SA(N,X,mymax,RT,eps,ns,nt,neps,maxevl,LB,UB,C,iprint,iseed1,&
         iseed2,T,VM,xopt,fopt,nacc,nfcnev,nobds,IER,&
         fstar,XP,nacp,omega,eps_r,eps_i,NR,npole0,npole1,npole2)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(a, e12.5)
    !------------------------------------------------------------------
  END SUBROUTINE SIMULATED_ANNEALING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FCN(N,theta,H,X1,E1,E2,NR,npole0,npole1,npole2)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: NR, N, npole0, npole1, npole2
    !
    REAL(C_DOUBLE) :: theta(N), X1(NR), E1(NR), E2(NR), S, H, par(N)
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: I, J, ind
    !
    COMPLEX(C_DOUBLE_COMPLEX) :: CX, U, JW
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Do stuff
    !------------------------------------------------------------------
    S = 0.0
    H = 0.0
    CX = (0.0, 1.0)
    !
    !---- Calculating target function
    DO I=1,NR
       S = S + ABS(E1(I) + CX*E2(I))**2
    ENDDO
    S = SQRT(S)
    DO I=1,NR
       JW  = CX*X1(I)
       U   = theta(1)
       ind = 2
       !
       DO J=1,npole0
          U   = U - theta(ind)/JW
          ind = ind + 1
       ENDDO
       !
       DO J=1,npole1
          U   = U - theta(ind)/(JW-theta(ind+1))
          ind = ind + 2
       ENDDO
       !
       DO J=1,npole2
          U   = U - (theta(ind)-JW*theta(ind+1))/(X1(I)**2-theta(ind+2)+JW*theta(ind+3))
          ind = ind + 4
       ENDDO
       !
       IF (AIMAG(U) < 0.0) THEN
          H = H + 1000.0*ABS(E1(I) + CX*E2(I) - U)**2
       ELSE
          H = H + ABS(E1(I) + CX*E2(I) - U)**2
       ENDIF
    ENDDO
    H = SQRT(H)/S
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FCN

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SA(N,X,mymax,RT,eps,ns,nt,neps,maxevl,LB,UB,C,iprint, &
       iseed1,iseed2,T,VM,xopt,fopt,nacc,nfcnev,nobds,IER, &
       fstar,XP,nacp,omega,eps_r,eps_i,NR,npole0,npole1,npole2)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: nacp(N), N, ns, nt, neps, nacc, maxevl, iprint, &
         nobds, IER, nfcnev, iseed1, iseed2, NR, npole0, npole1, npole2
    !
    REAL(C_DOUBLE) :: X(N), LB(N), UB(N), C(N), VM(N), fstar(neps), xopt(N), &
         XP(N), omega(N), eps_r(N), eps_i(N), T, eps, RT, fopt
    !
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: NUP, ndown, NREJ, NNEW, Lnobds, H, I, J, M
    !
    REAL(C_DOUBLE) :: F, FP, P, PP, RATIO
    !
    LOGICAL(C_BOOL) :: quit
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Do stuff
    !------------------------------------------------------------------
    !---- Initialize random number generator
    CALL RMARIN(iseed1,iseed2)
    !
    !---- Set initial values.
    nacc = 0
    nobds = 0
    nfcnev = 0
    IER = 99
    !
    DO I=1,N
       xopt(I) = X(I)
       nacp(I) = 0
    ENDDO
    !
    DO I=1,neps
       fstar(I) = 1.0D+20
    ENDDO
    !
    !---- Initial temperature must be positive
    IF (T .LE. 0.0) THEN
       WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. '' &
            /,''  RESET THE VARIABLE T. ''/)')
       IER = 3
       RETURN
    ENDIF
    !
    !---- Initial value must be inside bounds
    DO I=1,N
       IF ((X(I) .GT. UB(I)) .OR. (X(I) .LT. LB(I))) THEN
          CALL PRT1
          IER = 2
          RETURN
       END IF
    ENDDO
    !
    !---- Evaluate the function with input X and return value as F
    CALL FCN(N,X,F,omega,eps_r,eps_i,NR,npole0,npole1,npole2)
    !
    !---- If the function is to be minimized, switch the sign of the function.
    !---- Note that all intermediate and final output switches the sign back
    !---- to eliminate any possible confusion for the user.
    IF(.NOT. mymax) F = -F
    nfcnev = nfcnev + 1
    fopt = F
    fstar(1) = F
    IF(iprint .GE. 1) CALL PRT2(mymax,N,X,F)
    !
    !---- Start the main loop. Note that it terminates if (i) the algorithm
    !---- succesfully optimizes the function or (ii) there are too many
    !---- function evaluations (more than maxevl).
100 NUP = 0
    NREJ = 0
    NNEW = 0
    ndown = 0
    Lnobds = 0
    !
    DO M=1,nt
       DO J=1,ns
          DO H=1,N
             !
             !---- Generate XP, the trial value of X. Note use of VM to choose XP.
             DO I=1,N
                IF (I .EQ. H) THEN
                   XP(I) = X(I) + (RANMAR()*2.- 1.) * VM(I)
                ELSE
                   XP(I) = X(I)
                END IF
                !
                !---- If XP is out of bounds, select a point in bounds for the trial.
                IF((XP(I) .LT. LB(I)) .OR. (XP(I) .GT. UB(I))) THEN
                   XP(I) = LB(I) + (UB(I) - LB(I))*RANMAR()
                   Lnobds = Lnobds + 1
                   nobds = nobds + 1
                   IF(iprint .GE. 3) CALL PRT3(mymax,N,XP,X,FP,F)
                END IF
             ENDDO
             !
             !---- Evaluate the function with the trial point XP and return as FP.
             CALL FCN(N,XP,FP,omega,eps_r,eps_i,NR,npole0,npole1,npole2)
             IF(.NOT. mymax) FP = -FP
             nfcnev = nfcnev + 1
             IF(iprint .GE. 3) CALL PRT4(mymax,N,XP,X,FP,F)
             !
             !---- If too many function evaluations occur, terminate the algorithm.
             IF(nfcnev .GE. maxevl) THEN
                CALL PRT5
                IF (.NOT. mymax) fopt = -fopt
                IER = 1
                RETURN
             END IF
             !
             !---- Accept the new point if the function value increases.
             IF(FP .GE. F) THEN
                IF(iprint .GE. 3) THEN
                   WRITE(*,'(''  POINT ACCEPTED'')')
                END IF
                DO I=1,N
                   X(I) = XP(I)
                ENDDO
                F = FP
                nacc = nacc + 1
                nacp(H) = nacp(H) + 1
                NUP = NUP + 1
                !
                !---- If greater than any other point, record as new optimum.
                IF (FP .GT. fopt) THEN
                   IF(iprint .GE. 3) THEN
                      WRITE(*,'(''  NEW OPTIMUM'')')
                   END IF
                   DO I=1,N
                      xopt(I) = XP(I)
                   ENDDO
                   fopt = FP
                   NNEW = NNEW + 1
                END IF
                !
                !---- If the point is lower, use the Metropolis criteria to decide on
                !---- acceptance or rejection.
             ELSE
                P = EXPREP((FP - F)/T)
                PP = RANMAR()
                IF (PP .LT. P) THEN
                   IF(iprint .GE. 3) CALL PRT6(mymax)
                   DO I=1,N
                      X(I) = XP(I)
                   ENDDO
                   F = FP
                   nacc = nacc + 1
                   nacp(H) = nacp(H) + 1
                   ndown = ndown + 1
                ELSE
                   NREJ = NREJ + 1
                   IF(iprint .GE. 3) CALL PRT7(mymax)
                END IF
             END IF
             !
          ENDDO
       ENDDO
       !
       !---- Adjust VM so that approximately half of all evaluations are accepted.
       DO I=1,N
          RATIO = DFLOAT(nacp(I)) /DFLOAT(ns)
          IF (RATIO .GT. .6) THEN
             VM(I) = VM(I)*(1. + C(I)*(RATIO - .6)/.4)
          ELSE IF (RATIO .LT. .4) THEN
             VM(I) = VM(I)/(1. + C(I)*((.4 - RATIO)/.4))
          END IF
          IF (VM(I) .GT. (UB(I)-LB(I))) THEN
             VM(I) = UB(I) - LB(I)
          END IF
       ENDDO
       !
       IF(iprint .GE. 2) THEN
          CALL PRT8(N,VM,xopt,X)
       END IF
       !
       DO I=1,N
          nacp(I) = 0
       ENDDO
       !
    ENDDO
    !
    IF(iprint .GE. 1) THEN
       CALL PRT9(mymax,N,T,xopt,VM,fopt,NUP,ndown,NREJ,Lnobds,NNEW)
    END IF
    !
    !---- Check termination criteria.
    quit = .FALSE.
    fstar(1) = F
    IF ((fopt - fstar(1)) .LE. eps) quit = .TRUE.
    DO I=1,neps
       IF (ABS(F - fstar(I)) .GT. eps) quit = .FALSE.
    ENDDO
    !
    !---- Terminate SA if appropriate.
    IF (quit) THEN
       DO I=1,N
          X(I) = xopt(I)
       ENDDO
       IER = 0
       IF (.NOT. mymax) fopt = -fopt
       IF(iprint .GE. 1) CALL PRT10
       RETURN
    END IF
    !
    !---- If termination criteria is not met, prepare for another loop.
    T = RT*T
    DO I=neps,2,-1
       fstar(I) = fstar(I-1)
    ENDDO
    F = fopt
    DO I=1,N
       X(I) = xopt(I)
    ENDDO
    !
    !---- Loop again.
    GO TO 100
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SA

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRTVEC(VECTOR,NCOLS,NAME)
    !------------------------------------------------------------------
    !---- This subroutine prints the double precision vector named VECTOR.
    !---- Elements 1 thru NCOLS will be printed. NAME is a character variable
    !---- that describes VECTOR. Note that if NAME is given in the call to
    !---- PRTVEC, it must be enclosed in quotes. If there are more than 10
    !---- elements in VECTOR, 10 elements will be printed on each line.
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR) NAME
    !
    INTEGER(C_INT) :: NCOLS
    !
    REAL(C_DOUBLE) :: VECTOR(NCOLS)
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: I, J, LL, lines
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,1001) NAME
    !
    IF (NCOLS .GT. 10) THEN
       lines = INT(NCOLS/10.)
       !
       DO I=1,lines
          LL = 10*(I - 1)
          WRITE(*,1000) (VECTOR(J),J = 1+LL, 10+LL)
       ENDDO
       !
       WRITE(*,1000) (VECTOR(J),J = 11+LL, NCOLS)
    ELSE
       WRITE(*,1000) (VECTOR(J),J = 1, NCOLS)
    ENDIF
    !
1000 FORMAT( 10(G12.5,1X))
1001 FORMAT(/,25X,A)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRTVEC

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RMARIN(IJ, KL)
    !------------------------------------------------------------------
    !---- This subroutine and the next function generate random numbers. See
    !---- the comments for SA for more information.
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: IJ, KL
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: I97, J97, i, j, k, l, ii, jj, m
    !
    REAL(C_DOUBLE) :: C, CD, CM, U(97), s, t
    !
    COMMON /raset1/ U, C, CD, CM, I97, J97
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Do stuff
    !------------------------------------------------------------------
    IF( IJ .LT. 0  .OR.  IJ .GT. 31328  .OR. &
         KL .LT. 0  .or.  KL .GT. 30081 ) THEN
       PRINT '(A)', 'The first random number seed must have a value between 0 and 31328'
       PRINT '(A)','The second seed must have a value between 0 and 30081'
       CALL TILT
    ENDIF
    !
    i = mod(IJ/177, 177) + 2
    j = mod(IJ    , 177) + 2
    k = mod(KL/169, 178) + 1
    l = mod(KL,     169)
    !
    DO ii = 1, 97
       s = 0.0
       t = 0.5
       DO jj = 1, 24
          m = mod(mod(i*j, 179)*k, 179)
          i = j
          j = k
          k = m
          l = mod(53*l+1, 169)
          IF (mod(l*m, 64) .ge. 32) THEN
             s = s + t
          ENDIF
          t = 0.5 * t
       ENDDO
       U(ii) = s
    ENDDO
    !
    C = 362436.0 / 16777216.0
    CD = 7654321.0 / 16777216.0
    CM = 16777213.0 /16777216.0
    I97 = 97
    J97 = 33
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RMARIN

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT1
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,'(/,''  THE STARTING VALUE (X) IS OUTSIDE THE BOUNDS '' &
         /,''  (LB AND UB). EXECUTION TERMINATED WITHOUT ANY'' &
         /,''  OPTIMIZATION. RESPECIFY X, UB OR LB SO THAT  '' &
         /,''  LB(I) .LT. X(I) .LT. UB(I), I = 1, N. ''/)')
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT1

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT2(mymax,N,X,F)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: N
    !
    REAL(C_DOUBLE) :: F, X(*)
    !
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,'(''  '')')
    CALL PRTVEC(X,N,'INITIAL X')
    IF (mymax) THEN
       WRITE(*,'(''  INITIAL F: '',/, G25.18)') F
    ELSE
       WRITE(*,'(''  INITIAL F: '',/, G25.18)') -F
    END IF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT2

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT3(mymax,N,XP,X,FP,F)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: N
    !
    REAL(C_DOUBLE) :: F, FP, X(*), XP(*)
    !
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,'(''  '')')
    CALL PRTVEC(X,N,'CURREnt X')
    IF (mymax) THEN
       WRITE(*,'(''  CURREnt F: '',G25.18)') F
    ELSE
       WRITE(*,'(''  CURREnt F: '',G25.18)') -F
    END IF
    CALL PRTVEC(XP,N,'TRIAL X')
    WRITE(*,'(''  POINT REJECTED SINCE OUT OF BOUNDS'')')
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT3

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT4(mymax,N,XP,X,FP,F)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: N
    !
    REAL(C_DOUBLE) :: F, FP, X(*), XP(*)
    !
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    CALL PRTVEC(X,N,'CURREnt X')
    IF (mymax) THEN
       WRITE(*,'(''  CURREnt F: '',G25.18)') F
       CALL PRTVEC(XP,N,'TRIAL X')
       WRITE(*,'(''  RESULTING F: '',G25.18)') FP
    ELSE
       WRITE(*,'(''  CURREnt F: '',G25.18)') -F
       CALL PRTVEC(XP,N,'TRIAL X')
       WRITE(*,'(''  RESULTING F: '',G25.18)') -FP
    END IF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT4

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT5
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,'(/,''  TOO MANY FUNCTION EVALUATIOns; CONSIDER '' &
         /,''  INCREASING maxevl OR eps, OR DECREASING '' &
         /,''  nt OR RT. THESE RESULTS ARE LIKELY TO BE '' &
         /,''  POOR.'',/)')
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT5

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT6(mymax)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    IF (mymax) THEN
       WRITE(*,'(''  THOUGH LOWER, POINT ACCEPTED'')')
    ELSE
       WRITE(*,'(''  THOUGH HIGHER, POINT ACCEPTED'')')
    END IF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT6

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT7(mymax)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    IF (mymax) THEN
       WRITE(*,'(''  LOWER POINT REJECTED'')')
    ELSE
       WRITE(*,'(''  HIGHER POINT REJECTED'')')
    END IF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT7

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT8(N,VM,xopt,X)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: N
    !
    REAL(C_DOUBLE) :: VM(*), xopt(*), X(*)
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,'(/, '' INTERMEDIATE RESULTS AFTER STEP LENGTH ADJUSTMEnt'',/)')
    CALL PRTVEC(VM,N,'NEW STEP LENGTH (VM)')
    CALL PRTVEC(xopt,N,'CURREnt OPTIMAL X')
    CALL PRTVEC(X,N,'CURREnt X')
    WRITE(*,'('' '')')
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT8

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT9(mymax,N,T,xopt,VM,fopt,NUP,ndown,NREJ,Lnobds,NNEW)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: N, NUP, ndown, NREJ, Lnobds, NNEW, TOTMOV
    !
    REAL(C_DOUBLE) :: T, fopt, xopt(*), VM(*)
    !
    LOGICAL(C_BOOL) :: mymax
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    TOTMOV = NUP + ndown + NREJ
    !
    WRITE(*,'(/,'' INTERMEDIATE RESULTS BEFORE NEXT TEMPERATURE REDUCTION'',/)')
    WRITE(*,'(''  CURREnt TEMPERATURE:            '',G12.5)') T
    IF (mymax) THEN
       WRITE(*,'(''  MAX FUNCTION VALUE SO FAR:  '',G25.18)') fopt
       WRITE(*,'(''  TOTAL MOVES:                '',I8)') TOTMOV
       WRITE(*,'(''     UPHILL:                  '',I8)') NUP
       WRITE(*,'(''     ACCEPTED DOWNHILL:       '',I8)') ndown
       WRITE(*,'(''     REJECTED DOWNHILL:       '',I8)') NREJ
       WRITE(*,'(''  OUT OF BOUNDS TRIALS:       '',I8)') Lnobds
       WRITE(*,'(''  NEW MAXIMA THIS TEMPERATURE:'',I8)') NNEW
    ELSE
       WRITE(*,'(''  MIN FUNCTION VALUE SO FAR:  '',G25.18)') -fopt
       WRITE(*,'(''  TOTAL MOVES:                '',I8)') TOTMOV
       WRITE(*,'(''     DOWNHILL:                '',I8)')  NUP
       WRITE(*,'(''     ACCEPTED UPHILL:         '',I8)')  ndown
       WRITE(*,'(''     REJECTED UPHILL:         '',I8)')  NREJ
       WRITE(*,'(''  TRIALS OUT OF BOUNDS:       '',I8)')  Lnobds
       WRITE(*,'(''  NEW MINIMA THIS TEMPERATURE:'',I8)')  NNEW
    END IF
    CALL PRTVEC(xopt,N,'CURREnt OPTIMAL X')
    CALL PRTVEC(VM,N,'STEP LENGTH (VM)')
    WRITE(*,'('' '')')
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT9

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRT10
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    WRITE(*,'(/,''  SA ACHIEVED TERMINATION CRITERIA. IER = 0. '',/)')
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRT10

  !====================================================================
  !====================================================================
  !====================================================================

  FUNCTION EXPREP(RDUM) RESULT(res)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: RDUM, res
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print
    !------------------------------------------------------------------
    IF (RDUM .GT. 174.) THEN
       res = 3.69E+35
    ELSE IF (RDUM .LT. -180.) THEN
       res = 0.0
    ELSE
       res = EXP(RDUM)
    END IF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END FUNCTION EXPREP

  !====================================================================
  !====================================================================
  !====================================================================

  FUNCTION RANMAR() RESULT(res2)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: I97, J97
    !
    REAL(C_DOUBLE) :: C, CD, CM, uni, U(97), res2
    !
    COMMON /raset1/ U, C, CD, CM, I97, J97
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Do stuff
    !------------------------------------------------------------------
    uni = U(I97) - U(J97)
    IF( uni .LT. 0.0 ) uni = uni + 1.0
    U(I97) = uni
    I97 = I97 - 1
    IF(I97 .EQ. 0) I97 = 97
    J97 = J97 - 1
    IF(J97 .EQ. 0) J97 = 97
    C = C - CD
    IF( C .LT. 0.0 ) C = C + CM
    uni = uni - C
    IF( uni .LT. 0.0 ) uni = uni + 1.0
    res2 = uni
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END FUNCTION RANMAR

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE DISPERSION_FIT
