MODULE EELS_DIRAC

  !====================================================================
  ! Routines related to the eels dirac case
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! EELS DIRAC DATA ('EE')
  !--------------------------------------------------------------------
  ! qEELS            ==> Charge of the electron (As)
  ! v0EELS           ==> Absolute velocity of the electron
  ! retardEELS       ==> Retardation for eels electron (meters)
  ! amplitudeEELS    ==> Coefficient of the incident field 
  ! coeffEELS        ==> Coefficient for denominator of incident field
  ! freqMinEELS      ==> Minimal frequency for eels signal (Hz)
  ! freqMaxEELS      ==> Maximal frequency for eels signal (Hz)
  ! 
  ! vEELS            ==> Velocity vector of the electron 
  ! r0EELS           ==> Anchor of the trajectory (meters)
  ! rEELS            ==> Translated (retarded) anchor 
  !--------------------------------------------------------------------
  REAL(C_DOUBLE)               :: qEELS, retardEELS, v0EELS, &
       amplitudeEELS, gammaEELS, coeffEELS, freqMinEELS, freqMaxEELS

  REAL(C_DOUBLE), DIMENSION(3) :: vEELS, r0EELS, rEELS 
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_EELS_DIRAC
    !------------------------------------------------------------------
    !---- Pre-processing for eels dirac case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    USE SPECIAL_FUNCTIONS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: normSquared, norm
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    normSquared     = DOT_PRODUCT(vEELS, vEELS)
    norm            = SQRT(normSquared)
    !
    IF(norm /= 1.0) THEN
        IF(idWrite) WRITE(6, '(a)')      'ERROR from PEPROCESS_EELS_DIRAC'
        IF(idWrite) WRITE(6, '(a)')      'Your velocity vector is NOT normalized to 1.0'
        IF(idWrite) WRITE(6, '(a,e12.5)')'|vEELS| = ', norm
        IF(idWrite) WRITE(6, '(a)')      'I will normalize it for you'
        vEELS = vEELS / norm
    ENDIF 
    !------------------------------------------------------------------
    !---- Warn if electron does not start far enough
    !------------------------------------------------------------------
    ! to do
!    IF (retardEELS < 1.0) THEN
!       IF(idWrite) WRITE(6, '(a)')       'WARNING from PREPROCESS_EELS_DIRAC'
!       IF(idWrite) WRITE(6, '(a)')       'The value you took for retardEELS might be too small'
!       IF(idWrite) WRITE(6, '(a)')       'A minimal value is XXXXXXX'
!       IF(idWrite) WRITE(6, '(a,e12.5)') 'Your value is ', retardEELS
!       IF(idWrite) WRITE(6, '(a,e12.5)') 'Your domain size is XXXXXX'
!    ENDIF
    !
    !------------------------------------------------------------------
    !---- Print sum-up
    !------------------------------------------------------------------
    IF(idWrite) WRITE(6, '(a)')        'Source informations'
    IF(idWrite) WRITE(6, 10)
    IF(idWrite) WRITE(6, '(a)')        'Running case: EELS Dirac'
    IF(idWrite) WRITE(6, '(a,e12.5)')  'The charge of EELS beam in As: ',qEELS
    IF(idWrite) WRITE(6, '(a,3e12.5)') 'Trajectory point in m: ',r0EELS
    IF(idWrite) WRITE(6, '(a,e12.5)')  'Retarded initial distance in m: ',retardEELS
    IF(idWrite) WRITE(6, '(a,e12.5)')  'Velocity in m/s: ',v0EELS
    IF(idWrite) WRITE(6, '(a,3e12.5)') 'Normalized velocity vector: ',vEELS
    IF(idWrite) WRITE(6, 10)
    !------------------------------------------------------------------
    !---- Precompute and rescale constants
    !------------------------------------------------------------------
    CALL VELOCITY_TO_UNITLESS(v0EELS)
    !
    gammaEELS     = SQRT(1/(1-v0EELS**2))
    amplitudeEELS = qEELS/(4*Pi*eps0)
    coeffEELS     = (gammaEELS**2-1)/v0EELS*2
    vEELS         = vEELS*v0EELS
    !------------------------------------------------------------------
    !---- Shift r0EELS along trajectory (ratardation)
    !------------------------------------------------------------------
    rEELS(1) = r0EELS(1) - vEELS(1)*retardEELS/v0EELS
    rEELS(2) = r0EELS(2) - vEELS(2)*retardEELS/v0EELS
    rEELS(3) = r0EELS(3) - vEELS(3)*retardEELS/v0EELS
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_EELS_DIRAC

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_EELS_DIRAC_TIME(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the EELS dirac at the dof of a cell 
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_EELS_DIRAC(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_EELS_DIRAC_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_EELS_DIRAC(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the EELS Dirac at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: dSquared, vDotD
    REAL(C_DOUBLE), DIMENSION(3)              :: d
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    d(1)        = rEELS(1) + vEELS(1)*time - x
    d(2)        = rEELS(2) + vEELS(2)*time - y
    d(3)        = rEELS(3) + vEELS(3)*time - z
    !
    dSquared    = DOT_PRODUCT(d, d) 
    vDotD       = DOT_PRODUCT(vEELS, d)
    !
    solE(:)     = amplitudeEELS*gammaEELS*d(:)/(dSquared + coeffEELS*vDotD**2)**(1.5)
    solH(:)     = 0.0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_EELS_DIRAC

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_DT_SOL_EELS_DIRAC_TIME_SF(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Time derivative of the solution of the eels dirac at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_DT_EELS_DIRAC(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_DT_SOL_EELS_DIRAC_TIME_SF

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_DT_EELS_DIRAC(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the time derivative of the eels 
    !---- dirac at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for dtE and dtH
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: dSquared, vSquared,     &
                                                 numerator, denumerator, &
                                                 vDotD
    REAL(C_DOUBLE), DIMENSION(3)              :: d
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    d(1)        = rEELS(1) + vEELS(1)*time - x
    d(2)        = rEELS(2) + vEELS(2)*time - y
    d(3)        = rEELS(3) + vEELS(3)*time - z
    !
    dSquared    = DOT_PRODUCT(d, d) 
    vSquared    = DOT_PRODUCT(vEELS, vEELS)
    vDotD       = DOT_PRODUCT(vEELS, d)
    !
    numerator   = vDotD * (1 + coeffEELS * vSquared)
    denumerator = ( dSquared + coeffEELS*vDotD**2 )
    !
    solE        = amplitudeEELS*gammaEELS * ( &
                        (vEELS / denumerator**(1.5)) &
                       -3.0*d*numerator / denumerator**(2.5))
    solH(:)     = 0.0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_DT_EELS_DIRAC

  !====================================================================
  !====================================================================
  !====================================================================
 
END MODULE EELS_DIRAC
