MODULE ENSIGHT_MESH

  !========================================================================
  ! Module containing the subroutines related to Ensight Gold solution format
  ! Ref: http://www-vis.lbl.gov/NERSC/Software/ensight/doc/OnlineHelp/UM-C11.pdf
  !========================================================================

  !------------------------------------------------------------------------
  ! Shared modules
  !------------------------------------------------------------------------
  USE ISO_C_BINDING
  !------------------------------------------------------------------------
  IMPLICIT NONE
  !------------------------------------------------------------------------

CONTAINS

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE READ_ENSIGHT_MESH
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE IO_FILES
    USE MESH_ELEMENTS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE PARALLEL_STRUCTURES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR)      :: keyword
    CHARACTER(LEN=4,   KIND=C_CHAR)      :: type

    INTEGER(C_INT)                       :: eltTag, iElt, nElts, nUnit, i
    INTEGER(C_INT), DIMENSION(nvCellMax) :: iBuffer
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Reading and filling Mesh arrays
    !----------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(Mesh%name)//'.geo')
    !
    Readingloop: DO
       READ(nUnit, *) keyword
       SELECT CASE (trim(keyword))
          !
          !---- Vertices
       CASE('coordinates')
          READ(nUnit,*) nElts
          IF(ALLOCATED(MeshVert)) DEALLOCATE(MeshVert)
          ALLOCATE(MeshVert(nElts))
          DO i=1,3
             DO iElt=1,nElts
                READ(nUnit,*) MeshVert(iElt)%coor(i)
             ENDDO
          ENDDO
          !
          !---- Linear tetrahedra
       CASE('tetra4')
          type = 'TET1'
          !
          READ(nUnit,*) nElts
          cellBlockSize = MAX(nElts, cellBlockSize)
          DO iElt=1,nElts
             READ(nUnit,*) iBuffer(1:4), eltTag
             CALL ADD_ELEMENT(iBuffer(1:4), eltTag, type)
          ENDDO
          !
          !---- Quadratic tetrahedra
       CASE('tetra10')
          type = 'TET2'
          !
          READ(nUnit,*) nElts
          cellBlockSize = MAX(nElts, cellBlockSize)
          DO iElt=1,nElts
             READ(nUnit,*) iBuffer(1:10), eltTag
             CALL ADD_ELEMENT(iBuffer(1:10), eltTag, type)
          ENDDO
          !
          !---- Linear triangles
       CASE('tria3')
          type = 'TRI1'
          !
          READ(nUnit,*) nElts
          faceBlockSize = MAX(nElts, faceBlockSize)
          DO iElt=1,nElts
             READ(nUnit,*) iBuffer(1:3), eltTag
             CALL ADD_ELEMENT(iBuffer(1:3), eltTag, type)
          ENDDO
          !
          !---- Quadratic triangles
       CASE('tria6')
          type = 'TRI2'
          !
          READ(nUnit,*) nElts
          faceBlockSize = MAX(nElts, faceBlockSize)
          DO iElt=1,nElts
             READ(nUnit,*) iBuffer(1:6), eltTag
             CALL ADD_ELEMENT(iBuffer(1:6), eltTag, type)
          ENDDO
          !
          !---- Linear hexahedra
       CASE('hexa8')
          type = 'HEX1'
          !
          READ(nUnit,*) nElts
          cellBlockSize = MAX(nElts, cellBlockSize)
          DO iElt=1,nElts
             READ(nUnit,*) iBuffer(1:8), eltTag
             CALL ADD_ELEMENT(iBuffer(1:8), eltTag, type)
          ENDDO
          !
          !---- Linear quadrangles
       CASE('quad4')
          type = 'QUA1'
          !
          READ(nUnit,*) nElts
          faceBlockSize = MAX(nElts, faceBlockSize)
          DO iElt=1,nElts
             READ(nUnit,*) iBuffer(1:4), eltTag
             CALL ADD_ELEMENT(iBuffer(1:4), eltTag, type)
          ENDDO
          !
       CASE DEFAULT
          !
       END SELECT
    ENDDO Readingloop
    !
    CLOSE(nUnit)
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh reading'    
    !----------------------------------------------------------------------
  END SUBROUTINE READ_ENSIGHT_MESH

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_ENSIGHT_MESH
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE IO_FILES
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE MESH_TFSF_FACES
    USE MESH_STATS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)                              :: nUnit, iElt, iCell, iFace
    INTEGER(C_INT), DIMENSION(nsTriMaxEnsight)  :: triPerm
    INTEGER(C_INT), DIMENSION(nsQuadMaxEnsight) :: quadPerm
    INTEGER(C_INT), DIMENSION(nsTetMaxEnsight)  :: tetPerm
    INTEGER(C_INT), DIMENSION(nsHexMaxEnsight)  :: hexPerm
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Reset cell tags to initial values depending on zone
    !----------------------------------------------------------------------
    CALL PROCESS_TFSF_ZONES_BEFORE_WRITING
    !----------------------------------------------------------------------
    !---- In Ensight mesh we need the .case file
    !----------------------------------------------------------------------
    CALL WRITE_ENSIGHT_CASE
    !----------------------------------------------------------------------
    !---- Write mesh
    !----------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(Mesh%name)//'.geo')
    !
    WRITE(nUnit, '(a)') 'Ensight ASCII Model Geometry'
    WRITE(nUnit, '(a)') 'INRIA 2016 '
    WRITE(nUnit, '(a)') 'node id given'
    WRITE(nUnit, '(a)') 'element id given'
    WRITE(nUnit, '(a)') 'extents'
    !
    !---- Bounding Box
    WRITE(nUnit, '(2(e12.5,1x))') Mesh%coorMin(1), Mesh%coorMax(1)
    WRITE(nUnit, '(2(e12.5,1x))') Mesh%coorMin(2), Mesh%coorMax(2)
    WRITE(nUnit, '(2(e12.5,1x))') Mesh%coorMin(3), Mesh%coorMax(3)
    !
    WRITE(nUnit, '(a)') 'part'
    WRITE(nUnit, '(i1 )') 1
    WRITE(nUnit, '(a)') 'Volume'
    WRITE(nUnit, '(a)') 'coordinates'
    !
    !---- Vertices
    WRITE(nUnit, '(i10)') Mesh%nVert
    DO iElt=1,Mesh%nVert
       WRITE(nUnit, '(e12.5,1x)') MeshVert(iElt)%coor(1)
    ENDDO
    DO iElt=1,Mesh%nVert
       WRITE(nUnit, '(e12.5,1x)') MeshVert(iElt)%coor(2)
    ENDDO
    DO iElt=1,Mesh%nVert
       WRITE(nUnit, '(e12.5,1x)') MeshVert(iElt)%coor(3)
    ENDDO
    !
    !---- Linear tetrahedra
    IF (Mesh%nCellTetLin > 0) THEN
       WRITE(nUnit, '(a)'   ) 'tetra4'
       WRITE(nUnit, '(i10 )')  Mesh%nCellTetLin
       !
       DO iElt=1,Mesh%nCellTetLin
          iCell = MeshCellTetLin(iElt)%cellNum
          WRITE(nUnit, '(5i10)') MeshCell(iCell)%vertex(:), Mesh%nCellTetLin
       ENDDO
    ENDIF
    !
    !---- Quadratic tetrahedra
    IF (Mesh%nCellTetQuadr > 0) THEN
       WRITE(nUnit, '(a)') 'tetra10'
       WRITE(nUnit, '(i10 )') Mesh%nCellTetQuadr
       tetPerm = nvTetInriaToEnsight
       !
       DO iElt=1,Mesh%nCellTetQuadr
          iCell = MeshCellTetQuadr(iElt)%cellNum
          WRITE(nUnit, '(11i10)') MeshCell(iCell)%vertex(tetPerm(:)), Mesh%nCellTetQuadr
       ENDDO
    ENDIF
    !
    !---- Linear triangles
    IF (Mesh%nFacTriLin > 0) THEN
       WRITE(nUnit, '(a)') 'tria3'
       WRITE(nUnit, '(i10)')  Mesh%nFacTriLin
       !
       DO iElt=1,Mesh%nFacTriLin
          iFace = MeshFacTriLin(iElt)%faceNum
          WRITE(nUnit, '(4i10)') MeshFac(iFace)%vertex(:), Mesh%nFacTriLin
       ENDDO
    ENDIF
    !
    !---- Quadratic triangles
    IF (Mesh%nFacTriQuadr > 0) THEN    
       WRITE(nUnit, '(a)') 'tria6'
       WRITE(nUnit, '(i10)' ) Mesh%nFacTriQuadr
       triPerm = nsTriInriaToEnsight
       !
       DO iElt=1,Mesh%nFacTriQuadr
          iFace = MeshFacTriQuadr(iElt)%faceNum
          WRITE(nUnit, '(7i10)') MeshFac(iFace)%vertex(triPerm(:)), MeshFac(iFace)%log
       ENDDO
    ENDIF
    !
    !---- Linear hexahedra
    IF (Mesh%nCellHexLin > 0) THEN
       WRITE(nUnit, '(a)') 'hexa8'
       WRITE(nUnit, '(i10)' ) Mesh%nCellHexLin
       !
       DO iElt=1,Mesh%nCellHexLin
          iCell = MeshCellHexLin(iElt)%cellNum
          WRITE(nUnit, '(9i10)') MeshCell(iCell)%vertex(:), MeshCell(iCell)%log
       ENDDO
    ENDIF
    !
    !---- Linear quadrangles
    IF (Mesh%nCellHexQuadr > 0) THEN
       WRITE(nUnit, '(a)') 'quad4'
       WRITE(nUnit, '(i10)' ) Mesh%nFacQuadLin
       hexPerm = nvHexInriaToEnsight
       !
       DO iElt=1,Mesh%nFacQuadLin
          iFace = MeshFacQuadLin(iElt)%faceNum
          WRITE(nUnit, '(5i10)') MeshFac(iFace)%vertex(hexPerm(:)), MeshFac(iFace)%log
       ENDDO
    ENDIF
    !
    CLOSE(nUnit)
    !----------------------------------------------------------------------
    !-- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_ENSIGHT_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_ENSIGHT_SOLUTION
    !------------------------------------------------------------------
    !---- Read the 3D fields in Ensight format
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    USE DOF_PERMUTATIONS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: nVerts, iElt, iDof, nUnit
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_ENSIGHT_SOLUTION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_ENSIGHT_CASE
    !------------------------------------------------------------------
    !---- Write the 3D fields in Medit format
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    USE DOF_PERMUTATIONS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                 :: nUnit, iStep, iCount, stepNumber

    CHARACTER(KIND=C_CHAR, LEN=10) :: iStepC
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write solution
    !------------------------------------------------------------------
    CALL EXTRACT_SOLUTION_STEP
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(Mesh%name)//'.case')
    !
    WRITE(nUnit, '(a)'   ) 'FORMAT'
    WRITE(nUnit, '(a)'   ) 'type: ensight gold'
    WRITE(nUnit, '(a)'   ) ' '
    WRITE(nUnit, '(a)'   ) 'GEOMETRY'
    WRITE(nUnit, '(a)'   ) 'model: '//TRIM(Mesh%name)//'.geo'
    WRITE(nUnit, '(a)'   ) ' '
    WRITE(nUnit, '(a)'   ) 'VARIABLE'
    WRITE(nUnit, '(a)'   ) 'vector per node: E E_'//TRIM(Mesh%solutionName)//'_****.dat'
    WRITE(nUnit, '(a)'   ) 'vector per node: H H_'//TRIM(Mesh%solutionName)//'_****.dat'
    WRITE(nUnit, '(a)'   ) ' '
    WRITE(nUnit, '(a)'   ) 'TIME'
    WRITE(nUnit, '(a,i0)') 'time set:                ',1
    WRITE(nUnit, '(a,i0)') 'number of steps:         ',n3dView
    WRITE(nUnit, '(a,i0)') 'filename start number:   ',Mesh%initialStep
    WRITE(nUnit, '(a,i0)') 'filename increment:      ',1
    WRITE(nUnit, '(a)', advance='NO') 'time values:  '
    !
    iCount = 0
    DO iStep=1,n3dView
       iCount     = iCount + 1
       stepNumber = Mesh%initialStep + iStep - 1
       IF (iCount <  10) WRITE(nUnit, '(i6.4)', advance='NO') stepNumber
       IF (iCount == 10) WRITE(nUnit, '(i6.4)'              ) stepNumber
       IF (iCount == 10) iCount = 0
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_ENSIGHT_CASE

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_ENSIGHT_SOLUTION
    !------------------------------------------------------------------
    !---- Write the 3D fields in Medit format
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    USE DOF_PERMUTATIONS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: fileName
    CHARACTER(LEN=4, KIND=C_CHAR)    :: padding
    CHARACTER(LEN=10, KIND=C_CHAR)   :: stepC

    INTEGER(C_INT)                   :: iCoor, iDof, iElt, nUnit, padSize
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Before writing the solution, we handle the lame solution
    !---- name convention used by Ensight Gold...
    !------------------------------------------------------------------
    CALL EXTRACT_SOLUTION_STEP
    padding = ''
    !
    SELECT CASE (Mesh%solutionStep)
    CASE(0:9)
       padding = '000'
    CASE(10:99)
       padding = '00'
    CASE(100:999)
       padding = '0'
    END SELECT
    !
    CALL INT_TO_CHAR(Mesh%solutionStep, stepC)
    stepC = TRIM(padding)//TRIM(stepC)
    Mesh%solutionName = TRIM(Mesh%solutionName)//'_'//TRIM(stepC)
    !------------------------------------------------------------------
    !---- Write E solution
    !------------------------------------------------------------------    
    fileName = 'E_'//TRIM(Mesh%solutionName)//'.dat'
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(fileName))
    !
    WRITE(nUnit, '(a,i0)') 'E N=', Mesh%solutionStep
    WRITE(nUnit, '(a)'   ) 'part'
    WRITE(nUnit, '(i0)'  ) 1
    WRITE(nUnit, '(a)'   ) 'coordinates'
    !
    DO iCoor=1,3
       DO iElt=1,Mesh%nCell
          DO iDof=1,dgEM(iElt)%nDofV
             WRITE(nUnit, '(e12.5)') dgEM(iElt)%E%vec(iCoor,iDof)
          ENDDO
       ENDDO
    ENDDO
    !
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Write H solution
    !------------------------------------------------------------------    
    fileName = 'H_'//TRIM(Mesh%solutionName)//'.dat'
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(fileName))
    !
    WRITE(nUnit, '(a,i0)') 'H N=', Mesh%solutionStep
    WRITE(nUnit, '(a)'   ) 'part'
    WRITE(nUnit, '(i0)'  ) 1
    WRITE(nUnit, '(a)'   ) 'coordinates'
    !
    DO iCoor=1,3
       DO iElt=1,Mesh%nCell
          DO iDof=1,dgEM(iElt)%nDofV
             WRITE(nUnit, '(e12.5)') dgEM(iElt)%H%vec(iCoor,iDof)
          ENDDO
       ENDDO
    ENDDO
    !
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_ENSIGHT_SOLUTION

  !========================================================================
  !========================================================================
  !========================================================================

END MODULE ENSIGHT_MESH
