MODULE FIELDS_VIEW

  !====================================================================
  ! Everything related to fields view
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_VERTICES
  USE NUMERICAL_TYPES
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! General data
  !
  ! viewType       ==> CONT for continuous, DISC for discontinuous
  ! outputFormat   ==> output format for solution (ENSIGHT, MEDIT)
  ! viewOrder*     ==> final polynomial order
  ! nSolFace       ==> # of solutions on faces
  ! globalDofPos   ==> array containing all dof positions for fields averaging  
  ! verts*         ==> vertices  of cells and faces
  ! Eval, Hval     ==> arrays used for plane solutions
  ! planeVertices  ==> array  used for plane cells
  !--------------------------------------------------------------------
  CHARACTER(LEN=4, KIND=C_CHAR)               :: viewType
  CHARACTER(LEN=8, KIND=C_CHAR)               :: outputFormat

  LOGICAL(C_BOOL)                             :: computeVideo = .FALSE.

  INTEGER(C_INT)                              :: viewOrderV, viewOrderS, &
       nSolFace, initialStep
  INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: planeVertices, cellLogs

  REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: globalDofPos, Eval, Hval

  !--------------------------------------------------------------------
  ! Data structure associated to a couple element/dof
  !
  ! iElt ==> element index
  ! iDof ==> dof     index
  !--------------------------------------------------------------------
  TYPE pair_structure
     INTEGER(C_INT) :: iElt, iDof
  END TYPE pair_structure
  
  !--------------------------------------------------------------------
  ! Data structure associated to a set of couples element/dof, used to
  ! process continuity of fields in visualization
  !
  ! type      ==> CELL or FACE
  ! chunkSize ==> #     of pairs in the chunk
  ! pair      ==> array of pairs structures
  !--------------------------------------------------------------------
  TYPE chunk_structure
     CHARACTER(LEN=4, KIND=C_CHAR)                   :: type
     
     INTEGER(C_INT)                                  :: chunkSize
     
     TYPE(pair_structure), DIMENSION(:), ALLOCATABLE :: pair
  END TYPE chunk_structure

  !--------------------------------------------------------------------
  ! Data structure associated to a viewing plane
  !
  ! recLevel ==> recursive level
  ! sampling ==> # of quads per direction
  ! coords   ==> extremal coordinates of the plane
  !--------------------------------------------------------------------
  TYPE plane_structure
     INTEGER(C_INT)                 :: recLevel
     INTEGER(C_INT), DIMENSION(2)   :: sampling

     REAL(C_DOUBLE), DIMENSION(3,3) :: coords     
  END TYPE plane_structure

  !--------------------------------------------------------------------
  ! Stuff related to planes
  !
  ! nPlanes        ==> # of planes
  ! nDofPlane      ==> # of dof on each plane cell
  ! planeBlockSize ==> block size of planes
  ! plane          ==> array of planes
  !--------------------------------------------------------------------
  INTEGER(C_INT)                                   :: nPlanes, nDofPlane
  INTEGER(C_INT)                                   :: planeBlockSize = 1
  INTEGER(C_INT), PARAMETER                        :: planeFaceLog = 1

  REAL(C_DOUBLE),        DIMENSION(:), ALLOCATABLE :: dofXPlane, dofYPlane, dofZPlane  
  
  TYPE(plane_structure), DIMENSION(:), ALLOCATABLE :: plane
  
  !--------------------------------------------------------------------
  ! Stuff related to chunks
  !
  ! nChunks        ==> number of pairs in the chunk
  ! chunkBlockSize ==> block size for chunks
  ! pairBlockSize  ==> block size for pairs
  !--------------------------------------------------------------------
  INTEGER(C_INT)                                   :: nChunks
  INTEGER(C_INT)                                   :: chunkBlockSize = 500
  INTEGER(C_INT)                                   :: pairBlockSize  = 30
  
  TYPE(chunk_structure), DIMENSION(:), ALLOCATABLE :: chunk
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_CONFIG_VIEW(configFile)
    !------------------------------------------------------------------
    !---- This routine reads view config file
    !
    !---- IN
    !---- configFile : file containing view parameters
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE IO_FILES
    USE ENSIGHT_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: configFile
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR)           :: tag, readChain
    
    INTEGER(C_INT)                            :: nUnit, iRead, iPlane, iPt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Open file
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(configFile))
    !------------------------------------------------------------------
    !---- Read mesh file name
    !------------------------------------------------------------------
    tag = '#MeshFile'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) Mesh%name
    !------------------------------------------------------------------
    !---- Read solution file name
    !------------------------------------------------------------------
    tag = '#DataFile'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) Mesh%solutionName
    !------------------------------------------------------------------
    !---- Read solution file name
    !------------------------------------------------------------------
    tag = '#Video'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0  ) computeVideo = .FALSE.
    IF (iRead == 0  ) READ(nUnit,*) computeVideo
    IF (computeVideo) READ(nUnit,*) n3dView
    !
    !---- Default value
    IF (.NOT. computeVideo) n3dView = 1
    !------------------------------------------------------------------
    !---- Read viewType
    !------------------------------------------------------------------
    tag = '#ViewType'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) THEN
       READ(nUnit,*) viewType
    ENDIF
    !------------------------------------------------------------------
    !---- Read outputFormat
    !------------------------------------------------------------------
    tag = '#OutputFormat'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) READ(nUnit,*) outputFormat
    !
    SELECT CASE(outputFormat)
    CASE('ENSIGHT')
       WRITE_MESH     => WRITE_ENSIGHT_MESH
       WRITE_SOLUTION => WRITE_ENSIGHT_SOLUTION
    CASE('MEDIT')
       WRITE_MESH     => WRITE_MEDIT_MESH
       WRITE_SOLUTION => WRITE_MEDIT_SOLUTION
    CASE('MEDITB')
       WRITE_MESH     => WRITE_MEDIT_BINARY_MESH
       WRITE_SOLUTION => WRITE_MEDIT_BINARY_SOLUTION
    CASE DEFAULT
       WRITE(6,'(a)') 'Error in routine READ_CONFIG_VIEW'
       WRITE(6,'(a)') 'Unknown output format ',outputFormat
       WRITE(6,'(a)') 'Authorized types are PARAVIEW, MEDIT and MEDITB'
       CALL TILT
    END SELECT
    !------------------------------------------------------------------
    !---- Read viewOrder
    !------------------------------------------------------------------
    tag = '#ViewOrder'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) THEN
       READ(nUnit,*) viewOrderV
       READ(nUnit,*) viewOrderS
    ENDIF
    !------------------------------------------------------------------
    !---- Read viewPlanes
    !------------------------------------------------------------------
    tag = '#ViewPlanes'
    CALL REWIND_FILE_AND_FIND(nUnit, tag, iRead)
    IF (iRead <  0) CALL COULD_NOT_FIND_TAG(tag)
    IF (iRead == 0) THEN
       DO ! End of reading is controlled by IOSTAT
          READ(nUnit,*,IOSTAT=iRead) readChain
          IF (iRead == 0) THEN ! Not the end of file yet
             IF (INDEX(readChain,'#')==1) THEN ! Each new plane must be tagged by a #
                CALL EXTEND_PLANE_ARRAY(plane, nPlanes)          
                READ(nUnit,*) plane(nPlanes)%sampling(:)
                READ(nUnit,*) plane(nPlanes)%recLevel
                !
                DO iPt=1,3
                   READ(nUnit,*) plane(nPlanes)%coords(:,iPt)
                ENDDO
             ENDIF
          ELSE
             EXIT
          ENDIF
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Close file
    !------------------------------------------------------------------
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_CONFIG_VIEW

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_SOLUTION_FILES
    !------------------------------------------------------------------
    !---- This routine checks the existence of solution files
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: solName
    
    LOGICAL(C_BOOL)                  :: fileExists
    
    INTEGER(C_INT)                   :: iStep, stepNumber, fileSize
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check solution files
    !------------------------------------------------------------------
    CALL EXTRACT_SOLUTION_STEP
    Mesh%initialStep = Mesh%solutionStep
    !
    DO iStep=1,n3dView
       CALL EXTRACT_SOLUTION_STEP       
       stepNumber        = Mesh%initialStep + iStep - 1
       Mesh%solutionStep = stepNumber
       CALL ADD_STEP_TO_SOLUTION_NAME
       solName           = TRIM(Mesh%solutionName)//TRIM(Mesh%solutionExtension)
       CALL DOES_FILE_EXIST(TRIM(solName), fileExists)
       CALL GET_FILE_SIZE_IN_BYTES(TRIM(solName), fileSize)
       !
       IF ((.NOT. fileExists) .OR. (fileSize == 0)) THEN
          WRITE(6,'(a)')       'Warning in READ_CONFIG_VIEW'
          WRITE(6,'(a,i0,a)')  'You asked me to process ',n3dView,' solution files'
          WRITE(6,'(a,i0,a)')  'I could only find ',iStep-1,' valid files'
          WRITE(6,'(2(a,i0))') 'I will only process solution files from #', &
               Mesh%initialStep,' to #',stepNumber-1
          n3dView = iStep - 1
          EXIT
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_SOLUTION_FILES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SORT_INPUT_COMPATIBILITY
    !------------------------------------------------------------------
    !---- This routine checks if the user request is possible, and
    !---- modifies it if necessary
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Sort cases
    !------------------------------------------------------------------
    SELECT CASE (outputFormat)
    CASE('MEDIT', 'MEDITB') ! MEDIT format
       !---- Analyse viewOrderV input
       SELECT CASE (viewOrderV)
       CASE(1:2)
          ! Nothing
       CASE(3:)
          WRITE(6,'(a)') 'You asked for a visualization order equal to 3 or more'
          WRITE(6,'(a)') 'This is not compatible with the MEDIT format'
          WRITE(6,'(a)') 'I will use the maximal possible order, ie 2'
          viewOrderV = 2
       END SELECT
       !
       !---- Analyse viewOrderS input
       SELECT CASE (viewOrderS)
       CASE(1:2)
          ! Nothing
       CASE(3:)
          WRITE(6,'(a)') 'You asked for a visualization order equal to 3 or more'
          WRITE(6,'(a)') 'This is not compatible with the MEDIT format'
          WRITE(6,'(a)') 'I will use the maximal possible order, ie 2'
          viewOrderS = 2
       END SELECT

    CASE('ENSIGHT') ! ENSIGHT format
       !---- Analyse viewOrderV input
       SELECT CASE (viewOrderV)
       CASE(1:2)
          ! Nothing
       CASE(3:)
          WRITE(6,'(a)') 'You asked for a visualization order equal to 3 or more'
          WRITE(6,'(a)') 'This is not compatible with the ENSIGHT format'
          WRITE(6,'(a)') 'I will use the maximal possible order, ie 2'
          viewOrderV = 2
       END SELECT
       !
       !---- Analyse viewOrderS input
       SELECT CASE (viewOrderS)
       CASE(1)
          ! Nothing
       CASE(2:)
          WRITE(6,'(a)') 'You asked for a visualization order equal to 2 or more'
          WRITE(6,'(a)') 'This is not compatible with the ENSIGHT format'
          WRITE(6,'(a)') 'I will use the maximal possible order, ie 1'
          viewOrderS = 1
       END SELECT       
    CASE DEFAULT
       !
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SORT_INPUT_COMPATIBILITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_VIEW_PLANES
    !------------------------------------------------------------------
    !---- Add view planes to the mesh
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_ELEMENTS
    USE BASES_LAGRANGE
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR) :: faceType
    
    INTEGER(C_INT)                :: nQuadU, nQuadV, iPlane, &
         recLevel, maxRecLevel

    REAL(C_DOUBLE), DIMENSION(3)  :: p0, pU, pV, pt, dU, dV
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (nPlanes == 0) RETURN
    !------------------------------------------------------------------
    !---- Preliminary steps
    !------------------------------------------------------------------
    CALL BASES_LAGRANGE_DOF_QUAD
    !
    nSolFace = 0
    CALL CHANGE_TYPE_WITH_ORDER(viewOrderS, 'QUA ', faceType)
    CALL ORDER_TO_NDOFS_LAGRANGE(viewOrderS, faceType, nDofPlane)
    !------------------------------------------------------------------
    !---- Build planes
    !------------------------------------------------------------------
    DO iPlane=1,nPlanes
       WRITE(6,'(a,i0,a,i0)') 'Building plane ',iPlane,' out of ',nPlanes
       !
       p0(:)       = plane(iPlane)%coords(:,1)
       pU(:)       = plane(iPlane)%coords(:,2)
       pV(:)       = plane(iPlane)%coords(:,3)
       nQuadU      = plane(iPlane)%sampling(1)
       nQuadV      = plane(iPlane)%sampling(2)
       dU(:)       = (pU(:) - p0(:))/nQuadU
       dV(:)       = (pV(:) - p0(:))/nQuadV
       recLevel    = 0       
       maxRecLevel = plane(iPlane)%recLevel
       !
       CALL BUILD_SUBPLANE(faceType, p0, pU, pV, dU, dV, &
            nQuadU, nQuadV, planeFaceLog, recLevel, maxRecLevel)
    ENDDO
    !
    CALL BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished building planes'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_VIEW_PLANES

  !====================================================================
  !====================================================================
  !====================================================================

  RECURSIVE SUBROUTINE BUILD_SUBPLANE(faceType, p0, pU, pV, dU, dV, &
       nQuadU, nQuadV, planeFaceLog, recLevel, maxRecLevel)
    !------------------------------------------------------------------
    !---- Build a plane or a part of plane for viewing
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_ELEMENTS_DOFS
    USE BASES_LAGRANGE
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR), INTENT(INOUT) :: faceType
    
    INTEGER(C_INT),                INTENT(IN)    :: nQuadU, nQuadV, &
         planeFaceLog
    INTEGER(C_INT),                INTENT(INOUT) :: recLevel, maxRecLevel
    
    REAL(C_DOUBLE), DIMENSION(:),  INTENT(IN)    :: p0, pU, pV, dU, dV
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                              :: addQuad, tryRecursive

    INTEGER(C_INT)                               :: iQuadU, iQuadV, &
         nQuadUQ, nQuadVQ, localRecLevel

    REAL(C_DOUBLE), DIMENSION(3)                 :: p0Q, pUQ, pVQ, dUQ, dVQ
    REAL(C_DOUBLE), DIMENSION(3,4)               :: verts
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Preliminary steps
    !------------------------------------------------------------------
    recLevel      = recLevel + 1
    localRecLevel = recLevel
    !
    IF (.NOT. ALLOCATED(dofXPlane)) ALLOCATE(dofXPlane(nDofPlane))
    IF (.NOT. ALLOCATED(dofYPlane)) ALLOCATE(dofYPlane(nDofPlane))
    IF (.NOT. ALLOCATED(dofZPlane)) ALLOCATE(dofZPlane(nDofPlane))
    !------------------------------------------------------------------
    !---- Build subplane
    !------------------------------------------------------------------
    DO iQuadU=1,nQuadU
       DO iQuadV=1,nQuadV
          !---- Main vertices of the quadrangle
          verts(:,1) = p0(:) + REAL(iQuadU-1)*dU(:) + REAL(iQuadV-1)*dV(:)
          verts(:,2) = p0(:) + REAL(iQuadU-1)*dU(:) + REAL(iQuadV  )*dV(:)
          verts(:,3) = p0(:) + REAL(iQuadU  )*dU(:) + REAL(iQuadV  )*dV(:)
          verts(:,4) = p0(:) + REAL(iQuadU  )*dU(:) + REAL(iQuadV-1)*dV(:)
          !
          CALL GET_QUAD_DOF(verts(:,1), verts(:,2), verts(:,3), verts(:,4), &
               nDofPlane, viewOrderS, dofXPlane, dofYPlane, dofZPlane)
          !
          CALL DO_ADD_QUAD_IN_PLANE(dofXPlane, dofYPlane, dofZPlane,        &
               addQuad, tryRecursive)
          !
          IF (addQuad) CALL ADD_PLANE_QUAD(dofXPlane, dofYPlane, dofZPlane, &
               planeFaceLog, faceType)
          !
          IF (.NOT. addQuad) THEN
             IF (.NOT. tryRecursive) CYCLE
             !
             recLevel = localRecLevel
             IF (recLevel > maxRecLevel) CYCLE
             !
             p0Q(:)  = verts(:,1)
             pUQ(:)  = verts(:,2)
             pVQ(:)  = verts(:,4)
             nQuadUQ = 2
             nQuadVQ = 2
             dUQ(:)  = (pUQ(:) - p0Q(:))/nQuadUQ
             dVQ(:)  = (pVQ(:) - p0Q(:))/nQuadVQ
             !
             CALL BUILD_SUBPLANE(faceType, p0Q, pUQ, pVQ, dUQ, dVQ, &
                  nQuadUQ, nQuadVQ, planeFaceLog, recLevel, maxRecLevel)
          ENDIF
       ENDDO
       !
       IF (localRecLevel == 1) WRITE(6,'(a,i3,a)') 'Processed ',100*iQuadU/nQuadU,' %'
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_SUBPLANE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_PLANE_QUAD(dofX, dofY, dofZ, planeFaceLog, faceType)
    !------------------------------------------------------------------
    !---- Add a quad to a viewing plane
    !
    !---- IN
    !---- dof*         : positions of quad dofs
    !---- planeFaceLog : log of quad to add
    !---- faceType     : type of quad to add
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR), INTENT(INOUT) :: faceType
    
    INTEGER(C_INT),                INTENT(IN)    :: planeFaceLog
    
    REAL(C_DOUBLE), DIMENSION(:),  INTENT(IN)    :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                               :: iFace, iDof

    REAL(C_DOUBLE), DIMENSION(3)                 :: pt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. ALLOCATED(planeVertices)) ALLOCATE(planeVertices(nDofPlane))
    !------------------------------------------------------------------
    !---- Add quad
    !------------------------------------------------------------------    
    DO iDof=1,nDofPlane
       pt(1) = dofX(iDof)
       pt(2) = dofY(iDof)
       pt(3) = dofZ(iDof)
       !
       CALL ADD_VERTEX(pt)
       planeVertices(iDof) = Mesh%nVert
    ENDDO
    !
    CALL ADD_FACE(planeVertices, planeFaceLog, faceType, iFace)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_PLANE_QUAD

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_VIEW_PLANES_SOLUTION
    !------------------------------------------------------------------
    !---- Compute view planes solutions
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_STATS
    USE MESH_FACES
    USE MESH_VERTICES
    USE MESH_ELEMENTS_DOFS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iQuad, iFac, log, order

    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, v4
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (nPlanes == 0) RETURN
    !------------------------------------------------------------------
    !---- Build planes
    !------------------------------------------------------------------
    DO iQuad=1,Mesh%nFacQuad
       iFac = MeshFacQuad(iQuad)%faceNum
       IF (MeshFac(iFac)%log /= planeFaceLog) CYCLE
       !
       v1(:) = MeshVert(MeshFac(iFac)%vertex(1))%coor(:)
       v2(:) = MeshVert(MeshFac(iFac)%vertex(2))%coor(:)
       v3(:) = MeshVert(MeshFac(iFac)%vertex(3))%coor(:)
       v4(:) = MeshVert(MeshFac(iFac)%vertex(4))%coor(:)
       !
       order = MeshFac(iFac)%geoOrder
       CALL GET_QUAD_DOF(v1, v2, v3, v4, nDofPlane, order, &
            dofXPlane, dofYPlane, dofZPlane)
       CALL ADD_PLANE_QUAD_SOL(dofXPlane, dofYPlane, dofZPlane)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_VIEW_PLANES_SOLUTION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_PLANE_QUAD_SOL(dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Add solution of a viewing plane quad
    !
    !---- IN
    !---- dof* : positions of quad dofs
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_ELEMENTS
    USE MESH_FIND_POINT_IN_ELEMENTS
    USE BASES_LAGRANGE
    USE DOF_PERMUTATIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(:),  INTENT(IN) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)             :: cellType
    
    INTEGER(C_INT)                            :: iCell, iVert, iDof, &
         cellOrder, nDofV

    REAL(C_DOUBLE), DIMENSION(3)              :: pt
    REAL(C_DOUBLE), DIMENSION(3,4)            :: localVertsCell
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: phi    
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate arrays if necessary
    !------------------------------------------------------------------
    IF (.NOT. ALLOCATED(Eval)) ALLOCATE(Eval(3,nDofPlane))
    IF (.NOT. ALLOCATED(Hval)) ALLOCATE(Hval(3,nDofPlane))    
    !------------------------------------------------------------------
    !---- Add quad solution
    !------------------------------------------------------------------    
    Eval(:,:) = 0.0
    Hval(:,:) = 0.0
    !
    DO iDof=1,nDofPlane
       pt(1) = dofX(iDof)
       pt(2) = dofY(iDof)
       pt(3) = dofZ(iDof)
       CALL FIND_POINT_IN_CELLS(pt, iCell)
       !
       !---- iVert loops over the cell vertices
       DO iVert=1,4
          localVertsCell(:,iVert) = MeshVert(MeshCell(iCell)%vertex(iVert))%coor(:)
       ENDDO
       !
       cellOrder = dgEM(iCell)%order
       cellType  = MeshCell(iCell)%cellType
       CALL BASES_LAGRANGE_POLY_TET(localVertsCell, pt(1), pt(2), pt(3), cellOrder, phi)
       CALL ORDER_TO_NDOFV_LAGRANGE(cellOrder, cellType, nDofV)
       !
       DO iVert=1,nDofV
          Eval(:,iDof) = Eval(:,iDof) + dgEM(iCell)%E%vec(:,iVert)*phi(iVert)
          Hval(:,iDof) = Hval(:,iDof) + dgEM(iCell)%H%vec(:,iVert)*phi(iVert)
       ENDDO
    ENDDO
    !
    CALL EXTEND_NUMERICAL_FACE_ARRAY(EMf, nSolFace)
    EMf(nSolFace)%nDofS = nDofPlane
    EMf(nSolFace)%order = viewOrderS
    ALLOCATE(EMf(nSolFace)%E%vec(3,nDofPlane))
    ALLOCATE(EMf(nSolFace)%H%vec(3,nDofPlane))
    EMf(nSolFace)%E%vec(:,1:nDofPlane) = Eval(:,1:nDofPlane)
    EMf(nSolFace)%H%vec(:,1:nDofPlane) = Hval(:,1:nDofPlane)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_PLANE_QUAD_SOL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DO_ADD_QUAD_IN_PLANE(dofX, dofY, dofZ, addQuad, tryRecursive)
    !------------------------------------------------------------------
    !---- Check if quad must be added in viewing plane
    !
    !---- IN
    !
    !---- OUT
    !---- addQuad   : TRUE if quad must be added    
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_ELEMENTS
    USE MESH_FIND_POINT_IN_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),              INTENT(OUT) :: addQuad, tryRecursive

    REAL(C_DOUBLE), DIMENSION(:), INTENT(IN)  :: dofX, dofY, dofZ    
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: iCell, iDof, &
         nPtsNotOk, cellLog

    REAL(C_DOUBLE), DIMENSION(3)              :: pt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. ALLOCATED(cellLogs)) ALLOCATE(cellLogs(nDofPlane))
    !------------------------------------------------------------------
    !---- Check if quad must be added
    !------------------------------------------------------------------    
    addQuad      = .TRUE.
    tryRecursive = .FALSE.
    cellLogs(:)  = 0
    nPtsNotOk    = 0
    !
    DO iDof=1,nDofPlane
       pt(1) = dofX(iDof)
       pt(2) = dofY(iDof)
       pt(3) = dofZ(iDof)
       CALL FIND_POINT_IN_CELLS(pt, iCell)
       !
       IF (iCell == 0) THEN
          addQuad = .FALSE.
          RETURN
       ENDIF
       !
       cellLog        = MeshCell(iCell)%log
       cellLogs(iDof) = cellLog
    ENDDO
    !
    cellLog = cellLogs(1)
    DO iDof=1,nDofPlane
       IF (cellLog /= cellLogs(iDof)) nPtsNotOk = nPtsNotOk + 1
    ENDDO
    !
    IF (nPtsNotOk == 0)                        RETURN
    IF (nPtsNotOk >  0)        addQuad      = .FALSE.
    IF (nPtsNotOk < nDofPlane) tryRecursive = .TRUE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DO_ADD_QUAD_IN_PLANE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADJUST_MESH_ORDER
    !------------------------------------------------------------------
    !---- This routine modifies mesh order depending on input
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_ELEMENTS_DOFS
    USE BASES_LAGRANGE
    USE REFERENCE_CELLS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=4)             :: newType, cellType

    INTEGER(C_INT)                            :: order, iCell, iVert, &
         nVerts, nVertsCell
    INTEGER(C_INT), DIMENSION(4)              :: vertices

    REAL(C_DOUBLE), DIMENSION(3)              :: v1, v2, v3, v4, v5, &
         v6, v7, v8, buffer
    REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: dofX, dofY, dofZ

    TYPE(vertex),   DIMENSION(:), ALLOCATABLE :: vertsCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Pre-process steps
    !------------------------------------------------------------------
    nVertsCell = 0
    IF (ALLOCATED(vertsCell)) DEALLOCATE(vertsCell)
    IF (.NOT. ALLOCATED(xDofTet)) CALL BASES_LAGRANGE_DOF_TET
    !------------------------------------------------------------------
    !---- Adjust mesh
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       !---- Get new vertices coordinates
       cellType = MeshCell(iCell)%cellType
       CALL ORDER_TO_NDOFV_LAGRANGE(viewOrderV, cellType, nVerts)
       IF (ALLOCATED(dofX)) DEALLOCATE(dofX, dofY, dofZ)
       ALLOCATE(dofX(nVerts), dofY(nVerts), dofZ(nVerts))
       CALL GET_CELL_DOF_FROM_INDEX(iCell, viewOrderV, nVerts, &
            dofX, dofY, dofZ)
       !
       !---- Store new vertices coordinates
       DO iVert=1,nVerts
          buffer(1) = dofX(iVert)
          buffer(2) = dofY(iVert)
          buffer(3) = dofZ(iVert)
          !
          CALL EXTEND_VERTEX_ARRAY(vertsCell, nVertsCell)
          vertsCell(nVertsCell)%coor(:) = buffer(:)
       ENDDO
       !
       !---- Update cell
       MeshCell(iCell)%geoOrder = viewOrderV       
       MeshCell(iCell)%nVerts   = nVerts
       !
       DEALLOCATE(MeshCell(iCell)%vertex)
       ALLOCATE(MeshCell(iCell)%vertex(nVerts))
       !
       DO iVert=1,nVerts
          MeshCell(iCell)%vertex(iVert) = nVertsCell - nVerts + iVert
       ENDDO
       !
       CALL CHANGE_TYPE_WITH_ORDER(viewOrderV, cellType, newType)
       MeshCell(iCell)%cellType = newType
    ENDDO
    !------------------------------------------------------------------
    CALL COPY_VERTICES(vertsCell, MeshVert, nVertsCell)
    Mesh%nVert = nVertsCell
    IF (ALLOCATED(vertsCell)) DEALLOCATE(vertsCell)
    CALL BUILD_DERIVED_CELLS
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished adjusting mesh'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADJUST_MESH_ORDER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADJUST_SOLUTION_ORDER
    !------------------------------------------------------------------
    !---- This routine modifies fields order depending on input
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE BASES_LAGRANGE
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=4) :: cellType

    INTEGER(C_INT)                :: order, nDofView, iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Project fields
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       order    = dgEM(iCell)%order
       cellType = MeshCell(iCell)%cellType
       CALL PROJECT_VECTOR_FIELD(dgEM(iCell)%E, cellType, &
            order, viewOrderV, nDofView)
       CALL PROJECT_VECTOR_FIELD(dgEM(iCell)%H, cellType, &
            order, viewOrderV, nDofView)
       !
       dgEM(iCell)%order = viewOrderV
       dgEM(iCell)%nDofV = nDofView
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADJUST_SOLUTION_ORDER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_CONTINUITY
    !------------------------------------------------------------------
    !---- This routine pre-processes the continuity of the fields through
    !---- cells and faces interfaces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR) :: cType
    
    LOGICAL(C_BOOL)               :: isEqual, keepLooking, addNewLog

    INTEGER(C_INT)                :: iCell, iMult, nVertsCount,   &
         multiplicity, iVertLocal, iVert, iVertGlobal, iElt, log, &
         nUniqueLogs, iLog, logMultiplicity, type, iFac

    INTEGER(C_INT), DIMENSION(20) :: logList

    REAL(C_DOUBLE), DIMENSION(3)  :: diffCoor
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (viewType == 'DISC') RETURN
    !------------------------------------------------------------------
    !---- Fill array for vertices sorting
    !---- We use an ugly convention where 1 is for cells and 2 for faces...
    !------------------------------------------------------------------
    IF (ALLOCATED(globalDofPos)) DEALLOCATE(globalDofPos)
    ALLOCATE(globalDofPos(7,Mesh%nVert+1))
    globalDofPos(:,:) = 0.0
    !
    nVertsCount = 0
    !
    DO iCell=1,Mesh%nCell
       DO iVert=1,MeshCell(iCell)%nVerts
          nVertsCount                   = nVertsCount + 1
          iVertGlobal                   = MeshCell(iCell)%vertex(iVert)
          log                           = MeshCell(iCell)%log
          globalDofPos(1:3,nVertsCount) = MeshVert(iVertGlobal)%coor(1:3)
          globalDofPos(4,  nVertsCount) = REAL(iCell)
          globalDofPos(5,  nVertsCount) = REAL(iVert)
          globalDofPos(6,  nVertsCount) = REAL(log)
          globalDofPos(7,  nVertsCount) = REAL(1)
       ENDDO
    ENDDO
    !
    DO iFac=1,Mesh%nFac
       IF (MeshFac(iFac)%log == planeFaceLog) CYCLE
       !
       DO iVert=1,MeshFac(iFac)%nVerts
          nVertsCount                   = nVertsCount + 1
          iVertGlobal                   = MeshFac(iFac)%vertex(iVert)
          log                           = MeshFac(iFac)%log
          globalDofPos(1:3,nVertsCount) = MeshVert(iVertGlobal)%coor(1:3)
          globalDofPos(4,  nVertsCount) = REAL(iFac)
          globalDofPos(5,  nVertsCount) = REAL(iVert)
          globalDofPos(6,  nVertsCount) = REAL(log)
          globalDofPos(7,  nVertsCount) = REAL(2)
       ENDDO
           ENDDO
    !
    CALL SORT_ARRAY(7,nVertsCount,globalDofPos(:,1:nVertsCount))
    !------------------------------------------------------------------
    !---- Average fields
    !------------------------------------------------------------------
    iVert = 1
    DO WHILE(iVert <= nVertsCount)
       keepLooking  = .TRUE.
       multiplicity = 1
       !
       !---- Check multiplicity of the dof
       DO WHILE(keepLooking)
          diffCoor(1:3) = globalDofPos(1:3,iVert)-globalDofPos(1:3,iVert+multiplicity)
          isEqual   = ALL(ABS(diffCoor) < Mesh%tolerance1D)
          IF (      isEqual) multiplicity = multiplicity + 1
          IF (.NOT. isEqual) keepLooking  = .FALSE.
       ENDDO
       !
       !---- Fill list with unique logs
       logList(:)  =-1
       nUniqueLogs = 0
       DO iMult=0,multiplicity-1
          log = FLOOR(globalDofPos(6,iVert+iMult))
          !
          addNewLog = .TRUE.
          DO iLog=1,nUniqueLogs
             IF (log == logList(iLog)) addNewLog = .FALSE.
          ENDDO
          !
          IF (addNewLog) THEN
             nUniqueLogs          = nUniqueLogs + 1
             logList(nUniqueLogs) = log
          ENDIF
       ENDDO
       !
       !---- Store information about the chunks
       DO iLog=1,nUniqueLogs
          CALL EXTEND_CHUNK_ARRAY(chunk, nChunks)
          !
          DO iMult=0,multiplicity-1 ! Careful to the loop bounds
             iElt       = FLOOR(globalDofPos(4,iVert+iMult))
             iVertLocal = FLOOR(globalDofPos(5,iVert+iMult))
             log        = FLOOR(globalDofPos(6,iVert+iMult))
             type       = FLOOR(globalDofPos(7,iVert+iMult))
             IF (type == 1) cType = 'CELL'
             IF (type == 2) cType = 'FACE'
             IF (log /= logList(iLog)) CYCLE
             CALL EXTEND_PAIR_ARRAY(chunk(nChunks)%pair, chunk(nChunks)%chunkSize)
             !
             chunk(nChunks)%pair(chunk(nChunks)%chunkSize)%iElt = iElt
             chunk(nChunks)%pair(chunk(nChunks)%chunkSize)%iDof = iVertLocal
             chunk(nChunks)%type                                = cType
          ENDDO
       ENDDO
       !
       !---- Increment counter
       iVert = iVert + multiplicity
    ENDDO
    !------------------------------------------------------------------
    IF (ALLOCATED(globalDofPos)) DEALLOCATE(globalDofPos)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_CONTINUITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_CONTINUITY
    !------------------------------------------------------------------
    !---- This routine processes the continuity of the fields through
    !---- cells and faces interfaces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)  :: type
    
    INTEGER(C_INT)                 :: iChunk, iPair, iElt, iDof

    REAL(C_DOUBLE), DIMENSION(3,2) :: fieldValue
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (viewType == 'DISC') RETURN
    !------------------------------------------------------------------
    !---- Average fields
    !------------------------------------------------------------------
    DO iChunk=1,nChunks
       fieldValue(:,:) = 0.0
       !
       DO iPair=1,chunk(iChunk)%chunkSize
          iElt = chunk(iChunk)%pair(iPair)%iElt
          iDof = chunk(iChunk)%pair(iPair)%iDof
          type = chunk(iChunk)%type
          !
          SELECT CASE (type)
          CASE('CELL')
             fieldValue(:,1) = fieldValue(:,1) + dgEM(iElt)%E%vec(:,iDof)
             fieldValue(:,2) = fieldValue(:,2) + dgEM(iElt)%H%vec(:,iDof)
          CASE('FACE')
             fieldValue(:,1) = fieldValue(:,1) + EMf(iElt)%E%vec(:,iDof)
             fieldValue(:,2) = fieldValue(:,2) + EMf(iElt)%H%vec(:,iDof)
          END SELECT
       ENDDO
       !
       fieldValue(:,:) = fieldValue(:,:)/REAL(chunk(iChunk)%chunkSize)
       !
       DO iPair=1,chunk(iChunk)%chunkSize
          iElt = chunk(iChunk)%pair(iPair)%iElt
          iDof = chunk(iChunk)%pair(iPair)%iDof
          type = chunk(iChunk)%type
          !
          SELECT CASE (type)
          CASE('CELL')          
             dgEM(iElt)%E%vec(:,iDof) = fieldValue(:,1)
             dgEM(iElt)%H%vec(:,iDof) = fieldValue(:,2)
          CASE('FACE')
             EMf(iElt)%E%vec(:,iDof)  = fieldValue(:,1)
             EMf(iElt)%H%vec(:,iDof)  = fieldValue(:,2)
          END SELECT
       ENDDO       
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_CONTINUITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_CHUNK_ARRAY(chunkArray, chunkArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to chunkArray
    !
    !---- IN
    !---- chunkArray     : array of "chunk" structure
    !---- chunkArraySize : number of elements in chunkArray
    !
    !---- OUT
    !---- chunkArray     : modified in the process
    !---- chunkArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                    INTENT(INOUT) :: chunkArraySize
    
    TYPE(chunk_structure), DIMENSION(:), ALLOCATABLE :: chunkArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                  :: isArrayFull, isAllocated

    INTEGER(C_INT)                                   :: iChunk
    
    TYPE(chunk_structure), DIMENSION(:), ALLOCATABLE :: tempChunk
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(chunkArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(chunkArray) < chunkArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempChunk)) DEALLOCATE(tempChunk)
          ALLOCATE(tempChunk(chunkArraySize))
          tempChunk = chunkArray
          DEALLOCATE(chunkArray)
       ENDIF
       !
       ALLOCATE(chunkArray(chunkArraySize + chunkBlockSize))
       DO iChunk=chunkArraySize+1,chunkArraySize+chunkBlockSize
          CALL INITIALIZE_CHUNK_STRUCTURE(chunkArray(iChunk))
       ENDDO
       !
       chunkArray(1:chunkArraySize) = tempChunk(1:chunkArraySize)
       IF (ALLOCATED(tempChunk)) DEALLOCATE(tempChunk)
    ENDIF
    !
    chunkArraySize = chunkArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_CHUNK_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_CHUNK_STRUCTURE(chunkStructure)
    !------------------------------------------------------------------
    !---- Initialize chunk structure
    !
    !---- IN
    !---- chunkStructure : chunk structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(chunk_structure), INTENT(INOUT) :: chunkStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize chunk structure
    !------------------------------------------------------------------
    chunkStructure%chunkSize = 0
    chunkStructure%type      = ''
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_CHUNK_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_PAIR_ARRAY(pairArray, pairArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to pairArray
    !
    !---- IN
    !---- pairArray     : array of "pair" structure
    !---- pairArraySize : number of elements in pairArray
    !
    !---- OUT
    !---- pairArray     : modified in the process
    !---- pairArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(INOUT) :: pairArraySize
    
    TYPE(pair_structure), DIMENSION(:), ALLOCATABLE :: pairArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                 :: isArrayFull, isAllocated

    INTEGER(C_INT)                                  :: iPair
    
    TYPE(pair_structure), DIMENSION(:), ALLOCATABLE :: tempPair
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(pairArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(pairArray) < pairArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempPair)) DEALLOCATE(tempPair)
          ALLOCATE(tempPair(pairArraySize))
          tempPair = pairArray
          DEALLOCATE(pairArray)
       ENDIF
       !
       ALLOCATE(pairArray(pairArraySize + pairBlockSize))
       DO iPair=pairArraySize+1,pairArraySize+pairBlockSize
          CALL INITIALIZE_PAIR_STRUCTURE(pairArray(iPair))
       ENDDO
       !
       pairArray(1:pairArraySize) = tempPair(1:pairArraySize)
       IF (ALLOCATED(tempPair)) DEALLOCATE(tempPair)
    ENDIF
    !
    pairArraySize = pairArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_PAIR_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_PAIR_STRUCTURE(pairStructure)
    !------------------------------------------------------------------
    !---- Initialize pair structure
    !
    !---- IN
    !---- pairStructure : pair structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(pair_structure), INTENT(INOUT) :: pairStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize pair structure
    !------------------------------------------------------------------
    pairStructure%iElt = 0
    pairStructure%iDof = 0    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_PAIR_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_PLANE_ARRAY(planeArray, planeArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to planeArray
    !
    !---- IN
    !---- planeArray     : array of "plane" structure
    !---- planeArraySize : number of elements in planeArray
    !
    !---- OUT
    !---- planeArray     : modified in the process
    !---- planeArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(INOUT) :: planeArraySize
    
    TYPE(plane_structure), DIMENSION(:), ALLOCATABLE :: planeArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                 :: isArrayFull, isAllocated

    INTEGER(C_INT)                                  :: iPlane
    
    TYPE(plane_structure), DIMENSION(:), ALLOCATABLE :: tempPlane
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(planeArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(planeArray) < planeArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempPlane)) DEALLOCATE(tempPlane)
          ALLOCATE(tempPlane(planeArraySize))
          tempPlane = planeArray
          DEALLOCATE(planeArray)
       ENDIF
       !
       ALLOCATE(planeArray(planeArraySize + planeBlockSize))
       DO iPlane=planeArraySize+1,planeArraySize+planeBlockSize
          CALL INITIALIZE_PLANE_STRUCTURE(planeArray(iPlane))
       ENDDO
       !
       planeArray(1:planeArraySize) = tempPlane(1:planeArraySize)
       IF (ALLOCATED(tempPlane)) DEALLOCATE(tempPlane)
    ENDIF
    !
    planeArraySize = planeArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_PLANE_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_PLANE_STRUCTURE(planeStructure)
    !------------------------------------------------------------------
    !---- Initialize plane structure
    !
    !---- IN
    !---- planeStructure : plane structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(plane_structure), INTENT(INOUT) :: planeStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize plane structure
    !------------------------------------------------------------------
    planeStructure%recLevel    = 0
    planeStructure%sampling(:) = 0
    planeStructure%coords(:,:) = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_PLANE_STRUCTURE    

  !========================================================================
  !========================================================================
  !========================================================================

END MODULE FIELDS_VIEW
