MODULE GMSH_MESH

  !====================================================================
  ! Module containing the subroutines related to Gmsh mesh format
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_GMSH_MESH
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE IO_FILES
    USE MESH_ELEMENTS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR)      :: keyword
    CHARACTER(LEN=4,   KIND=C_CHAR)      :: type

    INTEGER(C_INT)                       :: eltTag, eltType, num, &
         nTags, otherTag, iVert, iElt, nElts, nVert, nUnit
    INTEGER(C_INT), DIMENSION(nvCellMax) :: iBuffer

    REAL(C_DOUBLE), DIMENSION(3)         :: rBuffer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Reading and filling Mesh arrays
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(Mesh%name)//'.msh')
    !
    Readingloop: DO
       READ(nUnit, *) keyword 
       !
       SELECT CASE (trim(keyword))
          !
          !---- Vertices
       CASE('$Nodes')
          READ(nUnit,*) nVert
          vertexBlockSize = nVert
          DO iVert=1,nVert
             READ(nUnit, *) num, rBuffer(1:3)
             CALL ADD_VERTEX(rBuffer(1:3))
          ENDDO
          !
          !---- Faces and Cells
       CASE('$Elements')
          READ(nUnit, *) nElts
          cellBlockSize = nElts/10
          faceBlockSize = nElts/10
          !
          DO iElt=1,nElts
             READ(nUnit,*) num, eltType, nTags, eltTag, otherTag, iBuffer(1:BUFFLENGTH(eltType))
             !
             CALL SORT_GMSH_TYPE(eltType, type)
             CALL ADD_ELEMENT(iBuffer(1:BUFFLENGTH(eltType)), eltTag, type)
          ENDDO
       CASE('$EndElements')
          EXIT Readingloop
       CASE DEFAULT
       END SELECT
    ENDDO Readingloop
    !
    CLOSE(nUnit)
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh reading'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_GMSH_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_GMSH_MESH
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_TFSF_FACES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Reset cell tags to initial values depending on zone
    !----------------------------------------------------------------------
    CALL PROCESS_TFSF_ZONES_BEFORE_WRITING
    !----------------------------------------------------------------------
    !---- Write
    !----------------------------------------------------------------------
    WRITE(6, '(a)') 'Writing mesh for GMSH format is not yet implemented'
    WRITE(6, '(a)') 'I wont crash but I will just not do it'
    !----------------------------------------------------------------------
    !-- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_GMSH_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SORT_GMSH_TYPE(eltType, type)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                INTENT(IN)  :: eltType

    CHARACTER(LEN=4, KIND=C_CHAR), INTENT(OUT) :: type
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !---- Check if type is known
    SELECT CASE(eltType)
    CASE(2,3,4,5,9,10,11,12)
       ! Do nothing
    CASE DEFAULT
       WRITE(6,'(a)') 'Error in routine SORT_GMSH_TYPE'
       WRITE(6,'(a)') 'Element type is unknown'
       CALL TILT
    END SELECT
    !------------------------------------------------------------------
    !---- Sort type
    !------------------------------------------------------------------
    SELECT CASE(eltType)
    CASE(2)  ! Linear triangle
       type =  'TRI1'
    CASE(9)  ! Curved triangle
       type =  'TRI2'
    CASE(3)  ! Linear quadrangle
       type =  'QUA1'
    CASE(10) ! Curved quadrangle
       type =  'QUA2'
    CASE(4)  ! Linear tetrahedron
       type =  'TET1'
    CASE(11) ! Curved tetrahedron
       type =  'TET2'
    CASE(5)  ! Linear hexahedron
       type =  'HEX1'
    CASE(12) ! Curved hexahedron
       type =  'HEX2'
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SORT_GMSH_TYPE

  !========================================================================
  !========================================================================
  !========================================================================

  FUNCTION BUFFLENGTH(eltType)
    !----------------------------------------------------------------------
    !---- Length of reading buffer for gmsh format
    !
    !---- IN
    !---- eltType : type of element
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE ISO_C_BINDING
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: eltType
    INTEGER(C_INT)             :: BUFFLENGTH
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Test element type
    !----------------------------------------------------------------------
    SELECT CASE(eltType)
    CASE(2)  ! Linear triangle
       BUFFLENGTH = 3
    CASE(3)  ! Linear quadrangle
       BUFFLENGTH = 4
    CASE(4)  ! Linear tetrahedron
       BUFFLENGTH = 4
    CASE(5)  ! Linear hexahedron
       BUFFLENGTH = 8
    CASE(9)  ! Curvilinear triangle
       BUFFLENGTH = 6
    CASE(10) ! Curvilinear quadrangle
       BUFFLENGTH = 9
    CASE(11) ! Curvilinear tetrahedron
       BUFFLENGTH = 10
    CASE(12) ! Curvilinear hexahedron
       BUFFLENGTH = 27
    END SELECT
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END FUNCTION BUFFLENGTH

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE GMSH_MESH
