MODULE GRAPH_STRUCTURES

  !====================================================================
  ! Module containing methods related to graph partitionning
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_DATA
  USE MESH_VERTICES
  USE MESH_FACES
  USE MESH_CELLS
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a node in a graph
  !
  ! nNeighbors     ==> number of other nodes connected to current node
  ! neighbors      ==> indices of these nodes
  !--------------------------------------------------------------------
  TYPE graph_node
     INTEGER(C_INT)                            :: nNeighb
     INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: neighb, edgeWeight
  END TYPE graph_node

  !--------------------------------------------------------------------
  ! Data structure associated to any submesh information
  !
  ! tab            ==> tab
  !--------------------------------------------------------------------
  TYPE submesh_info
     INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: tab
  END TYPE submesh_info

  !--------------------------------------------------------------------
  ! Data structure associated to the list of face shared with a 
  ! neighbor subdomain
  !
  ! tab            ==> tab
  !--------------------------------------------------------------------
  TYPE submesh_faces
     TYPE(submesh_info), DIMENSION(:), ALLOCATABLE :: faces, perm
  END TYPE submesh_faces
  
  !--------------------------------------------------------------------
  ! Data structure associated to a mesh dual graph
  ! In this structure, every cell is a node, and every face is an edge
  ! between two nodes
  !
  ! name           ==> graph name
  ! nNodes         ==> number of nodes in the graph
  ! nEdges         ==> number of edges in the graph
  ! n*PerPart      ==> number of each type per partition
  ! nodes          ==> nodes of the graph
  ! *PerPart       ==> indirections from local partition to global mesh
  !                    (ie *PerPart(localNum) = globalNum)
  ! globalIndir    ==> indirection of verticles from global numering
  !                    to local partitions
  !--------------------------------------------------------------------
  TYPE dual_graph
     CHARACTER(LEN=1000, KIND=C_CHAR)                 :: name
     
     INTEGER(C_INT)                                   :: nNodes, nEdges
     INTEGER(C_INT),      DIMENSION(:),   ALLOCATABLE :: partition,    &
          nVertsPerPart, nFacesPerPart, nCellsPerPart, nNeighbPerPart
     INTEGER(C_INT),      DIMENSION(:,:), ALLOCATABLE :: globalIndir,  &
          connectivity

     TYPE(graph_node),    DIMENSION(:),   ALLOCATABLE :: nodes

     TYPE(submesh_info),  DIMENSION(:),   ALLOCATABLE :: vertsPerPart, &
          facesPerPart, cellsPerPart, neighbPerPart,                   &
          nFacesPerNeighb

     TYPE(submesh_faces), DIMENSION(:),   ALLOCATABLE :: facesPerNeighb
  END TYPE dual_graph

  !--------------------------------------------------------------------
  ! Data for edges weights and others
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)           :: hybridCut   = .TRUE.
  LOGICAL(C_BOOL)           :: periodicCut = .TRUE.  
  
  INTEGER(C_INT), PARAMETER :: hybridEdgeWeight   = 99999
  INTEGER(C_INT), PARAMETER :: periodicEdgeWeight = 99999

  !--------------------------------------------------------------------
  ! Mesh dual graph
  !--------------------------------------------------------------------
  TYPE(dual_graph) :: Graph

  !--------------------------------------------------------------------
  ! Minimum amount of mesh structures for local copy
  !--------------------------------------------------------------------
  TYPE(mesh_structure)                          :: globalGraphMesh
  TYPE(vertex),       DIMENSION(:), ALLOCATABLE :: globalVert
  TYPE(face),         DIMENSION(:), ALLOCATABLE :: globalFac
  TYPE(derived_face), DIMENSION(:), ALLOCATABLE :: globalFacTriLin,  &
       globalFacTriQuadr, globalFacQuadLin, globalFacQuadQuadr
  TYPE(cell),         DIMENSION(:), ALLOCATABLE :: globalCell
  TYPE(derived_cell), DIMENSION(:), ALLOCATABLE :: globalCellTetLin, &
       globalCellTetQuadr, globalCellHexLin, globalCellHexQuadr
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MESH_TO_GRAPH
    !------------------------------------------------------------------
    !---- Build the dual graph of a mesh structure for partitionning
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, iFace, nEdges, iNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Complete informations for the graph
    !------------------------------------------------------------------
    Graph%name   = TRIM(Mesh%name)//'.dgraph'
    Graph%nNodes = Mesh%nCell
    Graph%nEdges = 0
    ALLOCATE(Graph%nodes(Graph%nNodes))
    !
    !---- Correct connectivity for periodic faces
    CALL CORRECT_PERIODIC_CONNECTIVITY
    !
    !---- Loop over graph nodes
    DO iCell=1,Mesh%nCell
       !---- Count the number of neighbors of this node and allocate structure
       nEdges = 0
       DO iNeighb=1,MeshCell(iCell)%nNeighb
          IF (MeshCell(iCell)%neighb(iNeighb) <= Mesh%nCell) nEdges = nEdges + 1
       ENDDO
       Graph%nodes(iCell)%nNeighb = nEdges
       ALLOCATE(Graph%nodes(iCell)%neighb    (nEdges))
       ALLOCATE(Graph%nodes(iCell)%edgeWeight(nEdges))
       Graph%nodes(iCell)%edgeWeight(:) = 1
       !
       !---- Increment the global number of edges
       Graph%nEdges = Graph%nEdges + nEdges
       !
       !---- Loop once again over neighbors of current node, and fill
       !---- the neighbors array
       nEdges = 0
       DO iNeighb=1,MeshCell(iCell)%nNeighb
          IF (MeshCell(iCell)%neighb(iNeighb) <= Mesh%nCell) THEN
             nEdges = nEdges + 1
             Graph%nodes(iCell)%neighb(nEdges) = MeshCell(iCell)%neighb(iNeighb)
          ENDIF
       ENDDO
    ENDDO
    !
    !---- Take into account that each edge was counted twice
    Graph%nEdges = Graph%nEdges/2
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished mesh to graph conversion'
    !------------------------------------------------------------------
  END SUBROUTINE MESH_TO_GRAPH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_EDGE_WEIGHTS
    !------------------------------------------------------------------
    !---- Process the weight of the graph edges
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)  :: iCell, nEdges, iNeighb, iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Process hybrid edges cutting
    !---- If hybridCut is set to FALSE, we set the weight of all
    !---- hybrid faces to a very high value to avoid cutting
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       nEdges = 0
       DO iNeighb=1,MeshCell(iCell)%nNeighb
          IF (MeshCell(iCell)%neighb(iNeighb) <= Mesh%nCell) THEN
             nEdges = nEdges + 1
             iFace  = MeshCell(iCell)%facNum(iNeighb)
             !
             !---- Hybrid edge
             IF (MeshFac(iFace)%isHybrid   .AND. (.NOT. hybridCut))   &
                  Graph%nodes(iCell)%edgeWeight(nEdges) = hybridEdgeWeight
             !
             !---- Periodic edge
             IF (MeshFac(iFace)%isPeriodic .AND. (.NOT. periodicCut)) &
                  Graph%nodes(iCell)%edgeWeight(nEdges) = periodicEdgeWeight
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished processing edge weights'
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_EDGE_WEIGHTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_GRAPH
    !------------------------------------------------------------------
    !---- Write graph in METIS format
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=3, KIND=C_CHAR) :: metisTag
    
    INTEGER(C_INT)                :: nUnit, iNode, iNeighb, nNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write graph
    !---- The binary tag for Metis works as follows :
    !---- 001 = weights on edges     (faces)
    !---- 010 = weights on vertices  (cells)
    !---- 100 = size of the vertices (cells)
    !---- and any combination of the latter    
    !------------------------------------------------------------------
    metisTag = '001'
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(Graph%name))
    !
    !---- Write header informations
    WRITE(nUnit, '(2i9,1x,a)') Graph%nNodes, Graph%nEdges, metisTag
    !
    !---- Write connectivity
    DO iNode=1,Graph%nNodes
       nNeighb = Graph%nodes(iNode)%nNeighb
       DO iNeighb=1,nNeighb-1
          WRITE(nUnit, '(2i9)', ADVANCE='NO')      &
               Graph%nodes(iNode)%neighb(iNeighb), &
               Graph%nodes(iNode)%edgeWeight(iNeighb)
       ENDDO
       WRITE(nUnit, '(2i9)')                    &
            Graph%nodes(iNode)%neighb(nNeighb), &
            Graph%nodes(iNode)%edgeWeight(nNeighb)
    ENDDO
    !
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished graph writing'
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_GRAPH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PARTITION_GRAPH(nDomains)
    !------------------------------------------------------------------
    !---- Partition a graph with METIS
    !
    !---- IN
    !---- nDomains : number of subDomains
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),       INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: command, cDomains
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Partition graph
    !------------------------------------------------------------------
    CALL INT_TO_CHAR(nDomains,cDomains)
    command = 'gpmetis '//TRIM(Graph%name)//' -contig '//TRIM(cDomains)    
    CALL SYSTEM(TRIM(command))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PARTITION_GRAPH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_PARTITION(nDomains)
    !------------------------------------------------------------------
    !---- Read METIS partition
    !
    !---- IN
    !---- nDomains : number of submeshes
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),       INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: fileName, cDomains
    
    INTEGER(C_INT)                   :: nUnit, iNode
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read partition
    !------------------------------------------------------------------
    !---- Convert int to char
    CALL INT_TO_CHAR(nDomains,cDomains)
    !
    !---- Allocate partition array
    ALLOCATE(Graph%partition(Graph%nNodes))
    !
    !---- Open file
    CALL GET_NEW_FILE_UNIT(nUnit)
    fileName = TRIM(Graph%name)//'.part.'//TRIM(cDomains)
    OPEN(UNIT=nUnit, FILE=TRIM(fileName))
    !
    !---- Read partition and close file
    DO iNode=1,Graph%nNodes
       READ(nUnit, *) Graph%partition(iNode)
    ENDDO
    !
    !---- This stupid numbering starts at 0, so correct it
    Graph%partition(:) = Graph%partition(:) + 1
    !
    CLOSE(nUnit)    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished partition reading'
    !------------------------------------------------------------------
  END SUBROUTINE READ_PARTITION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_SUBMESHES(nDomains)
    !------------------------------------------------------------------
    !---- Build submeshes after partition was read
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_ELEMENTS
    USE IO_FILES
    USE MEDIT_MESH
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),    INTENT(INOUT) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: subName, cDomain
    
    INTEGER(C_INT)                   :: iPart, iNode, iCell, iVert, &
         iFac, globalIndex, cellNum, faceNum
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build indirections and copy global mesh
    !------------------------------------------------------------------
    CALL BUILD_CELL_INDIRECTIONS(nDomains)
    CALL BUILD_VERT_INDIRECTIONS(nDomains)
    CALL BUILD_FACE_INDIRECTIONS(nDomains)
    CALL COPY_MESH
    !------------------------------------------------------------------
    !---- Adjust the allocation block sizes
    !------------------------------------------------------------------
    cellBlockSize   = cellBlockSize/nDomains
    dCellBlockSize  = cellBlockSize
    faceBlockSize   = faceBlockSize/nDomains
    dFaceBlockSize  = faceBlockSize
    cFaceBlockSize  = faceBlockSize
    vertexBlockSize = vertexBlockSize/nDomains
    !------------------------------------------------------------------
    !---- Build submeshes and write them
    !------------------------------------------------------------------
    DO iPart=1,nDomains
       !---- Reset Mesh counters and set mesh name
       CALL INITIALIZE_MESH_STRUCTURE(Mesh)
       CALL DEALLOCATE_MESH_STRUCTURES
       CALL BUILD_SUBMESH_NAME(globalMesh%name,iPart,subName)
       Mesh%name = TRIM(subName)
       !
       CALL ADD_VERTICES_TO_SUBMESH(iPart)
       CALL ADD_CELLS_TO_SUBMESH(iPart)
       CALL ADD_FACES_TO_SUBMESH(iPart)
       CALL BUILD_DERIVED_CELLS
       CALL BUILD_DERIVED_FACES
       CALL CHECK_FOR_ARTIFICIAL_HYBRID_FACES(iPart)
       !
       CALL WRITE_MEDIT_MESH
       CALL DEALLOCATE_MESH_STRUCTURES
    ENDDO
    !------------------------------------------------------------------
    !---- Build additional informations on submeshes
    !------------------------------------------------------------------
    CALL FIND_SUBMESHES_NEIGHBORS(nDomains)
    CALL FIND_SUBMESHES_COMMON_FACES(nDomains)
    CALL WRITE_SUBMESHES_INFORMATION(nDomains)
    !------------------------------------------------------------------
    !---- Deallocate everyone
    !------------------------------------------------------------------
    CALL DEALLOCATE_GLOBAL_MESH
    CALL DEALLOCATE_GRAPH
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished building submeshes'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_SUBMESHES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COPY_MESH
    !------------------------------------------------------------------
    !---- Create a local copy of Mesh
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Copy global mesh
    !------------------------------------------------------------------
    CALL BUILD_DERIVED_CELLS
    CALL BUILD_DERIVED_FACES
    globalGraphMesh = Mesh
    !
    IF (ALLOCATED(globalVert))         DEALLOCATE(globalVert)
    IF (ALLOCATED(MeshVert))           ALLOCATE(globalVert(Mesh%nVert))
    IF (ALLOCATED(MeshVert))           globalVert         = MeshVert
    !
    IF (ALLOCATED(globalFac))          DEALLOCATE(globalFac)
    IF (ALLOCATED(MeshFac))            ALLOCATE(globalFac(Mesh%nFac))
    IF (ALLOCATED(MeshFac))            globalFac          = MeshFac
    !
    IF (ALLOCATED(globalFacTriLin))    DEALLOCATE(globalFacTriLin)
    IF (ALLOCATED(MeshFacTriLin))      ALLOCATE(globalFacTriLin(Mesh%nFacTriLin))
    IF (ALLOCATED(MeshFacTriLin))      globalFacTriLin    = MeshFacTriLin
    !
    IF (ALLOCATED(globalFacTriQuadr))  DEALLOCATE(globalFacTriQuadr)
    IF (ALLOCATED(MeshFacTriQuadr))    ALLOCATE(globalFacTriQuadr(Mesh%nFacTriQuadr))
    IF (ALLOCATED(MeshFacTriQuadr))    globalFacTriQuadr  = MeshFacTriQuadr
    !
    IF (ALLOCATED(globalFacQuadLin))   DEALLOCATE(globalFacQuadLin)
    IF (ALLOCATED(MeshFacQuadLin))     ALLOCATE(globalFacQuadLin(Mesh%nFacQuadLin))
    IF (ALLOCATED(MeshFacQuadLin))     globalFacQuadLin   = MeshFacQuadLin
    !
    IF (ALLOCATED(globalFacQuadQuadr)) DEALLOCATE(globalFacQuadQuadr)
    IF (ALLOCATED(MeshFacQuadQuadr))   ALLOCATE(globalFacQuadQuadr(Mesh%nFacQuadQuadr))
    IF (ALLOCATED(MeshFacQuadQuadr))   globalFacQuadQuadr = MeshFacQuadQuadr
    !
    IF (ALLOCATED(globalCell))         DEALLOCATE(globalCell)
    IF (ALLOCATED(MeshCell))           ALLOCATE(globalCell(Mesh%nCell))
    IF (ALLOCATED(MeshCell))           globalCell         = MeshCell
    !
    IF (ALLOCATED(globalCellTetLin))   DEALLOCATE(globalCellTetLin)
    IF (ALLOCATED(MeshCellTetLin))     ALLOCATE(globalCellTetLin(Mesh%nCellTetLin))
    IF (ALLOCATED(MeshCellTetLin))     globalCellTetLin   = MeshCellTetLin
    !
    IF (ALLOCATED(globalCellTetQuadr)) DEALLOCATE(globalCellTetQuadr)
    IF (ALLOCATED(MeshCellTetQuadr))   ALLOCATE(globalCellTetQuadr(Mesh%nCellTetQuadr))
    IF (ALLOCATED(MeshCellTetQuadr))   globalCellTetQuadr = MeshCellTetQuadr
    !
    IF (ALLOCATED(globalCellHexLin))   DEALLOCATE(globalCellHexLin)
    IF (ALLOCATED(MeshCellHexLin))     ALLOCATE(globalCellHexLin(Mesh%nCellHexLin))
    IF (ALLOCATED(MeshCellHexLin))     globalCellHexLin   = MeshCellHexLin
    !
    IF (ALLOCATED(globalCellHexQuadr)) DEALLOCATE(globalCellHexQuadr)
    IF (ALLOCATED(MeshCellHexQuadr))   ALLOCATE(globalCellHexQuadr(Mesh%nCellHexQuadr))
    IF (ALLOCATED(MeshCellHexQuadr))   globalCellHexQuadr = MeshCellHexQuadr
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished mesh copying'
    !------------------------------------------------------------------
  END SUBROUTINE COPY_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_CELL_INDIRECTIONS(nDomains)
    !------------------------------------------------------------------
    !---- Build cell lists from partitioned graph
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iNode, iPart
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build cell indirections
    !------------------------------------------------------------------
    !---- Allocate array counting the number of cells per part
    ALLOCATE(Graph%nCellsPerPart(nDomains))
    Graph%nCellsPerPart(:) = 0
    !
    !---- Count the number of nodes per partition
    DO iNode=1,Graph%nNodes
       iPart                      = Graph%partition(iNode)
       Graph%nCellsPerPart(iPart) = Graph%nCellsPerPart(iPart) + 1
    ENDDO
    !
    !---- Allocate indirection arrays
    ALLOCATE(Graph%cellsPerPart(nDomains))
    DO iPart=1,nDomains
       ALLOCATE(Graph%cellsPerPart(iPart)%tab(Graph%nCellsPerPart(iPart)))
    ENDDO
    Graph%nCellsPerPart(:) = 0
    !
    !---- Fill indirection arrays
    DO iNode=1,Graph%nNodes
       iPart                                                     = Graph%partition(iNode)
       Graph%nCellsPerPart(iPart)                                = Graph%nCellsPerPart(iPart) + 1
       Graph%cellsPerPart(iPart)%tab(Graph%nCellsPerPart(iPart)) = iNode
    ENDDO    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished submeshes cells building'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_CELL_INDIRECTIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_VERT_INDIRECTIONS(nDomains)
    !------------------------------------------------------------------
    !---- Build vertices lists from partitioned graph
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE ARRAY_OPERATIONS    
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)             :: cellType
    
    INTEGER(C_INT)                            :: iPart, numCell,  &
         nGenVerts, nVerts, nLinVerts, nHoVerts, geoOrder, iCell, &
         iVert, iNeighb, nNeighb, numFace
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: verts, order
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build vertices indirections
    !------------------------------------------------------------------
    !---- Allocate array counting the number of vertices per part
    ALLOCATE(Graph%nVertsPerPart(nDomains))
    ALLOCATE(Graph%vertsPerPart (nDomains))
    ALLOCATE(Graph%globalIndir  (Mesh%nVert,nDomains))
    !
    Graph%nVertsPerPart(:) = 0
    Graph%globalIndir(:,:) = 0
    !
    !---- Loop over domains
    DO iPart=1,nDomains
       !---- Initiliaze stuff
       nGenVerts = 0
       IF (ALLOCATED(verts)) DEALLOCATE(verts)
       IF (ALLOCATED(order)) DEALLOCATE(order)
       !
       DO iCell=1,Graph%nCellsPerPart(iPart)
          numCell = Graph%cellsPerpart(iPart)%tab(iCell)
          nNeighb = MeshCell(numCell)%nNeighb
          !
          DO iNeighb=1,nNeighb
             numFace = MeshCell(numCell)%facNum(iNeighb)
             nVerts  = MeshFac(numFace)%nVerts
             DO iVert=1,nVerts
                CALL EXTEND_INTEGER_ARRAY(verts, nGenVerts, Mesh%nVert/nDomains)
                verts(nGenVerts) = MeshFac(numFace)%vertex(iVert)
             ENDDO
          ENDDO
       ENDDO
       !
       !---- Find duplicates
       ALLOCATE(order(SIZE(verts)))       
       CALL FIND_ARRAY_DUPLICATES(verts(1:nGenVerts), nGenVerts, &
            order(1:nGenVerts), nVerts)
       Graph%nVertsPerPart(iPart) = nVerts
       ALLOCATE(Graph%vertsPerPart(iPart)%tab(nVerts))
       Graph%vertsPerPart(iPart)%tab(1:nVerts) = verts(1:nVerts)
       !
       !---- Fill global to local indirection
       DO iVert=1,nVerts
          Graph%globalIndir(verts(iVert),iPart) = iVert
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (ALLOCATED(verts)) DEALLOCATE(verts)
    IF (ALLOCATED(order)) DEALLOCATE(order)
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished submeshes vertices building'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_VERT_INDIRECTIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_FACE_INDIRECTIONS(nDomains)
    !------------------------------------------------------------------
    !---- Build face lists from partitioned graph
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE ARRAY_OPERATIONS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                            :: countFace, isArtificial
    LOGICAL(C_BOOL), DIMENSION(:), ALLOCATABLE :: artificialFaces
    
    INTEGER(C_INT)             :: iPart, nFaces, iFac, nVerts, iVert, &
         log, vert, iCell1, iCell2, iPart1, iPart2, numLocal, iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build face indirections
    !------------------------------------------------------------------
    !---- Allocate array counting the number of faces per part
    ALLOCATE(Graph%nFacesPerPart(nDomains))
    ALLOCATE(Graph%facesPerPart (nDomains))
    ALLOCATE(artificialFaces(Mesh%nFac))
    artificialFaces(:)     = .FALSE.
    Graph%nFacesPerPart(:) = 0
    !
    !---- Loop over domains
    DO iPart=1,nDomains
       !---- Fill faces array
       DO iFac=1,Mesh%nFac
          !---- Sort face
          CALL SORT_GRAPH_FACE(iFac, iPart, countFace, isArtificial)
          IF (.NOT. countFace) CYCLE
          !
          !---- Tag artificial faces
          IF (isArtificial) artificialFaces(iFac) = .TRUE.
          !
          !---- Count face
          CALL EXTEND_INTEGER_ARRAY(Graph%facesPerPart(iPart)%tab, &
               Graph%nFacesPerPart(iPart), MAX(10,Mesh%nFacB))
          Graph%facesPerPart(iPart)%tab(Graph%nFacesPerPart(iPart)) = iFac
       ENDDO
    ENDDO
    !
    !---- Modify artificial faces
    DO iFac=1,Mesh%nFac
       IF (.NOT. artificialFaces(iFac)) CYCLE
       MeshFac(iFac)%isArtificial = .TRUE.
       CALL MODIFY_INTEGER(MeshFac(iFac)%log, 2, 1)
       CALL MODIFY_INTEGER(MeshFac(iFac)%log, 1, 0)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    DEALLOCATE(artificialFaces)
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished submeshes faces building'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_FACE_INDIRECTIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SORT_GRAPH_FACE(iFac, iPart, countFace, isArtificial)
    !------------------------------------------------------------------
    !---- Check if a face is inside a given subdomain, determine
    !---- wether it must be added or not, and if it is an artificial
    !---- face or not
    !
    !---- IN
    !---- iFac         : global index of the face
    !---- iPart        : global index of the subdomain
    !
    !---- OUT
    !---- countFace    : TRUE if face must be counted
    !---- isArtificial : TRUE if face is artificial
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE ARRAY_OPERATIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL), INTENT(OUT) :: countFace, isArtificial
    
    INTEGER(C_INT),  INTENT(IN)  :: iFac, iPart
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)              :: isFound
    
    INTEGER(C_INT)               :: nVerts, iVert, vert, iCell1, &
         iCell2, iPart1, iPart2, nLinVerts, iFacPer, iFacMatch
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Sort face
    !------------------------------------------------------------------
    countFace    = .FALSE.
    isArtificial = .FALSE.
    !
    !---- Discriminate the case where the vertices are in the
    !---- subdomain, but the face isn't
    iCell1 = MeshFac(iFac)%neighb(1)
    iCell2 = MeshFac(iFac)%neighb(2)    
    IF (iCell1 > Mesh%nCell) iCell1 = iCell2 ! Boundary case
    IF (iCell2 > Mesh%nCell) iCell2 = iCell1 ! Boundary case
    iPart1 = Graph%partition(iCell1)
    iPart2 = Graph%partition(iCell2)
    IF ((iPart1 /= iPart)          .AND. (iPart2 /= iPart)) RETURN
    IF ((MeshFac(iFac)%isBoundary) .AND. (iPart1 /= iPart)) RETURN
    !
    !---- Check if it is an actual artificial face
    IF ((iPart1 /= iPart) .OR. (iPart2 /= iPart)) THEN
       countFace    = .TRUE.
       isArtificial = .TRUE.
       RETURN
    ENDIF    
    !
    !---- Check if it is a boundary, hybrid or TFSF face
    IF ( (MeshFac(iFac)%isBoundary) .OR. &
         (MeshFac(iFac)%isHybrid)   .OR. &
         (MeshFac(iFac)%isTfsf))    THEN
       countFace = .TRUE.
       RETURN
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SORT_GRAPH_FACE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_VERTICES_TO_SUBMESH(iPart)
    !------------------------------------------------------------------
    !---- Add vertices to submesh
    !
    !---- IN
    !---- iPart : index of submesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iPart
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iVert, globalIndex
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Add vertices
    !------------------------------------------------------------------
    DO iVert=1,Graph%nVertsPerPart(iPart)
       globalIndex = Graph%vertsPerPart(iPart)%tab(iVert)
       CALL ADD_VERTEX(globalVert(globalIndex)%coor(:))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_VERTICES_TO_SUBMESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_CELLS_TO_SUBMESH(iPart)
    !------------------------------------------------------------------
    !---- Add cells to submesh
    !
    !---- IN
    !---- iPart : index of submesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iPart
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iCell, cellNum, log
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Add cells
    !------------------------------------------------------------------
    DO iCell=1,Graph%nCellsPerPart(iPart)
       cellNum = Graph%cellsPerPart(iPart)%tab(iCell)
       CALL GET_GLOBAL_TFSF_MATLOG(cellNum, log)
       CALL ADD_ELEMENT(Graph%globalIndir(globalCell(cellNum)%vertex,iPart), &
            log, globalCell(cellNum)%cellType)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_CELLS_TO_SUBMESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_GLOBAL_TFSF_MATLOG(iCell, log)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: iCell
    INTEGER(C_INT), INTENT(OUT) :: log
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute log including TFSF zone
    !------------------------------------------------------------------
    log = 0
    !
    SELECT CASE(globalCell(iCell)%zone)
    CASE('SF')
       log = globalCell(iCell)%log + 1000
    CASE('TF')
       log = globalCell(iCell)%log
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_GLOBAL_TFSF_MATLOG

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_FACES_TO_SUBMESH(iPart)
    !------------------------------------------------------------------
    !---- Add faces to submesh
    !
    !---- IN
    !---- iPart : index of submesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iPart
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iFace, faceNum
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Add faces
    !------------------------------------------------------------------
    DO iFace=1,Graph%nFacesPerPart(iPart)
       faceNum = Graph%facesPerPart(iPart)%tab(iFace)
       CALL ADD_ELEMENT(Graph%globalIndir(globalFac(faceNum)%vertex,iPart), &
            globalFac(faceNum)%log, globalFac(faceNum)%faceType)
    ENDDO
    !
    CALL BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_FACES_TO_SUBMESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_SUBMESHES_NEIGHBORS(nDomains)
    !------------------------------------------------------------------
    !---- Find submeshes neighbors
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_CELLS
    USE ARRAY_OPERATIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iPart, iCell, nFaces, iFace, cellNum, &
         cellNeighb, neighbPart, iNeighb, nNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Find neighbors
    !------------------------------------------------------------------
    ALLOCATE(Graph%nNeighbPerPart(nDomains))
    ALLOCATE(Graph%neighbPerPart(nDomains))    
    ALLOCATE(Graph%connectivity(nDomains,nDomains))
    Graph%nNeighbPerPart = 0
    Graph%connectivity   = 0
    !
    !---- Loop over domains
    DO iPart=1,nDomains
       !---- Loop over cells of current subdomain and look for neighbors
       DO iCell=1,Graph%nCellsPerPart(iPart)
          cellNum = Graph%cellsPerPart(iPart)%tab(iCell)
          nNeighb = globalCell(cellNum)%nNeighb
          !
          DO iNeighb=1,nNeighb
             cellNeighb = globalCell(cellNum)%neighb(iNeighb)
             IF (cellNeighb > globalMesh%nCell) CYCLE ! Boundary face
             neighbPart = Graph%partition(cellNeighb)
             !
             !---- Update connectivity array 
             IF (neighbPart /= iPart) Graph%connectivity(iPart,neighbPart) &
                  = Graph%connectivity(iPart,neighbPart) + 1
          ENDDO
       ENDDO
       !
       !---- Count number of submesh neighbors
       DO iNeighb=1,nDomains
          IF (Graph%connectivity(iPart,iNeighb) /= 0) THEN
             CALL EXTEND_INTEGER_ARRAY(Graph%neighbPerPart(iPart)%tab, &
                  Graph%nNeighbPerPart(iPart), nDomains)
             Graph%neighbPerPart(iPart)%tab(Graph%nNeighbPerPart(iPart)) = iNeighb
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished finding submeshes neighbors'
    !------------------------------------------------------------------
  END SUBROUTINE FIND_SUBMESHES_NEIGHBORS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_SUBMESHES_COMMON_FACES(nDomains)
    !------------------------------------------------------------------
    !---- Find common faces with neighbor subdomains
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_CELLS
    USE MESH_PERIODIC_FACES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                           :: foundFace
    INTEGER(C_INT)                            :: iPart, nFaces, nNeighb, &
         neighbNum, iNeighb, iFace, faceNum, iSide, perm, iLocNeighb
    INTEGER(C_INT), DIMENSION(2)              :: iCell, iSub
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: nCountFaces
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate the right number of common faces per neighbor 
    !---- This information is still contained in the connectivity
    !---- array, we are just going to extract it properly
    !------------------------------------------------------------------
    ALLOCATE(Graph%nFacesPerNeighb(nDomains))
    !
    !---- Loop over domains
    DO iPart=1,nDomains
       nNeighb = Graph%nNeighbPerPart(iPart)
       ALLOCATE(Graph%nFacesPerNeighb(iPart)%tab(nNeighb))
       !
       DO iNeighb=1,nNeighb ! This is a local neighbor numbering
          neighbNum = Graph%neighbPerPart(iPart)%tab(iNeighb) ! This is a global numbering
          nFaces    = Graph%connectivity(iPart,neighbNum)
          Graph%nFacesPerNeighb(iPart)%tab(iNeighb) = nFaces
       ENDDO
    ENDDO
    !
    IF (ALLOCATED(Graph%connectivity)) DEALLOCATE(Graph%connectivity)
    !------------------------------------------------------------------
    !---- Allocate the list of faces per submesh neighbor
    !------------------------------------------------------------------
    ALLOCATE(Graph%facesPerNeighb(nDomains))
    !
    !---- Loop over domains
    DO iPart=1,nDomains
       nNeighb = Graph%nNeighbPerPart(iPart)
       ALLOCATE(Graph%facesPerNeighb(iPart)%faces(nNeighb))
       ALLOCATE(Graph%facesPerNeighb(iPart)%perm(nNeighb))
       !
       !---- Loop over neighbors of each submesh
       DO iNeighb=1,nNeighb
          nFaces = Graph%nFacesPerNeighb(iPart)%tab(iNeighb)
          ALLOCATE(Graph%facesPerNeighb(iPart)%faces(iNeighb)%tab(nFaces))
          ALLOCATE(Graph%facesPerNeighb(iPart)%perm(iNeighb)%tab(nFaces))
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Fill the list of faces per submesh neighbor
    !------------------------------------------------------------------
    DO iPart=1,nDomains
       IF (ALLOCATED(nCountFaces)) DEALLOCATE(nCountFaces)
       ALLOCATE(nCountFaces(Graph%nNeighbPerPart(iPart)))
       nCountFaces(:) = 0
       !
       DO iFace=1,Graph%nFacesPerPart(iPart)
          faceNum = Graph%facesPerPart(iPart)%tab(iFace)
          !
          !---- Drop if it is not an artificial face
          IF (.NOT. globalFac(faceNum)%isArtificial) CYCLE
          !
          !---- Find the global index of the neighbor
          iCell(:) = globalFac(faceNum)%neighb(:)
          iSub(:)  = Graph%partition(iCell(:))
          iSide    = 0
          IF (iSub(1) == iPart) iSide = 2
          IF (iSub(2) == iPart) iSide = 1
          !
          !---- Find the neighbor index in the local submesh numbering
          DO iNeighb=1,Graph%nNeighbPerPart(iPart)
             neighbNum = Graph%neighbPerPart(iPart)%tab(iNeighb)
             IF (iSub(iSide) == neighbNum) iLocNeighb = iNeighb
          ENDDO
          !
          !---- Increment local counter and fill arrays
          perm                    = globalFac(faceNum)%perm(iSide)
          nCountFaces(iLocNeighb) = nCountFaces(iLocNeighb) + 1
          nFaces                  = nCountFaces(iLocNeighb) 
          Graph%facesPerNeighb(iPart)%faces(iLocNeighb)%tab(nFaces) = faceNum
          Graph%facesPerNeighb(iPart)%perm(iLocNeighb)%tab(nFaces) = perm
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- As a convention, matching artificial faces of any kind must
    !---- be ordered in the same way in their respective submeshes.
    !---- This might not be the case for periodic artificial faces
    !---- because of their local numbering in their respective submeshes.
    !---- The following routine fixes this problem
    !------------------------------------------------------------------
    CALL FIX_ARTIFICIAL_FACES_LISTS(nDomains)
    !------------------------------------------------------------------
    IF (ALLOCATED(nCountFaces)) DEALLOCATE(nCountFaces)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished finding submeshes common faces'
    !------------------------------------------------------------------
  END SUBROUTINE FIND_SUBMESHES_COMMON_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIX_ARTIFICIAL_FACES_LISTS(nDomains)
    !------------------------------------------------------------------
    !---- Fix the the artificial faces lists numbering
    !---- This is problem is specific to periodic artificial faces
    !
    !---- IN
    !---- nDomains : # of subMeshes
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_CELLS
    USE REFERENCE_CELLS
    USE MESH_PERIODIC_FACES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)            :: foundFace
    INTEGER(C_INT)             :: iPart1, iPart2, iNeighb1, iNeighb2, &
         iFace1, iFace2, faceNum1, faceNum2, iComp, iDir, faceIndex,  &
         permIndex, permNum2
    INTEGER(C_INT), DIMENSION(nvFaceMax) :: globVert, neighbVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check that artificial faces properly match
    !------------------------------------------------------------------
    DO iPart1=1,nDomains ! Loop on submeshes
       DO iNeighb1=1,Graph%nNeighbPerPart(iPart1) ! Loop on neighbors
          !---- Find the neighbor's global submesh index
          iPart2 = Graph%neighbPerPart(iPart1)%tab(iNeighb1)
          !
          DO iFace1=1,Graph%nFacesPerNeighb(iPart1)%tab(iNeighb1) ! Loop on faces in common
             foundFace = .FALSE.
             !
             !---- Retrieve vertices indices in the neighbor numbering
             faceNum1 = Graph%facesPerNeighb(iPart1)%faces(iNeighb1)%tab(iFace1)
             IF (.NOT. globalFac(faceNum1)%isArtificial) CYCLE
             IF (.NOT. globalFac(faceNum1)%isPeriodic)   CYCLE
             !
             !---- Find the local index of current submesh in 
             !---- the neighbor's neighbors list
             DO iNeighb2=1,Graph%nNeighbPerPart(iPart2)
                IF (Graph%neighbPerPart(iPart2)%tab(iNeighb2) == iPart1) EXIT
             ENDDO
             !
             !---- If the face was found, check if it is a parallel periodic face
             DO iFace2=1,Graph%nFacesPerNeighb(iPart2)%tab(iNeighb2)
                faceNum2 = Graph%facesPerNeighb(iPart2)%faces(iNeighb2)%tab(iFace2)
                permNum2 = Graph%facesPerNeighb(iPart2)%perm(iNeighb2)%tab(iFace2)
                !
                CALL DO_PERIODIC_FACES_MATCH(globalFac(faceNum1), globalFac(faceNum2), &
                     globalVert, foundFace, iDir, iComp)
                IF ((foundFace) .AND. (iFace1 /= iFace2)) THEN
                   faceIndex = Graph%facesPerNeighb(iPart2)%faces(iNeighb2)%tab(iFace1)
                   Graph%facesPerNeighb(iPart2)%faces(iNeighb2)%tab(iFace1) = faceNum2
                   Graph%facesPerNeighb(iPart2)%faces(iNeighb2)%tab(iFace2) = faceIndex
                   !
                   permIndex = Graph%facesPerNeighb(iPart2)%perm(iNeighb2)%tab(iFace1)
                   Graph%facesPerNeighb(iPart2)%perm(iNeighb2)%tab(iFace1) = permNum2
                   Graph%facesPerNeighb(iPart2)%perm(iNeighb2)%tab(iFace2) = permIndex
                   EXIT
                ENDIF
             ENDDO
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FIX_ARTIFICIAL_FACES_LISTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_SUBMESHES_INFORMATION(nDomains)
    !------------------------------------------------------------------
    !---- Write additional information about submeshes
    !
    !---- IN
    !---- nDomains : number of submeshes    
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),       INTENT(IN) :: nDomains
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: fileName, cDomain
    CHARACTER(LEN=4,    KIND=C_CHAR) :: faceType    

    INTEGER(C_INT)                   :: iPart, iNeighb, iFace,  nUnit, &
         faceNum, cellNum, cellLog, iCell1, iCell2, iCell, nVerts,     &
         nLinVerts, nHoVerts, geoOrder, iVert, vertIndex, iPart2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write additional information
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    !
    DO iPart=1,nDomains
       CALL BUILD_SUBINFO_NAME(globalMesh%name,iPart,fileName)
       OPEN(UNIT=nUnit, FILE=TRIM(fileName))
       !
       !---- Write number of neighbors
       WRITE(nUnit, '(a)')  '#nNeighbors'
       WRITE(nUnit, '(i9)') Graph%nNeighbPerPart(iPart)
       WRITE(nUnit, '(a)')  ' '
       !
       !---- Write list of neighbors       
       WRITE(nUnit, '(a)')  '#neighbors'
       DO iNeighb=1,Graph%nNeighbPerPart(iPart)
          WRITE(nUnit, '(i9)') Graph%neighbPerPart(iPart)%tab(iNeighb)
       ENDDO       
       WRITE(nUnit, '(a)')  ' '
       !
       !---- Write number of common faces with each neighbor
       WRITE(nUnit, '(a)')  '#nFacesPerNeighbor'
       DO iNeighb=1,Graph%nNeighbPerPart(iPart)
          WRITE(nUnit, '(i9)') Graph%nFacesPerNeighb(iPart)%tab(iNeighb)
       ENDDO       
       WRITE(nUnit, '(a)')  ' '
       !
       !---- Write the vertices of the common faces with each neighbor
       !---- We cannot write just the face indices, because their order
       !---- can be modified by the writing/reading process
       !---- Instead, we write the face vertices
       WRITE(nUnit, '(a)')  '#facesPerNeighbor'
       DO iNeighb=1,Graph%nNeighbPerPart(iPart)
          DO iFace=1,Graph%nFacesPerNeighb(iPart)%tab(iNeighb)
             faceNum  = Graph%facesPerNeighb(iPart)%faces(iNeighb)%tab(iFace)
             faceType = globalFac(faceNum)%faceType
             CALL FACE_TYPE_TO_VERTS(faceType, nVerts, nLinVerts, nHoVerts, geoOrder)
             !
             WRITE(nUnit, '(i9)'  ) nLinVerts
             WRITE(nUnit, '(99i9)') Graph%globalIndir(globalFac(faceNum)%vertex(1:nLinVerts),iPart)
          ENDDO
       ENDDO       
       WRITE(nUnit, '(a)')  ' '
       !
       !---- Write list of permutations with each neighbor
       WRITE(nUnit, '(a)')  '#facesPermPerNeighbor'
       DO iNeighb=1,Graph%nNeighbPerPart(iPart)
          DO iFace=1,Graph%nFacesPerNeighb(iPart)%tab(iNeighb)
             WRITE(nUnit, '(i9)') Graph%facesPerNeighb(iPart)%perm(iNeighb)%tab(iFace)
          ENDDO
       ENDDO
       WRITE(nUnit, '(a)')  ' '
       !
       !---- Write list of adjacent cell logs
       WRITE(nUnit, '(a)')  '#cellLogsPerNeighbor'
       DO iNeighb=1,Graph%nNeighbPerPart(iPart)
          DO iFace=1,Graph%nFacesPerNeighb(iPart)%tab(iNeighb)
             faceNum = Graph%facesPerNeighb(iPart)%faces(iNeighb)%tab(iFace)
             iCell1  = globalFac(faceNum)%neighb(1)
             iCell2  = globalFac(faceNum)%neighb(2)
             !
             IF (Graph%partition(iCell1) == iPart) cellNum = iCell2
             IF (Graph%partition(iCell2) == iPart) cellNum = iCell1
             !
             cellLog = globalCell(cellNum)%log
             WRITE(nUnit, '(i9)') cellLog
          ENDDO
       ENDDO
       WRITE(nUnit, '(a)')  ' '       
       !
       !---- Write global cell indices for future visualization purpose
       WRITE(nUnit, '(a)')  '#globalCellNeighbors'
       DO iCell=1,Graph%nCellsPerPart(iPart)
          cellNum = Graph%cellsPerPart(iPart)%tab(iCell)
          WRITE(nUnit, '(99i9)') cellNum, globalCell(cellNum)%neighb(:)
       ENDDO
       !
       !---- End of file
       WRITE(nUnit, '(a)')  '#End'
       !
       CLOSE(nUnit)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a)') 'Finished writing submeshes informations'
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_SUBMESHES_INFORMATION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_GLOBAL_MESH
    !------------------------------------------------------------------
    !---- Deallocate cells
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate global mesh
    !------------------------------------------------------------------
    IF (ALLOCATED(globalVert))         DEALLOCATE(globalVert)
    IF (ALLOCATED(globalFac))          DEALLOCATE(globalFac)
    IF (ALLOCATED(globalFacTriLin))    DEALLOCATE(globalFacTriLin)
    IF (ALLOCATED(globalFacTriQuadr))  DEALLOCATE(globalFacTriQuadr)
    IF (ALLOCATED(globalFacQuadLin))   DEALLOCATE(globalFacQuadLin)
    IF (ALLOCATED(globalFacQuadQuadr)) DEALLOCATE(globalFacQuadQuadr)
    IF (ALLOCATED(globalCell))         DEALLOCATE(globalCell)
    IF (ALLOCATED(globalCellTetLin))   DEALLOCATE(globalCellTetLin)
    IF (ALLOCATED(globalCellTetQuadr)) DEALLOCATE(globalCellTetQuadr)
    IF (ALLOCATED(globalCellHexLin))   DEALLOCATE(globalCellHexLin)
    IF (ALLOCATED(globalCellHexQuadr)) DEALLOCATE(globalCellHexQuadr)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_GLOBAL_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_GRAPH
    !------------------------------------------------------------------
    !---- Deallocate graph structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate graph structures
    !------------------------------------------------------------------
    IF (ALLOCATED(graph%partition))       DEALLOCATE(graph%partition)
    IF (ALLOCATED(graph%nVertsPerPart))   DEALLOCATE(graph%nVertsPerPart)
    IF (ALLOCATED(graph%vertsPerPart))    DEALLOCATE(graph%vertsPerPart)
    IF (ALLOCATED(graph%nFacesPerPart))   DEALLOCATE(graph%nFacesPerpart)
    IF (ALLOCATED(graph%nCellsPerpart))   DEALLOCATE(graph%nCellsPerPart)
    IF (ALLOCATED(graph%nNeighbPerPart))  DEALLOCATE(graph%nNeighbPerPart)
    IF (ALLOCATED(graph%globalIndir))     DEALLOCATE(graph%globalIndir)
    IF (ALLOCATED(graph%connectivity))    DEALLOCATE(graph%connectivity)
    IF (ALLOCATED(graph%nodes))           DEALLOCATE(graph%nodes)
    IF (ALLOCATED(graph%vertsPerPart))    DEALLOCATE(graph%vertsPerPart)
    IF (ALLOCATED(graph%facesPerPart))    DEALLOCATE(graph%facesPerPart)
    IF (ALLOCATED(graph%cellsPerPart))    DEALLOCATE(graph%cellsPerPart)
    IF (ALLOCATED(graph%neighbPerpart))   DEALLOCATE(graph%neighbPerPart)
    IF (ALLOCATED(graph%nFacesPerNeighb)) DEALLOCATE(graph%nFacesPerNeighb)
    IF (ALLOCATED(graph%facesperNeighb))  DEALLOCATE(graph%facesPerNeighb)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_GRAPH  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CORRECT_PERIODIC_CONNECTIVITY
    !------------------------------------------------------------------
    !---- Correct connectivity for periodic faces in the case of
    !---- a parallel partitioning
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFacPer, iFac, iFacMatch, iCell, iCellMatch, &
         numLocal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate global mesh
    !------------------------------------------------------------------
    DO iFacPer=1,Mesh%nFacPer
       iFac       = MeshFacPer(iFacPer)%faceNum
       iFacMatch  = CompFacPer(iFacPer)%matchingFace
       iCell      = MeshFac(iFac)%neighb(1)
       iCellMatch = MeshFac(iFacMatch)%neighb(1)
       !
       MeshFac(iFac)%neighb(2)          = iCellMatch
       numLocal                         = MeshFac(iFac)%numLocal(1)
       MeshCell(iCell)%neighb(numLocal) = iCellMatch
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CORRECT_PERIODIC_CONNECTIVITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_FOR_ARTIFICIAL_HYBRID_FACES(iSub)
    !------------------------------------------------------------------
    !---- Check if current subdomain contains artificial hybrid faces
    !
    !---- IN
    !---- iSub : index of the subdomain
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iSub
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)            :: containsArtificialHybridFaces
    
    INTEGER(C_INT)             :: iFacArt, iFac, nArtificialHybridFaces
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (hybridCut) RETURN
    !------------------------------------------------------------------
    !---- Deallocate global mesh
    !------------------------------------------------------------------
    containsArtificialHybridFaces = .FALSE.
    nArtificialHybridFaces        = 0
    !
    DO iFacArt=1,Mesh%nFacArt
       iFac = MeshFacArt(iFacArt)%faceNum
       IF (MeshFac(iFac)%isHybrid) THEN
          containsArtificialHybridFaces = .TRUE.
          nArtificialHybridFaces        = nArtificialHybridFaces + 1
       ENDIF
    ENDDO
    !
    IF (containsArtificialHybridFaces) THEN
       WRITE(6, '(a,i9,a,i9,a)') 'Subdomain ',iSub,' contains '&
            ,nArtificialHybridFaces,' artificial hybrid faces'
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_FOR_ARTIFICIAL_HYBRID_FACES

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE GRAPH_STRUCTURES
