MODULE MATERIALS_STRUCTURES

  !====================================================================
  ! Definition and data structure associated to materials
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a material
  !  
  ! eps(3)   ==> diagonal permittivity tensor
  ! mu(3)    ==> diagonal permeability tensor
  ! isDisp   ==> TRUE if material is dispersive, FALSE otherwise
  ! iDispMat ==> index of the corresponding dispersive material
  ! nCells   ==> # of cells composing this material
  !--------------------------------------------------------------------
  TYPE material
     LOGICAL(C_BOOL)              :: isDisp

     INTEGER(C_INT)               :: matLog, iDispMat, nCells

     REAL(C_DOUBLE), DIMENSION(3) :: eps, mu
  END TYPE material

  !--------------------------------------------------------------------
  ! Data structure associated to a zero-order dispersive pole
  !
  ! sigma ==> Pole parameters in PHz
  !--------------------------------------------------------------------
  TYPE zero_order_pole
     REAL(C_DOUBLE) :: sigma
  END TYPE zero_order_pole

  !--------------------------------------------------------------------
  ! Data structure associated to a first-order dispersive pole
  !
  ! a, b ==> Pole parameters in Hz
  !--------------------------------------------------------------------
  TYPE first_order_pole
     REAL(C_DOUBLE) :: a, b
  END TYPE first_order_pole

  !--------------------------------------------------------------------
  ! Data structure associated to a second-order dispersive pole
  !
  ! c, e ==> Pole parameters in (Hz)**2
  ! d, f ==> Pole parameters in  Hz
  !--------------------------------------------------------------------
  TYPE second_order_pole
     REAL(C_DOUBLE) :: c, d, e, f
  END TYPE second_order_pole

  !--------------------------------------------------------------------
  ! Data structure associated to a non-local dispersive pole
  !
  ! omega2 ==> Pole parameters in (Hz)**2
  ! gamma  ==> Pole parameters in  Hz
  ! beta   ==> Pole parameters (m.s-1)
  !--------------------------------------------------------------------
  TYPE non_local_pole
     REAL(C_DOUBLE) :: omega2, gamma, beta
  END TYPE non_local_pole

  !--------------------------------------------------------------------
  ! Data structure associated to a dispersive material
  ! Arbitrary numbers of first- and second-order poles can be set
  !
  ! dispType   ==> Dispersion model ('gen', 'nl',..)
  ! iMat       ==> Index of the corresponding material
  ! nPole****  ==> number of zero-, first-, second-order and non-local poles
  ! pole0(:)   ==> Zero-order pole parameters,   to be allocated to size nPoles(1)
  ! pole1(:)   ==> First-order pole parameters,  to be allocated to size nPoles(2)
  ! pole2(:)   ==> Second-order pole parameters, to be allocated to size nPoles(3)
  !--------------------------------------------------------------------
  TYPE dispersive_material
     CHARACTER(LEN=5, KIND=C_CHAR)                      :: dispType

     INTEGER(C_INT)                                     :: nPole0, nPole1, &
          nPole2, nPoleNL, iMat
     
     TYPE(zero_order_pole),   DIMENSION(:), ALLOCATABLE :: pole0

     TYPE(first_order_pole),  DIMENSION(:), ALLOCATABLE :: pole1

     TYPE(second_order_pole), DIMENSION(:), ALLOCATABLE :: pole2

     TYPE(non_local_pole),    DIMENSION(:), ALLOCATABLE :: poleNL
  END TYPE dispersive_material

  !--------------------------------------------------------------------
  ! Misc. stuff
  !
  ! useDisp ==> TRUE if at least one material is dispersive
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL) :: useDisp

  !--------------------------------------------------------------------
  ! Arrays of material and dispersive material structures
  !--------------------------------------------------------------------
  TYPE(material),            DIMENSION(:), ALLOCATABLE, TARGET :: Mat
  TYPE(dispersive_material), DIMENSION(:), ALLOCATABLE, TARGET :: MatDisp
  !--------------------------------------------------------------------

CONTAINS

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE EVAL_PERMITTIVITY(iMatDisp, omega, epsR, epsI)
    !----------------------------------------------------------------------
    !---- Evaluate real and imaginary parts of permittivity function
    !---- at puslation omega for material nMat
    !
    !---- IN
    !---- iMatDisp : index of material in MatDisp structure
    !---- omega    : pulsation in PHz
    !
    !---- OUT			
    !---- epsR     : real part of permittivity
    !---- epsI     : imaginary part of permittivity
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT) :: iMatDisp
    
    REAL(C_DOUBLE) :: omega, epsR, epsI
    !----------------------------------------------------------------------
    !---- Local stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT) :: iPole, iMat
    
    REAL(C_DOUBLE) :: locEpsR, locEpsI
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Initialize
    !----------------------------------------------------------------------
    iMat = MatDisp(iMatDisp)%iMat
    epsR = Mat(iMat)%eps(1)
    epsI = 0.0
    !----------------------------------------------------------------------
    !---- Compute permittivities
    !----------------------------------------------------------------------
    !---- Zero order poles
    DO iPole=1,SIZE(MatDisp(iMatDisp)%pole0)
       CALL EVAL_PERMITTIVITY_POLE0(iMatDisp, iPole, omega, locEpsR, locEpsI)
       epsR = epsR + locEpsR
       epsI = epsI + locEpsI
    ENDDO
    !
    !---- First order poles
    DO iPole=1,SIZE(MatDisp(iMatDisp)%pole1)
       CALL EVAL_PERMITTIVITY_POLE1(iMatDisp, iPole, omega, locEpsR, locEpsI)
       epsR = epsR + locEpsR
       epsI = epsI + locEpsI
    ENDDO
    !
    !---- Second order poles
    DO iPole=1,SIZE(MatDisp(iMatDisp)%pole2)
       CALL EVAL_PERMITTIVITY_POLE2(iMatDisp, iPole, omega, locEpsR, locEpsI)
       epsR = epsR + locEpsR
       epsI = epsI + locEpsI
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE EVAL_PERMITTIVITY

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE EVAL_PERMITTIVITY_POLE0(iMatDisp, nPole, omega, epsR, epsI)
    !----------------------------------------------------------------------
    !---- Evaluate real and imaginary contributions of zero-order pole nPole
    !---- to permittivity function at puslation omega for material iMatDisp
    !
    !---- IN
    !---- iMatDisp : index of material in MatDisp structure
    !---- nPole    : index of the pole in MatDisp%pole0 structure
    !---- omega    : pulsation in PHz
    !
    !---- OUT			
    !---- epsR     : real part of permittivity
    !---- epsI     : imaginary part of permittivity
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT) :: iMatDisp, nPole
    
    REAL(C_DOUBLE) :: omega, epsR, epsI
    !----------------------------------------------------------------------
    !---- Local stuff
    !----------------------------------------------------------------------
    REAL(C_DOUBLE) :: sigma
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Compute permittivity contribution
    !----------------------------------------------------------------------
    sigma = MatDisp(iMatDisp)%pole0(nPole)%sigma
    epsR  = 0.0
    epsI  = sigma/omega
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE EVAL_PERMITTIVITY_POLE0

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE EVAL_PERMITTIVITY_POLE1(iMatDisp, nPole, omega, epsR, epsI)
    !----------------------------------------------------------------------
    !---- Evaluate real and imaginary contributions of first-order pole nPole
    !---- to permittivity function at puslation omega for material iMatDisp
    !
    !---- IN
    !---- iMatDisp : index of material in MatDisp structure
    !---- nPole    : index of the pole in MatDisp%pole1 structure
    !---- omega    : pulsation in PHz
    !
    !---- OUT			
    !---- epsR     : real part of permittivity
    !---- epsI     : imaginary part of permittivity
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT) :: iMatDisp, nPole
    
    REAL(C_DOUBLE) :: omega, epsR, epsI
    !----------------------------------------------------------------------
    !---- Local stuff
    !----------------------------------------------------------------------
    REAL(C_DOUBLE) :: a, b, iDenom
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Compute permittivity contribution
    !----------------------------------------------------------------------
    a      = MatDisp(iMatDisp)%pole1(nPole)%a
    b      = MatDisp(iMatDisp)%pole1(nPole)%b
    iDenom = 1.0/(omega**2 + b**2)
    epsR   = a*b*iDenom
    epsI   = a*omega*iDenom
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE EVAL_PERMITTIVITY_POLE1

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE EVAL_PERMITTIVITY_POLE2(iMatDisp, nPole, omega, epsR, epsI)
    !----------------------------------------------------------------------
    !---- Evaluate real and imaginary contributions of second-order pole nPole
    !---- to permittivity function at pulsation omega for material iMatDisp
    !
    !---- IN
    !---- iMatDisp : index of material in MatDisp structure
    !---- nPole    : index of the pole in MatDisp%pole1 structure
    !---- omega    : pulsation in PHz
    !
    !---- OUT			
    !---- epsR     : real part of permittivity
    !---- epsI     : imaginary part of permittivity
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT) :: iMatDisp, nPole
    
    REAL(C_DOUBLE) :: omega, epsR, epsI
    !----------------------------------------------------------------------
    !---- Local stuff
    !----------------------------------------------------------------------
    REAL(C_DOUBLE) :: c, d, e, f, iDenom
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Compute permittivity contribution
    !----------------------------------------------------------------------
    c      = MatDisp(iMatDisp)%pole2(nPole)%c
    d      = MatDisp(iMatDisp)%pole2(nPole)%d
    e      = MatDisp(iMatDisp)%pole2(nPole)%e
    f      = MatDisp(iMatDisp)%pole2(nPole)%f
    iDenom = 1.0/(omega**4 + (f**2 - 2.0*e)*omega**2 + e**2)
    epsR   = ((d*f - c)*omega**2 + e*c)*iDenom
    epsI   = (d*omega**3 + (f*c - d*e)*omega)*iDenom
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE EVAL_PERMITTIVITY_POLE2

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_PERMITTIVITY(filename, nUnit, iMatDisp, omegaStart, omegaEnd, nSamples)
    !----------------------------------------------------------------------
    !---- Write plottable permittivity function files (real and imaginary)
    !
    !---- IN
    !---- filename   : name of file to write to
    !---- nUnit      : file unit for writing
    !---- iMatDisp   : index of material in MatDisp structure
    !---- omegaStart : minimal pulsation in PHz
    !---- omegaEnd   : maximal pulsation in PHz
    !---- nSamples   : number of evaluation points for permittivity function
    !
    !---- OUT			
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE IO_FILES
    !----------------------------------------------------------------------
    !---- I/O stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT)                  :: iMatDisp, nSamples, nUnit
    
    CHARACTER(LEN=100, KIND=C_CHAR) :: filename
    
    REAL(C_DOUBLE)                  :: omegaStart, omegaEnd
    !----------------------------------------------------------------------
    !---- Local stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT)                  :: iSample
    
    REAL(C_DOUBLE)                  :: dOmega, locOmega, epsR, epsI
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Initialize data and open output file
    !----------------------------------------------------------------------
    dOmega = (omegaEnd - omegaStart)/REAL(nSamples)
    !
    OPEN(UNIT=nUnit, FILE=TRIM(filename))
    !----------------------------------------------------------------------
    !---- Compute permittivities
    !----------------------------------------------------------------------
    DO iSample=1,nSamples+1
       locOmega = omegaStart + REAL(iSample-1)*dOmega
       CALL EVAL_PERMITTIVITY(iMatDisp, locOmega, epsR, epsI)
       WRITE(nUnit, '(e15.7,a,e15.7,a,e15.7)') locOmega, TABCHAR, epsR, TABCHAR, epsI
    ENDDO
    !
    CLOSE(nUnit)
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_PERMITTIVITY

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE INIT_MAT_STRUCT(mat)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    TYPE(material), INTENT(INOUT) :: mat
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Initialize Mat structure
    !----------------------------------------------------------------------
    mat%eps(:)   =  1.0
    mat%mu(:)    =  1.0
    mat%isDisp   = .FALSE.
    mat%matLog   =  0
    mat%iDispMat =  0
    mat%nCells   =  0
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE INIT_MAT_STRUCT

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE INIT_MATDISP_STRUCT(matDisp)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    TYPE(dispersive_material), INTENT(INOUT) :: matDisp
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Initialize MatDisp structure
    !----------------------------------------------------------------------
    matDisp%iMat     =  0
    matDisp%dispType = 'NONE'
    matDisp%nPole0   =  0
    matDisp%nPole1   =  0
    matDisp%nPole2   =  0
    matDisp%nPoleNL  =  0
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE INIT_MATDISP_STRUCT

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE READ_MAT_MAIN(nUnit, matLog, nPole0, nPole1, nPole2, &
       nPoleNL, isGenDisp, isNL)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE IO_FILES
    USE SPECIAL_FUNCTIONS
    USE PARALLEL_STRUCTURES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    LOGICAL(C_BOOL),     INTENT(OUT) :: isGenDisp, isNL
    
    INTEGER(C_INT),      INTENT(IN)  :: nUnit, matLog
    INTEGER(C_INT),      INTENT(OUT) :: nPole0, nPole1, nPole2, nPoleNL
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)                   :: nPoles
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Read material data
    !----------------------------------------------------------------------
    Mat(matLog)%matLog = matLog
    !
    !---- Read epsilon and mu tensors
    CALL READ_MATERIAL_TENSOR(nUnit, Mat(matLog)%eps)
    CALL READ_MATERIAL_TENSOR(nUnit, Mat(matLog)%mu)
    !
    !---- Read poles
    READ(nUnit,*) nPole0, nPole1, nPole2, nPoleNL
    nPoles    = nPole0 + nPole1 + nPole2
    isGenDIsp = .FALSE.
    isNL      = .FALSE.
    !
    IF ((nPoles >  0) .AND. (nPoleNL == 0)) isGenDisp = .TRUE.
    IF ((nPoles == 0) .AND. (nPoleNL >  0)) isNL      = .TRUE.
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    IF ((nPoleNL > 0) .AND. (nPoles > 0)) THEN
       IF(idWrite) WRITE(6,'(a)') 'Error in READ_MAT_MAIN'
       IF(idWrite) WRITE(6,'(a)') 'A material contains both local and non-local poles'
       IF(idWrite) WRITE(6,'(a)') 'This case is not allowed'
       CALL TILT
    ENDIF
    !----------------------------------------------------------------------
  END SUBROUTINE READ_MAT_MAIN

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE READ_MATERIAL_TENSOR(nUnit, tensor)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE IO_FILES
    USE SPECIAL_FUNCTIONS
    USE PARALLEL_STRUCTURES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nUnit

    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: tensor
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR)          :: string
    
    INTEGER(C_INT)                            :: nVal

    REAL(C_DOUBLE)                            :: isoVal
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Read tensor
    !----------------------------------------------------------------------
    READ(nUnit,'(a)') string
    CALL GET_NUMBER_OF_DOUBLE_IN_STRING(string, nVal)
    !
    BACKSPACE(nUnit)
    IF (nVal == 3) READ(nUnit,*) tensor(:)
    IF (nVal == 1) READ(nUnit,*) isoVal
    IF (nVal == 1) tensor(:) = isoVal
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    IF (nVal == 0) THEN
       IF(idWrite) WRITE(6,'(a)') 'Error in READ_MATERIAL_TENSOR'
       IF(idWrite) WRITE(6,'(a)') 'I could not read the material tensors'
       CALL TILT
    ENDIF
    !
    IF (nVal == 2) THEN
       IF(idWrite) WRITE(6,'(a)') 'Error in READ_MATERIAL_TENSOR'
       IF(idWrite) WRITE(6,'(a)') 'You provided two value for epsilon or mu tensor'
       IF(idWrite) WRITE(6,'(a)') 'This case is not allowed'
       IF(idWrite) WRITE(6,'(a)') 'Please provide either one value (isotropic &
            material) or three values (diagonally anisotropic material)'       
       CALL TILT
    ENDIF
    !----------------------------------------------------------------------
  END SUBROUTINE READ_MATERIAL_TENSOR

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE FILL_MATDISP_MAIN(dispMatLog, matLog, dispType, nPole0, &
       nPole1, nPole2, nPoleNL)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    CHARACTER(LEN=5, KIND=C_CHAR), INTENT(IN)    :: dispType

    INTEGER(C_INT),                INTENT(IN)    :: dispMatLog, matLog, &
         nPole0, nPole1, nPole2, nPoleNL
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Fill a dispersive material
    !----------------------------------------------------------------------
    MatDisp(dispMatLog)%iMat     =  matLog
    MatDisp(dispMatLog)%dispType =  TRIM(dispType)
    MatDisp(dispMatLog)%nPole0   =  nPole0
    MatDisp(dispMatLog)%nPole1   =  nPole1
    MatDisp(dispMatLog)%nPole2   =  nPole2
    MatDisp(dispMatLog)%nPoleNL  =  nPoleNL
    !
    Mat(matLog)%isDisp           = .TRUE.
    Mat(matLog)%iDispmat         =  dispMatLog
    !
    ALLOCATE(MatDisp(dispMatLog)%pole0 (nPole0 ))
    ALLOCATE(MatDisp(dispMatLog)%pole1 (nPole1 ))
    ALLOCATE(MatDisp(dispMatLog)%pole2 (nPole2 ))
    ALLOCATE(MatDisp(dispMatLog)%poleNL(nPoleNL))
    !
    !---- Second order pole requires a zero order pole
    IF ((nPole0 == 0) .AND. (nPole2 > 0)) THEN
       IF (ALLOCATED(MatDisp(dispMatLog)%pole0)) DEALLOCATE(MatDisp(dispMatLog)%pole0)
       ALLOCATE(MatDisp(dispMatLog)%pole0(1))       
       MatDisp(dispMatLog)%nPole0         = 1
       MatDisp(dispMatLog)%pole0(1)%sigma = 0.0
    ENDIF
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE FILL_MATDISP_MAIN

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE READ_POLES_0(nUnit, matDispLog, nPole0)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nUnit, nPole0, matDispLog
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)             :: iPole
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Read 0th order pole
    !----------------------------------------------------------------------
    DO iPole=1,nPole0
       READ(nUnit,*) MatDisp(matDispLog)%pole0(iPole)%sigma
       !
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole0(iPole)%sigma)
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE READ_POLES_0

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE READ_POLES_1(nUnit, matDispLog, nPole1)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nUnit, nPole1, matDispLog
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)             :: iPole
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Read 1st order pole
    !----------------------------------------------------------------------
    DO iPole=1,nPole1
       READ(nUnit,*) &
            MatDisp(matDispLog)%pole1(iPole)%a, &
            MatDisp(matDispLog)%pole1(iPole)%b
       !
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole1(iPole)%a)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole1(iPole)%b)
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE READ_POLES_1

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE READ_POLES_2(nUnit, matDispLog, nPole2)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nUnit, nPole2, matDispLog
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)             :: iPole
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Read 2d order pole
    !----------------------------------------------------------------------
    DO iPole=1,nPole2
       READ(nUnit,*) &
            MatDisp(matDispLog)%pole2(iPole)%c, &
            MatDisp(matDispLog)%pole2(iPole)%d, &
            MatDisp(matDispLog)%pole2(iPole)%e, &
            MatDisp(matDispLog)%pole2(iPole)%f
       !
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole2(iPole)%c)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole2(iPole)%c)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole2(iPole)%d)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole2(iPole)%e)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole2(iPole)%e)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%pole2(iPole)%f)
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE READ_POLES_2

  !========================================================================
  !========================================================================
  !========================================================================
  
  SUBROUTINE READ_POLES_NL(nUnit, matDispLog, nPoleNL)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nUnit, nPoleNL, matDispLog
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)             :: iPole
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Read non-local pole
    !----------------------------------------------------------------------
    DO iPole=1,nPoleNL
       READ(nUnit,*) &
            MatDisp(matDispLog)%poleNL(iPole)%omega2, &
            MatDisp(matDispLog)%poleNL(iPole)%gamma, &
            MatDisp(matDispLog)%poleNL(iPole)%beta
       !
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%poleNL(iPole)%omega2)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%poleNL(iPole)%omega2)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%poleNL(iPole)%gamma)
       CALL FREQUENCY_TO_INVERSE_METERS(MatDisp(matDispLog)%poleNL(iPole)%beta)
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE READ_POLES_NL

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE EXTEND_MATDISP_ARRAY(matDispArray, matDispArraySize)
    !----------------------------------------------------------------------
    !---- Add a new slot to matDispArray
    !
    !---- IN
    !---- matDispArray     : array of matDisp
    !---- matDispArraySize : number of elements in matDispArray
    !
    !---- OUT
    !---- matDispArray     : modified in the process
    !---- matDispArraySize : modified in the process
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- I/O stuff
    !----------------------------------------------------------------------
    INTEGER(C_INT),                        INTENT(INOUT) :: matDispArraySize
    
    TYPE(dispersive_material), DIMENSION(:), ALLOCATABLE :: matDispArray
    !----------------------------------------------------------------------
    !---- Local stuff
    !----------------------------------------------------------------------
    LOGICAL(C_BOOL)                                      :: isArrayFull, isAllocated

    TYPE(dispersive_material), DIMENSION(:), ALLOCATABLE :: tempMatDisp
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Check if cell array is full
    !----------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(matDispArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(matDispArray) < matDispArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !----------------------------------------------------------------------
    !---- Extend array if necessary
    !----------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempMatDisp)) DEALLOCATE(tempMatDisp)
          ALLOCATE(tempMatDisp(matDispArraySize))
          tempMatDisp(1:matDispArraySize) = matDispArray(1:MatDispArraySize)
          DEALLOCATE(matDispArray)
       ENDIF
       !
       ALLOCATE(matDispArray(matDispArraySize + 1))
       CALL INIT_MATDISP_STRUCT(matDispArray(matDispArraySize + 1))
       !
       matDispArray(1:matDispArraySize) = tempMatDisp(1:matDispArraySize)
       IF (ALLOCATED(tempMatDisp)) DEALLOCATE(tempMatDisp)
    ENDIF
    !
    matDispArraySize = matDispArraySize + 1
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE EXTEND_MATDISP_ARRAY

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE FIND_BACKGROUND_MATERIAL(dFace, dFaceSize, logRef)
    !----------------------------------------------------------------------
    !---- Find the homogeneous background material around given
    !---- derived face list
    !
    !---- IN
    !---- dFace     : derived face list
    !---- dFaceSize : size of derived face list
    !
    !---- OUT
    !---- logRef    : background log
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(IN)  :: dFaceSize
    INTEGER(C_INT),                   INTENT(OUT) :: logRef
    
    TYPE(derived_face), DIMENSION(:), ALLOCATABLE :: dFace
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    LOGICAL(C_BOOL)                               :: isHomogeneous
    
    INTEGER(C_INT)                                :: idFace, iFace, &
         nLogChange, iSide, logSide, nLogChangeGlobal
    INTEGER(C_INT), DIMENSION(2)                  :: iCell

    REAL(C_DOUBLE), DIMENSION(3)                  :: epsRef, muRef, eps, mu
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Check background cells
    !----------------------------------------------------------------------
    isHomogeneous = .TRUE.
    logRef        = 0
    epsRef(:)     = Mat(logRef)%eps(:)
    muRef(:)      = Mat(logRef)%mu(:)
    nLogChange    = 0
    !
    DO idFace=1,dFaceSize
       iFace    = dFace(idFace)%faceNum
       iCell(:) = MeshFac(iFace)%neighb(:)
       !
       DO iSide=1,2
          IF (iCell(iSide) <= Mesh%nCell) THEN
             logSide = MeshCell(iCell(iSide))%log
             !
             !---- Handle the case of PML
             IF (logSide >= 100) logSide = logSide - 100
             !
             eps(:)  = Mat(logSide)%eps(:)
             mu(:)   = Mat(logSide)%mu(:)
             !
             IF (ANY(eps /= epsRef) .OR. ANY(mu /= muRef)) THEN
                logRef     = logSide
                epsRef(:)  = eps(:)
                muRef(:)   = mu(:)
                nLogChange = nLogChange + 1
             ENDIF
          ENDIF
       ENDDO
    ENDDO
    !
    CALL GLOBAL_OP(nLogChange, MPI_MAX, nLogChangeGlobal)
    IF (nLogChangeGlobal > 1) isHomoGeneous = .FALSE.
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    IF (isHomogeneous) THEN
       IF(idWrite) WRITE(6,'(a)')    'Background material is homogeneous'
       IF(idWrite) WRITE(6,'(a,i0)') 'Background material log is ',logRef
       IF(idWrite) WRITE(6, 10)
    ENDIF
    !
    IF (.NOT. isHomogeneous) THEN
       IF(idWrite) WRITE(6,'(a)')      'Error in FIND_BACKGROUND_MATERIAL'
       IF(idWrite) WRITE(6,'(a)')      'Background material is not homogeneous'
       IF(idWrite) WRITE(6,'(a,i0,a)') 'I found ',nLogChangeGlobal,' different &
            materials touching the injection surface'
       CALL TILT
    ENDIF
    !
    IF (Mat(logRef)%isDisp) THEN
       IF(idWrite) WRITE(6,'(a)') 'Warning in FIND_BACKGROUND_MATERIAL'
       IF(idWrite) WRITE(6,'(a)') 'You are trying to impose a plane wave &
            directly inside a dispersive material'
       IF(idWrite) WRITE(6,'(a)') 'This is not allowed'       
       !CALL TILT
    ENDIF
    !------------------------------------------------------------------
10  FORMAT(60('-'))    
    !----------------------------------------------------------------------
  END SUBROUTINE FIND_BACKGROUND_MATERIAL  

  !========================================================================
  !========================================================================
  !========================================================================

END MODULE MATERIALS_STRUCTURES
