MODULE MATRICES_STRUCTURES

  !======================================================================
  ! Data structures for matrices and vectors, and several usual operations
  !======================================================================

  !----------------------------------------------------------------------
  ! Shared module
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  !----------------------------------------------------------------------
  IMPLICIT NONE
  !----------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Data structure of vectors, v := [v_1, ..., v_ndof]^T
  ! ndof := v%ndof
  ! v_i  := v%tab(i)
  !----------------------------------------------------------------------
  TYPE vector_integer
     INTEGER(C_INT) :: ndof

     INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: tab
  END TYPE vector_integer
  !
  TYPE vector_real
     INTEGER(C_INT) :: ndof

     REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: tab
  END TYPE vector_real
  !
  TYPE vector_complex
     INTEGER(C_INT) :: ndof

     COMPLEX(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: tab
  END TYPE vector_complex

  !----------------------------------------------------------------------
  ! Data structure of square matrices, M[ndof*ndof]
  ! ndof := M%ndof
  ! M_ij := M%tab(i,j)
  !----------------------------------------------------------------------
  TYPE matrix_sq_integer
     INTEGER(C_INT) :: ndof

     INTEGER(C_INT), DIMENSION(:,:), ALLOCATABLE :: tab
  END TYPE matrix_sq_integer
  !
  TYPE matrix_sq_real
     INTEGER(C_INT) :: ndof

     REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: tab
  END TYPE matrix_sq_real
  !
  TYPE matrix_sq_complex
     INTEGER(C_INT) :: ndof

     COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: tab
  END TYPE matrix_sq_complex

  !----------------------------------------------------------------------
  ! Data structure of rectangular matrices, M[ndof1*ndof2]
  ! ndof1 := M%ndof1, ndof2 := M%ndof2
  ! M_ij  := M%tab(i,j)
  !----------------------------------------------------------------------
  TYPE matrix_rect_integer
     INTEGER(C_INT) :: ndof1, ndof2

     REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: tab
  END TYPE matrix_rect_integer
  !
  TYPE matrix_rect_real
     INTEGER(C_INT) :: ndof1, ndof2

     REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: tab
  END TYPE matrix_rect_real
  !
  TYPE matrix_rect_complex
     INTEGER(C_INT) :: ndof1, ndof2

     COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: tab
  END TYPE matrix_rect_complex

  !----------------------------------------------------------------------
  ! Structure of pointer to an integer matrix
  !
  ! ptr            ==> pointer to the associated matrix
  !----------------------------------------------------------------------
  TYPE sq_matrix_ptr
     TYPE(matrix_sq_integer), POINTER :: ptr
  END TYPE sq_matrix_ptr

  !----------------------------------------------------------------------
  ! Structure of pointer to a real rectangular matrix
  !
  ! ptr            ==> pointer to the associated matrix
  !----------------------------------------------------------------------
  TYPE rect_matrix_ptr
     TYPE(matrix_rect_real), POINTER :: ptr
  END TYPE rect_matrix_ptr

  !--------------------------------------------------------------------
  ! Vector field structures
  !--------------------------------------------------------------------
  TYPE real_vector_field
     REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: vec
  END TYPE real_vector_field
  !
  TYPE complex_vector_field
     COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: vec
  END TYPE complex_vector_field

  !----------------------------------------------------------------------
  ! Interface for overloading matrix-vector multiplication
  !----------------------------------------------------------------------
  INTERFACE MATVECMUL
     MODULE PROCEDURE MATVECMUL_REAL, MATVECMUL_CX
  END INTERFACE MATVECMUL

  !----------------------------------------------------------------------
  ! Interface for overloading vector-vector reductions
  !----------------------------------------------------------------------
  INTERFACE VECVECRED
     MODULE PROCEDURE VECVECRED_REAL, VECVECRED_CX
  END INTERFACE VECVECRED
  !----------------------------------------------------------------------  

CONTAINS

  !======================================================================
  !======================================================================
  !======================================================================
  
  SUBROUTINE MATMATMUL(mat1,mat2,m,res)
    !--------------------------------------------------------------------
    !---- A square matrix-matrix multiplication (M1*M2)
    !
    !---- IN
    !---- mat1, mat2 : m*m matrices
    !---- m          : size of mat1 and mat2
    !
    !---- OUT
    !---- res        : resulting matrix
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                    :: m

    REAL(C_DOUBLE), INTENT(INOUT), DIMENSION(:,:) :: res
    REAL(C_DOUBLE), INTENT(IN),    DIMENSION(:,:) :: mat1, mat2
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: iRow, iCol
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Multiply
    !--------------------------------------------------------------------
    res = 0.0d0
    !
    DO iRow=1,m
       DO iCol=1,m
          res(iRow,iCol) = res(iRow,iCol) + mat1(iRow,iCol)*mat2(iCol,iRow)
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE MATMATMUL

  !======================================================================
  !======================================================================
  !======================================================================
  
  SUBROUTINE MATVECMUL_REAL(mat,vec,m,n,res)
    !--------------------------------------------------------------------
    !---- A matrix-vector multiplication for real vectors (M*V)
    !
    !---- IN
    !---- mat  : m*n matrix (m rows, n columns)
    !---- vec  : vector of size n (n rows)
    !---- m, n : sizes of mat
    !
    !---- OUT
    !---- res  : resulting vector
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                    :: m, n

    REAL(C_DOUBLE), INTENT(INOUT), DIMENSION(:)   :: res
    REAL(C_DOUBLE), INTENT(IN),    DIMENSION(:)   :: vec
    REAL(C_DOUBLE), INTENT(IN),    DIMENSION(:,:) :: mat
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: iRow, iCol
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Multiply
    !--------------------------------------------------------------------
    res = 0.0d0
    !
    DO iCol=1,n
       DO iRow=1,m
          res(iRow) = res(iRow) + mat(iRow,iCol)*vec(iCol)
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE MATVECMUL_REAL

  !======================================================================
  !======================================================================
  !======================================================================
  
  SUBROUTINE MATVECMUL_CX(mat,vec,m,n,res)
    !--------------------------------------------------------------------
    !---- A matrix-vector multiplication for complex vectors (M*V)
    !
    !---- IN
    !---- mat  : m*n matrix (m rows, n columns)
    !---- vec  : vector of size n (n rows)
    !---- m, n : sizes of mat
    !
    !---- OUT
    !---- res  : resulting vector
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),            INTENT(IN)                    :: m, n

    COMPLEX(C_DOUBLE_COMPLEX), INTENT(INOUT), DIMENSION(:)   :: res
    COMPLEX(C_DOUBLE_COMPLEX), INTENT(IN),    DIMENSION(:)   :: vec
    REAL(C_DOUBLE),            INTENT(IN),    DIMENSION(:,:) :: mat
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: iRow, iCol
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Multiply
    !--------------------------------------------------------------------
    res = 0.0d0
    !
    DO iCol=1,m
       DO iRow=1,n
          res(iRow) = res(iRow) + mat(iRow,iCol)*vec(iCol)
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE MATVECMUL_CX

  !======================================================================
  !======================================================================
  !======================================================================
  
  PURE SUBROUTINE VECVECRED_REAL(vec1,vec2,m,res)
    !--------------------------------------------------------------------
    !---- A vector-vector reduction for real vectors (vec1.vec2)
    !
    !---- IN
    !---- vec1, vec2  : vectors of size m
    !---- m           : sizes of vec1 and vec2
    !
    !---- OUT
    !---- res         : resulting scalar
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                 :: m

    REAL(C_DOUBLE), INTENT(INOUT)              :: res
    REAL(C_DOUBLE), INTENT(IN),   DIMENSION(:) :: vec1, vec2
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: iRow
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform reduction
    !--------------------------------------------------------------------
    res = 0.0d0
    !
    DO iRow=1,m
       res = res + vec1(iRow)*vec2(iRow)
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE VECVECRED_REAL

  !======================================================================
  !======================================================================
  !======================================================================
  
  PURE SUBROUTINE VECVECRED_CX(vec1,vec2,m,res)
    !--------------------------------------------------------------------
    !---- A vector-vector reduction for complex vectors (vec1.vec2)
    !
    !---- IN
    !---- vec1, vec2  : vectors of size m
    !---- m           : sizes of vec1 and vec2
    !
    !---- OUT
    !---- res         : resulting scalar
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),            INTENT(IN)                 :: m

    COMPLEX(C_DOUBLE_COMPLEX), INTENT(INOUT)              :: res
    COMPLEX(C_DOUBLE_COMPLEX), INTENT(IN),   DIMENSION(:) :: vec1, vec2
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: iRow
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform reduction
    !--------------------------------------------------------------------
    res = 0.0d0
    !
    DO iRow=1,m
       res = res + vec1(iRow)*vec2(iRow)
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE VECVECRED_CX

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE INVERSE_3X3(mat,threshold,iMat,dJac)
    !--------------------------------------------------------------------
    !---- Computation of the inverse of matrix mat (3x3)
    !
    !---- IN
    !---- mat       : matrix
    !---- threshold : limit below which the determinant is considered to be 0
    !
    !---- OUT
    !---- iMat      : inverse matrix
    !---- dJac      : determinant
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN),  DIMENSION(3,3) :: mat
    REAL(C_DOUBLE), INTENT(OUT), DIMENSION(3,3) :: iMat
    REAL(C_DOUBLE), INTENT(OUT)                 :: dJac
    REAL(C_DOUBLE), INTENT(IN)                  :: threshold
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE) :: coFactor(3,3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute inverse
    !--------------------------------------------------------------------
    CALL COMPUTE_DETERMINANT(mat, 3, dJac)
    !
    IF (dJac <= threshold) THEN
       WRITE(6,'(a)') "Error in INVERSE_3X3"
       WRITE(6,'(a)') "Matrix is singular"
       CALL TILT
    ENDIF
    !
    coFactor(1,1) = +(mat(2,2)*mat(3,3)-mat(2,3)*mat(3,2))
    coFactor(1,2) = -(mat(2,1)*mat(3,3)-mat(2,3)*mat(3,1))
    coFactor(1,3) = +(mat(2,1)*mat(3,2)-mat(2,2)*mat(3,1))
    coFactor(2,1) = -(mat(1,2)*mat(3,3)-mat(1,3)*mat(3,2))
    coFactor(2,2) = +(mat(1,1)*mat(3,3)-mat(1,3)*mat(3,1))
    coFactor(2,3) = -(mat(1,1)*mat(3,2)-mat(1,2)*mat(3,1))
    coFactor(3,1) = +(mat(1,2)*mat(2,3)-mat(1,3)*mat(2,2))
    coFactor(3,2) = -(mat(1,1)*mat(2,3)-mat(1,3)*mat(2,1))
    coFactor(3,3) = +(mat(1,1)*mat(2,2)-mat(1,2)*mat(2,1))
    !
    iMat = TRANSPOSE(coFactor)/dJac
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE INVERSE_3X3

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE COMPUTE_DETERMINANT(matrix, matSize, det)
    !--------------------------------------------------------------------
    !---- Computation of the determinant of square matrix
    !
    !---- IN
    !---- matrix  : square matrix
    !---- matSize : size of matrix
    !
    !---- OUT
    !---- det     : determinant
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                 :: matSize
    
    REAL(C_DOUBLE), INTENT(IN), DIMENSION(:,:) :: matrix
    REAL(C_DOUBLE), INTENT(OUT)                :: det
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                             :: iRow
    
    REAL(C_DOUBLE)                             :: rowSwap
    REAL(C_DOUBLE), DIMENSION(matSize,matSize) :: mat
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute determinant
    !--------------------------------------------------------------------
    det = 0.0
    !
    !---- Copy matrix to local array
    mat = matrix
    !
    !---- Convert to upper triangular form
    CALL CONVERT_TO_UPPER_TRIANGULAR(mat, matSize, rowSwap)
    !
    !---- Compute det by finding product of diagonal elements
    det = rowSwap
    DO iRow=1,matSize
       det = det*mat(iRow,iRow)
    END DO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE COMPUTE_DETERMINANT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE CONVERT_TO_UPPER_TRIANGULAR(mat, matSize, rowSwap)
    !--------------------------------------------------------------------
    !---- Convert a square matrix to upper triangular form
    !
    !---- IN
    !---- mat     : square matrix
    !---- matSize : size of matrix
    !
    !---- OUT
    !---- mat     : modified in the process
    !---- rowSwap : equals +1 or -1 depending on the number of row swapw
    !----           performed during the conversion
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                    :: matSize

    REAL(C_DOUBLE), INTENT(OUT)                   :: rowSwap
    REAL(C_DOUBLE), INTENT(INOUT), DIMENSION(:,:) :: mat
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    LOGICAL(C_BOOL)                               :: detExists = .TRUE.
    
    INTEGER(C_INT)                                :: iRow, iMoreRow
    
    REAL(C_DOUBLE)                                :: pivot, matVal
    REAL(C_DOUBLE), DIMENSION(matSize)            :: tempRow
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Convert
    !--------------------------------------------------------------------
    rowSwap = 1.0
    !
    !---- Loop over rows
    DO iRow=1,matSize-1       
       pivot = mat(iRow,iRow)
       !
       !---- If pivot is 0, we must find another row below the current one
       !---- for which the value in the same column is not zero
       IF (pivot == 0.0) THEN
          detExists = .FALSE.
          !
          DO iMoreRow=iRow+1,matSize
             IF (mat(iMoreRow,iRow) /= 0.0) THEN
                !---- We found a row that matches, now we swap it with
                !---- the row that was containing a zero on the diagonal
                tempRow(:)      = mat(iMoreRow,:)
                mat(iMoreRow,:) = mat(iRow,:)
                mat(iRow,:)     = tempRow(:)
                pivot           = mat(iRow,iRow)
                !
                !---- If determinant is computed after that, the number
                !---- of row swapping must be counted, since after a row
                !---- swap the sign of the determinant changes
                rowSwap   = -1.0*rowSwap
                detExists = .TRUE.
                EXIT
             ENDIF
          ENDDO
          !
          !---- If no row was found for swapping, then the determinant
          !---- is zero and we must exit. To do that, we just set
          !---- rowSwap to zero
          IF (.NOT. detExists) THEN
             rowSwap = 0.0
             RETURN
          ENDIF
       ENDIF
       !
       !---- For all rows following iRow, compute new values
       DO iMoreRow=iRow+1,matSize
          matVal          = mat(iMoreRow,iRow)/pivot
          mat(iMoreRow,:) = mat(iMoreRow,:) - matVal*mat(iRow,:)
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE CONVERT_TO_UPPER_TRIANGULAR

  !======================================================================
  !======================================================================
  !======================================================================  

  FUNCTION SPECIAL_DET_2X2(v1,v2) RESULT(det)
    !--------------------------------------------------------------------
    !---- Special 2x2 determinant computation :
    !
    !     | v1(1) v1(2) |
    !     | v2(1) v2(2) |
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN), DIMENSION(2) :: v1, v2
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE) det
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute determinant
    !--------------------------------------------------------------------
    det = v1(1)*v2(2) - v2(1)*v1(2)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END FUNCTION SPECIAL_DET_2X2
  
  !======================================================================
  !======================================================================
  !======================================================================  

  FUNCTION SPECIAL_DET_3X3(v1,v2,v3) RESULT(det)
    !--------------------------------------------------------------------
    !---- Special 3x3 determinant computation :
    !
    !     | v1(1) v1(2) v1(3) |
    !     | v2(1) v2(2) v2(3) |
    !     | v3(1) v3(2) v3(3) |
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN), DIMENSION(3) :: v1, v2, v3
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE) det
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute determinant
    !--------------------------------------------------------------------
    det = v1(1)*(v2(2)*v3(3) - v2(3)*v3(2)) &
        - v1(2)*(v2(1)*v3(3) - v2(3)*v3(1)) &
        + v1(3)*(v2(1)*v3(2) - v2(2)*v3(1))
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END FUNCTION SPECIAL_DET_3X3

  !======================================================================
  !======================================================================
  !======================================================================  

  FUNCTION SPECIAL_DET_4X4(v1,v2,v3,v4) RESULT(det)
    !--------------------------------------------------------------------
    !---- Special 4x4 determinant computation :
    !
    !     | v1(1) v1(2) v1(3) 1 |
    !     | v2(1) v2(2) v2(3) 1 |
    !     | v3(1) v3(2) v3(3) 1 |
    !     | v4(1) v4(2) v4(3) 1 |
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN), DIMENSION(3) :: v1, v2, v3, v4
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE) det
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute determinant
    !--------------------------------------------------------------------
    det = v2(1)*(v3(2)*v4(3) - v3(3)*v4(2)) - v3(1)*(v2(2)*v4(3) - v2(3)*v4(2)) &
        + v4(1)*(v2(2)*v3(3) - v2(3)*v3(2)) - v1(1)*(v3(2)*v4(3) - v3(3)*v4(2)) &
        + v3(1)*(v1(2)*v4(3) - v1(3)*v4(2)) - v4(1)*(v1(2)*v3(3) - v1(3)*v3(2)) &
        + v1(1)*(v2(2)*v4(3) - v2(3)*v4(2)) - v2(1)*(v1(2)*v4(3) - v1(3)*v4(2)) &
        + v4(1)*(v1(2)*v2(3) - v1(3)*v2(2)) - v1(1)*(v2(2)*v3(3) - v2(3)*v3(2)) &
        + v2(1)*(v1(2)*v3(3) - v1(3)*v3(2)) - v3(1)*(v1(2)*v2(3) - v1(3)*v2(2)) 
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END FUNCTION SPECIAL_DET_4X4  

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE REDUCTION(A, B, dim1, dim2, dim3, dim4)
    !--------------------------------------------------------------------
    !---- Elimination of the null rows and columns of the matrix  A[dim1*dim2] 
    !---- to obtain the matrix B[dim3*dim4]
    !
    !---- IN
    !---- A          : Input matrix with null rows and columns
    !---- dim1,dim2  : sizes of A
    !
    !---- OUT
    !---- B          : Output matrix with no null rows or columns
    !---- dim3, dim4 : sizes of B
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                    :: dim1, dim2, dim3, dim4

    REAL(C_DOUBLE), INTENT(IN),    DIMENSION(:,:) :: A ! dim1*dim2

    REAL(C_DOUBLE), INTENT(INOUT), DIMENSION(:,:) :: B ! dim3*dim4
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                        :: i, j, k, n

    REAL(C_DOUBLE), DIMENSION(dim2)       :: P

    REAL(C_DOUBLE), DIMENSION(dim1)       :: Q

    REAL(C_DOUBLE), DIMENSION(dim3, dim4) :: C
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform reduction
    !--------------------------------------------------------------------
    k = 0

    DO i=1,dim1
       P(1:dim2) = ABS(A(i,1:dim2))

       IF (MAXVAL(P) > 1.0e-12) THEN
          k = k + 1
          n = 0

          DO j=1,dim2
             Q(1:dim1) = ABS(A(1:dim1,j))

             IF (MAXVAL(Q) > 1.0e-12) THEN
                n = n + 1
                C(k,n) = A(i,j)
             ENDIF
          ENDDO
       ENDIF
    ENDDO

    B = C
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE REDUCTION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INVERSE_MATRIX(ncho, ndim, mat, inv_mat, error)
    !------------------------------------------------------------------
    !---- Inversion by Gauss elimination with pivoting (ncho=GP) or 
    !---- Cholesky factorization (ncho=CK)
    !
    !---- IN
    !---- ncho : inversion method
    !---- ndim : dimension of square matrix
    !---- mat  : matrix to invert
    !
    !---- OUT
    !---- inv_mat : inverted matrix
    !---- error   : error of the inversion given by Frobenius norm
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=2, KIND=C_CHAR), INTENT(IN)         :: ncho

    INTEGER(C_INT), INTENT(IN)                        :: ndim

    REAL(C_DOUBLE), INTENT(OUT)                       :: error
    REAL(C_DOUBLE), INTENT(IN),  DIMENSION(ndim,ndim) :: mat
    REAL(C_DOUBLE), INTENT(OUT), DIMENSION(ndim,ndim) :: inv_mat
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: isPositiveDefinite
    
    INTEGER(C_INT)  :: iDim

    REAL(C_DOUBLE)  :: limitError
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: tab
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Performe inversion
    !------------------------------------------------------------------
    ALLOCATE(tab(ndim,ndim))
    tab = mat
    !
    SELECT CASE(ncho)
    CASE('GP')
       CALL GAUSS_PIVOT(tab, ndim)
       limitError = 1.0e-10
    CASE('CK')
       CALL CHOLESKY(tab, ndim)
       limitError = 1.0e-10
    CASE DEFAULT
       WRITE(6,'(a)') 'Unknown inversion method in module INVERSION'
       WRITE(6,'(a)') 'Received inversion method: ', ncho
       CALL TILT
    END SELECT
    !
    inv_mat = tab
    !------------------------------------------------------------------
    !---- Evaluate error with Frobenius norm : ||M*[M]^-1 - Id||
    !------------------------------------------------------------------
    tab = 0.0d0
    DO iDim=1,ndim
       tab(iDim,iDim) = 1.0d0
    ENDDO
    tab = tab - MATMUL(mat, inv_mat)
    !
    CALL FROBENIUS_NORM(tab,ndim,error)
    IF (ALLOCATED(tab)) DEALLOCATE(tab)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (.NOT. (error < limitError)) THEN ! This test also catches NaN
       WRITE(6,'(a)') 'Matrix inversion failed in module INVERSION'
       WRITE(6,'(a,e12.5)') 'Error in Frobenius norm is: ',error
       !
       !---- If the method is CK, test if the matrix is positive definite
       IF (ncho == 'CK') THEN
          CALL IS_POSITIVE_DEFINITE(mat,ndim,isPositiveDefinite)
          !
          IF (.NOT. isPositiveDefinite) THEN
             WRITE(6,'(a)') 'You asked for inversion with Cholesky method'
             WRITE(6,'(a)') 'However, the matrix you gave is not positive definite'
          ENDIF
       ENDIF
       !
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE INVERSE_MATRIX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POSITIVE_DEFINITE(mat, nDim, isPositiveDefinite)
    !------------------------------------------------------------------
    !---- Test if input matrix is positive definite
    !---- This routine uses Sylvester's criterion : mat is PD if and
    !---- only if all its leading principal minors are positive 
    !
    !---- IN
    !---- mat                : matrix to test
    !---- nDim               : size of mat
    !
    !---- OUT
    !---- isPositiveDefinite : TRUE if mat is positive definite
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL), INTENT(OUT)                      :: isPositiveDefinite
    
    INTEGER(C_INT),  INTENT(IN)                       :: nDim
    
    REAL(C_DOUBLE),  INTENT(IN), DIMENSION(ndim,ndim) :: mat
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iDim

    REAL(C_DOUBLE) :: det
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Test if matrix is positive definite
    !------------------------------------------------------------------
    isPositiveDefinite = .TRUE.
    !
    DO iDim=1,nDim
       CALL COMPUTE_DETERMINANT(mat(1:iDim,1:iDim),iDim,det)
       IF (det < 0.0) THEN
          isPositiveDefinite = .FALSE.
          EXIT
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POSITIVE_DEFINITE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FROBENIUS_NORM(mat, ndim, norm)
    !------------------------------------------------------------------
    !---- Compute Frobenius norm of square matrix
    !
    !---- IN
    !---- mat  : matrix to compute the norm of
    !---- ndim : size of mat
    !
    !---- OUT
    !---- norm : Frobenius norm of mat
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                       :: ndim

    REAL(C_DOUBLE), INTENT(OUT)                      :: norm
    REAL(C_DOUBLE), INTENT(IN), DIMENSION(ndim,ndim) :: mat
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, j
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Evaluate norm
    !------------------------------------------------------------------
    norm = 0.0d0
    DO i=1,ndim
       DO j=1,ndim
          norm = norm + (mat(i,j))**2
       ENDDO
    ENDDO
    norm = SQRT(norm)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FROBENIUS_NORM

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GAUSS_PIVOT(mat, ndim)
    !------------------------------------------------------------------
    !---- Gauss pivot elimination
    !
    !---- IN
    !---- mat  : matrix to invert
    !---- ndim : size of matrix
    !
    !---- OUT
    !---- mat  : inverted matrix (erased in the process)
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                          :: ndim

    REAL(C_DOUBLE), INTENT(INOUT), DIMENSION(ndim,ndim) :: mat
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, ipivot, j, k

    REAL(C_DOUBLE) :: pivotmax, pivotmin, var
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Perform Gauss pivot
    !------------------------------------------------------------------
    ipivot    =  0
    pivotmax  =-1.0e+12
    pivotmin  = 1.0e+12

    DO k=1,ndim
       var = 1.0d0/mat(k,k)
       mat(k,k) = 1.0d0

       DO j=1,ndim
          mat(k,j) = mat(k,j)*var
       ENDDO

       IF (ABS(var) < 1.0e-12) ipivot = ipivot + 1

       pivotmax  = MAX(pivotmax, ABS(var))
       pivotmin  = MIN(pivotmin, ABS(var))

       DO i=1,ndim
          IF (i /= k) THEN
             var = mat(i,k)
             mat(i,k)  = 0.0d0

             DO j=1,ndim
                mat(i,j) = mat(i,j) - var*mat(k,j)
             ENDDO
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GAUSS_PIVOT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHOLESKY(mat, ndim)
    !------------------------------------------------------------------
    !---- Cholesky factorization
    !
    !---- IN
    !---- mat  : matrix to invert
    !---- ndim : size of matrix
    !
    !---- OUT
    !---- mat  : inverted matrix (erased in the process)
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !  
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                          :: ndim

    REAL(C_DOUBLE), INTENT(INOUT), DIMENSION(ndim,ndim) :: mat
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, j, k

    REAL(C_DOUBLE) :: som

    REAL(C_DOUBLE), DIMENSION(ndim,ndim) :: Id, L, S
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Perform Cholesky factorization
    !------------------------------------------------------------------
    L  = 0.0d0
    S  = 0.0d0
    Id = 0.0d0

    DO i=1,ndim
       Id(i,i) = 1.0d0
    ENDDO
    !
    !-- L = lower triangular matrix
    DO j=1,ndim
       som = mat(j,j)

       DO k=1,j-1
          som = som - L(j,k)**2
       ENDDO

       L(j,j) = SQRT(som)

       DO i=j+1,ndim
          som = mat(i,j) 

          DO k=1,j-1
             som = som - L(j,k)*L(i,k)
          ENDDO

          L(i,j) = som/L(j,j)
       ENDDO
    ENDDO
    !
    !-- S = conjugate transpose of L
    DO j=1,ndim
       DO i=1,ndim
          som = Id(i,j)

          DO k=1,i-1
             som = som - L(i,k)*S(k,j)
          ENDDO

          S(i,j) = som/L(i,i)
       ENDDO
    ENDDO
    !
    !-- [mat]^-1 = [S]^T * S 
    DO i=1,ndim
       DO j=1,ndim
          mat(i,j) = SUM(S(:,i)*S(:,j))
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHOLESKY

  !======================================================================
  !======================================================================
  !======================================================================

END MODULE MATRICES_STRUCTURES
