MODULE MESH_APPS

  !====================================================================
  ! All subroutines called by mesh apps
  !====================================================================

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_INTERFACES_SUB(readName)
    !------------------------------------------------------------------
    !---- Subroutine for interfaces building
    !
    !---- IN
    !---- readName : name of the mesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Build interfaces
    !------------------------------------------------------------------
    CALL globalTimer%START()
    addInnerFaces     = .FALSE.
    addInterfaceFaces = .TRUE.
    Mesh%name         = readname
    CALL PREPROCESS_MESH
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    Mesh%name = TRIM(Mesh%name)//'_interfaces'
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_INTERFACES_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_LAYER_MODEL_SUB(dataFile, smoothBorder, param)
    !------------------------------------------------------------------
    !---- Subroutine for layer model building
    !
    !---- IN
    !---- dataFile : name of the data file
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE IO_FILES
    USE POINT_CLOUDS
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_STATS
    USE MESH_STRUCTURES
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR)             :: dataFile

    LOGICAL(C_BOOL), OPTIONAL                    :: smoothBorder

    REAL(C_DOUBLE),  OPTIONAL                    :: param
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                               :: iSurf, nSurf, nCols, &
         nArg, nUnit

    REAL(C_DOUBLE)                               :: xyStep

    TYPE(point_cloud), DIMENSION(:), ALLOCATABLE :: layers

    TYPE(Timer)                                  :: globalTimer
    !------------------------------------------------------------------
    !---- Read configuration file and data
    !---- Nothing is assumed by this building tool, so you have to provide
    !---- *all* the layers, even the bottom and top surfaces
    !---- The configuration file is of the following form :
    !----   xyStep
    !----   layer1.dat   additionalThickness   tag tukeymode
    !----   layer2.dat   additionalThickness   tag tukeymode
    !----   ...
    !----
    !---- The additionalThickness value will be added to the values
    !---- of layer*.dat files. The tag value will be that of the
    !---- final mesh file
    !---- If tukeyMode =1, smoothly the border of each layer
    !------------------------------------------------------------------
    CALL globalTimer%START()
    CALL GET_SIZE_OF_FILE(TRIM(dataFile), nSurf, nCols)
    ALLOCATE(layers(nSurf))
    WRITE(6,'(a,i9,a)') 'I found ',nSurf,' surfaces'
    !
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(dataFile))
    DO iSurf=1,nSurf
       READ(nUnit, *) layers(iSurf)%dataFile, layers(iSurf)%offset, &
            layers(iSurf)%tag, layers(iSurf)%xyStep, layers(iSurf)%subSize
       CALL READ_LAYER_DATA(layers(iSurf))
       IF(smoothBorder) CALL MULT_BY_TUKEY_WINDOW(layers(iSurf), param)
    ENDDO
    CLOSE(nUnit)
    !
    DO iSurf=2,nSurf
       layers(iSurf)%altitude = layers(iSurf)%altitude + layers(iSurf-1)%altitude
       IF(smoothBorder) CALL MULT_BY_TUKEY_WINDOW(layers(iSurf), param)
    ENDDO
    !------------------------------------------------------------------
    !---- Build mesh
    !------------------------------------------------------------------
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    Mesh%name      = 'layers'
    Mesh%extension = '.mesh'
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL BUILD_VERTICES_FROM_LAYER_ARRAY(layers)
    CALL BUILD_FACES_FROM_LAYER_ARRAY(layers)
    CALL BUILD_DERIVED_FACES
    CALL WRITE_MESH(Mesh)
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    IF (ALLOCATED(layers))   DEALLOCATE(layers)
    IF (ALLOCATED(MeshVert)) DEALLOCATE(MeshVert)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_LAYER_MODEL_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHANGE_ELEMENT_TAG_SUB(readName, eltType, oldTag, newTag, nChange)
    !------------------------------------------------------------------
    !---- Subroutine for element tag modification
    !
    !---- IN
    !---- readName : name of the mesh
    !---- eltType  : type of element for which the tag must be modified
    !---- oldTag   : tag to modify
    !---- newTag   : modified tag
    !
    !---- OUT
    !---- nChange  : number of modified elements
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName, eltType

    INTEGER(C_INT)                   :: oldTag, newTag, nChange
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                   :: iElt, iCell, iFace

    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Change tag
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    !
    nChange = 0
    !
    SELECT CASE(TRIM(eltType))
    CASE('HEX')
       DO iElt=1,Mesh%nCellHex
          iCell = MeshCellHex(iElt)%cellNum
          IF(MeshCell(iCell)%log == oldTag) THEN
             MeshCell(iCell)%log = newTag
             nChange             = nChange + 1
          ENDIF
       ENDDO
    CASE('TET')
       DO iElt=1,Mesh%nCellTet
          iCell = MeshCellTet(iElt)%cellNum
          IF(MeshCell(iCell)%log == oldTag) THEN
             MeshCell(iCell)%log = newTag
             nChange             = nChange + 1
          ENDIF
       ENDDO
    CASE('QUA')
       DO iElt=1,Mesh%nFacQuad
          iFace = MeshFacQuad(iElt)%faceNum
          IF(MeshFac(iFace)%log == oldTag) THEN
             MeshFac(iFace)%log = newTag
             nChange            = nChange + 1
          ENDIF
       ENDDO
    CASE('TRI')
       DO iElt=1,Mesh%nFacTri
          iFace = MeshFacTri(iElt)%faceNum
          IF(MeshFac(iFace)%log == oldTag) THEN
             MeshFac(iFace)%log = newTag
             nChange            = nChange + 1
          ENDIF
       ENDDO
    CASE DEFAULT
       WRITE(6, '(a)') 'Error in CHANGE_ELEMENT_TAG_SUB'
       WRITE(6, '(a)') 'I received an unknown element type: ', TRIM(eltType)
       WRITE(6, '(a)') 'Allowed element types are TRI, QUA, TET and HEX'
    END SELECT
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHANGE_ELEMENT_TAG_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CONVERT_MESH_SUB(readName, outputExt)
    !------------------------------------------------------------------
    !---- Subroutine for mesh conversion
    !
    !---- IN
    !---- readName  : name of the mesh to convert
    !---- outputExt : output extension
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName, outputExt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    CALL globalTimer%START()
    addInnerFaces     = .TRUE.
    addInterfaceFaces = .TRUE.
    Mesh%name = readName
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    !
    CALL CHECK_OUT_MESH_EXTENSION(outputExt)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CONVERT_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_MESH_STATS_SUB(readName)
    !------------------------------------------------------------------
    !---- Subroutine for mesh stats processing and printing
    !
    !---- IN
    !---- readName : name of the mesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Get stats
    !------------------------------------------------------------------
    CALL globalTimer%START()
    addInnerFaces     = .TRUE.
    addInterfaceFaces = .TRUE.
    Mesh%name         =  TRIM(readname)
    CALL PREPROCESS_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_MESH_STATS_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PARTITION_MESH_SUB(readName, nDomains, hybridCutIn, periodicCutIn)
    !------------------------------------------------------------------
    !---- Subroutine for mesh partitioning
    !
    !---- IN
    !---- readName      : name of the mesh
    !---- nDomains      : number of subdomains
    !---- hybridCutIn   : TRUE if hybrid face cutting is authorized
    !---- periodicCutIn : TRUE if periodic face cutting is authorized
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_STRUCTURES
    USE GRAPH_STRUCTURES
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL), OPTIONAL        :: hybridCutIn, periodicCutIn

    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName

    INTEGER(C_INT)                   :: nDomains
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer, localTimer
    !------------------------------------------------------------------
    !---- Partition mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    CALL localTimer%START()    
    Mesh%name         =  TRIM(readname)
    addInnerFaces     = .TRUE.
    addInterfaceFaces = .FALSE.
    IF (PRESENT(hybridCutIn))   hybridCut   = hybridCutIn
    IF (PRESENT(periodicCutIn)) periodicCut = periodicCutIn
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL PREPROCESS_MESH
    !
    CALL DEALLOCATE_GRAPH
    CALL MESH_TO_GRAPH
    CALL PROCESS_EDGE_WEIGHTS
    CALL WRITE_GRAPH
    CALL PRINT_TIMERS(localTimer)
    !
    CALL PARTITION_GRAPH(nDomains)
    !
    CALL localTimer%START()
    CALL READ_PARTITION (nDomains)
    CALL BUILD_SUBMESHES(nDomains)
    CALL PRINT_TIMERS(localTimer)
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PARTITION_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE REMOVE_DUPLICATE_POINTS_SUB(readName)
    !------------------------------------------------------------------
    !---- Subroutine for duplicate points removal
    !
    !---- IN
    !---- readName : name of the mesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_STRUCTURES
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR)            :: readName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                             :: isModified

    INTEGER(C_INT)                              :: iVert, iFac, iCell, &
         nSorted, vert, nVerts
    INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: sorted

    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: copyVert

    TYPE(Timer)                                 :: globalTimer
    !------------------------------------------------------------------
    !---- Find duplicate points
    !------------------------------------------------------------------
    CALL globalTimer%START()
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(readName)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    !
    CALL REMOVE_VERTICES_DUPLICATES(isModified)
    IF (isModified) THEN
       CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
       CALL WRITE_MESH
    ENDIF
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REMOVE_DUPLICATE_POINTS_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCALE_MESH_SUB(readName, scaling)
    !------------------------------------------------------------------
    !---- Subroutine for mesh scaling
    !
    !---- IN
    !---- readName : name of the mesh
    !---- scaling  : scaling to apply
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STRUCTURES
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName

    REAL(C_DOUBLE)                   :: scaling
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Scale mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    !
    CALL SCALE_MESH_COORDINATES(scaling)
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCALE_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TRANSLATE_MESH_SUB(readName, vector)
    !------------------------------------------------------------------
    !---- Subroutine for mesh translation
    !
    !---- IN
    !---- readName : name of the mesh
    !---- vector   : translation vector
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName

    REAL(C_DOUBLE), DIMENSION(3)     :: vector
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Tanslate mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    !
    CALL TRANSLATE_MESH_COORDINATES(vector)
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TRANSLATE_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ROTATE_MESH_SUB(readName, vector, angle)
    !------------------------------------------------------------------
    !---- Subroutine for mesh rotation
    !
    !---- IN
    !---- readName : name of the mesh
    !---- vector   : axes vector
    !---- angle    : angle of rotation
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName

    REAL(C_DOUBLE)                   :: angle
    REAL(C_DOUBLE), DIMENSION(3)     :: vector
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Translate mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    !
    CALL ROTATE_MESH_COORDINATES(vector, angle)
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ROTATE_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CLEAN_DIOGENES_SUB
    !------------------------------------------------------------------
    !---- Clean all diogenes stuff
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Clean stuff
    !------------------------------------------------------------------
    CALL SYSTEM('&
         rm L2*; &
         rm Probe*; &
         rm Timers*; &
         rm sub*; &
         rm *dgraph*; &
         rm T_P*; &
         rm R_P*; &
         rm A_P*; &
         rm Cabs_P*; &
         rm Csca_P*; &
         rm Cext_P*; &
         rm R_seq; &
         rm T_seq; &
         rm Probe_*_seq; &
         rm *error*; &
         rm partition_mesh; &
         rm *~; &
         rm diff*; &
         rm *.case; &
         rm *.dat; &
         rm *.sol; &
         rm *_view.*; &
         rm diogenes-dgtd;' )
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CLEAN_DIOGENES_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE VIEW_DATA_SUB(configFile)
    !------------------------------------------------------------------
    !---- Build visualizations
    !
    !---- IN
    !---- configFile : configuration file for visualization
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE FIELDS_VIEW
    USE ENSIGHT_MESH
    USE MESH_DATA
    USE MESH_STATS
    USE MESH_STRUCTURES
    USE MESH_INTERFACES
    USE NUMERICAL_TYPES
    USE TIMERS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: configFile
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: localMeshName, localSolName

    INTEGER(C_INT)                   :: iStep, stepNumber

    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Process
    !------------------------------------------------------------------
    addInnerFaces     = .TRUE.
    addInterfaceFaces = .TRUE.
    CALL globalTimer%START()
    !
    CALL READ_CONFIG_VIEW(configFile)
    CALL SORT_INPUT_COMPATIBILITY
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL CHECK_IN_SOLUTION_EXTENSION(Mesh%solutionName)
    CALL CHECK_SOLUTION_FILES
    CALL EXTRACT_SOLUTION_STEP
    localMeshName = TRIM(Mesh%name)
    localSolName  = TRIM(Mesh%solutionName) ! Name without step
    !
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL READ_MESH
    CALL ADJUST_MESH_ORDER
    CALL BUILD_DERIVED_CELLS
    CALL BUILD_DERIVED_FACES
    CALL COMPUTE_MESH_BOUNDING_BOX
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    CALL BUILD_VIEW_PLANES    
    CALL PREPROCESS_CONTINUITY    
    !
    DO iStep=1,n3dView
       stepNumber        = Mesh%initialStep + iStep - 1
       WRITE(6,'(a,i0)') 'Processing solution #',stepNumber
       Mesh%name         = TRIM(localMeshName)
       Mesh%solutionName = TRIM(localSolName) ! Name without step
       Mesh%solutionStep = stepNumber
       !
       CALL ADD_STEP_TO_SOLUTION_NAME
       CALL READ_SOLUTION
       CALL ADJUST_SOLUTION_ORDER
       CALL BUILD_VIEW_PLANES_SOLUTION
       CALL PROCESS_CONTINUITY
       CALL UPDATE_SOLUTION_NAME_FOR_WRITING
       IF (computeVideo) CALL ADD_STEP_TO_SOLUTION_NAME
       CALL WRITE_SOLUTION
    ENDDO
    !
    CALL BUILD_MESH_BOUNDARIES
    CALL BUILD_MESH_INTERFACES    
    IF (computeVideo) CALL EXTRACT_SOLUTION_STEP
    CALL WRITE_MESH
    !
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE VIEW_DATA_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_CUBIC_MESH_SUB(configFile)
    !------------------------------------------------------------------
    !---- Build a regular cubic mesh file
    !
    !---- IN
    !---- configFile : configuration file
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE CUBIC_MESH
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STRUCTURES
    USE MESH_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: configFile
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                  :: isModified

    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Process
    !------------------------------------------------------------------
    CALL globalTimer%START()
    !
    CALL READ_CONFIG_CUBIC_MESH(configFile)
    CALL BUILD_CUBE_VERTICES_AND_CELLS
    CALL BUILD_DERIVED_CELLS
    CALL BUILD_CUBE_FACES
    CALL BUILD_DERIVED_FACES
    CALL REMOVE_VERTICES_DUPLICATES(isModified)
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_CUBIC_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CONVERT_SOLUTION_SUB(meshName, solutionName, solutionExt)
    !------------------------------------------------------------------
    !---- Subroutine for solution format conversion
    !
    !---- IN
    !---- meshName     : name of the mesh     to convert
    !---- solutionName : name of the solution to convert
    !---- solutionExt  : required output solution extension
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: meshName, solutionName, solutionExt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: meshExt

    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Convert mesh and solution formats
    !------------------------------------------------------------------
    CALL globalTimer%START()
    !
    !---- Mesh part
    CALL SOLUTION_TO_MESH_EXTENSIONS(solutionExt, meshExt)
    CALL CONVERT_MESH_SUB(meshName, meshExt)
    !
    !---- Solution part
    Mesh%solutionName = solutionName
    CALL CHECK_IN_SOLUTION_EXTENSION(Mesh%solutionName)
    CALL READ_SOLUTION
    CALL CHECK_OUT_SOLUTION_EXTENSION(solutionExt)
    CALL WRITE_SOLUTION
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CONVERT_SOLUTION_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SYMMETRISE_MESH_SUB(readName, plane)
    !------------------------------------------------------------------
    !---- Subroutine for symmetry
    !
    !---- IN
    !---- readName : name of the mesh
    !---- plane    : plane for the symmetry
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_STRUCTURES
    USE MESH_STATS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    CHARACTER(LEN=2,    KIND=C_CHAR) :: plane
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Translate mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_FACES
    CALL BUILD_DERIVED_CELLS
    !
    CALL COMPUTE_MESH_BOUNDING_BOX
    CALL SYMMETRISE_MESH_COORDINATES(plane, Mesh%coorMin, Mesh%coorMax)
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SYMMETRISE_MESH_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_MESH_SIZE_SUB(readName)
    !------------------------------------------------------------------
    !---- Subroutine for building medit file with element sizes
    !
    !---- IN
    !---- readName : name of the mesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_STRUCTURES
    USE NUMERICAL_TYPES
    USE RAW_MESH
    USE IO_FIELDS
    USE MESH_STATS
    USE BASES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: configName

    INTEGER(C_INT)                   :: iCell
    
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Translate mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL PREPROCESS_MESH
    !
    IF (ALLOCATED(dgEM)) DEALLOCATE(dgEM)
    ALLOCATE(dgEM(Mesh%nCell))
    DO iCell=1,Mesh%nCell
       CALL INITIALIZE_NUMERICAL_ELEMENT(dgEM(iCell))
       !
       dgEM(iCell)%nDofV = MeshCell(iCell)%nVerts
       dgEM(iCell)%order = MeshCell(iCell)%geoOrder
       !
       ALLOCATE(dgEM(iCell)%E%vec(dgEM(iCell)%nDofV,3))
       ALLOCATE(dgEM(iCell)%H%vec(dgEM(iCell)%nDofV,3))
       !
       dgEM(iCell)%E%vec(:,1)  = LOG(1.0/MeshCell(iCell)%measure)
       dgEM(iCell)%H%vec(:,1)  = LOG(1.0/MeshCell(iCell)%measure)
       !
       dgEM(iCell)%E%vec(:,2:) = 0.0
       dgEM(iCell)%H%vec(:,2:) = 0.0
    ENDDO
    !
    CALL BASES_INITIALIZATION('L')
    CALL BUILD_ORDER_ARRAY
    CALL WRITE_RAW_MESH(orderArray)
    CALL WRITE_RAW_SOLUTION
    configName = 'sizeConfig'
    CALL WRITE_MESH_SIZE_CONFIG_FILE(configName)
    CALL SYSTEM('view_data sizeConfig')
    CALL SYSTEM('rm '//TRIM(Mesh%name)//'.rawMesh')
    CALL SYSTEM('rm '//TRIM(Mesh%name)//'_0.rawSol')
    !
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_MESH_SIZE_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_PERIODICITY_SUB(readName)
    !------------------------------------------------------------------
    !---- Subroutine for checking mesh periodicity
    !
    !---- IN
    !---- readName : name of the mesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_STRUCTURES
    USE NUMERICAL_TYPES
    USE MESH_STATS
    USE MESH_NEIGHBORS
    USE MESH_NORMALS_AND_MEASURES
    USE MESH_PERIODIC_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Translate mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    Mesh%name = readname
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)    
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_CELLS
    CALL BUILD_DERIVED_FACES
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    CALL COMPUTE_MESH_BOUNDING_BOX
    CALL BUILD_MESH_NEIGHBORS_AND_FACES
    CALL BUILD_GLOBAL_MESH    
    CALL BUILD_MESH_NORMALS_AND_MEASURES
    CALL COMPUTE_MESH_MINIMAL_SURFACE_AND_VOLUME
    CALL PROCESS_PERIODIC_FACES
    CALL WRITE_COLORED_PERIODIC_MESH
    CALL PRINT_TIMERS(globalTimer)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_PERIODICITY_SUB

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_MESH_BOUNDARY_SUB(readName, boundaryTag)
    !------------------------------------------------------------------
    !---- Subroutine for adding mesh boundary faces when they miss
    !
    !---- IN
    !---- readName    : mesh name
    !---- boundaryTag : tag for boundary faces
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE TIMERS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_INTERFACES
    USE MESH_NEIGHBORS
    USE MESH_STRUCTURES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    
    INTEGER(C_INT)                   :: boundaryTag
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(Timer)                      :: globalTimer
    !------------------------------------------------------------------
    !---- Translate mesh
    !------------------------------------------------------------------
    CALL globalTimer%START()
    addInnerFaces     = .FALSE.
    addInterfaceFaces = .FALSE.
    Mesh%name         = readname
    !
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_MPI_PARALLEL_MESH
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_CELLS
    CALL BUILD_DERIVED_FACES
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    CALL COMPUTE_MESH_BOUNDING_BOX    
    CALL PREPROCESS_CONFORMAL_INNER_FACES
    CALL PROCESS_CONFORMAL_INNER_FACES
    CALL REMOVE_FACE_DUPLICATES    
    CALL FORCE_MESH_BOUNDARIES(boundaryTag)
    !
    CALL CHECK_OUT_MESH_EXTENSION(Mesh%extension)
    Mesh%name = TRIM(Mesh%name)//'_boundary'
    CALL WRITE_MESH
    CALL PRINT_TIMERS(globalTimer)    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_MESH_BOUNDARY_SUB

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_APPS
