MODULE MESH_AXES

  !====================================================================
  ! Routines related to mesh face axes
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_DATA
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_FACE_AXES
    !------------------------------------------------------------------
    !---- This routine builds coordinate axes (not necessarily 
    !---- orthogonal) on faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MATH_PHYS_CONSTANTS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iFace

    REAL(C_DOUBLE)               :: nu, nw
    REAL(C_DOUBLE), DIMENSION(3) :: n, u, w
    !------------------------------------------------------------------
    !---- Compute coordinate axes
    !------------------------------------------------------------------
    DO iFace=1,Mesh%nFac
       SELECT CASE(MeshFac(iFace)%faceType)
       CASE("TRI1","TRI2")
          CALL TRIANGULAR_FACE_AXES(iFace)
       CASE("QUA1","QUA2")
          CALL QUADRANGULAR_FACE_AXES(iFace)
       END SELECT
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !---- Loop over faces and check that the dot products of 
    !---- normal and coordinate axes are zero-like value 'absoluteEps'
    DO iFace=1,Mesh%nFac
       u(:) = MeshFac(iFace)%axis(:,1)
       w(:) = MeshFac(iFace)%axis(:,2)
       n(:) = MeshFac(iFace)%normal(:)
       nu   = DOT_PRODUCT(n,u)
       nw   = DOT_PRODUCT(n,w)
       !
       IF ((nu > absoluteEps ) .OR. (nw > absoluteEps)) THEN
          WRITE(6,'(a)')        "Error in FACE_AXES"
          WRITE(6,'(a,i0)')     "Dot product of normal and coordinate &
               axes was not zero on face ",iFace
          WRITE(6, '(a,e15.7)') "(n|u) = ",nu
          WRITE(6, '(a,e15.7)') "(n|w) = ",nw
          CALL TILT
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished face axes'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_FACE_AXES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TRIANGULAR_FACE_AXES(iFace)
    !------------------------------------------------------------------
    !---- Compute coordinate axes for triangular face iFace
    !
    !---- IN
    !---- iFace : global face index
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_FACES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(3) :: iTriVert

    REAL(C_DOUBLE)               :: norm
    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, ridge1, ridge2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute coordinate axes
    !------------------------------------------------------------------
    !---- Get tet vertex indices
    iTriVert(1:3)           = MeshFac(iFace)%vertex(1:3)
    !
    !---- Get coordinates of the vertices
    v1(:)                   = MeshVert(iTriVert(1))%coor(:)
    v2(:)                   = MeshVert(iTriVert(2))%coor(:)
    v3(:)                   = MeshVert(iTriVert(3))%coor(:)
    !
    !---- Build the ridge vectors
    ridge1(:)               = v2(:) - v1(:)
    ridge2(:)               = v3(:) - v1(:)
    !
    !---- Normalize ridge vectors to obtain coordinate axes
    CALL NORMALIZE_VECTOR(ridge1,norm)
    CALL NORMALIZE_VECTOR(ridge2,norm)
    !
    !---- Fill the axes of MeshFac structure
    MeshFac(iFace)%axis(:,1) = ridge1(:)
    MeshFac(iFace)%axis(:,2) = ridge2(:)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TRIANGULAR_FACE_AXES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE QUADRANGULAR_FACE_AXES(iFace)
    !------------------------------------------------------------------
    !---- Compute coordinate axes for quadrangular face iFace
    !
    !---- IN
    !---- iFace : global face index
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_FACES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(4) :: iQuadVert

    REAL(C_DOUBLE)               :: norm
    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, v4, ridge1, ridge2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute coordinate axes
    !------------------------------------------------------------------
    !---- Get hex vertex indices
    iQuadVert(1:4)          = MeshFac(iFace)%vertex(1:4)
    !
    !---- Get coordinates of the vertices
    v1(:)                   = MeshVert(iQuadVert(1))%coor(:)
    v2(:)                   = MeshVert(iQuadVert(2))%coor(:)
    v3(:)                   = MeshVert(iQuadVert(3))%coor(:)
    v4(:)                   = MeshVert(iQuadVert(4))%coor(:)
    !
    !---- Build the ridge vectors
    ridge1(:)               = v2(:) - v1(:)
    ridge2(:)               = v4(:) - v1(:)
    !
    !---- Normalize ridge vectors to obtain coordinate axes
    CALL NORMALIZE_VECTOR(ridge1,norm)
    CALL NORMALIZE_VECTOR(ridge2,norm)
    !
    !---- Fill the axes of MeshFac structure
    MeshFac(iFace)%axis(:,1) = ridge1(:)
    MeshFac(iFace)%axis(:,2) = ridge2(:)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE QUADRANGULAR_FACE_AXES

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_AXES
