MODULE MESH_CELLS

  !====================================================================
  ! Definition and data structure associated to a generic cell of the mesh
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_FACES
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a generic cell of the mesh
  !
  ! cellType         ==> type of cell
  ! zone             ==> Total or Scattered field. Default is TF
  ! isHO             ==> TRUE if the cell is geometrically high-order
  ! nVerts           ==> number of            vertices of the cell
  ! nLinVerts        ==> number of linear     vertices of the cell, ie the
  !                      regular vertices of the cell
  ! nHoVerts         ==> number of high-order vertices of the cell, ie 
  !                      the additional quadratic vertices
  ! geoOrder         ==> geometrical order of the cell
  ! nFaces           ==> number of geometrical faces of the cell
  ! nNeighb          ==> total number of neighbors (not necessarily
  !                      equal to nFaces)
  ! log              ==> material index of the cell
  ! measure          ==> measure of the cell (i.e. the volume)
  ! surface          ==> total surface of the cell  
  ! nFacNeighb(k)    ==> number of neighbors per geometrical face (k=1:nFaces)
  ! nNeighbFac(k)    ==> number of the geometrical face holding the
  !                      neighbor (k=1:nNeighb)
  ! vertex(k)        ==> index of the kth vertex of the cell (k=1:nVerts)
  ! normal(i,k)      ==> ith component of the normal of the kth face
  !                      (i=1:3, k=1:nNeighb)
  ! neighb(k)        ==> index of the ith neighbor of the cell (k=1:nNeighb)
  ! globalIndex      ==> index of the cell in    GLOBAL numbering
  !                      this is used in parallel for visualization purposes
  ! globalNeighb(k)  ==> same as neighb but with GLOBAL numbering (k=1:nNeighb)
  !                      this is used in parallel for visualization purposes
  ! perm(i,k)        ==> mapping of the vertices between the local
  !                      numbering of the kth neighbor of the cell and
  !                      the local numbering of the adjacent cell
  !                      (i=1:2, k=1:nNeighb). As a convention, 
  !                      i=1 is the current cell, i=2 is the neighbor cell
  ! permType(i,k)    ==> pointer to the right cell/face vertices 
  !                      permutation array (i=1:nFaces,k=1:nVerts)
  ! facSurf(k)       ==> surface of common face with kth neighbor (k=1:nNeighb)
  ! facNum(k)        ==> index of the kth face (k=1:nNeighb)
  ! facLog(k)        ==> log   of the kth face (k=1:nNeighb), this field
  !                      is only used for visualization purposes
  !
  ! Additional details about how faces work:
  ! Faces are uniquely defined. For a hybrid mesh, we chose to consider
  ! only the smallest faces at the non-conformal interface (e.g. at a
  ! quad/tri interface, the only faces that exist are the triangles, the
  ! quad face does not exist in the MeshFac structure).
  ! For this reason, a cell has:
  !       - nFaces geometrical faces (Tet=4 faces, Hex=6 faces, ...)
  !       - nNeighbors numerical faces (eg a simply hybrid hex has 5
  !         quad faces and 2 or more triangular faces)
  !--------------------------------------------------------------------
  TYPE cell
     CHARACTER(LEN=4, KIND=C_CHAR)               :: cellType
     CHARACTER(LEN=2, KIND=C_CHAR)               :: zone

     LOGICAL(C_BOOL)                             :: isHO

     INTEGER(C_INT)                              :: log, nFaces, geoOrder, &
          nVerts, nLinVerts, nHoVerts, nNeighb, globalIndex
     INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: vertex, nFacNeighb,    &
          nNeighbFac, neighb, globalNeighb, facNum, facLog
     INTEGER(C_INT), DIMENSION(:,:), ALLOCATABLE :: perm
     INTEGER(C_INT), DIMENSION(:,:), POINTER     :: permType

     REAL(C_DOUBLE)                              :: measure, surface
     REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: facSurf
     REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: normal
  END TYPE cell

  !--------------------------------------------------------------------
  ! Data structure associated to a derived cell type
  !
  ! cellNum       ==> index of the cell in the MeshCell structure
  !--------------------------------------------------------------------
  TYPE derived_cell
     INTEGER(C_INT) :: cellNum
  END TYPE derived_cell

  !--------------------------------------------------------------------
  ! Global definition of the mesh cells
  !--------------------------------------------------------------------
  TYPE(cell),         DIMENSION(:), ALLOCATABLE, TARGET :: MeshCell

  TYPE(derived_cell), DIMENSION(:), ALLOCATABLE         :: MeshCellLin, &
       MeshCellQuadr, MeshCellCub, MeshCellQuart, MeshCellTet,          &
       MeshCellTetLin, MeshCellTetQuadr, MeshCellTetCub,                &
       MeshCellTetQuart, MeshCellHex, MeshCellHexLin, MeshCellHexQuadr

  !--------------------------------------------------------------------
  ! Miscellaneous stuff
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: cellBlockSize  = 5000
  INTEGER(C_INT) :: dCellBlockSize = 1000
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_CELL(buffer, eltTag, eltType)
    !------------------------------------------------------------------
    !---- Fill all cell lists for a given cell
    !
    !---- IN
    !---- buffer  : vertices of the cell
    !---- eltTag  : logical tag of the cell
    !---- eltType : geometric type of cell
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE DOF_PERMUTATIONS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(INOUT) :: eltType    

    INTEGER(C_INT),                INTENT(IN)    :: eltTag
    INTEGER(C_INT), DIMENSION(:),  INTENT(IN)    :: buffer
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)                :: hoType
    
    INTEGER(C_INT)                               :: nVerts, nLinVerts, &
         nHoVerts, geoOrder, nFaces, nNeighbors, iNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Filling structure
    !------------------------------------------------------------------
    !---- Basic cell informations
    CALL EXTEND_CELL_ARRAY(MeshCell, Mesh%nCell)
    CALL CELL_TYPE_TO_VERTS (eltType, nVerts, nLinVerts, nHoVerts, geoOrder)
    !
    !---- Handle HO mesh
    IF (.NOT. Mesh%useHO) THEN
       nVerts   = nLinVerts
       geoOrder = 1
       nHoVerts = 0
       hoType   = eltType
       CALL CHANGE_TYPE_WITH_ORDER(1, hoType, eltType)
    ENDIF
    !
    !---- Fill informations
    CALL CELL_TYPE_TO_NFACES(eltType, nFaces)
    MeshCell(Mesh%nCell)%cellType               =  eltType
    MeshCell(Mesh%nCell)%nVerts                 =  nVerts
    MeshCell(Mesh%nCell)%nLinVerts              =  nLinVerts
    MeshCell(Mesh%nCell)%nHoVerts               =  nHoVerts
    MeshCell(Mesh%nCell)%geoOrder               =  geoOrder
    MeshCell(Mesh%nCell)%nFaces                 =  nFaces
    ALLOCATE(MeshCell(Mesh%nCell)%vertex(nVerts))
    MeshCell(Mesh%nCell)%vertex(1:nVerts)       =  buffer(1:nVerts)
    MeshCell(Mesh%nCell)%log                    =  eltTag
    !
    !---- In the conforming case, neighbors and faces match together
    !---- Arrays are initialized by default to the conforming case
    !---- They must be reallocated to proper sizes and values if 
    !---- hybrid faces are present
    ALLOCATE(MeshCell(Mesh%nCell)%neighb    (nFaces))
    ALLOCATE(MeshCell(Mesh%nCell)%nFacNeighb(nFaces))
    ALLOCATE(MeshCell(Mesh%nCell)%nNeighbFac(nFaces))
    ALLOCATE(MeshCell(Mesh%nCell)%facSurf   (nFaces))
    ALLOCATE(MeshCell(Mesh%nCell)%facNum    (nFaces))
    ALLOCATE(MeshCell(Mesh%nCell)%perm    (2,nFaces))
    ALLOCATE(MeshCell(Mesh%nCell)%normal  (3,nFaces))
    MeshCell(Mesh%nCell)%neighb(:)              =  0    
    MeshCell(Mesh%nCell)%nFacNeighb(:)          =  1
    MeshCell(Mesh%nCell)%facNum(:)              =  0
    MeshCell(Mesh%nCell)%nNeighb                =  nFaces
    DO iNeighb=1,nFaces
       MeshCell(Mesh%nCell)%nNeighbFac(iNeighb) =  iNeighb
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_CELL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_DERIVED_CELL(eltTag, eltType, cellNum)
    !------------------------------------------------------------------
    !---- Fill all derived cell lists for a given cell
    !
    !---- IN
    !---- eltTag  : logical tag of the cell
    !---- eltType : geometric type of cell
    !---- cellNum : index of the corresponding cell structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE DOF_PERMUTATIONS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: eltType    

    INTEGER(C_INT),                INTENT(IN) :: eltTag, cellNum
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Filling structure
    !------------------------------------------------------------------
    SELECT CASE(eltType)
    CASE('TET','TET1')
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTet,      Mesh%nCellTet)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellLin,      Mesh%nCellLin)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTetLin,   Mesh%nCellTetLin)
       MeshCellTet(Mesh%nCellTet)%cellNum            =  cellNum
       MeshCell(cellNum)%permType                    => nvTet
       MeshCellLin(Mesh%nCellLin)%cellNum            =  cellNum
       MeshCell(cellNum)%isHO                        =  .FALSE.
       MeshCellTetLin(Mesh%nCellTetLin)%cellNum      =  cellNum
    CASE('TET2')
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTet,      Mesh%nCellTet)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellQuadr,    Mesh%nCellQuadr)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTetQuadr, Mesh%nCellTetQuadr)
       MeshCellTet(Mesh%nCellTet)%cellNum            =  cellNum
       MeshCell(cellNum)%permType                    => nvTet
       MeshCellQuadr(Mesh%nCellQuadr)%cellNum        =  cellNum
       MeshCell(cellNum)%isHO                        =  .TRUE.
       MeshCellTetQuadr(Mesh%nCellTetQuadr)%cellNum  =  cellNum
    CASE('TET3')
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTet,      Mesh%nCellTet)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellCub,      Mesh%nCellCub)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTetCub,   Mesh%nCellTetCub)
       MeshCellTet(Mesh%nCellTet)%cellNum            =  cellNum
       MeshCell(cellNum)%permType                    => nvTet
       MeshCellCub(Mesh%nCellCub)%cellNum            =  cellNum
       MeshCell(cellNum)%isHO                        =  .TRUE.
       MeshCellTetCub(Mesh%nCellTetCub)%cellNum      =  cellNum
    CASE('TET4')
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTet,      Mesh%nCellTet)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellQuart,    Mesh%nCellQuart)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellTetQuart, Mesh%nCellTetQuart)
       MeshCellTet(Mesh%nCellTet)%cellNum            =  cellNum
       MeshCell(cellNum)%permType                    => nvTet
       MeshCellQuart(Mesh%nCellQuart)%cellNum        =  cellNum
       MeshCell(cellNum)%isHO                        =  .TRUE.
       MeshCellTetQuart(Mesh%nCellTetQuart)%cellNum  =  cellNum
    CASE('HEX','HEX1')
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellHex,      Mesh%nCellHex)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellLin,      Mesh%nCellLin)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellHexLin,   Mesh%nCellHexLin)
       MeshCellHex(Mesh%nCellHex)%cellNum            =  cellNum
       MeshCell(cellNum)%permType                    => nvHex
       MeshCellLin(Mesh%nCellLin)%cellNum            =  cellNum
       MeshCell(cellNum)%isHO                        =  .FALSE.
       MeshCellHexLin(Mesh%nCellHexLin)%cellNum      =  cellNum
    CASE('HEX2')
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellHex,      Mesh%nCellHex)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellQuadr,    Mesh%nCellQuadr)
       CALL EXTEND_DERIVED_CELL_ARRAY(MeshCellHexQuadr, Mesh%nCellHexQuadr)
       MeshCellHex(Mesh%nCellHex)%cellNum            =  cellNum
       MeshCell(cellNum)%permType                    => nvHex
       MeshCellQuadr(Mesh%nCellQuadr)%cellNum        =  cellNum
       MeshCell(cellNum)%isHO                        =  .TRUE.
       MeshCellHexQuadr(Mesh%nCellHexQuadr)%cellNum  =  cellNum          
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_DERIVED_CELL

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE BUILD_DERIVED_CELLS
    !------------------------------------------------------------------
    !---- Build the derived cell lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build derived cells from scratch
    !------------------------------------------------------------------
    dCellBlockSize = cellBlockSize/2
    CALL DEALLOCATE_DERIVED_CELLS
    DO iCell=1,Mesh%nCell
       CALL ADD_DERIVED_CELL(MeshCell(iCell)%log, MeshCell(iCell)%cellType, iCell)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_DERIVED_CELLS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_CELL_ARRAY(cellArray, cellArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to cellArray
    !
    !---- IN
    !---- cellArray     : array of 'cell' structure
    !---- cellArraySize : number of elements in cellArray
    !
    !---- OUT
    !---- cellArray     : modified in the process
    !---- cellArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),         INTENT(INOUT) :: cellArraySize
    
    TYPE(cell), DIMENSION(:), ALLOCATABLE :: cellArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                       :: isArrayFull, isAllocated

    INTEGER(C_INT)                        :: iCell
    
    TYPE(cell), DIMENSION(:), ALLOCATABLE :: tempCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(cellArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(cellArray) < cellArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempCell)) DEALLOCATE(tempCell)
          ALLOCATE(tempCell(cellArraySize))
          tempCell = cellArray
          DEALLOCATE(cellArray)
       ENDIF
       !
       ALLOCATE(cellArray(cellArraySize + cellBlockSize))
       DO iCell=cellArraySize+1,cellArraySize+cellBlockSize
          CALL INITIALIZE_CELL_STRUCTURE(cellArray(iCell))
       ENDDO
       !
       cellArray(1:cellArraySize) = tempCell(1:cellArraySize)
       IF (ALLOCATED(tempCell)) DEALLOCATE(tempCell)
    ENDIF
    !
    cellArraySize = cellArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_CELL_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_DERIVED_CELL_ARRAY(dCellArray, dCellArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to dCellArray
    !
    !---- IN
    !---- dCellArray     : array of 'derived_cell' structure
    !---- dCellArraySize : number of elements in derivedCellArray
    !
    !---- OUT
    !---- dCellArray     : modified in the process
    !---- dCellArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                                :: dCellArraySize
    
    TYPE(derived_cell), DIMENSION(:), ALLOCATABLE :: dCellArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                               :: isArrayFull, isAllocated
    
    INTEGER(C_INT)                                :: iCell
    
    TYPE(derived_cell), DIMENSION(:), ALLOCATABLE :: tempCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(dCellArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(dCellArray) < dCellArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF    
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempCell)) DEALLOCATE(tempCell)
          ALLOCATE(tempCell(dCellArraySize))
          tempCell = dCellArray
          DEALLOCATE(dCellArray)
       ENDIF
       !
       ALLOCATE(dCellArray(dCellArraySize + dCellBlockSize))
       DO iCell=dCellArraySize+1,dCellArraySize+dCellBlockSize
          CALL INITIALIZE_DERIVED_CELL_STRUCTURE(dCellArray(iCell))
       ENDDO
       !
       dCellArray(1:dCellArraySize) = tempCell(1:dCellArraySize)
       IF (ALLOCATED(tempCell)) DEALLOCATE(tempCell)
    ENDIF
    !
    dCellArraySize = dCellArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_DERIVED_CELL_ARRAY  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_CELL_STRUCTURE(cellStructure)
    !------------------------------------------------------------------
    !---- Initialize cell structure 
    !
    !---- IN
    !---- cellStructure : cell structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(cell) :: cellStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize cell structure
    !------------------------------------------------------------------
    cellStructure%cellType    = 'NONE'
    cellStructure%zone        = 'TF'
    cellStructure%isHO        = .FALSE.
    cellStructure%nVerts      = 0
    cellStructure%nLinVerts   = 0
    cellStructure%nHoVerts    = 0
    cellStructure%geoOrder    = 0
    cellStructure%nFaces      = 0
    cellStructure%nNeighb     = 0
    cellStructure%log         = 0
    cellStructure%measure     = 0.0d0
    cellStructure%surface     = 0.0d0
    cellStructure%globalIndex = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_CELL_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_DERIVED_CELL_STRUCTURE(derivedCellStructure)
    !------------------------------------------------------------------
    !---- Initialize derived cell structure 
    !
    !---- IN
    !---- derivedCellStructure : derived cell structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(derived_cell) :: derivedCellStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize derived cell structure
    !------------------------------------------------------------------
    derivedCellStructure%cellNum =  0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_DERIVED_CELL_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_CELLS
    !------------------------------------------------------------------
    !---- Deallocate all cell lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate cells
    !------------------------------------------------------------------
    IF (ALLOCATED(MeshCell)) DEALLOCATE(MeshCell)
    CALL DEALLOCATE_DERIVED_CELLS
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_CELLS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_DERIVED_CELLS
    !------------------------------------------------------------------
    !---- Deallocate all derived cell lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate cells
    !------------------------------------------------------------------
    CALL RESET_MESH_DERIVED_CELL_COUNTERS(Mesh)
    !
    IF (ALLOCATED(MeshCellLin))      DEALLOCATE(MeshCellLin)
    IF (ALLOCATED(MeshCellQuadr))    DEALLOCATE(MeshCellQuadr)
    IF (ALLOCATED(MeshCellCub))      DEALLOCATE(MeshCellCub)
    IF (ALLOCATED(MeshCellQuart))    DEALLOCATE(MeshCellQuart)
    IF (ALLOCATED(MeshCellTet))      DEALLOCATE(MeshCellTet)
    IF (ALLOCATED(MeshCellTetLin))   DEALLOCATE(MeshCellTetLin)
    IF (ALLOCATED(MeshCellTetQuadr)) DEALLOCATE(MeshCellTetQuadr)
    IF (ALLOCATED(MeshCellTetCub))   DEALLOCATE(MeshCellTetCub)
    IF (ALLOCATED(MeshCellTetQuart)) DEALLOCATE(MeshCellTetQuart)
    IF (ALLOCATED(MeshCellHex))      DEALLOCATE(MeshCellHex)
    IF (ALLOCATED(MeshCellHexLin))   DEALLOCATE(MeshCellHexLin)
    IF (ALLOCATED(MeshCellHexQuadr)) DEALLOCATE(MeshCellHexQuadr)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_DERIVED_CELLS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPARE_CELL_STRUCTURES(cell1, cell2, isIdentical)
    !------------------------------------------------------------------
    !---- Tell if 2 cell structure arrays are identical
    !
    !---- IN
    !---- cell*       : cell structures to compare
    !
    !---- OUT
    !---- isIdentical : TRUE if structures are identical
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),          INTENT(OUT) :: isIdentical
    
    TYPE(cell), DIMENSION(:), INTENT(IN)  :: cell1, cell2
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                        :: size1, size2, iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compare structures
    !------------------------------------------------------------------
    isIdentical = .TRUE.
    size1       = SIZE(cell1)
    size2       = SIZE(cell2)
    !
    IF (size1 /= size2) THEN
       isIdentical = .FALSE.
       RETURN
    ENDIF
    !
    DO iCell=1,size1
       IF (cell1(iCell)%cellType         /= cell2(iCell)%cellType)        isIdentical = .FALSE.
       IF (cell1(iCell)%isHO         .NEQV. cell2(iCell)%isHO)            isIdentical = .FALSE.
       IF (cell1(iCell)%nVerts           /= cell2(iCell)%nVerts)          isIdentical = .FALSE.
       IF (cell1(iCell)%nLinVerts        /= cell2(iCell)%nLinVerts)       isIdentical = .FALSE.
       IF (cell1(iCell)%nHoVerts         /= cell2(iCell)%nHoVerts)        isIdentical = .FALSE.
       IF (cell1(iCell)%geoOrder         /= cell2(iCell)%geoOrder)        isIdentical = .FALSE.
       IF (cell1(iCell)%nFaces           /= cell2(iCell)%nFaces)          isIdentical = .FALSE.
       IF (cell1(iCell)%nNeighb          /= cell2(iCell)%nNeighb)         isIdentical = .FALSE.
       IF (cell1(iCell)%log              /= cell2(iCell)%log)             isIdentical = .FALSE.
       IF (cell1(iCell)%measure          /= cell2(iCell)%measure)         isIdentical = .FALSE.
       IF (cell1(iCell)%surface          /= cell2(iCell)%surface)         isIdentical = .FALSE.
       !
       IF (ALLOCATED(cell1(iCell)%vertex)     .AND. ALLOCATED(cell2(iCell)%vertex))     THEN
          IF (.NOT. ALL(cell1(iCell)%vertex      == cell2(iCell)%vertex))     isIdentical = .FALSE.
       ENDIF
       !
       IF (ALLOCATED(cell1(iCell)%nFacNeighb) .AND. ALLOCATED(cell2(iCell)%nFacNeighb)) THEN
          IF (.NOT. ALL(cell1(iCell)%nFacNeighb  == cell2(iCell)%nFacNeighb)) isIdentical = .FALSE.
       ENDIF
       !
       IF (ALLOCATED(cell1(iCell)%nNeighbFac) .AND. ALLOCATED(cell2(iCell)%nNeighbFac)) THEN
          IF (.NOT. ALL(cell1(iCell)%nNeighbFac  == cell2(iCell)%nNeighbFac)) isIdentical = .FALSE.
       ENDIF
       !
       IF (ALLOCATED(cell1(iCell)%neighb)     .AND. ALLOCATED(cell2(iCell)%neighb))     THEN
          IF (.NOT. ALL(cell1(iCell)%neighb      == cell2(iCell)%neighb))     isIdentical = .FALSE.
       ENDIF
       !
       IF (ALLOCATED(cell1(iCell)%facNum)     .AND. ALLOCATED(cell2(iCell)%facNum))     THEN
          IF (.NOT. ALL(cell1(iCell)%facNum      == cell2(iCell)%facNum))     isIdentical = .FALSE.
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPARE_CELL_STRUCTURES  

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_CELLS
