MODULE MESH_DATA

  !====================================================================
  ! Definition and data structure associated to the mesh
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! General data
  !--------------------------------------------------------------------
  ! name            ==> name of the mesh
  ! extension       ==> extension of the mesh when read
  ! hAvg            ==> average distance between mesh vertices
  ! hMax            ==> maximum distance between mesh vertices
  ! hMin            ==> minimum distance between mesh vertices
  ! surfMin         ==> minimal surface of 2D element
  ! volMin          ==> minimal volume  of 3D element
  ! measure         ==> measure of the mesh (i.e. total volume of the mesh)
  ! coorMax         ==> maximal values of the bounding box (xmax, ymax, zmax)
  ! coorMin         ==> minimal values of the bounding box (xmin, ymin, zmin)
  ! coorCenter      ==> center of the mesh
  ! boxSize         ==> size of the bounding box
  ! tfsfMax         ==> maximal values of the tfsf box (xmax, ymax, zmax)
  ! tfsfMin         ==> minimal values of the tfsf box (xmin, ymin, zmin)
  ! tolerance1D     ==> geometrical tolerance for test on 1D quantities
  ! tolerance2D     ==> geometrical tolerance for test on 2D quantities
  ! tolerance3D     ==> geometrical tolerance for test on 3D quantities
  !
  ! solName         ==> name        of the associated solution
  ! solExtension    ==> extension   of the associated solution
  ! solStep         ==> step        of the associated solution
  ! initialStep     ==> intial step of the associated solution
  !--------------------------------------------------------------------
  ! Cells data
  !--------------------------------------------------------------------
  ! nVert           ==> # of vertices of the mesh
  ! nCell           ==> # of cells
  ! nCellTet        ==> # of tetrahedral cells
  ! nCellHex        ==> # of hexahedral cells
  ! nCellLin        ==> # of linear cells
  ! nCellQuadr      ==> # of quadratic cells
  ! nCellCub        ==> # of cubic cells
  ! nCellQuart      ==> # of quartic cells
  ! nCellTetLin     ==> # of tetrahedral linear cells
  ! nCellTetQuadr   ==> # of tetrahedral quadratic cells
  ! nCellTetCub     ==> # of tetrahedral cubic cells
  ! nCellTetQuart   ==> # of tetrahedral quartic cells
  ! nCellHexLin     ==> # of hexahedral linear cells
  ! nCellHexCurv    ==> # of hexahedral curved cells
  !
  ! nCellTfZone     ==> # of cells in the TF zone
  ! nCellSfZone     ==> # of cells in the SF zone
  !--------------------------------------------------------------------
  ! Faces data
  !--------------------------------------------------------------------
  ! nFac            ==> total # of faces to allocate as face structure
  !                     nFac = nFacB + nFacTfsf + nFacPoy + nFacI
  !                     nFac = nFacTri + nFacQuad
  !                     nFac = nFacLin + nFacCurv
  !
  ! nFacTri         ==> # of triangular   faces of the mesh
  ! nFacQuad        ==> # of quadrangular faces of the mesh
  !
  ! nFacB           ==> total # of boundary faces (including artificial faces)
  !                     nFacB    = nFacAbs + nFacPec + nFacPmc + nFacPer + nFacArt
  ! nFacPhy         ==> # of physical boundary faces of the mesh
  !                     nFacBphy = nFacBabs + nFacBpec + nFacBpmc + nFacBper
  ! nFacAbs         ==> # of absorbing  boundary faces of the mesh
  ! nFacPec         ==> # of PEC        boundary faces of the mesh
  ! nFacPmc         ==> # of PMC        boundary faces of the mesh
  ! nFacPer         ==> # of periodic   boundary faces of the mesh
  ! nFacArt         ==> # of artificial boundary faces of the mesh
  ! nFacArtPer      ==> # of artificial periodic faces
  !
  ! nFacI           ==> # of purely inner faces of the mesh
  ! nFacTfsf        ==> # of TF/SF        faces of the mesh
  ! nFacPoy         ==> # of Poynting     faces of the mesh
  ! nFacHyb         ==> # of hybrid       faces of the mesh
  ! nFacInter       ==> # of material/material interface faces
  !
  ! nFacLin         ==> # of linear      faces of the mesh
  ! nFacQuadr       ==> # of quadratic   faces of the mesh
  ! nFacTriLin      ==> # of linear tri  faces of the mesh
  ! nFacTriQuadr    ==> # of curved tri  faces of the mesh
  ! nFacQuadLin     ==> # of linear quad faces of the mesh
  ! nFacQuadQuadr   ==> # of curved quad faces of the mesh
  !
  ! isTfsfDisjoint  ==> TRUE if TF/SF surface is disjoint
  ! useHO           ==> TRUE if high-order elements must be considered
  !                     if FALSE, all elements are build as linear elements
  !--------------------------------------------------------------------
  TYPE mesh_structure
     CHARACTER(LEN=1000, KIND=C_CHAR) :: name, extension, solutionName, &
          solutionExtension

     LOGICAL(C_BOOL)                  :: isTfsfDisjoint, useHO

     INTEGER(C_INT)                   :: nVert, nCell, nCellTet,        &
          nCellHex, nCellLin, nCellQuadr, nCellCub, nCellQuart,         &
          nCellTetLin, nCellTetQuadr, nCellTetCub, nCellTetQuart,       &
          nCellHexLin, nCellHexQuadr, nFac, nFacTri, nFacQuad, nFacB,   &
          nFacPhy, nFacAbs, nFacPec, nFacPmc, nFacPer, nFacArt, nFacI,  &
          nFacTfsf, nFacPoy, nFacLin, nFacQuadr, nFacTriLin,            &
          nFacTriQuadr, nFacQuadLin, nFacQuadQuadr, nFacInter, nFacHyb, &
          nCellTfZone, nCellSfZone, nSfZone, nFacArtPer, solutionStep,  &
          initialStep

     REAL(C_DOUBLE)                   :: hAvg, hMax, hMin, measure,     &
          surfMin, volMin, tolerance1D, tolerance2D, tolerance3D
     REAL(C_DOUBLE), DIMENSION(3)     :: coorMax, coorMin, tfsfMax,     &
          tfsfMin, coorCenter, boxSize
  END TYPE mesh_structure

  !--------------------------------------------------------------------
  ! Global mesh information
  !--------------------------------------------------------------------
  ! Mesh       ==> local Mesh
  ! globalMesh ==> global mesh in case of parallel run
  !--------------------------------------------------------------------
  TYPE(mesh_structure) :: Mesh, globalMesh
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_MESH_EMPTY(locMesh)
    !------------------------------------------------------------------
    !---- Test if mesh contains any kind of elements
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: errortest
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing if locMesh is empty
    !------------------------------------------------------------------
    errortest = locMesh%nVert + locMesh%nFac + locMesh%nCell
    !
    IF (errortest == 0) THEN
       WRITE(6, '(a)') 'Mesh seems empty'
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_MESH_EMPTY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_MESH_STRUCTURE(locMesh)
    !------------------------------------------------------------------
    !---- Fill mesh with default values
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    locMesh%hAvg              = 0.0
    locMesh%hMax              =-1.0e+20
    locMesh%hMin              = 1.0e+20
    locMesh%surfMin           = 1.0e+20
    locMesh%volMin            = 1.0e+20
    locMesh%measure           = 0.0
    locMesh%coorMax(:)        =-1.0e+20
    locMesh%coorMin(:)        = 1.0e+20
    locMesh%coorCenter(:)     = 0.0
    locMesh%boxSize(:)        = 0.0
    locMesh%tfsfMax(:)        =-1.0e+20
    locMesh%tfsfMin(:)        = 1.0e+20
    locMesh%useHO             = .TRUE.
    locMesh%solutionStep      = 0
    !
    CALL RESET_MESH_VERTEX_COUNTERS(locMesh)
    CALL RESET_MESH_FACE_COUNTERS  (locMesh)
    CALL RESET_MESH_CELL_COUNTERS  (locMesh)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_MESH_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPARE_MESH_STRUCTURES(mesh1, mesh2, isIdentical)
    !------------------------------------------------------------------
    !---- Tell if 2 mesh structures are identical
    !
    !---- IN
    !---- mesh*       : mesh structures to compare
    !
    !---- OUT
    !---- isIdentical : TRUE if structures are identical
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),      INTENT(OUT) :: isIdentical

    TYPE(mesh_structure), INTENT(IN)  :: mesh1, mesh2
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compare structures
    !------------------------------------------------------------------
    isIdentical = .TRUE.
    !
    IF (mesh1%nVert         /= mesh2%nVert)         isIdentical = .FALSE.
    IF (mesh1%nFac          /= mesh2%nFac)          isIdentical = .FALSE.
    IF (mesh1%nFacTri       /= mesh2%nFacTri)       isIdentical = .FALSE.
    IF (mesh1%nFacTriLin    /= mesh2%nFacTriLin)    isIdentical = .FALSE.
    IF (mesh1%nFacTriQuadr  /= mesh2%nFacTriQuadr)  isIdentical = .FALSE.
    IF (mesh1%nFacQuad      /= mesh2%nFacQuad)      isIdentical = .FALSE.
    IF (mesh1%nFacQuadLin   /= mesh2%nFacQuadLin)   isIdentical = .FALSE.
    IF (mesh1%nFacQuadQuadr /= mesh2%nFacQuadQuadr) isIdentical = .FALSE.
    IF (mesh1%nFacB         /= mesh2%nFacB)         isIdentical = .FALSE.
    IF (mesh1%nFacPhy       /= mesh2%nFacPhy)       isIdentical = .FALSE.
    IF (mesh1%nFacAbs       /= mesh2%nFacAbs)       isIdentical = .FALSE.
    IF (mesh1%nFacPec       /= mesh2%nFacPec)       isIdentical = .FALSE.
    IF (mesh1%nFacPmc       /= mesh2%nFacPmc)       isIdentical = .FALSE.
    IF (mesh1%nFacPer       /= mesh2%nFacPer)       isIdentical = .FALSE.
    IF (mesh1%nFacArt       /= mesh2%nFacArt)       isIdentical = .FALSE.
    IF (mesh1%nFacArtPer    /= mesh2%nFacArtPer)    isIdentical = .FALSE.
    IF (mesh1%nFacI         /= mesh2%nFacI)         isIdentical = .FALSE.
    IF (mesh1%nFacTfsf      /= mesh2%nFacTfsf)      isIdentical = .FALSE.
    IF (mesh1%nFacPoy       /= mesh2%nFacPoy)       isIdentical = .FALSE.
    IF (mesh1%nFacLin       /= mesh2%nFacLin)       isIdentical = .FALSE.
    IF (mesh1%nFacQuadr     /= mesh2%nFacQuadr)     isIdentical = .FALSE.
    IF (mesh1%nFacInter     /= mesh2%nFacInter)     isIdentical = .FALSE.
    IF (mesh1%nFacHyb       /= mesh2%nFacHyb)       isIdentical = .FALSE.
    IF (mesh1%nCell         /= mesh2%nCell)         isIdentical = .FALSE.
    IF (mesh1%nCellTet      /= mesh2%nCellTet)      isIdentical = .FALSE.
    IF (mesh1%nCellHex      /= mesh2%nCellHex)      isIdentical = .FALSE.
    IF (mesh1%nCellLin      /= mesh2%nCellLin)      isIdentical = .FALSE.
    IF (mesh1%nCellQuadr    /= mesh2%nCellQuadr)    isIdentical = .FALSE.
    IF (mesh1%nCellCub      /= mesh2%nCellCub)      isIdentical = .FALSE.
    IF (mesh1%nCellQuart    /= mesh2%nCellQuart)    isIdentical = .FALSE.
    IF (mesh1%nCellTetLin   /= mesh2%nCellTetLin)   isIdentical = .FALSE.
    IF (mesh1%nCellTetQuadr /= mesh2%nCellTetQuadr) isIdentical = .FALSE.
    IF (mesh1%nCellTetCub   /= mesh2%nCellTetCub)   isIdentical = .FALSE.
    IF (mesh1%nCellTetQuart /= mesh2%nCellTetQuart) isIdentical = .FALSE.
    IF (mesh1%nCellHexLin   /= mesh2%nCellHexLin)   isIdentical = .FALSE.
    IF (mesh1%nCellHexQuadr /= mesh2%nCellHexQuadr) isIdentical = .FALSE.
    IF (mesh1%nCellTfZone   /= mesh2%nCellTfZone)   isIdentical = .FALSE.
    IF (mesh1%nCellSfZone   /= mesh2%nCellSfZone)   isIdentical = .FALSE.
    IF (mesh1%hAvg          /= mesh2%hAvg)          isIdentical = .FALSE.
    IF (mesh1%hMax          /= mesh2%hMax)          isIdentical = .FALSE.
    IF (mesh1%hMin          /= mesh2%hMin)          isIdentical = .FALSE.
    IF (mesh1%surfMin       /= mesh2%surfMin)       isIdentical = .FALSE.
    IF (mesh1%volMin        /= mesh2%volMin)        isIdentical = .FALSE.
    IF (mesh1%measure       /= mesh2%measure)       isIdentical = .FALSE.
    !
    IF (.NOT. ALL(mesh1%coorMax == mesh2%coorMax)) isIdentical = .FALSE.
    IF (.NOT. ALL(mesh1%coorMin == mesh2%coorMin)) isIdentical = .FALSE.
    IF (.NOT. ALL(mesh1%tfsfMax == mesh2%tfsfMax)) isIdentical = .FALSE.
    IF (.NOT. ALL(mesh1%tfsfMin == mesh2%tfsfMin)) isIdentical = .FALSE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPARE_MESH_STRUCTURES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RESET_MESH_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Put all mesh counters to zero
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    CALL RESET_MESH_VERTEX_COUNTERS(locMesh)
    CALL RESET_MESH_FACE_COUNTERS  (locMesh)
    CALL RESET_MESH_CELL_COUNTERS  (locMesh)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RESET_MESH_COUNTERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RESET_MESH_VERTEX_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Put all mesh vertices counters to zero
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    locMesh%nVert = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RESET_MESH_VERTEX_COUNTERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RESET_MESH_FACE_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Put all mesh face counters to zero
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    locMesh%nFac = 0
    CALL RESET_MESH_DERIVED_FACE_COUNTERS(locMesh)
    locMesh%isTfsfDisjoint = .TRUE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RESET_MESH_FACE_COUNTERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RESET_MESH_DERIVED_FACE_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Put all mesh derived face counters to zero
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    locMesh%nFacTri       = 0
    locMesh%nFacTriLin    = 0
    locMesh%nFacTriQuadr  = 0
    locMesh%nFacQuad      = 0
    locMesh%nFacQuadLin   = 0
    locMesh%nFacQuadQuadr = 0
    locMesh%nFacB         = 0
    locMesh%nFacPhy       = 0
    locMesh%nFacAbs       = 0
    locMesh%nFacPec       = 0
    locMesh%nFacPmc       = 0
    locMesh%nFacPer       = 0
    locMesh%nFacArt       = 0
    locMesh%nFacArtPer    = 0
    locMesh%nFacI         = 0
    locMesh%nFacTfsf      = 0
    locMesh%nFacPoy       = 0
    locMesh%nFacLin       = 0
    locMesh%nFacQuadr     = 0
    locMesh%nFacInter     = 0
    locMesh%nFacHyb       = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RESET_MESH_DERIVED_FACE_COUNTERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RESET_MESH_CELL_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Put all mesh cell counters to zero
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    locMesh%nCell = 0
    CALL RESET_MESH_DERIVED_CELL_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RESET_MESH_CELL_COUNTERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RESET_MESH_DERIVED_CELL_COUNTERS(locMesh)
    !------------------------------------------------------------------
    !---- Put all mesh derived cell counters to zero
    !
    !---- IN
    !---- locMesh : mesh to modify
    !
    !---- OUT
    !---- locMesh : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(mesh_structure), INTENT(OUT) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize Mesh structure
    !------------------------------------------------------------------
    locMesh%nCellTet      = 0
    locMesh%nCellHex      = 0
    locMesh%nCellLin      = 0
    locMesh%nCellQuadr    = 0
    locMesh%nCellCub      = 0
    locMesh%nCellQuart    = 0
    locMesh%nCellTetLin   = 0
    locMesh%nCellTetQuadr = 0
    locMesh%nCellTetCub   = 0
    locMesh%nCellTetQuart = 0
    locMesh%nCellHexLin   = 0
    locMesh%nCellHexQuadr = 0
    locMesh%nCellTfZone   = 0
    locMesh%nCellSfZone   = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RESET_MESH_DERIVED_CELL_COUNTERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTRACT_SOLUTION_STEP
    !------------------------------------------------------------------
    !---- Extract the index of the solution step and update name
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=10, KIND=C_CHAR) :: solStepC

    LOGICAL(C_BOOL)                :: isAnInt

    INTEGER(C_INT)                 :: sizeUnd
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check for the existence of an underscore and a dot in 
    !---- the name, and retrieve the solution index in between.
    !---- Then, rebuild solution name without index
    !------------------------------------------------------------------
    sizeUnd  = SCAN(ADJUSTL(Mesh%solutionName), '_', .TRUE.)
    solStepC = TRIM(Mesh%solutionName(sizeUnd+1:))    
    !
    !---- If what is after the underscore is not an int
    CALL IS_AN_INT(solStepC, isAnInt)
    IF (.NOT. isAnInt) RETURN
    !
    CALL CHAR_TO_INT(solStepC, Mesh%solutionStep)
    Mesh%solutionName = TRIM(Mesh%solutionName(1:sizeUnd-1))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTRACT_SOLUTION_STEP

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_STEP_TO_SOLUTION_NAME
    !------------------------------------------------------------------
    !---- Modify solution name for solution reading from file
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR) :: stepC
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Add the step number at the end of the solution file name
    !------------------------------------------------------------------
    CALL INT_TO_CHAR(Mesh%solutionStep,stepC)
    !
    Mesh%solutionName = TRIM(Mesh%solutionName)//'_'//TRIM(stepC)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_STEP_TO_SOLUTION_NAME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE UPDATE_SOLUTION_NAME_FOR_WRITING
    !------------------------------------------------------------------
    !---- Modify solution name for solution writing from file
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- If video is not used, we just output one solution file, 
    !---- so we remove the step in the name
    !------------------------------------------------------------------
    CALL EXTRACT_SOLUTION_STEP
    !
    Mesh%name         = TRIM(Mesh%name)//'_view'
    Mesh%solutionName = TRIM(Mesh%solutionName)//'_view'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE UPDATE_SOLUTION_NAME_FOR_WRITING

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_DATA
