MODULE MESH_ELEMENTS

  !====================================================================
  ! Routines related to general actions on elements
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE DOF_PERMUTATIONS
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Procedure pointers for conversion from polynomial order to number
  ! of degrees of freedom
  !--------------------------------------------------------------------
  PROCEDURE(), POINTER :: ORDER_TO_NDOFV => NULL()
  PROCEDURE(), POINTER :: ORDER_TO_NDOFS => NULL()

  !--------------------------------------------------------------------
  ! Buffer for permutations at reading
  ! In case of doubt see headers in DOF permutations
  !--------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(6)  :: triPerm
  INTEGER(C_INT), DIMENSION(9)  :: quadPerm
  INTEGER(C_INT), DIMENSION(35) :: tetPerm
  INTEGER(C_INT), DIMENSION(27) :: hexPerm
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_ELEMENT(buffer, eltTag, eltType)
    !------------------------------------------------------------------
    !---- Fill element structures depending on element type, tag 
    !---- and options
    !
    !---- IN
    !---- buffer            : vertices of the element
    !---- eltTag            : logical tag of the element
    !---- eltType           : type of element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                INTENT(INOUT) :: eltTag
    INTEGER(C_INT), DIMENSION(:)                 :: buffer

    CHARACTER(LEN=4, KIND=C_CHAR), INTENT(INOUT) :: eltType
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                               :: nVerts, length, facNum
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check extension to determine the right format permutation
    !---- for high-order vertices
    !------------------------------------------------------------------
    SELECT CASE(TRIM(Mesh%extension))
    CASE('.mesh','.meshb')
       triPerm (1:nsTriMaxMedit)    = nsTriMeditToInria (1:nsTriMaxMedit)
       tetPerm (1:nsTetMaxMedit)    = nvTetMeditToInria (1:nsTetMaxMedit)
       quadPerm(1:nsQuadMaxMedit)   = nsQuadMeditToInria(1:nsQuadMaxMedit)
       hexPerm (1:nsHexMaxMedit)    = nvHexMeditToInria (1:nsHexMaxMedit)
    CASE('.nmesh')
       triPerm (1:nsTriMaxInria)    = nsTriIdentity (1:nsTriMaxInria)
       tetPerm (1:nsTetMaxInria)    = nvTetIdentity (1:nsTetMaxInria)
       quadPerm(1:nsQuadMaxInria)   = nsQuadIdentity(1:nsQuadMaxInria)
       hexPerm (1:nsHexMaxInria)    = nvHexIdentity (1:nsHexMaxInria)
    CASE('.msh')
       triPerm (1:nsTriMaxGmsh)     = nsTriGmshToInria (1:nsTriMaxGmsh)
       tetPerm (1:nsTetMaxGmsh)     = nvTetGmshToInria (1:nsTetMaxGmsh)
       quadPerm(1:nsQuadMaxGmsh)    = nsQuadGmshToInria(1:nsQuadMaxGmsh)
       hexPerm (1:nsHexMaxGmsh)     = nvHexGmshToInria (1:nsHexMaxGmsh)
    CASE('.rawMesh')
       triPerm (1:nsTriMaxRaw)      = nsTriIdentity (1:nsTriMaxRaw)
       tetPerm (1:nsTetMaxRaw)      = nvTetIdentity (1:nsTetMaxRaw)
       quadPerm(1:nsQuadMaxRaw)     = nsQuadIdentity(1:nsQuadMaxRaw)
       hexPerm (1:nsHexMaxRaw)      = nvHexIdentity (1:nsHexMaxRaw)
    CASE('.geo')
       triPerm (1:nsTriMaxEnsight)  = nsTriInriaToEnsight (1:nsTriMaxEnsight)
       tetPerm (1:nsTetMaxEnsight)  = nvTetInriaToEnsight (1:nsTetMaxEnsight)
       quadPerm(1:nsQuadMaxEnsight) = nsQuadInriaToEnsight(1:nsQuadMaxEnsight)
       hexPerm (1:nsHexMaxEnsight)  = nvHexInriaToEnsight (1:nsHexMaxEnsight)
    END SELECT
    !------------------------------------------------------------------
    !---- Sort eltType
    !------------------------------------------------------------------
    length = SIZE(buffer)
    !
    SELECT CASE(eltType)
    CASE('TRI1','TRI2')
       buffer(1:length) = buffer(triPerm(1:length))
       CALL ADD_FACE(buffer(1:length), eltTag, eltType, facNum)
    CASE('QUA1','QUA2')
       buffer(1:length) = buffer(quadPerm(1:length))
       CALL ADD_FACE(buffer(1:length), eltTag, eltType, facNum)
    CASE('TET1','TET2','TET3','TET4')
       buffer(1:length) = buffer(tetPerm(1:length))
       CALL ADD_CELL(buffer(1:length), eltTag, eltType)
    CASE('HEX1','HEX2')
       buffer(1:length) = buffer(hexPerm(1:length))
       CALL ADD_CELL(buffer(1:length), eltTag, eltType)
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_ELEMENT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_MESH_STRUCTURES
    !------------------------------------------------------------------
    !---- Deallocate everything related to the mesh
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate everything
    !------------------------------------------------------------------
    CALL DEALLOCATE_VERTICES
    CALL DEALLOCATE_FACES
    CALL DEALLOCATE_CELLS
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_MESH_STRUCTURES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_FACE_LINEAR_VERTICES_FROM_CELL(iCell, iFacLoc, vFace)
    !------------------------------------------------------------------
    !---- Get face linear vertices from the cell index and the local face index
    !
    !---- IN
    !---- iCell    : cell index
    !---- iFacLoc  : local face index
    !
    !---- OUT
    !---- vFace    : face vertices
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE DOF_PERMUTATIONS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(INOUT) :: iCell, iFacLoc
    INTEGER(C_INT), DIMENSION(:), INTENT(INOUT) :: vFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)               :: cellType, faceType

    INTEGER(C_INT)                              :: nVerts, nLinVerts, nHoVerts, geoOrder
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call the appropriate method depending on faceType
    !------------------------------------------------------------------
    cellType = MeshCell(iCell)%cellType
    CALL CELL_TYPE_TO_FACE_TYPE(cellType,faceType)
    CALL FACE_TYPE_TO_VERTS(faceType,nVerts,nLinVerts,nHoVerts,geoOrder)
    !
    SELECT CASE (cellType)
    CASE('TET','TET1','TET2','TET3','TET4')
       vFace(1:nLinVerts) = MeshCell(iCell)%vertex(nvTet(iFacLoc,1:nLinVerts))
    CASE('HEX','HEX1','HEX2')
       vFace(1:nLinVerts) = MeshCell(iCell)%vertex(nvHex(iFacLoc,1:nLinVerts))
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_FACE_LINEAR_VERTICES_FROM_CELL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_FACE_VERTICES_FROM_CELL(iCell, iFacLoc, vFace)
    !------------------------------------------------------------------
    !---- Get face vertices from the cell index and the local face index
    !
    !---- IN
    !---- iCell    : cell index
    !---- iFacLoc  : local face index
    !
    !---- OUT
    !---- vFace    : face vertices
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE DOF_PERMUTATIONS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(INOUT) :: iCell, iFacLoc
    INTEGER(C_INT), DIMENSION(:), INTENT(INOUT) :: vFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)               :: cellType, faceType

    INTEGER(C_INT)                              :: nVerts, nLinVerts, nHoVerts, geoOrder
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call the appropriate method depending on faceType
    !------------------------------------------------------------------
    cellType = MeshCell(iCell)%cellType
    CALL CELL_TYPE_TO_FACE_TYPE(cellType,faceType)
    CALL FACE_TYPE_TO_VERTS(faceType,nVerts,nLinVerts,nHoVerts,geoOrder)
    !
    SELECT CASE (cellType)
    CASE('TET1','TET2','TET3','TET4')
       vFace(1:nVerts) = MeshCell(iCell)%vertex(nvTet(iFacLoc,1:nVerts))
    CASE('HEX1','HEX2')
       vFace(1:nVerts) = MeshCell(iCell)%vertex(nvHex(iFacLoc,1:nVerts))
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_FACE_VERTICES_FROM_CELL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_FACE_COORDINATES_FROM_CELL(iCell, iFacLoc, pFace)
    !------------------------------------------------------------------
    !---- Get face vertices coordinates from the cell index and the
    !---- local face index
    !
    !---- IN
    !---- iCell    : cell index
    !---- iFacLoc  : local face index
    !
    !---- OUT
    !---- pFace    : face vertices coordinates
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(INOUT) :: iCell, iFacLoc

    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(INOUT) :: pFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)                 :: cellType, faceType

    INTEGER(C_INT)                                :: nVerts, nLinVerts, &
         nHoVerts, geoOrder, iVert, vertLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call the appropriate method depending on faceType
    !------------------------------------------------------------------
    cellType          = MeshCell(iCell)%cellType
    CALL CELL_TYPE_TO_FACE_TYPE(cellType,faceType)
    CALL FACE_TYPE_TO_VERTS(faceType,nVerts,nLinVerts,nHoVerts,geoOrder)
    !
    DO iVert=1,nVerts
       vertLoc        = MeshCell(iCell)%vertex(MeshCell(iCell)%permType(iFacLoc,iVert))
       pFace(:,iVert) = MeshVert(vertLoc)%coor(:)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_FACE_COORDINATES_FROM_CELL

  !====================================================================
  !====================================================================
  !====================================================================  
  
  SUBROUTINE REMOVE_VERTICES_DUPLICATES(isModified)
    !------------------------------------------------------------------
    !---- Remove duplicate vertices
    !
    !---- IN
    !
    !---- OUT
    !---- isModified : TRUE if mesh was modified
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE ARRAY_OPERATIONS
    USE MESH_STATS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),                INTENT(OUT) :: isModified
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                              :: iVert, iFac, iCell, &
         nSorted, vert, nVerts, nRemoved
    INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: sorted
    
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: copyVert
    !------------------------------------------------------------------
    !---- Find duplicate points
    !------------------------------------------------------------------
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    ALLOCATE(copyVert(3,Mesh%nVert))
    DO iVert=1,Mesh%nVert
       copyVert(:,iVert) = MeshVert(iVert)%coor(:)
    ENDDO
    !
    ALLOCATE(sorted(Mesh%nVert))
    CALL FIND_ARRAY_DUPLICATES(copyVert, Mesh%nVert, 3, &
         Mesh%tolerance1D, sorted, nSorted)
    nRemoved = Mesh%nVert - nSorted
    WRITE(6,'(a,i0,a)') 'Found ',nRemoved,' duplicate point(s)'
    !------------------------------------------------------------------------
    !---- Write modified mesh if needed
    !------------------------------------------------------------------------
    IF (nRemoved == 0) THEN
       isModified = .FALSE.
       RETURN
    ELSE
       DEALLOCATE(MeshVert)
       Mesh%nVert = nSorted
       ALLOCATE(MeshVert(Mesh%nVert))
       !
       DO iVert=1,Mesh%nVert
          MeshVert(iVert)%coor(:) = copyVert(:,iVert)
       ENDDO
       DEALLOCATE(copyVert)
       !
       DO iFac=1,Mesh%nFac
          nVerts = MeshFac(iFac)%nVerts
          DO iVert=1,nVerts
             vert = MeshFac(iFac)%vertex(iVert)
             MeshFac(iFac)%vertex(iVert) = sorted(vert)
          ENDDO
       ENDDO
       !
       DO iCell=1,Mesh%nCell
          nVerts = MeshCell(iCell)%nVerts
          DO iVert=1,nVerts
             vert = MeshCell(iCell)%vertex(iVert)
             MeshCell(iCell)%vertex(iVert) = sorted(vert)
          ENDDO
       ENDDO
       !
       isModified = .TRUE.
       RETURN
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REMOVE_VERTICES_DUPLICATES  

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_ELEMENTS
