MODULE MESH_ELEMENTS_CENTERS

  !====================================================================
  ! Routines related to finding element centers
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE DOF_PERMUTATIONS
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Interface for overloading COMPUTE_CENTER_OF_FACE
  !--------------------------------------------------------------------
  INTERFACE COMPUTE_CENTER_OF_FACE
     MODULE PROCEDURE COMPUTE_CENTER_OF_FACE_FROM_INDEX, &
          COMPUTE_CENTER_OF_FACE_FROM_VERTICES,          &
          COMPUTE_CENTER_OF_FACE_FROM_INDEX_AND_LISTS
  END INTERFACE COMPUTE_CENTER_OF_FACE

  !--------------------------------------------------------------------
  ! Interface for overloading COMPUTE_CENTER_OF_CELL
  !--------------------------------------------------------------------
  INTERFACE COMPUTE_CENTER_OF_CELL
     MODULE PROCEDURE COMPUTE_CENTER_OF_CELL_FROM_INDEX, &
          COMPUTE_CENTER_OF_CELL_FROM_VERTICES
  END INTERFACE COMPUTE_CENTER_OF_CELL
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_CENTER_OF_FACE_FROM_INDEX(iFac, pCenter)
    !------------------------------------------------------------------
    !---- Compute the coordinates of the center of a face from its index
    !
    !---- IN
    !---- iFac    : index of the face
    !
    !---- OUT
    !---- pCenter : coordinates of the center of the face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: iFac
   
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: pCenter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: nVerts, iVert, vertLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute center of face
    !------------------------------------------------------------------
    pCenter = 0.0
    !
    nVerts = MeshFac(iFac)%nLinVerts
    DO iVert=1,nVerts
       vertLoc = MeshFac(iFac)%vertex(iVert)
       pCenter = pCenter + MeshVert(vertLoc)%coor
    ENDDO
    pCenter = pCenter/DBLE(nVerts)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_CENTER_OF_FACE_FROM_INDEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_CENTER_OF_FACE_FROM_INDEX_AND_LISTS(facList, &
       vertList, iFac, pCenter)
    !------------------------------------------------------------------
    !---- Compute the coordinates of the center of a face from its
    !---- index and a face list
    !
    !---- IN
    !---- iFac     : index of the face
    !---- facList  : list of faces
    !---- vertList : list of vertices
    !
    !---- OUT
    !---- pCenter  : coordinates of the center of the face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: iFac
   
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: pCenter

    TYPE(face),   DIMENSION(:), ALLOCATABLE   :: facList

    TYPE(vertex), DIMENSION(:), ALLOCATABLE   :: vertList
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: nVerts, iVert, vertLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute center of face
    !------------------------------------------------------------------
    pCenter = 0.0
    !
    nVerts = facList(iFac)%nLinVerts
    DO iVert=1,nVerts
       vertLoc = facList(iFac)%vertex(iVert)
       pCenter = pCenter + vertList(vertLoc)%coor
    ENDDO
    pCenter = pCenter/DBLE(nVerts)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_CENTER_OF_FACE_FROM_INDEX_AND_LISTS
  
  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_CENTER_OF_FACE_FROM_VERTICES(vertices, pCenter)
    !------------------------------------------------------------------
    !---- Compute the coordinates of the center of a face from its vertices
    !
    !---- IN
    !---- vertices : indices of the vertices
    !
    !---- OUT
    !---- pCenter  : coordinates of the center of the face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(:), INTENT(IN)  :: vertices
   
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: pCenter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: nVerts, iVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute center of face
    !------------------------------------------------------------------
    pCenter = 0.0
    !
    nVerts  = SIZE(vertices)
    DO iVert=1,nVerts
       pCenter = pCenter + MeshVert(vertices(iVert))%coor
    ENDDO
    pCenter = pCenter/DBLE(nVerts)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_CENTER_OF_FACE_FROM_VERTICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_CENTER_OF_CELL_FROM_INDEX(iCell, pCenter)
    !------------------------------------------------------------------
    !---- Compute the coordinates of the center of a cell from its index
    !
    !---- IN
    !---- iCell   : index of the cell
    !
    !---- OUT
    !---- pCenter : coordinates of the center of the cell
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: iCell

    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: pCenter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: nVerts, iVert, vertLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute center of face
    !------------------------------------------------------------------
    pCenter = 0.0
    !
    nVerts = MeshCell(iCell)%nLinVerts
    DO iVert=1,nVerts
       vertLoc = MeshCell(iCell)%vertex(iVert)
       pCenter = pCenter + MeshVert(vertLoc)%coor
    ENDDO
    pCenter = pCenter/REAL(nVerts)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_CENTER_OF_CELL_FROM_INDEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_CENTER_OF_CELL_FROM_VERTICES(vertices, pCenter)
    !------------------------------------------------------------------
    !---- Compute the coordinates of the center of a cell from its vertices
    !
    !---- IN
    !---- vertices : indices of the vertices
    !
    !---- OUT
    !---- pCenter  : coordinates of the center of the cell
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(:), INTENT(IN)  :: vertices

    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: pCenter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: nVerts, iVert, vertLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute center of face
    !------------------------------------------------------------------
    pCenter = 0.0
    !
    nVerts  = SIZE(vertices)
    DO iVert=1,nVerts
       pCenter = pCenter + MeshVert(vertices(iVert))%coor
    ENDDO
    pCenter = pCenter/REAL(nVerts)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_CENTER_OF_CELL_FROM_VERTICES

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_ELEMENTS_CENTERS
