MODULE MESH_ELEMENTS_DOFS

  !====================================================================
  ! Routines related to the dofs computations for elements
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE DOF_PERMUTATIONS
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Interfaces for overloading DOF position computation in cells
  !----------------------------------------------------------------------
  INTERFACE GET_CELL_DOF
     MODULE PROCEDURE GET_CELL_DOF_FROM_INDEX, GET_TET_DOF, GET_HEX_DOF
  END INTERFACE GET_CELL_DOF

  !----------------------------------------------------------------------
  ! Interfaces for overloading DOF position computation in faces
  !----------------------------------------------------------------------
  INTERFACE GET_FACE_DOF
     MODULE PROCEDURE GET_FACE_DOF_FROM_INDEX, GET_TRI_DOF, GET_QUAD_DOF
  END INTERFACE GET_FACE_DOF    
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_CELL_DOF_FROM_INDEX(iCell, order, nDofV, dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Get the coordinates of the d.o.f. of a linear cell 
    !---- from its index
    !
    !---- IN
    !---- iCell : global cell index
    !---- order : polynomial order
    !---- nDofV : number of volumic d.o.f.
    !
    !---- OUT
    !---- dof*  : computed coordinates of the d.o.f.
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nDofV, iCell, order

    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)             :: cellType

    INTEGER(C_INT)                            :: iDof, nLinVerts
    INTEGER(C_INT), DIMENSION(8)              :: vertNum

    REAL(C_DOUBLE), DIMENSION(3)              :: v1, v2, v3, v4, v5, v6, v7, v8
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get linear coordinates of degrees of freedom
    !------------------------------------------------------------------
    nLinVerts            = MeshCell(iCell)%nLinVerts
    vertNum(1:nLinVerts) = MeshCell(iCell)%vertex(1:nLinVerts)
    cellType             = MeshCell(iCell)%cellType
    !
    SELECT CASE(cellType)
    CASE('TET1','TET2','TET3','TET4')
       v1(:)      = MeshVert(vertNum(1))%coor(:)
       v2(:)      = MeshVert(vertNum(2))%coor(:)
       v3(:)      = MeshVert(vertNum(3))%coor(:)
       v4(:)      = MeshVert(vertNum(4))%coor(:)
       !
       CALL GET_TET_DOF(v1, v2, v3, v4, nDofV, order, dofX, dofY, dofZ)
    CASE('HEX1','HEX2')
       v1(:)      = MeshVert(vertNum(1))%coor(:)
       v2(:)      = MeshVert(vertNum(2))%coor(:)
       v3(:)      = MeshVert(vertNum(3))%coor(:)
       v4(:)      = MeshVert(vertNum(4))%coor(:)
       v5(:)      = MeshVert(vertNum(5))%coor(:)
       v6(:)      = MeshVert(vertNum(6))%coor(:)
       v7(:)      = MeshVert(vertNum(7))%coor(:)
       v8(:)      = MeshVert(vertNum(8))%coor(:)
       !
       CALL GET_HEX_DOF(v1, v2, v3, v4, v5, v6, v7, v8, &
            nDofV, order, dofX, dofY, dofZ)
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_CELL_DOF_FROM_INDEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_TET_DOF(v1, v2, v3, v4, nDofV, order, dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Get the coordinates of the d.o.f. of a TET cell from its vertices
    !
    !---- IN
    !---- v*    : coordinates of the vertices
    !---- order : polynomial order
    !---- nDofV : number of volumic d.o.f.
    !
    !---- OUT
    !---- dof*  : computed coordinates of the d.o.f.
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_CELLS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nDofV, order

    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: v1, v2, v3, v4
    
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: iDof

    REAL(C_DOUBLE), DIMENSION(3,3)            :: jac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get linear coordinates of degrees of freedom
    !------------------------------------------------------------------
    jac(1,:) = v2(:) - v1(:)
    jac(2,:) = v3(:) - v1(:)
    jac(3,:) = v4(:) - v1(:)
    !
    DO iDof=1,nDofV
       dofX(iDof) = v1(1)                       &
            + jac(1,1)*xDofTet(order)%tab(iDof) &
            + jac(2,1)*yDofTet(order)%tab(iDof) &
            + jac(3,1)*zDofTet(order)%tab(iDof)
       dofY(iDof) = v1(2)                       &
            + jac(1,2)*xDofTet(order)%tab(iDof) &
            + jac(2,2)*yDofTet(order)%tab(iDof) &
            + jac(3,2)*zDofTet(order)%tab(iDof)
       dofZ(iDof) = v1(3)                       &
            + jac(1,3)*xDofTet(order)%tab(iDof) &
            + jac(2,3)*yDofTet(order)%tab(iDof) &
            + jac(3,3)*zDofTet(order)%tab(iDof)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_TET_DOF

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_HEX_DOF(v1, v2, v3, v4, v5, v6, v7, v8, &
       nDofV, order, dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Get the coordinates of the d.o.f. of a HEX cell from its vertices
    !
    !---- IN
    !---- v*    : coordinates of the vertices
    !---- order : polynomial order
    !---- nDofV : number of volumic d.o.f.
    !
    !---- OUT
    !---- dof*  : computed coordinates of the d.o.f.
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_CELLS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nDofV, order

    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: v1, v2, v3, v4, v5, v6, v7, v8
    
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: iDof

    REAL(C_DOUBLE), DIMENSION(3,3)            :: jac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get linear coordinates of degrees of freedom
    !------------------------------------------------------------------
    jac(1,:) = v2(:) - v1(:)
    jac(2,:) = v4(:) - v1(:)
    jac(3,:) = v5(:) - v1(:)
    !
    DO iDof=1,nDofV
       dofX(iDof) = v1(1)                                   &
            + 0.5d0*jac(1,1)*(1.0+xDofHex(order)%tab(iDof)) &
            + 0.5d0*jac(2,1)*(1.0+yDofHex(order)%tab(iDof)) &
            + 0.5d0*jac(3,1)*(1.0+zDofHex(order)%tab(iDof))
       dofY(iDof) = v1(2)                                   &
            + 0.5d0*jac(1,2)*(1.0+xDofHex(order)%tab(iDof)) &
            + 0.5d0*jac(2,2)*(1.0+yDofHex(order)%tab(iDof)) &
            + 0.5d0*jac(3,2)*(1.0+zDofHex(order)%tab(iDof))
       dofZ(iDof) = v1(3)                                   &
            + 0.5d0*jac(1,3)*(1.0+xDofHex(order)%tab(iDof)) &
            + 0.5d0*jac(2,3)*(1.0+yDofHex(order)%tab(iDof)) &
            + 0.5d0*jac(3,3)*(1.0+zDofHex(order)%tab(iDof))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_HEX_DOF

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_FACE_DOF_FROM_INDEX(iFace, order, nDofS, nDofV, dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Get the coordinates of the d.o.f. of a linear face
    !---- from its index
    !
    !---- IN
    !---- iFace : global face index
    !---- order : polynomial order
    !---- nDofS : number of surfacic d.o.f.
    !---- nDofV : number of volumic  d.o.f.
    !
    !---- OUT
    !---- dof*  : computed coordinates of the d.o.f.
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_CELLS
    USE MESH_FACES
    USE REFERENCE_CELLS
    USE NUMERICAL_TYPES
    USE MESH_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nDofS, nDofV, order, iFace

    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)             :: cellType
    
    INTEGER(C_INT)                            :: iCell, iDof, iPerm
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. ALLOCATED(dof))   ALLOCATE(dof  (nDofSMax))
    IF (.NOT. ALLOCATED(dofXV)) ALLOCATE(dofXV(nDofVMax))
    IF (.NOT. ALLOCATED(dofYV)) ALLOCATE(dofYV(nDofVMax))
    IF (.NOT. ALLOCATED(dofZV)) ALLOCATE(dofZV(nDofVMax))
    !------------------------------------------------------------------
    !---- Get linear coordinates of degrees of freedom
    !------------------------------------------------------------------
    iPerm    = MeshFac(iFace)%perm(1)
    iCell    = MeshFac(iFace)%neighb(1)
    cellType = MeshCell(iCell)%cellType
    CALL ORDER_TO_NDOFV(order,cellType,nDofV)
    CALL GET_CELL_DOF(iCell,order,nDofV,dofXV,dofYV,dofZV)
    !
    SELECT CASE(cellType)
    CASE('TET1')
       dof(1:nDofS) = iPermDofTet(order)%tab(1:nDofS,iPerm)
    CASE('HEX1')
       dof(1:nDofS) = iPermDofHex(order)%tab(1:nDofS,iPerm)
    END SELECT
    !
    dofX(:) = dofXV(dof(1:nDofS))
    dofY(:) = dofYV(dof(1:nDofS))
    dofZ(:) = dofZV(dof(1:nDofS))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_FACE_DOF_FROM_INDEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_TRI_DOF(v1, v2, v3, nDofS, order, dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Get the coordinates of the d.o.f. of a TRI face from its vertices
    !
    !---- IN
    !---- v*    : coordinates of the vertices
    !---- order : polynomial order
    !---- nDofV : number of surfacic d.o.f.
    !
    !---- OUT
    !---- dof*  : computed coordinates of the d.o.f.
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_CELLS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nDofS, order

    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: v1, v2, v3
    
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                 :: iDof

    REAL(C_DOUBLE), DIMENSION(2,3) :: jac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get linear coordinates of degrees of freedom
    !------------------------------------------------------------------
    jac(1,:) = v2(:) - v1(:)
    jac(2,:) = v3(:) - v1(:)
    !
    DO iDof=1,nDofS
       dofX(iDof) = v1(1)                       &
            + jac(1,1)*xDofTri(order)%tab(iDof) &
            + jac(2,1)*yDofTri(order)%tab(iDof) 
       dofY(iDof) = v1(2)                       &
            + jac(1,2)*xDofTri(order)%tab(iDof) &
            + jac(2,2)*yDofTri(order)%tab(iDof) 
       dofZ(iDof) = v1(3)                       &
            + jac(1,3)*xDofTri(order)%tab(iDof) &
            + jac(2,3)*yDofTri(order)%tab(iDof) 
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_TRI_DOF

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_QUAD_DOF(v1, v2, v3, v4, nDofS, order, dofX, dofY, dofZ)
    !------------------------------------------------------------------
    !---- Get the coordinates of the d.o.f. of a TRI face from its vertices
    !
    !---- IN
    !---- v*    : coordinates of the vertices
    !---- order : polynomial order
    !---- nDofV : number of surfacic d.o.f.
    !
    !---- OUT
    !---- dof*  : computed coordinates of the d.o.f.
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_VERTICES
    USE MESH_CELLS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: nDofS, order

    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: v1, v2, v3, v4
    
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: dofX, dofY, dofZ
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: iDof

    REAL(C_DOUBLE), DIMENSION(2,3)            :: jac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Get linear coordinates of degrees of freedom
    !------------------------------------------------------------------
    jac(1,:) = v2(:) - v1(:)
    jac(2,:) = v4(:) - v1(:)
    !
    DO iDof=1,nDofS
       dofX(iDof) = v1(1)                                 &
            + 0.5*jac(1,1)*(1.0+xDofQua(order)%tab(iDof)) &
            + 0.5*jac(2,1)*(1.0+yDofQua(order)%tab(iDof))
       dofY(iDof) = v1(2)                                 &
            + 0.5*jac(1,2)*(1.0+xDofQua(order)%tab(iDof)) &
            + 0.5*jac(2,2)*(1.0+yDofQua(order)%tab(iDof)) 
       dofZ(iDof) = v1(3)                                 &
            + 0.5*jac(1,3)*(1.0+xDofQua(order)%tab(iDof)) &
            + 0.5*jac(2,3)*(1.0+yDofQua(order)%tab(iDof))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_QUAD_DOF

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_ELEMENTS_DOFS
