MODULE MESH_FACES

  !====================================================================
  ! Definition and data structure associated to a generic face of the mesh
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a generic face of the mesh
  !
  ! faceType        ==> type of face:
  !                     'TRI1' = linear triangle
  !                     'QUA1' = linear quadrangle
  !                     'TRI2' = curved triangle
  !                     'QUA2' = curved quadrangle  
  ! isHO            ==> TRUE if this is a  high-order  face
  ! isTfsf          ==> TRUE if this is a  TFSF        face
  ! isInterface     ==> TRUE if this is an interface   face
  ! isInner         ==> TRUE if this is an inner       face
  ! isBoundary      ==> TRUE if this is a  boundary    face
  ! isPeriodic      ==> TRUE if this is a  periodic    face
  ! isArtificial    ==> TRUE if this is an artificial  face
  ! isHybrid        ==> TRUE if this is a  hybrid      face
  ! boundaryFaceNum ==> index of the associated boundary face if relevant
  ! nVerts          ==> number of            vertices of the face
  ! nLinVerts       ==> number of linear     vertices of the face
  ! nHoVerts        ==> number of high-order vertices of the face
  ! geoOrder        ==> geometrical order             of the face
  ! numFace         ==> index                         of the face
  ! log             ==> logical                       of the face
  ! measure         ==> measure of the face (i.e. the aera)
  ! vertex(k)       ==> index of the kth vertex of the face (k=1:nVerts)
  ! numLocal(i)     ==> local number of the face in the neighbor cells (i=1:2)
  ! numGeom(i)      ==> local geometric number of the face in the neighbor cells (i=1:2)
  ! normal(i)       ==> ith component of the normal of the face  (i=1:3)
  ! neighb(i)       ==> index of the neighbors attached to each side
  !                     of the face (i=1:2)
  ! perm(i)         ==> mapping of the vertices between the local numbering
  !                     of the face and the local numbering of the
  !                     attached elements (i=1:2)
  ! axis(i,k)       ==> ith coordinate of the kth axis of the inner face 
  !                     (i=1:3, k=1:2)  
  !--------------------------------------------------------------------
  TYPE face
     CHARACTER(LEN=4, KIND=C_CHAR)               :: faceType

     LOGICAL(C_BOOL)                             :: isHO, isTfsf,   &
          isInterface, isInner, isBoundary, isArtificial, isHybrid, &
          isPeriodic

     INTEGER(C_INT)                              :: numFace, log,   &
          geoOrder, nVerts, nLinVerts, nHoVerts, boundaryFaceNum
     INTEGER(C_INT), DIMENSION(2)                :: perm, numLocal, &
          numGeom, neighb
     INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: vertex

     REAL(C_DOUBLE)                              :: measure 
     REAL(C_DOUBLE), DIMENSION(3)                :: normal
     REAL(C_DOUBLE), DIMENSION(3,2)              :: axis  
  END TYPE face

  !--------------------------------------------------------------------
  ! Structure of pointer to a face
  ! This structure is used in MESH_CELL
  !
  ! ptr            ==> pointer to the associated face
  !--------------------------------------------------------------------
  TYPE face_ptr
     TYPE(face), POINTER      :: ptr
  END TYPE face_ptr

  !--------------------------------------------------------------------
  ! Data structure associated to a derived face type
  !
  ! faceNum        ==> index of the face in the MeshFac structure
  !--------------------------------------------------------------------
  TYPE derived_face
     INTEGER(C_INT)           :: faceNum
  END TYPE derived_face

  !--------------------------------------------------------------------
  ! Complementary data structure associated to a TF/SF face
  !
  ! cellNum        ==> index of the associated cell in the MeshCell structure
  ! zone           ==> 'TF' (total field) or 'SF' (scattered field)
  ! sign           ==> =+1 if TF, =-1 if SF (used in the flux computation)
  ! localFaceNum   ==> local number of the face in the associated cell
  !--------------------------------------------------------------------
  TYPE complement_tfsf_face
     INTEGER(C_INT)                :: cellNum, localFaceNum
     
     REAL(C_DOUBLE)                :: sign

     CHARACTER(LEN=2, KIND=C_CHAR) :: zone
  END TYPE complement_tfsf_face

  !--------------------------------------------------------------------
  ! Complementary data structure associated to a periodic face
  !
  ! matchingFace   ==> periodic face matching current face
  ! direction      ==> (+1,0,0) if the face is periodic alonge x direction,
  !                    and its matching face coordinates are found by
  !                    adding       the size of the domain to its coordinates
  !                    (-1,0,0) if the face is periodic alonge x direction,
  !                    and its matching face coordinates are found by
  !                    substracting the size of the domain to its coordinates
  !                    (0,+1,0) if ...
  !                    (0,-1,0) if ...
  !--------------------------------------------------------------------
  TYPE complement_periodic_face
     INTEGER(C_INT)                :: matchingFace

     REAL(C_DOUBLE), DIMENSION(3)  :: direction
  END TYPE complement_periodic_face

  !--------------------------------------------------------------------
  ! Complementary data structure associated to a material interface face
  !
  ! cellIn         ==> index of the cell that is inside  the main material
  ! cellOut        ==> index of the cell that is outside the main material
  ! iNeighb        ==> local index of the face in cellIn connecting to cellOut
  !--------------------------------------------------------------------
  TYPE complement_interface_face
     INTEGER(C_INT)                :: cellIn, cellOut, iNeighb
  END TYPE complement_interface_face  

  !--------------------------------------------------------------------
  ! Arrays of faces
  !--------------------------------------------------------------------
  TYPE(face),             TARGET, DIMENSION(:), ALLOCATABLE :: MeshFac

  TYPE(derived_face),             DIMENSION(:), ALLOCATABLE :: MeshFacI, &
       MeshFacB, MeshFacPhy, MeshFacPec, MeshFacPmc, MeshFacAbs,         &
       MeshFacPer, MeshFacPoy, MeshFacLin, MeshFacQuadr, MeshFacTri,     &
       MeshFacTriLin, MeshFacTriQuadr, MeshFacQuad, MeshFacQuadLin,      &
       MeshFacQuadQuadr, MeshFacInter, MeshFacArt, MeshFacHyb, MeshFacTfsf

  TYPE(complement_tfsf_face),     DIMENSION(:), ALLOCATABLE :: CompFacTfsf
  
  TYPE(complement_periodic_face), DIMENSION(:), ALLOCATABLE :: CompFacPer

  !--------------------------------------------------------------------
  ! Miscellaneous stuff
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL) :: addInnerFaces     = .FALSE.
  LOGICAL(C_BOOL) :: addInterfaceFaces = .FALSE.
  LOGICAL(C_BOOL) :: checkForFaces     = .TRUE.

  INTEGER(C_INT)  :: faceBlockSize     = 2000
  INTEGER(C_INT)  :: dFaceBlockSize    = 500
  INTEGER(C_INT)  :: cFaceBlockSize    = 100
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_FACE(buffer, eltLog, eltType, faceNum)
    !------------------------------------------------------------------
    !---- Fill all face arrays for a given face
    !
    !---- IN
    !---- buffer  : vertices of the face
    !---- eltLog  : logical tag of the face
    !---- eltType : geometric type of face
    !
    !---- OUT
    !---- faceNum : index of the created face    
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES_LOGS
    USE MESH_TYPE_CONVERSIONS    
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(INOUT) :: eltType

    INTEGER(C_INT),                INTENT(IN)    :: eltLog
    INTEGER(C_INT),                INTENT(OUT)   :: faceNum
    INTEGER(C_INT), DIMENSION(:),  INTENT(IN)    :: buffer
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)                :: hoType
    
    LOGICAL(C_BOOL)                              :: isConformalInnerTag
    
    INTEGER(C_INT)                               :: nVerts, nLinVerts, &
         nHoVerts, geoOrder
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Exclude inner face if it is not any other kind of face
    !------------------------------------------------------------------
    CALL IS_CONFORMAL_INNER_LOG(eltLog, isConformalInnerTag)
    IF ((.NOT. addInnerFaces) .AND. (isConformalInnerTag)) RETURN
    !------------------------------------------------------------------
    !---- If the face does not already exist, proceed to basic operations
    !------------------------------------------------------------------
    !---- Basic face informations
    CALL EXTEND_FACE_ARRAY(MeshFac, Mesh%nFac)
    CALL FACE_TYPE_TO_VERTS(eltType, nVerts, nLinVerts, nHoVerts, geoOrder)
    !
    !---- Handle HO mesh
    IF (.NOT. Mesh%useHO) THEN
       nVerts   = nLinVerts
       geoOrder = 1
       nHoVerts = 0
       hoType   = eltType
       CALL CHANGE_TYPE_WITH_ORDER(1, hoType, eltType)
    ENDIF
    !
    !---- Fill informations
    faceNum                             = Mesh%nFac
    MeshFac(Mesh%nFac)%faceType         = eltType
    MeshFac(Mesh%nFac)%nVerts           = nVerts
    MeshFac(Mesh%nFac)%nLinVerts        = nLinVerts
    MeshFac(Mesh%nFac)%nHoVerts         = nHoVerts
    MeshFac(Mesh%nFac)%geoOrder         = geoOrder
    ALLOCATE(MeshFac(Mesh%nFac)%vertex(nVerts))
    MeshFac(Mesh%nFac)%vertex(1:nVerts) = buffer(1:nVerts)
    MeshFac(Mesh%nFac)%log              = eltLog
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_FACE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_DERIVED_FACE(eltLog, eltType, faceNum)
    !------------------------------------------------------------------
    !---- Fill all derived face arrays for a given face
    !
    !---- IN
    !---- eltLog  : logical tag of the face
    !---- eltType : geometric type of face
    !---- faceNum : index of the corresponding face structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES_LOGS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: eltType

    INTEGER(C_INT),                INTENT(IN) :: eltLog, faceNum
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                           :: isConformalInnerTag
    
    INTEGER(C_INT)                            :: tag1, tag10, &
         tag100, tag1000, tag10000
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Exclude inner face if it is not any other kind of face
    !------------------------------------------------------------------
    CALL IS_CONFORMAL_INNER_LOG(eltLog, isConformalInnerTag)
    IF ((.NOT. addInnerFaces) .AND. (isConformalInnerTag)) RETURN
    !------------------------------------------------------------------
    !---- Handle geometric informations
    !------------------------------------------------------------------
    SELECT CASE(eltType)
    CASE('TRI','TRI1')
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTri,       Mesh%nFacTri)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacLin,       Mesh%nFacLin)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTriLin,    Mesh%nFacTriLin)
       MeshFacTri(Mesh%nFacTri)%faceNum              =  faceNum       
       MeshFacLin(Mesh%nFacLin)%faceNum              =  faceNum
       MeshFac(faceNum)%isHO                         =  .FALSE.
       MeshFacTriLin(Mesh%nFacTriLin)%faceNum        =  faceNum          
    CASE('TRI2')
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTri,       Mesh%nFacTri)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacQuadr,     Mesh%nFacQuadr)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTriQuadr,  Mesh%nFacTriQuadr)
       MeshFacTri(Mesh%nFacTri)%faceNum              =  faceNum
       MeshFacQuadr(Mesh%nFacQuadr)%faceNum          =  faceNum
       MeshFac(faceNum)%isHO                         =  .TRUE.
       MeshFacTriQuadr(Mesh%nFacTriQuadr)%faceNum    =  faceNum
    CASE('QUA','QUA1')
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacQuad,      Mesh%nFacQuad)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacLin,       Mesh%nFacLin)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacQuadLin,   Mesh%nFacQuadLin)
       MeshFacQuad(Mesh%nFacQuad)%faceNum            =  faceNum 
       MeshFacLin(Mesh%nFacLin)%faceNum              =  faceNum
       MeshFac(faceNum)%isHO                         =  .FALSE.         
       MeshFacQuadLin(Mesh%nFacQuadLin)%faceNum      =  faceNum
    CASE('QUA2')
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacQuad,      Mesh%nFacQuad)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacQuadr,     Mesh%nFacQuadr)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacQuadQuadr, Mesh%nFacQuadQuadr)
       MeshFacQuad(Mesh%nFacQuad)%faceNum            =  faceNum          
       MeshFacQuadr(Mesh%nFacQuadr)%faceNum          =  faceNum
       MeshFac(faceNum)%isHO                         =  .TRUE.
       MeshFacQuadQuadr(Mesh%nFacQuadQuadr)%faceNum  =  faceNum
    END SELECT
    !------------------------------------------------------------------
    !---- Decompose tag to increment the right counters
    !------------------------------------------------------------------    
    CALL DECOMPOSE_INTEGER(eltLog, 1, tag1    )
    CALL DECOMPOSE_INTEGER(eltLog, 2, tag10   )
    CALL DECOMPOSE_INTEGER(eltLog, 3, tag100  )
    CALL DECOMPOSE_INTEGER(eltLog, 4, tag1000 )
    CALL DECOMPOSE_INTEGER(eltLog, 5, tag10000)
    !------------------------------------------------------------------
    !---- Sort first integer (inner or boundary)
    !------------------------------------------------------------------
    SELECT CASE(tag1)
    CASE(1)      ! Inner face
       IF (addInnerFaces) THEN
          CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacI, Mesh%nFacI)
          MeshFacI(Mesh%nFacI)%faceNum           = faceNum
          MeshFac(faceNum)%isInner               = .TRUE.
       ENDIF
    CASE(2)      ! PEC face
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacB,   Mesh%nFacB)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPhy, Mesh%nFacPhy)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPec, Mesh%nFacPec)
       MeshFacB(Mesh%nFacB)%faceNum              = faceNum
       MeshFacPhy(Mesh%nFacPhy)%faceNum          = faceNum
       MeshFacPec(Mesh%nFacPec)%faceNum          = faceNum
       MeshFac(faceNum)%isBoundary               = .TRUE.
       MeshFac(faceNum)%boundaryFaceNum          = Mesh%nFacB
    CASE(4)      ! Absorbing face
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacB,   Mesh%nFacB)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPhy, Mesh%nFacPhy)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacAbs, Mesh%nFacAbs)
       MeshFacB(Mesh%nFacB)%faceNum              = faceNum
       MeshFacPhy(Mesh%nFacPhy)%faceNum          = faceNum
       MeshFacAbs(Mesh%nFacAbs)%faceNum          = faceNum
       MeshFac(faceNum)%isBoundary               = .TRUE.
       MeshFac(faceNum)%boundaryFaceNum          = Mesh%nFacB
    CASE(6)      ! Periodic face
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacB,   Mesh%nFacB)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPhy, Mesh%nFacPhy)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPer, Mesh%nFacPer)
       MeshFacB(Mesh%nFacB)%faceNum              = faceNum
       MeshFacPhy(Mesh%nFacPhy)%faceNum          = faceNum
       MeshFacPer(Mesh%nFacPer)%faceNum          = faceNum
       MeshFac(faceNum)%isBoundary               = .TRUE.
       MeshFac(faceNum)%isPeriodic               = .TRUE.
       MeshFac(faceNum)%boundaryFaceNum          = Mesh%nFacB
    CASE(8)      ! PMC face
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacB,   Mesh%nFacB)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPhy, Mesh%nFacPhy)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacPmc, Mesh%nFacPmc)
       MeshFacB(Mesh%nFacB)%faceNum              = faceNum
       MeshFacPhy(Mesh%nFacPhy)%faceNum          = faceNum
       MeshFacPmc(Mesh%nFacPmc)%faceNum          = faceNum       
       MeshFac(faceNum)%isBoundary               = .TRUE.
       MeshFac(faceNum)%boundaryFaceNum          = Mesh%nFacB
    CASE DEFAULT
       ! Do nothing
    END SELECT
    !------------------------------------------------------------------
    !---- Sort second integer (artificial or not)
    !------------------------------------------------------------------    
    SELECT CASE(tag10)
    CASE(0)      ! Non-artificial face
       ! Do nothing
    CASE(1)      ! Artificial face
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacB,   Mesh%nFacB)
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacArt, Mesh%nFacArt)
       MeshFacB(Mesh%nFacB)%faceNum              = faceNum
       MeshFacArt(Mesh%nFacArt)%faceNum          = faceNum       
       MeshFac(faceNum)%isBoundary               = .TRUE.
       MeshFac(faceNum)%isArtificial             = .TRUE.
       MeshFac(faceNum)%boundaryFaceNum          = Mesh%nFacB
    CASE DEFAULT
       ! Do nothing
    END SELECT
    !------------------------------------------------------------------
    !---- Sort third integer (tfsf or not)
    !------------------------------------------------------------------    
    SELECT CASE(tag100)
    CASE(0)      ! Non-tfsf face
       ! Do nothing
    CASE(1)      ! Tfsf face
       IF (MeshFac(faceNum)%isArtificial) THEN
          CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTfsf, Mesh%nFacTfsf)
          MeshFacTfsf(Mesh%nFacTfsf)%faceNum     = faceNum
          MeshFac(faceNum)%isTfsf                = .TRUE.
       ELSE
          CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTfsf, Mesh%nFacTfsf)
          MeshFacTfsf(Mesh%nFacTfsf)%faceNum     = faceNum
          MeshFac(faceNum)%isTfsf                = .TRUE.
          !
          CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacTfsf, Mesh%nFacTfsf)
          MeshFacTfsf(Mesh%nFacTfsf)%faceNum     = faceNum
       ENDIF
    CASE DEFAULT
       ! Do nothing
    END SELECT
    !------------------------------------------------------------------
    !---- Sort fourth integer (conformal or not)
    !------------------------------------------------------------------    
    SELECT CASE(tag1000)
    CASE(0)      ! Conformal face
       ! Do nothing
    CASE(1)      ! Hybrid face
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacHyb, Mesh%nFacHyb)
       MeshFacHyb(Mesh%nFacHyb)%faceNum          = faceNum
       MeshFac(faceNum)%isHybrid                 = .TRUE.       
    CASE DEFAULT
       ! Do nothing
    END SELECT
    !------------------------------------------------------------------
    !---- Sort fifth integer (interface or not)
    !------------------------------------------------------------------    
    SELECT CASE(tag10000)
    CASE(0)      ! Not an interface
       ! Do nothing
    CASE DEFAULT ! Interface
       CALL EXTEND_DERIVED_FACE_ARRAY(MeshFacInter, Mesh%nFacInter)
       MeshFacInter(Mesh%nFacInter)%faceNum      = faceNum
       MeshFac(faceNum)%isInterface              = .TRUE.
    END SELECT    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_DERIVED_FACE

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Build the derived face lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace, iCell, locFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build derived faces from scratch
    !------------------------------------------------------------------
    dFaceBlockSize = faceBlockSize/2
    cFaceBlockSize = faceBlockSize/2
    CALL DEALLOCATE_DERIVED_FACES
    !
    DO iFace=1,Mesh%nFac
       CALL ADD_DERIVED_FACE(MeshFac(iFace)%log, &
            MeshFac(iFace)%faceType, iFace)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_DERIVED_FACES  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_FACE_ARRAY(faceArray, faceArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to faceArray
    !
    !---- IN
    !---- faceArray     : array of 'face' structure
    !---- faceArraySize : number of elements in faceArray
    !
    !---- OUT
    !---- faceArray     : modified in the process
    !---- faceArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),         INTENT(INOUT) :: faceArraySize
    
    TYPE(face), DIMENSION(:), ALLOCATABLE :: faceArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                       :: isArrayFull, isAllocated

    INTEGER(C_INT)                        :: iFace
    
    TYPE(face), DIMENSION(:), ALLOCATABLE :: tempFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(faceArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(faceArray) < faceArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
          ALLOCATE(tempFace(faceArraySize))
          tempFace = faceArray
          DEALLOCATE(faceArray)
       ENDIF
       !
       ALLOCATE(faceArray(faceArraySize + faceBlockSize))
       DO iFace=faceArraySize+1,faceArraySize+faceBlockSize
          CALL INITIALIZE_FACE_STRUCTURE(faceArray(iFace), iFace)
       ENDDO
       !
       faceArray(1:faceArraySize) = tempFace(1:faceArraySize)
       IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
    ENDIF
    !
    faceArraySize = faceArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_FACE_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_DERIVED_FACE_ARRAY(dFaceArray, dFaceArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to derivedFaceArray
    !
    !---- IN
    !---- dFaceArray     : array of 'derived_face' structure
    !---- dFaceArraySize : number of elements in derivedFaceArray
    !
    !---- OUT
    !---- dFaceArray     : modified in the process
    !---- dFaceArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(INOUT) :: dFaceArraySize
    
    TYPE(derived_face), DIMENSION(:), ALLOCATABLE :: dFaceArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                               :: isArrayFull, isAllocated
    
    INTEGER(C_INT)                                :: iFace
    
    TYPE(derived_face), DIMENSION(:), ALLOCATABLE :: tempFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(dFaceArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(dFaceArray) < dFaceArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF    
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
          ALLOCATE(tempFace(dFaceArraySize))
          tempFace = dFaceArray
          DEALLOCATE(dFaceArray)
       ENDIF
       !
       ALLOCATE(dFaceArray(dFaceArraySize + dFaceBlockSize))
       DO iFace=dFaceArraySize+1,dFaceArraySize+dFaceBlockSize
          CALL INITIALIZE_DERIVED_FACE_STRUCTURE(dFaceArray(iFace))
       ENDDO
       !
       dFaceArray(1:dFaceArraySize) = tempFace(1:dFaceArraySize)
       IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
    ENDIF
    !
    dFaceArraySize = dFaceArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_DERIVED_FACE_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_FACE_STRUCTURE(faceStructure, iFace)
    !------------------------------------------------------------------
    !---- Initialize face structure
    !
    !---- IN
    !---- faceStructure : face structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)    :: iFace
    
    TYPE(face),     INTENT(INOUT) :: faceStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize face structure
    !------------------------------------------------------------------
    faceStructure%faceType        = 'NONE'
    faceStructure%isHO            = .FALSE.
    faceStructure%isTfsf          = .FALSE.
    faceStructure%isInterface     = .FALSE.
    faceStructure%isInner         = .FALSE.
    faceStructure%isBoundary      = .FALSE.
    faceStructure%isArtificial    = .FALSE.
    faceStructure%isHybrid        = .FALSE.
    faceStructure%isPeriodic      = .FALSE.
    faceStructure%nVerts          = 0
    faceStructure%nLinVerts       = 0
    faceStructure%nHoVerts        = 0
    faceStructure%geoOrder        = 0        
    faceStructure%numFace         = iFace
    faceStructure%boundaryFaceNum = 0
    faceStructure%log             = 0
    faceStructure%measure         = 0.0
    faceStructure%numLocal(:)     = 0
    faceStructure%numGeom(:)      = 0
    faceStructure%normal(:)       = 0.0
    faceStructure%neighb(:)       = 0
    faceStructure%perm(:)         = 0
    faceStructure%axis(:,:)       = 0.0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_FACE_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MERGE_FACES(face1, face2)
    !------------------------------------------------------------------
    !---- Merge faces into each other
    !---- WARNING : This method is *NOT* commutative
    !
    !---- IN
    !---- face* : faces to merge
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES_LOGS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(face), INTENT(INOUT) :: face1, face2
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    TYPE(face)                :: newFace
    !------------------------------------------------------------------
    !---- Merge faces
    !------------------------------------------------------------------
    CALL INITIALIZE_FACE_STRUCTURE(newFace, 0)
    CALL COMPUTE_LOG_SUM(face1%log, face2%log, newFace%log)
    newFace%nVerts      = MAX(face1%nVerts,      face2%nVerts     )
    newFace%nLinVerts   = MAX(face1%nLinVerts,   face2%nLinVerts  )
    newFace%nHoVerts    = MAX(face1%nHoVerts,    face2%nHoVerts   )
    newFace%geoOrder    = MAX(face1%geoOrder,    face2%geoOrder   )
    newFace%numFace     = MAX(face1%numFace,     face2%numFace    )
    newFace%measure     = MAX(face1%measure,     face2%measure    )
    newFace%numLocal(:) = MAX(face1%numLocal(:), face2%numLocal(:))
    newFace%numGeom(:)  = MAX(face1%numGeom(:),  face2%numGeom(:) )
    newFace%normal(:)   = MAX(face1%normal(:),   face2%normal(:)  )
    newFace%neighb(1)   = MAX(face1%neighb(1),   face2%neighb(1)  )
    newFace%neighb(2)   = MAX(face1%neighb(2),   face2%neighb(2)  )
    newFace%perm(:)     = MAX(face1%perm(:),     face2%perm(:)    )
    ALLOCATE(newFace%vertex(SIZE(face1%vertex)))
    newFace%vertex      = face1%vertex
    newFace%faceType    = face1%faceType
    IF (face1%faceType == 'NONE') newFace%faceType = face2%faceType
    IF (face2%faceType == 'NONE') newFace%faceType = face1%faceType    
    !
    face1 = newFace
    face2 = newFace
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MERGE_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_DERIVED_FACE_STRUCTURE(derivedFaceStructure)
    !------------------------------------------------------------------
    !---- Initialize derived face structure
    !
    !---- IN
    !---- derivedFaceStructure : derived face structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(derived_face), INTENT(INOUT) :: derivedFaceStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize derived face structure
    !------------------------------------------------------------------
    derivedFaceStructure%faceNum = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_DERIVED_FACE_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_FACES
    !------------------------------------------------------------------
    !---- Deallocate all face lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate faces
    !------------------------------------------------------------------
    IF (ALLOCATED(MeshFac)) DEALLOCATE(MeshFac)
    CALL DEALLOCATE_DERIVED_FACES
    CALL DEALLOCATE_COMPLEMENTARY_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Deallocate all derived face lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate faces
    !------------------------------------------------------------------
    CALL RESET_MESH_DERIVED_FACE_COUNTERS(Mesh)
    !
    IF (ALLOCATED(MeshFacI))         DEALLOCATE(MeshFacI)
    IF (ALLOCATED(MeshFacB))         DEALLOCATE(MeshFacB)
    IF (ALLOCATED(MeshFacPhy))       DEALLOCATE(MeshFacPhy)
    IF (ALLOCATED(MeshFacPec))       DEALLOCATE(MeshFacPec)
    IF (ALLOCATED(MeshFacPmc))       DEALLOCATE(MeshFacPmc)
    IF (ALLOCATED(MeshFacAbs))       DEALLOCATE(MeshFacAbs)
    IF (ALLOCATED(MeshFacPer))       DEALLOCATE(MeshFacPer)
    IF (ALLOCATED(MeshFacArt))       DEALLOCATE(MeshFacArt)
    IF (ALLOCATED(MeshFacTfsf))      DEALLOCATE(MeshFacTfsf)
    IF (ALLOCATED(MeshFacPoy))       DEALLOCATE(MeshFacPoy)
    IF (ALLOCATED(MeshFacLin))       DEALLOCATE(MeshFacLin)
    IF (ALLOCATED(MeshFacQuadr))     DEALLOCATE(MeshFacQuadr)
    IF (ALLOCATED(MeshFacTri))       DEALLOCATE(MeshFacTri)
    IF (ALLOCATED(MeshFacTriLin))    DEALLOCATE(MeshFacTriLin)
    IF (ALLOCATED(MeshFacTriQuadr))  DEALLOCATE(MeshFacTriQuadr)
    IF (ALLOCATED(MeshFacQuad))      DEALLOCATE(MeshFacQuad)    
    IF (ALLOCATED(MeshFacQuadLin))   DEALLOCATE(MeshFacQuadLin)    
    IF (ALLOCATED(MeshFacQuadQuadr)) DEALLOCATE(MeshFacQuadQuadr)
    IF (ALLOCATED(MeshFacInter))     DEALLOCATE(MeshFacInter)
    IF (ALLOCATED(MeshFacHyb))       DEALLOCATE(MeshFacHyb)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_DERIVED_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_COMPLEMENTARY_FACES
    !------------------------------------------------------------------
    !---- Deallocate all complementary face lists
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate faces
    !------------------------------------------------------------------
    IF (ALLOCATED(CompFacPer))      DEALLOCATE(CompFacPer)
    IF (ALLOCATED(CompFacTfsf))     DEALLOCATE(CompFacTfsf)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_COMPLEMENTARY_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPARE_FACE_STRUCTURES(face1, face2, isIdentical)
    !------------------------------------------------------------------
    !---- Tell if 2 face structure arrays are identical
    !
    !---- IN
    !---- face*       : face structures to compare
    !
    !---- OUT
    !---- isIdentical : TRUE if structures are identical
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),          INTENT(OUT) :: isIdentical
    
    TYPE(face), DIMENSION(:), INTENT(IN)  :: face1, face2
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                        :: size1, size2, iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compare structures
    !------------------------------------------------------------------
    isIdentical = .TRUE.
    size1       = SIZE(face1)
    size2       = SIZE(face2)
    !
    IF (size1 /= size2) THEN
       isIdentical = .FALSE.
       RETURN
    ENDIF
    !
    DO iFace=1,size1
       IF (face1(iFace)%faceType         /= face2(iFace)%faceType)        isIdentical = .FALSE.
       IF (face1(iFace)%isHO         .NEQV. face2(iFace)%isHO)            isIdentical = .FALSE.
       IF (face1(iFace)%isTfsf       .NEQV. face2(iFace)%isTfsf)          isIdentical = .FALSE.
       IF (face1(iFace)%isInterface  .NEQV. face2(iFace)%isInterface)     isIdentical = .FALSE.
       IF (face1(iFace)%isInner      .NEQV. face2(iFace)%isInner)         isIdentical = .FALSE.
       IF (face1(iFace)%isBoundary   .NEQV. face2(iFace)%isBoundary)      isIdentical = .FALSE.
       IF (face1(iFace)%isPeriodic   .NEQV. face2(iFace)%isPeriodic)      isIdentical = .FALSE.
       IF (face1(iFace)%isArtificial .NEQV. face2(iFace)%isArtificial)    isIdentical = .FALSE.
       IF (face1(iFace)%isHybrid     .NEQV. face2(iFace)%isHybrid)        isIdentical = .FALSE.
       IF (face1(iFace)%nVerts           /= face2(iFace)%nVerts)          isIdentical = .FALSE.
       IF (face1(iFace)%nLinVerts        /= face2(iFace)%nLinVerts)       isIdentical = .FALSE.
       IF (face1(iFace)%nHoVerts         /= face2(iFace)%nHoVerts)        isIdentical = .FALSE.
       IF (face1(iFace)%geoOrder         /= face2(iFace)%geoOrder)        isIdentical = .FALSE.
       IF (face1(iFace)%numFace          /= face2(iFace)%numFace)         isIdentical = .FALSE.
       IF (face1(iFace)%log              /= face2(iFace)%log)             isIdentical = .FALSE.
       IF (face1(iFace)%measure          /= face2(iFace)%measure)         isIdentical = .FALSE.
       IF (face1(iFace)%boundaryFaceNum  /= face2(iFace)%boundaryFaceNum) isIdentical = .FALSE.
       !
       IF (.NOT. ALL(face1(iFace)%numLocal == face2(iFace)%numLocal))     isIdentical = .FALSE.
       IF (.NOT. ALL(face1(iFace)%numGeom  == face2(iFace)%numGeom))      isIdentical = .FALSE.
       IF (.NOT. ALL(face1(iFace)%normal   == face2(iFace)%normal))       isIdentical = .FALSE.
       IF (.NOT. ALL(face1(iFace)%neighb   == face2(iFace)%neighb))       isIdentical = .FALSE.
       IF (.NOT. ALL(face1(iFace)%perm     == face2(iFace)%perm))         isIdentical = .FALSE.
       IF (.NOT. ALL(face1(iFace)%axis     == face2(iFace)%axis))         isIdentical = .FALSE.
       !
       IF (ALLOCATED(face1(iFace)%vertex) .AND. ALLOCATED(face2(iFace)%vertex)) THEN
          IF (.NOT. ALL(face1(iFace)%vertex  == face2(iFace)%vertex))     isIdentical = .FALSE.
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPARE_FACE_STRUCTURES

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_FACES
