MODULE MESH_FIND_FACES

  !====================================================================
  ! Everything related to finding faces and derived faces 
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_DERIVED_FACE_IN_LIST_FROM_VERTICES(facVert, facList, &
       sizeList, foundFace)
    !------------------------------------------------------------------
    !---- Find the number of a derived face from its vertices
    !
    !---- IN
    !---- facVert   : vertices of the derived face
    !---- facList   : list of faces to look up in
    !---- sizeList  : size of facList
    !
    !---- OUT
    !---- foundFace : index of the face in the derived face list
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_FACES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(IN)  :: sizeList    
    INTEGER(C_INT),                   INTENT(OUT) :: foundFace
    INTEGER(C_INT),     DIMENSION(:), INTENT(IN)  :: facVert

    TYPE(derived_face), DIMENSION(:), INTENT(IN)  :: facList
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                               :: isEqual

    INTEGER(C_INT)                                :: idFace, iFace, &
         nVertsIn, nVertsLoc, nVerts
    INTEGER(C_INT), DIMENSION(nvFaceMax)          :: vert, locVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Find face number
    !------------------------------------------------------------------
    foundFace        = 0
    vert(:)          = 0
    nVertsIn         = SIZE(facVert)
    vert(1:nVertsIn) = facVert(1:nVertsIn)
    CALL INSERTION_SORT_INT(vert(1:nVertsIn))
    !
    DO idFace=1,sizeList
       !---- It might be that the face is not filled yet
       iFace = facList(idFace)%faceNum
       IF (.NOT. ALLOCATED(MeshFac(iFace)%vertex)) CYCLE
       !
       locVert(:)           = 0
       nVertsLoc            = MeshFac(iFace)%nLinVerts
       locVert(1:nVertsLoc) = MeshFac(iFace)%vertex(1:nVertsLoc)
       CALL INSERTION_SORT_INT(locVert(1:nVertsLoc))
       nVerts  = MAX(nVertsIn, nVertsLoc)
       isEqual = ALL(locVert(1:nVerts) == vert(1:nVerts))
       !
       IF (isEqual) THEN
          foundFace = idFace
          RETURN
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FIND_DERIVED_FACE_IN_LIST_FROM_VERTICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_DERIVED_FACE_IN_LIST_FROM_INDEX(faceNum, facList, &
       sizeList, foundFace)
    !------------------------------------------------------------------
    !---- Find the number of a derived face from its vertices
    !
    !---- IN
    !---- faceNum   : global index of the face
    !---- facList   : list of faces to look up in
    !---- sizeList  : size of facList
    !
    !---- OUT
    !---- foundFace : index of the face in the derived face list
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_FACES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(IN)  :: faceNum, sizeList
    INTEGER(C_INT),                   INTENT(OUT) :: foundFace

    TYPE(derived_face), DIMENSION(:), INTENT(IN)  :: facList
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                                :: idFace, iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Find face number
    !------------------------------------------------------------------
    foundFace = 0
    !
    DO idFace=1,sizeList
       iFace = facList(idFace)%faceNum
       IF (iFace == faceNum) THEN
          foundFace = idFace
          RETURN
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FIND_DERIVED_FACE_IN_LIST_FROM_INDEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_FACE_IN_LIST_FROM_VERTICES(facVert, facList, sizeList, &
       foundFace)
    !------------------------------------------------------------------
    !---- Find the index of a face from its vertices
    !
    !---- IN
    !---- facVert   : vertices of the face
    !---- facList   : list of faces to look up in
    !---- sizeList  : size of facList
    !
    !---- OUT
    !---- foundFace : index of the face in the face list
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_FACES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: sizeList
    INTEGER(C_INT),               INTENT(OUT) :: foundFace
    INTEGER(C_INT), DIMENSION(:), INTENT(IN)  :: facVert

    TYPE(face),     DIMENSION(:), INTENT(IN)  :: facList
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                           :: isEqual

    INTEGER(C_INT)                            :: iFace, nVertsIn, &
         nVertsLoc, nVerts
    INTEGER(C_INT), DIMENSION(nvFaceMax)      :: vert, locVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Find face number
    !------------------------------------------------------------------
    foundFace        = 0
    vert             = 0
    nVertsIn         = SIZE(facVert)
    vert(1:nVertsIn) = facVert(1:nVertsIn)
    CALL INSERTION_SORT_INT(vert(1:nVertsIn))
    !
    DO iFace=1,sizeList
       !---- It might be that the face is not filled yet
       IF (.NOT. ALLOCATED(facList(iFace)%vertex)) CYCLE
       !
       locVert              = 0
       nVertsLoc            = facList(iFace)%nVerts
       locVert(1:nVertsLoc) = facList(iFace)%vertex(1:nVertsLoc)
       CALL INSERTION_SORT_INT(locVert(1:nVertsLoc))
       nVerts  = MAX(nVertsIn, nVertsLoc)
       isEqual = ALL(locVert(1:nVerts) == vert(1:nVerts))
       !
       IF (isEqual) THEN
          foundFace = iFace
          RETURN
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FIND_FACE_IN_LIST_FROM_VERTICES

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_FIND_FACES
