MODULE MESH_FIND_POINT_IN_ELEMENTS

  !====================================================================
  ! Routines related to finding point in elements
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE DOF_PERMUTATIONS
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Interface for overloading IS_POINT_IN_FACE
  !--------------------------------------------------------------------
  INTERFACE IS_POINT_IN_FACE
     MODULE PROCEDURE IS_POINT_IN_FACE_FROM_INDEX, &
          IS_POINT_IN_FACE_FROM_VERTICES
  END INTERFACE IS_POINT_IN_FACE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_POINT_IN_CELLS(ptIn, cellNum)
    !------------------------------------------------------------------
    !---- Find the cell containing a given point
    !
    !---- IN
    !---- ptIn     : point coordinates
    !
    !---- OUT
    !---- cellNum  : index of the cell containing the point
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------    
    INTEGER(C_INT),               INTENT(OUT) :: cellNum

    REAL(C_DOUBLE), DIMENSION(:), INTENT(IN)  :: ptIn
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                           :: isInCell
    
    CHARACTER(LEN=4, KIND=C_CHAR)             :: cellType

    INTEGER(C_INT)                            :: iCell, iFound, globalFound

    REAL(C_DOUBLE), DIMENSION(3)              :: pt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Find unique a cell containing given point
    !------------------------------------------------------------------
    pt = ptIn
    !
    DO ! Loop over cells and count how many times it was found
       iFound  = 0
       cellNum = 0
       !
       DO iCell=1,Mesh%nCell
          CALL IS_POINT_IN_CELL(pt, iCell, isInCell)
          !
          IF (isInCell) THEN
             cellNum = iCell
             iFound  = iFound + 1
          ENDIF
       ENDDO
       !
       CALL GLOBAL_OP(iFound, MPI_SUM, globalFound)
       iFound = globalFound
       !
       !---- Point was found only once
       IF (iFound == 1) EXIT
       !
       !---- If the point is found several times, it lies exactly on a
       !---- face or a ridge. In this case, the coordinates of the point
       !---- are slightly moved toward the center of the mesh, and the
       !---- loop restarts
       IF (iFound /= 1) THEN
          pt(:) = pt(:) + 10.0*Mesh%tolerance1D*(Mesh%coorCenter(:) - ptIn(:))
          pt(:) = pt(:) + 0.1*Mesh%tolerance1D
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FIND_POINT_IN_CELLS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_CELL(pt, iCell, isInCell)
    !------------------------------------------------------------------
    !---- Tells if a point is in a given cell
    !
    !---- IN
    !---- pt       : point coordinates
    !---- iCell    : index of the cell
    !
    !---- OUT
    !---- isInCell : TRUE if point is in cell
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),              INTENT(INOUT) :: isInCell

    INTEGER(C_INT),               INTENT(INOUT) :: iCell

    REAL(C_DOUBLE), DIMENSION(:), INTENT(INOUT) :: pt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)               :: cellType
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call the appropriate method depending on cellType
    !------------------------------------------------------------------
    cellType = MeshCell(iCell)%cellType
    !
    SELECT CASE(cellType)
    CASE('TET1', 'TET2', 'TET3', 'TET4')
       CALL IS_POINT_IN_TET(pt, iCell, isInCell)
    CASE('HEX1', 'HEX2')
       CALL IS_POINT_IN_HEX(pt, iCell, isInCell)       
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_CELL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_TET(pt, iCell, isInCell)
    !------------------------------------------------------------------
    !---- Tells if a point is in a given tet
    !
    !---- IN
    !---- pt       : point coordinates
    !---- iCell    : index of the tet
    !
    !---- OUT
    !---- isInCell : TRUE if point is in tet
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MATRICES_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),              INTENT(OUT) :: isInCell

    INTEGER(C_INT),               INTENT(IN)  :: iCell

    REAL(C_DOUBLE), DIMENSION(:), INTENT(IN)  :: pt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: det0, det
    REAL(C_DOUBLE), DIMENSION(3)              :: v1, v2, v3, v4
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if point is in tet
    !------------------------------------------------------------------
    isInCell = .FALSE.
    !
    !---- Retrieve vertices coordinates and form matrix
    v1(:) = MeshVert(MeshCell(iCell)%vertex(1))%coor(:)
    v2(:) = MeshVert(MeshCell(iCell)%vertex(2))%coor(:)
    v3(:) = MeshVert(MeshCell(iCell)%vertex(3))%coor(:)
    v4(:) = MeshVert(MeshCell(iCell)%vertex(4))%coor(:)
    !
    !---- Successively check determinants
    det0 = SPECIAL_DET_4X4(v1,v2,v3,v4)
    !
    det  = SPECIAL_DET_4X4(pt,v2,v3,v4)/det0
    IF (det < 0.0) RETURN
    !
    det  = SPECIAL_DET_4X4(v1,pt,v3,v4)/det0
    IF (det < 0.0) RETURN
    !
    det  = SPECIAL_DET_4X4(v1,v2,pt,v4)/det0
    IF (det < 0.0) RETURN
    !
    det  = SPECIAL_DET_4X4(v1,v2,v3,pt)/det0
    IF (det < 0.0) RETURN
    !
    !---- If not returned, then the probe is in this tet
    isInCell = .TRUE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_TET

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_HEX(pt, iCell, isInCell)
    !------------------------------------------------------------------
    !---- Tells if a point is in a given hex
    !
    !---- IN
    !---- pt       : point coordinates
    !---- iCell    : index of the hex
    !
    !---- OUT
    !---- isInCell : TRUE if point is in hex
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MATRICES_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),              INTENT(OUT) :: isInCell

    INTEGER(C_INT),               INTENT(IN)  :: iCell

    REAL(C_DOUBLE), DIMENSION(:), INTENT(IN)  :: pt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: xMin, xMax, yMin, yMax, zMin, zMax
    REAL(C_DOUBLE), DIMENSION(3)              :: v1, v2, v4, v5
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if point is in hex
    !------------------------------------------------------------------
    isInCell = .FALSE.
    !
    !---- Retrieve vertices coordinates
    v1(:) = MeshVert(MeshCell(iCell)%vertex(1))%coor(:)
    v2(:) = MeshVert(MeshCell(iCell)%vertex(2))%coor(:)
    v4(:) = MeshVert(MeshCell(iCell)%vertex(4))%coor(:)
    v5(:) = MeshVert(MeshCell(iCell)%vertex(5))%coor(:)
    !
    !---- Retrieve min/max values
    xMin  = v1(1)
    xMax  = v2(1)
    yMin  = v1(2)
    yMax  = v4(2)
    zMin  = v1(3)
    zMax  = v5(3)
    !
    !---- Successively check value bounds
    IF (.NOT. ((xMin <= pt(1)) .AND. (pt(1) <= xMax))) RETURN
    IF (.NOT. ((yMin <= pt(2)) .AND. (pt(2) <= yMax))) RETURN
    IF (.NOT. ((zMin <= pt(3)) .AND. (pt(3) <= zMax))) RETURN
    !
    !---- If not returned, then the probe is in this tet
    isInCell = .TRUE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_HEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_FACE_FROM_INDEX(pPoint, iFace, isInFace)
    !------------------------------------------------------------------
    !---- Check if a point pPoint is in a face given by its
    !---- vertices coordinates pFace
    !
    !---- IN
    !---- pPoint   : coordinates of point
    !---- iFace    : index of the face
    !
    !---- OUT
    !---- isInFace : TRUE if point is in face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),              INTENT(INOUT) :: isInFace

    INTEGER(C_INT),               INTENT(IN)    :: iFace

    REAL(C_DOUBLE), DIMENSION(:), INTENT(INOUT) :: pPoint
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)               :: faceType

    INTEGER(C_INT)                              :: nVerts, iVert

    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: pFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call the appropriate method depending on faceType
    !------------------------------------------------------------------
    faceType = MeshFac(iFace)%faceType
    nVerts   = MeshFac(iFace)%nLinVerts
    ALLOCATE(pFace(3,nVerts))
    DO iVert=1,nVerts
       pFace(:,iVert) = MeshVert(MeshFac(iFace)%vertex(iVert))%coor(:)
    ENDDO
    !
    SELECT CASE(faceType)
    CASE('TRI1', 'TRI2')
       CALL IS_POINT_IN_TRIANGLE(pPoint,pFace,isInFace)
    CASE('QUA1', 'QUA2')
       CALL IS_POINT_IN_QUAD    (pPoint,pFace,isInFace)
    END SELECT
    !------------------------------------------------------------------
    DEALLOCATE(pFace)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_FACE_FROM_INDEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_FACE_FROM_VERTICES(pPoint, pFace, isInFace)
    !------------------------------------------------------------------
    !---- Check if a point pPoint is in a face given by its
    !---- vertices coordinates pFace
    !
    !---- IN
    !---- pPoint   : coordinates of point
    !---- pFace    : coordinates of the vertices of the face
    !
    !---- OUT
    !---- isInFace : TRUE if point is in face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),                INTENT(INOUT) :: isInFace

    REAL(C_DOUBLE), DIMENSION(:),   INTENT(INOUT) :: pPoint
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(INOUT) :: pFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                                :: nVerts
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Select routine depending on input
    !------------------------------------------------------------------
    nVerts = SIZE(pFace,2)
    !
    SELECT CASE(nVerts)
    CASE(3) ! Triangle
       CALL IS_POINT_IN_TRIANGLE(pPoint,pFace,isInFace)
    CASE(4) ! Quadrangle
       CALL IS_POINT_IN_QUAD    (pPoint,pFace,isInFace)
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_FACE_FROM_VERTICES  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_TRIANGLE(pt, pTri, isInTri)
    !------------------------------------------------------------------
    !---- Check if a given point is in a triangular face given
    !---- by its vertices coordinates
    !
    !---- IN
    !---- pt      : coordinates of point
    !---- pTri    : coordinates of the vertices of the quad
    !
    !---- OUT
    !---- isInTri : TRUE if point is in triangle
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),                INTENT(OUT) :: isInTri

    REAL(C_DOUBLE), DIMENSION(:),   INTENT(IN)  :: pt
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(IN)  :: pTri
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                              :: prod1, prod2, d, distance
    REAL(C_DOUBLE), DIMENSION(3)                :: AB, AC, BC, AM, BM, normal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Tri is defined with 3 points (A,B,C)
    !
    !     C 
    !     |\
    !     | \ 
    !     |  \
    !     | x \  
    !     |  M \ 
    !     |_____\
    !    A       B 
    !------------------------------------------------------------------
    isInTri   = .FALSE.
    !
    !---- Compute AB, AC, AM and a normal vector
    AB(:) = pTri(:,2) - pTri(:,1) 
    AC(:) = pTri(:,3) - pTri(:,1)
    BC(:) = pTri(:,3) - pTri(:,2)
    AM(:) = pt(:)     - pTri(:,1)
    BM(:) = pt(:)     - pTri(:,2)
    !
    !---- Check if (0 < AM.AB < AB.AB)
    prod1 = DOT_PRODUCT(AM,AB)
    prod2 = DOT_PRODUCT(AB,AB)
    IF ((prod1 < 0.0) .OR. (prod1 > prod2)) RETURN
    !
    !---- Check if (0 < AM.AC < AC.AC)
    prod1 = DOT_PRODUCT(AM,AC)
    prod2 = DOT_PRODUCT(AC,AC)
    IF ((prod1 < 0.0) .OR. (prod1 > prod2)) RETURN
    !
    !---- Check if (0 < BM.BC < BC.BC)
    prod1 = DOT_PRODUCT(BM,BC)
    prod2 = DOT_PRODUCT(BC,BC)
    IF ((prod1 < 0.0) .OR. (prod1 > prod2)) RETURN    
    !
    !---- Compute plane equation ax + by + cz + d = 0
    !---- (in fact a, b and c are equal to the normal 
    !---- components, so we only compute d = - A.n
    CALL CROSS_PROD(AB,AC,normal)
    d = -DOT_PRODUCT(pTri(:,1),normal)
    !
    !---- Compute distance of M to plane
    !---- distance = |a*x + b*y + c*z + d|
    distance = ABS(DOT_PRODUCT(normal,pt) + d)    
    IF (distance > Mesh%tolerance1D) RETURN
    !
    isInTri = .TRUE.    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_TRIANGLE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_POINT_IN_QUAD(pt, pQuad, isInQuad)
    !------------------------------------------------------------------
    !---- Check if a given point is in a quad face given by its
    !---- vertices coordinates
    !
    !---- IN
    !---- pt       : coordinates of point
    !---- pQuad    : coordinates of the vertices of the quad
    !
    !---- OUT
    !---- isInQuad : TRUE if point is in quad
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),                INTENT(INOUT) :: isInQuad

    REAL(C_DOUBLE), DIMENSION(:),   INTENT(INOUT) :: pt
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(INOUT) :: pQuad
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                                :: prod1, prod2, d, distance
    REAL(C_DOUBLE), DIMENSION(3)                  :: AB, AD, AM, normal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Quad is defined with 4 points (A,B,C,D)
    !
    !    D ______ C
    !     |      |
    !     |  x   |
    !     | M    |
    !     |______|
    !    A        B
    !------------------------------------------------------------------
    isInQuad  = .FALSE.
    !
    !---- Compute AB, AD, AM and a normal vector
    AB(:) = pQuad(:,2) - pQuad(:,1) 
    AD(:) = pQuad(:,4) - pQuad(:,1) 
    AM(:) = pt(:)      - pQuad(:,1)
    CALL CROSS_PROD(AB,AD,normal)
    !
    !---- Check if (0 < AM.AB < AB.AB)
    prod1 = DOT_PRODUCT(AM,AB)
    prod2 = DOT_PRODUCT(AB,AB)
    IF ((prod1 < 0.0) .OR. (prod1 > prod2)) RETURN
    !
    !---- Check if (0 < AM.AD < AD.AD)
    prod1 = DOT_PRODUCT(AM,AD)
    prod2 = DOT_PRODUCT(AD,AD)
    IF ((prod1 < 0.0) .OR. (prod1 > prod2)) RETURN
    !
    !---- Compute plane equation ax + by + cz + d = 0
    !---- (in fact a, b and c are equal to the normal 
    !---- components, so we only compute d = - A.n
    d = -DOT_PRODUCT(pQuad(:,1),normal)
    !
    !---- Compute distance of M to plane
    !---- distance = |a*x + b*y + c*z + d|
    distance = ABS(DOT_PRODUCT(normal,pt) + d)    
    IF (distance > Mesh%tolerance1D) RETURN
    !
    isInQuad = .TRUE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE IS_POINT_IN_QUAD

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_FIND_POINT_IN_ELEMENTS
