MODULE MESH_INTERFACES

  !====================================================================
  ! Routines related to the computation of interface faces between
  ! materials with different logs
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_MESH_INTERFACES
    !------------------------------------------------------------------
    !---- Build all material interfaces in the mesh
    !
    !---- IN
    !
    !---- OUT
    !    
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE MESH_ELEMENTS
    USE PARALLEL_STRUCTURES
    USE MESH_TYPE_CONVERSIONS
    USE MESH_NEIGHBORS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)        :: cellType, faceType

    INTEGER(C_INT)                       :: iCell1, iCell2, iNeighb, &
         log1, log2, nNeighb, nVerts, nLinVerts, nHoVerts, geoOrder, &
         logInter, iFac, faceNum
    INTEGER(C_INT), DIMENSION(nvCellMax) :: buffer

    TYPE(cell)                           :: locCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Fill the structure
    !------------------------------------------------------------------
    DO iCell1=1,Mesh%nCell
       locCell = MeshCell(iCell1)
       log1    = locCell%log
       nNeighb = locCell%nNeighb
       !
       DO iNeighb=1,nNeighb
          !---- Get the neighbor of current tetrahedron
          iCell2 = locCell%neighb(iNeighb)
          !
          !---- Cycle if it is a boundary face
          IF (iCell2 > Mesh%nCell) CYCLE
          !
          !---- Cycle if logs are identical or log1 > log2. This
          !---- convention avoids the creation of interface faces in double
          log2 = MeshCell(iCell2)%log
          IF (log1 >= log2) CYCLE
          !
          CALL CELL_TYPE_TO_FACE_TYPE(locCell%cellType, faceType)
          CALL FACE_TYPE_TO_VERTS(faceType, nVerts, nLinVerts, nHoVerts, geoOrder)
          buffer(1:nVerts) = locCell%vertex(locCell%permType(iNeighb,1:nVerts))
          !
          CALL GET_INTERFACE_LOG(log1, log2, logInter)
          CALL ADD_FACE(buffer(1:nVerts), logInter, faceType, faceNum)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Remove duplicates and recompute derived face lists
    !------------------------------------------------------------------
    CALL REMOVE_FACE_DUPLICATES
    CALL BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
    IF (idWrite) WRITE(6,'(a)') 'Finished building mesh interfaces'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_MESH_INTERFACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_MATERIAL_BOUNDARY(logMat, faceList, compFaceList, sizeFaceList)
    !------------------------------------------------------------------
    !---- Build structures defining a material boundary
    !
    !---- IN
    !---- logMat       : material log
    !
    !---- OUT
    !---- sizeFaceList : size of lists
    !---- faceList     : derived face list
    !---- compFaceList : complementary interface list
    !        
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE MESH_ELEMENTS
    USE PARALLEL_STRUCTURES
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                  DIMENSION(:), INTENT(IN)  :: logMat
    INTEGER(C_INT),                                INTENT(OUT) :: sizeFaceList

    TYPE(derived_face),              DIMENSION(:), ALLOCATABLE :: faceList

    TYPE(complement_interface_face), DIMENSION(:), ALLOCATABLE :: compFaceList
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                                             :: iCell, &
         log, iNeighb, nNeighb, iCellNeighb, logNeighb, faceNum,         &
         nMatCell, iMatCell, sizeCompFaceList
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Count the number of elements
    !------------------------------------------------------------------
    nMatCell = 0
    DO iCell=1,Mesh%nCell
       log = MeshCell(iCell)%log
       IF (ANY(log == logMat)) nMatCell = nMatCell + 1
    ENDDO
    !------------------------------------------------------------------
    !---- Build boundary
    !------------------------------------------------------------------
    sizeFaceList     = 0
    sizeCompFaceList = 0
    iMatCell         = 0
    !
    DO iCell=1,Mesh%nCell
       log     = MeshCell(iCell)%log
       nNeighb = MeshCell(iCell)%nNeighb
       IF (.NOT. (ANY(log == logMat))) CYCLE
       iMatCell = iMatCell + 1
       !
       DO iNeighb=1,nNeighb
          iCellNeighb = MeshCell(iCell)%neighb(iNeighb)
          !---- Neighbor is a physical cell
          IF (iCellNeighb <= Mesh%nCell) logNeighb = MeshCell(iCellNeighb)%log
          !---- Neighbor is a ghost    cell
          IF (iCellNeighb >  Mesh%nCell) logNeighb = 9999999
          !
          IF (.NOT. (ANY(logNeighb == logMat))) THEN
             faceNum = MeshCell(iCell)%facNum(iNeighb)
             CALL EXTEND_DERIVED_FACE_ARRAY(faceList, sizeFaceList)
             CALL EXTEND_COMPLEMENT_INTERFACE_FACE_ARRAY(compFaceList, sizeCompFaceList)
             faceList(sizeFaceList)%faceNum         = faceNum
             compFaceList(sizeCompFaceList)%cellIn  = iMatCell
             compFaceList(sizeCompFaceList)%cellOut = nMatCell + sizeCompFaceList
             compFaceList(sizeCompFaceList)%iNeighb = iNeighb
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished building boundary material'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_MATERIAL_BOUNDARY

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE BUILD_MESH_BOUNDARIES
    !------------------------------------------------------------------
    !---- Build mesh boundaries from known face logs stored in MeshCell
    !---- array. This is mainly used for visualization purpose
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE MESH_ELEMENTS
    USE MESH_FACES_LOGS
    USE PARALLEL_STRUCTURES
    USE MESH_TYPE_CONVERSIONS
    USE MESH_NEIGHBORS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)                     :: faceType

    LOGICAL(C_BOOL)                                   :: isBoundaryLog
    
    INTEGER(C_INT)                                    :: iCell, iNeighb, &
         nNeighb, faceLog, nVerts, nLinVerts, nHoVerts, geoOrder, faceNum
    INTEGER(C_INT), DIMENSION(nvCellMax)              :: buffer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Fill the structure
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       nNeighb = MeshCell(iCell)%nNeighb
       !
       DO iNeighb=1,nNeighb
          faceLog = MeshCell(iCell)%facLog(iNeighb)
          CALL IS_CONFORMAL_BOUNDARY_LOG(faceLog, isBoundaryLog)
          !
          IF (isBoundaryLog) THEN
             CALL CELL_TYPE_TO_FACE_TYPE(MeshCell(iCell)%cellType, faceType)
             CALL FACE_TYPE_TO_VERTS(faceType, nVerts, nLinVerts, nHoVerts, geoOrder)
             buffer(1:nVerts) = MeshCell(iCell)%vertex(MeshCell(iCell)%permType(iNeighb,1:nVerts))
             !
             CALL ADD_FACE(buffer(1:nVerts), faceLog, faceType, faceNum)
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Remove duplicates and recompute derived face lists
    !------------------------------------------------------------------
    CALL REMOVE_FACE_DUPLICATES
    CALL BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished building mesh boundary faces'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_MESH_BOUNDARIES

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE FORCE_MESH_BOUNDARIES(boundaryTag)
    !------------------------------------------------------------------
    !---- Create mesh boundary in the case of missing neighbors
    !---- and force it to a given tag
    !
    !---- IN
    !---- boundaryTag : tag for boundary faces
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE MESH_ELEMENTS
    USE PARALLEL_STRUCTURES
    USE MESH_TYPE_CONVERSIONS
    USE MESH_NEIGHBORS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),           INTENT(IN) :: boundaryTag
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)        :: faceType
    
    INTEGER(C_INT)                       :: iFace, iSide, iCell
    INTEGER(C_INT), DIMENSION(nvCellMax) :: buffer    
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Fill the structure
    !------------------------------------------------------------------
    DO iFace=1,Mesh%nFac
       DO iSide=1,2
          iCell = MeshFac(iFace)%neighb(iSide)
          IF (iCell == 0) MeshFac(iFace)%log = boundaryTag
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Remove duplicates and recompute derived face lists
    !------------------------------------------------------------------
    CALL REMOVE_FACE_DUPLICATES
    CALL BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished building mesh boundary faces'
    !------------------------------------------------------------------
  END SUBROUTINE FORCE_MESH_BOUNDARIES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_INTERFACE_LOG(log1, log2, logInter)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: log1, log2
    INTEGER(C_INT), INTENT(OUT) :: logInter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute log
    !------------------------------------------------------------------
    logInter = 10000*(log1+1)*(log2+1)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_INTERFACE_LOG

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_COMPLEMENT_INTERFACE_FACE_ARRAY(cFaceArray, cFaceArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to a complementary face array
    !
    !---- IN
    !---- cFaceArray     : array of "complementary_face" structure
    !---- cFaceArraySize : number of elements in cFaceArray
    !
    !---- OUT
    !---- cFaceArray     : modified in the process
    !---- cFaceArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                  INTENT(INOUT)             :: cFaceArraySize
    
    TYPE(complement_interface_face), DIMENSION(:), ALLOCATABLE :: cFaceArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                            :: isArrayFull, isAllocated
    
    INTEGER(C_INT)                                             :: iFace
    
    TYPE(complement_interface_face), DIMENSION(:), ALLOCATABLE :: tempFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(cFaceArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(cFaceArray) < cFaceArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF    
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN    
       IF (isAllocated) THEN
          IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
          ALLOCATE(tempFace(cFaceArraySize))
          tempFace = cFaceArray
          DEALLOCATE(cFaceArray)
       ENDIF
       !
       ALLOCATE(cFaceArray(cFaceArraySize + cFaceBlockSize))
       DO iFace=cFaceArraySize+1,cFaceArraySize+cFaceBlockSize
          CALL INITIALIZE_COMPLEMENT_INTERFACE_FACE_STRUCTURE(cFaceArray(iFace))
       ENDDO
       !
       cFaceArray(1:cFaceArraySize) = tempFace(1:cFaceArraySize)
       IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
    ENDIF
    !
    cFaceArraySize = cFaceArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_COMPLEMENT_INTERFACE_FACE_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_COMPLEMENT_INTERFACE_FACE_STRUCTURE(cFaceStructure)
    !------------------------------------------------------------------
    !---- Initialize complementary interface face structure
    !
    !---- IN
    !---- cFaceStructure : complementary interface face structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(complement_interface_face), INTENT(INOUT) :: cFaceStructure
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize complementary interface face structure
    !------------------------------------------------------------------
    cFaceStructure%cellIn  = 0
    cFaceStructure%cellOut = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_COMPLEMENT_INTERFACE_FACE_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_INTERFACES
