MODULE MESH_NEIGHBORS

  !====================================================================
  ! Routines related to neighbors finding and face building
   !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_FACES
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! nFaceToSort   ==> total number of faces to sort
  ! nFacHyb       ==> total number of hybrid faces
  ! faceSort      ==> array used locally for the lexicographic sort of face
  !                   vertices. This makes the inner faces appear 2 by 2
  ! hybTol2D      ==> 2D tolerance for hybrid faces cleaning
  ! localHybFac   ==> local array of hybrid faces for sorting
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL), DIMENSION(:),   ALLOCATABLE :: removeFace  

  INTEGER(C_INT)                               :: nFaceToSort, nFacHyb
  INTEGER(C_INT),  DIMENSION(:,:), ALLOCATABLE :: faceSort

  REAL(C_DOUBLE)                               :: hybTol2D

  TYPE(face),      DIMENSION(:),   ALLOCATABLE :: localHybFac
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_MESH_NEIGHBORS_AND_FACES
    !------------------------------------------------------------------
    !---- This routine builds faces that were not included in the
    !---- mesh, and determines the neighbors of each cell
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Do stuff
    !------------------------------------------------------------------
    CALL PREPROCESS_CONFORMAL_INNER_FACES
    CALL PROCESS_CONFORMAL_INNER_FACES
    CALL REMOVE_FACE_DUPLICATES
    CALL PREPROCESS_HYBRID_FACES
    CALL REMOVE_LARGE_HYBRID_FACES
    CALL BUILD_DERIVED_FACES
    CALL PROCESS_BOUNDARY_FACES
    CALL BUILD_DERIVED_FACES
    CALL PREPROCESS_HYBRID_FACES
    CALL PROCESS_HYBRID_FACES
    CALL BUILD_DERIVED_FACES
    CALL ATTACH_FACES_TO_CELLS    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (addInnerFaces) CALL TEST_CELLS_FACES
    CALL TEST_FACES_NEIGHBORS    
    CALL TEST_CELLS_NEIGHBORS    
    IF (addInnerFaces) CALL TEST_NEIGHBORS_COHERENCE
    !------------------------------------------------------------------
    IF (ALLOCATED(faceSort))    DEALLOCATE(faceSort)
    IF (ALLOCATED(removeFace))  DEALLOCATE(removeFace)
    IF (ALLOCATED(localHybFac)) DEALLOCATE(localHybFac)
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh neighbors and faces'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
  END SUBROUTINE BUILD_MESH_NEIGHBORS_AND_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_CONFORMAL_INNER_FACES
    !------------------------------------------------------------------
    !---- This routine fills a temporary array used to find neighbors
    !---- for conformal inner faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                       :: iFace, iCell, faceNum, &
         nVerts, nLinVerts, nHoVerts, geoOrder, nFaces
    INTEGER(C_INT), DIMENSION(nvFaceMax) :: faceVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- For each cell and face, retrieve vertex numbers and locally
    !---- sort them by ascending order. Then, perform a global sort
    !---- For array faceSort:
    !---- - Col 1 to 4 are for vertices
    !---- - Col 5      is  for local face number
    !---- - Col 6      is  for cell number
    !------------------------------------------------------------------
    nFaceToSort = 4*Mesh%nCellTet + 6*Mesh%nCellHex
    IF (ALLOCATED(faceSort)) DEALLOCATE(faceSort)
    ALLOCATE(faceSort(6,nFaceToSort+1)) ! This is not a mistake
    faceSort(:,:) = 0
    !
    faceNum  = 0
    DO iCell=1,Mesh%nCell
       nFaces = MeshCell(iCell)%nFaces
       !
       DO iFace=1,nFaces
          faceVert(:) = 0
          CALL GET_FACE_LINEAR_VERTICES_FROM_CELL(iCell, iFace, faceVert)
          CALL INSERTION_SORT_INT(faceVert(1:4))
          faceNum               = faceNum + 1
          faceSort(1:4,faceNum) = faceVert(1:4)
          faceSort(5,  faceNum) = iFace
          faceSort(6,  faceNum) = iCell
       ENDDO
    ENDDO
    !
    CALL SORT_ARRAY(6,nFaceToSort,faceSort)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_CONFORMAL_INNER_FACES

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE PROCESS_CONFORMAL_INNER_FACES
    !------------------------------------------------------------------
    !---- This routine processes conformal inner faces
    !---- 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_ELEMENTS
    USE MESH_FIND_FACES
    USE MESH_TYPE_CONVERSIONS
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)             :: cellType, faceType
    
    LOGICAL(C_BOOL)                           :: isEqual
    
    INTEGER(C_INT)                            :: iFace, iCell, locFace, &
         nVerts, nLinVerts, nHoVerts, geoOrder, iCell2, locFace2,       &
         faceNum, foundFace
    INTEGER(C_INT)                            :: logInner   = 1
    INTEGER(C_INT)                            :: logUnknown = 0
    INTEGER(C_INT), DIMENSION(nvFaceMax)      :: faceVert

    TYPE(face),     DIMENSION(:), ALLOCATABLE :: locFaceArray
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Copy faces that were already read and allocate temp arrays
    !------------------------------------------------------------------
    ALLOCATE(locFaceArray(Mesh%nFac))
    locFaceArray(1:Mesh%nFac) = MeshFac(1:Mesh%nFac)
    !------------------------------------------------------------------
    !---- Adapt the faceBlockSize depending on the options
    !---- If addInnerFaces is TRUE, then we are about to add a large
    !---- amount of faces, so we better make it larger
    !------------------------------------------------------------------
    IF (addInnerFaces) faceBlockSize = 4*Mesh%nCellTet + 6*Mesh%nCellHex
    !------------------------------------------------------------------
    !---- Process inner faces
    !------------------------------------------------------------------
    iFace = 1
    DO WHILE(iFace <= nFaceToSort)
       locFace  = faceSort(5,iFace)
       iCell    = faceSort(6,iFace)
       cellType = MeshCell(iCell)%cellType
       CALL CELL_TYPE_TO_FACE_TYPE(cellType, faceType)
       CALL GET_FACE_VERTICES_FROM_CELL(iCell, locFace, faceVert)
       CALL FACE_TYPE_TO_VERTS(faceType, nVerts, nLinVerts, nHoVerts, geoOrder)
       isEqual = ALL(faceSort(1:4,iFace) == faceSort(1:4,iFace+1))
       !
       !---- Process faces
       SELECT CASE (isEqual)
       CASE(.TRUE.) ! Conformal inner face
          locFace2 = faceSort(5,iFace+1)
          iCell2   = faceSort(6,iFace+1)
          !
          MeshCell(iCell )%neighb(locFace)  = iCell2
          MeshCell(iCell2)%neighb(locFace2) = iCell
          !
          SELECT CASE (addInnerFaces)
          CASE(.TRUE.)
             CALL ADD_FACE(faceVert(1:nVerts), logInner, faceType, faceNum)
             MeshFac(faceNum)%neighb(1)   = iCell
             MeshFac(faceNum)%neighb(2)   = iCell2
             MeshFac(faceNum)%numLocal(1) = locFace
             MeshFac(faceNum)%numLocal(2) = locFace2
             MeshFac(faceNum)%numGeom(1)  = locFace
             MeshFac(faceNum)%numGeom(2)  = locFace2
          CASE(.FALSE.)
             CALL FIND_FACE_IN_LIST_FROM_VERTICES(faceVert(1:nVerts), &
                  locFaceArray, SIZE(locFaceArray), foundFace)
             !
             IF (foundFace /= 0) THEN
                MeshFac(foundFace)%neighb(1)   = iCell
                MeshFac(foundFace)%neighb(2)   = iCell2
                MeshFac(foundFace)%numLocal(1) = locFace
                MeshFac(foundFace)%numLocal(2) = locFace2
                MeshFac(foundFace)%numGeom(1)  = locFace
                MeshFac(foundFace)%numGeom(2)  = locFace2
             ENDIF
          END SELECT
          !
          iFace = iFace + 2
       CASE(.FALSE.) ! Either boundary or hybrid face
          CALL ADD_FACE(faceVert(1:nVerts), logUnknown, faceType, faceNum)
          !
          MeshFac(faceNum)%neighb(1)   = iCell
          MeshFac(faceNum)%numLocal(1) = locFace
          MeshFac(faceNum)%numGeom(1)  = locFace
          !
          iFace = iFace + 1
       END SELECT
    ENDDO
    !------------------------------------------------------------------
    IF (ALLOCATED(locFaceArray)) DEALLOCATE(locFaceArray)
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished processing conformal inner faces'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_CONFORMAL_INNER_FACES

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE REMOVE_FACE_DUPLICATES
    !------------------------------------------------------------------
    !---- This routine removes duplicate faces from the MeshFac array
    !---- 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                       :: isEqual
    
    INTEGER(C_INT)                        :: iFace, iCell, nVerts, &
         nLinVerts, nHoVerts, geoOrder, nFaces, nRemoved, log1,    &
         log2, faceNumKeep, faceNumRemove, nRemovedGlobal
    INTEGER(C_INT), DIMENSION(nvFaceMax)  :: faceVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Remove duplicates
    !------------------------------------------------------------------
    IF (ALLOCATED(removeFace)) DEALLOCATE(removeFace)
    ALLOCATE(removeFace(Mesh%nFac))
    removeFace  = .FALSE.
    !
    nFaceToSort = Mesh%nFac
    IF (ALLOCATED(faceSort)) DEALLOCATE(faceSort)
    ALLOCATE(faceSort(5,nFaceToSort+1)) ! This is not a mistake
    faceSort = 0
    !
    DO iFace=1,nFaceToSort
       nLinVerts             = MeshFac(iFace)%nLinVerts
       faceVert(:)           = 0
       faceVert(1:nLinVerts) = MeshFac(iFace)%vertex(1:nLinVerts)
       CALL INSERTION_SORT_INT(faceVert(1:4))
       faceSort(1:4,iFace)   = faceVert(1:4)
       faceSort(5  ,iFace)   = iFace
    ENDDO
    CALL SORT_ARRAY(5,nFaceToSort,faceSort)
    !
    nRemoved = 0
    iFace    = 1
    DO WHILE(iFace <= nFaceToSort)
       isEqual = ALL(faceSort(1:4,iFace) == faceSort(1:4,iFace+1))
       !
       SELECT CASE (isEqual)
       CASE(.TRUE.) ! If there is a duplicate, tag it
          faceNumKeep               = faceSort(5,iFace)
          faceNumRemove             = faceSort(5,iFace+1)
          removeFace(faceNumRemove) = .TRUE.
          nRemoved                  = nRemoved + 1
          CALL MERGE_FACES(MeshFac(faceNumKeep), MeshFac(faceNumRemove))
          iFace                     = iFace + 2
       CASE(.FALSE.)
          iFace                     = iFace + 1
       END SELECT
    ENDDO
    !
    CALL CLEAN_FACE_LIST(nRemoved)
    CALL GLOBAL_OP(nRemoved, MPI_SUM, nRemovedGlobal)
    nRemoved = nRemovedGlobal
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a,i0,a)') 'Finished removing ',nRemoved,' face duplicates'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REMOVE_FACE_DUPLICATES

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE CLEAN_FACE_LIST(nRemoved)
    !------------------------------------------------------------------
    !---- This routine removes the faces in MeshFac that are tagged
    !---- in the removeFace list
    !---- 
    !
    !---- IN
    !---- nRemoved : number of faces to remove
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),            INTENT(IN) :: nRemoved
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                        :: iFace

    TYPE(face), DIMENSION(:), ALLOCATABLE :: localMeshFac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (nRemoved == 0) RETURN
    !------------------------------------------------------------------
    !---- Clean list
    !------------------------------------------------------------------
    nFaceToSort = Mesh%nFac
    ALLOCATE(localMeshFac(Mesh%nFac - nRemoved))
    Mesh%nFac   = 0
    !
    DO iFace=1,nFaceToSort
       SELECT CASE(removeFace(iFace))
       CASE(.FALSE.) ! Face that must be kept
          Mesh%nFac                       = Mesh%nFac + 1
          localMeshFac(Mesh%nFac)         = MeshFac(iFace)
          localMeshFac(Mesh%nFac)%numFace = Mesh%nFac
       END SELECT
    ENDDO
    !
    IF (ALLOCATED(MeshFac))      DEALLOCATE(MeshFac)
    ALLOCATE(MeshFac(Mesh%nFac))
    MeshFac = localMeshFac
    IF (ALLOCATED(localMeshFac)) DEALLOCATE(localMeshFac)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CLEAN_FACE_LIST

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_HYBRID_FACES
    !------------------------------------------------------------------
    !---- This routine fills a temporary array used to find neighbors
    !---- for hybrid faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_FACES_LOGS
    USE PARALLEL_STRUCTURES
    USE MESH_TYPE_CONVERSIONS
    USE MESH_NORMALS_AND_MEASURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)              :: isHybridInnerFace, isHybridArtificialFace

    INTEGER(C_INT)               :: iFace, nLinVerts

    REAL(C_DOUBLE)               :: area, globalHybTol2D
    REAL(C_DOUBLE), DIMENSION(3) :: normal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Fill an array of faces to make things easier
    !------------------------------------------------------------------
    hybTol2D = 1.0e20
    nFacHyb  = 0
    !
    !---- Add all hybrid faces
    DO iFace=1,Mesh%nFac
       isHybridInnerFace                               = .FALSE.
       isHybridArtificialFace                          = .FALSE.
       IF (MeshFac(iFace)%log == 0) isHybridInnerFace  = .TRUE.
       IF (MeshFac(iFace)%log == 0) MeshFac(iFace)%log = 1001
       CALL IS_HYBRID_ARTIFICIAL_LOG(MeshFac(iFace)%log, isHybridArtificialFace)
       CALL IS_HYBRID_INNER_LOG     (MeshFac(iFace)%log, isHybridInnerFace)
       IF ((.NOT. isHybridInnerFace) .AND. (.NOT. isHybridArtificialFace)) CYCLE
       !
       CALL EXTEND_FACE_ARRAY(localHybFac, nFacHyb)
       localHybFac(nFacHyb)         = MeshFac(iFace)
       nLinVerts                    = localHybFac(nFacHyb)%nLinVerts
       SELECT CASE(MeshFac(iFace)%faceType)
       CASE('TRI1', 'TRI2')
          CALL COMPUTE_TRI_NORMAL_AND_MEASURE (localHybFac(nFacHyb)%vertex(1:nLinVerts), &
               normal,area)
       CASE('QUA1', 'QUA2')
          CALL COMPUTE_QUAD_NORMAL_AND_MEASURE(localHybFac(nFacHyb)%vertex(1:nLinVerts), &
               normal,area)
       END SELECT
       localHybFac(nFacHyb)%measure = area
       hybTol2D                     = MIN(hybTol2D, area)
    ENDDO
    !
    !---- Compute tolerance for tests on hybrid faces
    hybTol2D = hybTol2D*1.0e-5
    CALL GLOBAL_OP(hybTol2D, MPI_MIN, globalHybTol2D)
    hybTol2D = globalHybTol2D
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_HYBRID_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE REMOVE_LARGE_HYBRID_FACES
    !------------------------------------------------------------------
    !---- In this routine, we find the hybrid faces that match each other, 
    !---- identify the large and small ones, and drop the large ones
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_ELEMENTS
    USE MESH_FIND_POINT_IN_ELEMENTS
    USE MESH_ELEMENTS_CENTERS
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                :: isInFace

    INTEGER(C_INT)                 :: iFaceHyb1, iFaceHyb2, nLinVerts, &
         iVert, vertLoc, nRemoved, iFace, iFace1, iFace2, nRemovedGlobal

    REAL(C_DOUBLE)                 :: area1, area2
    REAL(C_DOUBLE), DIMENSION(3)   :: pCenter1, pCenter2
    REAL(C_DOUBLE), DIMENSION(3,4) :: pFac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (nFacHyb == 0) RETURN
    !------------------------------------------------------------------
    !---- Loop on hybrid faces and clean the largest ones
    !---- A test on the areas is made so iFacHyb1 is always the largest
    !---- face, and iFacHyb2 the smallest face
    !------------------------------------------------------------------
    IF (ALLOCATED(removeFace)) DEALLOCATE(removeFace)
    ALLOCATE(removeFace(Mesh%nFac))
    removeFace = .FALSE.
    !
    nRemoved   = 0
    DO iFaceHyb1=1,nFacHyb
       area1     = localHybFac(iFaceHyb1)%measure
       nLinVerts = localHybFac(iFaceHyb1)%nLinVerts
       iFace1    = localHybFac(iFaceHyb1)%numFace
       !
       !---- Retrieve vertices coordinates for largest face
       DO iVert=1,nLinVerts
          vertLoc       = localHybFac(iFaceHyb1)%vertex(iVert)
          pFac(:,iVert) = MeshVert(vertLoc)%coor(:)
       ENDDO
       !
       !---- Loop on hybrid faces
       DO iFaceHyb2=1,nFacHyb
          area2     = localHybFac(iFaceHyb2)%measure
          nLinVerts = localHybFac(iFaceHyb2)%nLinVerts
          iFace2    = localHybFac(iFaceHyb2)%numFace
          !
          !---- Cycle if iFacHyb1 was already tagged as a large hybrid face
          !           if iFacHyb2 was already tagged as a large hybrid face
          !           if iFacHyb1 is not the biggest face of the two
          !           if it is the same face
          IF (removeFace(iFace1))         CYCLE          
          IF (removeFace(iFace2))         CYCLE
          IF ((area1 - area2) < hybTol2D) CYCLE          
          IF (iFaceHyb1 == iFaceHyb2)     CYCLE
          !
          CALL COMPUTE_CENTER_OF_FACE(localHybFac(iFaceHyb2)%vertex(1:nLinVerts),pCenter2)
          CALL IS_POINT_IN_FACE(pCenter2, pFac(:,1:nLinVerts), isInFace)
          IF (.NOT. isInFace)             CYCLE
          !
          !---- We identified a large hybrid face, we tag it
          removeFace(iFace1) = .TRUE.
          nRemoved           = nRemoved + 1
       ENDDO
    ENDDO
    !
    CALL CLEAN_FACE_LIST(nRemoved)
    CALL GLOBAL_OP(nRemoved, MPI_SUM, nRemovedGlobal)
    nRemoved = nRemovedGlobal
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a,i0,a)') 'Finished removing ',nRemoved,' hybrid faces'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE REMOVE_LARGE_HYBRID_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_HYBRID_FACES
    !------------------------------------------------------------------
    !---- Process hybrid faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_ELEMENTS
    USE MESH_FACES_LOGS
    USE MESH_FIND_POINT_IN_ELEMENTS
    USE MESH_ELEMENTS_CENTERS
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                :: isInQuad, isHybridInner, isHybridArtificial

    INTEGER(C_INT)                 :: iHex, iCell, faceNum, iFaceHyb, &
         iGeoFace, iFace, locFace

    REAL(C_DOUBLE), DIMENSION(3)   :: pCenter
    REAL(C_DOUBLE), DIMENSION(3,4) :: pQuad

    TYPE(cell)                     :: cellCopy
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (nFacHyb == 0) RETURN
    !------------------------------------------------------------------
    !---- Loop on inner hybrid faces and add the conformal-side neighbor
    !------------------------------------------------------------------
    DO iFaceHyb=1,nFacHyb
       iFace = localHybFac(iFaceHyb)%numFace
       CALL IS_HYBRID_INNER_LOG(MeshFac(iFace)%log, isHybridInner)
       IF (isHybridInner) THEN
          iCell   = MeshFac(iFace)%neighb(1)
          locFace = MeshFac(iFace)%numLocal(1)
          MeshCell(iCell)%facNum(locFace) = iFace
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Loop on hexahedra and add hybrid neighbors
    !------------------------------------------------------------------
    DO iHex=1,Mesh%nCellHex
       iCell    = MeshCellHex(iHex)%cellNum
       cellCopy = MeshCell(iCell)
       !
       !---- Count the number of additional neighbors
       DO iGeoFace=1,cellCopy%nFaces
          IF (cellCopy%neighb(iGeoFace) /= 0) CYCLE
          CALL GET_FACE_COORDINATES_FROM_CELL(iCell, iGeoFace, pQuad)
          !
          DO iFace=1,Mesh%nFac
             CALL IS_HYBRID_INNER_LOG(MeshFac(iFace)%log, isHybridInner)
             CALL IS_HYBRID_ARTIFICIAL_LOG(MeshFac(iFace)%log, isHybridArtificial)
             IF ((.NOT. isHybridInner) .AND. (.NOT. isHybridArtificial)) CYCLE
             !
             CALL COMPUTE_CENTER_OF_FACE(iFace, pCenter)
             CALL IS_POINT_IN_QUAD(pCenter, pQuad, isInQuad)
             IF (isInQuad) CALL ADD_HYBRID_NEIGHBOR(MeshCell(iCell), iFace, iGeoFace, iCell)
          ENDDO
       ENDDO
    ENDDO    
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_HYBRID_FACES

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE ADD_HYBRID_NEIGHBOR(cellIn, facNum, geoFac, cellNum)
    !------------------------------------------------------------------
    !---- Add a hybrid neighbor to a cell
    !
    !---- IN
    !---- cellIn  : cell to which we want to add the neighbor
    !---- facNum  : index of the corresponding face
    !---- geoFac  : local index of the geometrical face holding the neighbor
    !---- cellNum : index of cellIn
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_FACES_LOGS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)    :: facNum, cellNum, geoFac
    
    TYPE(cell),     INTENT(INOUT) :: cellIn
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)               :: isHybridInner, isHybridArtificial
    
    INTEGER(C_INT)                :: iNeighb, iNeighbCell, iNeighbFac, &
         iFac, iZero, nTotNeighb, boundaryFaceNum

    TYPE(cell)                    :: locCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if there is an available slot in neighbors
    !------------------------------------------------------------------
    iZero = 0
    DO iNeighb=1,cellIn%nNeighb
       IF (cellIn%neighb(iNeighb) == 0) iZero = iZero + 1
    ENDDO
    !
    !---- If there is no slot available, dig a new one ! Haha !
    IF (iZero == 0) THEN
       !---- Create local copy of the cell
       locCell = cellIn
       !
       !---- Reallocate neighbors array
       IF (ALLOCATED(cellIn%neighb))     DEALLOCATE(cellIn%neighb)
       IF (ALLOCATED(cellIn%nNeighbFac)) DEALLOCATE(cellIn%nNeighbFac)
       IF (ALLOCATED(cellIn%normal))     DEALLOCATE(cellIn%normal)
       IF (ALLOCATED(cellIn%perm))       DEALLOCATE(cellIn%perm)
       IF (ALLOCATED(cellIn%facSurf))    DEALLOCATE(cellIn%facSurf)
       IF (ALLOCATED(cellIn%facNum))     DEALLOCATE(cellIn%facNum)
       !
       nTotNeighb = cellIn%nNeighb + 1
       ALLOCATE(cellIn%neighb    (  nTotNeighb))
       ALLOCATE(cellIn%nNeighbFac(  nTotNeighb))
       ALLOCATE(cellIn%normal    (3,nTotNeighb))
       ALLOCATE(cellIn%perm      (2,nTotNeighb))
       ALLOCATE(cellIn%facSurf   (  nTotNeighb))
       ALLOCATE(cellIn%facNum    (  nTotNeighb))
       !
       cellIn%nNeighb          =  nTotNeighb
       cellIn%neighb           =  0
       cellIn%nNeighbFac       =  0
       cellIn%normal           =  0
       cellIn%perm             =  0
       cellIn%facSurf          =  0
       cellIn%facNum           =  0       
       !
       !---- Copy non-zero elements
       DO iNeighb=1,nTotNeighb-1
          cellIn%neighb (  iNeighb)     =  locCell%neighb (  iNeighb)
          cellIn%normal (:,iNeighb)     =  locCell%normal (:,iNeighb)
          cellIn%perm   (:,iNeighb)     =  locCell%perm   (:,iNeighb)
          cellIn%facSurf(  iNeighb)     =  locCell%facSurf(  iNeighb)
          cellIn%facNum (  iNeighb)     =  locCell%facNum (  iNeighb)
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Add neighbor
    !---- There two possible cases here
    !---- (i)  Inner hybrid face
    !---- (ii) Artificial hybrid face, hybrid side
    !------------------------------------------------------------------
    CALL IS_HYBRID_INNER_LOG     (MeshFac(facNum)%log, isHybridInner)
    CALL IS_HYBRID_ARTIFICIAL_LOG(MeshFac(facNum)%log, isHybridArtificial)
    !
    IF (isHybridInner) THEN
       iNeighbCell = MeshFac(facNum)%neighb(1)
       iNeighbFac  = MeshFac(facNum)%numLocal(1)
       !
       !---- Look for the first vacant neighbor slot in cellIn
       DO iNeighb=1,cellIn%nNeighb
          IF (cellIn%neighb(iNeighb) == 0) THEN
             !
             !---- Complete local cell
             cellIn%neighb    (iNeighb)               = iNeighbCell
             cellIn%nNeighbFac(iNeighb)               = geoFac
             cellIn%facNum    (iNeighb)               = facNum
             cellIn%nFacNeighb(geoFac)                = cellIn%nFacNeighb(geoFac) + 1
             !
             !---- Complete neighbor cell
             MeshCell(iNeighbCell)%neighb(iNeighbFac) = cellNum
             MeshCell(iNeighbCell)%facNum(iNeighbFac) = facNum
             !
             !---- Complete hybrid face
             MeshFac(facNum)%neighb(2)                = cellNum
             MeshFac(facNum)%numLocal(2)              = iNeighb
             MeshFac(facNum)%numGeom(2)               = geoFac
             !
             !---- Stop looping
             RETURN
          ENDIF
       ENDDO
    ENDIF
    !
    IF (isHybridArtificial) THEN
       boundaryFaceNum = MeshFac(facNum)%boundaryFaceNum
       !
       !---- Look for the first vacant neighbor slot in cellIn
       DO iNeighb=1,cellIn%nNeighb
          IF (cellIn%neighb(iNeighb) == 0) THEN
             !
             !---- Complete local cell
             cellIn%neighb    (iNeighb)               = Mesh%nCell + boundaryFaceNum
             cellIn%nNeighbFac(iNeighb)               = geoFac
             cellIn%facNum    (iNeighb)               = facNum
             cellIn%nFacNeighb(geoFac)                = cellIn%nFacNeighb(geoFac) + 1
             !
             !---- Complete hybrid face
             MeshFac(facNum)%neighb(1)                = cellNum
             MeshFac(facNum)%neighb(2)                = Mesh%nCell + boundaryFaceNum
             MeshFac(facNum)%numLocal(1)              = iNeighb
             MeshFac(facNum)%numLocal(2)              = 1
             MeshFac(facNum)%numGeom(1)               = geoFac
             MeshFac(facNum)%numGeom(2)               = 1
             !
             !---- Stop looping
             RETURN
          ENDIF
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_HYBRID_NEIGHBOR

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE PROCESS_BOUNDARY_FACES
    !------------------------------------------------------------------
    !---- Process both conformal (physical and artificial) and hybrid
    !---- (artificial) boundary faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: isConformalBoundaryFace, isHybridArtificialFace

    INTEGER(C_INT)  :: iFaceB, iFace, iCell, locFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Process conformal boundary faces
    !------------------------------------------------------------------
    DO iFaceB=1,Mesh%nFacB
       iFace                           = MeshFacB(iFaceB)%faceNum
       !
       iCell                           = MeshFac(iFace)%neighb(1)
       IF (iCell == 0) CYCLE ! Artificial hybrid face on hybrid side
       locFace                         = MeshFac(iFace)%numLocal(1)
       MeshFac(iFace)%boundaryFaceNum  = iFaceB
       !
       MeshCell(iCell)%neighb(locFace) = Mesh%nCell + iFaceB
       MeshCell(iCell)%facNum(locFace) = iFace
       MeshFac(iFace)%neighb(2)        = Mesh%nCell + iFaceB
       MeshFac(iFace)%numLocal(2)      = 1
       MeshFac(iFace)%numGeom(2)       = 1
    ENDDO
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished processing boundary faces'
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_BOUNDARY_FACES  

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE ATTACH_FACES_TO_CELLS
    !------------------------------------------------------------------
    !---- Fill the facNum attribute in cells, using faces informations
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace, iCell, locFace, iNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Attach faces to cells
    !------------------------------------------------------------------
    DO iFace=1,Mesh%nFac
       !---- First side
       iCell   = MeshFac(iFace)%neighb(1)
       locFace = MeshFac(iFace)%numLocal(1)
       IF (iCell <= Mesh%nCell) THEN
          MeshCell(iCell)%facNum(locFace) = iFace
          IF (MeshCell(iCell)%neighb(locFace) == 0) &
               MeshCell(iCell)%neighb(locFace) = MeshFac(iFace)%neighb(2)
       ENDIF
       !
       !---- Second side
       iCell   = MeshFac(iFace)%neighb(2)
       locFace = MeshFac(iFace)%numLocal(2)
       IF (iCell <= Mesh%nCell) THEN
          MeshCell(iCell)%facNum(locFace) = iFace
          IF (MeshCell(iCell)%neighb(locFace) == 0) &
               MeshCell(iCell)%neighb(locFace) = MeshFac(iFace)%neighb(1)
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished attaching faces to cells'        
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ATTACH_FACES_TO_CELLS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_CELLS_FACES
    !------------------------------------------------------------------
    !---- This routine tests that the faces were properly attached to cells
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: localError, globalError

    INTEGER(C_INT)  :: iCell, iNeighb, nMissing, iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (Mesh%nCell == 0) RETURN
    !------------------------------------------------------------------
    !---- Test faces
    !------------------------------------------------------------------
    globalError = .FALSE.
    !
    DO iCell=1,Mesh%nCell
       localError = .FALSE.
       !
       DO iNeighb=1,MeshCell(iCell)%nNeighb
          iFace = MeshCell(iCell)%facNum(iNeighb)
          IF (iFace == 0) localError  = .TRUE.
          IF (iFace == 0) globalError = .TRUE.
       ENDDO
       !
       IF (localError) nMissing = nMissing + 1
    ENDDO

    !
    IF (globalError) THEN
       WRITE(6,'(a)')      'Error in MESH_NEIGHBORS_AND_FACES'
       WRITE(6,'(a,i0,a)') 'I found ',nMissing,' cells missing at least one face'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing cells faces'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TEST_CELLS_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_FACES_NEIGHBORS
    !------------------------------------------------------------------
    !---- This routine tests the existence of faces neighbors
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: localError, globalError

    INTEGER(C_INT)  :: iFac, iCell1, iCell2, nNoMatch
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (Mesh%nCell == 0) RETURN
    !------------------------------------------------------------------
    !---- Test neighbors
    !------------------------------------------------------------------
    nNoMatch   = 0
    localError = .FALSE.
    !    
    DO iFac=1,Mesh%nFac
       iCell1 = MeshFac(iFac)%neighb(1)
       iCell2 = MeshFac(iFac)%neighb(2)
       !
       IF ((iCell1 == 0) .OR. (iCell2 == 0)) THEN
          localError = .TRUE.
          nNoMatch   = nNoMatch + 1
       ENDIF
    ENDDO
    !
    IF (localError) THEN
       WRITE(6,'(a)')       'Error in MESH_NEIGHBORS_AND_FACES'
       WRITE(6,'(a,i0,a)')  'I found ',nNoMatch,' faces missing at least one neighbor'
       CALL TILT          
    ENDIF
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing faces neighbors'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TEST_FACES_NEIGHBORS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_CELLS_NEIGHBORS
    !------------------------------------------------------------------
    !---- This routine tests the existence of cells neighbors
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: localError

    INTEGER(C_INT)  :: iCell, iCellNeighb, nNoMatch, iNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (Mesh%nCell == 0) RETURN
    !------------------------------------------------------------------
    !---- Test neighbors
    !------------------------------------------------------------------
    nNoMatch = 0
    !
    DO iCell=1,Mesh%nCell
       localError = .FALSE.
       !
       DO iNeighb=1,MeshCell(iCell)%nNeighb
          iCellNeighb = MeshCell(iCell)%neighb(iNeighb)
          IF (iCellNeighb == 0) localError  = .TRUE.
       ENDDO
       !
       IF (localError) nNoMatch = nNoMatch + 1
    ENDDO
    !
    IF (localError) THEN
       WRITE(6,'(a)')       'Error in MESH_NEIGHBORS_AND_FACES'
       WRITE(6,'(a,i0,a)')  'I found ',nNoMatch,' cells missing at &
            least one neighbor'
       CALL TILT          
    ENDIF
    !
    localError = .FALSE.
    DO iCell=1,Mesh%nCell
       IF (SIZE(MeshCell(iCell)%facNum) /= MeshCell(iCell)%nNeighb) localError = .TRUE.
    ENDDO
    !
    IF (localError) THEN
       WRITE(6,'(a)')  'Error in MESH_NEIGHBORS_AND_FACES'
       WRITE(6,'(a)')  'I found incompatibilities between the number &
            of neighbors and the size of the neighbor arrays in some cells'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing cells neighbors'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TEST_CELLS_NEIGHBORS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_NEIGHBORS_COHERENCE
    !------------------------------------------------------------------
    !---- This routine tests the coherence of neighbors between
    !---- cell and face structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, iNeighb, nNeighb, numFace, facNeighbor1, &
         facNeighbor2, cellNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (Mesh%nCell == 0) RETURN
    !------------------------------------------------------------------
    !---- Test neighbors
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       !
       nNeighb = MeshCell(iCell)%nNeighb
       DO iNeighb=1,nNeighb
          numFace      = MeshCell(iCell)%facNum(iNeighb)
          facNeighbor1 = MeshFac(numFace)%neighb(1)
          facNeighbor2 = MeshFac(numFace)%neighb(2)
          !
          IF ((facNeighbor1 /= iCell) .AND. (facNeighbor2 /= iCell)) THEN
             WRITE(6,'(a)') 'Error in MESH_NEIGHBORS_AND_FACES'
             WRITE(6,'(a)') 'I found incoherences in the neighbors arrays'
             WRITE(6,'(a)') 'At least one cell does not appear in the &
                  neighbor array of one of its faces'
             CALL TILT
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing neighbors coherence'
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TEST_NEIGHBORS_COHERENCE

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_NEIGHBORS
