MODULE MESH_NORMALS_AND_MEASURES

  !====================================================================
  ! Routines related to mesh normals
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_DATA
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_MESH_NORMALS_AND_MEASURES
    !------------------------------------------------------------------
    !---- This routine builds:
    !---- - The normal  array for cells
    !---- - The normal  array for inner faces
    !---- - The normal  array for boundary faces
    !---- - The surface value for cells
    !---- - The measure value for cells
    !---- - The measure value for inner faces
    !---- - The measure value for boundary faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE MESH_ELEMENTS_CENTERS
    USE PARALLEL_STRUCTURES
    USE MATH_PHYS_CONSTANTS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iCell, iFac, nNeighb, iNeighb, &
         numLocal, facNum

    REAL(C_DOUBLE)               :: faceArea, cellArea, plusOrMinus
    REAL(C_DOUBLE), DIMENSION(3) :: faceCenter, cellCenter, vector, normal

    TYPE(face)                   :: locFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute cell measures
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       CALL COMPUTE_CELL_MEASURE(iCell)
    ENDDO
    !------------------------------------------------------------------
    !---- Compute normals and surfaces for all faces
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFac
       CALL COMPUTE_FACE_NORMAL_AND_MEASURE(iFac)
       !
       !---- Switch the orientation of the normal
       !---- As a convention, face normals are oriented from neighbor of lower
       !---- index of neighbor to higher index
       iNeighb = MIN(MeshFac(iFac)%neighb(1), MeshFac(iFac)%neighb(2))
       CALL COMPUTE_CENTER_OF_CELL(iNeighb,cellCenter)
       CALL COMPUTE_CENTER_OF_FACE(iFac,faceCenter)
       !
       !---- This vector is from the center of the cell of lower index to
       !---- the center of the face. This is the orientation we want
       vector      = cellCenter - faceCenter
       plusOrMinus = DOT_PRODUCT(vector,MeshFac(iFac)%normal)
       CALL NORMALIZE_SCALAR(plusOrMinus)
       MeshFac(iFac)%normal = plusOrMinus*MeshFac(iFac)%normal
    ENDDO
    !------------------------------------------------------------------
    !---- Complete normals and surface info on cells
    !------------------------------------------------------------------    
    DO iFac=1,Mesh%nFac
       normal = MeshFac(iFac)%normal
       CALL COMPUTE_CENTER_OF_FACE(iFac,faceCenter)
       !
       !---- Make normal inward to the cell
       DO iNeighb=1,2
          iCell                              = MeshFac(iFac)%neighb(iNeighb)
          IF (iCell > Mesh%nCell) CYCLE      ! This is a ghost cell
          numLocal                           = MeshFac(iFac)%numLocal(iNeighb)
          MeshCell(iCell)%facSurf(numLocal)  = MeshFac(iFac)%measure
          CALL COMPUTE_CENTER_OF_CELL(iCell,cellCenter)
          vector                             = cellCenter - faceCenter
          plusOrMinus                        = DOT_PRODUCT(vector,normal)
          CALL NORMALIZE_SCALAR(plusOrMinus)
          MeshCell(iCell)%normal(:,numLocal) = plusOrMinus*normal(:)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Compute cellArea
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       cellArea = 0.0
       nNeighb  = MeshCell(iCell)%nNeighb
       !
       DO iNeighb=1,nNeighb
          faceArea = MeshCell(iCell)%facSurf(iNeighb)
          cellArea = cellArea + faceArea
       ENDDO
       MeshCell(iCell)%surface = cellArea
    ENDDO
    !------------------------------------------------------------------
    !---- Compute Mesh measure
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       Mesh%measure = Mesh%measure + MeshCell(iCell)%measure
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    CALL TEST_BOUNDARY_NORMALS
    IF (addInnerFaces) CALL TEST_CELL_NORMALS
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh normals and measures'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_MESH_NORMALS_AND_MEASURES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_FACE_NORMAL_AND_MEASURE(iFac)
    !------------------------------------------------------------------
    !---- Compute face normal and measure
    !
    !---- IN
    !---- iFac : face index
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),           INTENT(IN) :: iFac
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)        :: faceType
    
    INTEGER(C_INT)                       :: nLinVerts
    INTEGER(C_INT), DIMENSION(nvFaceMax) :: faceVerts

    REAL(C_DOUBLE)                       :: measure
    REAL(C_DOUBLE),         DIMENSION(3) :: normal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine right routine call for normal computation
    !------------------------------------------------------------------
    faceType               = MeshFac(iFac)%faceType
    nLinVerts              = MeshFac(iFac)%nLinVerts
    faceVerts(1:nLinVerts) = MeshFac(iFac)%vertex(1:nLinVerts)
    !
    SELECT CASE(faceType)
    CASE("TRI1", "TRI2")
       CALL COMPUTE_TRI_NORMAL_AND_MEASURE (faceVerts(1:nLinVerts),normal,measure)
    CASE("QUA1", "QUA2")
       CALL COMPUTE_QUAD_NORMAL_AND_MEASURE(faceVerts(1:nLinVerts),normal,measure)
    END SELECT
    !
    !---- Fill MeshFac structure
    MeshFac(iFac)%normal  = normal
    MeshFac(iFac)%measure = measure
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_FACE_NORMAL_AND_MEASURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_TRI_NORMAL_AND_MEASURE(verts, normal, measure)
    !------------------------------------------------------------------
    !---- Computation of the unit normal vector of a tri face 
    !
    !---- IN
    !---- verts   : vertices numbers
    !
    !---- OUT
    !---- normal  : unitary normal vector
    !---- measure : area of the face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_VERTICES
    USE SPECIAL_FUNCTIONS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(3), INTENT(IN)  :: verts

    REAL(C_DOUBLE),               INTENT(OUT) :: measure
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: normal
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, ridge1, ridge2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute normal
    !------------------------------------------------------------------
    !---- Get coordinates of the vertices
    v1(:)     = MeshVert(verts(1))%coor(:)
    v2(:)     = MeshVert(verts(2))%coor(:)
    v3(:)     = MeshVert(verts(3))%coor(:)
    !
    !---- Build the ridges vectors
    ridge1(:) = v2(:) - v1(:)
    ridge2(:) = v3(:) - v1(:)
    !
    !---- Normal to face is colinear to ridge1 x ridge2
    CALL CROSS_PROD(ridge1,ridge2,normal)
    !
    !---- Compute normal and measure
    normal = Tref_surf*normal
    CALL NORMALIZE_VECTOR(normal,measure)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_TRI_NORMAL_AND_MEASURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_QUAD_NORMAL_AND_MEASURE(verts, normal, measure)
    !------------------------------------------------------------------
    !---- Computation of the unit normal vector of a quad face 
    !
    !---- IN
    !---- verts   : vertices numbers
    !
    !---- OUT
    !---- normal  : unitary normal vector
    !---- measure : area of the face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_VERTICES
    USE SPECIAL_FUNCTIONS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(4), INTENT(IN)  :: verts

    REAL(C_DOUBLE),               INTENT(OUT) :: measure
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: normal
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, v4, ridge1, ridge2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute normal
    !------------------------------------------------------------------
    !---- Get coordinates of the vertices
    v1(:)     = MeshVert(verts(1))%coor(:)
    v2(:)     = MeshVert(verts(2))%coor(:)
    v3(:)     = MeshVert(verts(3))%coor(:)
    v4(:)     = MeshVert(verts(4))%coor(:)    
    !
    !---- Build the ridges vectors
    ridge1(:) = v2(:) - v1(:)
    ridge2(:) = v4(:) - v1(:)
    !
    !---- Normal to face is colinear to ridge1 x ridge2
    CALL CROSS_PROD(ridge1,ridge2,normal)
    !
    !---- Compute normal and measure
    CALL NORMALIZE_VECTOR(normal,measure)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_QUAD_NORMAL_AND_MEASURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_CELL_MEASURE(iCell)
    !------------------------------------------------------------------
    !---- Compute measure of cell
    !
    !---- IN
    !---- iCell : cell index
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)    :: iCell
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR) :: cellType
    
    INTEGER(C_INT)                :: nLinVerts
    INTEGER(C_INT), DIMENSION(8)  :: cellVerts
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine right routine call for measure computation
    !------------------------------------------------------------------
    cellType               = MeshCell(iCell)%cellType
    nLinVerts              = MeshCell(iCell)%nLinVerts
    cellVerts(1:nLinVerts) = MeshCell(iCell)%vertex(1:nLinVerts)
    !
    SELECT CASE(cellType)
    CASE("TET1", "TET2")
       CALL COMPUTE_TET_MEASURE(iCell,cellVerts(1:nLinVerts))
    CASE("HEX1", "HEX2")
       CALL COMPUTE_HEX_MEASURE(iCell,cellVerts(1:nLinVerts))
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_CELL_MEASURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_TET_MEASURE(iCell, verts)
    !------------------------------------------------------------------
    !---- Compute measure of tetrahedral cell
    !
    !---- IN
    !---- iCell : cell index
    !---- verts : list of cell vertices
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_VERTICES
    USE SPECIAL_FUNCTIONS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN) :: iCell
    INTEGER(C_INT), DIMENSION(4), INTENT(IN) :: verts
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, v4, ridge1, ridge2, ridge3
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute cell measure
    !------------------------------------------------------------------
    !---- Get coordinates of the vertices
    v1(:)                   = MeshVert(verts(1))%coor(:)
    v2(:)                   = MeshVert(verts(2))%coor(:)
    v3(:)                   = MeshVert(verts(3))%coor(:)
    v4(:)                   = MeshVert(verts(4))%coor(:)
    !
    !---- Build the ridge vectors
    ridge1(:)               = v2(:) - v1(:)
    ridge2(:)               = v3(:) - v1(:)
    ridge3(:)               = v4(:) - v1(:)
    !
    !---- Compute the measure
    MeshCell(iCell)%measure = Tref_vol*MIXT_PROD(ridge1,ridge2,ridge3)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_TET_MEASURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_HEX_MEASURE(iCell, verts)
    !------------------------------------------------------------------
    !---- Compute measure of hexahedral cell
    !
    !---- IN
    !---- iCell : cell index
    !---- verts : list of cell vertices
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_VERTICES
    USE SPECIAL_FUNCTIONS
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN) :: iCell
    INTEGER(C_INT), DIMENSION(8), INTENT(IN) :: verts
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3) :: v1, v2, v3, v5, ridge1, ridge2, ridge3
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute cell measure
    !------------------------------------------------------------------
    !---- Get coordinates of the vertices
    v1(:)                   = MeshVert(verts(1))%coor(:)
    v2(:)                   = MeshVert(verts(2))%coor(:)
    v3(:)                   = MeshVert(verts(3))%coor(:)
    v5(:)                   = MeshVert(verts(5))%coor(:)
    !
    !---- Build the ridge vectors
    ridge1(:)               = v2(:) - v1(:)
    ridge2(:)               = v3(:) - v1(:)
    ridge3(:)               = v5(:) - v1(:)
    !
    !---- Compute the measure
    MeshCell(iCell)%measure = MIXT_PROD(ridge1,ridge2,ridge3)
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_HEX_MEASURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_CELL_NORMALS
    !------------------------------------------------------------------
    !---- This routine tests that the sum normal*area on each cell
    !---- is equal to zero
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iCell, iFac, nNeighb, facNum, cellNum, numLocal

    REAL(C_DOUBLE)               :: faceArea
    REAL(C_DOUBLE), DIMENSION(3) :: nVector, checkArea
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Loop over cells and check that faceArea*norm sums to zero
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       nNeighb      = MeshCell(iCell)%nNeighb
       checkArea(:) = 0.0
       !
       DO iFac=1,nNeighb
          faceArea     = MeshCell(iCell)%facSurf(iFac)
          nVector(:)   = MeshCell(iCell)%normal(:,iFac)
          checkArea(:) = checkArea(:) + faceArea*nVector(:)
       ENDDO
       !
       IF (NORM2(checkArea) > Mesh%tolerance2D) THEN
          WRITE(6,'(a)')    "Error in MESH_NORMALS"
          WRITE(6,'(a,i0)') "Sum of normals was not zero on cell ",iCell
          WRITE(6,'(a,a)' ) "Cell type is ", MeshCell(iCell)%cellType
          WRITE(6,'(2(a,e15.7))') "The computed sum ",NORM2(checkArea)," &
               was higher than tolerance ",Mesh%tolerance2D
          CALL TILT
       ENDIF
       !
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing cell normals'
    !------------------------------------------------------------------
  END SUBROUTINE TEST_CELL_NORMALS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_BOUNDARY_NORMALS
    !------------------------------------------------------------------
    !---- This routine tests that the sum normal*area on all boundary 
    !---- faces is equal to zero
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iFac, iFacB

    REAL(C_DOUBLE)               :: faceArea
    REAL(C_DOUBLE), DIMENSION(3) :: nVector, checkArea
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Loop over boundary faces and check that faceArea*norm sums to zero
    !------------------------------------------------------------------
    CALL CHECKSUM_NORMALS_ON_DERIVED_FACES(MeshFacB, Mesh%nFacB, checkArea)
    !
    IF (NORM2(checkArea) > Mesh%tolerance2D) THEN
       WRITE(6,'(a)') 'Error in MESH_NORMALS'
       WRITE(6,'(a)') 'Sum of normals on boundary faces was not zero'
       WRITE(6,'(2(a,e15.7))') 'The computed sum ',NORM2(checkArea),' &
            was higher than tolerance ',Mesh%tolerance2D
       WRITE(6,'(a)') 'You might be missing boundary faces or tags'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing boundary normals'
    !------------------------------------------------------------------
  END SUBROUTINE TEST_BOUNDARY_NORMALS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECKSUM_NORMALS_ON_DERIVED_FACES(dFaceList, sizeList, checkSum)
    !------------------------------------------------------------------
    !---- This routine tests that the sum normal*area on all faces
    !---- of a given derived face list
    !
    !---- IN
    !---- dFaceList : derived face list
    !---- sizeList  : size of derived face list
    !
    !---- OUT
    !---- checkSum  : total sum
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(IN)  :: sizeList
    
    REAL(C_DOUBLE),     DIMENSION(3), INTENT(OUT) :: checkSum
    
    TYPE(derived_face), DIMENSION(:), ALLOCATABLE :: dFaceList
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                                :: iFac, idFac

    REAL(C_DOUBLE)                                :: faceArea
    REAL(C_DOUBLE), DIMENSION(3)                  :: nVector, localCheckSum
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Loop over derived faces and check that faceArea*norm sums to zero
    !------------------------------------------------------------------
    localCheckSum(:) = 0.0
    !
    DO idFac=1,sizeList
       iFac             = dFaceList(idFac)%faceNum
       faceArea         = MeshFac(iFac)%measure
       nVector(:)       = MeshFac(iFac)%normal(:)
       localCheckSum(:) = localCheckSum(:) + faceArea*nVector(:)
    ENDDO
    !
    CALL GLOBAL_OP(localCheckSum, MPI_SUM, checkSum)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECKSUM_NORMALS_ON_DERIVED_FACES

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_NORMALS_AND_MEASURES
