MODULE MESH_PERIODIC_FACES

  !====================================================================
  ! Everything related to periodic faces
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_PERIODIC_FACES
    !------------------------------------------------------------------
    !---- Process periodic faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: success

    INTEGER(C_INT)  :: iNoMatch
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Match faces
    !------------------------------------------------------------------
    CALL MATCH_PERIODIC_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    CALL TEST_PERIODIC_FACES_MATCHING(success, iNoMatch)
    !
    IF (.NOT. success) THEN
       WRITE(6,'(a)')    'Error in PROCESS_PERIODIC_FACES'
       WRITE(6,'(i0,a)') iNomatch,' periodic faces could not find a match'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished periodic faces processing'
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_PERIODIC_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MATCH_PERIODIC_FACES
    !------------------------------------------------------------------
    !---- Match faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL) :: faceMatch
    INTEGER(C_INT)  :: iFaceper, iFacePer1, iFace1, iFacePer2, &
         iFace2, iComp, iDir
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Fill complementary structure for periodic faces
    !------------------------------------------------------------------
    CALL ALLOCATE_AND_INITIALIZE_COMPFACPER
    !
    !---- Look for matching faces
    DO iFacePer1=1,Mesh%nFacPer
       iFace1 = MeshFacPer(iFacePer1)%faceNum
       !
       !---- If the face was already processed, cycle
       IF (CompFacper(iFacePer1)%matchingFace /= 0) CYCLE
       !
       DO iFacePer2=1,Mesh%nFacPer
          iFace2 = MeshFacPer(iFacePer2)%faceNum
          !
          CALL DO_PERIODIC_FACES_MATCH(MeshFac(iFace1), MeshFac(iFace2), &
               MeshVert, faceMatch, iDir, iComp)
          !
          IF (faceMatch) THEN
             CompFacPer(iFacePer1)%matchingFace = iFace2
             CompFacPer(iFacePer2)%matchingFace = iFace1
             IF (iDir == 1) CompFacPer(iFacePer1)%direction(iComp) = 1.0
             IF (iDir == 1) CompFacPer(iFacePer2)%direction(iComp) =-1.0             
             IF (iDir == 2) CompFacPer(iFacePer1)%direction(iComp) =-1.0
             IF (iDir == 2) CompFacPer(iFacePer2)%direction(iComp) = 1.0
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MATCH_PERIODIC_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DO_PERIODIC_FACES_MATCH(face1, face2, vertList, facematch, &
       iDir, iComp)
    !------------------------------------------------------------------
    !---- Match faces
    !
    !---- IN
    !---- vertList : list of vertices
    !---- face*    : faces to test in faceList
    !
    !---- OUT
    !---- faceMatch : TRUE if faces match periodically
    !---- iDir      : direction in which faces match (+ or -)
    !---- iComp     : component in which faces match (1, 2 or 3)
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL),            INTENT(OUT) :: faceMatch

    INTEGER(C_INT),             INTENT(OUT) :: iDir, iComp
    
    TYPE(face),                 INTENT(IN)  :: face1, face2
    
    TYPE(vertex), DIMENSION(:), ALLOCATABLE :: vertList
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                          :: nLinVerts1, nLinVerts2, &
         iVert1, iVert2
    INTEGER(C_INT), DIMENSION(nvFaceMax)    :: vertFace1, vertFace2,   &
         matchingVerts

    REAL(C_DOUBLE), DIMENSION(3)            :: vector, vert1, vert2, dVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check for a match
    !------------------------------------------------------------------
    faceMatch               = .FALSE.
    nLinVerts1              = face1%nLinVerts
    vertFace1(1:nLinVerts1) = face1%vertex(1:nLinVerts1)
    nLinVerts2              = face2%nLinVerts
    vertFace2(1:nLinVerts2) = face2%vertex(1:nLinVerts2)
    !
    DO iComp=1,3
       DO iDir=1,2
          !---- Compute moving vector
          vector(:)                = 0.0
          vector(iComp)            = globalMesh%boxSize(iComp)
          IF (iDir == 2) vector(:) =-vector(:)
          matchingVerts(:)         = 0
          !
          !---- Loop over vertices in first face
          DO iVert1=1,nLinVerts1
             vert1(:) = vertList(vertFace1(iVert1))%coor(:)
             vert1(:) = vert1(:) + vector(:)
             !
             !---- Look for a match in second face
             DO iVert2=1,nLinVerts2
                vert2(:) = vertList(vertFace2(iVert2))%coor(:)
                dVert(:) = ABS(vert2(:) - vert1(:))
                IF (ALL (dVert(:) < globalMesh%tolerance1D)) &
                     matchingVerts(iVert1) = iVert2
             ENDDO
             !
             !---- Test if this was a matching face
             IF (ALL(matchingVerts(1:nLinVerts1) /= 0)) THEN
                faceMatch = .TRUE.
                RETURN
             ENDIF
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DO_PERIODIC_FACES_MATCH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_PERIODIC_FACES_MATCHING(success, iNoMatch)
    !------------------------------------------------------------------
    !---- Test matching of periodic faces
    !
    !---- IN
    !
    !---- OUT
    !---- success : TRUE if the test was successful
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)             :: success

    INTEGER(C_INT), INTENT(OUT) :: iNoMatch
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)              :: iFaceper, iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    success = .TRUE.
    !------------------------------------------------------------------
    !---- Test matching of periodic faces
    !------------------------------------------------------------------
    iNoMatch = 0
    DO iFacePer=1,Mesh%nFacPer
       IF (CompFacPer(iFacePer)%matchingFace == 0) iNoMatch = iNoMatch + 1
    ENDDO
    IF (iNoMatch /= 0) success = .FALSE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !    
    !------------------------------------------------------------------
  END SUBROUTINE TEST_PERIODIC_FACES_MATCHING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_COLORED_PERIODIC_MESH
    !------------------------------------------------------------------
    !---- Write a periodic mesh with colored faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MEDIT_MESH
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFac, iColor, iFacPer, iFacMatch
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write colored periodic faces
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFac
       MeshFac(iFac)%log = -1
    ENDDO
    !
    iColor = 0
    DO iFacPer=1,Mesh%nFacPer
       iFac = MeshFacPer(iFacPer)%faceNum
       IF (MeshFac(iFac)%log /= -1) CYCLE
       !
       iColor                 = iColor + 1
       MeshFac(iFac)%log      = iColor
       iFacMatch              = compFacper(iFacPer)%matchingFace
       MeshFac(iFacMatch)%log = iColor
    ENDDO
    !
    Mesh%name = TRIM(Mesh%name)//'_matched'
    CALL WRITE_MEDIT_MESH
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !    
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_COLORED_PERIODIC_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ALLOCATE_AND_INITIALIZE_COMPFACPER
    !------------------------------------------------------------------
    !---- Allocate and initialize complementary structure for
    !---- periodic faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFaceper
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate and initialize
    !------------------------------------------------------------------
    IF (ALLOCATED(CompFacPer)) DEALLOCATE(CompFacPer)
    ALLOCATE(CompFacPer(Mesh%nFacPer))
    DO iFacePer=1,Mesh%nFacPer
       CompFacPer(iFacePer)%matchingFace = 0
       CompFacPer(iFacePer)%direction(:) = 0.0
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_AND_INITIALIZE_COMPFACPER

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE MESH_PERIODIC_FACES
