MODULE MESH_PERMUTATIONS

  !====================================================================
  ! Routines related to face permutations
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_DATA
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_MESH_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine builds:
    !---- - The permutations between internal faces and adjacent cells
    !---- - The permutations between boundary faces and adjacent cell
    !---- - The permutations between cell faces and cell volume
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: idFace, iFace, iCell, nFaces, iNeighb, nNeighb
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Permutations
    !------------------------------------------------------------------
    CALL GENERAL_FACE_PERMUTATIONS
    CALL HYBRID_FACE_PERMUTATIONS
    CALL ARTIFICIAL_FACE_PERMUTATIONS
    CALL PERIODIC_FACE_PERMUTATIONS
    CALL CELL_PERMUTATIONS
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    CALL TEST_PERMUTATIONS
    !---- Test internal faces permutations
    DO idFace=1,Mesh%nFacI
       iFace = MeshFacI(idFace)%faceNum
       !
       IF ((MeshFac(iFace)%perm(1) == 0) .OR. &
           (MeshFac(iFace)%perm(2) == 0)) THEN
          WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
          WRITE(6,'(a,i0)') "A permutation was not found on an internal face"
          CALL TILT
       ENDIF
    ENDDO
    !
    !---- Test boundary faces permutations
    DO idFace=1,Mesh%nFacB
       iFace = MeshFacB(idFace)%faceNum
       !
       IF (MeshFac(iFace)%perm(1) == 0) THEN
          WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
          WRITE(6,'(a,i0)') "A permutation was not found on a boundary face"
          CALL TILT
       ENDIF
    ENDDO
    !
    !---- Test cell permutations
    DO iCell=1,Mesh%nCell
       nNeighb = MeshCell(iCell)%nNeighb
       !
       DO iNeighb=1,nNeighb
          IF ((MeshCell(iCell)%perm(1,iNeighb) == 0) .OR. &
              (MeshCell(iCell)%perm(2,iNeighb) == 0)) THEN
             WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
             WRITE(6,'(a,i0)') "A permutation was not found on a cell"
             CALL TILT
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh permutations'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_MESH_PERMUTATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GENERAL_FACE_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine builds permutations between faces
    !---- and their associated cells, in the general case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)        :: faceType

    INTEGER(C_INT)                       :: iFace, iCell, nFaceVerts, &
         nCellVerts, perm, iNeighb
    INTEGER(C_INT), DIMENSION(nvFaceMax) :: faceVerts
    INTEGER(C_INT), DIMENSION(nvCellMax) :: cellVerts
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Face permutations
    !------------------------------------------------------------------
    DO iFace=1,Mesh%nFac
       !---- Get face vertices
       nFaceVerts                       = MeshFac(iFace)%nLinVerts
       faceVerts(1:nFaceVerts)          = MeshFac(iFace)%vertex(1:nFaceVerts)
       !
       DO iNeighb=1,2
          !---- Get neighbor
          iCell                         = MeshFac(iFace)%neighb(iNeighb)
          !
          !---- If neighbor is a ghost cell, take other neighbor
          IF (iCell > Mesh%nCell) iCell = MeshFac(iFace)%neighb(MOD(iNeighb+1,2))
          !
          !---- Get neighbor vertices
          nCellVerts                    = MeshCell(iCell)%nLinVerts
          cellVerts(1:nCellVerts)       = MeshCell(iCell)%vertex(1:nCellVerts)
          !
          !---- Find permutation and store it
          faceType                      = MeshFac(iFace)%faceType
          CALL FIND_PERM_NUMBER(faceType,faceVerts(1:nFaceVerts),cellVerts(1:nCellVerts),perm)
          MeshFac(iFace)%perm(iNeighb)  = perm
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GENERAL_FACE_PERMUTATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HYBRID_FACE_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine corrects permutations on the hybrid faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE MESH_TYPE_CONVERSIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR) :: faceType, cellType

    INTEGER(C_INT) :: iFaceHyb, iFace, iCell, iNeighb, locFace, &
         nFaceVerts, nCellVerts, perm, nVerts, nHoVerts, geoOrder
    INTEGER(C_INT), DIMENSION(nvFaceMax) :: faceVerts
    INTEGER(C_INT), DIMENSION(nvCellMax) :: cellVerts
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Correct permutations
    !------------------------------------------------------------------
    DO iFaceHyb=1,Mesh%nFacHyb
       iFace                           = MeshFacHyb(iFaceHyb)%faceNum
       !
       !---- Loop on face neighbors
       DO iNeighb=1,2
          locFace                      = MeshFac(iFace)%numGeom(iNeighb)
          iCell                        = MeshFac(iFace)%neighb(iNeighb)
          !
          IF (iCell <= Mesh%nCell) THEN 
             nCellVerts                   = MeshCell(iCell)%nLinVerts
             cellVerts(1:nCellVerts)      = MeshCell(iCell)%vertex(1:nCellVerts)
             cellType                     = MeshCell(iCell)%cellType
             CALL CELL_TYPE_TO_FACE_TYPE(cellType,faceType)
             CALL FACE_TYPE_TO_VERTS(faceType, nVerts, nFaceVerts, nHoVerts, geoOrder)
             faceVerts(1:nFaceVerts)      = cellVerts(MeshCell(iCell)%permType(locFace,1:nFaceVerts))
             !
             CALL FIND_PERM_NUMBER(faceType,faceVerts(1:nFaceVerts),cellVerts(1:nCellVerts),perm)
             MeshFac(iFace)%perm(iNeighb) = perm
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE HYBRID_FACE_PERMUTATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PERIODIC_FACE_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine corrects permutations on the periodic faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE DOF_PERMUTATIONS
    USE REFERENCE_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)        :: faceType
    
    INTEGER(C_INT)                       :: idFace, iFace, iFaceMatch, &
         iCell, nLinVerts, iVert, iVertMatch, perm, iCellMatch, numLocal
    INTEGER(C_INT), DIMENSION(nvFaceMax) :: vertFace, vertFaceMatch, matchingVerts
    INTEGER(C_INT), DIMENSION(nvCellMax) :: vertCell

    REAL(C_DOUBLE), DIMENSION(3)         :: vert, vertMatch, domainSize
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Correct permutations
    !------------------------------------------------------------------
    domainSize = globalMesh%coorMax - globalMesh%coorMin
    !
    DO idFace=1,Mesh%nFacPer
       !---- If this is a periodic artificial face, cycle
       iFace = MeshFacPer(idFace)%faceNum
       IF (MeshFac(iFace)%isArtificial) CYCLE
       !
       !---- Otherwise, correct the permutation
       iFaceMatch = CompFacPer(idFace)%matchingFace
       nLinVerts  = MeshFac(iFaceMatch)%nLinVerts
       faceType   = MeshFac(iFaceMatch)%faceType
       iCell      = MeshFac(iFace)%neighb(1)
       numLocal   = MeshFac(iFace)%numLocal(1)
       !
       !---- For all vertices of iFac, find the matching vertices of iFacMatch
       vertFace     (1:nLinVerts) = MeshFac(iFace)%vertex(1:nLinVerts)
       vertFaceMatch(1:nLinVerts) = MeshFac(iFaceMatch)%vertex(1:nLinVerts)
       matchingVerts(:)           = 0
       !
       DO iVert=1,nLinVerts
          vert = MeshVert(vertFace(iVert))%coor
          vert = vert + CompFacper(idFace)%direction*domainSize
          !
          DO iVertMatch=1,nLinVerts
             vertMatch = MeshVert(vertFaceMatch(iVertMatch))%coor - vert
             !
             IF (ALL (ABS(vertMatch(1:nLinVerts)) < Mesh%tolerance1D)) &
                  matchingVerts(iVert) = iVertMatch
          ENDDO
       ENDDO
       !
       !---- Find permutation between local periodic face and
       !---- distant matching cell
       iCellMatch            = MeshFac(iFaceMatch)%neighb(1)
       vertCell(1:nLinVerts) = MeshCell(iCellMatch)%vertex(1:nLinVerts)
       vertFace(1:nLinVerts) = vertFaceMatch(matchingVerts(1:nLinVerts))
       CALL FIND_PERM_NUMBER(faceType, vertFace, vertCell, perm)
       !
       MeshFac(iFace)%perm(2)           = perm
       MeshCell(iCell)%perm(2,numLocal) = perm
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PERIODIC_FACE_PERMUTATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ARTIFICIAL_FACE_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine corrects permutations on the artificial faces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    USE DOF_PERMUTATIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace, iNeighb, numFace, iCell, numLocal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. parallelRun) RETURN
    !------------------------------------------------------------------
    !---- Correct permutations
    !------------------------------------------------------------------
    DO iNeighb=1,nbNeighb
       DO iFace=1,neighb(iNeighb)%nbFac
          numFace                          = neighb(iNeighb)%faces(iFace)
          MeshFac(numFace)%perm(2)         = neighb(iNeighb)%perm(iFace)
          iCell                            = MeshFac(numFace)%neighb(1)
          numLocal                         = MeshFac(numFace)%numLocal(1)
          MeshCell(iCell)%perm(2,numLocal) = neighb(iNeighb)%perm(iFace)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ARTIFICIAL_FACE_PERMUTATIONS  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine builds permutations on cells
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE DOF_PERMUTATIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace, iCell, iCell1, iCell2, nNeighb, iPerm1, iPerm2

    TYPE(face)     :: locFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Cell permutations
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       nNeighb = MeshCell(iCell)%nNeighb
       !
       DO iFace=1,nNeighb
          locFace = MeshFac(MeshCell(iCell)%facNum(iFace))
          !
          !---- Neighbors
          iCell1  = locFace%neighb(1)
          iPerm1  = locFace%perm(1)
          iCell2  = locFace%neighb(2)
          iPerm2  = locFace%perm(2)
          !
          !---- Fill permutations
          IF (iCell == iCell1) THEN
             MeshCell(iCell)%perm(1,iFace) = iPerm1
             MeshCell(iCell)%perm(2,iFace) = iPerm2
          ELSE
             MeshCell(iCell)%perm(1,iFace) = iPerm2
             MeshCell(iCell)%perm(2,iFace) = iPerm1
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_PERMUTATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TEST_PERMUTATIONS
    !------------------------------------------------------------------
    !---- This routine tests permutations
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: idFace, iFace, iCell, nFaces, iNeighb, nNeighb, &
         facePerm, cellPerm, numLocal, iSide
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Test permutations
    !------------------------------------------------------------------
    !---- Test internal faces permutations
    DO idFace=1,Mesh%nFacI
       iFace = MeshFacI(idFace)%faceNum
       !
       IF ((MeshFac(iFace)%perm(1) == 0) .OR. &
           (MeshFac(iFace)%perm(2) == 0)) THEN
          WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
          WRITE(6,'(a,i0)') "A permutation was not found on an internal face"
          CALL TILT
       ENDIF
    ENDDO
    !
    !---- Test boundary faces permutations
    DO idFace=1,Mesh%nFacB
       iFace = MeshFacB(idFace)%faceNum
       !
       IF (MeshFac(iFace)%perm(1) == 0) THEN
          WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
          WRITE(6,'(a,i0)') "A permutation was not found on a boundary face"
          CALL TILT
       ENDIF
    ENDDO
    !
    !---- Test cell permutations
    DO iCell=1,Mesh%nCell
       nNeighb = MeshCell(iCell)%nNeighb
       !
       DO iNeighb=1,nNeighb
          IF ((MeshCell(iCell)%perm(1,iNeighb) == 0) .OR. &
              (MeshCell(iCell)%perm(2,iNeighb) == 0)) THEN
             WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
             WRITE(6,'(a,i0)') "A permutation was not found on a cell"
             CALL TILT
          ENDIF
       ENDDO
    ENDDO
    !
    !---- Test coherence between cells and faces
    DO iCell=1,Mesh%nCell
       nNeighb = MeshCell(iCell)%nNeighb
       !
       DO iNeighb=1,nNeighb
          iFace    = MeshCell(iCell)%facNum(iNeighb)
          IF (iFace == 0) CYCLE
          IF (MeshFac(iFace)%neighb(1) == iCell) iSide = 1
          IF (MeshFac(iFace)%neighb(2) == iCell) iSide = 2
          facePerm = MeshFac(iFace)%perm(iSide)
          cellPerm = MeshCell(iCell)%perm(1,iNeighb)
          !
          IF (facePerm /= cellPerm) THEN
             WRITE(6,'(a)')    "Error in MESH_PERMUTATIONS"
             WRITE(6,'(a,i0)') "Cell and face permutations do not match"
             CALL TILT
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished testing mesh permutations'
    !------------------------------------------------------------------
  END SUBROUTINE TEST_PERMUTATIONS
  
  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_PERMUTATIONS
