MODULE MESH_STATS

  !====================================================================
  ! Module containing the subroutines related to mesh statistics
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_MESH_BOUNDING_BOX
    !------------------------------------------------------------------
    !---- Compute the bounding box of Mesh structure
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iVert, iCoor

    REAL(C_DOUBLE)               :: localValue
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute bounding box and mesh center
    !------------------------------------------------------------------
    Mesh%coorMin(:)    = 1e9
    Mesh%coorMax(:)    =-1e9
    Mesh%coorCenter(:) = 0.0
    !
    DO iVert=1,Mesh%nVert
       DO iCoor=1,3
          localValue             = MeshVert(iVert)%coor(iCoor)
          Mesh%coorMin(iCoor)    = MIN(Mesh%coorMin(iCoor), localValue) 
          Mesh%coorMax(iCoor)    = MAX(Mesh%coorMax(iCoor), localValue)
          !
          Mesh%coorCenter(iCoor) = Mesh%coorCenter(iCoor) + MeshVert(iVert)%coor(iCoor)
       ENDDO
    ENDDO
    !
    CALL GLOBAL_OP(Mesh%coorCenter(:), MPI_SUM, globalMesh%coorCenter(:))
    CALL GLOBAL_OP(Mesh%nVert,         MPI_SUM, globalMesh%nVert)
    CALL GLOBAL_OP(Mesh%coorMin,       MPI_MIN, globalMesh%coorMin)
    CALL GLOBAL_OP(Mesh%coorMax,       MPI_MAX, globalMesh%coorMax)    
    !
    Mesh%coorCenter(:)    = globalMesh%coorCenter(:)/globalMesh%nVert
    Mesh%boxSize(:)       = ABS(globalMesh%coorMax(:)-globalMesh%coorMin(:))
    globalMesh%boxSize(:) = Mesh%boxSize(:)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished computing mesh bounding box'
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_MESH_BOUNDING_BOX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_TFSF_BOUNDING_BOX
    !------------------------------------------------------------------
    !---- Compute the TFSF bounding box of Mesh structure
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: idFace, iFace, iVert, nVerts, iCoor, localVert

    REAL(C_DOUBLE) :: localValue
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute bounding box
    !------------------------------------------------------------------
    DO idFace=1,Mesh%nFacTfsf
       iFace = MeshFacTfsf(idFace)%faceNum
       nVerts = MeshFac(iFace)%nVerts
       DO iVert=1,nVerts
          localVert = MeshFac(iFace)%vertex(iVert)
          DO iCoor=1,3
             localValue          = MeshVert(localVert)%coor(iCoor)
             Mesh%tfsfMin(iCoor) = MIN(Mesh%tfsfMin(iCoor), localValue) 
             Mesh%tfsfMax(iCoor) = MAX(Mesh%tfsfMax(iCoor), localValue) 
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished computing TF/SF bounding box'
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_TFSF_BOUNDING_BOX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_MESH_HMIN_HMAX_HAVG
    !------------------------------------------------------------------
    !---- Compute minimal, maximal and average ridge size
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, iFac, nLinVerts, iVert, iVert1, iVert2, &
         locVert, nAvg

    REAL(C_DOUBLE) :: hMin, hMax, hAvg
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize min and max values
    !------------------------------------------------------------------
    hMin = 1.0e+20
    hMax =-1.0e+20
    hAvg = 0.0
    nAvg = 0
    !------------------------------------------------------------------
    !---- Compute hMin, hMax and hAvg
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       nLinVerts = MeshCell(iCell)%nLinVerts
       CALL UPDATE_HMIN_HMAX_HAVG(nLinVerts, iCell, hMin, hMax, hAvg, nAvg)
    ENDDO
    !
    Mesh%hMax =  hMax
    Mesh%hMin =  hMin
    Mesh%hAvg =  hAvg/nAvg
    !
    !---- Modifying these values may affect code behavior seriously
    Mesh%tolerance1D = (hmin**1)*1.0e-7
    Mesh%tolerance2D = (hMin**2)*1.0e-3
    Mesh%tolerance3D = (hMin**3)*1.0e-2
    !
    CALL GLOBAL_OP(Mesh%tolerance1D, MPI_MIN, globalMesh%tolerance1D)
    CALL GLOBAL_OP(Mesh%tolerance2D, MPI_MIN, globalMesh%tolerance2D)
    CALL GLOBAL_OP(Mesh%tolerance3D, MPI_MIN, globalMesh%tolerance3D)
    !
    Mesh%tolerance1D = globalMesh%tolerance1D
    Mesh%tolerance2D = globalMesh%tolerance2D
    Mesh%tolerance3D = globalMesh%tolerance3D
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished computing mesh hMin/hMax'
    !------------------------------------------------------------------    
  END SUBROUTINE COMPUTE_MESH_HMIN_HMAX_HAVG

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE UPDATE_HMIN_HMAX_HAVG(nVerts, iCell, hmin, hmax, havg, navg)
    !------------------------------------------------------------------
    !---- Update the intermediate values to compute hmin, hmax and havg
    !
    !---- IN
    !---- nVerts : number of vertices in the current cell
    !---- iCell  : current cell
    !---- hmin   : temporary minimal ridge size
    !---- hmax   : temporary maximal ridge size
    !---- havg   : temporary average ridge size
    !---- navg   : number of evaluated ridges for average size
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(INOUT) :: nVerts, iCell
    INTEGER(C_INT), INTENT(INOUT) :: navg

    REAL(C_DOUBLE), INTENT(INOUT) :: hmin, hmax, havg
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR) :: cellType
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Select right routine call depending on type
    !------------------------------------------------------------------
    cellType = MeshCell(iCell)%cellType
    SELECT CASE(cellType)
    CASE('HEX','HEX1','HEX2')
       CALL UPDATE_HMIN_HMAX_HAVG_HEX(nVerts, iCell, hmin, hmax, havg, navg)
    CASE('TET','TET1','TET2','TET3','TET4')
       CALL UPDATE_HMIN_HMAX_HAVG_TET(nVerts, iCell, hmin, hmax, havg, navg)
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
  END SUBROUTINE UPDATE_HMIN_HMAX_HAVG

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE UPDATE_HMIN_HMAX_HAVG_HEX(nVerts, iCell, hmin, hmax, havg, navg)
    !------------------------------------------------------------------
    !---- Update the intermediate values to compute hmin, hmax and havg
    !---- on an hexahedron
    !
    !---- IN
    !---- nVerts : number of vertices in the current cell
    !---- iCell  : current cell
    !---- hmin   : temporary minimal ridge size
    !---- hmax   : temporary maximal ridge size
    !---- havg   : temporary average ridge size
    !---- navg   : number of evaluated ridges for average size
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN)    :: nVerts, iCell
    INTEGER(C_INT),   INTENT(INOUT) :: navg

    REAL(C_DOUBLE),   INTENT(INOUT) :: hmin, hmax, havg
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                  :: iVert, iVert1, iVert2, locVert

    REAL(C_DOUBLE),  DIMENSION(3)   :: distance
    REAL(C_DOUBLE),  DIMENSION(3,8) :: vertices  
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Update hmin and hmax
    !------------------------------------------------------------------
    !---- Fill local coordinates array
    DO iVert=1,nVerts
       locVert           = MeshCell(iCell)%vertex(iVert)
       vertices(:,iVert) = MeshVert(locVert)%coor(:)
    ENDDO
    !
    !---- Update values : we only compute the distance from vertex 1
    !---- to vertices 2, 4 and 5, since hexahedra are supposed to be
    !---- parallelepipedic
    CALL COMPUTE_R3_DISTANCE(vertices(:,1),vertices(:,2),distance(1))
    CALL COMPUTE_R3_DISTANCE(vertices(:,1),vertices(:,4),distance(2))
    CALL COMPUTE_R3_DISTANCE(vertices(:,1),vertices(:,5),distance(3))
    !
    DO iVert=1,3
       IF (distance(iVert) == 0.0) THEN
          WRITE(6,'(a)') 'Problem in COMPUTE_MESH_HMIN_HMAX_HAVG'
          WRITE(6,'(a)') 'A distance between two vertices of an hex was zero'
          WRITE(6,'(a)') 'I will skip it, but you should look into it'
          CYCLE
       ENDIF
       !
       hmin = MIN(hmin, distance(iVert))
       hmax = MAX(hmax, distance(iVert))
       havg = havg + distance(iVert)
       navg = navg + 1
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
  END SUBROUTINE UPDATE_HMIN_HMAX_HAVG_HEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE UPDATE_HMIN_HMAX_HAVG_TET(nVerts, iCell, hmin, hmax, havg, navg)
    !------------------------------------------------------------------
    !---- Update the intermediate values to compute hmin, hmax and havg
    !---- on a tetrahedron
    !
    !---- IN
    !---- nVerts : number of vertices in the current cell
    !---- iCell  : current cell
    !---- hmin   : temporary minimal ridge size
    !---- hmax   : temporary maximal ridge size
    !---- havg   : temporary average ridge size
    !---- navg   : number of evaluated ridges for average size
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN)    :: nVerts, iCell
    INTEGER(C_INT),   INTENT(INOUT) :: navg

    REAL(C_DOUBLE),   INTENT(INOUT) :: hmin, hmax, havg
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                  :: iVert, locVert

    REAL(C_DOUBLE),  DIMENSION(6)   :: distance
    REAL(C_DOUBLE),  DIMENSION(3,4) :: vertices
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Update hmin and hmax
    !------------------------------------------------------------------
    !---- Fill local coordinates array
    DO iVert=1,nVerts
       locVert           = MeshCell(iCell)%vertex(iVert)
       vertices(:,iVert) = MeshVert(locVert)%coor(:)
    ENDDO
    !
    !---- Update values
    CALL COMPUTE_R3_DISTANCE(vertices(:,1),vertices(:,2),distance(1))
    CALL COMPUTE_R3_DISTANCE(vertices(:,1),vertices(:,3),distance(2))
    CALL COMPUTE_R3_DISTANCE(vertices(:,1),vertices(:,4),distance(3))
    CALL COMPUTE_R3_DISTANCE(vertices(:,2),vertices(:,3),distance(4))
    CALL COMPUTE_R3_DISTANCE(vertices(:,2),vertices(:,4),distance(5))
    CALL COMPUTE_R3_DISTANCE(vertices(:,3),vertices(:,4),distance(6))
    !
    DO iVert=1,6
       IF (distance(iVert) == 0.0) THEN
          WRITE(6,'(a)') 'Problem in COMPUTE_MESH_HMIN_HMAX_HAVG'
          WRITE(6,'(a)') 'A distance between two vertices of a tet was zero'
          WRITE(6,'(a)') 'I will skip it, but you should look into it'
          CYCLE
       ENDIF
       !
       hmin = MIN(hmin, distance(iVert))
       hmax = MAX(hmax, distance(iVert))
       havg = havg + distance(iVert)
       navg = navg + 1
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
  END SUBROUTINE UPDATE_HMIN_HMAX_HAVG_TET

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_GLOBAL_MESH
    !------------------------------------------------------------------
    !---- Build global mesh for statistics printing in the parallel case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Process parallel case
    !------------------------------------------------------------------
    CALL GLOBAL_OP(Mesh%coorMin,       MPI_MIN, globalMesh%coorMin)
    CALL GLOBAL_OP(Mesh%coorMax,       MPI_MAX, globalMesh%coorMax)
    CALL GLOBAL_OP(Mesh%hMin,          MPI_MIN, globalMesh%hMin)
    CALL GLOBAL_OP(Mesh%hMax,          MPI_MAX, globalMesh%hMax)
    CALL GLOBAL_OP(Mesh%hAvg,          MPI_SUM, globalMesh%hAvg)
    globalMesh%hAvg                    = globalMesh%hAvg/nbProc
    CALL GLOBAL_OP(Mesh%surfMin,       MPI_MIN, globalMesh%surfMin)
    CALL GLOBAL_OP(Mesh%volMin,        MPI_MIN, globalMesh%volMin)    
    CALL GLOBAL_OP(Mesh%measure,       MPI_SUM, globalMesh%measure)
    CALL GLOBAL_OP(Mesh%coorMax,       MPI_MAX, globalMesh%coorMax)
    CALL GLOBAL_OP(Mesh%coorMin,       MPI_MIN, globalMesh%coorMin)
    CALL GLOBAL_OP(Mesh%tfsfMax,       MPI_MAX, globalMesh%tfsfMax)
    CALL GLOBAL_OP(Mesh%tfsfMin,       MPI_MIN, globalMesh%tfsfMin)
    !------------------------------------------------------------------
    !---- Cells data
    !------------------------------------------------------------------
    CALL GLOBAL_OP(Mesh%nVert,         MPI_SUM, globalMesh%nVert)
    CALL GLOBAL_OP(Mesh%nCell,         MPI_SUM, globalMesh%nCell)
    CALL GLOBAL_OP(Mesh%nCellTet,      MPI_SUM, globalMesh%nCellTet)
    CALL GLOBAL_OP(Mesh%nCellHex,      MPI_SUM, globalMesh%nCellHex)
    CALL GLOBAL_OP(Mesh%nCellLin,      MPI_SUM, globalMesh%nCellLin)
    CALL GLOBAL_OP(Mesh%nCellQuadr,    MPI_SUM, globalMesh%nCellQuadr)
    CALL GLOBAL_OP(Mesh%nCellTetLin,   MPI_SUM, globalMesh%nCellTetLin)
    CALL GLOBAL_OP(Mesh%nCellTetQuadr, MPI_SUM, globalMesh%nCellTetQuadr)
    CALL GLOBAL_OP(Mesh%nCellHexLin,   MPI_SUM, globalMesh%nCellHexLin)
    CALL GLOBAL_OP(Mesh%nCellHexQuadr, MPI_SUM, globalMesh%nCellHexQuadr)
    CALL GLOBAL_OP(Mesh%nCellTfZone,   MPI_SUM, globalMesh%nCellTfZone)
    CALL GLOBAL_OP(Mesh%nCellSfZone,   MPI_SUM, globalMesh%nCellSfZone)
    !------------------------------------------------------------------
    !---- Faces data
    !------------------------------------------------------------------
    CALL GLOBAL_OP(Mesh%nFac,          MPI_SUM, globalMesh%nFac)
    CALL GLOBAL_OP(Mesh%nFacTri,       MPI_SUM, globalMesh%nFacTri)
    CALL GLOBAL_OP(Mesh%nFacQuad,      MPI_SUM, globalMesh%nFacQuad)
    CALL GLOBAL_OP(Mesh%nFacB,         MPI_SUM, globalMesh%nFacB)
    CALL GLOBAL_OP(Mesh%nFacPhy,       MPI_SUM, globalMesh%nFacPhy)
    CALL GLOBAL_OP(Mesh%nFacAbs,       MPI_SUM, globalMesh%nFacAbs)
    CALL GLOBAL_OP(Mesh%nFacPec,       MPI_SUM, globalMesh%nFacPec)
    CALL GLOBAL_OP(Mesh%nFacPmc,       MPI_SUM, globalMesh%nFacPmc)
    CALL GLOBAL_OP(Mesh%nFacPer,       MPI_SUM, globalMesh%nFacPer)
    CALL GLOBAL_OP(Mesh%nFacArt,       MPI_SUM, globalMesh%nFacArt)
    CALL GLOBAL_OP(Mesh%nFacI,         MPI_SUM, globalMesh%nFacI)
    CALL GLOBAL_OP(Mesh%nFacTfsf,      MPI_SUM, globalMesh%nFacTfsf)
    CALL GLOBAL_OP(Mesh%nFacPoy,       MPI_SUM, globalMesh%nFacPoy)
    CALL GLOBAL_OP(Mesh%nFacLin,       MPI_SUM, globalMesh%nFacLin)
    CALL GLOBAL_OP(Mesh%nFacQuadr,     MPI_SUM, globalMesh%nFacQuadr)
    CALL GLOBAL_OP(Mesh%nFacTriLin,    MPI_SUM, globalMesh%nFacTriLin)
    CALL GLOBAL_OP(Mesh%nFacTriQuadr,  MPI_SUM, globalMesh%nFacTriQuadr)
    CALL GLOBAL_OP(Mesh%nFacQuadLin,   MPI_SUM, globalMesh%nFacQuadLin)
    CALL GLOBAL_OP(Mesh%nFacQuadQuadr, MPI_SUM, globalMesh%nFacQuadQuadr)
    CALL GLOBAL_OP(Mesh%nFacInter,     MPI_SUM, globalMesh%nFacInter)
    CALL GLOBAL_OP(Mesh%nFacHyb,       MPI_SUM, globalMesh%nFacHyb)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished global mesh generation'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_GLOBAL_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_MESH_MINIMAL_SURFACE_AND_VOLUME
    !------------------------------------------------------------------
    !---- Compute minimal values of face area and cell volume
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, iFac

    REAL(C_DOUBLE) :: surfMin, volMin
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize min and max values
    !------------------------------------------------------------------
    surfMin = 1.0e+20
    volMin  = 1.0e+20
    !------------------------------------------------------------------
    !---- Compute surfMin
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFac
       surfMin = MIN(surfMin, MeshFac(iFac)%measure)
    ENDDO
    Mesh%surfMin = surfMin
    !------------------------------------------------------------------
    !---- Compute volMin
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       volMin = MIN(volMin, MeshCell(iCell)%measure)
    ENDDO
    Mesh%volMin = volMin
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished computing minimal surface and volume'
    !------------------------------------------------------------------    
  END SUBROUTINE COMPUTE_MESH_MINIMAL_SURFACE_AND_VOLUME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_MESH_STATS(locMesh)
    !------------------------------------------------------------------
    !---- Display some statistics about the mesh
    !
    !---- IN
    !---- locMesh : mesh of which statistics will be printed
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Module
    !------------------------------------------------------------------
    USE MESH_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    type(mesh_structure), INTENT(IN) :: locMesh
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. (idWrite)) RETURN
    !------------------------------------------------------------------
    !---- Print statistics
    !------------------------------------------------------------------
    WRITE(6,10)
    WRITE(6,'(a)')       'Mesh characteristics               '
    WRITE(6,10)
    WRITE(6,'(a,i9)')    'VERTICES                         : ',locMesh%nVert
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a,i9)')    'FACES                            : ',locMesh%nFac
    WRITE(6,'(a,i9)')    '                Triangular faces : ',locMesh%nFacTri
    WRITE(6,'(a,i9)')    '              Quadrangular faces : ',locMesh%nFacQuad
    WRITE(6,'(a,i9)')    '                    Linear faces : ',locMesh%nFacLin
    WRITE(6,'(a,i9)')    '                 Quadratic faces : ',locMesh%nFacQuadr
    WRITE(6,'(a,i9)')    '         Triangular linear faces : ',locMesh%nFacTriLin
    WRITE(6,'(a,i9)')    '      Triangular quadratic faces : ',locMesh%nFacTriQuadr
    WRITE(6,'(a,i9)')    '       Quadrangular linear faces : ',locMesh%nFacQuadLin
    WRITE(6,'(a,i9)')    '    Quadrangular quadratic faces : ',locMesh%nFacQuadQuadr
    WRITE(6,'(a,i9)')    '                    Hybrid faces : ',locMesh%nFacHyb    
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a,i9)')    '                  Boundary faces : ',locMesh%nFacB
    WRITE(6,'(a,i9)')    '         Physical boundary faces : ',locMesh%nFacPhy
    WRITE(6,'(a,i9)')    '                       PEC faces : ',locMesh%nFacPec
    WRITE(6,'(a,i9)')    '                       ABC faces : ',locMesh%nFacAbs
    WRITE(6,'(a,i9)')    '                       PBC faces : ',locMesh%nFacPer
    WRITE(6,'(a,i9)')    '                       PMC faces : ',locMesh%nFacPmc
    WRITE(6,'(a,i9)')    '                Artificial faces : ',locMesh%nFacArt
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a,i9)')    '                     TF/SF faces : ',locMesh%nFacTfsf
    WRITE(6,'(a,i9)')    '                  Poynting faces : ',locMesh%nFacPoy
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a,i9)')    '                  Internal faces : ',locMesh%nFacI
    WRITE(6,'(a,i9)')    '                 Interface faces : ',locMesh%nFacInter
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a,i9)')    'CELLS                            : ',locMesh%nCell
    WRITE(6,'(a,i9)')    '               Tetrahedral cells : ',locMesh%nCellTet
    WRITE(6,'(a,i9)')    '                Hexahedral cells : ',locMesh%nCellHex
    WRITE(6,'(a,i9)')    '                    Linear cells : ',locMesh%nCellLin
    WRITE(6,'(a,i9)')    '                 Quadratic cells : ',locMesh%nCellQuadr
    WRITE(6,'(a,i9)')    '        Tetrahedral linear cells : ',locMesh%nCellTetLin
    WRITE(6,'(a,i9)')    '     Tetrahedral quadratic cells : ',locMesh%nCellTetQuadr
    WRITE(6,'(a,i9)')    '         Hexahedral linear cells : ',locMesh%nCellHexLin
    WRITE(6,'(a,i9)')    '      Hexahedral quadratic cells : ',locMesh%nCellHexQuadr
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a,i9)')    '       # of cells in the TF zone : ',locMesh%nCellTfZone
    WRITE(6,'(a,i9)')    '       # of cells in the SF zone : ',locMesh%nCellSfZone
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a)')       'SIZES                              '
    WRITE(6,'(a,e15.7)') '                            hmin = ',locMesh%hmin
    WRITE(6,'(a,e15.7)') '                            hmax = ',locMesh%hmax
    WRITE(6,'(a,e15.7)') '                            havg = ',locMesh%havg    
    WRITE(6,'(a,f15.7)') '                       hmax/hmin = ',locMesh%hmax/locMesh%hmin
    WRITE(6,'(a,e15.7)') '                         surfMin = ',locMesh%surfMin
    WRITE(6,'(a,e15.7)') '                          volMin = ',locMesh%volMin
    WRITE(6,'(a,e15.7)') '                    Total volume = ',locMesh%measure
    WRITE(6,'(a)')       '                                   '
    WRITE(6,'(a)')       'BOUNDING BOX                       '  
    WRITE(6,'(a,e15.7)') '                            Xmax = ',locMesh%coorMax(1)
    WRITE(6,'(a,e15.7)') '                            Ymax = ',locMesh%coorMax(2)
    WRITE(6,'(a,e15.7)') '                            Zmax = ',locMesh%coorMax(3)
    WRITE(6,'(a,e15.7)') '                            Xmin = ',locMesh%coorMin(1)
    WRITE(6,'(a,e15.7)') '                            Ymin = ',locMesh%coorMin(2)
    WRITE(6,'(a,e15.7)') '                            Zmin = ',locMesh%coorMin(3)
    WRITE(6,10)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_MESH_STATS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE WRITE_MESH_SIZE_CONFIG_FILE(fileName)
    !------------------------------------------------------------------
    !---- This routine writes a viewConfig file to produce a solution
    !---- file with the mesh size in it
    !
    !---- IN
    !---- fileName : name of config file
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: fileName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                   :: nUnit
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Write solution
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(fileName))
    !
    WRITE(nUnit, '(a)')   '#MeshFile'
    WRITE(nUnit, '(a,a)') TRIM(Mesh%name)//'.rawMesh'
    WRITE(nUnit, '(a)')   ' '
    WRITE(nUnit, '(a)')   '#DataFile'
    WRITE(nUnit, '(a,a)') TRIM(Mesh%name)//'_0.rawSol'
    WRITE(nUnit, '(a)')   ' '
    WRITE(nUnit, '(a)')   '#OutputFormat'
    WRITE(nUnit, '(a)')   'MEDIT'
    WRITE(nUnit, '(a)')   ' '
    WRITE(nUnit, '(a)')   '#ViewType'
    WRITE(nUnit, '(a)')   'DISC'
    WRITE(nUnit, '(a)')   'DISC'    
    WRITE(nUnit, '(a)')   ' '
    WRITE(nUnit, '(a)')   '#ViewOrders'
    WRITE(nUnit, '(a)')   '1'
    WRITE(nUnit, '(a)')   '1'
    WRITE(nUnit, '(a)')   ' '
    WRITE(nUnit, '(a)')   '#ViewPlanes'
    WRITE(nUnit, '(a)')   '0'
    !
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE WRITE_MESH_SIZE_CONFIG_FILE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_STATS
