MODULE MESH_STRUCTURES

  !====================================================================
  ! High-level routines for mesh processing
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_DATA
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Procedure pointer to read/write mesh and solution subroutines
  !--------------------------------------------------------------------
  PROCEDURE(), POINTER :: READ_MESH      => NULL()
  PROCEDURE(), POINTER :: WRITE_MESH     => NULL()
  PROCEDURE(), POINTER :: READ_SOLUTION  => NULL()
  PROCEDURE(), POINTER :: WRITE_SOLUTION => NULL()
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_MESH
    !------------------------------------------------------------------
    !---- Read mesh and build required data structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_STATS
    USE MESH_AXES
    USE MESH_NEIGHBORS
    USE MESH_PERMUTATIONS
    USE MESH_INTERFACES
    USE MESH_NORMALS_AND_MEASURES
    USE MESH_PERIODIC_FACES
    USE MESH_TFSF_FACES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Perform all pre-processings
    !------------------------------------------------------------------
    CALL INITIALIZE_MESH_STRUCTURE(Mesh)
    CALL CHECK_MPI_PARALLEL_MESH
    CALL CHECK_IN_MESH_EXTENSION(Mesh%name)
    CALL READ_MESH
    CALL BUILD_DERIVED_CELLS
    CALL BUILD_DERIVED_FACES
    CALL COMPUTE_MESH_HMIN_HMAX_HAVG
    CALL COMPUTE_MESH_BOUNDING_BOX
    CALL BUILD_MESH_NEIGHBORS_AND_FACES
    CALL BUILD_GLOBAL_MESH    
    CALL BUILD_MESH_NORMALS_AND_MEASURES
    CALL COMPUTE_MESH_MINIMAL_SURFACE_AND_VOLUME
    CALL READ_PARALLEL_INFORMATION
    IF (addInnerFaces) CALL BUILD_FACE_AXES
    CALL PROCESS_PERIODIC_FACES
    IF (addInnerFaces) CALL BUILD_MESH_PERMUTATIONS
    CALL PROCESS_TFSF_ZONES_AFTER_READING
    CALL COMPUTE_TFSF_BOUNDING_BOX
    CALL IS_TFSF_DISJOINT
    IF (addInterfaceFaces) CALL BUILD_MESH_INTERFACES
    CALL BUILD_GLOBAL_MESH
    CALL PRINT_MESH_STATS(globalMesh)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh processing'
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_IN_MESH_EXTENSION(readname)
    !------------------------------------------------------------------
    !---- Examines the mesh extension and initialize the reading pointer
    !---- to the appropriate subroutine
    !
    !---- IN
    !---- readname : full mesh name (with extension)
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE GMSH_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH
    USE NACHOS_MESH
    USE RAW_MESH
    USE IO_FILES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readName
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: extension

    LOGICAL(C_BOOL)                  :: fileExists
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    CALL DOES_FILE_EXIST(readName, fileExists)
    IF (.NOT. fileExists) THEN
       WRITE(6, '(a)' ) 'Error in CHECK_IN_MESH_EXTENSION'
       WRITE(6, '(3a)') 'File ',TRIM(readName),' not found '
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Get extension
    !------------------------------------------------------------------
    CALL REMOVE_EXTENSION(readname, extension)
    !------------------------------------------------------------------
    !---- Test file extension
    !------------------------------------------------------------------
    IF (TRIM(extension) == '.') THEN
       IF (idWrite) WRITE(6,'(a)') 'WARNING: No mesh extension given'
       IF (idWrite) WRITE(6,'(a)') 'WARNING: Assuming Medit/Vizir mesh extension'
       Mesh%extension = '.mesh'
       READ_MESH  => READ_MEDIT_MESH
       RETURN
    ENDIF
    !
    !---- Medit binary extension
    IF (TRIM(extension) == '.meshb') THEN
       IF (idWrite) WRITE(6,'(a)') 'Mesh format seems to be Medit/Vizir binary'
       Mesh%extension = '.meshb'
       READ_MESH  => READ_MEDIT_BINARY_MESH
    ENDIF
    !
    !---- Medit extension
    IF (TRIM(extension) == '.mesh') THEN
       IF (idWrite) WRITE(6,'(a)') 'Mesh format seems to be Medit/Vizir'
       Mesh%extension = '.mesh'
       READ_MESH  => READ_MEDIT_MESH
    ENDIF
    !
    !---- Nachos extension
    IF (TRIM(extension) == '.nmesh') THEN
       IF (idWrite) WRITE(6,'(a)') 'Mesh format seems to be Nachos'
       Mesh%extension = '.nmesh'
       READ_MESH  => READ_NACHOS_MESH
    ENDIF
    !
    !---- Gmsh extension
    IF (TRIM(extension) == '.msh') THEN
       IF (idWrite) WRITE(6,'(a)') 'Mesh format seems to be Gmsh'
       Mesh%extension = '.msh'
       READ_MESH  => READ_GMSH_MESH
    ENDIF
    !
    !---- Raw extension
    IF (TRIM(extension) == '.rawMesh') THEN
       IF (idWrite) WRITE(6,'(a)') 'Mesh format seems to be Raw'
       Mesh%extension = '.rawMesh'
       READ_MESH  => READ_RAW_MESH
    ENDIF
    !------------------------------------------------------------------
    !---- Filling Mesh structure
    !------------------------------------------------------------------
    Mesh%name = TRIM(readname)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (TRIM(Mesh%extension) == '') THEN
       IF (idWrite) WRITE(6, '(a)') 'Error in CHECK_IN_MESH_EXTENSION'
       IF (idWrite) WRITE(6, '(a)') 'Unknown mesh extension provided'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_IN_MESH_EXTENSION

  !====================================================================
  !====================================================================
  !==================================================================== 

  SUBROUTINE CHECK_OUT_MESH_EXTENSION(outputExt)
    !------------------------------------------------------------------
    !---- Examines the output extension and initialize the writing pointer
    !---- to the appropriate subroutine
    !
    !---- IN
    !---- outputExt : output mesh extension
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE GMSH_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH
    USE NACHOS_MESH
    USE RAW_MESH
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: outputExt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Test file extension
    !------------------------------------------------------------------
    IF (TRIM(outputExt) == '.') THEN
       WRITE_MESH => WRITE_MEDIT_MESH
    ENDIF
    !
    !---- Medit extension
    IF (TRIM(outputExt) == '.mesh') THEN
       WRITE_MESH => WRITE_MEDIT_MESH
    ENDIF
    !
    !---- Medit binary extension
    IF (TRIM(outputExt) == '.meshb') THEN
       WRITE_MESH => WRITE_MEDIT_BINARY_MESH
    ENDIF
    !
    !---- Nachos extension
    IF (TRIM(outputExt) == '.nmesh') THEN
       WRITE_MESH => WRITE_NACHOS_MESH
    ENDIF
    !
    !---- Gmsh extension
    IF (TRIM(outputExt) == '.msh') THEN
       WRITE_MESH => WRITE_GMSH_MESH
    ENDIF
    !
    !---- Raw extension
    !---- WARNING : this format requires an argument, so it is not 
    !---- possible to directly convert to it
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_OUT_MESH_EXTENSION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_IN_SOLUTION_EXTENSION(readname)
    !------------------------------------------------------------------
    !---- Examines the solution extension and initialize the reading
    !---- pointer to the appropriate subroutine
    !
    !---- IN
    !---- readname : full solution name (with extension)
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE GMSH_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH    
    USE NACHOS_MESH
    USE RAW_MESH
    USE IO_FILES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readname
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: extension
    
    LOGICAL(C_BOOL)                  :: fileExists
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    CALL DOES_FILE_EXIST(readName, fileExists)
    IF (.NOT. fileExists) THEN
       WRITE(6, '(a)' ) 'Error in CHECK_IN_MESH_EXTENSION'
       WRITE(6, '(3a)') 'File ',TRIM(readname),' not found '
       CALL TILT
    ENDIF    
    !------------------------------------------------------------------
    !---- Get extension
    !------------------------------------------------------------------
    CALL REMOVE_EXTENSION(readname, extension)
    !------------------------------------------------------------------
    !---- Test file extension
    !------------------------------------------------------------------
    IF (TRIM(extension) == '.') THEN
       IF (idWrite) WRITE(6,'(a)') 'WARNING: No solution extension given'
       IF (idWrite) WRITE(6,'(a)') 'WARNING: Assuming Medit/Vizir solution extension'
       Mesh%solutionExtension = '.sol'
       READ_SOLUTION  => READ_MEDIT_SOLUTION
       RETURN
    ENDIF
    !
    !---- Medit binary extension
    IF (TRIM(extension) == '.solb') THEN
       IF (idWrite) WRITE(6,'(a)') 'Solution format seems to be Medit/Vizir binary'
       Mesh%solutionExtension = '.meshb'
       !       READ_SOLUTION  => READ_MEDIT_BINARY_SOLUTION
       IF (idWrite) WRITE(6,'(a)') 'NOT IMPLEMENTED YET'
       CALL TILT
    ENDIF
    !
    !---- Medit extension
    IF (TRIM(extension) == '.sol') THEN
       IF (idWrite) WRITE(6,'(a)') 'Solution format seems to be Medit/Vizir'
       Mesh%solutionExtension = '.sol'
       READ_SOLUTION  => READ_MEDIT_SOLUTION
    ENDIF
    !
    !---- Raw extension
    IF (TRIM(extension) == '.rawSol') THEN
       IF (idWrite) WRITE(6,'(a)') 'Solution format seems to be Raw'
       Mesh%solutionExtension = '.rawSol'
       READ_SOLUTION  => READ_RAW_SOLUTION
    ENDIF    
    !------------------------------------------------------------------
    !---- Filling Mesh structure
    !------------------------------------------------------------------
    Mesh%solutionName = TRIM(readname)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (TRIM(Mesh%solutionExtension) == '') THEN
       IF (idWrite) WRITE(6, '(a)') 'Error in CHECK_IN_SOLUTION_EXTENSION'       
       IF (idWrite) WRITE(6, '(a)') 'Unknown solution extension provided'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_IN_SOLUTION_EXTENSION

  !====================================================================
  !====================================================================
  !==================================================================== 

  SUBROUTINE CHECK_OUT_SOLUTION_EXTENSION(solutionExtension)
    !------------------------------------------------------------------
    !---- Examines the output extension and initialize the writing pointer
    !---- to the appropriate subroutine
    !
    !---- IN
    !---- solutionExtension : output solution extension
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE GMSH_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH
    USE NACHOS_MESH
    USE RAW_MESH
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN) :: solutionExtension
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Test file extension
    !------------------------------------------------------------------
    IF (TRIM(solutionExtension) == '.') THEN
       WRITE_SOLUTION => WRITE_MEDIT_SOLUTION
    ENDIF
    !
    !---- Medit extension
    IF (TRIM(solutionExtension) == '.sol') THEN
       WRITE_SOLUTION => WRITE_MEDIT_SOLUTION
    ENDIF
    !
    !---- Medit binary extension
    IF (TRIM(solutionExtension) == '.solb') THEN
       WRITE_SOLUTION => WRITE_MEDIT_BINARY_SOLUTION
    ENDIF
    !
    !---- Raw extension
    IF (TRIM(solutionExtension) == '.rawSol') THEN
       WRITE_SOLUTION => WRITE_RAW_SOLUTION
    ENDIF    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_OUT_SOLUTION_EXTENSION

  !====================================================================
  !====================================================================
  !==================================================================== 

  SUBROUTINE SOLUTION_TO_MESH_EXTENSIONS(solutionExt, meshExt)
    !------------------------------------------------------------------
    !---- Provides the corresponding mesh extension given a solution extension
    !
    !---- IN
    !---- solutionExt : solution extension
    !
    !---- OUT
    !---- meshExt     : corresponding mesh extension
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE GMSH_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH
    USE NACHOS_MESH
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)    :: solutionExt
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(INOUT) :: meshExt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Test file extension
    !------------------------------------------------------------------
    meshExt = ''
    !
    !---- Medit extension
    IF (TRIM(solutionExt) == '.sol') THEN
       meshExt = '.mesh'
    ENDIF
    !
    !---- Medit binary extension
    IF (TRIM(solutionExt) == '.solb') THEN
       meshExt = '.meshb'
    ENDIF
    !
    !---- Raw extension
    IF (TRIM(solutionExt) == '.rawSol') THEN
       meshExt = '.rawMesh'
    ENDIF    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (TRIM(meshExt) == '') THEN
       IF (idWrite) WRITE(6, '(a)') 'Error in SOLUTION_TO_MESH_EXTENSIONS'
       IF (idWrite) WRITE(6, '(a)') 'Unknown solution extension provided'
       CALL TILT
    ENDIF       
    !------------------------------------------------------------------
  END SUBROUTINE SOLUTION_TO_MESH_EXTENSIONS

  !====================================================================
  !====================================================================
  !==================================================================== 

  SUBROUTINE MESH_TO_SOLUTION_EXTENSIONS(meshExt, solutionExt)
    !------------------------------------------------------------------
    !---- Provides the corresponding solution extension given a mesh extension
    !
    !---- IN
    !---- meshExt     : mesh extension    
    !
    !---- OUT
    !---- solutionExt : corresponding solution extension
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE GMSH_MESH
    USE MEDIT_MESH
    USE MEDIT_BINARY_MESH
    USE NACHOS_MESH
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(INOUT) :: solutionExt
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)    :: meshExt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Test file extension
    !------------------------------------------------------------------
    solutionExt = ''
    !
    !---- Medit extension
    IF (TRIM(meshExt) == '.mesh') THEN
       solutionExt = '.sol'
    ENDIF
    !
    !---- Medit binary extension
    IF (TRIM(meshExt) == '.meshb') THEN
       solutionExt = '.solb'
    ENDIF
    !
    !---- Raw extension
    IF (TRIM(meshExt) == '.rawMesh') THEN
       solutionExt = '.rawSol'
    ENDIF    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (TRIM(solutionExt) == '') THEN
       IF (idWrite) WRITE(6, '(a)') 'Error in MESH_TO_SOLUTION_EXTENSIONS'
       IF (idWrite) WRITE(6, '(a)') 'Unknown mesh extension provided'
       CALL TILT
    ENDIF       
    !------------------------------------------------------------------
  END SUBROUTINE MESH_TO_SOLUTION_EXTENSIONS

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_STRUCTURES
