MODULE MESH_TFSF_FACES

  !====================================================================
  ! Everything related to TFSF faces
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_TFSF_ZONES_AFTER_READING
    !------------------------------------------------------------------
    !---- This routine fills the complementary structures for TF/SF faces
    !---- depending on the log of the cells
    !---- WARNING : the log is then modified in the case of SF cells
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, idFace, iFace, iTfsf, log
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate and complete informations of TF/SF faces
    !------------------------------------------------------------------
    ALLOCATE(CompFacTfsf(Mesh%nFacTfsf))
    !
    idFace = 0
    DO WHILE (idFace < Mesh%nFacTfsf)
       DO iTfsf=1,2
          idFace    = idFace + 1
          IF (idFace > Mesh%nFacTfsf) EXIT
          iFace     = MeshFacTfsf(idFace)%faceNum
          iCell     = MeshFac(iFace)%neighb(iTfsf)
          IF (iCell > Mesh%nCell) THEN
             idFace = idFace - 1
             CYCLE
          ENDIF
          CompFacTfsf(idFace)%cellNum      = iCell
          CompFacTfsf(idFace)%localFaceNum = MeshFac(iFace)%numLocal(iTfsf)
          log                              = MeshCell(iCell)%log
          !
          IF (log >= 1000) THEN
             CompFacTfsf(idFace)%zone = 'SF'
             CompFacTfsf(idFace)%sign =-1.0d0
          ELSE
             CompFacTfsf(idFace)%zone = 'TF'
             CompFacTfsf(idFace)%sign = 1.0d0
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Modify cell tags according to TF/SF zone
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       log = MeshCell(iCell)%log
       !
       IF (log >= 1000) THEN
          MeshCell(iCell)%zone = 'SF'
          MeshCell(iCell)%log  = log - 1000
          Mesh%nCellSfZone     = Mesh%nCellSfZone + 1
       ELSE
          MeshCell(iCell)%zone = 'TF'
          Mesh%nCellTfZone     = Mesh%nCellTfZone + 1          
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished TF/SF zones processing'
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_TFSF_ZONES_AFTER_READING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_TFSF_ZONES_BEFORE_WRITING
    !------------------------------------------------------------------
    !---- This routine resets the cells log values depending on the
    !---- zone they are in (TF or SF)
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Modify cell tags according to TF/SF zone
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       IF (MeshCell(iCell)%zone == 'SF') MeshCell(iCell)%log = MeshCell(iCell)%log + 1000
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_TFSF_ZONES_BEFORE_WRITING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE IS_TFSF_DISJOINT
    !------------------------------------------------------------------
    !---- Check if TF/SF surface is disjoint
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iTfsf, iFac, iCell, localNum

    REAL(C_DOUBLE)               :: faceArea
    REAL(C_DOUBLE), DIMENSION(3) :: checkArea, checkSum, nVector
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if TF/SF surface is disjoint
    !------------------------------------------------------------------
    checkArea(:) = 0.0
    !
    DO iTfsf=1,Mesh%nFacTfsf
       IF (compFacTfsf(iTfsf)%zone == 'SF') CYCLE
       !
       iFac         = MeshFacTfsf(iTfsf)%faceNum
       faceArea     = MeshFac(iFac)%measure
       localNum     = compFacTfsf(iTfsf)%localFaceNum
       iCell        = compFacTfsf(iTfsf)%cellNum
       nVector(:)   = MeshCell(iCell)%normal(:,localNum)
       checkArea(:) = checkArea(:) + faceArea*nVector(:)
    ENDDO
    !
    CALL GLOBAL_OP(checkArea, MPI_SUM, checkSum)
    IF (NORM2(checkSum) > Mesh%tolerance2D) Mesh%isTfsfDisjoint = .FALSE.
    !------------------------------------------------------------------
    !---- Testing for errors 
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (Mesh%isTfsfDisjoint .AND. idWrite) WRITE(6,'(a)') 'TF/SF interface is disjoint'    
    !------------------------------------------------------------------
  END SUBROUTINE IS_TFSF_DISJOINT

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_TFSF_FACES
