MODULE MESH_TYPE_CONVERSIONS

  !====================================================================
  ! Routines related to type (cell, face or vertex) conversions
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FACE_TYPE_TO_VERTS(facType, nVerts, nLinVerts, nHoVerts, geoOrder)
    !------------------------------------------------------------------
    !---- Returns the number of vertices for a given type of face
    !
    !---- IN
    !---- facType   : type of face
    !
    !---- OUT
    !---- nVerts    : number of vertices
    !---- nLinVerts : number of linear vertices
    !---- nHoVerts  : number of high-order vertices
    !---- geoOrder  : geometrical order of the face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: facType

    INTEGER(C_INT),                INTENT(OUT) :: nVerts, nLinVerts, &
         nHoVerts, geoOrder
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine number of vertices
    !------------------------------------------------------------------
    SELECT CASE(TRIM(facType))
    CASE('TRI','TRI1')
       nVerts    = 3
       nLinVerts = 3
       nHoVerts  = 0
       geoOrder  = 1
    CASE('TRI2')
       nVerts    = 6
       nLinVerts = 3
       nHoVerts  = 3
       geoOrder  = 2
    CASE('QUA','QUA1')
       nVerts    = 4
       nLinVerts = 4
       nHoVerts  = 0
       geoOrder  = 1
    CASE('QUA2')
       nVerts    = 9
       nLinVerts = 4
       nHoVerts  = 5
       geoOrder  = 2
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FACE_TYPE_TO_VERTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_TYPE_TO_FACE_TYPE(cellType, faceType)
    !------------------------------------------------------------------
    !---- Returns the default face type for a given cell type
    !
    !---- IN
    !---- cellType : type of cell
    !
    !---- OUT
    !---- faceType : type of face
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: cellType    
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(OUT) :: faceType    
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine face type
    !------------------------------------------------------------------
    SELECT CASE(TRIM(cellType))
    CASE('TET')
       faceType = 'TRI'
    CASE('TET1')
       faceType = 'TRI1'
    CASE('TET2')
       faceType = 'TRI2'
    CASE('TET3')
       faceType = 'TRI3'
    CASE('TET4')
       faceType = 'TRI4'
    CASE('HEX')
       faceType = 'QUA'
    CASE('HEX1')
       faceType = 'QUA1'
    CASE('HEX2')
       faceType = 'QUA2'
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_TYPE_TO_FACE_TYPE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_TYPE_TO_VERTS(cellType, nVerts, nLinVerts, nHoVerts, geoOrder)
    !------------------------------------------------------------------
    !---- Returns the number of vertices for a given type of cell
    !
    !---- IN
    !---- cellType  : type of cell
    !---- nLinVerts : number of linear vertices
    !---- nHoVerts  : number of high-order vertices
    !---- geoOrder  : geometrical order of the cell
    !
    !---- OUT
    !---- nVerts   : number of vertices
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: cellType    

    INTEGER(C_INT),                INTENT(OUT) :: nVerts, nLinVerts, &
         nHoVerts, geoOrder
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine number of vertices
    !------------------------------------------------------------------
    SELECT CASE(cellType)
    CASE('TET','TET1')
       nVerts    = 4
       nLinVerts = 4
       nHoVerts  = 0
       geoOrder  = 1
    CASE('TET2')
       nVerts    = 10
       nLinVerts = 4
       nHoVerts  = 6
       geoOrder  = 2
    CASE('TET3')
       nVerts    = 20
       nLinVerts = 4
       nHoVerts  = 14
       geoOrder  = 3
    CASE('TET4')
       nVerts    = 35
       nLinVerts = 4
       nHoVerts  = 31
       geoOrder  = 4
    CASE('HEX','HEX1')
       nVerts    = 8
       nLinVerts = 8
       nHoVerts  = 0
       geoOrder  = 1
    CASE('HEX2')
       nVerts    = 27
       nLinVerts = 8
       nHoVerts  = 19
       geoOrder  = 2
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_TYPE_TO_VERTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_TYPE_TO_NFACES(cellType, nFaces)
    !------------------------------------------------------------------
    !---- Returns the number of faces for a given type of cell
    !
    !---- IN
    !---- cellType : type of cell
    !
    !---- OUT
    !---- nFaces   : number of faces
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: cellType

    INTEGER(C_INT),                INTENT(OUT) :: nFaces
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine number of faces
    !------------------------------------------------------------------
    SELECT CASE(cellType)
    CASE('TET','TET1','TET2','TET3','TET4')
       nFaces = 4
    CASE('HEX','HEX1','HEX2')
       nFaces = 6
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_TYPE_TO_NFACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHANGE_TYPE_WITH_ORDER(order, cellTypeIn, cellTypeOut)
    !------------------------------------------------------------------
    !---- Modify the type with a given order
    !
    !---- IN
    !---- order       : actual polynomial order
    !---- cellTypeIn  : input cell type
    !
    !---- OUT
    !---- cellTypeOut : output cell type
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(IN)  :: cellTypeIn
    CHARACTER(LEN=*, KIND=C_CHAR), INTENT(OUT) :: cellTypeOut

    INTEGER(C_INT),                INTENT(IN)  :: order
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1, KIND=C_CHAR)              :: cOrder
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine new type
    !------------------------------------------------------------------
    CALL INT_TO_CHAR(order, cOrder)
    cellTypeOut = TRIM(cellTypeIn(1:3))//cOrder
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHANGE_TYPE_WITH_ORDER

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_TYPE_CONVERSIONS
