MODULE MESH_VERTICES

  !====================================================================
  ! Definition and data structure associated to the vertices of the mesh 
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a vertex of the mesh
  ! 
  ! coor(k) ==> kth coordinate of the vertex (k=1:3)
  ! volume  ==> measure associated to the vertices of the mesh (i.e. the sum of 
  !             the volumes of the elements sharing the same vertex)
  !--------------------------------------------------------------------
  TYPE vertex
     REAL(C_DOUBLE)               :: volume
     REAL(C_DOUBLE), DIMENSION(3) :: coor
  END TYPE vertex

  !--------------------------------------------------------------------
  ! Global definition of the vertices of the mesh
  !--------------------------------------------------------------------
  TYPE(vertex), DIMENSION(:), ALLOCATABLE :: MeshVert

  !--------------------------------------------------------------------
  ! Miscellaneous stuff
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: vertexBlockSize = 500
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ADD_VERTEX(buffer)
    !------------------------------------------------------------------
    !---- Fill the vertices structure
    !
    !---- IN
    !---- buffer : vertex coordinates
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(:) :: buffer
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Filling structure
    !------------------------------------------------------------------
    CALL EXTEND_VERTEX_ARRAY(MeshVert, Mesh%nVert)
    MeshVert(Mesh%nVert)%coor(:) = buffer(:)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ADD_VERTEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCALE_MESH_COORDINATES(scaling)
    !------------------------------------------------------------------
    !---- Apply scaling to vertices coordinates
    !
    !---- IN
    !---- scaling : scaling value
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: scaling
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Scaling vertices coordinates
    !------------------------------------------------------------------
    MeshVert(:)%coor(1) = MeshVert(:)%coor(1)*scaling
    MeshVert(:)%coor(2) = MeshVert(:)%coor(2)*scaling
    MeshVert(:)%coor(3) = MeshVert(:)%coor(3)*scaling
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCALE_MESH_COORDINATES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TRANSLATE_MESH_COORDINATES(vector)
    !------------------------------------------------------------------
    !---- Apply translation to vertices coordinates
    !
    !---- IN
    !---- vector : translation vector
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3) :: vector
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Translate vertices coordinates
    !------------------------------------------------------------------
    MeshVert(:)%coor(1) = MeshVert(:)%coor(1)+vector(1)
    MeshVert(:)%coor(2) = MeshVert(:)%coor(2)+vector(2)
    MeshVert(:)%coor(3) = MeshVert(:)%coor(3)+vector(3)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TRANSLATE_MESH_COORDINATES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ROTATE_MESH_COORDINATES(axes, theta)
    !------------------------------------------------------------------
    !---- Apply rotation to vertices coordinates
    !
    !---- IN
    !---- axes   : axes of the rotation
    !     theta  : angle of the rotation
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATRICES_STRUCTURES
    USE MESH_DATA
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE)               :: theta
    REAL(C_DOUBLE), DIMENSION(3) :: axes
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                 :: iVert

    REAL(C_DOUBLE)                 :: c, s, norm
    REAL(C_DOUBLE), DIMENSION(3)   :: res
    REAL(C_DOUBLE), DIMENSION(3,3) :: R
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Precompute matrix R
    !------------------------------------------------------------------
    CALL NORMALIZE_VECTOR(axes, norm)
    CALL DEGREES_TO_RADIANS(theta)
    !
    c = COS(theta)
    s = SIN(theta)
    !
    R = RESHAPE( (/ &
         axes(1)**2 *(1-c)+c,               &
         axes(1)*axes(2)*(1-c) - axes(3)*s, &
         axes(1)*axes(3)*(1-c) + axes(2)*s, &
         axes(1)*axes(2)*(1-c) + axes(3)*s, &
         axes(2)**2 *(1-c)+c,               &
         axes(2)*axes(3)*(1-c) - axes(1)*s, &
         axes(1)*axes(3)*(1-c) - axes(2)*s, &
         axes(2)*axes(3)*(1-c) + axes(1)*s, &
         axes(3)**2 *(1-c)+c/),             &
         (/3,3/) )
    !------------------------------------------------------------------
    !---- Rotate vertices coordinates
    !------------------------------------------------------------------
    DO iVert = 1,Mesh%nVert
       CALL MATVECMUL(R, MeshVert(iVert)%coor(:), 3, 3, res)
       MeshVert(iVert)%coor(:) = res
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ROTATE_MESH_COORDINATES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SYMMETRISE_MESH_COORDINATES(plane, coorMin, coorMax)
    !------------------------------------------------------------------
    !---- Apply symmetry to vertices coordinates
    !
    !---- IN
    !---- plane   : symmetry plane
    !---- coorMin : minimal coordinates of the mesh
    !---- coorMax : maximal coordinates of the mesh
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=2, KIND=C_CHAR), INTENT(IN) :: plane

    REAL(C_DOUBLE), DIMENSION(3),  INTENT(IN) :: coorMin, coorMax
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: iVert, iCoor
    !
    REAL(C_DOUBLE)                            :: planeCoor
    REAL(C_DOUBLE), DIMENSION(3)              :: vector
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Determine the coordinates of the symmetry plane
    !------------------------------------------------------------------
    vector = 0
    !
    SELECT CASE(plane)
    CASE("x1")
       iCoor=1
       planeCoor=coorMin(1)
    CASE("x2")
       iCoor=1
       planeCoor=coorMax(1)
    CASE("y1")
       iCoor=2
       planeCoor=coorMin(2)
    CASE("y2")
       iCoor=2
       planeCoor=coorMax(2)
    CASE("z1")
       iCoor=3
       planeCoor=coorMin(3)
    CASE("z2")
       iCoor=3
       planeCoor=coorMax(3)
    END SELECT
    !------------------------------------------------------------------
    !---- Proceed to symmetry for each point
    !------------------------------------------------------------------
    DO iVert = 1,Mesh%nVert
       vector(iCoor) = 2.0*(planeCoor - MeshVert(iVert)%coor(iCoor))
       !
       MeshVert(iVert)%coor(:) = MeshVert(iVert)%coor(:)+vector(:)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SYMMETRISE_MESH_COORDINATES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_VERTEX_ARRAY(vertexArray, vertexArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to vertexArray
    !
    !---- IN
    !---- vertexArray     : array of "vertex" structure
    !---- vertexArraySize : number of elements in vertexArray
    !
    !---- OUT
    !---- vertexArray     : modified in the process
    !---- vertexArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),           INTENT(INOUT) :: vertexArraySize
    
    TYPE(vertex), DIMENSION(:), ALLOCATABLE :: vertexArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                         :: isArrayFull, isAllocated

    INTEGER(C_INT)                          :: iVert
    
    TYPE(vertex), DIMENSION(:), ALLOCATABLE :: tempVertex
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(vertexArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(vertexArray) < vertexArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF        
    !------------------------------------------------------------------
    !---- Extend array
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempVertex)) DEALLOCATE(tempVertex)
          ALLOCATE(tempVertex(vertexArraySize))
          tempVertex = vertexArray
          DEALLOCATE(vertexArray)
       ENDIF
       !
       ALLOCATE(vertexArray(vertexArraySize + vertexBlockSize))
       DO iVert=vertexArraySize+1,vertexArraySize+vertexBlockSize
          CALL INITIALIZE_VERTEX_STRUCTURE(vertexArray(iVert))
       ENDDO
       !
       vertexArray(1:vertexArraySize) = tempVertex(1:vertexArraySize)
       IF (ALLOCATED(tempVertex)) DEALLOCATE(tempVertex)
    ENDIF
    !
    vertexArraySize = vertexArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_VERTEX_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_VERTEX_STRUCTURE(vert)
    !------------------------------------------------------------------
    !---- Initialize vertex structure 
    !
    !---- IN
    !---- vert : vertex structure
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(vertex) :: vert
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize MeshtFacB structure
    !------------------------------------------------------------------
    vert%coor(:) = 0.0
    vert%volume  = 0.0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_VERTEX_STRUCTURE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEALLOCATE_VERTICES
    !------------------------------------------------------------------
    !---- Deallocate vertices
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate vertices
    !------------------------------------------------------------------
    IF (ALLOCATED(MeshVert)) DEALLOCATE(MeshVert)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_VERTICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_VERTEX_IN_LIST(vertIn, vertList, vertSize, vertIndex)
    !------------------------------------------------------------------
    !---- Find a vertex in a list of vertices
    !
    !---- IN
    !---- vertIn    : coordinates of vertex to seek
    !---- vertList  : list of vertices
    !---- vertSize  : size of vertices list
    !
    !---- OUT
    !---- vertIndex : index of vertex in given list
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SORT
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: vertSize
    INTEGER(C_INT),               INTENT(OUT) :: vertIndex    
    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: vertIn

    TYPE(vertex),   DIMENSION(:), ALLOCATABLE :: vertList
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: iVert
    
    REAL(C_DOUBLE), DIMENSION(3)              :: vert, locVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Find vertex
    !------------------------------------------------------------------
    vertIndex = 0
    vert(1:3)   = vertIn(1:3)
    CALL INSERTION_SORT(vert)
    !
    DO iVert=1,vertSize
       locVert(1:3) = vertList(iVert)%coor(1:3)
       CALL INSERTION_SORT(locVert)
       IF ( (ABS(vert(1)-locVert(1)) < Mesh%tolerance1D) .AND. &
            (ABS(vert(2)-locVert(2)) < Mesh%tolerance1D) .AND. &
            (ABS(vert(3)-locVert(3)) < Mesh%tolerance1D)) THEN
          !
          vertIndex = iVert
          RETURN
       ENDIF
    ENDDO    
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FIND_VERTEX_IN_LIST

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COPY_VERTICES(vertexArrayIn, vertexArrayOut, vertexArraySize)
    !------------------------------------------------------------------
    !---- Create a copy of a vertex array
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),             INTENT(IN)  :: vertexArraySize
    
    TYPE(vertex), DIMENSION(:), ALLOCATABLE :: vertexArrayIn
    TYPE(vertex), DIMENSION(:), ALLOCATABLE :: vertexArrayOut
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Copy global vertexs
    !------------------------------------------------------------------
    IF (ALLOCATED(vertexArrayOut)) DEALLOCATE(vertexArrayOut)
    IF (ALLOCATED(vertexArrayIn))  ALLOCATE(vertexArrayOut(vertexArraySize))
    IF (ALLOCATED(vertexArrayIn))  vertexArrayOut = vertexArrayIn
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COPY_VERTICES

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MESH_VERTICES
