MODULE NUMERICAL_TYPES

  !====================================================================
  ! All the types used in the DG routines
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE MESH_DATA
  USE MESH_CELLS
  USE MATRICES_STRUCTURES
  USE MATERIALS_STRUCTURES
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Various values
  !--------------------------------------------------------------------
  ! nCellEff        ==> effective # of cells, for allocation purposes
  ! nPCells         ==> # of cells of each polynomial order P*
  ! nPFaces         ==> # of faces of each polynomial order P* (usefull for hdg)
  ! genPCells       ==> global value of genPCells
  ! nQCells         ==> # of cells of each polynomial order Q*
  ! nQFaces         ==> # of faces of each polynomial order Q* (usefull for hdg)
  ! genQCells       ==> global value of genQCells
  ! nPmlCells       ==> # of PML cells
  ! genPmlCells     ==> global value of nPMLCells
  ! nSfCells        ==> # of scattered-field cells (non Pml)
  ! genSfCells      ==> global value of nSfCells (non Pml)
  ! nSfPmlCells     ==> # of scattered-field cells (Pml)
  ! genSfPmlCells   ==> global value of nSfCells (Pml)
  ! nDispCellsTot   ==> total # of dispersive cells = gen + NL
  ! nDispCellsGen   ==> # of generalized dispersion model cells
  ! nDispCellsNl    ==> # of nonlocal dispersion model cells
  ! genDispCellsTot ==> total # of dispersive cells = gen + NL (global)
  ! genDispCellsGen ==> # of generalized dispersion model cells (global)
  ! genDispCellsNl  ==> # of nonlocal dispersion model cells (global)
  !
  ! nMat            ==> total # of materials, excluding vacuum
  ! nMatDisp        ==> total # of dispersive materials
  !--------------------------------------------------------------------
  INTEGER(C_INT)                 :: nCellEff, nPmlCells, &
       genPmlCells, nSfCells, genSfCells, nDispCellsTot, &
       nDispCellsGen, nDispCellsNl, genDispCellsTot,     &
       genDispCellsGen, genDispCellsNl, nSfPmlCells,     &
       genSfPmlCells, nMat, nMatDisp

  INTEGER(C_INT), DIMENSION(0:4) :: nPCells(0:4), genPCells(0:4)
  INTEGER(C_INT), DIMENSION(0:4) :: nPFaces(0:4), genPFaces(0:4)
  INTEGER(C_INT), DIMENSION(0:4) :: nQCells(0:4), genQCells(0:4)
  INTEGER(C_INT), DIMENSION(0:4) :: nQFaces(0:4), genQFaces(0:4)

  !--------------------------------------------------------------------
  ! Non-local boundary faces
  !--------------------------------------------------------------------
  ! nBoundaryNlTot    ==> total # of nonlocal boundary faces
  ! nBoundaryNl       ==> # of nonlocal boundary faces after the
  !                       processing of periodic nonlocal faces
  ! nBoundaryNlPer    ==> # of periodic   nonlocal boundary faces
  ! nBoundaryNlPhy    ==> # of hard-wall  nonlocal boundary faces
  ! nBoundaryNlArt    ==> # of artificial nonlocal boundary faces
  ! boundaryNl*       ==> nonlocal boundary faces of given kind (see above)
  ! compBoundaryNl*   ==> complementary structure for nonlocal boundary faces
  !--------------------------------------------------------------------
  INTEGER(C_INT)                                             :: &
       nBoundaryNlTot, nBoundaryNl, nBoundaryNlPer,             &
       nBoundaryNlPhy, nBoundaryNlArt

  TYPE(derived_face),              DIMENSION(:), ALLOCATABLE :: &
       boundaryNlTot, boundaryNl, boundaryNlPer, boundaryNlPhy

  TYPE(complement_interface_face), DIMENSION(:), ALLOCATABLE :: &
       compBoundaryNlTot, compBoundaryNl, compBoundaryNlPhy,    &
       compBoundaryNlPer

  !--------------------------------------------------------------------
  ! Array of procedure pointer
  !--------------------------------------------------------------------
  ! proc          ==> procedure pointer
  !--------------------------------------------------------------------
  TYPE proc_ptr_array
     PROCEDURE(), NOPASS, POINTER :: proc
  END TYPE proc_ptr_array

  !--------------------------------------------------------------------
  ! Numerical element
  !--------------------------------------------------------------------
  ! E, H          ==> EM fields
  ! isPml         ==> TRUE if element is in the PML region
  ! isDisp        ==> TRUE if element is             dispersive
  ! isNL          ==> TRUE if element is non-local   dispersive
  ! isGenDisp     ==> TRUE if element is generalized dispersive
  ! nDofV         ==> volumic  # of degrees of freedom
  ! nDofS         ==> surfacic # of degrees of freedom
  ! order         ==> polynomial order for field representation
  ! pmlCellNum    ==> # of this element in pml the numbering
  ! dispCellNum   ==> # of this element in disp the numbering
  ! orderNeighbor ==> polynomial order of neighbor cells
  ! iJac          ==> Inverse jacobian matrix
  ! dJac          ==> Determinant of the jacobian matrix
  ! eps, mu       ==> material characteristics
  ! mat           ==> pointer to material structures
  ! massMtx       ==> Mass matrix of the element (M)
  ! *StfMtx       ==> Stiffness matrices of the element (K)
  ! srfMtx(:)     ==> Surface   matrices of the element (S)
  ! MATVECMUL_*   ==> Pointer to matrix-vector multiplication routines
  ! GET_*_DOF     ==> Pointer to routine giving the d.o.f.
  ! zone          ==> 'TF' (total field) or 'SF' (scattered field)
  ! iTot          ==> Index in of dispersive element (Gen+Nl)
  ! iGen          ==> Index of generalized disp element
  ! iNl           ==> Index of nonlocal disp element
  !--------------------------------------------------------------------
  TYPE numerical_element_dg
     TYPE(real_vector_field)                          :: E, H

     LOGICAL(C_BOOL)                                  :: isPml, isDisp, &
          isGenDisp, isNL

     INTEGER(C_INT)                                   :: nDofS, nDofV, order, &
          pmlCellNum, iTot, iGen, iNl
     INTEGER(C_INT), DIMENSION(:), ALLOCATABLE        :: orderNeighbor

     REAL(C_DOUBLE)                                   :: dJac
     REAL(C_DOUBLE), DIMENSION(3,3)                   :: iJac

     TYPE(matrix_sq_real),                    POINTER :: massMtx, iMassMtx, &
          xStfMtx, yStfMtx, zStfMtx

     TYPE(matrix_sq_integer), DIMENSION(:),   POINTER :: iPermDof

     TYPE(rect_matrix_ptr), DIMENSION(:), ALLOCATABLE :: srfMtx

     TYPE(material),                          POINTER :: mat

     PROCEDURE(), NOPASS,                     POINTER :: MATVECMUL_SRF, MATVECMUL_VLM
     PROCEDURE(), NOPASS,                     POINTER :: COMPUTE_VOLUMIC_RHS
     PROCEDURE(), NOPASS,                     POINTER :: COMPUTE_FLUX
     PROCEDURE(), NOPASS,                     POINTER :: COMPUTE_INC_FLUX
     PROCEDURE(), NOPASS,                     POINTER :: RHS_UPDATE
     TYPE(proc_ptr_array),  DIMENSION(:), ALLOCATABLE :: COMPUTE_SURFACIC_RHS
  END TYPE numerical_element_dg

  !--------------------------------------------------------------------
  ! Numerical element for hdg method
  !--------------------------------------------------------------------
  ! E, H                         ==> EM fields
  ! nDofV                        ==> volumic  # of degrees of freedom
  ! nDofS                        ==> surfacic # of degrees of freedom
  ! nDofF(k)                     ==> # of degrees of freedom of the kth
  !                                  face of the cell
  ! sumDofF(k)                   ==> sum (1:k) of # of degrees of freedom
  !                                  of faces of the cell
  ! order                        ==> polynomial order for field representation
  ! orderF(k)                    ==> order of the kth face of the cell
  ! eps, mu, rho, sigma, tau, z  ==> material characteristics
  ! iPermDof                     ==> permutations of DoFs on th element faces
  ! massMtx                      ==> Mass matrix of the element (M)
  ! iMassMtx                     ==> Inv Mass matrix of the element (M^(-1))
  ! *StfMtx                      ==> Stiffness matrices of the element (K)
  ! srfMtx*(:)                   ==> Surface   matrices of the element (S)
  !--------------------------------------------------------------------
  TYPE numerical_element_hdg
     TYPE(complex_vector_field)                       :: E, H

     INTEGER(C_INT)                                   :: nDofS, nDofV, order
     INTEGER(C_INT),        DIMENSION(:), ALLOCATABLE :: nDofF, orderF, sumDofF

     REAL(C_DOUBLE)                                   :: eps, mu, rho, sigma, tau, z

     TYPE(sq_matrix_ptr)                              :: iPermDof

     TYPE(matrix_sq_real),  POINTER                   :: massMtx, iMassMtx, &
          xStfMtx, yStfMtx, zStfMtx

     TYPE(rect_matrix_ptr), DIMENSION(:), ALLOCATABLE :: srfMtxE, srfMtxF, srfMtxG
  END TYPE numerical_element_hdg

  !--------------------------------------------------------------------
  ! Numerical face
  !--------------------------------------------------------------------
  ! E, H          ==> EM fields
  ! nDofS         ==> surfacic # of degrees of freedom
  ! order         ==> polynomial order for field representation
  !--------------------------------------------------------------------
  TYPE numerical_face
     TYPE(real_vector_field) :: E, H

     INTEGER(C_INT)          :: nDofS, order
  END TYPE numerical_face

  !--------------------------------------------------------------------
  ! Numerical face for hdg method
  !--------------------------------------------------------------------
  ! L         ==> hybrid variable (Lambda field)
  ! order     ==> polynomial order of the face in HDG method
  ! nDof      ==> surfacic # of degrees of freedom
  !--------------------------------------------------------------------
  TYPE numerical_face_hdg
     TYPE(complex_vector_field)   :: L

     INTEGER(C_INT)               :: nDof, order
  END TYPE numerical_face_hdg

  !--------------------------------------------------------------------
  ! Miscellaneous stuff
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: eltBlockSize  = 1000

  !--------------------------------------------------------------------
  ! Dispersive element
  !--------------------------------------------------------------------
  ! J#*, P#*      ==> Pol. currents of the gen. disp. model
  ! Jnl, Qnl      ==> Pol. current and charge density of nl disp. model
  ! cellNum       ==> Index of   the corresponding numerical element
  ! neighb        ==> Neighbours in the nonlocal case (nonlocal numbering)
  ! ptr           ==> Pointer to the corresponding numerical element
  ! mat           ==> Pointer to the corresponding material
  ! matDisp       ==> Pointer to the corresponding dispersive material
  ! COMPUTE_RHS*  ==> Pointer to corresponding RHS computation (gen, nl)
  ! ADD_RHS*      ==> Pointer to the right adding to amperes law (gen, nl)
  ! COMPUTE_VOL*  ==> Pointer to the volumic PDE part (only for nonlocal at the moment)
  ! RHS_UPDATE    ==> Pointer to the RHS (PDE) part (only for nonlocal at the moment)
  !--------------------------------------------------------------------
  TYPE dispersive_element
     INTEGER(C_INT)                                     :: cellNum

     INTEGER(C_INT),          DIMENSION(:), ALLOCATABLE :: neighb

     TYPE(real_vector_field), DIMENSION(:), ALLOCATABLE :: J0, P1, J2, P2
     TYPE(real_vector_field)                            :: Jnl, Qnl

     TYPE(numerical_element_dg),                POINTER :: ptr

     TYPE(material),                            POINTER :: mat
     TYPE(dispersive_material),                 POINTER :: matDisp

     PROCEDURE(), NOPASS,                       POINTER :: COMPUTE_RHS_DISP_CURRENTS
     PROCEDURE(), NOPASS,                       POINTER :: ADD_RHS_DISP_CURRENTS
     PROCEDURE(), NOPASS,                       POINTER :: COMPUTE_VOLUMIC_RHS
     PROCEDURE(), NOPASS,                       POINTER :: RHS_UPDATE
     PROCEDURE(), NOPASS,                       POINTER :: COMPUTE_J_FLUX
     PROCEDURE(), NOPASS,                       POINTER :: COMPUTE_Q_FLUX

     TYPE(proc_ptr_array),    DIMENSION(:), ALLOCATABLE :: COMPUTE_SURFACIC_RHS
  END TYPE dispersive_element

  !--------------------------------------------------------------------
  ! Derived dispersive element
  !--------------------------------------------------------------------
  ! iDispCell     ==> index of the associated dispersive element
  !--------------------------------------------------------------------
  TYPE derived_dispersive_element
     INTEGER(C_INT) :: iDispCell
  END TYPE derived_dispersive_element

  !--------------------------------------------------------------------
  ! PML element
  !--------------------------------------------------------------------
  ! cellNum       ==> Index of   the corresponding numerical element
  ! ptr           ==> Pointer to the corresponding numerical element
  ! g?*           ==> additional fields for the CFS-PML
  ! alpha         ==> CFS-PML parameter for complex shifting
  ! sigma         ==> CFS-PML parameter for absorption level
  ! kappa         ==> CFS-PML parameter for real stretching
  !--------------------------------------------------------------------
  TYPE pml_element
     TYPE(real_vector_field)             :: gEx, gEy, gEz, gHx, gHy, gHz
     TYPE(real_vector_field)             :: fEx, fEy, fEz, fHx, fHy, fHz

     INTEGER(C_INT)                      :: cellNum

     REAL(C_DOUBLE), DIMENSION(3)        :: alpha, sigma, kappa

     TYPE(numerical_element_dg), POINTER :: ptr
  END TYPE pml_element

  !--------------------------------------------------------------------
  ! SF element
  !--------------------------------------------------------------------
  ! cellNum       ==> Index of   the corresponding numerical element
  !--------------------------------------------------------------------
  TYPE sf_element
     INTEGER(C_INT) :: cellNum
  END TYPE sf_element

  !--------------------------------------------------------------------
  ! Discretized fields
  !--------------------------------------------------------------------
  TYPE(numerical_element_dg),       DIMENSION(:), ALLOCATABLE, TARGET :: dgEM
  TYPE(numerical_element_hdg),      DIMENSION(:), ALLOCATABLE, TARGET :: hdgEM
  TYPE(dispersive_element),         DIMENSION(:), ALLOCATABLE, TARGET :: disp
  TYPE(derived_dispersive_element), DIMENSION(:), ALLOCATABLE         :: dispNL, dispGen
  TYPE(pml_element),                DIMENSION(:), ALLOCATABLE         :: pml
  TYPE(sf_element),                 DIMENSION(:), ALLOCATABLE         :: Sf, SfPml
  TYPE(numerical_face),             DIMENSION(:), ALLOCATABLE         :: EMf
  TYPE(numerical_face_hdg),         DIMENSION(:), ALLOCATABLE         :: Lambda

  !--------------------------------------------------------------------
  ! Maxwell vector fields for update computation
  !--------------------------------------------------------------------
  TYPE(real_vector_field), DIMENSION(:), ALLOCATABLE :: &
       Ek, En, &
       Hk, Hn

  !--------------------------------------------------------------------
  ! Disp RK fields
  ! nPole   ==> n vector fields for poles of order #
  !--------------------------------------------------------------------
  TYPE disp_RK_field
     TYPE(real_vector_field),   DIMENSION(:), ALLOCATABLE   :: nPole
  END TYPE disp_RK_field

  !--------------------------------------------------------------------
  ! Disp vector fields for update computation
  !--------------------------------------------------------------------
  TYPE(real_vector_field)                            :: J0, J1, J2

  TYPE(real_vector_field), DIMENSION(:), ALLOCATABLE :: &
       Jnlk, Jnln, &
       Qnlk, Qnln

  TYPE(disp_RK_field),     DIMENSION(:), ALLOCATABLE :: &
       J0k, J0n, &
       P1k, P1n, &
       J2k, J2n, &
       P2k, P2n

  !--------------------------------------------------------------------
  ! PML vector fields for update computation
  !--------------------------------------------------------------------
  TYPE(real_vector_field), DIMENSION(:), ALLOCATABLE :: &
       gExk, gExn, &
       gEyk, gEyn, &
       gEzk, gEzn, &
       gHxk, gHxn, &
       gHyk, gHyn, &
       gHzk, gHzn

  !--------------------------------------------------------------------
  ! Local vectors for RHS computations
  !--------------------------------------------------------------------
  INTEGER(C_INT),    DIMENSION(:),   ALLOCATABLE :: dof, dofN

  REAL(C_DOUBLE),    DIMENSION(:),   ALLOCATABLE :: &
       RHS_Ex, RHS_Ey, RHS_Ez, &
       RHS_Hx, RHS_Hy, RHS_Hz, &
       RHSx,   RHSy,   RHSz,   &
       RHS_Jx, RHS_Jy, RHS_Jz, &
       RHS_Q,                  &
       UxL,    UyL,    UzL,    &
       UxP,    UyP,    UzP,    &
       UxM,    UyM,    UzM,    &
       Ux,     Uy,     Uz,     &
       iUx,    iUy,    iUz,    &
       ExV,    EyV,    EzV,    &
       HxV,    HyV,    HzV,    &
       JxV,    JyV,    JzV,    &
       QV,                     &
       ExL2,   EyL2,   EzL2,   &
       HxL2,   HyL2,   HzL2,   &
       JxL2,   JyL2,   JzL2,   &
       QL2,                    &
       ExS,    EyS,    EzS,    &
       HxS,    HyS,    HzS,    &
       dofXS,  dofYS,  dofZS,  &
       dofXV,  dofYV,  dofZV
  REAL(C_DOUBLE),    DIMENSION(:,:), ALLOCATABLE :: Kx, Ky, Kz

  COMPLEX(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: &
       ExVcx,  EyVcx,  EzVcx,  &
       HxVcx,  HyVcx,  HzVcx,  &
       ExScx,  EyScx,  EzScx,  &
       HxScx,  HyScx,  HzScx

  !--------------------------------------------------------------------
  ! Solver data
  !--------------------------------------------------------------------
  ! timeMethod       ==> Time-domain method
  !                      Methods currently available:
  !                      - 'LSRK14-4' (for upwindFactor = 1.0,       from [1])
  !                      - 'LSRK12-4' (for upwindFactor = 0.0,       from [1])
  !                      - 'LSRK05-4' (for any upwindFactor value,   from [2])
  !                      - 'LSRK21-5' (for upwindFactor = 1.0, from Niegemann)
  !                      - 'LSRK22-4' (for upwindFactor = 1.0, from Niegemann)
  ! polyBase         ==> Polynom base used for field expansion
  !                      Basis currently available:
  !                      - 'L' (Lagrange base)
  ! upwindFactor     ==> Upwinding factor for flux computation (in [0,1])
  ! orderMin*        ==> Minimal authorized polynomial order for tetra/hexa
  ! orderMax*        ==> Maximal authorized polynomial order for tetra/hexa
  ! useCurv          ==> TRUE enables the use of curvilinear elements
  ! useRCM           ==> TRUE enables the use of RCM renumbering
  !
  ! [1] Efficient LSRK schemes with optimized stability regions,
  !     Niegemann et al., JCP 231 (2012), 364-372
  ! [2] 4th order 2N storage RK schemes,
  !     Carpenter et al., Tech report NASA, 1994
  !--------------------------------------------------------------------
  CHARACTER(LEN=8, KIND=C_CHAR) :: timeMethod
  CHARACTER(LEN=1, KIND=C_CHAR) :: polyBase, cOrderMin, cOrderMinH,          &
       cOrderMax, cOrderMaxH

  LOGICAL(C_BOOL)               :: useCurv, useRCM

  INTEGER(C_INT)                :: orderMin, orderMinH, orderMax, orderMaxH, &
       nDofVMin, nDofVMinH, nDofVMax, nDofVMaxH, nDofSMin, nDofSMinH,        &
       nDofSMax, nDofSMaxH, genOrderMin, genOrderMinH, genOrderMax,          &
       genOrderMaxH

  REAL(C_DOUBLE)                :: upwindFactor

  !--------------------------------------------------------------------
  ! PML data
  !--------------------------------------------------------------------
  ! usePml        ==> TRUE if PML is activated
  ! pmlAlpha      ==> complex shifting
  ! pmlSigma      ==> dissipation level
  ! pmlKappa      ==> real stretching
  ! pmlThickness  ==> thickness of the pml
  ! pmlBorder     ==> inner Pml box in direction of incident signal (used for SF)
  ! pmlDamp       ==> attenuation coefficient for SF (fixed frequency)
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)               :: usePml
  LOGICAL(C_BOOL), DIMENSION(3) :: pmlDirection

  INTEGER(C_INT)                :: nPmlDirections

  REAL(C_DOUBLE)                :: pmlAlpha, pmlSigma, pmlKappa, &
       pmlThickness, pmlBorder, pmlDamp
  REAL(C_DOUBLE), DIMENSION(3)  :: pmlInnerBoxMax, pmlInnerBoxMin, &
       globalPmlInnerBoxMax, globalPmlInnerBoxMin

  !--------------------------------------------------------------------
  ! Parallel data
  !--------------------------------------------------------------------
  ! mpiPattern       ==> MPI communication pattern
  !--------------------------------------------------------------------
  CHARACTER(LEN=2, KIND=C_CHAR) :: mpiPattern

  !--------------------------------------------------------------------
  ! General source data
  !--------------------------------------------------------------------
  ! maxTime          ==> Maximal physical time (seconds)
  ! incidentType     ==> Defines how the incident field is applied
  !                      Methods currently available:
  !                      - 'AB' to impose field via Silver-Muller faces
  !                      - 'TF' to impose field via TF/SF faces
  !                      - 'SF' for fully-scattered formulation
  ! sourceType       ==> Type of source
  !                      Methods currently available:
  !                      - 'PW' for plane wave
  !                      - 'EE' for EELS             (via SF formulation only)
  !                      - 'DI' for dipole           (via inversed TF/SF only)
  !                      - 'DM' for modulated dipole (via inversed TF/SF only)
  !                      - 'FI' for step-index fiber mode
  !                      - 'CC' for cubic cavity mode
  !                      - 'NL' for nonlocal cubic cavity mode
  !                      - 'SC' for spherical cavity mode
  ! backgroundMat    ==> Background material for 'AB' and 'TF' incidentType
  !--------------------------------------------------------------------
  CHARACTER(LEN=2, KIND=C_CHAR) :: incidentType, sourceType

  INTEGER(C_INT)                :: backgroundMat

  REAL(C_DOUBLE)                :: maxTime

  !--------------------------------------------------------------------
  ! Observable data
  !--------------------------------------------------------------------
  ! enableObs        ==> FALSE disables all observable computations
  ! freqMinObs       ==> Minimal frequency for observables computations (Hz)
  ! freqMaxObs       ==> Maximal frequency for observables computations (Hz)
  ! deltaFreqObs     ==> Frequency step    for observables computations (Hz)
  ! deltaDsEELS      ==> Spatial step for EELS integral computation (m)
  ! computeCsAbs     ==> TRUE enables absorption cross-section computation
  ! computeCsSca     ==> TRUE enables scattering cross-section computation
  ! computeCsExt     ==> TRUE enables extinction cross-section computation
  ! csIndex          ==> face log on which CS must be computed
  ! computeEELS      ==> TRUE enables EELS computation
  ! computeR         ==> TRUE enables reflection   coefficient computation
  ! computeT         ==> TRUE enables transmission coefficient computation
  ! computeA         ==> TRUE enable  absorption   coefficient computation
  ! computeAbsVolMat ==> TRUE enables volumic absorption computations
  ! absVolMatNum     ==> # of materials for absVol
  ! absVolMat        ==> List of materials for volumic absorption computation
  ! computePoySurf   ==> TRUE enables Poynting flux computation on given surfaces
  ! poySurfNum       ==> # of Poynting surfaces
  ! poySurf          ==> List of the surface tags for Poynting flux computation
  ! computePoy       ==> Specifies the poynting computation ('RT', 'CS', 'OFF')
  ! poyTags(:)       ==> Poynting tags (see mesh_faces)
  ! computeRCS       ==> TRUE enables radar cross-section computations
  ! freqRCSNum       ==> # of frequencies for RCS
  ! freqRCS          ==> List of frequencies   for RCS computation (Hz)
  ! thetaMinRCS      ==> Minimal theta angle   for RCS (degrees)
  ! thetaMaxRCS      ==> Maximal theta angle   for RCS (degrees)
  ! deltaThetaRCS    ==> Theta angle   step    for RCS (degrees)
  ! phiMinRCS        ==> Minimal phi   angle   for RCS (degrees)
  ! phiMaxRCS        ==> Maximal phi   angle   for RCS (degrees)
  ! deltaPhiRCS      ==> Phi   angle   step    for RCS (degrees)
  ! time3dView       ==> TRUE outputs the full 3D time-domain solution
  ! timeMin3dView    ==> Starting time for 3D time-domain solution output (seconds)
  ! timeMax3dView    ==> Finish   time for 3D time-domain solution output (seconds)
  ! n3dView          ==> # of         3D time-domain solution output (seconds)
  ! deltaTime3dView  ==> Time step for 3D time-domain solution output (seconds)
  ! freq3dView       ==> TRUE outputs the full       3D frequency-domain solution
  ! freq3dViewNum    ==> # of frequencies for   3D frequency-domain solution
  ! freq3d           ==> List of frequencies for the 3D frequency solution (Hz)
  ! nProbePts        ==> # of probe points
  ! probePts         ==> Position of probe points
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)                             :: enableObs, computeCsAbs,  &
       computeCsSca, computeCsExt, computeR, computeT, computeA,           &
       computeAbsVolMat, computePoySurf, computeRCS, time3dView,           &
       final3dView, freq3dView, computeEELS

  CHARACTER(LEN=2, KIND=C_CHAR)               :: computePoy

  INTEGER(C_INT)                              :: absVolMatNum, poySurfNum, &
       freqRCSNum, freq3dViewNum, nProbePts, genProbePts, csIndex,         &
       nFreqObs, npoyTag, n3dView
  INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: absVolMat, poySurf, poyTags

  REAL(C_DOUBLE)                              :: freqMinObs, freqMaxObs,   &
       deltaFreqObs, thetaMinRCS, thetaMaxRCS, phiMinRCS, phiMaxRCS,       &
       deltaThetaRCS, deltaPhiRCS, timeMin3dView, timeMax3dView,           &
       deltaTime3dView, deltaDsEELS
  REAL(C_DOUBLE), DIMENSION(:),   ALLOCATABLE :: freqRCS, freq3d
  REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: probePts
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_NUMERICAL_ELEMENT_ARRAY(eltArray, eltArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to numerical element array
    !
    !---- IN
    !---- eltArray     : array of "numerical_element" structure
    !---- eltArraySize : # of elements in eltArray
    !
    !---- OUT
    !---- eltArray     : modified in the process
    !---- eltArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                         INTENT(INOUT) :: eltArraySize

    TYPE(numerical_element_dg), DIMENSION(:), ALLOCATABLE :: eltArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                       :: isArrayFull, isAllocated

    INTEGER(C_INT)                                        :: iElt

    TYPE(numerical_element_dg), DIMENSION(:), ALLOCATABLE :: tempElt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(eltArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(eltArray) < eltArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
          ALLOCATE(tempElt(eltArraySize))
          tempElt = eltArray
          DEALLOCATE(eltArray)
       ENDIF
       !
       ALLOCATE(eltArray(eltArraySize + eltBlockSize))
       DO iElt=eltArraySize+1,eltArraySize+eltBlockSize
          CALL INITIALIZE_NUMERICAL_ELEMENT(eltArray(iElt))
       ENDDO
       !
       eltArray(1:eltArraySize) = tempElt(1:eltArraySize)
       IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
    ENDIF
    !
    eltArraySize = eltArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_NUMERICAL_ELEMENT_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_NUMERICAL_FACE_ARRAY(faceArray, faceArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to numerical element array
    !
    !---- IN
    !---- faceArray     : array of "numerical_face" structure
    !---- faceArraySize : # of faces in faceArray
    !
    !---- OUT
    !---- faceArray     : modified in the process
    !---- faceArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                   INTENT(INOUT) :: faceArraySize

    TYPE(numerical_face), DIMENSION(:), ALLOCATABLE :: faceArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                 :: isArrayFull, isAllocated

    INTEGER(C_INT)                                  :: iFace

    TYPE(numerical_face), DIMENSION(:), ALLOCATABLE :: tempFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(faceArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(faceArray) < faceArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
          ALLOCATE(tempFace(faceArraySize))
          tempFace = faceArray
          DEALLOCATE(faceArray)
       ENDIF
       !
       ALLOCATE(faceArray(faceArraySize + faceBlockSize))
       DO iFace=faceArraySize+1,faceArraySize+faceBlockSize
          CALL INITIALIZE_NUMERICAL_FACE(faceArray(iFace))
       ENDDO
       !
       faceArray(1:faceArraySize) = tempFace(1:faceArraySize)
       IF (ALLOCATED(tempFace)) DEALLOCATE(tempFace)
    ENDIF
    !
    faceArraySize = faceArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_NUMERICAL_FACE_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ALLOCATE_NUMERICAL_ELEMENTS
    !------------------------------------------------------------------
    !---- Allocate EM array and initialize it
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate array
    !------------------------------------------------------------------
    IF (ALLOCATED(dgEM)) DEALLOCATE(dgEM)
    ALLOCATE(dgEM(Mesh%nCell))
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       CALL INITIALIZE_NUMERICAL_ELEMENT(dgEM(iCell))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_NUMERICAL_ELEMENTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ALLOCATE_NUMERICAL_FACES
    !------------------------------------------------------------------
    !---- Allocate EMf array and initialize it
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate array
    !------------------------------------------------------------------
    IF (ALLOCATED(EMf)) DEALLOCATE(EMf)
    ALLOCATE(EMf(Mesh%nFac))
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFac
       CALL INITIALIZE_NUMERICAL_FACE(EMf(iFac))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_NUMERICAL_FACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_NUMERICAL_ELEMENT(elt)
    !------------------------------------------------------------------
    !---- Initialize a numerical element
    !
    !---- IN
    !---- elt : numerical element
    !
    !---- OUT
    !---- elt : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(numerical_element_dg), INTENT(INOUT) :: elt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    elt%isPml                =  .FALSE.
    elt%isDisp               =  .FALSE.
    elt%isGenDisp            =  .FALSE.
    elt%isNL                 =  .FALSE.
    elt%pmlCellNum           =  0
    elt%nDofS                =  0
    elt%nDofV                =  0
    elt%order                =  0
    elt%iTot                 =  0
    elt%iGen                 =  0
    elt%iNL                  =  0
    elt%dJac                 =  0.0
    elt%mat                  => NULL()
    elt%iJac                 =  0.0
    elt%massMtx              => NULL()
    elt%iMassMtx             => NULL()
    elt%xStfMtx              => NULL()
    elt%yStfMtx              => NULL()
    elt%zStfMtx              => NULL()
    elt%MATVECMUL_SRF        => NULL()
    elt%MATVECMUL_VLM        => NULL()
    elt%COMPUTE_VOLUMIC_RHS  => NULL()
    elt%COMPUTE_FLUX         => NULL()
    elt%COMPUTE_INC_FLUX     => NULL()
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_NUMERICAL_ELEMENT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_NUMERICAL_FACE(face)
    !------------------------------------------------------------------
    !---- Initialize a numerical face
    !
    !---- IN
    !---- face : numerical face
    !
    !---- OUT
    !---- face : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(numerical_face), INTENT(INOUT) :: face
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    face%nDofS = 0
    face%order = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_NUMERICAL_FACE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ALLOCATE_NUMERICAL_ELEMENTS_FROM_MESH_TYPES
    !------------------------------------------------------------------
    !---- Allocate EM array and initialize them based on mesh types
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate array
    !------------------------------------------------------------------
    IF (ALLOCATED(dgEM)) DEALLOCATE(dgEM)
    ALLOCATE(dgEM(Mesh%nCell))
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       CALL INITIALIZE_NUMERICAL_ELEMENT_FROM_MESH_TYPE(&
            MeshCell(iCell), dgEM(iCell))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_NUMERICAL_ELEMENTS_FROM_MESH_TYPES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ALLOCATE_NUMERICAL_FACES_FROM_MESH_TYPES
    !------------------------------------------------------------------
    !---- Allocate EMf array and initialize them based on mesh types
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFac
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate array
    !------------------------------------------------------------------
    IF (ALLOCATED(EMf)) DEALLOCATE(EMf)
    ALLOCATE(EMf(Mesh%nFac))
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFac
       CALL INITIALIZE_NUMERICAL_FACE_FROM_MESH_TYPE(MeshFac(iFac), EMf(iFac))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_NUMERICAL_FACES_FROM_MESH_TYPES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_NUMERICAL_ELEMENT_FROM_MESH_TYPE(meshElt, numElt)
    !------------------------------------------------------------------
    !---- Initialize a numerical element based on mesh type
    !
    !---- IN
    !---- meshElt : mesh element
    !---- numElt  : numerical element
    !
    !---- OUT
    !---- numElt  : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(cell),                 INTENT(IN)    :: meshElt

    TYPE(numerical_element_dg), INTENT(INOUT) :: numElt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    CALL INITIALIZE_NUMERICAL_ELEMENT(numElt)
    !
    numElt%nDofV = meshElt%nVerts
    numElt%order = meshElt%geoOrder
    !
    ALLOCATE(numElt%E%vec(3,numElt%nDofV))
    ALLOCATE(numElt%H%vec(3,numElt%nDofV))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_NUMERICAL_ELEMENT_FROM_MESH_TYPE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_NUMERICAL_FACE_FROM_MESH_TYPE(meshFace, numFace)
    !------------------------------------------------------------------
    !---- Initialize a numerical face based on mesh type
    !
    !---- IN
    !---- meshFace : mesh face
    !---- numFace  : numerical face
    !
    !---- OUT
    !---- numFace  : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_FACES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(face),           INTENT(IN)    :: meshFace

    TYPE(numerical_face), INTENT(INOUT) :: numFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    CALL INITIALIZE_NUMERICAL_FACE(numFace)
    !
    numFace%nDofS = meshFace%nVerts
    numFace%order = meshFace%geoOrder
    !
    ALLOCATE(numFace%E%vec(3,numFace%nDofS))
    ALLOCATE(numFace%H%vec(3,numFace%nDofS))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_NUMERICAL_FACE_FROM_MESH_TYPE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_DISPERSIVE_ELEMENT_ARRAY(eltArray, eltArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to dispersive element array
    !
    !---- IN
    !---- eltArray     : array of "dispersive_element" structure
    !---- eltArraySize : # of elements in eltArray
    !
    !---- OUT
    !---- eltArray     : modified in the process
    !---- eltArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                       INTENT(INOUT) :: eltArraySize

    TYPE(dispersive_element), DIMENSION(:), ALLOCATABLE :: eltArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                     :: isArrayFull, isAllocated

    INTEGER(C_INT)                                      :: iElt

    TYPE(dispersive_element), DIMENSION(:), ALLOCATABLE :: tempElt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(eltArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(eltArray) < eltArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
          ALLOCATE(tempElt(eltArraySize))
          tempElt = eltArray
          DEALLOCATE(eltArray)
       ENDIF
       !
       ALLOCATE(eltArray(eltArraySize + eltBlockSize))
       DO iElt=eltArraySize+1,eltArraySize+eltBlockSize
          CALL INITIALIZE_DISPERSIVE_ELEMENT(eltArray(iElt))
       ENDDO
       !
       eltArray(1:eltArraySize) = tempElt(1:eltArraySize)
       IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
    ENDIF
    !
    eltArraySize = eltArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_DISPERSIVE_ELEMENT_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_DISPERSIVE_ELEMENT(elt)
    !------------------------------------------------------------------
    !---- Initialize a dispersive element
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(dispersive_element), INTENT(INOUT) :: elt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    elt%cellNum                   =  0
    elt%ptr                       => NULL()
    elt%mat                       => NULL()
    elt%matDisp                   => NULL()
    elt%COMPUTE_RHS_DISP_CURRENTS => NULL()
    elt%ADD_RHS_DISP_CURRENTS     => NULL()
    elt%COMPUTE_VOLUMIC_RHS       => NULL()
    elt%RHS_UPDATE                => NULL()
    elt%COMPUTE_J_FLUX            => NULL()
    elt%COMPUTE_Q_FLUX            => NULL()
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_DISPERSIVE_ELEMENT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_DERIVED_DISPERSIVE_ELEMENT_ARRAY(eltArray, eltArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to derived dispersive element array
    !
    !---- IN
    !---- eltArray     : array of "derived_dispersive_element" structure
    !---- eltArraySize : # of elements in eltArray
    !
    !---- OUT
    !---- eltArray     : modified in the process
    !---- eltArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),                               INTENT(INOUT) :: eltArraySize

    TYPE(derived_dispersive_element), DIMENSION(:), ALLOCATABLE :: eltArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                                             :: isArrayFull, isAllocated

    INTEGER(C_INT)                                              :: iElt

    TYPE(derived_dispersive_element), DIMENSION(:), ALLOCATABLE :: tempElt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(eltArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(eltArray) < eltArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
          ALLOCATE(tempElt(eltArraySize))
          tempElt = eltArray
          DEALLOCATE(eltArray)
       ENDIF
       !
       ALLOCATE(eltArray(eltArraySize + eltBlockSize))
       DO iElt=eltArraySize+1,eltArraySize+eltBlockSize
          CALL INITIALIZE_DERIVED_DISPERSIVE_ELEMENT(eltArray(iElt))
       ENDDO
       !
       eltArray(1:eltArraySize) = tempElt(1:eltArraySize)
       IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
    ENDIF
    !
    eltArraySize = eltArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_DERIVED_DISPERSIVE_ELEMENT_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_DERIVED_DISPERSIVE_ELEMENT(elt)
    !------------------------------------------------------------------
    !---- Initialize a derived dispersive element
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(derived_dispersive_element), INTENT(INOUT) :: elt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    elt%iDispCell = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_DERIVED_DISPERSIVE_ELEMENT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EXTEND_SF_ELEMENT_ARRAY(eltArray, eltArraySize)
    !------------------------------------------------------------------
    !---- Add a new slot to SF element array
    !
    !---- IN
    !---- eltArray     : array of "SF_element" structure
    !---- eltArraySize : # of elements in eltArray
    !
    !---- OUT
    !---- eltArray     : modified in the process
    !---- eltArraySize : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(INOUT) :: eltArraySize

    TYPE(sf_element), DIMENSION(:), ALLOCATABLE :: eltArray
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    LOGICAL(C_BOOL)                             :: isArrayFull, isAllocated

    INTEGER(C_INT)                              :: iElt

    TYPE(sf_element), DIMENSION(:), ALLOCATABLE :: tempElt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Check if cell array is full
    !------------------------------------------------------------------
    isArrayFull = .FALSE.
    isAllocated = ALLOCATED(eltArray)
    !
    IF (isAllocated) THEN
       IF (SIZE(eltArray) < eltArraySize + 1) isArrayFull = .TRUE.
    ELSE
       isArrayFull = .TRUE.
    ENDIF
    !------------------------------------------------------------------
    !---- Extend array if necessary
    !------------------------------------------------------------------
    IF (isArrayFull) THEN
       IF (isAllocated) THEN
          IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
          ALLOCATE(tempElt(eltArraySize))
          tempElt = eltArray
          DEALLOCATE(eltArray)
       ENDIF
       !
       ALLOCATE(eltArray(eltArraySize + eltBlockSize))
       DO iElt=eltArraySize+1,eltArraySize+eltBlockSize
          CALL INITIALIZE_SF_ELEMENT(eltArray(iElt))
       ENDDO
       !
       eltArray(1:eltArraySize) = tempElt(1:eltArraySize)
       IF (ALLOCATED(tempElt)) DEALLOCATE(tempElt)
    ENDIF
    !
    eltArraySize = eltArraySize + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EXTEND_SF_ELEMENT_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_SF_ELEMENT(elt)
    !------------------------------------------------------------------
    !---- Initialize a derived dispersive element
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(sf_element), INTENT(INOUT) :: elt
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize
    !------------------------------------------------------------------
    elt%cellNum = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_SF_ELEMENT

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE NUMERICAL_TYPES
