MODULE PARALLEL_STRUCTURES

  !====================================================================
  ! Parallel routines for COMMON library
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------
  INCLUDE 'mpif.h'
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Structure of a neighbor processor
  !
  ! isExchangeCompleted ==> TRUE if exchange with neighbor is done
  ! isSameLog           ==> TRUE if material log is identical on
  !                         both sides of the face
  ! id                  ==> neighbor id
  ! nbFac               ==> number of faces shared by current proc
  !                         with this neighbor
  ! exchangeSize        ==> number of dof exchanged between procs
  ! *Request            ==> send and receive request numbers
  ! *Code               ==> send and receive codes
  ! faces               ==> list of faces shared with this neighbor
  ! perm                ==> list of permutations for faces shared
  !                         with this neighbor
  !--------------------------------------------------------------------
  TYPE neighbor_proc
     LOGICAL(C_BOOL)                            :: isExchangeCompleted
     LOGICAL(C_BOOL), DIMENSION(:), ALLOCATABLE :: isSameLog
     
     INTEGER(C_INT)                             :: id, nbFac,  &
          exchangeSize, receiveCode, receiveRequest, sendCode, &
          sendRequest
     INTEGER(C_INT),  DIMENSION(:), ALLOCATABLE :: faces, perm, logs
  END TYPE neighbor_proc

  !--------------------------------------------------------------------
  ! Data used in parallel environment
  !--------------------------------------------------------------------
  ! parallelRun      ==> TRUE if run is parallel
  ! idWrite          ==> TRUE if local process is supposed to write
  ! defaultComm      ==> Default MPI communicator
  ! nbProc           ==> Effective number of processes
  ! idProc           ==> Process identifier
  ! mpiError         ==> MPI error code
  ! nbNeighb         ==> Number of neighbors for the current submesh
  ! neighbors        ==> data structure accounting for neighbors procs
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)           :: parallelRun = .FALSE.
  LOGICAL(C_BOOL)           :: idWrite     = .TRUE.

  INTEGER(C_INT), PARAMETER :: defaultComm = MPI_COMM_WORLD
  INTEGER(C_INT), PARAMETER :: mainProc    = 0
  INTEGER(C_INT)            :: nbProc      = 1
  INTEGER(C_INT)            :: idProc      = 0
  INTEGER(C_INT)            :: mpiError, nbNeighb, maxBufferSize, &
       status(MPI_STATUS_SIZE)  

  !--------------------------------------------------------------------
  ! Array of neighbor procs
  !--------------------------------------------------------------------
  TYPE(neighbor_proc), DIMENSION(:), ALLOCATABLE :: neighb
  
  !--------------------------------------------------------------------
  ! Interface for overloading global operation on all subdomains
  !--------------------------------------------------------------------
  INTERFACE GLOBAL_OP
     MODULE PROCEDURE GLOBAL_OP_INT, GLOBAL_OP_DOUBLE, &
          GLOBAL_OP_ARRAY_INT, GLOBAL_OP_ARRAY_DOUBLE
  END INTERFACE GLOBAL_OP

  !--------------------------------------------------------------------
  ! Interface for overloading reduction operation on main proc
  !--------------------------------------------------------------------
  INTERFACE REDUCE_OP
     MODULE PROCEDURE REDUCE_OP_INT, REDUCE_OP_DOUBLE, &
          REDUCE_OP_ARRAY_INT, REDUCE_OP_ARRAY_DOUBLE
  END INTERFACE REDUCE_OP
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PARALLEL_INITIALIZE
    !------------------------------------------------------------------
    !---- Initialize parallel environment and creates default communicator
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize parallel environment
    !------------------------------------------------------------------
    CALL MPI_INIT     (mpiError)
    CALL MPI_COMM_RANK(defaultComm,idProc,mpiError)
    CALL MPI_COMM_SIZE(defaultComm,nbProc,mpiError)
    parallelRun                     = .TRUE.
    IF (nbProc == 1)    parallelRun = .FALSE.
    !
    !---- Set only one proc for writing
    idWrite                         = .TRUE.
    IF (idProc /= mainProc) idWrite = .FALSE.
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PARALLEL_INITIALIZE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PARALLEL_FINALIZE
    !------------------------------------------------------------------
    !---- Finalize parallel environment
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Finalize parallel environment
    !------------------------------------------------------------------
    CALL MPI_FINALIZE(mpiError)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PARALLEL_FINALIZE
  
  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CHECK_MPI_PARALLEL_MESH
    !------------------------------------------------------------------
    !---- Handle global mesh and parallel mesh names
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=10,   KIND=C_CHAR) :: cIdProc
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readname, extension

    LOGICAL(C_BOOL)                  :: fileExists
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Proceed to adjustements for parallel compatibility
    !------------------------------------------------------------------
    !---- globalMesh is used in any case
    CALL INITIALIZE_MESH_STRUCTURE(globalMesh)
    globalMesh%name = Mesh%name
    CALL REMOVE_EXTENSION(globalMesh%name, extension)
    !
    !---- Check if this is a parallel run
    IF (.NOT. parallelRun) RETURN
    !
    !---- Modify mesh name to subMesh name
    IF (nbProc > 1) THEN
       !---- The local mesh structure takes the name of the submesh file
       !---- WARNING : medit binary extension is used in any case
       CALL INT_TO_CHAR(idProc+1,cIdProc)
       readname  = Mesh%name
       CALL REMOVE_EXTENSION(readname, extension)
       Mesh%name = 'subMesh_'//TRIM(readname)//'_'//TRIM(cIdProc)//'.mesh'
    ENDIF
    !
    IF (idWrite) WRITE(6, '(a,i0,a)') 'MPI parallel run with ',nbProc,' cores'    
    !------------------------------------------------------------------
    !---- Check that all subMeshes exist
    !------------------------------------------------------------------
    CALL DOES_FILE_EXIST(Mesh%name, fileExists)
    IF (.NOT. fileExists) THEN
       WRITE(6, '(a)')      'Error in CHECK_MPI_PARALLEL_MESH'
       WRITE(6, '(3a)')     'Mesh file ',TRIM(Mesh%name),' not found '
       WRITE(6, '(a,i0,a)') 'You asked for a parallel run on ',nbProc,' cores'
       WRITE(6, '(a)')      'It seems that I could not find some of the subMesh files'
       WRITE(6, '(a)')      'Please consider re-partitionning your mesh with the &
            right number of subdomains'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CHECK_MPI_PARALLEL_MESH

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_PARALLEL_INFORMATION
    !------------------------------------------------------------------
    !---- Read additional parallel information files
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Shared modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_FIND_FACES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100,  KIND=C_CHAR) :: keyword, cIdProc
    CHARACTER(LEN=1000, KIND=C_CHAR) :: fileName

    INTEGER(C_INT)                   :: nUnit, iNeighb, nbFac, iFac, &
         faceNum, nVerts, neighbLog, iCell, iVert, foundVert
    
    REAL(C_DOUBLE), DIMENSION(3)     :: vert
    INTEGER(C_INT), DIMENSION(3)     :: verts, readVerts
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. parallelRun) RETURN
    !------------------------------------------------------------------
    !---- Read additional informations
    !------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    CALL BUILD_SUBINFO_NAME(globalMesh%name,idProc+1,fileName)
    OPEN(UNIT=nUnit, FILE=fileName)
    !
    !---- Start reading
    Readingloop: DO
       READ(nUnit, *) keyword 
       SELECT CASE (trim(keyword))
          !
          !---- Read number of neighbors
       CASE('#nNeighbors')
          READ(nUnit,*) nbNeighb
          !
          !---- Allocate neighbor structures
          ALLOCATE(neighb(nbNeighb))
          !
          !---- Read list of neighbors
       CASE('#neighbors')
          DO iNeighb=1,nbNeighb
             READ(nUnit,*) neighb(iNeighb)%id
             neighb(iNeighb)%id = neighb(iNeighb)%id - 1
          ENDDO
          !
          !---- Read number of common faces with each neighbor
       CASE('#nFacesPerNeighbor')
          DO iNeighb=1,nbNeighb
             READ(nUnit,*) nbFac
             neighb(iNeighb)%nbFac = nbFac
             !
             ALLOCATE(neighb(iNeighb)%faces    (nbFac))
             ALLOCATE(neighb(iNeighb)%perm     (nbFac))
             ALLOCATE(neighb(iNeighb)%logs     (nbFac))
             ALLOCATE(neighb(iNeighb)%isSameLog(nbFac))
          ENDDO
          !
          !---- Read vertices of the common faces with each neighbor
          !---- and find the associated face
       CASE('#facesPerNeighbor')
          DO iNeighb=1,nbNeighb
             DO iFac=1,neighb(iNeighb)%nbFac
                READ(nUnit,*) nVerts
                READ(nUnit,*) readVerts
                !
                CALL FIND_DERIVED_FACE_IN_LIST_FROM_VERTICES(readVerts(1:nVerts), &
                     MeshFacArt, Mesh%nFacArt, faceNum)
                !
                IF (faceNum == 0) THEN
                   WRITE(6,'(a)') 'Error in READ_PARALLEL_INFORMATION'
                   WRITE(6,'(a)') 'I could not find some of the artificial faces'
                   CALL TILT
                ENDIF
                !
                neighb(iNeighb)%faces(iFac) = MeshFacArt(faceNum)%faceNum
             ENDDO
          ENDDO
          !
          !---- Read list of permutations with each neighbor
       CASE('#facesPermPerNeighbor')
          DO iNeighb=1,nbNeighb
             DO iFac=1,neighb(iNeighb)%nbFac
                READ(nUnit,*) neighb(iNeighb)%perm(iFac)
             ENDDO
          ENDDO
          !
          !---- Read list of cell logs for each neighbor
       CASE('#cellLogsPerNeighbor')
          DO iNeighb=1,nbNeighb
             DO iFac=1,neighb(iNeighb)%nbFac
                READ(nUnit,*) neighb(iNeighb)%logs(iFac)
             ENDDO
          ENDDO
          !
          !---- Read global neighbors numbering for all cells
       CASE('#globalCellNeighbors')
          DO iCell=1,Mesh%nCell
             ALLOCATE(MeshCell(iCell)%globalNeighb(MeshCell(iCell)%nNeighb))
             READ(nUnit,*) MeshCell(iCell)%globalIndex, MeshCell(iCell)%globalNeighb(:)
          ENDDO
          !
          !---- End of file
       CASE('#End')
          EXIT Readingloop
       CASE DEFAULT
       END SELECT
    ENDDO Readingloop
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Process isSameLog array
    !------------------------------------------------------------------
    DO iNeighb=1,nbNeighb
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum                         = neighb(iNeighb)%faces(iFac)
          neighbLog                       = neighb(iNeighb)%logs(iFac)
          iCell                           = MeshFac(faceNum)%neighb(1)
          neighb(iNeighb)%isSameLog(iFac) = (MeshCell(iCell)%log == neighbLog)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
    IF (idWrite) WRITE(6,'(a)') 'Finished reading parallel information'
    !------------------------------------------------------------------
  END SUBROUTINE READ_PARALLEL_INFORMATION
  
  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE GLOBAL_OP_INT(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform operation on int over all procs
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: localValue, operation
    INTEGER(C_INT), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF    
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    !
    CALL MPI_ALLREDUCE(localValue, globalValue, 1, &
         MPI_INTEGER, operation, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE GLOBAL_OP_INT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE GLOBAL_OP_ARRAY_INT(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform operation on int array over all procs
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: operation
    INTEGER(C_INT), DIMENSION(:), INTENT(IN)  :: localValue
    INTEGER(C_INT), DIMENSION(:), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                            :: length
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    length      = SIZE(localValue)
    !
    CALL MPI_ALLREDUCE(localValue, globalValue, length, &
         MPI_INTEGER, operation, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE GLOBAL_OP_ARRAY_INT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE GLOBAL_OP_COMPLEX(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform operation on complex over all procs
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: operation

    COMPLEX(C_DOUBLE_COMPLEX), INTENT(IN)  :: localValue
    COMPLEX(C_DOUBLE_COMPLEX), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = (0.0,0.0)
    !
    CALL MPI_ALLREDUCE(localValue, globalValue, 1, &
         MPI_DOUBLE_COMPLEX, operation, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE GLOBAL_OP_COMPLEX


  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE GLOBAL_OP_DOUBLE(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform operation on double over all procs
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: operation

    REAL(C_DOUBLE), INTENT(IN)  :: localValue
    REAL(C_DOUBLE), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    !
    CALL MPI_ALLREDUCE(localValue, globalValue, 1, &
         MPI_DOUBLE_PRECISION, operation, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE GLOBAL_OP_DOUBLE

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE GLOBAL_OP_ARRAY_DOUBLE(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform operation on double array over all procs
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: operation

    REAL(C_DOUBLE), DIMENSION(:), INTENT(IN)  :: localValue
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                            :: length
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    length      = SIZE(localValue)
    !
    CALL MPI_ALLREDUCE(localValue, globalValue, length, &
         MPI_DOUBLE_PRECISION, operation, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE GLOBAL_OP_ARRAY_DOUBLE

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE GLOBAL_OP_ARRAY_COMPLEX(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform operation on complex array over all procs
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: operation

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), INTENT(IN)  :: localValue
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                            :: length
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = (0.0,0.0)
    length      = SIZE(localValue)
    !
    CALL MPI_ALLREDUCE(localValue, globalValue, length, &
         MPI_DOUBLE_COMPLEX, operation, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE GLOBAL_OP_ARRAY_COMPLEX

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE REDUCE_OP_INT(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform reduce operation on int to main proc
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: localValue, operation
    INTEGER(C_INT), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    !
    CALL MPI_REDUCE(localValue, globalValue, 1, &
         MPI_INTEGER, operation, mainProc, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE REDUCE_OP_INT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE REDUCE_OP_ARRAY_INT(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform reduce operation on int array to main proc
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: operation
    INTEGER(C_INT), DIMENSION(:), INTENT(IN)  :: localValue
    INTEGER(C_INT), DIMENSION(:), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                            :: length
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    length      = SIZE(localValue)
    !
    CALL MPI_REDUCE(localValue, globalValue, length, &
         MPI_INTEGER, operation, mainProc, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE REDUCE_OP_ARRAY_INT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE REDUCE_OP_DOUBLE(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform reduce operation on double to main proc
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: operation

    REAL(C_DOUBLE), INTENT(IN)  :: localValue
    REAL(C_DOUBLE), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    !
    CALL MPI_REDUCE(localValue, globalValue, 1, &
         MPI_DOUBLE_PRECISION, operation, mainProc, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE REDUCE_OP_DOUBLE

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE REDUCE_OP_ARRAY_DOUBLE(localValue, operation, globalValue)
    !--------------------------------------------------------------------
    !---- Perform reduce operation on double array to main proc
    !
    !---- IN
    !---- localValue  : value on local processor
    !---- operation   : operation to perform
    !
    !---- OUT
    !---- globalValue : value summed on all processors
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),               INTENT(IN)  :: operation

    REAL(C_DOUBLE), DIMENSION(:), INTENT(IN)  :: localValue
    REAL(C_DOUBLE), DIMENSION(:), INTENT(OUT) :: globalValue
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                            :: length
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       globalValue = localValue
       RETURN
    ENDIF
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    globalValue = 0
    length      = SIZE(localValue)
    !
    CALL MPI_REDUCE(localValue, globalValue, length, &
         MPI_DOUBLE_PRECISION, operation, mainProc, defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE REDUCE_OP_ARRAY_DOUBLE  

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PARALLEL_BARRIER
    !--------------------------------------------------------------------
    !---- MPI_BARRIER operation
    !
    !---- IN
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Shared modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (.NOT. parallelRun) RETURN
    !--------------------------------------------------------------------
    !---- Perform sum
    !--------------------------------------------------------------------
    CALL MPI_BARRIER(defaultComm, mpiError)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PARALLEL_BARRIER

  !======================================================================
  !======================================================================
  !======================================================================

END MODULE PARALLEL_STRUCTURES
