MODULE PLANE_WAVE

  !====================================================================
  ! Routines related to the plane wave case
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! PLANE WAVE DATA ('PW')
  !--------------------------------------------------------------------
  ! computeErrPW     ==> TRUE enables the computation of the overall L2 error
  ! monoPW           ==> TRUE if PW is monochromatic
  ! thetaPW          ==> Theta angle for plane wave (degrees)
  ! phiPW            ==> Phi   angle for plane wave (degrees)
  ! ampliPW          ==> Amplitude for plane wave (V.m-1)
  ! retardPW         ==> Retardation for plane wave (seconds)
  ! freqMinPW        ==> Minimal frequency for plane wave (Hz)
  ! freqMaxPW        ==> Maximal frequency for plane wave (Hz)
  ! freqPW           ==> Mean of freqMinPW and freqMaxPW
  ! omegaMinPW       ==> Minimal angular frequency for plane wave (Hz)
  ! omegaMaxPW       ==> Maximal angular frequency for plane wave (Hz)
  ! omegaPW          ==> Mean of omegaMinPW and omegaMaxPW
  ! sigmaPW          ==> Control parameter for the gaussian exponential,
  !                      computed from freqMinPW and freqMaxPW
  ! polarPW          ==> Polarization vector for plane wave
  ! waveVecPW        ==> Wave vector, computed from thetaPW and phiPW
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)              :: computeErrPW, monoPW

  REAL(C_DOUBLE)               :: thetaPW, phiPW, ampliPW, retardPW,  &
       sigmaPW, freqMinPW, freqMaxPW, freqPW, omegaMinPW, omegaMaxPW, &
       omegaPW
  REAL(C_DOUBLE), DIMENSION(3) :: polarPW, waveVecPW  

  !--------------------------------------------------------------------
  ! Threshold for the computation of the adequate gaussian parameter
  !--------------------------------------------------------------------
  REAL(C_DOUBLE), PARAMETER :: epsGaussian = 1.0e-1
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_PLANE_WAVE
    !------------------------------------------------------------------
    !---- Pre-processing for plane wave case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES    
    USE MATH_PHYS_CONSTANTS
    USE SPECIAL_FUNCTIONS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: dProd, norm
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Handle waveVector and polarization
    !------------------------------------------------------------------    
    waveVecPW(:) = 0.0
    waveVecPW(1) = SIN(thetaPW)*COS(phiPW)
    waveVecPW(2) = SIN(thetaPW)*SIN(phiPW)
    waveVecPW(3) = COS(thetaPW)
    !
    CALL NORMALIZE_VECTOR(polarPW,norm)    
    !------------------------------------------------------------------
    !---- Handle monochromatic/polychromatic
    !------------------------------------------------------------------    
    IF (freqMinPW == freqMaxPW) monoPW = .TRUE.
    !
    !---- Monochromatic
    IF (monoPW) THEN
       freqPW     = freqMinPW       
       CALL FREQUENCY_TO_INVERSE_METERS(freqPW)
       omegaPW    = 2.0*Pi*freqPW
       sigmaPW    = 1.0e+10
       retardPW   = 0.0
    ENDIF
    !
    !---- Polychromatic
    IF (.NOT. monoPW) THEN
       CALL FREQUENCY_TO_INVERSE_METERS(freqMinPW)
       CALL FREQUENCY_TO_INVERSE_METERS(freqMaxPW)
       freqPW     = 0.5*(freqMinPW+freqMaxPW)
       omegaPW    = 2.0*Pi*freqPW
       omegaMinPW = 2.0*Pi*freqMinPW
       omegaMaxPW = 2.0*Pi*freqMaxPW
       sigmaPW    = SQRT(2.0*LOG(1.0/epsGaussian))/ABS(omegaPW-omegaMinPW)
       CALL SECONDS_TO_METERS(retardPW)
       !
       !---- Test if retardPW is sufficiently large
       IF (retardPW < 5.0*sigmaPW) THEN
          IF(idWrite) WRITE(6, '(a)')       'WARNING from PREPROCESS_PLANE_WAVE'
          IF(idWrite) WRITE(6, '(a)')       'The value you took for retardPW might be too small'
          IF(idWrite) WRITE(6, '(a)')       'A minimal value is 5*sigmaPW'
          IF(idWrite) WRITE(6, '(a,e12.5)') 'Your value is ', retardPW/c0
          IF(idWrite) WRITE(6, '(a,e12.5)') '5*sigmaPW  is ', 5.0*sigmaPW/c0
       ENDIF
    ENDIF
    !------------------------------------------------------------------    
    !---- Test compatibility of waveVecPW and polarPW
    !------------------------------------------------------------------    
    dProd = DOT_PRODUCT(waveVecPW,polarPW)
    IF (ABS(dProd) > absoluteEps) THEN
       IF(idWrite) WRITE(6, '(a)')        'Your choices of wave-vector and &
            polarization are incompatible'
       IF(idWrite) WRITE(6, '(a)')        'As a reminder, the later two must &
            be orthogonal for a plane wave'
       IF(idWrite) WRITE(6, '(a)')        'I will assume you made a wise choice of &
            waveVector and I will modify the polarization accordingly'
       IF(idWrite) WRITE(6, '(a,3e12.5)') 'Wave-vector is       ',waveVecPW       
       IF(idWrite) WRITE(6, '(a,3e12.5)') 'Old polarization was ',polarPW
       !
       CALL FIND_NEW_POLARIZATION(waveVecPW,polarPW)
       !
       IF(idWrite) WRITE(6, '(a,3e12.5)') 'New polarization is  ',polarPW
       IF(idWrite) WRITE(6, '(a)')        'Proceed with caution when using this solution'
       IF(idWrite) WRITE(6, '(a)')        'Please refer to manual for more informations'
    ENDIF
    !------------------------------------------------------------------
    !---- Print sum-up
    !------------------------------------------------------------------
    IF(idWrite) WRITE(6, '(a)')          'Source informations'
    IF(idWrite) WRITE(6, 10)
    IF(idWrite) WRITE(6, '(a)')          'Running case: plane wave'
    IF(idWrite) WRITE(6, '(a,e12.5)')    'Theta angle (<z,x>) in degrees : ',thetaPW
    IF(idWrite) WRITE(6, '(a,e12.5)')    'Phi   angle (<x,y>) in degrees : ',phiPW      
    IF(idWrite) WRITE(6, '(a,e12.5)')    'Central frequency in Hz        : ',freqPW*c0
    IF(idWrite) WRITE(6, '(a,e12.5)')    'Central pulsation in Hz        : ',omegaPW*c0
    IF(idWrite) WRITE(6, '(a,3e12.5)')   'Wave-vector                    : ',waveVecPW
    IF(idWrite) WRITE(6, '(a,3e12.5)')   'Polarization                   : ',polarPW
    !
    IF (.NOT. monoPW) THEN
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Min frequency in Hz            : ',freqMinPW*c0
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Max frequency in Hz            : ',freqMaxPW*c0
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Min pulsation in Hz            : ',omegaMinPW*c0
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Max pulsation in Hz            : ',omegaMaxPW*c0
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Retardation in seconds         : ',retardPW/c0
       IF(idWrite) WRITE(6, '(a,e12.5)') 'Gaussian parameter in seconds  : ',sigmaPW/c0
    ENDIF
    !
    IF(idWrite) WRITE(6, 10)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_PLANE_WAVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FIND_NEW_POLARIZATION(waveVector,polar)
    !------------------------------------------------------------------
    !---- Find a suitable polarization given the old polarization and
    !---- the wave-vector
    !
    !---- IN
    !---- waveVector : wave-vector of the plane wave
    !---- polar      : incorrect polarization 
    !
    !---- OUT
    !---- polar      : modified in the process
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    USE MATH_PHYS_CONSTANTS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3), INTENT(INOUT) :: polar
    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)    :: waveVector
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: errPolar, norm
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute new waveVector
    !------------------------------------------------------------------    
    errPolar = DOT_PRODUCT(polar,waveVector)
    polar    = polar - errPolar*waveVector
    CALL NORMALIZE_VECTOR(polar,norm)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF (DOT_PRODUCT(waveVector,polar) > absoluteEps) THEN
       IF(idWrite) WRITE(6, '(a)') 'Error in FIND_NEW_POLARIZATION'
       IF(idWrite) WRITE(6, '(a)') 'Your wave-vector and polarization are incompatible'
       IF(idWrite) WRITE(6, '(a)') 'I tried to correct the polarization, but &
            apparently I failed'
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE FIND_NEW_POLARIZATION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_PLANE_WAVE_TIME(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the plane wave at the dof of a cell 
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_PLANE_WAVE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_PLANE_WAVE_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_DX_SOL_PLANE_WAVE(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Spatial x derivative of the solution of the plane wave at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_DX_PLANE_WAVE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_DX_SOL_PLANE_WAVE
 
  !====================================================================
  !====================================================================
  !====================================================================


  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_DY_SOL_PLANE_WAVE(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Spatial y derivative of the solution of the plane wave at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_DY_PLANE_WAVE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_DY_SOL_PLANE_WAVE
 
  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_DZ_SOL_PLANE_WAVE(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Spatial z derivative of the solution of the plane wave at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_DZ_PLANE_WAVE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_DZ_SOL_PLANE_WAVE
 
  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_PLANE_WAVE_FREQ(iCell, nDofV, order, omegaValue)
    !------------------------------------------------------------------
    !---- Solution of the plane wave at the dof of a cell
    !
    !---- IN
    !---- iCell      : number of the cell
    !---- nDofV      : number of dof in the cell
    !---- order      : polynomial order in the cell    
    !---- omegaValue : angular frequency
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: iCell, nDofV, order

    REAL(C_DOUBLE), INTENT(IN)  :: omegaValue
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)              :: iDof

    COMPLEX(C_DOUBLE_COMPLEX)   :: solE(3), solH(3)
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV,                             &
         dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL FREQ_SOL_PLANE_WAVE(omegaValue,                            &
            dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExVcx(iDof) = solE(1)
       EyVcx(iDof) = solE(2)
       EzVcx(iDof) = solE(3)
       HxVcx(iDof) = solH(1)
       HyVcx(iDof) = solH(2)
       HzVcx(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_PLANE_WAVE_FREQ

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_PLANE_WAVE_TIME_SF(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the plane wave at the dof of a cell including 
    !---- the attenuation due to the Pml
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE)               :: attenuation, kTimesR, pmlDepth, &
         factor1, factor2

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH, r
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_PLANE_WAVE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       !
       !---- E(x,y,z,omega)_pml = E0 exp(-i*omega*t) *
       !----                      exp(i*k*r*omega*s_pml)
       !---- Approximation: pmlKappa >> pmlAlpha*pmlSigma/(pmlAlpha**2 + omega**2)
       !---- E(x,y,z,omega)_pml ~ E0 exp(-i*omega*t) *
       !----                      exp(i*k*r*omega*pmlKappa) * 
       !----                      exp(-k*r*(pmlSigma*omega**2)/(pmlAlpha**2 + omega**2)
       !---- Approximation: Set omega in damping term to omegaPW
       !---- E(x,y,z,t)_pml ~ E(x,y,z,t) * exp(-k*r*(pmlSigma*omegaPW**2)/(pmlAlpha**2 + omegaPW**2)
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_PLANE_WAVE_TIME_SF

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FACE_SOL_PLANE_WAVE_TIME(iFace, nDofS, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the plane wave at the dof of a face
    !
    !---- IN
    !---- iFace : number of the cell
    !---- nDofS : number of dof in the face
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iFace, nDofS, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_FACE_DOF(iFace, order, nDofS, nDofV, dofXS(1:nDofS), dofYS(1:nDofS), dofZS(1:nDofS))
    !
    DO iDof=1,nDofS
       CALL TIME_SOL_PLANE_WAVE(time, dofXS(iDof), dofYS(iDof), dofZS(iDof), solE, solH)
       !
       ExS(iDof) = solE(1)
       EyS(iDof) = solE(2)
       EzS(iDof) = solE(3)
       HxS(iDof) = solH(1)
       HyS(iDof) = solH(2)
       HzS(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FACE_SOL_PLANE_WAVE_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FACE_SOL_PLANE_WAVE_FREQ(iFac, nDofS, nDofV, order, omegaValue)
    !------------------------------------------------------------------
    !---- Solution of the plane wave at the dof of a face
    !
    !---- IN
    !---- iFac       : number of the face
    !---- nDofS      : number of dof in the face
    !---- nDofV      : number of dof in the cell
    !---- order      : polynomial order in the cell    
    !---- omegaValue : angular frequency
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_ELEMENTS_DOFS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: iFac, nDofS, nDofV, order

    REAL(C_DOUBLE), INTENT(IN)  :: omegaValue
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)              :: iDof

    COMPLEX(C_DOUBLE_COMPLEX)   :: solE(3), solH(3)
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_FACE_DOF(iFac, order, nDofS, nDofV,               &
         dofXS(1:nDofS), dofYS(1:nDofS), dofZS(1:nDofS))
    !
    DO iDof=1,nDofS
       CALL FREQ_SOL_PLANE_WAVE(omegaValue,                    &
            dofXS(iDof), dofYS(iDof), dofZS(iDof), solE, solH)
       !
       ExScx(iDof) = solE(1)
       EyScx(iDof) = solE(2)
       EzScx(iDof) = solE(3)
       HxScx(iDof) = solH(1)
       HyScx(iDof) = solH(2)
       HzScx(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FACE_SOL_PLANE_WAVE_FREQ

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_PLANE_WAVE(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the plane wave at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATERIALS_STRUCTURES
    USE SPECIAL_FUNCTIONS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: val, Eval, optIndex, optImpedance
    REAL(C_DOUBLE), DIMENSION(3)              :: E0, H0
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    optIndex     = SQRT(Mat(backgroundMat)%mu(1)*Mat(backgroundMat)%eps(1))
    optImpedance = SQRT(Mat(backgroundMat)%mu(1)/Mat(backgroundMat)%eps(1))
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    val  = time - retardPW - (x*waveVecPW(1) + y*waveVecPW(2) + z*waveVecPW(3))*optIndex
    Eval = SIN(omegaPW*val)*EXP(-(val**2)/(2.0*sigmaPW**2))
    E0   = ampliPW*polarPW/optIndex
    CALL CROSS_PROD(waveVecPW, E0, H0)
    H0   = H0*optImpedance
    !
    !---- E(x,y,z,t) = E0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- with    T  = t - t0 - K.x,
    !----         K  = k/omega,
    !----        |K| = 1/c = n/c0
    !----         E0 = E0 
    !---- H(x,y,z,t) = H0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- with    H0 = (K x E0)/Z
    solE(1) = E0(1)*Eval
    solE(2) = E0(2)*Eval
    solE(3) = E0(3)*Eval
    solH(1) = H0(1)*Eval
    solH(2) = H0(2)*Eval
    solH(3) = H0(3)*Eval
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_PLANE_WAVE  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_DT_SOL_PLANE_WAVE_TIME_SF(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Time derivative of the solution of the plane wave at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE)               :: attenuation, kTimesR, pmlDepth, &
                                    factor1, factor2

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH, r
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------    
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_DT_PLANE_WAVE(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       !---- E(x,y,z,omega)_pml = E0 exp(-i*omega*t) *
       !----                      exp(i*k*r*omega*s_pml)
       !---- Approximation: pmlKappa >> pmlAlpha*pmlSigma/(pmlAlpha**2 + omega**2)
       !---- E(x,y,z,omega)_pml ~ E0 exp(-i*omega*t) *
       !----                      exp(i*k*r*omega*pmlKappa) * 
       !----                      exp(-k*r*(pmlSigma*omega**2)/(pmlAlpha**2 + omega**2)
       !---- Approximation: Set omega in damping term to omegaPW
       !---- dtE(x,y,z,t)_pml ~ dtE(x,y,z,t) * exp(-k*r*(pmlSigma*omegaPW**2)/(pmlAlpha**2 + omegaPW**2) 
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_DT_SOL_PLANE_WAVE_TIME_SF

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_DT_PLANE_WAVE(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the time derivative of the plane 
    !---- wave at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for dtE and dtH
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: logMat, nMat, zmat
    
    REAL(C_DOUBLE)                            :: val, Eval
    REAL(C_DOUBLE), DIMENSION(3)              :: E0, H0
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    nMat = 1.0
    zMat = 1.0
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    val  = time - retardPW - (x*waveVecPW(1) + y*waveVecPW(2) + z*waveVecPW(3))*nMat
    Eval =  omegaPW*COS(omegaPW*val)*EXP(-(val**2)/(2.0*sigmaPW**2)) &
           -val/(sigmaPW**2)*SIN(omegaPW*val)*EXP(-(val**2)/(2.0*sigmaPW**2))
    E0   = ampliPW*polarPW
    CALL CROSS_PROD(waveVecPW, E0, H0)
    H0   = H0/zMat
    !
    !----   E(x,y,z,t) = E0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- dtE(x,y,z,t) = E0*(omega*cos(omega*T)*exp(-T^2/(2*sigma^2))
    !----               -T/sigma^2*sin(omega*T)*exp(-T^2/(2*sigma^2)))
    !---- with      T  = t - t0 - K.x,
    !----           K  = k/omega,
    !----          |K| = 1/c = n/c0
    !----           E0 = E0 
    !----   H(x,y,z,t) = H0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- dtH(x,y,z,t) = H0*(omega*cos(omega*T)*exp(-T^2/(2*sigma^2))
    !----               -T/sigma^2*sin(omega*T)*exp(-T^2/(2*sigma^2)))
    !---- with    H0 = (K x E0)/Z
    solE(1) = E0(1)*Eval
    solE(2) = E0(2)*Eval
    solE(3) = E0(3)*Eval
    solH(1) = H0(1)*Eval
    solH(2) = H0(2)*Eval
    solH(3) = H0(3)*Eval
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_DT_PLANE_WAVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_DX_PLANE_WAVE(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the x derivative of the 
    !---- plane wave at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: logMat, nMat, zmat
    
    REAL(C_DOUBLE)                            :: val, Eval
    REAL(C_DOUBLE), DIMENSION(3)              :: E0, H0
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    nMat = 1.0
    zMat = 1.0
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    val  = time - retardPW - (x*waveVecPW(1) + y*waveVecPW(2) + z*waveVecPW(3))*nMat
    Eval = waveVecPW(1)*nMat*EXP(-(val**2)/(2.0*sigmaPW**2)) &
           *(val/sigmaPW**2*SIN(omegaPW*val) - omegaPW*COS(omegaPW*val))  
    E0   = ampliPW*polarPW
    CALL CROSS_PROD(waveVecPW, E0, H0)
    H0   = H0/zMat
    !
    !----   E(x,y,z,t) = E0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- dxE(x,y,z,t) = E0*kx*exp(-T^2/(2*sigma^2))
    !----                     *(T/sigma^2*sin(omega*T)-cos(omega*T))
    !---- with      T  = t - t0 - K.x,
    !----           K  = k/omega,
    !----          |K| = 1/c = n/c0
    !----           E0 = E0 
    !----   H(x,y,z,t) = H0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- with      H0 = (K x E0)/Z
    !---- and     dxH  = (K x dxE)/Z
    solE(1) = E0(1)*Eval
    solE(2) = E0(2)*Eval
    solE(3) = E0(3)*Eval
    solH(1) = H0(1)*Eval
    solH(2) = H0(2)*Eval
    solH(3) = H0(3)*Eval
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_DX_PLANE_WAVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_DY_PLANE_WAVE(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the y derivative of the 
    !---- plane wave at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: logMat, nMat, zmat
    
    REAL(C_DOUBLE)                            :: val, Eval
    REAL(C_DOUBLE), DIMENSION(3)              :: E0, H0
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    nMat = 1.0
    zMat = 1.0
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    val  = time - retardPW - (x*waveVecPW(1) + y*waveVecPW(2) + z*waveVecPW(3))*nMat
    Eval = waveVecPW(2)*nMat*EXP(-(val**2)/(2.0*sigmaPW**2)) &
           *(val/sigmaPW**2*SIN(omegaPW*val) - omegaPW*COS(omegaPW*val))  
    E0   = ampliPW*polarPW
    CALL CROSS_PROD(waveVecPW, E0, H0)
    H0   = H0/zMat
    !
    !----   E(x,y,z,t) = E0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- dxE(x,y,z,t) = E0*ky*exp(-T^2/(2*sigma^2))
    !----                     *(T/sigma^2*sin(omega*T)-cos(omega*T))
    !---- with      T  = t - t0 - K.x,
    !----           K  = k/omega,
    !----          |K| = 1/c = n/c0
    !----           E0 = E0 
    !----   H(x,y,z,t) = H0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- with      H0 = (K x E0)/Z
    !---- and     dxH  = (K x dxE)/Z
    solE(1) = E0(1)*Eval
    solE(2) = E0(2)*Eval
    solE(3) = E0(3)*Eval
    solH(1) = H0(1)*Eval
    solH(2) = H0(2)*Eval
    solH(3) = H0(3)*Eval
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_DY_PLANE_WAVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_DZ_PLANE_WAVE(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Time-domain solution of the z derivative of the 
    !---- plane wave at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: logMat, nMat, zmat
    
    REAL(C_DOUBLE)                            :: val, Eval
    REAL(C_DOUBLE), DIMENSION(3)              :: E0, H0
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    nMat = 1.0
    zMat = 1.0
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    val  = time - retardPW - (x*waveVecPW(1) + y*waveVecPW(2) + z*waveVecPW(3))*nMat
    Eval = waveVecPW(3)*nMat*EXP(-(val**2)/(2.0*sigmaPW**2)) &
           *(val/sigmaPW**2*SIN(omegaPW*val) - omegaPW*COS(omegaPW*val))  
    E0   = ampliPW*polarPW
    CALL CROSS_PROD(waveVecPW, E0, H0)
    H0   = H0/zMat
    !
    !----   E(x,y,z,t) = E0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- dxE(x,y,z,t) = E0*kz*exp(-T^2/(2*sigma^2))
    !----                     *(T/sigma^2*sin(omega*T)-cos(omega*T))
    !---- with      T  = t - t0 - K.x,
    !----           K  = k/omega,
    !----          |K| = 1/c = n/c0
    !----           E0 = E0 
    !----   H(x,y,z,t) = H0*sin(omega*T)*exp(-T^2/(2*sigma^2))
    !---- with      H0 = (K x E0)/Z
    !---- and     dxH  = (K x dxE)/Z
    solE(1) = E0(1)*Eval
    solE(2) = E0(2)*Eval
    solE(3) = E0(3)*Eval
    solH(1) = H0(1)*Eval
    solH(2) = H0(2)*Eval
    solH(3) = H0(3)*Eval
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_DZ_PLANE_WAVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FREQ_SOL_PLANE_WAVE(omegaValue, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Frequency-domain solution of the plane wave at given coordinates
    !
    !---- IN
    !---- omegaValue : angular frequency
    !---- x, y, z    : coordinates for the solution
    !
    !---- OUT
    !---- sol*       : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),                          INTENT(IN)  :: omegaValue, x, y, z

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                                       :: norm, val
    REAL(C_DOUBLE), DIMENSION(3)                         :: E, H

    COMPLEX(C_DOUBLE_COMPLEX)                            :: Eval
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    val  = - omegaValue*(x*WaveVecPW(1) + y*WaveVecPW(2) + z*WaveVecPW(3))
    Eval = EXP(iunit*val)
    E(:) = ampliPW*polarPW(:)
    CALL CROSS_PROD(WaveVecPW, E, H)
    CALL NORMALIZE_VECTOR(WaveVecPW,norm)
    H(:) = H(:)/norm
    !
    solE(1) = E(1)*Eval
    solE(2) = E(2)*Eval
    solE(3) = E(3)*Eval
    solH(1) = H(1)*Eval
    solH(2) = H(2)*Eval
    solH(3) = H(3)*Eval
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FREQ_SOL_PLANE_WAVE  

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FREQ_POWER_PLANE_WAVE(freq, power)
    !------------------------------------------------------------------
    !---- Spectral power of the plane wave at given frequency
    !
    !---- IN
    !---- freq  : frequency in Hz
    !
    !---- OUT
    !---- power : power in ? unit
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN)  :: freq
    REAL(C_DOUBLE), INTENT(OUT) :: power
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)              :: omega, omegaPwLoc, sigmaPwLoc, &
         freqPwLoc, expVal, mOmega, pOmega, coef, E
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert to the right units
    !------------------------------------------------------------------
    freqPwLoc  = freqPW
    sigmaPwLoc = sigmaPW
    CALL INVERSE_METERS_TO_FREQUENCY(freqPwLoc)
    CALL FREQUENCY_TO_INVERSE_METERS(sigmaPwLoc)    
    !------------------------------------------------------------------
    !---- Compute power
    !------------------------------------------------------------------    
    omega      = 2.0*Pi*freq
    omegaPwLoc = 2.0*Pi*freqPwLoc
    mOmega     = sigmaPwLoc*(omega - omegaPwLoc)
    pOmega     = sigmaPwLoc*(omega + omegaPwLoc)
    expVal     = EXP(-0.5*mOmega**2) - EXP(-0.5*pOmega**2)
    !
    !---- Symmetric definition of Fourier transform
    coef       = 0.5*sigmaPwLoc  
    E          = ampliPW*coef*expVal
    power      = ABS(E)**2
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FREQ_POWER_PLANE_WAVE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE PLANE_WAVE
