MODULE QUADRATURES_2D

  !======================================================================
  ! Quadrature rules for triangles and quadrilaterals
  ! For triangles      : nearly-optimal quadrature rules    from order 1 to 20, from [1]
  ! For quadrilaterals : classical gauss points integration from order 1 to 4
  !
  ! [1] A set of symmetric quadrature rules on triangles and tetrahedra
  !     Zhang et al., JCM 27 (2009), 89-96  
  !======================================================================

  !----------------------------------------------------------------------
  ! Shared modules
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  !----------------------------------------------------------------------
  IMPLICIT NONE
  !----------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Global arrays
  !----------------------------------------------------------------------
  ! n2d        ==> number of quadrature points
  ! qx2d, qy2d ==> quadrature points 
  ! qw2d       ==> weights
  !----------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: nq2d

  REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: qx2d, qy2d, qw2d
  !----------------------------------------------------------------------

CONTAINS

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE BUILD_QUADRATURE_2D(faceType, quadOrder)
    !--------------------------------------------------------------------
    !---- Call quadrature routine depending on face type and order
    !
    !---- IN
    !---- faceType : type of face
    !---- order    : order of the integrand
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=*), INTENT(IN)    :: faceType
    
    INTEGER(C_INT),                INTENT(INOUT) :: quadOrder
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Call right quadrature routines
    !--------------------------------------------------------------------
    SELECT CASE(faceType)
    CASE('TRI', 'TRI1', 'TRI2', 'TRI3', 'TRI4')
       CALL QUADRATURES_TRIANGLE  (quadOrder)
    CASE('QUA', 'QUA1', 'QUA2')
       CALL QUADRATURES_QUADRANGLE(quadOrder)
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE BUILD_QUADRATURE_2D

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE QUADRATURES_TRIANGLE(quadOrder)
    !--------------------------------------------------------------------
    !---- Initialize quadrature rule depending on integrand order
    !
    !---- IN
    !---- order : order of the integrand
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),     INTENT(IN) :: quadOrder
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                 :: nPtsPermTot
    INTEGER(C_INT), DIMENSION(3)   :: nPtsPermTri

    REAL(C_DOUBLE)                 :: a, b, w
    REAL(C_DOUBLE), DIMENSION(6,2) :: temp111
    REAL(C_DOUBLE), DIMENSION(3,2) :: temp21
    REAL(C_DOUBLE), DIMENSION(1,2) :: temp3
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !---- Deallocate arrays if they were allocated previously
    IF (ALLOCATED(nq2d)) DEALLOCATE(nq2d)
    IF (ALLOCATED(qw2d)) DEALLOCATE(qw2d)
    IF (ALLOCATED(qx2d)) DEALLOCATE(qx2d)
    IF (ALLOCATED(qy2d)) DEALLOCATE(qy2d)
    !
    !---- Check the required quadrature order
    IF ((quadOrder < 0) .OR. (quadOrder > 20)) THEN
       WRITE(6, '(a)')      'Error in QUADRATURES_2D'
       WRITE(6, '(a,i0,a)') 'Quadrature rule of order ',quadOrder,' is not &
            available on triangles'
       CALL TILT
    ENDIF
    !--------------------------------------------------------------------
    !---- Allocate and initialize arrays
    !--------------------------------------------------------------------
    !---- Fill nq2d and nPtsPermTri
    ALLOCATE(nq2d(0:20))
    nq2d        = (/1, 1, 3, 6, 6, 7, 12, 15, 16, 19, 25, 28, 33, 37, 48, 57, 57, 66, 84, 84, 106/)
    nPtsPermTri = (/1, 3, 6/)
    !
    !---- Allocate weights and points arrays
    ALLOCATE(qx2d(nq2d(quadOrder)))
    ALLOCATE(qy2d(nq2d(quadOrder)))
    ALLOCATE(qw2d(nq2d(quadOrder)))
    !--------------------------------------------------------------------
    !---- Build appropriate rule
    !--------------------------------------------------------------------
    SELECT CASE (quadOrder)
    CASE (0,1)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.3333333333333333
       w = 1
       CALL PERM3(a, w, temp3, nPtsPermTot)
    CASE (2)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.1666666666666667
       w = 0.3333333333333333
       CALL PERM21(a, w, temp21, nPtsPermTot)
    CASE (3)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.08343123540964358
       w = 0.1006463893858574
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.4463828804146981
       w = 0.232686943947476
       CALL PERM21(a, w, temp21, nPtsPermTot)
    CASE (4)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.09157621350977073
       w = 0.1099517436553219
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.4459484909159649
       w = 0.2233815896780115
       CALL PERM21(a, w, temp21, nPtsPermTot)
    CASE (5)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.3333333333333333
       w = 0.225
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.1012865073234563
       w = 0.1259391805448272
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.4701420641051151
       w = 0.1323941527885062
       CALL PERM21(a, w, temp21, nPtsPermTot)
    CASE (6)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.219429982549783
       w = 0.171333124152981
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.480137964112215
       w = 0.08073108959303098
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.1416190159239681
       b = 0.01937172436124079
       w = 0.04063455979366066
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (7)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.211545925991567
       w = 0.082028453840595
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.06327272641437519
       w = 0.05054977388608355
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.4076375309280243
       w = 0.07310089504730739
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Quatrieme permutation
       a = 0.312874485358357
       b = 0.0393124184281643
       w = 0.06382710527967368
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (8)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.3333333333333333
       w = 0.1443156076777872
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.05054722831703097
       w = 0.03245849762319808
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.1705693077517602
       w = 0.1032173705347182
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Quatrieme permutation
       a = 0.4592925882927231
       w = 0.09509163426728465
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Cinquieme permutation
       a = 0.2631128296346381
       b = 0.008394777409957609
       w = 0.027230314174435
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (9)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.3333333333333333
       w = 0.09713579628279882
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.1882035356190327
       w = 0.07964773892721025
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.04472951339445271
       w = 0.02557767565869803
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Quatrieme permutation
       a = 0.4896825191987376
       w = 0.03133470022713907
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Cinquieme permutation
       a = 0.4370895914929366
       w = 0.07782754100477428
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Sixieme permutation
       a = 0.2219629891607657
       b = 0.03683841205473629
       w = 0.04328353937728938
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (10)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.3333333333333333
       w = 0.08318553908730103
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.1848771691055548
       w = 0.07864108985217015
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.02114458310910991
       w = 0.007213065728130081
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Quatrieme permutation
       a = 0.4277169124511194
       w = 0.0740144134671236
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Cinquieme permutation
       a = 0.499998
       w = 0.005775042259099829
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Sixieme permutation
       a = 0.3361802769144466
       b = 0.03753078378313841
       w = 0.04016655082346215
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Septieme permutation
       a = 0.1380631838790965
       b = 0.03513783768207891
       w = 0.02981405367539252
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (11)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.3333333333333333
       w = 0.0879768437512549
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.2073440273884721
       w = 0.07215933836671422
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.094289508762816
       w = 0.03808164961037525
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Quatrieme permutation
       a = 0.4389077467980207
       w = 0.06932873512069092
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Cinquieme permutation
       a = 0.494637024787977
       w = 0.01885497098833655
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Sixieme permutation
       a = 0.02598983403722602
       w = 0.008744747158861149
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Septieme permutation
       a = 0.2772233150705471
       b = 0.04484194598885781
       w = 0.04105610529389746
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.8588662836018728
       b = 0.1411317163981272
       w = 0.007363033458404331
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (12)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.2714625070149261
       w = 0.06254121319590276
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.1092578276593543
       w = 0.02848605206887755
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.02464636343633559
       w = 0.007931642509973639
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Quatrieme permutation
       a = 0.4401116486585931
       w = 0.04991833492806095
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Cinquieme permutation
       a = 0.4882037509455415
       w = 0.02426683808145203
       CALL PERM21(a, w, temp21, nPtsPermTot)         
       !---- Sixieme permutation
       a = 0.2554542286385174
       b = 0.1162960196779266
       w = 0.04322736365941422
       CALL PERM111(a, b, w, temp111, nPtsPermTot)         
       !---- Septieme permutation
       a = 0.2916556797383409
       b = 0.02303415635526714
       w = 0.02178358503860756
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.1272797172335894
       b = 0.02138249025617059
       w = 0.01508367757651144
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (13)
       nPtsPermTot = 0
       !---- Premiere permutation                                          
       a = 0.3333333333333333
       w = 0.05242442599039866
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation                                          
       a = 0.229468729341773
       w = 0.04729382649194289
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                                         
       a = 0.1143794784970616
       w = 0.03115446231812814
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                                         
       a = 0.02481607024681091
       w = 0.007978703001561666
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                                         
       a = 0.4144599997122385
       w = 0.04703882080534875
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation                                            
       a =0.4686865457169068
       w =0.03148965370264994
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Septieme permutation                                            
       a =0.4950619228323134
       w =0.01126935161218213
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Huitieme permutation                                           
       a = 0.2686938332921753
       b = 0.09499772387915931
       w = 0.03686306819424886
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                                          
       a = 0.2917489931261045
       b = 0.01814725621289835
       w = 0.0174388570349182
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                                          
       a = 0.1263807043544204
       b = 0.02221898969729588
       w = 0.01551492813985976
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (14)
       nPtsPermTot = 0
       !---- Premiere permutation                   
       a = 0.2691822372771082
       w = 0.02655375139763148
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation                   
       a = 0.01584306362489598
       w = 0.003455427088993395
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                  
       a = 0.08198104351874962
       w = 0.01078306283748246
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                  
       a = 0.3902168969247423
       w = 0.03860125046871679
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4544162411767542
       w = 0.03635469288219466
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation             
       a = 0.499608260905523
       w = 0.005953067695078087
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.2663320074893452
       b = 0.1403403557376521
       w = 0.03922351455181438
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.1781623277441465
       b = 0.07957407119524386
       w = 0.0237627124475065
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.8931888463096356
       b = 0.08830122381641832
       w = 0.009486179389398509
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.3571657417088998
       b = 0.03192431162852029
       w = 0.02280783745143184
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation                   
       a = 0.2201536204443806
       b = 0.01250540563272446
       w = 0.01053579664146701
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (15)
       nPtsPermTot = 0
       !---- Premiere permutation                   
       a = 0.2879965484640022
       w = 0.03561509503647825
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation                   
       a = 0.19172513401718
       w = 0.03446485003815803
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                  
       a = 0.01321793666292992
       w = 0.002370565077095076
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                  
       a = 0.4945263487334957
       w = 0.009783474971461916
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4687732859518785
       w = 0.02358537694176325
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation             
       a = 0.0729701018887564
       b = 0.0896086526984031
       w = 0.007937622780942462
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.3456040545691723
       b = 0.1628662598046491
       w = 0.03295602080801961
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.3134766869730596
       b = 0.07650916368612136
       w = 0.02224364444573269
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.1859089312646532
       b = 0.08105519653380036
       w = 0.0211936039581939
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.07256122263051047
       b = 0.9115962488602088
       w = 0.006788214618713039
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation
       a = 0.1811159703832591
       b = 0.01533969944114118
       w = 0.01022858365264153
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Douzieme permutation
       a = 0.3271171127221582
       b = 0.6574862967771837
       w = 0.01240929536994518
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (16)
       nPtsPermTot = 0
       !---- Premiere permutation
       a = 0.2879965484640022
       w = 0.03561509503647825
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation
       a = 0.19172513401718
       w = 0.03446485003815803
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation
       a = 0.01321793666292992
       w = 0.002370565077095076
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation
       a = 0.4945263487334957
       w = 0.009783474971461916
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4687732859518785
       w = 0.02358537694176325
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation             
       a = 0.0729701018887564
       b = 0.0896086526984031
       w = 0.007937622780942462
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.3456040545691723
       b = 0.1628662598046491
       w = 0.03295602080801961
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.3134766869730596
       b = 0.07650916368612136
       w = 0.02224364444573269
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.1859089312646532
       b = 0.08105519653380036
       w = 0.0211936039581939
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.07256122263051047
       b = 0.9115962488602088
       w = 0.006788214618713039
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation                   
       a = 0.1811159703832591
       b = 0.01533969944114118
       w = 0.01022858365264153
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Douzieme permutation                    
       a = 0.3271171127221582
       b = 0.6574862967771837
       w = 0.01240929536994518
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (17)
       nPtsPermTot = 0
       !---- Premiere permutation                   
       a = 0.2875413741915859
       w = 0.03564986201080552
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation                   
       a = 0.07911335977549125
       w = 0.01471168452587801
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                  
       a = 0.01492813520633954
       w = 0.002915504033919137
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                  
       a = 0.4175175465024245
       w = 0.0337778604201483
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4668010945264359
       w = 0.01633410847657116
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation                  
       a = 0.4931948181403382
       w = 0.01147584786036648
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.07495507701144774
       b = 0.3565882464867537
       w = 0.0156008465601933
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.2708307463146822
       b = 0.165282797475811
       w = 0.02943361449440034
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.1800833875421717
       b = 0.1127488069889171
       w = 0.01872909571993711
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.2846420126813479
       b = 0.06224181260488845
       w = 0.01124763630663286
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation                   
       a = 0.3308192683978194
       b = 0.01321703510235169
       w = 0.01056191692271306
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Douzieme permutation                    
       a = 0.1822396941723672
       b = 0.03653377874734052
       w = 0.01392020379011369
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Treizieme permutation                    
       a = 0.1877244862138654
       b = 0.001079715952422338
       w = 0.002928996324757894
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quatorzieme permutation                    
       a = 0.07752405970460868
       b = 0.01547774667650556
       w = 0.006811922884074084
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (18)
       nPtsPermTot = 0
       !---- Premiere permutation                   
       a = 0.2668208784784349
       w = 0.02338880517576498
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation                   
       a = 0.08502136817527391
       w = 0.01295505481482721
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                  
       a = 0.01459843628946509
       w = 0.0007272074440818744
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                  
       a = 0.3731491240975792
       w = 0.02389764300272942
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4765144136671617
       w = 0.01411617123360641
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation             
       a = 0.4967342441740967
       w = 0.004125581247992263
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.4757478947467256
       b = 0.1504846529705858
       w = 0.01980326746769443
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.2551463120371493
       b = 0.161477880758523
       w = 0.02354050977929026
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.181455961883646
       b = 0.107294768621172
       w = 0.01794192174859743
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.3691503603369549
       b = 0.07994322597809553
       w = 0.01573919048947826
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation                   
       a = 0.2769862890493225
       b = 0.05519109623775229
       w = 0.01524445560087352
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Douzieme permutation                    
       a = 0.9132959320265938
       b = 0.06561830927120929
       w = 0.006106293144062769
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Treizieme permutation                    
       a = 0.3794150158222324
       b = 0.01390393383051861
       w = 0.008721193527958436
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quatorzieme permutation                    
       a = 0.02408633677871775
       b = 0.00233515074160661
       w = 0.0008584669633312858
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quinzieme permutation                    
       a = 0.159044177519736
       b = 0.03359109880431897
       w = 0.01127422240684624
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Seizieme permutation                    
       a = 0.130561673324793
       b = 8.428933734364707e-06
       w = 0.001882641220603348
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dix-septieme permutation                    
       a = 0.2535509373499845
       b = 0.7375771919817597
       w = 0.005949272858429614
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (19)
       nPtsPermTot = 0
       !---- Premiere permutation                   
       a = 0.2668208784784349
       w = 0.02338880517576498
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Deuxieme permutation                   
       a = 0.08502136817527391
       w = 0.01295505481482721
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                  
       a = 0.01459843628946509
       w = 0.0007272074440818744
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                  
       a = 0.3731491240975792
       w = 0.02389764300272942
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4765144136671617
       w = 0.01411617123360641
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation             
       a = 0.4967342441740967
       w = 0.004125581247992263
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.4757478947467256
       b = 0.1504846529705858
       w = 0.01980326746769443
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.2551463120371493
       b = 0.161477880758523
       w = 0.02354050977929026
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.181455961883646
       b = 0.107294768621172
       w = 0.01794192174859743
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.3691503603369549
       b = 0.07994322597809553
       w = 0.01573919048947826
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation                   
       a = 0.2769862890493225
       b = 0.05519109623775229
       w = 0.01524445560087352
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Douzieme permutation                    
       a = 0.9132959320265938
       b = 0.06561830927120929
       w = 0.006106293144062769
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Treizieme permutation                    
       a = 0.3794150158222324
       b = 0.01390393383051861
       w = 0.008721193527958436
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quatorzieme permutation                    
       a = 0.02408633677871775
       b = 0.00233515074160661
       w = 0.0008584669633312858
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quinzieme permutation                    
       a = 0.159044177519736
       b = 0.03359109880431897
       w = 0.01127422240684624
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Seizieme permutation                    
       a = 0.130561673324793
       b = 8.428933734364707e-06
       w = 0.001882641220603348
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dix-septieme permutation                    
       a = 0.2535509373499845
       b = 0.7375771919817597
       w = 0.005949272858429614
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    CASE (20)
       nPtsPermTot = 0
       !---- Premiere permutation                   
       a = 0.3333333333333333
       w = 0.02788015254750532
       CALL PERM3(a, w, temp3, nPtsPermTot)
       !---- Deuxieme permutation                   
       a = 0.2182144653663702
       w = 0.02014779806138516
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Troisieme permutation                  
       a = 0.08737788026835197
       w = 0.009120838354595963
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Quatrieme permutation                  
       a = 0.006479792680126985
       w = 0.0006254164009509602
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Cinquieme permutation                  
       a = 0.4761242870456229
       w = 0.006740756718435428
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Sixieme permutation             
       a = 0.4949757431494962
       w = 0.007166401485081101
       CALL PERM21(a, w, temp21, nPtsPermTot)
       !---- Septieme permutation             
       a = 0.3294045895067025
       b = 0.2198711148522821
       w = 0.02533023834446573
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Huitieme permutation
       a = 0.1806363490092006
       b = 0.1521078411982259
       w = 0.00759645627481338
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Neuvieme permutation                    
       a = 0.3057311279878027
       b = 0.1284977793629375
       w = 0.01516439192878541
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dixieme permutation                   
       a = 0.4717748839681798
       b = 0.1228325775669285
       w = 0.0136339837179289
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Onzieme permutation                   
       a = 0.1670261081476675
       b = 0.09004213174898912
       w = 0.012075605047566
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Douzieme permutation                    
       a = 0.3852148463157488
       b = 0.05362040286764569
       w = 0.01419601903963345
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Treizieme permutation                    
       a = 0.2643495759538629
       b = 0.04463615185971796
       w = 0.01223849091582233
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quatorzieme permutation                    
       a = 0.1518207239119767
       b = 0.03392122875574324
       w = 0.008245806768008486
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Quinzieme permutation                    
       a = 0.07142935815309673
       b = 0.03342573353718147
       w = 0.004955038663398934
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Seizieme permutation                    
       a = 0.04002520548256743
       b = 0.009501993341412331
       w = 0.002403483507953331
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dix-septieme permutation                    
       a = 0.3564806569929756
       b = 0.009759239001244081
       w = 0.006790591513868497
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dix-huitieme permutation                    
       a = 0.1143035547263665
       b = 0.004999396516337759
       w = 0.002636259775316881
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Dix-neuvieme permutation                    
       a = 0.1424439896654094
       b = 0.007206947768123408
       w = 0.0001228880789111185
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Vingtieme permutation                    
       a = 0.2564715300140338
       b = 0.638809191546181
       w = 0.009883684419276005
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
       !---- Vingt-et-unieme permutation                    
       a = 0.2270696517416056
       b = 0.007718765286441409
       w = 0.004846431069443004
       CALL PERM111(a, b, w, temp111, nPtsPermTot)
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE QUADRATURES_TRIANGLE

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE QUADRATURES_QUADRANGLE(quadOrder)
    !--------------------------------------------------------------------
    !---- Selects the quadrature rule according to the order of quadrature
    !
    !---- IN
    !---- quadOrder : order of the integrand
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    USE QUADRATURES_1D
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: quadOrder
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)             :: i, j, com
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------    
    !---- Deallocate arrays if they were allocated previously
    IF (ALLOCATED(nq2d)) DEALLOCATE(nq2d)    
    IF (ALLOCATED(qx2d)) DEALLOCATE(qx2d)
    IF (ALLOCATED(qy2d)) DEALLOCATE(qy2d)
    IF (ALLOCATED(qw2d)) DEALLOCATE(qw2d)
    !
    !---- Check the required quadrature order
    IF ((quadOrder < 0) .OR. (quadOrder > 8)) THEN
       WRITE(6, '(a)')      'Error in QUADRATURES_2D'
       WRITE(6, '(a,i0,a)') 'Quadrature rule of order ',quadOrder,' is not &
            available on quadrilaterals'
       CALL TILT
    ENDIF    
    !--------------------------------------------------------------------
    !---- Allocate and initialize arrays
    !--------------------------------------------------------------------
    !---- Fill nq2d
    ALLOCATE(nq2d(0:8))
    nq2d = (/1, 4, 4, 9, 9, 16, 16, 25, 25/)
    !
    !---- Allocate weights and points arrays
    ALLOCATE(qx2d(nq2d(quadOrder)))
    ALLOCATE(qy2d(nq2d(quadOrder)))
    ALLOCATE(qw2d(nq2d(quadOrder)))    
    !--------------------------------------------------------------------
    !---- Build appropriate rule
    !--------------------------------------------------------------------
    com = 0
    !
    SELECT CASE (quadOrder)
    CASE(0)
       qx2d(:) = 0.0
       qy2d(:) = 0.0
       qw2d(:) = 4.0
    CASE(1:2)
       DO i=1,2
          DO j=1,2
             com       = com + 1
             qx2d(com) = q1p2(1,i)
             qy2d(com) = q1p2(1,j)
             qw2d(com) = q1p2(2,i)*q1p2(2,j)
          ENDDO
       ENDDO
    CASE(3:4)
       DO i=1,3
          DO j=1,3
             com       = com + 1
             qx2d(com) = q1p5(1,i)
             qy2d(com) = q1p5(1,j)
             qw2d(com) = q1p5(2,i)*q1p5(2,j)
          ENDDO
       ENDDO
    CASE(5:6)
       DO i=1,4
          DO j=1,4
             com       = com + 1
             qx2d(com) = q1p7(1,i)
             qy2d(com) = q1p7(1,j)
             qw2d(com) = q1p7(2,i)*q1p7(2,j)
          ENDDO
       ENDDO
    CASE(7:8)
       DO i=1,5
          DO j=1,5
             com       = com + 1
             qx2d(com) = q1p9(1,i)
             qy2d(com) = q1p9(1,j)
             qw2d(com) = q1p9(2,i)*q1p9(2,j)
          ENDDO
       ENDDO       
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE QUADRATURES_QUADRANGLE

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERMCALCTRI(xi1, xi2, xi3, x)
    !--------------------------------------------------------------------
    !---- Compute point coordinates from permutation data
    !---- This routine is called by each permutation routine below
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN)  :: xi1, xi2, xi3
    REAL(C_DOUBLE), INTENT(OUT) :: x(2)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE) :: refc(3,2)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill refc
    refc(1,1) = 0.0
    refc(1,2) = 0.0
    refc(2,1) = 1.0
    refc(2,2) = 0.0
    refc(3,1) = 0.0
    refc(3,2) = 1.0
    !
    !---- Compute coordinates
    x(1) = xi1*refc(1,1) + xi2*refc(2,1) + xi3*refc(3,1)
    x(2) = xi1*refc(1,2) + xi2*refc(2,2) + xi3*refc(3,2)
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERMCALCTRI

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM3(a, w, temp3, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm3
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp3(1,2)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, x(2)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    !
    !---- Call PERMCALCTRI and store point 1
    CALL PERMCALCTRI(xi1, xi1, xi1, x)
    !
    DO i=1,2
       temp3(1,i) = x(i)
    ENDDO
    !
    !---- Fill points and weights
    DO i=1,1
       qx2d(nPtsPermTot+i) = temp3(i,1)
       qy2d(nPtsPermTot+i) = temp3(i,2)          
       qw2d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 1
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM3

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM21(a, w, temp21, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm21
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp21(3,2)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, xi2, x(2)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    xi2 = 1.0 - 2.0*a
    !
    !---- Call PERMCALCTRI and store point 1
    CALL PERMCALCTRI(xi1, xi1, xi2, x)
    !
    DO i=1,2
       temp21(1,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 2
    CALL PERMCALCTRI(xi1, xi2, xi1, x)
    !
    DO i=1,2
       temp21(2,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 3
    CALL PERMCALCTRI(xi2, xi1, xi1, x)
    !
    DO i=1,2
       temp21(3,i) = x(i)
    ENDDO
    !
    !---- Fill points and weights
    DO i=1,3
       qx2d(nPtsPermTot+i) = temp21(i,1)
       qy2d(nPtsPermTot+i) = temp21(i,2)       
       qw2d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 3
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM21

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM111(a, b, w, temp111, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm111
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, b, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp111(6,2)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, xi2, xi3, x(2)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    xi2 = b
    xi3 = 1.0 - a - b
    !
    !---- Call PERMCALCTRI and store point 1
    CALL PERMCALCTRI(xi1, xi2, xi3, x)
    !
    DO i=1,2
       temp111(1,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 2
    CALL PERMCALCTRI(xi1, xi3, xi2, x)
    !
    DO i=1,2
       temp111(2,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 3
    CALL PERMCALCTRI(xi3, xi1, xi2, x)
    !
    DO i=1,2
       temp111(3,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 4
    CALL PERMCALCTRI(xi2, xi1, xi3, x)
    !
    DO i=1,2
       temp111(4,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 5
    CALL PERMCALCTRI(xi2, xi3, xi1, x)
    !
    DO i=1,2
       temp111(5,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTRI and store point 6
    CALL PERMCALCTRI(xi3, xi2, xi1, x)
    !
    DO i=1,2
       temp111(6,i) = x(i)
    ENDDO
    !
    !---- Fill points and weights
    DO i=1,6
       qx2d(nPtsPermTot+i) = temp111(i,1)
       qy2d(nPtsPermTot+i) = temp111(i,2)       
       qw2d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 6
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM111
  
  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE QUADRATURES_DUNAVANT_2D(ndof)
    !======================================================================
    ! 3 sets (25, 33, 42) of Dunavant quadrature rules given on the reference
    ! triangle : { x >=0,y>=0; x+y <= 1 }
    !======================================================================

!!!!!!!!!!!!!!!!!!!!!!!!!!!! CAREFUL HERE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! On this set of quadrature, triangle weights are already pondered with
    ! the area of the reference triangle (1/2)
!!!!!!!!!!!!!!!!!!!!!!!!!!!! CAREFUL HERE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    !----------------------------------------------------------------------
    !---- 25-point quadrature rule (n2d = 25) 
    !----------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(25) :: qx25 = (/                                    &
         0.33333333333333298d0,  0.02884473323268500d0,  0.48557763338365700d0,   &
         0.48557763338365700d0,  0.78103684902992598d0,  0.10948157548503699d0,   &
         0.10948157548503699d0,  0.14170721941487999d0,  0.30793983876412101d0,   &
         0.55035294182099903d0,  0.30793983876412101d0,  0.55035294182099903d0,   &
         0.14170721941487999d0,  0.02500353476268600d0,  0.24667256063990300d0,   &
         0.72832390459741103d0,  0.24667256063990300d0,  0.72832390459741103d0,   &
         0.02500353476268600d0,  0.00954081540029900d0,  0.06680325101220000d0,   &
         0.92365593358749998d0,  0.06680325101220000d0,  0.92365593358749998d0,   &
         0.00954081540029900d0 /)

    REAL(C_DOUBLE), DIMENSION(25) :: qy25 = (/                                    &
         0.33333333333333298d0,  0.48557763338365700d0,  0.48557763338365700d0,   &
         0.02884473323268500d0,  0.10948157548503699d0,  0.10948157548503699d0,   &
         0.78103684902992598d0,  0.30793983876412101d0,  0.55035294182099903d0,   &
         0.14170721941487999d0,  0.14170721941487999d0,  0.30793983876412101d0,   &
         0.55035294182099903d0,  0.24667256063990300d0,  0.72832390459741103d0,   &
         0.02500353476268600d0,  0.02500353476268600d0,  0.24667256063990300d0,   &
         0.72832390459741103d0,  0.06680325101220000d0,  0.92365593358749998d0,   &
         0.00954081540029900d0,  0.00954081540029900d0,  0.06680325101220000d0,   &
         0.92365593358749998d0 /)

    REAL(C_DOUBLE), DIMENSION(25) :: qw25 = (/ &
         0.04540899519137700d0,  0.01836297887823350d0,  0.01836297887823350d0,   &
         0.01836297887823350d0,  0.02266052971776400d0,  0.02266052971776400d0,   &
         0.02266052971776400d0,  0.03637895842271000d0,  0.03637895842271000d0,   &
         0.03637895842271000d0,  0.03637895842271000d0,  0.03637895842271000d0,   &
         0.03637895842271000d0,  0.01416362126552850d0,  0.01416362126552850d0,   &
         0.01416362126552850d0,  0.01416362126552850d0,  0.01416362126552850d0,   &
         0.01416362126552850d0,  0.00471083348186650d0,  0.00471083348186650d0,   &
         0.00471083348186650d0,  0.00471083348186650d0,  0.00471083348186650d0,   &
         0.00471083348186650d0 /)

    !----------------------------------------------------------------------
    !---- 33-point quadrature rule (n2d = 33)
    !----------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(33) :: qx33 = (/                                    &
         0.023565220452389998d0, 0.488217389773805022d0, 0.488217389773805022d0,  &
         0.120551215411078994d0, 0.439724392294459976d0, 0.439724392294459976d0,  &
         0.457579229975767998d0, 0.271210385012115973d0, 0.271210385012115973d0,  &
         0.744847708916828055d0, 0.127576145541586000d0, 0.127576145541586000d0,  &
         0.957365299093579036d0, 0.021317350453210000d0, 0.021317350453210000d0,  &
         0.115343494534697999d0, 0.275713269685514017d0, 0.608943235779788039d0,  &
         0.275713269685514017d0, 0.608943235779788039d0, 0.115343494534697999d0,  &
         0.022838332222257000d0, 0.281325580989939983d0, 0.695836086787802954d0,  &
         0.281325580989939983d0, 0.695836086787802954d0, 0.022838332222257000d0,  &
         0.025734050548330001d0, 0.116251915907596995d0, 0.858014033544073018d0,  &
         0.116251915907596995d0, 0.858014033544073018d0, 0.025734050548330001d0 /)

    REAL(C_DOUBLE), DIMENSION(33) :: qy33 = (/ &
         0.488217389773805022d0, 0.488217389773805022d0, 0.023565220452389998d0,  &
         0.439724392294459976d0, 0.439724392294459976d0, 0.120551215411078994d0,  &
         0.271210385012115973d0, 0.271210385012115973d0, 0.457579229975767998d0,  &
         0.127576145541586000d0, 0.127576145541586000d0, 0.744847708916828055d0,  &
         0.021317350453210000d0, 0.021317350453210000d0, 0.957365299093579036d0,  &
         0.275713269685514017d0, 0.608943235779788039d0, 0.115343494534697999d0,  &
         0.115343494534697999d0, 0.275713269685514017d0, 0.608943235779788039d0,  &
         0.281325580989939983d0, 0.695836086787802954d0, 0.022838332222257000d0,  &
         0.022838332222257000d0, 0.281325580989939983d0, 0.695836086787802954d0,  &
         0.116251915907596995d0, 0.858014033544073018d0, 0.025734050548330001d0,  &
         0.025734050548330001d0, 0.116251915907596995d0, 0.858014033544073018d0 /)

    REAL(C_DOUBLE), DIMENSION(33) :: qw33 = (/                                    &
         0.012865533220227500d0, 0.012865533220227500d0, 0.012865533220227500d0,  &
         0.021846272269019001d0, 0.021846272269019001d0, 0.021846272269019001d0,  &
         0.031429112108942503d0, 0.031429112108942503d0, 0.031429112108942503d0,  &
         0.017398056465354500d0, 0.017398056465354500d0, 0.017398056465354500d0,  &
         0.003083130525779500d0, 0.003083130525779500d0, 0.003083130525779500d0,  &
         0.020185778883190501d0, 0.020185778883190501d0, 0.020185778883190501d0,  &
         0.020185778883190501d0, 0.020185778883190501d0, 0.020185778883190501d0,  &
         0.011178386601151500d0, 0.011178386601151500d0, 0.011178386601151500d0,  &
         0.011178386601151500d0, 0.011178386601151500d0, 0.011178386601151500d0,  &
         0.008658115554329500d0, 0.008658115554329500d0, 0.008658115554329500d0,  &
         0.008658115554329500d0, 0.008658115554329500d0, 0.008658115554329500d0 /)

    !----------------------------------------------------------------------
    !---- 42-point quadrature rule (n2d = 42)
    !----------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(42) :: qx42 = (/ &
         0.022072179275642999d0, 0.488963910362179011d0, 0.488963910362179011d0,  &
         0.164710561319092008d0, 0.417644719340453996d0, 0.417644719340453996d0,  &
         0.453044943382322984d0, 0.273477528308838980d0, 0.273477528308838980d0,  &
         0.645588935174913003d0, 0.177205532412542999d0, 0.177205532412542999d0,  &
         0.876400233818255048d0, 0.061799883090873003d0, 0.061799883090873003d0,  &
         0.961218077502598023d0, 0.019390961248700999d0, 0.019390961248700999d0,  &
         0.057124757403648002d0, 0.172266687821356013d0, 0.770608554774996013d0,  &
         0.172266687821356013d0, 0.770608554774996013d0, 0.057124757403648002d0,  &
         0.092916249356972000d0, 0.336861459796345020d0, 0.570222290846682967d0,  &
         0.336861459796345020d0, 0.570222290846682967d0, 0.092916249356972000d0,  &
         0.014646950055653999d0, 0.298372882136258011d0, 0.686980167808088016d0,  &
         0.298372882136258011d0, 0.686980167808088016d0, 0.014646950055653999d0,  &
         0.001268330932872000d0, 0.118974497696957005d0, 0.879757171370170954d0,  &
         0.118974497696957005d0, 0.879757171370170954d0, 0.001268330932872000d0 /)

    REAL(C_DOUBLE), DIMENSION(42) :: qy42 = (/ &
         0.488963910362179011d0, 0.488963910362179011d0, 0.022072179275642999d0,  &
         0.417644719340453996d0, 0.417644719340453996d0, 0.164710561319092008d0,  &
         0.273477528308838980d0, 0.273477528308838980d0, 0.453044943382322984d0,  &
         0.177205532412542999d0, 0.177205532412542999d0, 0.645588935174913003d0,  &
         0.061799883090873003d0, 0.061799883090873003d0, 0.876400233818255048d0,  &
         0.019390961248700999d0, 0.019390961248700999d0, 0.961218077502598023d0,  &
         0.172266687821356013d0, 0.770608554774996013d0, 0.057124757403648002d0,  &
         0.057124757403648002d0, 0.172266687821356013d0, 0.770608554774996013d0,  &
         0.336861459796345020d0, 0.570222290846682967d0, 0.092916249356972000d0,  &
         0.092916249356972000d0, 0.336861459796345020d0, 0.570222290846682967d0,  &
         0.298372882136258011d0, 0.686980167808088016d0, 0.014646950055653999d0,  &
         0.014646950055653999d0, 0.298372882136258011d0, 0.686980167808088016d0,  &
         0.118974497696957005d0, 0.879757171370170954d0, 0.001268330932872000d0,  &
         0.001268330932872000d0, 0.118974497696957005d0, 0.879757171370170954d0 /)

    REAL(C_DOUBLE), DIMENSION(42) :: qw42 = (/ &
         0.010941790684714500d0, 0.010941790684714500d0, 0.010941790684714500d0,  &
         0.016394176772062501d0, 0.016394176772062501d0, 0.016394176772062501d0,  &
         0.025887052253646001d0, 0.025887052253646001d0, 0.025887052253646001d0,  &
         0.021081294368496501d0, 0.021081294368496501d0, 0.021081294368496501d0,  &
         0.007216849834888500d0, 0.007216849834888500d0, 0.007216849834888500d0,  &
         0.002461701801200000d0, 0.002461701801200000d0, 0.002461701801200000d0,  &
         0.012332876606282000d0, 0.012332876606282000d0, 0.012332876606282000d0,  &
         0.012332876606282000d0, 0.012332876606282000d0, 0.012332876606282000d0,  &
         0.019285755393530502d0, 0.019285755393530502d0, 0.019285755393530502d0,  &
         0.019285755393530502d0, 0.019285755393530502d0, 0.019285755393530502d0,  &
         0.007218154056767000d0, 0.007218154056767000d0, 0.007218154056767000d0,  &
         0.007218154056767000d0, 0.007218154056767000d0, 0.007218154056767000d0,  &
         0.002505114419250500d0, 0.002505114419250500d0, 0.002505114419250500d0,  &
         0.002505114419250500d0, 0.002505114419250500d0, 0.002505114419250500d0 /)

    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: ndof
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: n
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (ALLOCATED(nq2d)) DEALLOCATE(nq2d)
    IF (ALLOCATED(qw2d)) DEALLOCATE(qw2d)
    IF (ALLOCATED(qx2d)) DEALLOCATE(qx2d)
    IF (ALLOCATED(qy2d)) DEALLOCATE(qy2d)
    !--------------------------------------------------------------------
    !---- Select rule
    !--------------------------------------------------------------------
    ALLOCATE(nq2d(5))
    nq2d = RESHAPE ( (/25, 25, 25, 33, 42/),(/5/) )

    SELECT CASE(ndof)
    CASE(:34)
       n = 25 

       ALLOCATE(qx2d(n))
       ALLOCATE(qy2d(n))
       ALLOCATE(qw2d(n))

       qx2d = qx25
       qy2d = qy25
       qw2d = qw25
    CASE(35)
       n = 33

       ALLOCATE(qx2d(n))
       ALLOCATE(qy2d(n))
       ALLOCATE(qw2d(n))

       qx2d = qx33
       qy2d = qy33
       qw2d = qw33
    CASE(56)
       n = 42

       ALLOCATE(qx2d(n))
       ALLOCATE(qy2d(n))
       ALLOCATE(qw2d(n))

       qx2d = qx42
       qy2d = qy42
       qw2d = qw42
    CASE DEFAULT
       WRITE(6, '(a)')    'Error in QUADRATURES_DUNAVANT_2D'
       WRITE(6, '(a,i0,a)') 'Quadrature rule for ndof=',ndof,'was not planned'
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE QUADRATURES_DUNAVANT_2D

  !======================================================================
  !======================================================================
  !======================================================================

END MODULE QUADRATURES_2D
