MODULE QUADRATURES_3D

  !======================================================================
  ! Quadrature rules for tetrahedra and hexahedra
  ! For tetrahedra : nearly-optimal quadrature rules    from order 1 to 13, from [1]
  ! For hexahedra  : classical gauss points integration from order 1 to 4
  !
  ! [1] A set of symmetric quadrature rules on triangles and tetrahedra
  !     Zhang et al., JCM 27 (2009), 89-96
  !======================================================================

  !----------------------------------------------------------------------
  ! Shared modules
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  !----------------------------------------------------------------------
  IMPLICIT NONE
  !----------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Global arrays
  !----------------------------------------------------------------------
  ! nq3d             ==> number of quadrature points
  ! qx3d, qy3d, qz3d ==> quadrature points 
  ! qw3d             ==> weights
  !----------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: nq3d

  REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: qx3d, qy3d, qz3d, qw3d
  !----------------------------------------------------------------------

CONTAINS

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE BUILD_QUADRATURE_3D(cellType, quadOrder)
    !--------------------------------------------------------------------
    !---- Call quadrature routine depending on cell type and order
    !
    !---- IN
    !---- cellType : type of cell
    !---- order    : order of the integrand
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    CHARACTER(KIND=C_CHAR, LEN=*), INTENT(IN)    :: cellType
    
    INTEGER(C_INT),                INTENT(INOUT) :: quadOrder
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Call right quadrature routines
    !--------------------------------------------------------------------
    SELECT CASE(cellType)
    CASE('TET','TET1','TET2','TET3','TET4')
       CALL QUADRATURES_TETRAHEDRON(quadOrder)
    CASE('HEX','HEX1','HEX2')
       CALL QUADRATURES_HEXAHEDRON (quadOrder)
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE BUILD_QUADRATURE_3D

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE QUADRATURES_TETRAHEDRON(quadOrder)
    !--------------------------------------------------------------------
    !---- Initialize quadrature rule depending on integrand order
    !
    !---- IN
    !---- order : order of the integrand
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),      INTENT(IN) :: quadOrder
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                  :: nPtsPermTot
    INTEGER(C_INT), DIMENSION(5)    :: nPtsPermTet

    REAL(C_DOUBLE)                  :: a, b, c, w
    REAL(C_DOUBLE), DIMENSION(24,3) :: temp1111
    REAL(C_DOUBLE), DIMENSION(6,3)  :: temp22
    REAL(C_DOUBLE), DIMENSION(12,3) :: temp211
    REAL(C_DOUBLE), DIMENSION(4,3)  :: temp31
    REAL(C_DOUBLE), DIMENSION(1,3)  :: temp4
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !---- Deallocate arrays if they were allocated previously
    IF (ALLOCATED(nq3d)) DEALLOCATE(nq3d)
    IF (ALLOCATED(qw3d)) DEALLOCATE(qw3d)
    IF (ALLOCATED(qx3d)) DEALLOCATE(qx3d)
    IF (ALLOCATED(qy3d)) DEALLOCATE(qy3d)
    IF (ALLOCATED(qz3d)) DEALLOCATE(qz3d)    
    !
    !---- Check the required quadrature order
    IF ((quadOrder < 0) .OR. (quadOrder > 13)) THEN
       WRITE(6, '(a)')      'Error in QUADRATURES_3D'
       WRITE(6, '(a,i0,a)') 'Quadrature rule of order ',quadOrder,' is &
            not available on tetrahedra'
       CALL TILT
    ENDIF
    !--------------------------------------------------------------------
    !---- Allocate and initialize arrays
    !--------------------------------------------------------------------
    !---- Fill nq3d and nPtsPermTet
    ALLOCATE(nq3d(0:13))
    nq3d        = (/1, 1, 4, 8, 14, 14, 24, 37, 47, 63, 100, 125, 170, 171/)
    nPtsPermTet = (/1, 4, 6, 12, 24/)
    !
    !---- Allocate weights and points arrays
    ALLOCATE(qx3d(nq3d(quadOrder)))
    ALLOCATE(qy3d(nq3d(quadOrder)))
    ALLOCATE(qz3d(nq3d(quadOrder)))    
    ALLOCATE(qw3d(nq3d(quadOrder)))
    !--------------------------------------------------------------------
    !---- Build appropriate rule
    !--------------------------------------------------------------------
    SELECT CASE (quadOrder)
    CASE(0,1)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.25
       w = 1
       CALL PERM4(a, w, temp4, nptspermtot)
    CASE(2)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.1381966011250105
       w = 0.25
       CALL PERM31(a, w, temp31, nptspermtot)
    CASE(3)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.05774267960122022
       w = 0.05063194708616196
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Deuxieme permutation
       a = 0.329292837955806
       w = 0.199368052913838
       CALL PERM31(a, w, temp31, nptspermtot)
    CASE(4)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.1005262177656247
       w = 0.08858863962466471
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Deuxieme permutation
       a = 0.3143725697163862
       w = 0.13283797664525
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.4999957232733486
       w = 0.01904892248672353
       CALL PERM22(a, w, temp22, nptspermtot)
    CASE(5)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.09273525031089123
       w = 0.07349304311636194
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Deuxieme permutation
       a = 0.3108859192633006
       w = 0.1126879257180158
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.04550370412564965
       w = 0.04254602077708147
       CALL PERM22(a, w, temp22, nptspermtot)
    CASE(6)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.214602871259152
       w = 0.03992275025816749
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Deuxieme permutation
       a = 0.04067395853461135
       w = 0.01007721105532064
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.3223378901422755
       w = 0.05535718154365472
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Quatrieme permutation
       a = 0.06366100187501753
       b = 0.6030056647916492  
       w = 0.04821428571428571
       CALL PERM211(a, b, w, temp211, nptspermtot)
    CASE(7)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.25
       w = 0.04172292858868213
       CALL PERM4(a, w, temp4, nptspermtot)
       !---- Deuxieme permutation
       a = 0.03315743587212305
       w = 0.003717829037757149
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.1679454852650537
       w = 0.06874106165718033
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Quatrieme permutation
       a = 0.3238703232369335
       w = 0.04795369883231083
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Cinquieme permutation
       a = 0.0563270344455036
       b = 0.5543027261610421
       w = 0.02746342330987018
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Sixieme permutation
       a = 0.1274565137928612
       b = 0.7435480043814663
       w = 0.01225546946532354
       CALL PERM211(a, b, w, temp211, nptspermtot)
    CASE(8)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.25
       w = 0.04100862889077436
       CALL PERM4(a, w, temp4, nptspermtot)
       !---- Deuxieme permutation
       a = 0.1005904273170712
       w = 0.0191028297018
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.04344030090364619
       w = 0.007947849249927197
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Quatrieme permutation
       a = 0.3122798639971374
       w  = 0.04913816234934988
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Cinquieme permutation
       a = 0.1606066182435581
       w = 0.04605405808207845
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Sixieme permutation
       a = 0.06733495202929608
       w = 0.03791457544688722
       CALL PERM22(a, w, temp22, nptspermtot)
       !---- Septieme permutation
       a = 0.2037430994820635
       b = 0.5924325063970318
       w = 0.01299071064349277
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Huitieme permutation
       a = 0.0211338563946341
       b = 0.7097963926340775
       w = 0.007220316097780584
       CALL PERM211(a, b, w, temp211, nptspermtot)
    CASE(9)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.25
       w = 0.04584923392595652
       CALL PERM4(a, w, temp4, nptspermtot)
       !---- Deuxieme permutation
       a = 0.04328588551777821
       w = 0.007911813159370779
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.3184507199244446
       w = 0.02736326188709693
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Quatrieme permutation
       a = 0.03527963995391836
       w = 0.01072484669649231
       CALL PERM22(a, w, temp22, nptspermtot)
       !---- Cinquieme permutation
       a = 0.168981566420787
       b = 0.6269521727699744
       w = 0.01944627562376205
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Sixieme permutation
       a = 0.0291307329786369
       b = 0.720099894142564
       w = 0.008201508539659981
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Septieme permutation
       a = 0.125475842696303
       b = 0.4580657071255869
       w = 0.02603615811872117
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Huitieme permutation
       a = 0.4233320931306423
       b = 0.1466055438449103
       w = 0.008707839860291699
       CALL PERM211(a, b, w, temp211, nptspermtot)
    CASE(10)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.02701885861980682
       w = 0.002282970499933636
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Deuxieme permutation
       a = 0.1595878923908152
       b = 0.3795512012163156
       w = 0.01799803661247595
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Troisieme permutation
       a = 0.1042924754172202
       b = 0.7682504796110421
       w = 0.006714059185295455
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Quatrieme permutation
       a = 0.3800083425124605
       b = 0.2045945545429161
       w = 0.01708548436666846
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Cinquieme permutation
       a = 0.03346325863204042
       b = 0.5469430971984417
       w = 0.007214953928945244
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Sixieme permutation
       a = 0.01545942343587724
       b = 0.7711779584545552
       w = 0.002826635682743845
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Septieme permutation
       a = 0.09976653876015035
       b = 0.577942197506615
       w = 0.0163394048998746
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Huitieme permutation
       a = 0.2746466244699153
       b = 0.5765369132610021
       c = 0.01617266990221066
       w = 0.007196884245342613
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)
    CASE(11)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.25
       w = 0.02458301399002181 
       CALL PERM4(a, w, temp4, nptspermtot)
       !---- Deuxieme permutation
       a = 0.0294973586175923
       w = 0.002565421440974963
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = 0.1298601367473586
       b = 0.7145872539759059
       w = 0.006520675563038226
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Quatrieme permutation
       a = 0.09638366419593351
       b = 0.2069291740327215
       w = 0.01120908627598861
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Cinquieme permutation
       a = 0.1498208764571236
       b = 0.3955693457216548
       w = 0.01449460098478238
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Sixieme permutation
       a = 0.05471313648651016
       b = 0.5243472683385225
       w = 0.0100967780972099
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Septieme permutation
       a = 0.007309496236795716
       b = 0.6231227257513769
       w = 0.001504945531490739
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Huitieme permutation
       a = 0.02706573735276358
       b = 0.7936178324242757
       w = 0.00459155464778553
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Neuvieme permutation
       a = 0.3883137744146376
       b = 0.206399604283082
       w = 0.007577095342937119
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Dixieme permutation
       a = 0.2405635567961802
       b = 0.458656622408461
       w = 0.01641256873094032
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Onzieme permutation
       a = 0.5859115935740754
       b = 0.2958922515683304
       c = 0.1148101126811963
       w = 0.004011151589833515
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)
    CASE(12)
       nptspermtot = 0
       !---- Premiere permutation
       a = 0.1042446353412517
       w = 0.01322315541577823
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Deuxieme permutation
       a = 0.01519974572866304
       b = 0.6795580958398778
       w = 0.002292996365780243
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Troisieme permutation
       a = 0.2982058150471994
       b = 0.2684444462052035
       w = 0.008989932139025993
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Quatrieme permutation
       a = 0.4364652600505872
       b = 0.1265726129088264
       w = 0.003170444269943039
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Cinquieme permutation
       a = 0.2013060251415608
       b = 0.4837137060053495
       w = 0.01188304532460262
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Sixieme permutation
       a = 0.0279130588555698
       b = 0.8259181061152207
       w = 0.003720097652261537
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Septieme permutation
       a = 0.0129361860806406
       b = 0.9360530754434968
       w = 0.0003607231874108994
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Huitieme permutation
       a = 0.2646735532244872
       b = 0.4466740915306313
       w = 0.01062644292424726
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Neuvieme permutation
       a = 0.03011582913691036
       b = 0.4777900544019199
       w = 0.003176332481752918
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Dixieme permutation
       a = 0.1233855359711034
       b = 0.7489046563394141
       w = 0.002129148559768063
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Onzieme permutation
       a = 0.6168477271123808
       b = 0.1054999685947271
       c = 0.02799890670055219
       w = 0.008637971430698634
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)
       !---- Douzieme permutation 
       a = 0.4465571013004676
       b = 0.1347310059903672
       c = 0.07324398914644378
       w = 0.007650254547608708
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)
    CASE(13)
       nptspermtot = 0
       !---- Premiere permutation
       a = .25
       w = .01501368777308314675062970631615983
       CALL PERM4(a, w, temp4, nptspermtot)
       !---- Deuxieme permutation
       a = .15521609351908950314115784335704739
       w = .01822520928017342532379068941490097
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Troisieme permutation
       a = .33012266333967360024433192595196779
       w = .00700610921774146424038518693926311
       CALL PERM31(a, w, temp31, nptspermtot)
       !---- Quatrieme permutation
       a = .16680640389386249928937782601144234
       w = .01642354974394954829540573107905531
       CALL PERM22(a, w, temp22, nptspermtot)
       !---- Cinquieme permutation
       a = .02492378854777361779701400374860089
       w = .00512061009636059707262596949702171
       CALL PERM22(a, w, temp22, nptspermtot)
       !---- Sixieme permutation
       a = .09719762991575100143072243716240818
       w = .01119669865290491634382032086351956
       CALL PERM22(a, w, temp22, nptspermtot)
       !---- Septieme permutation
       a = .24785929015736256692746910620827934
       b = .43365324235685144718726061434767377
       w = .01561914973337995400953811302431969
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Huitieme permutation
       a = .02223159608186700290879521860892929
       b = .83690032040373400514509486595698594
       w = .00248442301331647441904056776338473
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Neuvieme permutation
       a = .10727869331305341049150459639584801
       b = .77498030597500180756587877274179289
       w = .00163859853481823893844525309440751
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Dixieme permutation
       a = .19817684388398981142331840582142759
       b = .58756930578220530259172017903595920
       w = .00590303044012492197171914655535865
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Onzieme permutation
       a = .06917924347737931647732534347465502
       b = .60420006666006644707935264871115302
       w = .01102208245821805240445097989201525
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Douzieme permutation 
       a = .02311471947193316000000000000001000
       b = .93087579279244424864920228882888307
       w = .00040645183996417822585155512755848
       CALL PERM211(a, b, w, temp211, nptspermtot)
       !---- Treizieme permutation 
       a = .11788928751019608922290117470644250
       b = .11651536422540720000000000000001000
       c = .04202400112551542095676634303719997
       w = .00268796997296854209745781926651729
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)
       !---- Quatorzieme permutation 
       a = .67703279860228426355032221326746594
       b = .04616537602461971083458041122176081
       c = .00084434031890503975729899692135905
       w = .00197950480552671190531894675510740
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)
       !---- Quinzieme permutation 
       a = .48489008867363312201080094154790828
       b = .35888294295520201572423646909421086
       c = .13818283491762872996955080907912355
       w = .00544631918142579120943187040108667
       CALL PERM1111(a, b, c, w, temp1111, nptspermtot)       
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE QUADRATURES_TETRAHEDRON

  !======================================================================
  !======================================================================
  !======================================================================
  
  SUBROUTINE QUADRATURES_HEXAHEDRON(quadOrder)
    !--------------------------------------------------------------------
    !---- Selects the quadrature rule according to the order of quadrature
    !
    !---- IN
    !---- quadOrder : order of the integrand
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    USE QUADRATURES_1D
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: quadOrder
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)             :: i, j, k, com
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !---- Deallocate arrays if they were allocated previously
    IF (ALLOCATED(nq3d)) DEALLOCATE(nq3d)    
    IF (ALLOCATED(qx3d)) DEALLOCATE(qx3d)
    IF (ALLOCATED(qy3d)) DEALLOCATE(qy3d)
    IF (ALLOCATED(qz3d)) DEALLOCATE(qz3d)
    IF (ALLOCATED(qw3d)) DEALLOCATE(qw3d)
    !
    !---- Check the required quadrature order
    IF ((quadOrder < 0) .OR. (quadOrder > 8)) THEN
       WRITE(6, '(a)')      'Error in QUADRATURES_3D'
       WRITE(6, '(a,i0,a)') 'Quadrature rule of order ',quadOrder,' is not &
            available on hexahedra'
       CALL TILT
    ENDIF    
    !--------------------------------------------------------------------
    !---- Allocate and initialize arrays
    !--------------------------------------------------------------------
    !---- Fill nq3d
    ALLOCATE(nq3d(0:8))
    nq3d = (/1, 8, 8, 27, 27, 64, 64, 125, 125/)
    !
    !---- Allocate weights and points arrays
    ALLOCATE(qx3d(nq3d(quadOrder)))
    ALLOCATE(qy3d(nq3d(quadOrder)))
    ALLOCATE(qz3d(nq3d(quadOrder)))    
    ALLOCATE(qw3d(nq3d(quadOrder)))    
    !--------------------------------------------------------------------
    !---- Build appropriate rule
    !--------------------------------------------------------------------
    com = 0
    !
    SELECT CASE (quadOrder)
    CASE(0)
       qx3d(:) = 0.0
       qy3d(:) = 0.0
       qz3d(:) = 0.0
       qw3d(:) = 8.0
    CASE(1:2)
       DO i=1,2
          DO j=1,2
             DO k=1,2
                com       = com + 1 
                qx3d(com) = q1p2(1,i) 
                qy3d(com) = q1p2(1,j)
                qz3d(com) = q1p2(1,k)
                qw3d(com) = q1p2(2,i)*q1p2(2,j)*q1p2(2,k)
             ENDDO
          ENDDO
       ENDDO       
    CASE(3:4)
       DO i=1,3
          DO j=1,3
             DO k=1,3
                com       = com + 1
                qx3d(com) = q1p5(1,i)
                qy3d(com) = q1p5(1,j)
                qz3d(com) = q1p5(1,k)
                qw3d(com) = q1p5(2,i)*q1p5(2,j)*q1p5(2,k)
             ENDDO
          ENDDO
       ENDDO       
    CASE(5:6)
       DO i=1,4
          DO j=1,4
             DO k=1,4
                com       = com + 1
                qx3d(com) = q1p7(1,i)
                qy3d(com) = q1p7(1,j)
                qz3d(com) = q1p7(1,k)
                qw3d(com) = q1p7(2,i)*q1p7(2,j)*q1p7(2,k)
             ENDDO
          ENDDO
       ENDDO       
    CASE(7:8)
       DO i=1,5
          DO j=1,5
             DO k=1,5
                com       = com + 1
                qx3d(com) = q1p9(1,i)
                qy3d(com) = q1p9(1,j)
                qz3d(com) = q1p9(1,k)
                qw3d(com) = q1p9(2,i)*q1p9(2,j)*q1p9(2,k)
             ENDDO
          ENDDO
       ENDDO       
    END SELECT
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE QUADRATURES_HEXAHEDRON  

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERMCALCTET(xi1, xi2, xi3, xi4, x)
    !--------------------------------------------------------------------
    !---- Compute point coordinates from permutation data
    !---- This routine is called by each permutation routine below
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN)  :: xi1, xi2, xi3, xi4
    REAL(C_DOUBLE), INTENT(OUT) :: x(3)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE) :: refc(4,3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill refc
    refc(1,1) = 0.0
    refc(1,2) = 0.0
    refc(1,3) = 0.0
    refc(2,1) = 1.0
    refc(2,2) = 0.0
    refc(2,3) = 0.0
    refc(3,1) = 0.0
    refc(3,2) = 1.0
    refc(3,3) = 0.0
    refc(4,1) = 0.0
    refc(4,2) = 0.0
    refc(4,3) = 1.0
    !
    !---- Compute coordinates
    x(1) = xi1*refc(1,1) + xi2*refc(2,1) + xi3*refc(3,1) + xi4*refc(4,1)
    x(2) = xi1*refc(1,2) + xi2*refc(2,2) + xi3*refc(3,2) + xi4*refc(4,2)
    x(3) = xi1*refc(1,3) + xi2*refc(2,3) + xi3*refc(3,3) + xi4*refc(4,3)    
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERMCALCTET

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM4(a, w, temp4, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm4
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp4(1,3)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, x(3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    !
    !---- Call PERMCALCTRI and store point 1
    CALL PERMCALCTET(xi1, xi1, xi1, xi1, x)
    !
    DO i=1,3
       temp4(1,i) = x(i)
    ENDDO
    !
    !---- Fill points and weights
    DO i=1,1
       qx3d(nPtsPermTot+i) = temp4(i,1)
       qy3d(nPtsPermTot+i) = temp4(i,2)
       qz3d(nPtsPermTot+i) = temp4(i,3)
       qw3d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 1
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM4

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM31(a, w, temp31, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm31
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp31(4,3)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, xi2, x(3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    xi2 = 1.0 - 3.0*a
    !
    !---- Call PERMCALCTET and store point 1
    CALL PERMCALCTET(xi1, xi1, xi1, xi2, x)
    !
    DO i=1,3
       temp31(1,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 2
    CALL PERMCALCTET(xi1, xi1, xi2, xi1, x)
    !
    DO i=1,3
       temp31(2,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 3
    CALL PERMCALCTET(xi1, xi2, xi1, xi1, x)
    !
    DO i=1,3
       temp31(3,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 4
    CALL PERMCALCTET(xi2, xi1, xi1, xi1, x)
    !
    DO i=1,3
       temp31(4,i) = x(i)
    ENDDO    
    !
    !---- Fill points and weights
    DO i=1,4
       qx3d(nPtsPermTot+i) = temp31(i,1)
       qy3d(nPtsPermTot+i) = temp31(i,2)
       qz3d(nPtsPermTot+i) = temp31(i,3)       
       qw3d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 4
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM31

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM22(a, w, temp22, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm22
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp22(6,3)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, xi2, x(3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    xi2 = 0.5 - a
    !
    !---- Call PERMCALCTET and store point 1
    CALL PERMCALCTET(xi1, xi1, xi2, xi2, x)
    !
    DO i=1,3
       temp22(1,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 2
    CALL PERMCALCTET(xi1, xi2, xi1, xi2, x)
    !
    DO i=1,3
       temp22(2,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 3
    CALL PERMCALCTET(xi2, xi1, xi1, xi2, x)
    !
    DO i=1,3
       temp22(3,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 4
    CALL PERMCALCTET(xi1, xi2, xi2, xi1, x)
    !
    DO i=1,3
       temp22(4,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 5
    CALL PERMCALCTET(xi2, xi1, xi2, xi1, x)
    !
    DO i=1,3
       temp22(5,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 6
    CALL PERMCALCTET(xi2, xi2, xi1, xi1, x)
    !
    DO i=1,3
       temp22(6,i) = x(i)
    ENDDO        
    !
    !---- Fill points and weights
    DO i=1,6
       qx3d(nPtsPermTot+i) = temp22(i,1)
       qy3d(nPtsPermTot+i) = temp22(i,2)
       qz3d(nPtsPermTot+i) = temp22(i,3)       
       qw3d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 6
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM22

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM211(a, b, w, temp211, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm211
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, b, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp211(12,3)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, xi2, xi3, x(3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    xi2 = b
    xi3 = 1.0 - 2.0*a - b
    !
    !---- Call PERMCALCTET and store point 1
    CALL PERMCALCTET(xi1, xi1, xi2, xi3, x)
    !
    DO i=1,3
       temp211(1,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 2
    CALL PERMCALCTET(xi1, xi2, xi1, xi3, x)
    !
    DO i=1,3
       temp211(2,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 3
    CALL PERMCALCTET(xi2, xi1, xi1, xi3, x)
    !
    DO i=1,3
       temp211(3,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 4
    CALL PERMCALCTET(xi1, xi1, xi3, xi2, x)
    !
    DO i=1,3
       temp211(4,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 5
    CALL PERMCALCTET(xi1, xi2, xi3, xi1, x)
    !
    DO i=1,3
       temp211(5,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 6
    CALL PERMCALCTET(xi2, xi1, xi3, xi1, x)
    !
    DO i=1,3
       temp211(6,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 7
    CALL PERMCALCTET(xi1, xi3, xi1, xi2, x)
    !
    DO i=1,3
       temp211(7,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 8
    CALL PERMCALCTET(xi1, xi3, xi2, xi1, x)
    !
    DO i=1,3
       temp211(8,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 9
    CALL PERMCALCTET(xi2, xi3, xi1, xi1, x)
    !
    DO i=1,3
       temp211(9,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 10
    CALL PERMCALCTET(xi3, xi1, xi1, xi2, x)
    !
    DO i=1,3
       temp211(10,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 11
    CALL PERMCALCTET(xi3, xi1, xi2, xi1, x)
    !
    DO i=1,3
       temp211(11,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 12
    CALL PERMCALCTET(xi3, xi2, xi1, xi1, x)
    !
    DO i=1,3
       temp211(12,i) = x(i)
    ENDDO    
    !
    !---- Fill points and weights
    DO i=1,12
       qx3d(nPtsPermTot+i) = temp211(i,1)
       qy3d(nPtsPermTot+i) = temp211(i,2)
       qz3d(nPtsPermTot+i) = temp211(i,3)       
       qw3d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 12
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM211

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE PERM1111(a, b, c, w, temp1111, nPtsPermTot)
    !--------------------------------------------------------------------
    !---- Define permutation perm1111
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: nPtsPermTot

    REAL(C_DOUBLE), INTENT(IN)  :: a, b, c, w
    REAL(C_DOUBLE), INTENT(OUT) :: temp1111(24,3)
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: xi1, xi2, xi3, xi4, x(3)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Compute coordinates
    !--------------------------------------------------------------------
    !---- Fill xi
    xi1 = a
    xi2 = b
    xi3 = c
    xi4 = 1.0 - a - b - c
    !
    !---- Call PERMCALCTET and store point 1
    CALL PERMCALCTET(xi1, xi2, xi3, xi4, x)
    !
    DO i=1,3
       temp1111(1,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 2
    CALL PERMCALCTET(xi1, xi2, xi4, xi3, x)
    !
    DO i=1,3
       temp1111(2,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 3
    CALL PERMCALCTET(xi1, xi4, xi2, xi3, x)
    !
    DO i=1,3
       temp1111(3,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 4
    CALL PERMCALCTET(xi4, xi1, xi2, xi3, x)
    !
    DO i=1,3
       temp1111(4,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 5
    CALL PERMCALCTET(xi1, xi3, xi2, xi4, x)
    !
    DO i=1,3
       temp1111(5,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 6
    CALL PERMCALCTET(xi1, xi3, xi4, xi2, x)
    !
    DO i=1,3
       temp1111(6,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 7
    CALL PERMCALCTET(xi1, xi4, xi3, xi2, x)
    !
    DO i=1,3
       temp1111(7,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 8
    CALL PERMCALCTET(xi4, xi1, xi3, xi2, x)
    !
    DO i=1,3
       temp1111(8,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 9
    CALL PERMCALCTET(xi3, xi1, xi2, xi4, x)
    !
    DO i=1,3
       temp1111(9,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 10
    CALL PERMCALCTET(xi3, xi1, xi4, xi2, x)
    !
    DO i=1,3
       temp1111(10,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 11
    CALL PERMCALCTET(xi3, xi4, xi1, xi2, x)
    !
    DO i=1,3
       temp1111(11,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 12
    CALL PERMCALCTET(xi4, xi3, xi1, xi2, x)
    !
    DO i=1,3
       temp1111(12,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 13
    CALL PERMCALCTET(xi2, xi1, xi3, xi4, x)
    !
    DO i=1,3
       temp1111(13,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 14
    CALL PERMCALCTET(xi2, xi1, xi4, xi3, x)
    !
    DO i=1,3
       temp1111(14,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 15
    CALL PERMCALCTET(xi2, xi4, xi1, xi3, x)
    !
    DO i=1,3
       temp1111(15,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 16
    CALL PERMCALCTET(xi4, xi2, xi1, xi3, x)
    !
    DO i=1,3
       temp1111(16,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 17
    CALL PERMCALCTET(xi2, xi3, xi1, xi4, x)
    !
    DO i=1,3
       temp1111(17,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 18
    CALL PERMCALCTET(xi2, xi3, xi4, xi1, x)
    !
    DO i=1,3
       temp1111(18,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 19
    CALL PERMCALCTET(xi2, xi4, xi3, xi1, x)
    !
    DO i=1,3
       temp1111(19,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 20
    CALL PERMCALCTET(xi4, xi2, xi3, xi1, x)
    !
    DO i=1,3
       temp1111(20,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 21
    CALL PERMCALCTET(xi3, xi2, xi1, xi4, x)
    !
    DO i=1,3
       temp1111(21,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 22
    CALL PERMCALCTET(xi3, xi2, xi4, xi1, x)
    !
    DO i=1,3
       temp1111(22,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 23
    CALL PERMCALCTET(xi3, xi4, xi2, xi1, x)
    !
    DO i=1,3
       temp1111(23,i) = x(i)
    ENDDO
    !
    !---- Call PERMCALCTET and store point 24
    CALL PERMCALCTET(xi4, xi3, xi2, xi1, x)
    !
    DO i=1,3
       temp1111(24,i) = x(i)
    ENDDO        
    !
    !---- Fill points and weights
    DO i=1,24
       qx3d(nPtsPermTot+i) = temp1111(i,1)
       qy3d(nPtsPermTot+i) = temp1111(i,2)
       qz3d(nPtsPermTot+i) = temp1111(i,3)       
       qw3d(nPtsPermTot+i) = w
    ENDDO
    !
    nPtsPermTot = nPtsPermTot + 24
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE PERM1111

  !======================================================================
  !======================================================================
  !======================================================================

END MODULE QUADRATURES_3D
