MODULE RAW_MESH

  !========================================================================
  ! Module containing the subroutines related to Raw mesh format
  !========================================================================

  !------------------------------------------------------------------------
  ! Shared modules
  !------------------------------------------------------------------------
  USE ISO_C_BINDING
  !------------------------------------------------------------------------
  IMPLICIT NONE
  !------------------------------------------------------------------------

CONTAINS

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE READ_RAW_MESH
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    USE MESH_FACES
    USE MESH_CELLS
    USE IO_FILES
    USE MESH_ELEMENTS
    USE REFERENCE_CELLS
    USE DOF_PERMUTATIONS
    USE PARALLEL_STRUCTURES
    USE MESH_TYPE_CONVERSIONS
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)             :: type, faceType

    LOGICAL(C_BOOL)                           :: isBoundary

    INTEGER(C_INT)                            :: order, nDof, nDofTot,  &
         error, iDof, log, nNeighb, nUnit, nVerts, nLinVerts, nHoVerts, &
         geoOrder, iNeighb, globalIndex, iCell
    INTEGER(C_INT), DIMENSION(nvCellMax)      :: iBuffer
    INTEGER(C_INT), DIMENSION(nvFaceMax)      :: faceVerts    
    INTEGER(C_INT), DIMENSION(99)             :: neighbors, faceTags

    REAL(C_DOUBLE), DIMENSION(3)              :: rBuffer

    TYPE(cell),     DIMENSION(:), ALLOCATABLE :: locCell
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Reading and filling Mesh arrays
    !----------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    nDofTot   = 0
    neighbors = 0
    OPEN(UNIT=nUnit, FILE=TRIM(Mesh%name)//'.rawMesh')
    !
    Readingloop: DO
       iBuffer(:) = 0
       READ(nUnit, *, IOSTAT=error) type
       IF (error /= 0) EXIT
       !
       READ(nUnit, *) order, nDof, log, nNeighb
       READ(nUnit, *) globalIndex, neighbors(1:nNeighb)
       READ(nUnit, *) faceTags(1:nNeighb)
       !
       !---- Add vertices
       DO iDof=1,nDof
          READ(nUnit,*) rBuffer(1:3)
          CALL ADD_VERTEX(rBuffer(1:3))
          nDofTot = nDofTot + 1
          iBuffer(iDof) = nDofTot
       ENDDO
       !
       !---- Add element
       CALL ADD_ELEMENT(iBuffer(1:nDof), log, type)
       ALLOCATE(MeshCell(Mesh%nCell)%facLog(nNeighb))
       MeshCell(Mesh%nCell)%facLog(1:nNeighb) = faceTags(1:nNeighb)
       !
       !---- Fill neighbors
       MeshCell(Mesh%nCell)%globalIndex = globalIndex
       MeshCell(Mesh%nCell)%nNeighb     = nNeighb
       IF (ALLOCATED(MeshCell(Mesh%nCell)%neighb)) DEALLOCATE(MeshCell(Mesh%nCell)%neighb)
       ALLOCATE(MeshCell(Mesh%nCell)%neighb(nNeighb))
       MeshCell(Mesh%nCell)%neighb(1:nNeighb) = neighbors(1:nNeighb)
    ENDDO Readingloop
    !
    CLOSE(nUnit)
    !----------------------------------------------------------------------
    !---- Re-order cells according to global indexing
    !---- This procedure allows to reconstruct material interfaces
    !---- through stored neighbor indices while maintaining vertices
    !---- duplication (each cell has its own vertices)
    !----------------------------------------------------------------------
    ALLOCATE(locCell(Mesh%nCell))
    !
    DO iCell=1,Mesh%nCell
       globalIndex          = MeshCell(iCell)%globalIndex
       locCell(globalIndex) = MeshCell(iCell)
    ENDDO
    !
    DO iCell=1,Mesh%nCell
       globalIndex                 = MeshCell(iCell)%globalIndex
       MeshCell(iCell)             = locCell(iCell)
       MeshCell(iCell)%globalIndex = globalIndex
    ENDDO
    !
    DEALLOCATE(locCell)
    !----------------------------------------------------------------------
    IF (idWrite) WRITE(6,'(a)') 'Finished mesh reading'
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE READ_RAW_MESH

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_RAW_MESH(orderArray)
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE BASES
    USE IO_FILES
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE MESH_ELEMENTS
    USE MESH_ELEMENTS_DOFS
    USE REFERENCE_CELLS
    USE MESH_FACES_LOGS
    USE MESH_TYPE_CONVERSIONS
    USE PARALLEL_STRUCTURES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(:), INTENT(IN) :: orderArray
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)            :: cellType
    CHARACTER(LEN=1000, KIND=C_CHAR)         :: fileName

    INTEGER(C_INT)                           :: nUnit, iCell, nDofV, iDof, &
         order, log, nNeighb, iNeighb, iProc, nCellsSum, nCellsSumGlobal
    INTEGER(C_INT), DIMENSION(99)            :: faceLogs
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    fileName = TRIM(globalMesh%name)//'.rawMesh'
    IF (idProc == mainProc) CALL REMOVE_FILE(TRIM(fileName))
    CALL PARALLEL_BARRIER
    !----------------------------------------------------------------------
    !---- If the run is sequential, the globalNeighb array does not exist,
    !---- so we create it before outputing solution
    !----------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       DO iCell=1,Mesh%nCell
          ALLOCATE(MeshCell(iCell)%globalNeighb(MeshCell(iCell)%nNeighb))
          MeshCell(iCell)%globalNeighb(:) = MeshCell(iCell)%neighb(:)
          MeshCell(iCell)%globalIndex     = iCell
       ENDDO
    ENDIF
    !----------------------------------------------------------------------
    !---- Re-allocate dof* arrays
    !---- This is to prevent segfaults in the case where P0 was used
    !----------------------------------------------------------------------
    IF (ALLOCATED(dofXV)) DEALLOCATE(dofXV)
    IF (ALLOCATED(dofYV)) DEALLOCATE(dofYV)
    IF (ALLOCATED(dofZV)) DEALLOCATE(dofZV)
    !
    ALLOCATE(dofXV(MAX(nDofVMax,nvCellMax)))
    ALLOCATE(dofYV(MAX(nDofVMax,nvCellMax)))
    ALLOCATE(dofZV(MAX(nDofVMax,nvCellMax)))
    !----------------------------------------------------------------------
    !---- Write cells
    !----------------------------------------------------------------------
    !---- nCellsSum is the summed number of cells, it must be right since
    !---- we need to reconstruct neighbors in solution processing
    nCellsSum = 0
    !
    DO iProc=1,nbProc
       IF (idProc == (iProc-1)) THEN
          CALL GET_NEW_FILE_UNIT(nUnit)
          OPEN(UNIT=nUnit, FILE=fileName, POSITION='APPEND')
          !
          DO iCell=1,Mesh%nCell
             order   = orderArray(iCell)
             CALL CHANGE_TYPE_WITH_ORDER(orderArray(iCell), MeshCell(iCell)%cellType, cellType)
             CALL ORDER_TO_NDOFV(order, cellType, nDofV)
             log     = MeshCell(iCell)%log
             nNeighb = MeshCell(iCell)%nNeighb
             !
             !---- Process face logs to remove the "artificial" tag
             faceLogs(1:nNeighb) = MeshFac(MeshCell(iCell)%facNum(1:nNeighb))%log
             DO iNeighb=1,nNeighb
                CALL REMOVE_ARTIFICIAL_LOG(faceLogs(iNeighb))
             ENDDO
             !
             !---- Write general data
             WRITE(nUnit, '(a)'   ) cellType
             WRITE(nUnit, '(4i9)' ) order, nDofV, log, nNeighb
             WRITE(nUnit, '(99i9)') MeshCell(iCell)%globalIndex, MeshCell(iCell)%globalNeighb(:)
             WRITE(nUnit, '(99i9)') faceLogs(1:nNeighb)
             !
             !---- Compute dof positions
             CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
             !
             !---- Write dof positions
             DO iDof=1,nDofV
                WRITE(nUnit, '(3e15.7)') dofXV(iDof), dofYV(iDof), dofZV(iDof)
             ENDDO
          ENDDO
          !
          nCellsSum = nCellsSum + Mesh%nCell
          CLOSE(nUnit)
       ENDIF
       !
       CALL GLOBAL_OP(nCellsSum, MPI_MAX, nCellsSumGlobal)
       nCellsSum = nCellsSumGlobal
       CALL PARALLEL_BARRIER
    ENDDO
    !----------------------------------------------------------------------
    !-- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_RAW_MESH

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_RAW_SOLUTION
    !----------------------------------------------------------------------
    !---- Write the 3D fields in Raw format
    !
    !---- IN
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    USE DOF_PERMUTATIONS
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Local variables
    !---------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: fileName
    CHARACTER(LEN=4, KIND=C_CHAR)    :: n3dViewC

    INTEGER(C_INT)                   :: iDof, iElt, nDofV, nUnit, &
         iProc, order
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    CALL INT_TO_CHAR(n3dView, n3dViewC)
    fileName = TRIM(globalMesh%name)//'_'//TRIM(n3dViewC)//'.rawSol'
    IF (idProc == mainProc) CALL REMOVE_FILE(TRIM(fileName))
    CALL PARALLEL_BARRIER
    !----------------------------------------------------------------------
    !---- Write solution
    !----------------------------------------------------------------------
    DO iProc=1,nbProc
       IF (idProc == (iProc-1)) THEN
          CALL GET_NEW_FILE_UNIT(nUnit)
          OPEN(UNIT=nUnit, FILE=fileName, POSITION='APPEND')
          !
          IF (ALLOCATED(dgEM))  CALL WRITE_RAW_DG_NUMERICAL_ELEMENT (nUnit)
          IF (ALLOCATED(hdgEM)) CALL WRITE_RAW_HDG_NUMERICAL_ELEMENT(nUnit)
          !
          CLOSE(nUnit)
       ENDIF
       !
       CALL PARALLEL_BARRIER
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_RAW_SOLUTION

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_RAW_DG_NUMERICAL_ELEMENT(nUnit)
    !----------------------------------------------------------------------
    !---- Write 3D DG numerical elements
    !
    !---- IN
    !---- nUnit : unit of file to write in
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE NUMERICAL_TYPES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nUnit
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)             :: iCell, iDof
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Write solution
    !----------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       WRITE(nUnit, '(2i9)') dgEM(iCell)%order, dgEM(iCell)%nDofV
       DO iDof=1,dgEM(iCell)%nDofV
          WRITE(nUnit, '(3e15.7)') dgEM(iCell)%E%vec(iDof,:)
       ENDDO
       DO iDof=1,dgEM(iCell)%nDofV
          WRITE(nUnit, '(3e15.7)') dgEM(iCell)%H%vec(iDof,:)
       ENDDO
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_RAW_DG_NUMERICAL_ELEMENT

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE WRITE_RAW_HDG_NUMERICAL_ELEMENT(nUnit)
    !----------------------------------------------------------------------
    !---- Write 3D HDG numerical elements
    !
    !---- IN
    !---- nUnit : unit of file to write in
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE NUMERICAL_TYPES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: nUnit
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT)             :: iCell, iDof
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Write solution
    !----------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       WRITE(nUnit, '(2i9)') hdgEM(iCell)%order, hdgEM(iCell)%nDofV
       DO iDof=1,hdgEM(iCell)%nDofV
          WRITE(nUnit, '(3e15.7)') ABS(hdgEM(iCell)%E%vec(iDof,:))
       ENDDO
       DO iDof=1,hdgEM(iCell)%nDofV
          WRITE(nUnit, '(3e15.7)') REAL(hdgEM(iCell)%E%vec(iDof,:))
       ENDDO
    ENDDO
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE WRITE_RAW_HDG_NUMERICAL_ELEMENT

  !========================================================================
  !========================================================================
  !========================================================================

  SUBROUTINE READ_RAW_SOLUTION
    !----------------------------------------------------------------------
    !---- Read the 3D fields in Raw format
    !
    !---- IN
    !
    !---- OUT
    !
    !----------------------------------------------------------------------
    !---- Modules
    !----------------------------------------------------------------------
    USE IO_FILES
    USE MESH_DATA
    USE DOF_PERMUTATIONS
    USE NUMERICAL_TYPES
    !----------------------------------------------------------------------
    !---- I/O variables
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Local variables
    !----------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, iDof, nUnit, globalIndex
    !----------------------------------------------------------------------
    !---- Testing for bad input
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
    !---- Allocate solution fields
    !----------------------------------------------------------------------
    IF (ALLOCATED(dgEM)) DEALLOCATE(dgEM)
    ALLOCATE(dgEM(Mesh%nCell))
    !----------------------------------------------------------------------
    !---- Read solution
    !----------------------------------------------------------------------
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(Mesh%solutionName)//'.rawSol')
    !
    DO iCell=1,Mesh%nCell
       globalIndex = MeshCell(iCell)%globalIndex
       !
       READ(nUnit,*) dgEM(globalIndex)%order, dgEM(globalIndex)%nDofV
       ALLOCATE(dgEM(globalIndex)%E%vec(3,dgEM(globalIndex)%nDofV))
       ALLOCATE(dgEM(globalIndex)%H%vec(3,dgEM(globalIndex)%nDofV))
       !
       DO iDof=1,dgEM(globalIndex)%nDofV
          READ(nUnit,*) dgEM(globalIndex)%E%vec(:,iDof)
       ENDDO
       DO iDof=1,dgEM(globalIndex)%nDofV
          READ(nUnit,*) dgEM(globalIndex)%H%vec(:,iDof)
       ENDDO
    ENDDO
    !
    CLOSE(nUnit)
    !----------------------------------------------------------------------
    !---- Testing for errors
    !----------------------------------------------------------------------
    !
    !----------------------------------------------------------------------
  END SUBROUTINE READ_RAW_SOLUTION

  !========================================================================
  !========================================================================
  !========================================================================

END MODULE RAW_MESH
