MODULE REFERENCE_CELLS

  !======================================================================
  ! Informations about the reference cells
  !======================================================================

  !----------------------------------------------------------------------
  ! Shared module
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  USE MATRICES_STRUCTURES
  !----------------------------------------------------------------------
  IMPLICIT NONE
  !----------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Definition of the reference tetrahedron
  ! Tr := {0<=x,y,z : (x+y+z)<=1}
  !----------------------------------------------------------------------
  REAL(C_DOUBLE), PARAMETER :: Tref_vol  = 1.0d0/6.0d0
  REAL(C_DOUBLE), PARAMETER :: Tref_surf = 1.0d0/2.0d0

  REAL(C_DOUBLE), DIMENSION(4), PARAMETER ::   & 
       Tref_face_area = (/ 0.5d0, 0.5d0, 0.5d0, 0.5d0*SQRT(3.0d0) /)

  REAL(C_DOUBLE), DIMENSION(3), PARAMETER ::   &
       Tref_Vert1 = (/ 0.0d0, 0.0d0, 0.0d0 /), &
       Tref_Vert2 = (/ 1.0d0, 0.0d0, 0.0d0 /), & 
       Tref_Vert3 = (/ 0.0d0, 1.0d0, 0.0d0 /), &
       Tref_Vert4 = (/ 0.0d0, 0.0d0, 1.0d0 /)

  !----------------------------------------------------------------------
  ! Definition of the reference hexahedron
  ! Hr := {0<=(x,y,z)<=1}
  !----------------------------------------------------------------------
  REAL(C_DOUBLE), PARAMETER :: Href_vol  = 8.0d0
  REAL(C_DOUBLE), PARAMETER :: Href_surf = 4.0d0

  REAL(C_DOUBLE), DIMENSION(6), PARAMETER ::   &
       Href_face_area = (/ 4.0d0, 4.0d0, 4.0d0, 4.0d0, 4.0d0, 4.0d0 /)

  REAL(C_DOUBLE), DIMENSION(3), PARAMETER ::   &
       Href_Vert1 = (/-1.0d0,-1.0d0,-1.0d0 /), &
       Href_Vert2 = (/ 1.0d0,-1.0d0,-1.0d0 /), &
       Href_Vert3 = (/ 1.0d0, 1.0d0,-1.0d0 /), &
       Href_Vert4 = (/-1.0d0, 1.0d0,-1.0d0 /), &
       Href_Vert5 = (/-1.0d0,-1.0d0, 1.0d0 /), &
       Href_Vert6 = (/ 1.0d0,-1.0d0, 1.0d0 /), &
       Href_Vert7 = (/ 1.0d0, 1.0d0, 1.0d0 /), &
       Href_Vert8 = (/-1.0d0, 1.0d0, 1.0d0 /)

  !----------------------------------------------------------------------
  ! Definition of some fixed values
  !
  ! nvFaceMax ==> maximal # of vertices in a      face (curved quad)
  ! nvTriMax  ==> maximal # of vertices in a tri  face
  ! nvQuadMax ==> maximal # of vertices in a quad face 
  ! nvCellMax ==> maximal # of vertices in a     cell (curved hex)
  ! nvTetMax  ==> maximal # of vertices in a tet cell 
  ! nvHexMax  ==> maximal # of vertices in a hex cell 
  !----------------------------------------------------------------------  
  INTEGER(C_INT), PARAMETER :: nvFaceMax = 9
  INTEGER(C_INT), PARAMETER :: nvTriMax  = 6
  INTEGER(C_INT), PARAMETER :: nvQuadMax = 9
  INTEGER(C_INT), PARAMETER :: nvCellMax = 27
  INTEGER(C_INT), PARAMETER :: nvTetMax  = 10
  INTEGER(C_INT), PARAMETER :: nvHexMax  = 27
  
  !----------------------------------------------------------------------
  ! Reference matrices on a reference cell
  !
  ! xDof*, yDof*, zDof* ==> dof coordinates
  ! iPermDof*           ==> permutations of the dof on the faces
  ! mass*               ==> mass matrix (M)
  ! iMass*              ==> inverse of the mass matrix (M^-1)
  ! rdx*, rdy*, rdz*    ==> stiffness matrices (Kx, Ky, Kz)
  ! mSurf*              ==> mass matrix on the face z=0 (S)
  ! mProj*              ==> projection matrices from an order to another
  !----------------------------------------------------------------------
  TYPE(vector_real),       DIMENSION(:),   ALLOCATABLE         :: xDofTet, yDofTet, zDofTet
  TYPE(vector_real),       DIMENSION(:),   ALLOCATABLE         :: xDofHex, yDofHex, zDofHex
  TYPE(vector_real),       DIMENSION(:),   ALLOCATABLE         :: xDofTri, yDofTri
  TYPE(vector_real),       DIMENSION(:),   ALLOCATABLE         :: xDofQua, yDofQua

  TYPE(matrix_sq_integer), DIMENSION(:),   ALLOCATABLE, TARGET :: iPermDofTet
  TYPE(matrix_sq_integer), DIMENSION(:),   ALLOCATABLE, TARGET :: iPermDofHex

  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: massTet
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: massHex

  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: iMassTet
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: iMassHex
  
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: rdxTet
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: rdyTet
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: rdzTet
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: rdxHex
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: rdyHex
  TYPE(matrix_sq_real),    DIMENSION(:),   ALLOCATABLE, TARGET :: rdzHex

  TYPE(matrix_rect_real),  DIMENSION(:,:), ALLOCATABLE, TARGET :: mSurfTet
  TYPE(matrix_rect_real),  DIMENSION(:,:), ALLOCATABLE, TARGET :: mSurfHex

  TYPE(matrix_rect_real),  DIMENSION(:,:), ALLOCATABLE, TARGET :: mProjTet
  TYPE(matrix_rect_real),  DIMENSION(:,:), ALLOCATABLE, TARGET :: mProjHex

  !----------------------------------------------------------------------
  ! For the moment, definition of matrix hybrid
  !-----------------------------------------------------------------------
  TYPE(matrix_rect_real),  DIMENSION(:), ALLOCATABLE, TARGET :: mSurfHyb
  TYPE(matrix_rect_real),  DIMENSION(:), ALLOCATABLE, TARGET :: mSurfHybF
  !----------------------------------------------------------------------

END MODULE REFERENCE_CELLS
