MODULE SORT

  !======================================================================
  ! Various sorting routines
  !======================================================================

  !----------------------------------------------------------------------
  ! Shared modules
  !----------------------------------------------------------------------
  USE ISO_C_BINDING
  !----------------------------------------------------------------------
  IMPLICIT NONE
  !----------------------------------------------------------------------

  !----------------------------------------------------------------------
  ! Interface for overloading lexicographic array sorting
  !----------------------------------------------------------------------
  INTERFACE SORT_ARRAY
     MODULE PROCEDURE SORT_ARRAY_INT, SORT_ARRAY_DOUBLE
  END INTERFACE SORT_ARRAY

  !----------------------------------------------------------------------
  ! Interface for overloading insertion sort
  !----------------------------------------------------------------------
  INTERFACE INSERTION_SORT
     MODULE PROCEDURE INSERTION_SORT_INT, INSERTION_SORT_DOUBLE
  END INTERFACE INSERTION_SORT
  !----------------------------------------------------------------------  

CONTAINS

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE INSERTION_SORT_INT(array)
    !--------------------------------------------------------------------
    !---- Sort array by insertion
    !---- This method is faster for very small arrays
    !
    !---- IN
    !---- array : array to sort
    !
    !---- OUT
    !---- array : modified in the process
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), DIMENSION(:), INTENT(INOUT) :: array
    !--------------------------------------------------------------------
    !---- Local variables 
    !--------------------------------------------------------------------
    INTEGER(C_INT)                              :: length, i, j, val
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform sort
    !--------------------------------------------------------------------
    length = SIZE(array)
    !
    DO i=2,length
       j = i
       !
       DO WHILE(array(j-1) > array(j))
          val          = array(j-1)
          array(j-1)   = array(j)
          array(j)     = val
          IF (j > 2) j = j - 1
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE INSERTION_SORT_INT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE INSERTION_SORT_DOUBLE(array)
    !--------------------------------------------------------------------
    !---- Sort array by insertion
    !---- This method is faster for very small arrays
    !
    !---- IN
    !---- array : array to sort
    !
    !---- OUT
    !---- array : modified in the process
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(:), INTENT(INOUT) :: array
    !--------------------------------------------------------------------
    !---- Local variables 
    !--------------------------------------------------------------------
    INTEGER(C_INT)                              :: length, i, j
    
    REAL(C_DOUBLE)                              :: val
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Perform sort
    !--------------------------------------------------------------------
    length = SIZE(array)
    !
    DO i=2,length
       j = i
       !
       DO WHILE(array(j-1) > array(j))
          val          = array(j-1)
          array(j-1)   = array(j)
          array(j)     = val
          IF (j > 2) j = j - 1
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Testing for errors
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE INSERTION_SORT_DOUBLE

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE SORT_ARRAY_INT(m,n,a)
    !--------------------------------------------------------------------
    !---- Lexicographic sort routine for integer arrays
    !
    !---- IN
    !---- m     : size of first component 
    !---- n     : size of second component
    !---- a     : array to sort
    !
    !---- OUT
    !---- a     : array is modified in the process
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(IN)    :: m, n
    INTEGER(C_INT), DIMENSION(:,:), INTENT(INOUT) :: a
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                                :: i, j, indx, isgn
    INTEGER(C_INT), DIMENSION(m)                  :: col
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (m <= 0) RETURN
    IF (n <= 1) RETURN
    !--------------------------------------------------------------------
    !---- Initializations
    !--------------------------------------------------------------------
    i    = 0
    indx = 0
    isgn = 0
    j    = 0
    !--------------------------------------------------------------------
    !---- Sort array
    !--------------------------------------------------------------------
    !---- Call the external heap sorter.
    DO
       CALL SORT_HEAP_EXTERNAL(n,indx,i,j,isgn)
       !
       !---- Interchange the I and J objects.
       IF (0 < indx) THEN
          CALL COL_SWAP_INT(m,n,i,j)
          !
          !---- Compare the I and J objects.
       ELSEIF (indx < 0) THEN
          CALL COL_COMPARE_INT(m,n,i,j,isgn)
       ELSEIF (indx == 0) THEN
          EXIT
       ENDIF
    ENDDO
    !--------------------------------------------------------------------
    
  CONTAINS

    !====================================================================
    !====================================================================
    !====================================================================

    SUBROUTINE COL_SWAP_INT(m,n,j1,j2)
      !------------------------------------------------------------------
      !---- Swaps columns j1 and j2 of an integer array
      !
      !---- IN
      !---- m      : size of first component 
      !---- n      : size of second component
      !---- a      : array 
      !---- j1, j2 : columns to swap
      !
      !---- OUT
      !---- a      : array is modified in the process
      !------------------------------------------------------------------
      !---- Modules
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
      !---- I/O variables
      !------------------------------------------------------------------
      INTEGER(C_INT), INTENT(IN) :: m, n, j1, j2
      !------------------------------------------------------------------
      !---- Local variables 
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
      !---- Testing for bad input
      !------------------------------------------------------------------
      IF (j1 == j2) RETURN
      !------------------------------------------------------------------
      !---- Perform swap
      !------------------------------------------------------------------
      col(1:m)  = a(1:m,j1)
      a(1:m,j1) = a(1:m,j2)
      a(1:m,j2) = col(1:m)
      !------------------------------------------------------------------
      !---- Testing for errors
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
    END SUBROUTINE COL_SWAP_INT

    !====================================================================
    !====================================================================
    !====================================================================

    SUBROUTINE COL_COMPARE_INT(m,n,i,j,isgn)
      !------------------------------------------------------------------
      !---- Compare columns i and j of an integer array
      !
      !---- IN
      !---- m      : size of first component 
      !---- n      : size of second component
      !---- a      : array 
      !---- i, j   : columns to compare
      !
      !---- OUT
      !---- isgn   : result of the comparison (-1 if col i < col j, 
      !----          0 if col i = col j, +1 if col i > col j)
      !------------------------------------------------------------------
      !---- Modules
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
      !---- I/O variables
      !------------------------------------------------------------------
      INTEGER(C_INT), INTENT(IN)  :: m, n, i, j
      INTEGER(C_INT), INTENT(OUT) :: isgn
      !------------------------------------------------------------------
      !---- Local variables 
      !------------------------------------------------------------------
      INTEGER(C_INT)              :: k
      !------------------------------------------------------------------
      !---- Testing for bad input
      !------------------------------------------------------------------
      IF (i == j) RETURN
      !------------------------------------------------------------------
      !---- Compare columns
      !------------------------------------------------------------------
      isgn = 0
      !
      !---- Comparison
      k = 1
      DO WHILE (k <= m)
         !
         IF ( a(k,i) < a(k,j) ) THEN
            isgn = -1
            RETURN
         ELSEIF ( a(k,j) < a(k,i) ) THEN
            isgn = +1
            RETURN
         ENDIF
         !
         k = k + 1
      ENDDO
      !------------------------------------------------------------------
    END SUBROUTINE COL_COMPARE_INT

    !--------------------------------------------------------------------
  END SUBROUTINE SORT_ARRAY_INT

  !======================================================================
  !======================================================================
  !======================================================================

  SUBROUTINE SORT_ARRAY_DOUBLE(m,n,a)
    !--------------------------------------------------------------------
    !---- Lexicographic sort routine for double arrays
    !
    !---- IN
    !---- m     : size of first component 
    !---- n     : size of second component
    !---- a     : array to sort
    !
    !---- OUT
    !---- a     : array is modified in the process
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT),                 INTENT(IN)    :: m, n
    
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(INOUT) :: a
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT)                                :: i, j, indx, isgn

    REAL(C_DOUBLE), DIMENSION(m)                  :: col
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    IF (m <= 0) RETURN
    IF (n <= 1) RETURN
    !--------------------------------------------------------------------
    !---- Initializations
    !--------------------------------------------------------------------
    i    = 0
    indx = 0
    isgn = 0
    j    = 0
    !--------------------------------------------------------------------
    !---- Sort array
    !--------------------------------------------------------------------
    !---- Call the external heap sorter.
    DO
       CALL SORT_HEAP_EXTERNAL(n,indx,i,j,isgn)
       !
       !---- Interchange the I and J objects.
       IF (0 < indx) THEN
          CALL COL_SWAP_DOUBLE(m,n,i,j)
          !
          !---- Compare the I and J objects.
       ELSEIF (indx < 0) THEN
          CALL COL_COMPARE_DOUBLE(m,n,i,j,isgn)
       ELSEIF (indx == 0) THEN
          EXIT
       ENDIF
    ENDDO
    !--------------------------------------------------------------------

  CONTAINS

    !====================================================================
    !====================================================================
    !====================================================================
    
    SUBROUTINE COL_SWAP_DOUBLE(m,n,j1,j2)
      !------------------------------------------------------------------
      !---- Swaps columns j1 and j2 of a double array
      !
      !---- IN
      !---- m      : size of first component 
      !---- n      : size of second component
      !---- a      : array 
      !---- j1, j2 : columns to swap
      !
      !---- OUT
      !---- a      : array is modified in the process
      !------------------------------------------------------------------
      !---- Modules
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
      !---- I/O variables
      !------------------------------------------------------------------
      INTEGER(C_INT), INTENT(IN) :: m, n, j1, j2
      !------------------------------------------------------------------
      !---- Local variables 
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
      !---- Testing for bad input
      !------------------------------------------------------------------
      IF (j1 == j2) RETURN
      !------------------------------------------------------------------
      !---- Perform swap
      !------------------------------------------------------------------
      col(1:m)  = a(1:m,j1)
      a(1:m,j1) = a(1:m,j2)
      a(1:m,j2) = col(1:m)
      !------------------------------------------------------------------
    END SUBROUTINE COL_SWAP_DOUBLE

    !====================================================================
    !====================================================================
    !====================================================================

    SUBROUTINE COL_COMPARE_DOUBLE(m,n,i,j,isgn)
      !------------------------------------------------------------------
      !---- Compare columns i and j of a double array
      !
      !---- IN
      !---- m      : size of first component 
      !---- n      : size of second component
      !---- a      : array 
      !---- i, j   : columns to compare
      !
      !---- OUT
      !---- isgn   : result of the comparison (-1 if col i < col j, 
      !----          0 if col i = col j, +1 if col i > col j)
      !------------------------------------------------------------------
      !---- Modules
      !------------------------------------------------------------------
      !
      !------------------------------------------------------------------
      !---- I/O variables
      !------------------------------------------------------------------
      INTEGER(C_INT), INTENT(IN)  :: m, n, i, j
      INTEGER(C_INT), INTENT(OUT) :: isgn
      !------------------------------------------------------------------
      !---- Local variables 
      !------------------------------------------------------------------
      INTEGER(C_INT)              :: k
      !------------------------------------------------------------------
      !---- Testing for bad input
      !------------------------------------------------------------------
      IF (i == j) RETURN
      !------------------------------------------------------------------
      !---- Compare columns
      !------------------------------------------------------------------
      isgn = 0
      !
      !---- Comparison
      k = 1
      DO WHILE (k <= m)
         !
         IF ( a(k,i) < a(k,j) ) THEN
            isgn = -1
            RETURN
         ELSEIF ( a(k,j) < a(k,i) ) THEN
            isgn = +1
            RETURN
         ENDIF
         !
         k = k + 1
      ENDDO
      !------------------------------------------------------------------
    END SUBROUTINE COL_COMPARE_DOUBLE
    
    !--------------------------------------------------------------------
  END SUBROUTINE SORT_ARRAY_DOUBLE

  !======================================================================
  !======================================================================
  !======================================================================
  
  SUBROUTINE SORT_HEAP_EXTERNAL(n,indx,i,j,isgn)
    !--------------------------------------------------------------------
    !---- Sorts a list of items in ascending order
    !
    !---- IN
    !---- n      : number of items to sort
    !---- isgn   : result of comparison of elements i and j
    !
    !---- OUT
    !---- indx   : communication signal
    !---- i, j   : two items to act on
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)    :: n, isgn
    INTEGER(C_INT), INTENT(OUT)   :: i, j
    INTEGER(C_INT), INTENT(INOUT) :: indx
    !--------------------------------------------------------------------
    !---- Local variables 
    !--------------------------------------------------------------------
    INTEGER(C_INT), SAVE :: i_save = 0
    INTEGER(C_INT), SAVE :: j_save = 0
    INTEGER(C_INT), SAVE :: k      = 0
    INTEGER(C_INT), SAVE :: k1     = 0
    INTEGER(C_INT), SAVE :: n1     = 0
    !--------------------------------------------------------------------
    !---- Sort
    !--------------------------------------------------------------------
    !---- INDX = 0: This is the first call.
    IF (indx == 0) THEN
       i_save = 0
       j_save = 0
       k      = n/2
       k1     = k
       n1     = n
       !
       !---- INDX < 0: The user is RETURNing the results of a comparison.
    ELSEIF (indx < 0) THEN
       IF (indx == -2) THEN
          IF (isgn < 0) i_save = i_save + 1
          !
          j_save = k1
          k1     = i_save
          indx   = -1
          i      = i_save
          j      = j_save
          RETURN
       ENDIF
       !
       IF (0 < isgn) THEN
          indx = 2
          i    = i_save
          j    = j_save
          RETURN
       ENDIF
       !
       IF (k <= 1) THEN
          IF ( n1 == 1 ) THEN
             i_save = 0
             j_save = 0
             indx   = 0
          ELSE
             i_save = n1
             n1     = n1 - 1
             j_save = 1
             indx   = 1
          ENDIF
          !
          i = i_save
          j = j_save
          RETURN
       ENDIF
       !
       k  = k - 1
       k1 = k
       !
       !---- 0 < INDX, the user was asked to make an interchange.
    ELSEIF (indx == 1) THEN
       k1 = k
    ENDIF
    !
    DO
       i_save = 2 * k1
       IF (i_save == n1) THEN
          j_save = k1
          k1     = i_save
          indx   = -1
          i      = i_save
          j      = j_save
          RETURN
       ELSEIF (i_save <= n1) THEN
          j_save = i_save + 1
          indx   = -2
          i      = i_save
          j      = j_save
          RETURN
       ENDIF
       !
       IF (k <= 1) EXIT
       !
       k  = k - 1
       k1 = k
    ENDDO
    !
    IF (n1 == 1) THEN
       i_save = 0
       j_save = 0
       indx   = 0
       i      = i_save
       j      = j_save
    ELSE
       i_save = n1
       n1     = n1 - 1
       j_save = 1
       indx   = 1
       i      = i_save
       j      = j_save
    ENDIF
    !--------------------------------------------------------------------
  END SUBROUTINE SORT_HEAP_EXTERNAL  

  !======================================================================
  !======================================================================
  !======================================================================

END MODULE SORT
