MODULE SPECIAL_FUNCTIONS

  !====================================================================
  ! Module containing special functions and various other operations
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Interface for overloading cross product
  !--------------------------------------------------------------------
  INTERFACE CROSS_PROD
     MODULE PROCEDURE CROSS_PROD_REAL, CROSS_PROD_COMPLEX
  END INTERFACE CROSS_PROD
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CROSS_PROD_REAL(u, v, w)
    !------------------------------------------------------------------
    !---- Cross-product of two R3 vectors
    !
    !---- IN
    !---- u, v       : R3 vectors
    !
    !---- OUT
    !---- w          : u x v
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: u, v
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: w
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute cross-product
    !------------------------------------------------------------------
    w(1) = u(2)*v(3) - u(3)*v(2)
    w(2) = u(3)*v(1) - u(1)*v(3)
    w(3) = u(1)*v(2) - u(2)*v(1)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CROSS_PROD_REAL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CROSS_PROD_COMPLEX(u, v, w)
    !------------------------------------------------------------------
    !---- Cross-product of two C3 vectors
    !
    !---- IN
    !---- u, v       : C3 vectors
    !
    !---- OUT
    !---- w          : u x v
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3), INTENT(IN)  :: u, v
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3), INTENT(OUT) :: w
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute cross-product
    !------------------------------------------------------------------
    w(1) = u(2)*v(3) - u(3)*v(2)
    w(2) = u(3)*v(1) - u(1)*v(3)
    w(3) = u(1)*v(2) - u(2)*v(1)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CROSS_PROD_COMPLEX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE NORMALIZE_VECTOR(v, norm)
    !------------------------------------------------------------------
    !---- Normalize an R3 vector
    !
    !---- IN
    !---- v    : R3 vector
    !
    !---- OUT
    !---- v    : modified in the process
    !---- norm : norm of the vector
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(OUT)   :: norm
    REAL(C_DOUBLE), DIMENSION(3), INTENT(INOUT) :: v
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Normalize vector
    !------------------------------------------------------------------
    norm = NORM2(v)
    v(:) = v(:)/norm
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE NORMALIZE_VECTOR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE NORMALIZE_SCALAR(scalar)
    !------------------------------------------------------------------
    !---- Normalize a scalar
    !
    !---- IN
    !---- scalar : scalar
    !
    !---- OUT
    !---- scalar : modified in the process
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: scalar
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                :: norm
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Normalize scalar
    !------------------------------------------------------------------
    norm   = ABS(scalar)
    scalar = scalar/norm
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE NORMALIZE_SCALAR

  !====================================================================
  !====================================================================
  !====================================================================

  FUNCTION MIXT_PROD(u, v, w)
    !------------------------------------------------------------------
    !---- Mixt product of three R3 vectors
    !
    !---- IN
    !---- u, v, w   : R3 vectors
    !
    !---- OUT
    !---- MIXT_PROD : (u x v).w
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: MIXT_PROD
    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: u, v, w
    REAL(C_DOUBLE), DIMENSION(3)              :: uv
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute cross-product
    !------------------------------------------------------------------
    CALL CROSS_PROD(u,v,uv)
    MIXT_PROD = DOT_PRODUCT(uv,w)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END FUNCTION MIXT_PROD

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CANTOR_PAIRING(int1, int2, int3, iCantor)
    !------------------------------------------------------------------
    !---- Symmetric Cantor pairing of three integers
    !---- To get symmetric results, input integers are ordered
    !
    !---- IN
    !---- i1, i2, i3 : Integers to pair
    !
    !---- OUT
    !---- iCantor    : Cantor pairing number
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    INTEGER(C_INT),       INTENT(IN)  :: int1, int2, int3
    INTEGER(C_LONG_LONG), INTENT(OUT) :: iCantor
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)       :: itemp, i1, i2, i3
    INTEGER(C_LONG_LONG) :: it
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize and order i1, i2, i3
    !------------------------------------------------------------------
    iCantor = 0
    i1      = int1
    i2      = int2
    i3      = int3
    !
    IF (i1 > i2) THEN
       itemp = i2
       i2    = i1
       i1    = itemp
    ENDIF
    IF (i2 > i3) THEN
       itemp = i3
       i3    = i2
       i2    = itemp
    ENDIF
    IF (i1 > i2) THEN
       itemp = i2
       i2    = i1
       i1    = itemp
    ENDIF
    !------------------------------------------------------------------
    !---- Apply Cantor pairing
    !------------------------------------------------------------------
    it      = (i1+i2)*(i1+i2+1)+i2
    iCantor = (it+i3)*(it+i3+1)+i3
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CANTOR_PAIRING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DEGREES_TO_RADIANS(angle)
    !------------------------------------------------------------------
    !---- Convert degrees to radians
    !
    !---- IN
    !---- angle : angle in degrees
    !
    !---- OUT
    !---- angle : converted into radians
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: angle
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                :: convAngle
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    convAngle = 180.0
    angle     = angle*Pi/convAngle
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEGREES_TO_RADIANS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SECONDS_TO_METERS(time)
    !------------------------------------------------------------------
    !---- Convert seconds to meters
    !
    !---- IN
    !---- time : time in seconds
    !
    !---- OUT
    !---- time : converted in meters
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: time
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    time = time*c0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SECONDS_TO_METERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE METERS_TO_SECONDS(time)
    !------------------------------------------------------------------
    !---- Convert meters to seconds
    !
    !---- IN
    !---- time : time in meters
    !
    !---- OUT
    !---- time : converted in seconds
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: time
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    time = time*invC0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE METERS_TO_SECONDS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FREQUENCY_TO_INVERSE_METERS(frequency)
    !------------------------------------------------------------------
    !---- Convert frequency to inverse meters
    !
    !---- IN
    !---- frequency : frequency in Hz (s^-1)
    !
    !---- OUT
    !---- frequency : converted in m^-1
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: frequency
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    frequency = frequency*invC0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FREQUENCY_TO_INVERSE_METERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INVERSE_METERS_TO_FREQUENCY(frequency)
    !------------------------------------------------------------------
    !---- Convert inverse meters to frequency
    !
    !---- IN
    !---- frequency : frequency in m^-1
    !
    !---- OUT
    !---- frequency : converted in Hz (s^-1)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: frequency
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    frequency = frequency*c0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INVERSE_METERS_TO_FREQUENCY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE VELOCITY_TO_UNITLESS(vel)
    !------------------------------------------------------------------
    !---- Convert velocity to normalized velocity
    !
    !---- IN
    !---- vel : velocity in [m/s]
    !
    !---- OUT
    !---- vel : converted in []
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: vel
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    vel = vel*invC0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE VELOCITY_TO_UNITLESS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE UNITLESS_TO_VELOCITY(vel)
    !------------------------------------------------------------------
    !---- Convert normalized velocity to velocity
    !
    !---- IN
    !---- vel : velocity in []
    !
    !---- OUT
    !---- vel : converted in [m/s]
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(INOUT) :: vel
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Convert
    !------------------------------------------------------------------
    vel = vel*c0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE UNITLESS_TO_VELOCITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_NUMBER_OF_FREQUENCIES(freqMin, freqMax, deltaFreq, nFreq)
    !------------------------------------------------------------------
    !---- Compute the number of frequencies for computations
    !
    !---- IN
    !---- freqM**   : minimal and maximal frequencies
    !---- deltaFreq : frequency step
    !
    !---- OUT
    !---- nFreq     : number of frequencies to compute
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT)   :: nFreq
    
    REAL(C_DOUBLE), INTENT(INOUT) :: freqMin, freqMax, deltaFreq
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute
    !------------------------------------------------------------------
    nFreq = FLOOR((freqMax - freqMin)/deltaFreq) + 1
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_NUMBER_OF_FREQUENCIES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_FOURIER_TRANSFORM_FACTOR(freq, time, dt, zFourier)
    !------------------------------------------------------------------
    !---- Compute the Fourier transform factor
    !
    !---- IN
    !---- freq     : frequency     in Hz
    !---- time     : physical time in seconds
    !---- dt       : timestep      in seconds
    !
    !---- OUT
    !---- zFourier : Fourier factor
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),            INTENT(IN)  :: freq, time, dt

    COMPLEX(C_DOUBLE_COMPLEX), INTENT(OUT) :: zFourier
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute Fourier transform factor
    !------------------------------------------------------------------
    zFourier = dt*EXP(iUnit*2.0*Pi*freq*time)*OverSqrtTwoPi
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_FOURIER_TRANSFORM_FACTOR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_R3_DISTANCE(pt1, pt2, distance)
    !------------------------------------------------------------------
    !---- Compute R3 distance between two points
    !
    !---- IN
    !---- pt*      : coordinates of points
    !
    !---- OUT
    !---- distance : distance between two points
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(OUT) :: distance
    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: pt1, pt2
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute distance
    !------------------------------------------------------------------
    distance = SQRT((pt1(1)-pt2(1))**2 + (pt1(2)-pt2(2))**2 + (pt1(3)-pt2(3))**2)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_R3_DISTANCE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MOVE_POINT_ARRAY_WITH_VECTOR(ptArrayIn, vector, ptArrayOut)
    !------------------------------------------------------------------
    !---- Move a point with a given vector
    !---- Point arrays must have size (3,nPts)
    !
    !---- IN
    !---- ptArrayIn  : input point array
    !---- vector     : vector to move ptArrayIn
    !
    !---- OUT
    !---- ptArrayOut : output point array
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3),   INTENT(IN)  :: vector
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(IN)  :: ptArrayin
    REAL(C_DOUBLE), DIMENSION(:,:), INTENT(OUT) :: ptArrayOut
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                              :: iVector
    REAL(C_DOUBLE), DIMENSION(3)                :: pt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute distance
    !------------------------------------------------------------------
    DO iVector=1,SIZE(ptArrayIn,2)
       CALL MOVE_POINT_WITH_VECTOR(ptArrayIn(:,iVector), vector, pt)
       ptArrayout(:,iVector) = pt
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MOVE_POINT_ARRAY_WITH_VECTOR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MOVE_POINT_WITH_VECTOR(ptIn, vector, ptOut)
    !------------------------------------------------------------------
    !---- Move a point with a given vector
    !
    !---- IN
    !---- ptIn   : input point
    !---- vector : vector to move ptIn
    !
    !---- OUT
    !---- ptOut  : output point
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), DIMENSION(3), INTENT(IN)  :: ptIn, vector
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: ptOut
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute distance
    !------------------------------------------------------------------
    ptOut(:) = ptIn(:) + vector(:)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MOVE_POINT_WITH_VECTOR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MODIFY_INTEGER(int, position, val)
    !------------------------------------------------------------------
    !---- Modify an int by putting val at given position
    !
    !---- IN
    !---- val      : value to put at right position
    !---- position : position where to put val
    !
    !---- OUT
    !---- int      : modified int
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)    :: val, position
    INTEGER(C_INT), INTENT(INOUT) :: int
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                :: newVal, oldVal
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Modify log
    !------------------------------------------------------------------
    CALL DECOMPOSE_INTEGER(int, position, oldVal)
    oldVal = oldVal*(10**(position-1))
    newVal = val   *(10**(position-1))
    int = int - oldVal + newVal
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MODIFY_INTEGER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DECOMPOSE_INTEGER(int, position, val)
    !------------------------------------------------------------------
    !---- Decompose a number in its integers
    !
    !---- IN
    !---- int      : integer to decompose
    !---- position : position of the integer we want to extract
    !----         (eg 1 for units, 2 for tens, 3 for hundreds, etc)
    !
    !---- OUT
    !---- val      : value of the extracted integer
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: int, position
    INTEGER(C_INT), INTENT(OUT) :: val
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                            :: length, locInt, counter
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: decomposed
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (int == 0) THEN
       val = 0
       RETURN
    ENDIF
    !------------------------------------------------------------------
    !---- Extract integer
    !------------------------------------------------------------------
    CALL GET_INTEGER_LENGTH(int, length)
    ALLOCATE(decomposed(length))
    !
    locInt  = int
    counter = 0
    DO WHILE(locInt /= 0)
       counter             = counter + 1
       decomposed(counter) = MOD(locInt, 10)
       locInt              = locInt/10
    ENDDO
    !------------------------------------------------------------------
    !---- Return required value
    !------------------------------------------------------------------
    IF (position <= length) val = decomposed(position)
    IF (position >  length) val = 0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (ALLOCATED(decomposed)) DEALLOCATE(decomposed)
    !------------------------------------------------------------------
  END SUBROUTINE DECOMPOSE_INTEGER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GET_INTEGER_LENGTH(int, length)
    !------------------------------------------------------------------
    !---- Compute the length of an integer
    !
    !---- IN
    !---- int    : input integer
    !
    !---- OUT
    !---- length : length of the integer
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: int
    INTEGER(C_INT), INTENT(OUT) :: length
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute length
    !------------------------------------------------------------------
    length = 1 + FLOOR(LOG10(REAL(int)))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_INTEGER_LENGTH

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE GET_MANTISSA_EXPONENT(valIn, mantissa, exponent)
    !------------------------------------------------------------------
    !---- Compute the mantissa and exponent of a real
    !
    !---- IN
    !---- valIn    : real value
    !
    !---- OUT
    !---- mantissa : mantissa of valIn
    !---- exponent : exponent of valIn
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(OUT) :: exponent

    REAL(C_DOUBLE), INTENT(IN)  :: valIn
    REAL(C_DOUBLE), INTENT(OUT) :: mantissa
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)              :: val
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute mantissa and exponenet
    !------------------------------------------------------------------
    exponent = 0.0
    mantissa = 0.0
    !
    IF (valIn <  0.0) val =-valIn
    IF (valIn >  0.0) val = valIn
    IF (valIn == 0.0) RETURN
    !
    exponent = INT(LOG10(val))
    mantissa = val*10.0**(-exponent)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GET_MANTISSA_EXPONENT

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE SPECIAL_FUNCTIONS
