MODULE SPHERIC_CAVITY

  !====================================================================
  ! Routines related to the spheric cavity case
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! SPHERIC CAVITY DATA ('SC')
  ! computeErrSC ==> TRUE enables the computation of the overall L2 error
  ! omegaSC      ==> mode pulsation
  ! kSC          ==> waveNumber
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL) :: computeErrSC

  REAL(C_DOUBLE)  :: omegaSC, kSC
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_SPHERIC_CAVITY
    !------------------------------------------------------------------
    !---- Pre-processing for spheric cavity case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SPECIAL_FUNCTIONS
    USE MATH_PHYS_CONSTANTS
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE) :: freqSC
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize data
    !------------------------------------------------------------------
    freqSC  = 0.13091174401040966770E9 ! in Hz
    omegaSC = 2.0*Pi*freqSC
    CALL FREQUENCY_TO_INVERSE_METERS(omegaSC)
    kSC     = omegaSC
    !------------------------------------------------------------------
    !---- Print sum-up
    !------------------------------------------------------------------
    IF(idWrite) WRITE(6, '(a)')       'Source informations'
    IF(idWrite) WRITE(6, 10)
    IF(idWrite) WRITE(6, '(a)')       'Running case: spheric cavity mode'
    IF(idWrite) WRITE(6, '(a,e15.7)') 'Omega in Hz: ',omegaSC*c0
    IF(idWrite) WRITE(6, 10)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_SPHERIC_CAVITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_SOL_SPHERIC_CAVITY_TIME(iCell, nDofV, order, time)
    !------------------------------------------------------------------
    !---- Solution of the spheric cavity mode at the dof of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of dof in the cell
    !---- order : polynomial order in the cell    
    !---- time  : physical time for solution
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT),   INTENT(IN) :: iCell, nDofV, order

    REAL(C_DOUBLE),   INTENT(IN) :: time
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: iDof

    REAL(C_DOUBLE), DIMENSION(3) :: solE, solH
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDof=1,nDofV
       CALL TIME_SOL_SPHERIC_CAVITY(time, dofXV(iDof), dofYV(iDof), dofZV(iDof), solE, solH)
       !
       ExV(iDof) = solE(1)
       EyV(iDof) = solE(2)
       EzV(iDof) = solE(3)
       HxV(iDof) = solH(1)
       HyV(iDof) = solH(2)
       HzV(iDof) = solH(3)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_SOL_SPHERIC_CAVITY_TIME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE TIME_SOL_SPHERIC_CAVITY(time, x, y, z, solE, solH)
    !------------------------------------------------------------------
    !---- Solution of the spheric cavity mode at given coordinates
    !
    !---- IN
    !---- time    : physical time for solution
    !---- x, y, z : coordinates for the solution
    !
    !---- OUT
    !---- sol*    : solutions for E and H
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE),               INTENT(IN)  :: time, x, y, z
    REAL(C_DOUBLE), DIMENSION(3), INTENT(OUT) :: solE, solH
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE)                            :: ct, st, cr, sr, rad, &
         rad2, kRad, kRad2, val1, val2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute solution
    !------------------------------------------------------------------
    rad   = SQRT(x**2 + y**2 + z**2)
    kRad  = kSC*rad
    rad2  = rad**2
    kRad2 = kRad**2
    !
    cr   = COS(kRad)
    sr   = SIN(kRad)
    ct   = COS(omegaSC*time)
    st   = SIN(omegaSC*time)
    !
    val1 = (kRad - 3.0/kRad)*sr + 3.0*cr
    val2 = (kRad - 1.0/kRad)*sr +     cr
    !
    solE(1) = z*x*val1*ct/(kRad2*rad2)
    solE(2) = z*y*val1*ct/(kRad2*rad2)
    solE(3) = z*z*val1*ct/(kRad2*rad2) - val2*ct/kRad2
    solH(1) =-y*(sr/kRad - cr)*st/(kRad*rad)
    solH(2) = x*(sr/kRad - cr)*st/(kRad*rad)
    solH(3) = 0.0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE TIME_SOL_SPHERIC_CAVITY

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE SPHERIC_CAVITY
