MODULE TIMERS

  !====================================================================
  ! Definition and procedures associated to timers
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a timer
  !--------------------------------------------------------------------
  TYPE :: Timer
     LOGICAL(C_BOOL) :: amIrunning = .FALSE.
     INTEGER(C_INT)  :: systemStart, systemRate
     REAL(C_DOUBLE)  :: cpuStart
   CONTAINS
     PROCEDURE       :: START       => START_TIMER
     PROCEDURE       :: STOP        => STOP_TIMER
     PROCEDURE       :: CPU_TIME    => RETURN_CPU_TIME
     PROCEDURE       :: SYSTEM_TIME => RETURN_SYSTEM_TIME
     PROCEDURE       :: RUNNING     => RETURN_AMIRUNNING
  END TYPE Timer

  !--------------------------------------------------------------------
  ! Arrays and various stuff for timing
  !--------------------------------------------------------------------
  REAL(C_DOUBLE)                            :: cpuMean, mpiMean, &
       maxCpuGap, maxMpiGap, mpiTime = 0.0
  REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: cpuTimeArray,     &
       cpuGapArray, mpiTimeArray, mpiGapArray
  !--------------------------------------------------------------------  

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE START_TIMER(this)
    !------------------------------------------------------------------    
    CLASS(Timer), INTENT(INOUT)	:: this
    !------------------------------------------------------------------
    this%amIrunning = .TRUE.
    CALL CPU_TIME(this%cpuStart)
    CALL SYSTEM_CLOCK(this%systemStart, this%systemRate)
    !------------------------------------------------------------------
  END SUBROUTINE START_TIMER

  !====================================================================
  !====================================================================
  !====================================================================  

  SUBROUTINE STOP_TIMER(this)
    !------------------------------------------------------------------
    class(Timer), INTENT(INOUT)	:: this
    !------------------------------------------------------------------    
    this%amIrunning = .FALSE.
    !------------------------------------------------------------------
  END SUBROUTINE STOP_TIMER

  !====================================================================
  !====================================================================
  !====================================================================  

  FUNCTION RETURN_SYSTEM_TIME(this) RESULT(time)
    !------------------------------------------------------------------    
    CLASS(Timer), INTENT(IN) :: this
    INTEGER(C_INT)           :: systemEnd
    REAL(C_DOUBLE)           :: time
    !------------------------------------------------------------------    
    IF (this%amIrunning) THEN
       CALL SYSTEM_CLOCK(systemEnd)
       time = REAL(systemEnd - this%systemStart)/REAL(this%systemRate)
    ELSE
       time = 0.0
    ENDIF
    !------------------------------------------------------------------
  END FUNCTION RETURN_SYSTEM_TIME

  !====================================================================
  !====================================================================
  !====================================================================  

  FUNCTION RETURN_CPU_TIME(this) RESULT(time)
    !------------------------------------------------------------------    
    CLASS(Timer), INTENT(IN) :: this
    REAL(C_DOUBLE)           :: cpuEnd,time
    !------------------------------------------------------------------
    IF (this%amIrunning) THEN
       CALL CPU_TIME(cpuEnd)
       time = cpuEnd - this%cpuStart
    ELSE
       time = 0.0
    ENDIF
    !------------------------------------------------------------------
  END FUNCTION RETURN_CPU_TIME

  !====================================================================
  !====================================================================
  !====================================================================  

  FUNCTION RETURN_AMIRUNNING(this) RESULT(amIrunning)
    !------------------------------------------------------------------    
    CLASS(Timer), INTENT(INOUT)	:: this
    LOGICAL(C_BOOL)             :: amIrunning
    !------------------------------------------------------------------
    amIrunning = this%amIrunning
    !------------------------------------------------------------------
  END FUNCTION RETURN_AMIRUNNING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PROCESS_TIMERS_LOOP(myTimer)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    USE MESH_DATA
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(Timer), INTENT(IN) :: myTimer
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    CHARACTER(LEN=100, KIND=C_CHAR) :: fileName, cProc
    
    INTEGER(C_INT)                  :: iProc, nUnit

    REAL(C_DOUBLE)                  :: cpuTime, wallTime
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Retrieve timers information
    !------------------------------------------------------------------
    cpuTime  = myTimer%CPU_TIME()
    wallTime = myTimer%SYSTEM_TIME()
    !------------------------------------------------------------------
    !---- Gather information and select writing proc
    !------------------------------------------------------------------
    ALLOCATE(cpuTimeArray(nbProc))
    ALLOCATE(mpiTimeArray(nbProc))
    !
    CALL MPI_GATHER(cpuTime, 1, MPI_DOUBLE_PRECISION, cpuTimeArray, &
         1, MPI_DOUBLE_PRECISION, 0, defaultComm, mpiError)
    CALL MPI_GATHER(mpiTime, 1, MPI_DOUBLE_PRECISION, mpiTimeArray, &
         1, MPI_DOUBLE_PRECISION, 0, defaultComm, mpiError)
    !
    IF (.NOT. idWrite) RETURN
    !------------------------------------------------------------------
    !---- Initial write and allocations
    !------------------------------------------------------------------
    ALLOCATE(cpuGapArray(nbProc))
    ALLOCATE(mpiGapArray(nbProc))
    !
    WRITE(6, '(a)') "Time spent on main loop (seconds)"
    !------------------------------------------------------------------
    !---- Process CPU and MPI timers
    !------------------------------------------------------------------
    cpuMean        = SUM(cpuTimeArray)/nbProc
    cpuGapArray(:) = ABS(cpuTimeArray(:)-cpuMean)*100.0/cpuMean
    maxCpuGap      = MAXVAL(cpuGapArray)
    mpiMean        = SUM(mpiTimeArray)/nbProc
    mpiGapArray(:) = ABS(mpiTimeArray(:)-mpiMean)*100.0/mpiMean
    maxMpiGap      = MAXVAL(mpiGapArray)
    !
    WRITE(6, '(a,f15.7)'  )    "Mean CPU time : ",cpuMean
    WRITE(6, '(a,f15.7,a)')    "Max  CPU gap  : ",maxCpuGap," %"
    IF (parallelRun) THEN
       WRITE(6, '(a,f15.7)'  ) "Mean MPI time : ",mpiMean
       WRITE(6, '(a,f15.7,a)') "Max  MPI gap  : ",maxMpiGap," %"
    ENDIF
    WRITE(6, '(a,f15.7)'  )    "Wall time     : ",wallTime
    !------------------------------------------------------------------
    !---- Write timing results in file
    !------------------------------------------------------------------
    CALL INT_TO_CHAR(nbProc,cProc)
    fileName = "Timers_"//TRIM(globalMesh%name)//"_"//TRIM(cProc)
    CALL GET_NEW_FILE_UNIT(nUnit)
    !
    OPEN(UNIT=nUnit, FILE=fileName)
    WRITE(nUnit,'(i9)'    ) nbProc
    WRITE(nUnit,'(5f15.7)') wallTime, cpuMean, maxCpuGap, mpiMean, maxMpiGap
    !
    DO iProc=1,nbProc
       WRITE(nUnit,'(i9,4f15.7)') iProc, cpuTimeArray(iProc), &
            cpuGapArray(iProc), mpiTimeArray(iProc), mpiGapArray(iProc)
    ENDDO
    !
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
    DEALLOCATE(cpuTimeArray, cpuGapArray, mpiTimeArray, mpiGapArray)
    !------------------------------------------------------------------
  END SUBROUTINE PROCESS_TIMERS_LOOP

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_TIMERS(myTimer)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(Timer), INTENT(IN) :: myTimer
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. idWrite) RETURN
    !------------------------------------------------------------------
    !---- Write
    !------------------------------------------------------------------
    WRITE(6, '(a,f15.7)') "Time elapsed (seconds) : ",myTimer%SYSTEM_TIME()
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_TIMERS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_TIMERS_PREPROCESSING(myTimer)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(Timer), INTENT(IN) :: myTimer
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. idWrite) RETURN
    !------------------------------------------------------------------
    !---- Write
    !------------------------------------------------------------------
    WRITE(6, '(a)') "Time spent on pre-processing (seconds)"
    WRITE(6, '(a,f15.7)') "CPU  time     : ", myTimer%CPU_TIME()
    WRITE(6, '(a,f15.7)') "Wall time     : ", myTimer%SYSTEM_TIME()
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_TIMERS_PREPROCESSING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_TIMERS_POSTPROCESSING(myTimer)
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(Timer), INTENT(IN) :: myTimer
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. idWrite) RETURN
    !------------------------------------------------------------------
    !---- Write
    !------------------------------------------------------------------
    WRITE(6, '(a)') "Time spent on post-processing (seconds)"
    WRITE(6, '(a,f15.7)') "CPU  time     : ", myTimer%CPU_TIME()
    WRITE(6, '(a,f15.7)') "Wall time     : ", myTimer%SYSTEM_TIME()
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------    
    !
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_TIMERS_POSTPROCESSING

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE TIMERS

