PROGRAM HORSE

  !========================================================================
  ! Main program of HORSE(*): 
  ! HDG methods for the solution of 3D time-harmonic Maxwell's equations
  !
  ! Available HDG methods:
  ! - Classical HDGM 
  ! - HDGM combined to a Schwarz algorithm 
  ! - HDGM combined to a Schwarz algorithm accelerated by a Krylov 
  !   subspace method
  !
  ! For a complete description of HDG methods implemented in HORSE refer to
  ! ~/horse/documents/HDGMinHORSE.pdf
  !
  ! (*) High Order solver for Radar cross Section Evaluation
  !========================================================================

  !------------------------------------------------------------------------
  ! Shared modules
  !------------------------------------------------------------------------
  USE MESH_FACES
  USE MESH_DATA
  USE HDG_MATRICES_STRUCTURES
  USE HDG_CORE
  USE HDG_SOLVER
  USE INPUT_DATA
  USE ISO_C_BINDING
  USE MATERIAL_STRUCTURES
  USE MAXWELL_STRUCTURES
  USE MESH_STRUCTURES
  USE OBSERVABLE_STRUCTURES
  USE PARALLEL_STRUCTURES
  USE SOURCE_STRUCTURES
  USE TIMERS
  !------------------------------------------------------------------------
  IMPLICIT NONE
  !------------------------------------------------------------------------
  ! Shared variables
  !------------------------------------------------------------------------
  TYPE(Timer) :: myTimer
  !------------------------------------------------------------------------
  !---- Pre-processing steps
  !
  !---- PARAL_INITIALIZE            opens MPI/OMP environments
  !---- PREPROCESS_DATA             reads all data files
  !---- PREPROCESS_MESH             reads mesh and sets up all related structures
  !---- PREPROCESS_MATERIALS        reads materials data and create structures
  !---- PREPROCESS_MAXWELL          sets up solution and matrices data structures
  !---- PREPROCESS_SOURCES          sets up sources
  !---- PREPROCESS_HDGMATRICES      sets up hdg matrices data structures
  !---- PREPROCESS_HDGCORE          sets up some usefull routines for hdg method
  !---- PREPROCESS_OBSERVABLE       sets up observables
  !---- PRINT_TIMERS_PREPROCESSING  write time spent on preprocessing
  !------------------------------------------------------------------------
  addInnerFaces = .TRUE.
  !
  CALL myTimer%Start()
  !
  CALL PARALLEL_INITIALIZE
  CALL PREPROCESS_DATA
  CALL PREPROCESS_MESH
  CALL PREPROCESS_MATERIALS
  CALL PREPROCESS_MAXWELL
  CALL PREPROCESS_SOURCES
  CALL PREPROCESS_HDGMATRICES
  CALL PREPROCESS_HDGCORE
  CALL PREPROCESS_OBSERVABLES
  CALL PRINT_TIMERS_PREPROCESSING(myTimer)
  !------------------------------------------------------------------------
  !---- Processing steps
  !
  !---- COMPUTE_HDG_SOLVER compute HDG Method
  !
  !     ROUTINE LOCATION 
  !     module HDG_core.f90
  !
  !     IN 
  !     hdgMethod  : defined in ~/horse/data/dataHorse
  !     solverPack : defined in ~/horse/data/dataHorse
  !
  !---- Print timing infos
  !------------------------------------------------------------------------
  CALL myTimer%Start()
  CALL COMPUTE_HDG_SOLVER(hdgMethod, solverPack)
  IF (idWrite) THEN
     WRITE(6, '(a)')       " Time spent in HDGM processing (seconds) "
     WRITE(6, '(a,f15.7)') " CPU  time     : ", myTimer%CPU_TIME()
     WRITE(6, '(a,f15.7)') " Wall time     : ", myTimer%SYSTEM_TIME()
  ENDIF
  !------------------------------------------------------------------------
  !---- Post-processing steps
  !
  !---- POSTPROCESS_OBSERVABLES      realizes the final operations for observables
  !---- PARAL_FINALIZE               closes MPI/OMP environments
  !---- PRINT_TIMERS_POSTPROCESSING  write time spent on postprocessing
  !------------------------------------------------------------------------
  CALL myTimer%Start()
  !
  CALL POSTPROCESS_OBSERVABLES
  CALL PARALLEL_BARRIER
  CALL PARALLEL_FINALIZE
  CALL PRINT_TIMERS_POSTPROCESSING(myTimer)
  !------------------------------------------------------------------------
END PROGRAM HORSE

