MODULE HDG_CORE

  !====================================================================
  ! Main classical hdgm core routines
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Procedure pointer to compute local solver
  !--------------------------------------------------------------------
  PROCEDURE(), POINTER :: LOCAL_SOLVER => NULL()
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_HDGCORE
    !------------------------------------------------------------------
    !---- Point to the appropriate routines for hdgm
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Point LOCAL_SOLVER to the appropriate routine
    !------------------------------------------------------------------
    IF (currentSrcType == 'NO') THEN
       LOCAL_SOLVER => LOCAL_SOLVER_WITHOUT_CURRENT_SRC
    ELSE
       LOCAL_SOLVER => LOCAL_SOLVER_WITH_CURRENT_SRC
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_HDGCORE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE LOCAL_SOLVER_WITHOUT_CURRENT_SRC(iCell)
    !------------------------------------------------------------------
    !---- Solve local linear systems in terms of DoFs of the EM field
    !
    !---- IN
    !---- Cell index : iCell
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_MATRICES_STRUCTURES
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: indLoc, j, k, kFac, nDofFk, nDofV, facNum

    COMPLEX(C_DOUBLE_COMPLEX)                            :: Ex, Ey, Ez
    COMPLEX(C_DOUBLE_COMPLEX)                            :: Hx, Hy, Hz
    COMPLEX(C_DOUBLE_COMPLEX)                            :: zero
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: LambdaLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Solve the local linear system in terms of the DoFs of the EM
    !---- field in cell with index 'iCell' in data structure 'MeshCell'
    !------------------------------------------------------------------
    zero = (0.d0, 0.d0)
    ALLOCATE(LambdaLoc(Qe(iCell)%ndof2))
    DO indLoc=1,Qe(iCell)%ndof2
       LambdaLoc(indLoc) = zero
    ENDDO
    !
    DO k=1,MeshCell(iCell)%nNeighb
       facNum = MeshCell(iCell)%facNum(k)
       nDofFk = Lambda(facNum)%nDof
       indLoc = 2*(hdgEM(iCell)%sumDofF(k) - nDofFk)
       DO j=1,nDofFk
          LambdaLoc(indLoc + j)          = Lambda(facNum)%L%vec(j,1)
          LambdaLoc(indloc + j + nDofFk) = Lambda(facNum)%L%vec(j,2)
       ENDDO
    ENDDO
    !
    nDofV = hdgEM(iCell)%nDofV
    DO k=1,nDofV
       Ex = zero
       Ey = zero
       Ez = zero
       Hx = zero
       Hy = zero
       Hz = zero
       DO j=1,Qe(iCell)%nDof2
          Ex = Ex + Qe(iCell)%tab(k,j)*LambdaLoc(j)
          Ey = Ey + Qe(iCell)%tab(k +   nDofV,j)*LambdaLoc(j)
          Ez = Ez + Qe(iCell)%tab(k + 2*nDofV,j)*LambdaLoc(j)
          Hx = Hx + Qe(iCell)%tab(k + 3*nDofV,j)*LambdaLoc(j)
          Hy = Hy + Qe(iCell)%tab(k + 4*nDofV,j)*LambdaLoc(j)
          Hz = Hz + Qe(iCell)%tab(k + 5*nDofV,j)*LambdaLoc(j)
       ENDDO
       hdgEM(iCell)%E%vec(k,1) = Ex 
       hdgEM(iCell)%E%vec(k,2) = Ey 
       hdgEM(iCell)%E%vec(k,3) = Ez
       hdgEM(iCell)%H%vec(k,1) = Hx
       hdgEM(iCell)%H%vec(k,2) = Hy
       hdgEM(iCell)%H%vec(k,3) = Hz
    ENDDO
    !
    DEALLOCATE(LambdaLoc)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE LOCAL_SOLVER_WITHOUT_CURRENT_SRC

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE LOCAL_SOLVER_WITH_CURRENT_SRC(iCell)
    !------------------------------------------------------------------
    !---- Solve local linear systems in terms of DoFs of the EM field
    !
    !---- IN
    !---- Cell index : iCell
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_MATRICES_STRUCTURES
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: indLoc, j, k, kFac, nDofFk, nDofV, facNum

    COMPLEX(C_DOUBLE_COMPLEX)                            :: Ex, Ey, Ez
    COMPLEX(C_DOUBLE_COMPLEX)                            :: Hx, Hy, Hz 
    COMPLEX(C_DOUBLE_COMPLEX)                            :: zero
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: LambdaLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Solve the local linear system in terms of the DoFs of the EM
    !---- field in cell with index 'iCell' in data structure 'MeshCell'
    !------------------------------------------------------------------
    zero = (0.d0, 0.d0)
    ALLOCATE(LambdaLoc(Qe(iCell)%ndof2))
    DO indLoc=1,Qe(iCell)%ndof2
       LambdaLoc(indLoc) = zero
    ENDDO
    !
    DO k=1,MeshCell(iCell)%nNeighb
       facNum = MeshCell(iCell)%facNum(k)
       nDofFk = Lambda(facNum)%nDof
       indLoc = 2*(hdgEM(iCell)%sumDofF(k) - nDofFk)
       DO j=1,nDofFk
          LambdaLoc(indLoc + j)          = Lambda(facNum)%L%vec(j,1)
          LambdaLoc(indloc + j + nDofFk) = Lambda(facNum)%L%vec(j,2)
       ENDDO
    ENDDO
    !
    nDofV = hdgEM(iCell)%nDofV
    DO k=1,nDofV
       Ex = zero
       Ey = zero
       Ez = zero
       Hx = zero
       Hy = zero
       Hz = zero
       DO j=1,Qe(iCell)%nDof2
          Ex = Ex + Qe(iCell)%tab(k,j)*LambdaLoc(j)
          Ey = Ey + Qe(iCell)%tab(k +   nDofV,j)*LambdaLoc(j)
          Ez = Ez + Qe(iCell)%tab(k + 2*nDofV,j)*LambdaLoc(j)
          Hx = Hx + Qe(iCell)%tab(k + 3*nDofV,j)*LambdaLoc(j)
          Hy = Hy + Qe(iCell)%tab(k + 4*nDofV,j)*LambdaLoc(j)
          Hz = Hz + Qe(iCell)%tab(k + 5*nDofV,j)*LambdaLoc(j)
       ENDDO
       hdgEM(iCell)%E%vec(k,1) = Ex - rhsJloc(iCell)%tab(k          ) 
       hdgEM(iCell)%E%vec(k,2) = Ey - rhsJloc(iCell)%tab(k +   nDofV)
       hdgEM(iCell)%E%vec(k,3) = Ez - rhsJloc(iCell)%tab(k + 2*nDofV)
       hdgEM(iCell)%H%vec(k,1) = Hx - rhsJloc(iCell)%tab(k + 3*nDofV)
       hdgEM(iCell)%H%vec(k,2) = Hy - rhsJloc(iCell)%tab(k + 4*nDofV)
       hdgEM(iCell)%H%vec(k,3) = Hz - rhsJloc(iCell)%tab(k + 5*nDofV)
    ENDDO
    !
    DEALLOCATE(LambdaLoc)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE LOCAL_SOLVER_WITH_CURRENT_SRC

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE HDG_CORE
