MODULE HDG_MATRICES_STRUCTURES

  !====================================================================
  ! Definition, data structure and routine for assembling 
  ! HDG elementary matrices
  !
  ! *For a complete description of HDG elementary matrices refer to
  !  ~/horse/documents/HDGMinHORSE.pdf
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared module
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  USE MATRICES_STRUCTURES
  USE INPUT_DATA
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! HDG elementary matrices
  !--------------------------------------------------------------------
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: Ae, Be, Ce, Ge

  TYPE(matrix_sq_complex),   DIMENSION(:),   ALLOCATABLE :: Ke

  TYPE(matrix_rect_complex), DIMENSION(:),   ALLOCATABLE :: Qe
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_HDGMATRICES
    !------------------------------------------------------------------
    !---- Compute HDG elementary matrices
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE NUMERICAL_TYPES
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, nDofV, nNeighb, sumDofF
    !------------------------------------------------------------------
    !---- Allocations of Ke and Qe matrices
    !------------------------------------------------------------------
    ALLOCATE(Ke(Mesh%nCell), Qe(Mesh%nCell))
    !------------------------------------------------------------------
    !---- Assembling hdg elementary matrices
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       CALL ALLOCATE_HDG_MATRICES(iCell)
       !
       CALL HDG_MATRIX_AE(iCell)
       CALL HDG_MATRIX_CE(iCell)
       CALL HDG_MATRIX_BE(iCell)
       CALL HDG_MATRIX_GE(iCell)
       CALL HDG_MATRIX_QE(iCell)
       CALL HDG_MATRIX_KE(iCell)
       !
       nDofV   = hdgEM(iCell)%nDofV
       nNeighb = MeshCell(iCell)%nNeighb
       sumDofF = hdgEM(iCell)%sumDofF(nNeighb)
       CALL COMPUTE_RHS_CURRENT_SRC(iCell, nDofV, Ae, sumDofF, Be)
       !
       CALL DEALLOCATE_HDG_MATRICES(iCell)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_HDGMATRICES

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE ALLOCATE_HDG_MATRICES(iCell)
    !------------------------------------------------------------------
    !---- Allocate and initialize hdg elementary matrices
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: nDofV, nNeighb, sumDofF

    COMPLEX(C_DOUBLE_COMPLEX) :: zero
    !------------------------------------------------------------------
    !---- Dynamic allocation of hdg matrices
    !------------------------------------------------------------------
    nDofV           = hdgEM(iCell)%nDofV
    nNeighb         = MeshCell(iCell)%nNeighb
    sumDofF         = hdgEM(iCell)%sumDofF(nNeighb)
    ALLOCATE(Ae(6*nDofV,6*nDofV))
    ALLOCATE(Ce(6*nDofV,2*sumDofF))
    ALLOCATE(Be(2*sumDofF,6*nDofV))
    ALLOCATE(Ge(2*sumDofF,2*sumDofF))
    Qe(iCell)%ndof1 = 6*nDofV
    Qe(iCell)%ndof2 = 2*sumDofF
    ALLOCATE(Qe(iCell)%tab(Qe(iCell)%ndof1,Qe(iCell)%ndof2))
    Ke(iCell)%ndof  = 2*sumDofF
    ALLOCATE(Ke(iCell)%tab(Ke(iCell)%ndof,Ke(iCell)%ndof))
    !------------------------------------------------------------------
    !---- Initialization of hdg matrices
    !------------------------------------------------------------------
    zero               = (0.d0, 0.d0)
    Ae(:,:)            = zero
    Ce(:,:)            = zero
    Be(:,:)            = zero
    Ge(:,:)            = zero
    Qe(iCell)%tab(:,:) = zero
    Ke(iCell)%tab(:,:) = zero
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_HDG_MATRICES

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE DEALLOCATE_HDG_MATRICES(iCell)
    !------------------------------------------------------------------
    !---- Deallocate elementary hdg matrices
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocation of Ae, Ce, Be and Ge hdg matrices
    !------------------------------------------------------------------
    DEALLOCATE( Ae, Ce, Be, Ge )
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DEALLOCATE_HDG_MATRICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HDG_MATRIX_AE(iCell)
    !------------------------------------------------------------------
    !---- Assembling the hdg elementary matrix Ae
    !---- *For a complete description of the matrix Ae refer to
    !----  eq.(16) in ~/horse/documents/HDGMinHORSE.pdf
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MATH_PHYS_CONSTANTS
    USE MESH_CELLS
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE REFERENCE_CELLS
    USE SOURCE_STRUCTURES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: i, ik, iPerm, j, jk, k, nDofS, nDofV
    INTEGER(C_INT), DIMENSION(4) :: iTetVert
    INTEGER(C_INT), DIMENSION(8) :: iHexVert

    REAL(C_DOUBLE)               :: aera, p1, p2, p3, tauCell
    REAL(C_DOUBLE), DIMENSION(3) :: cp2, cp3, cp4, ridge2_1, ridge3_1, &
         ridge4_1, ridge5_1, ridge2_3, ridge4_3, v1, v2, v3, v4, v5 
    REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: Dex, Dey, Dez, Eex, &
         Eexy, Eexz, Eey, Eeyz, Eez

    COMPLEX(C_DOUBLE_COMPLEX) :: tmp1, tmp2, tmpx, tmpxy, tmpxz, tmpy, &
         tmpyz, tmpz
    !------------------------------------------------------------------
    !----Mass matrix contributions
    !------------------------------------------------------------------
    nDofV = hdgEM(iCell)%nDofV
    tmp1  = (iunit*omega*hdgEM(iCell)%eps+hdgEM(iCell)%sigma)*MeshCell(iCell)%measure
    tmp2  = iunit*omega*hdgEM(iCell)%mu*MeshCell(iCell)%measure
    !
    DO k=1,3
       DO j=1,nDofV
          DO i=1,nDofV
             ik        = (k-1)*nDofV + i
             jk        = (k-1)*nDofV + j
             Ae(ik,jk) = tmp1*hdgEM(iCell)%massMtx%tab(i,j)
             !
             ik        = (k+2)*nDofV + i
             jk        = (k+2)*nDofV + j
             Ae(ik,jk) = tmp2*hdgEM(iCell)%massMtx%tab(i,j)
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Stiffness matrices contribution
    !------------------------------------------------------------------
    ALLOCATE( Dex(nDofV,nDofV), Dey(nDofV,nDofV), Dez(nDofV,nDofV) )
    !
    SELECT CASE(MeshCell(iCell)%cellType)
    CASE("TET1","TET2")
       !
       !---- Get tet vertex indices
       iTetVert(:) = MeshCell(iCell)%vertex(:)
       !
       !---- Get coordinates of the vertices
       v1(:) = MeshVert(iTetVert(1))%coor(:)
       v2(:) = MeshVert(iTetVert(2))%coor(:)
       v3(:) = MeshVert(iTetVert(3))%coor(:)
       v4(:) = MeshVert(iTetVert(4))%coor(:)
       !
       !---- Build ridge vectors
       ridge2_1(:) = v2(:) - v1(:)
       ridge3_1(:) = v3(:) - v1(:)
       ridge4_1(:) = v4(:) - v1(:)
       ridge2_3(:) = v2(:) - v3(:)
       ridge4_3(:) = v4(:) - v3(:)
       !
       !---- Compute cross product
       CALL CROSS_PROD( ridge3_1,  ridge4_3, cp2)
       CALL CROSS_PROD(-ridge2_1,  ridge4_1, cp3)
       CALL CROSS_PROD( ridge2_1, -ridge2_3, cp4)
       !
       DO j=1,nDofV
          DO i=1,nDofV 
             Dex(i,j) = hdgEM(iCell)%xStfMtx%tab(i,j)*cp2(1) & 
                  + hdgEM(iCell)%yStfMtx%tab(i,j)*cp3(1)     &
                  + hdgEM(iCell)%zStfMtx%tab(i,j)*cp4(1)
             Dey(i,j) = hdgEM(iCell)%xStfMtx%tab(i,j)*cp2(2) & 
                  + hdgEM(iCell)%yStfMtx%tab(i,j)*cp3(2)     &
                  + hdgEM(iCell)%zStfMtx%tab(i,j)*cp4(2)
             Dez(i,j) = hdgEM(iCell)%xStfMtx%tab(i,j)*cp2(3) & 
                  + hdgEM(iCell)%yStfMtx%tab(i,j)*cp3(3)     & 
                  + hdgEM(iCell)%zStfMtx%tab(i,j)*cp4(3)
          ENDDO
       ENDDO
    CASE("HEX1","HEX2")
       !
       !---- Get hex vertex indices
       iHexVert(:) = MeshCell(iCell)%vertex(:)
       !
       !---- Get coordinates of the vertices
       v1(:) = MeshVert(iHexVert(1))%coor(:)
       v2(:) = MeshVert(iHexVert(2))%coor(:)
       v4(:) = MeshVert(iHexVert(4))%coor(:)
       v5(:) = MeshVert(iHexVert(5))%coor(:)
       !
       !---- Build ridge vectors
       ridge2_1(:) = v2(:) - v1(:)
       ridge4_1(:) = v4(:) - v1(:)
       ridge5_1(:) = v5(:) - v1(:)
       !
       p1 = ridge4_1(2)*ridge5_1(3)
       p2 = ridge5_1(3)*ridge2_1(1)
       p3 = ridge2_1(1)*ridge4_1(2)
       !
       DO j=1,nDofV
          DO i=1,nDofV 
             Dex(i,j) = hdgEM(iCell)%xStfMtx%tab(i,j)*p1
             Dey(i,j) = hdgEM(iCell)%yStfMtx%tab(i,j)*p2
             Dez(i,j) = hdgEM(iCell)%zStfMtx%tab(i,j)*p3
          ENDDO
       ENDDO
    ENDSELECT
    
    DO j=1,nDofV
       DO i=1,nDofV
          Ae(i,nDofV*4+j)         = -Dez(i,j)
          Ae(i,nDofV*5+j)         =  Dey(i,j)
          Ae(nDofV+i,nDofV*3+j)   =  Dez(i,j)
          Ae(nDofV+i,nDofV*5+j)   = -Dex(i,j)
          Ae(2*nDofV+i,nDofV*3+j) = -Dey(i,j)
          Ae(2*nDofV+i,nDofV*4+j) =  Dex(i,j)
          Ae(3*nDofV+i,nDofV+j)   = -Dez(j,i)
          Ae(3*nDofV+i,nDofV*2+j) =  Dey(j,i)
          Ae(4*nDofV+i,j)         =  Dez(j,i)
          Ae(4*nDofV+i,2*nDofV+j) = -Dex(j,i)
          Ae(5*nDofV+i,j)         = -Dey(j,i)
          Ae(5*nDofV+i,nDofV+j)   =  Dex(j,i)
       ENDDO
    ENDDO
    !
    DEALLOCATE( Dex, Dey, Dez )
    !------------------------------------------------------------------
    !---- Sum of the integration on the faces
    !------------------------------------------------------------------
    ALLOCATE( Eex(nDofV,nDofV), Eexy(nDofV,nDofV), Eexz(nDofV,nDofV) )
    ALLOCATE( Eey(nDofV,nDofV), Eeyz(nDofV,nDofV) )
    ALLOCATE( Eez(nDofV,nDofV) )
    !
    DO j=1,nDofV
       DO i=1,nDofV
          Eex(i,j)  = (0.d0,0.d0)
          Eexy(i,j) = (0.d0,0.d0)
          Eexz(i,j) = (0.d0,0.d0)
          Eey(i,j)  = (0.d0,0.d0)
          Eeyz(i,j) = (0.d0,0.d0)
          Eez(i,j)  = (0.d0,0.d0)
       ENDDO
    ENDDO
    !
    tauCell = hdgEM(iCell)%tau
    nDofS   = hdgEM(iCell)%nDofS
    !
    DO k=1,MeshCell(iCell)%nNeighb
       iPerm = MeshCell(iCell)%perm(1,k)
       aera  = MeshCell(iCell)%facSurf(k)
       !
       tmpx  = tauCell*aera*(1 - (MeshCell(iCell)%normal(1,k))**2)
       tmpy  = tauCell*aera*(1 - (MeshCell(iCell)%normal(2,k))**2)
       tmpz  = tauCell*aera*(1 - (MeshCell(iCell)%normal(3,k))**2)
       !
       tmpxy = tauCell*MeshCell(iCell)%normal(1,k)*MeshCell(iCell)%normal(2,k)*aera
       tmpxz = tauCell*MeshCell(iCell)%normal(1,k)*MeshCell(iCell)%normal(3,k)*aera
       tmpyz = tauCell*MeshCell(iCell)%normal(2,k)*MeshCell(iCell)%normal(3,k)*aera
       !
       DO j=1,nDofS
          jk = hdgEM(iCell)%iPermDof%ptr%tab(j,iPerm)
          DO i=1,nDofS
             ik = hdgEM(iCell)%iPermDof%ptr%tab(i,iPerm)
             !
             Eex(ik,jk)  = Eex(ik,jk)  + tmpx*hdgEM(iCell)%srfMtxE(k)%ptr%tab(i,j)
             Eexy(ik,jk) = Eexy(ik,jk) + tmpxy*hdgEM(iCell)%srfMtxE(k)%ptr%tab(i,j)
             Eexz(ik,jk) = Eexz(ik,jk) + tmpxz*hdgEM(iCell)%srfMtxE(k)%ptr%tab(i,j)
             Eey(ik,jk)  = Eey(ik,jk)  + tmpy*hdgEM(iCell)%srfMtxE(k)%ptr%tab(i,j)
             Eeyz(ik,jk) = Eeyz(ik,jk) + tmpyz*hdgEM(iCell)%srfMtxE(k)%ptr%tab(i,j)
             Eez(ik,jk)  = Eez(ik,jk)  + tmpz*hdgEM(iCell)%srfMtxE(k)%ptr%tab(i,j)
          ENDDO
       ENDDO
    ENDDO
    !
    DO j=1,nDofV
       DO i=1,nDofV
          Ae(3*nDofV+i,3*nDofV+j) =  Ae(3*nDofV+i,3*nDofV+j) + Eex(i,j)
          Ae(3*nDofV+i,4*nDofV+j) = -Eexy(i,j)
          Ae(4*nDofV+i,3*nDofV+j) = -Eexy(i,j)
          Ae(3*nDofV+i,5*nDofV+j) = -Eexz(i,j)
          Ae(5*nDofV+i,3*nDofV+j) = -Eexz(i,j)
          Ae(4*nDofV+i,4*nDofV+j) =  Ae(4*nDofV+i,4*nDofV+j) + Eey(i,j)
          Ae(4*nDofV+i,5*nDofV+j) = -Eeyz(i,j)
          Ae(5*nDofV+i,4*nDofV+j) = -Eeyz(i,j)
          Ae(5*nDofV+i,5*nDofV+j) =  Ae(5*nDofV+i,5*nDofV+j) + Eez(i,j)
       ENDDO
    ENDDO
    !
    DEALLOCATE( Eex, Eexy, Eexz, Eey, Eeyz, Eez )
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE HDG_MATRIX_AE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HDG_MATRIX_CE(iCell)
    !------------------------------------------------------------------
    !---- Assembling the hdg elementary matrix Ce
    !---- *For a complete description of the matrix Ce refer to
    !----  eq.(16) in ~/horse/documents/HDGMinHORSE.pdf
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, i, ik, iPerm, j, jk, jk2, k, nDofFk,     &
         nDofS, nDofV, sumDofFk

    REAL(C_DOUBLE) :: aera, tauCell

    COMPLEX(C_DOUBLE_COMPLEX) :: tmp_nxuz, tmp_nxwz, tmp_nyux, tmp_nywx,&
         tmp_ux, tmp_uy, tmp_uz, tmp_wx, tmp_nzuy, tmp_nzwy,   &
         tmp_wy, tmp_wz
    !------------------------------------------------------------------
    !---- Compute hdg elementary matrix Ce
    !------------------------------------------------------------------
    nDofV   = hdgEM(iCell)%nDofV
    nDofS   = hdgEM(iCell)%nDofS
    tauCell = hdgEM(iCell)%tau
    !
    DO k=1,MeshCell(iCell)%nNeighb
       aera     = MeshCell(iCell)%facSurf(k)
       facNum  = MeshCell(iCell)%facNum(k)
       nDofFk   = hdgEM(iCell)%nDofF(k)
       sumDofFk = 2*(hdgEM(iCell)%sumDofF(k) - nDofFk)
       iPerm    = MeshCell(iCell)%perm(1,k)
       !
       tmp_nzuy = -(MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(2,1)  &
            - MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(3,1))*aera
       tmp_nxuz = -(MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(3,1)  &
            - MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(1,1))*aera
       tmp_nyux = -(MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(1,1)  &
            - MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(2,1))*aera
       tmp_nzwy = -(MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(2,2)  &
            - MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(3,2))*aera
       tmp_nxwz = -(MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(3,2)  &
            - MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(1,2))*aera
       tmp_nywx = -(MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(1,2)  &
            - MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(2,2))*aera
       !
       tmp_ux   = -tauCell*MeshFac(facNum)%axis(1,1)*aera
       tmp_uy   = -tauCell*MeshFac(facNum)%axis(2,1)*aera
       tmp_uz   = -tauCell*MeshFac(facNum)%axis(3,1)*aera
       tmp_wx   = -tauCell*MeshFac(facNum)%axis(1,2)*aera
       tmp_wy   = -tauCell*MeshFac(facNum)%axis(2,2)*aera
       tmp_wz   = -tauCell*MeshFac(facNum)%axis(3,2)*aera
       !
       DO j=1,nDofFk
          jk  = sumDofFk + j
          jk2 = sumDofFk + nDofFk + j
          DO i=1,nDofS
             ik = hdgEM(iCell)%iPermDof%ptr%tab(i,iPerm)
             !
             Ce(ik,jk)          = tmp_nzuy*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(ik,jk2)         = tmp_nzwy*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV+ik,jk)    = tmp_nxuz*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV+ik,jk2)   = tmp_nxwz*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*2+ik,jk)  = tmp_nyux*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*2+ik,jk2) = tmp_nywx*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             !
             Ce(nDofV*3+ik,jk)  = tmp_ux*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*3+ik,jk2) = tmp_wx*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*4+ik,jk)  = tmp_uy*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*4+ik,jk2) = tmp_wy*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*5+ik,jk)  = tmp_uz*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
             Ce(nDofV*5+ik,jk2) = tmp_wz*hdgEM(iCell)%srfMtxF(k)%ptr%tab(i,j)
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE HDG_MATRIX_CE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HDG_MATRIX_BE(iCell)
    !------------------------------------------------------------------
    !---- Assembling the hdg elementary matrix Be
    !---- *For a complete description of the matrix Be refer to
    !----  eq.(22) in ~/horse/documents/HDGMinHORSE.pdf
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, j, nDofV, nNeighb, sumDofF
    !------------------------------------------------------------------
    !---- Compute the hdg elementary matrix Be
    !------------------------------------------------------------------
    nDofV   = hdgEM(iCell)%nDofV
    nNeighb = MeshCell(iCell)%nNeighb
    sumDofF = hdgEM(iCell)%sumDofF(nNeighb)
    !
    DO i=1,2*sumDofF
       DO j=1,3*nDofV
          Be(i,j) = Ce(j,i)
       ENDDO
       !
       DO j=3*nDofV+1,6*nDofV
          Be(i,j) = -Ce(j,i)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE HDG_MATRIX_BE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HDG_MATRIX_GE(iCell)
    !------------------------------------------------------------------
    !---- Assembling the hdg elementary matrix Ge
    !---- *For a complete description of the matrix Ge refer to
    !----  eq.(22) in ~/horse/documents/HDGMinHORSE.pdf
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, ik1, ik2, j, jk1, jk2, k, nDofFk, sumDofFk, facNum

    REAL(C_DOUBLE) :: aera, scal_uw

    REAL(C_DOUBLE) :: kappa, tauCell
    !------------------------------------------------------------------
    !---- Compute the hdg elementary matrix Ge
    !------------------------------------------------------------------
    tauCell = hdgEM(iCell)%tau
    DO k=1,MeshCell(iCell)%nNeighb
       aera     = MeshCell(iCell)%facSurf(k)
       facNum  = MeshCell(iCell)%facNum(k)
       nDofFk   = hdgEM(iCell)%nDofF(k)
       sumDofFk = 2*(hdgEM(iCell)%sumDofF(k) - nDofFk)
       scal_uw  = MeshFac(facNum)%axis(1,1)*MeshFac(facNum)%axis(1,2) &
            + MeshFac(facNum)%axis(2,1)*MeshFac(facNum)%axis(2,2)     &
            + MeshFac(facNum)%axis(3,1)*MeshFac(facNum)%axis(3,2)
       !
       IF (MeshFac(facNum)%log == 4) THEN
          kappa = tauCell + 1.d0
       ELSE IF ((hdgMethod /= 0) .AND. (MeshFac(facNum)%log == 10)) THEN
          kappa = tauCell + hdgEM(iCell)%z
       ELSE
          kappa = tauCell
       ENDIF
       !
       DO j=1,nDofFk
          jk1 = sumDofFk + j
          jk2 = sumDofFk + nDofFk + j
          DO i=1,nDofFk
             ik1 = sumDofFk + i
             ik2 = sumDofFk + nDofFk + i
             !
             Ge(ik1,jk1) = -kappa*aera*hdgEM(iCell)%srfMtxG(k)%ptr%tab(i,j) 
             Ge(ik1,jk2) = -kappa*scal_uw*aera*hdgEM(iCell)%srfMtxG(k)%ptr%tab(i,j)
             Ge(ik2,jk1) = -kappa*scal_uw*aera*hdgEM(iCell)%srfMtxG(k)%ptr%tab(i,j)
             Ge(ik2,jk2) = -kappa*aera*hdgEM(iCell)%srfMtxG(k)%ptr%tab(i,j)
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE HDG_MATRIX_GE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HDG_MATRIX_QE(iCell)
    !------------------------------------------------------------------
    !---- Assembling the hdg elmentary matrix Qe := -[Ae]^(-1)*Ce
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: nDof1, nDof2, ierr, i

    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: ipiv
    !------------------------------------------------------------------
    !---- Size of the hdg elementary matrix Qe
    !------------------------------------------------------------------
    nDof1 = Qe(iCell)%ndof1
    nDof2 = Qe(iCell)%ndof2
    !------------------------------------------------------------------
    !---- Compute [Ae]^(-1) (result ==> Ae)
    !------------------------------------------------------------------
    !
    !---- ZGETRF : LAPACK routine computes an LU factorization of a 
    !---- general M-by-N matrix A using partial pivoting with row 
    !---- interchanges (cf LAPACK documentation for more details)
    ALLOCATE(ipiv(nDof1))
    CALL ZGETRF(nDof1, nDof1, Ae, nDof1, ipiv, ierr)
    !------------------------------------------------------------------
    !---- Compute [Ae]^(-1)*Ce (result ==> Ce)
    !------------------------------------------------------------------
    !
    !---- ZGETRS : LAPACK routine solves a system of linear equations 
    !---- A*X=B,  A**T*X=B, or  A**H*X=B with a general N-by-N matrix A 
    !---- using the LU factorization computed by ZGETRF (cf LAPACK 
    !---- documentation for more details)
    CALL ZGETRS('N', nDof1, nDof2, Ae, nDof1, ipiv, Ce, nDof1, ierr)
    DEALLOCATE(ipiv)
    !------------------------------------------------------------------
    !---- Compute the hdg elementary matrix Qe
    !------------------------------------------------------------------
    Ce(:,:) = -Ce(:,:)
    Qe(iCell)%tab(:,:) = Ce(:,:)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END SUBROUTINE HDG_MATRIX_QE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE HDG_MATRIX_KE(iCell)
    !------------------------------------------------------------------
    !---- Assembling the hdg elementary matrix Ke := Ge - Be*[Ae]^(-1)*Ce
    !
    !---- IN
    !---- iCell : element
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: AeSize, KeSize

    COMPLEX(C_DOUBLE_COMPLEX) :: one
    !--------------------------------------------------------------------
    !---- Compute Ge - Be*[Ae]^(-1)*Ce (result ==> Ge)
    !--------------------------------------------------------------------
    !
    !---- ZGEMM : LAPACK routine performs one of the matrix-matrix operations 
    !---- C := alpha*op(A)*op(B) + beta*C, where:  
    !----   - op(X) is one of op(X)=X or op(X)=X' or op(X)=conjg(X') 
    !----   - alpha and beta are scalars
    !----   - A, B and C are matrices
    !----   - op(A) an m by k matrix, op(B) a k by n matrix 
    !----   - C an m by n matrix
    !---- (cf LAPACK documentation for more details)
    AeSize = 6*hdgEM(iCell)%nDofV
    KeSize = Ke(iCell)%ndof
    one   = (1.d0, 0.d0)
    CALL ZGEMM('N', 'N', KeSize, KeSize, AeSize, one, Be, KeSize, Ce, AeSize, &
         one, Ge, KeSize)
    !------------------------------------------------------------------
    !---- Compute the hdg elmentary matrix Ke
    !------------------------------------------------------------------
    Ke(iCell)%tab(:,:) = Ge(:,:)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE HDG_MATRIX_KE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE HDG_MATRICES_STRUCTURES
