MODULE HDG_SOLVER

  !====================================================================
  ! Everything related with the hdg solver
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------


  !--------------------------------------------------------------------
  ! COMPUTE_LAMBDA_FIELD  ==> Procedure pointers to subroutine computing 
  !                           the global solver in terms od DoFs of the 
  !                           hybrid variable
  !--------------------------------------------------------------------
  PROCEDURE(), POINTER :: COMPUTE_LAMBDA_FIELD
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_HDG_SOLVER(Method, SolverPackage)
    !------------------------------------------------------------------
    !---- Compute hdgm
    !
    !---- IN
    !---- Method        : hdgm used (cf hdgMethod in dataHorse)
    !---- SolverPackage : solver package used for solving the global 
    !                     linear system in hdgm (cf solverPack in dataHorse)
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_CORE
    USE KRYLOV_CORE
    USE KRYLOV_STRUCTURES
    USE MAPHYS_STRUCTURES
    USE MESH_DATA
    USE MUMPS_STRUCTURES
    USE NUMERICAL_TYPES
    USE SCHWARZ_CORE
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                :: Method
    CHARACTER(LEN=8, KIND=C_CHAR), INTENT(IN) :: SolverPackage
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, k, nDof
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute absorbing boundary condition contribution for RHS
    !------------------------------------------------------------------
    CALL COMPUTE_RHS_ABS
    !------------------------------------------------------------------
    !---- Compute the hybrid variable i.e. Lambda field
    !------------------------------------------------------------------
    !
    !---- Make COMPUTE_LAMBDA_FIELD points to the right routine
    SELECT CASE(Method)
    CASE(0)
       !
       !---- Classical hdgm
       SELECT CASE(TRIM(SolverPackage))
       CASE("mumps")
          COMPUTE_LAMBDA_FIELD => COMPUTE_MUMPS_SOLVER
       CASE("maphys")
          COMPUTE_LAMBDA_FIELD => COMPUTE_MAPHYS_SOLVER
       END SELECT
    CASE(1)
       !
       !---- hdgm combined with a Schwarz algorithm
       COMPUTE_LAMBDA_FIELD => COMPUTE_SCHWARZ_ALGORITHM
    CASE(2)
       !
       !---- hdgm combined with a Schwarz algorithm
       COMPUTE_LAMBDA_FIELD => COMPUTE_KRYLOV_SCHWARZ_ALGORITHM 
    END SELECT
    !
    !---- Compute DoFs of Lambda field
    CALL COMPUTE_LAMBDA_FIELD
    !------------------------------------------------------------------
    !---- Solve local linear systems in terms of DoFs of the EM field
    !---- (local linear systems element-by-element)
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       CALL LOCAL_SOLVER(iCell)
    ENDDO
    !------------------------------------------------------------------
    !---- Specific computation for Schwarz algoritm accelerated by a 
    !---- Krylov subspace method
    !------------------------------------------------------------------
    IF (Method == 2) THEN
       DO iCell=1,Mesh%nCell
          nDof = hdgEM(iCell)%nDofV
          DO k=1,nDof
             hdgEM(iCell)%E%vec(k,1) = hdgEM(iCell)%E%vec(k,1) + EMsave(iCell)%vec(k,1)
             hdgEM(iCell)%E%vec(k,2) = hdgEM(iCell)%E%vec(k,2) + EMsave(iCell)%vec(k,2)
             hdgEM(iCell)%E%vec(k,3) = hdgEM(iCell)%E%vec(k,3) + EMsave(iCell)%vec(k,3)
             hdgEM(iCell)%H%vec(k,1) = hdgEM(iCell)%H%vec(k,1) + EMsave(iCell)%vec(k,4)
             hdgEM(iCell)%H%vec(k,2) = hdgEM(iCell)%H%vec(k,2) + EMsave(iCell)%vec(k,5)
             hdgEM(iCell)%H%vec(k,3) = hdgEM(iCell)%H%vec(k,3) + EMsave(iCell)%vec(k,6)
          ENDDO
       ENDDO
       !
       !---- Deallocate data structures 
       CALL KRYLOV_DATA_STRUCTURES_END
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_HDG_SOLVER

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE HDG_SOLVER
