MODULE KRYLOV_CORE

  !====================================================================
  ! Main hdgm combined with a Schwarz algorithm accelerated by a Krylov 
  ! subspace method routines
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_KRYLOV_SCHWARZ_ALGORITHM
    !------------------------------------------------------------------
    !---- Compute Schwarz algorithm accelerated by a Krylov subspace 
    !---- method
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_CORE
    USE INPUT_DATA
    USE KRYLOV_STRUCTURES
    USE MESH_DATA
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE SCHWARZ_MUMPS_STRUCTURES
    USE SCHWARZ_TYPES
    USE TIMERS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)            :: iCell, k, nDoFs

    TYPE(Timer)               :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of Krylov data structures
    !------------------------------------------------------------------
    CALL KRYLOV_DATA_STRUCTURES_INIT
    !------------------------------------------------------------------
    !---- Initialization of MUMPS control parameters
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_INIT
    !------------------------------------------------------------------
    !---- Compute local index of faces for assembling matrix and RHS in 
    !---- MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_FACES_INDICES
    !------------------------------------------------------------------
    !---- Assembling matrix in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- Analysis of the matrix in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_ANALYSIS
    !------------------------------------------------------------------
    !---- Factorization of the matrix in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_FACTORIZATION
    !------------------------------------------------------------------
    !---- Compute RHS in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_RHS_UPDATE    
    !------------------------------------------------------------------
    !---- Solve the local linear system
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_SOLVE
    !------------------------------------------------------------------
    !---- Compute and save EM field
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       CALL LOCAL_SOLVER(iCell)
    ENDDO
    DO iCell=1,Mesh%nCell
       nDoFs = hdgEM(icell)%nDofV
       DO k=1,nDoFs
          EMsave(iCell)%vec(k,1) = hdgEM(iCell)%E%vec(k,1)
          EMsave(iCell)%vec(k,2) = hdgEM(iCell)%E%vec(k,2)
          EMsave(iCell)%vec(k,3) = hdgEM(iCell)%E%vec(k,3)
          EMsave(iCell)%vec(k,4) = hdgEM(iCell)%H%vec(k,1)
          EMsave(iCell)%vec(k,5) = hdgEM(iCell)%H%vec(k,2)
          EMsave(iCell)%vec(k,6) = hdgEM(iCell)%H%vec(k,3)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- BiCGStab(l) iterative method
    !------------------------------------------------------------------
    CALL BICGSTAB(BiCGStabLevel)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_KRYLOV_SCHWARZ_ALGORITHM

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE KRYLOV_EM_EXCHANGE
    !------------------------------------------------------------------
    !---- Exchange EM field DoFs on interfaces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE KRYLOV_STRUCTURES
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    USE SCHWARZ_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, faceNumArt, cellNum, id, iFac, iNeighb, &
         k, nDoFs 
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Exchange EM field on interfaces
    !------------------------------------------------------------------
    DO iNeighb=1,nbNeighb
       id = 1
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum    = neighb(iNeighb)%faces(iFac)
          faceNumArt = indArt(faceNum)
          cellNum    = MeshFac(faceNum)%neighb(1)
          nDoFs      = hdgEM(cellNum)%nDofS
          DO k=1,nDoFs
             EMsend(id)     = EMold(faceNumArt)%vec(k,1)
             EMsend(id + 1) = EMold(faceNumArt)%vec(k,2)
             EMsend(id + 2) = EMold(faceNumArt)%vec(k,3)
             EMsend(id + 3) = EMold(faceNumArt)%vec(k,4)
             EMsend(id + 4) = EMold(faceNumArt)%vec(k,5)
             EMsend(id + 5) = EMold(faceNumArt)%vec(k,6)
             id             = id + 6
          ENDDO
       ENDDO
       CALL MPI_SENDRECV(EMsend, 6*nDoFsComm(iNeighb), MPI_DOUBLE_COMPLEX, &
            neighb(iNeighb)%id, 1, EMrecv, 6*nDoFsComm(iNeighb),           &
            MPI_DOUBLE_COMPLEX, neighb(iNeighb)%id, 1, defaultComm,        &
            status, mpiError)
       id = 1
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum    = neighb(iNeighb)%faces(iFac)
          faceNumArt = indArt(faceNum)
          nDoFs      = EMex(faceNumArt)%nDoFs
          DO k=1,nDoFs
             EMex(faceNumArt)%E%vec(k,1) = EMrecv(id) 
             EMex(faceNumArt)%E%vec(k,2) = EMrecv(id + 1)
             EMex(faceNumArt)%E%vec(k,3) = EMrecv(id + 2)
             EMex(faceNumArt)%H%vec(k,1) = EMrecv(id + 3)
             EMex(faceNumArt)%H%vec(k,2) = EMrecv(id + 4)
             EMex(faceNumArt)%H%vec(k,3) = EMrecv(id + 5)
             id                          = id + 6
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE KRYLOV_EM_EXCHANGE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE KRYLOV_CORE
