MODULE KRYLOV_STRUCTURES

  !====================================================================                              
  ! Initialize all necessary data structures for Schwarz algorithm 
  ! accelerated by a Krylov subspace method
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  USE MATRICES_STRUCTURES
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Additional data structures (compared to classical Schwarz algorithm, 
  ! cf module SCHWARZ_structures) usefull for Schwarz algorithm 
  ! accelerated by a Krylov subspace method
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: nDofEMart

  TYPE(complex_vector_field), DIMENSION(:), ALLOCATABLE :: EMnew, EMold 
  TYPE(complex_vector_field), DIMENSION(:), ALLOCATABLE :: EMsave
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE KRYLOV_DATA_STRUCTURES_INIT
    !------------------------------------------------------------------
    !---- Inititalization of all data structures for hdgm combined to a 
    !---- Schwarz algorithm accelerated by a Krylov subspace method
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE SCHWARZ_TYPES
    USE SCHWARZ_STRUCTURES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, faceNumArt, iCell, iFac, nDoFs, iNeighb

    COMPLEX(C_DOUBLE_COMPLEX) :: zero
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of Schwarz data structures
    !------------------------------------------------------------------
    CALL SCHWARZ_DATA_STRUCTURES_INIT
    !------------------------------------------------------------------
    !---- Initialization of usefull additional data structures for Krylov
    !------------------------------------------------------------------
    !
    !---- nDofEMart : number of DoFs of EM field on interfaces i.e. on 
    !---- artificial boundary faces
    nDofEMart = 0
    DO iFac=1,Mesh%nFacArt
       faceNum   = MeshFacArt(iFac)%faceNum
       iCell     = MeshFac(faceNum)%neighb(1)
       nDoFs     = hdgEM(iCell)%nDofS
       nDofEMart = nDofEMart + nDoFs
    ENDDO
    !
    zero = (0.d0, 0.d0)
    !
    !---- EMold:  DoFs of EM field on interfaces (at iteration "n")
    ALLOCATE(EMold(Mesh%nFacArt))
    ALLOCATE(EMnew(Mesh%nFacArt))
    DO iFac=1,Mesh%nFacArt
       faceNum    = MeshFacArt(iFac)%faceNum
       faceNumArt = indArt(faceNum)
       iCell      = MeshFac(faceNum)%neighb(1)
       nDoFs      = hdgEM(iCell)%nDofS
       ALLOCATE(EMold(faceNumArt)%vec(nDoFs,6))
       ALLOCATE(EMnew(faceNumArt)%vec(nDoFs,6))
    ENDDO
    DO iFac=1,Mesh%nFacArt
       EMold(iFac)%vec(:,:) = zero
       EMnew(iFac)%vec(:,:) = zero
    ENDDO
    !
    !---- EMsave: save DoFs of EM fields before iterative step
    ALLOCATE(EMsave(Mesh%nCell))
    DO iCell=1,Mesh%nCell
       nDoFs = hdgEM(iCell)%nDofV
       ALLOCATE(EMsave(iCell)%vec(nDoFs,6))
    ENDDO
    DO iCell=1,Mesh%nCell
       EMsave(iCell)%vec(:,:) = zero
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE KRYLOV_DATA_STRUCTURES_INIT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE KRYLOV_DATA_STRUCTURES_END
    !------------------------------------------------------------------
    !---- Deallocate all data structures defined for hdgm combined to a 
    !---- Schwarz algorithm accelerated by a Krylov subspace method
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE SCHWARZ_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Deallocate all data structures defined for Schwarz algorithm
    !------------------------------------------------------------------
    CALL SCHWARZ_DATA_STRUCTURES_END
    !------------------------------------------------------------------
    !---- Deallocate data structures defined for in this module
    !------------------------------------------------------------------
    !
    !---- EMold:  DoFs of EM field on interfaces (at iteration "n")
    DEALLOCATE(EMold,EMnew)
    !
    !---- EMsave: DoFs of EM fields before iterative step
    DO iCell=1,Mesh%nCell
       DEALLOCATE(EMsave(iCell)%vec)
    ENDDO
    DEALLOCATE(EMsave)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE KRYLOV_DATA_STRUCTURES_END

  !====================================================================
  !====================================================================
  !====================================================================  

END MODULE KRYLOV_STRUCTURES
