MODULE L2_ERROR

  !====================================================================
  ! Everything for L2 error computation for cases with exact solution
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  USE SOURCE_STRUCTURES
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Stuff related to L2 error computation
  !
  ! computeErr ==> logical for L2 error computation
  ! EXACT_SOl  ==> Procedure pointer to subroutine computing exact solution element-wise
  !--------------------------------------------------------------------
  LOGICAL(C_BOOL)           :: computeErr

  PROCEDURE(),    POINTER   :: EXACT_SOL => NULL()
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_L2_ERROR
    !------------------------------------------------------------------
    !---- Select the L2 error solution depending on case
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE NUMERICAL_TYPES
    USE PLANE_WAVE
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Make EXACT_SOL points to the right subroutine
    !------------------------------------------------------------------
    SELECT CASE(sourceType)
    CASE('PW')
       computeErr =  computeErrPW
       IF (.NOT. computeErr) RETURN
       EXACT_SOL  => CELL_SOL_PLANE_WAVE_FREQ
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_L2_ERROR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_L2_ERROR(omegaValue)
    !------------------------------------------------------------------
    !---- Compute L2 error 
    !
    !---- IN
    !---- omegaValue : angular frequency
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MESH_DATA
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN) :: omegaValue
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, nDofV, order

    REAL(C_DOUBLE) :: genL2err, genL2errE, genL2errH, L2err, L2errE,   &
         L2errH, sumL2err, sumL2errE, sumL2errH, vol

    COMPLEX(C_DOUBLE_COMPLEX)                                  ::      &
         L2errEx, L2errEy, L2errEz,                                    &
         L2errHx, L2errHy, L2errHz,                                    &
         one,     zero 
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(nDofVMax)             ::      &
         Esx, Esy, Esz,                                                &
         Hsx, Hsy, Hsz
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:),     ALLOCATABLE ::      &
         CmplxMassMtx
    !------------------------------------------------------------------
    !---- External function
    !------------------------------------------------------------------
    COMPLEX*16, EXTERNAL :: ZDOTC
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. computeErr) RETURN
    !------------------------------------------------------------------
    !---- Compute L2 error
    !------------------------------------------------------------------
    one  = (1.d0, 0.d0)
    zero = (0.d0, 0.d0)
    !
    L2errE = 0.d0
    L2errH = 0.d0
    L2err  = 0.d0
    DO iCell=1,Mesh%nCell
       nDofV = hdgEM(iCell)%nDofV
       order = hdgEM(iCell)%order
       vol   = MeshCell(iCell)%measure
       CALL EXACT_SOL(iCell, nDofV, order, omegaValue)
       ExVcx(:) = hdgEM(iCell)%E%vec(:,1) - ExVcx(:)
       EyVcx(:) = hdgEM(iCell)%E%vec(:,2) - EyVcx(:)
       EzVcx(:) = hdgEM(iCell)%E%vec(:,3) - EzVcx(:)
       HxVcx(:) = hdgEM(iCell)%H%vec(:,1) - HxVcx(:)
       HyVcx(:) = hdgEM(iCell)%H%vec(:,2) - HyVcx(:)
       HzVcx(:) = hdgEM(iCell)%H%vec(:,3) - HzVcx(:)
       !
       !---- Allocation and initialization of the complex mass matrix 
       !---- i.e. CmplxMassMtx
       !
       !---- Warning : the matrix in Lapack ZGEMV routine must be complex 
       !               and have the right size !
       ALLOCATE(CmplxMassMtx(nDofV,nDofV))
       CmplxMassMtx(:,:) = hdgEM(iCell)%massMtx%tab(:,:)
       !
       CALL ZGEMV('N', nDofV, nDofV, one, CmplxMassMtx, nDofV, ExVcx, 1, zero, Esx, 1)
       CALL ZGEMV('N', nDofV, nDofV, one, CmplxMassMtx, nDofV, EyVcx, 1, zero, Esy, 1)
       CALL ZGEMV('N', nDofV, nDofV, one, CmplxMassMtx, nDofV, EzVcx, 1, zero, Esz, 1)
       CALL ZGEMV('N', nDofV, nDofV, one, CmplxMassMtx, nDofV, HxVcx, 1, zero, Hsx, 1)
       CALL ZGEMV('N', nDofV, nDofV, one, CmplxMassMtx, nDofV, HyVcx, 1, zero, Hsy, 1)
       CALL ZGEMV('N', nDofV, nDofV, one, CmplxMassMtx, nDofV, HzVcx, 1, zero, Hsz, 1)
       !
       !---- Deallocate complex mass matrix
       DEALLOCATE(CmplxMassMtx)
       !
       L2errEx = ZDOTC(nDofV, ExVcx, 1, Esx, 1)
       L2errEy = ZDOTC(nDofV, EyVcx, 1, Esy, 1)
       L2errEz = ZDOTC(nDofV, EzVcx, 1, Esz, 1)
       L2errHx = ZDOTC(nDofV, HxVcx, 1, Hsx, 1)
       L2errHy = ZDOTC(nDofV, HyVcx, 1, Hsy, 1)
       L2errHz = ZDOTC(nDofV, HzVcx, 1, Hsz, 1)
       !
       sumL2errE = L2errEx + L2errEy + L2errEz
       sumL2errH = L2errHx + L2errHy + L2errHz
       sumL2err  = sumL2errE + sumL2errH
       !
       L2errE = L2errE + vol*sumL2errE
       L2errH = L2errH + vol*sumL2errH
       L2err  = L2err  + vol*sumL2err
    ENDDO
    !
    IF (parallelRun) THEN
       CALL GLOBAL_OP(L2errE, MPI_SUM, genL2errE)
       CALL GLOBAL_OP(L2errH, MPI_SUM, genL2errH)
       CALL GLOBAL_OP(L2err , MPI_SUM, genL2err)
       L2errE = genL2errE
       L2errH = genL2errH
       L2err  = genL2err
    ENDIF
    L2errE = SQRT(L2errE)
    L2errH = SQRT(L2errH)
    L2err  = SQRT(L2err)
    !------------------------------------------------------------------
    !---- Print L2 error
    !------------------------------------------------------------------
    IF (idWrite) THEN
       WRITE(6, 10)
       WRITE(6, '(a)')       " L2-norm error with the exact solution "
       WRITE(6, 10)
       WRITE(6, '(a,e15.6)') " L2-norm error for E  field : ", L2errE
       WRITE(6, '(a,e15.6)') " L2-norm error for H  field : ", L2errH
       WRITE(6, '(a,e15.6)') " L2-norm error for EM field : ", L2err
       WRITE(6, 10)
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_L2_ERROR

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE POSTPROCESS_L2_ERROR
    !------------------------------------------------------------------
    !---- Finalize L2 error operations
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF (.NOT. computeErr) THEN 
       RETURN
    ELSE
       CALL COMPUTE_L2_ERROR(omega)
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE POSTPROCESS_L2_ERROR

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE L2_ERROR
