MODULE RCS_STRUCTURES

  !====================================================================
  ! Everything related to RCS computations
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  USE MATRICES_STRUCTURES
  USE QUADRATURES_2D
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Quadrature rules order
  !--------------------------------------------------------------------
  INTEGER(C_INT), PARAMETER :: quadOrder = 20
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Exterior element to the Tfsf surface
  !--------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: RCSelt
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Scattered fields
  !--------------------------------------------------------------------
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:),   ALLOCATABLE :: ExSca, EySca, EzSca
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:),   ALLOCATABLE :: HxSca, HySca, HzSca
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: Esca,  Hsca
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Surface currents
  !--------------------------------------------------------------------
  TYPE(matrix_rect_complex), DIMENSION(:), ALLOCATABLE :: Ms, Js
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Radiation vectors
  !--------------------------------------------------------------------
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: NsTheta,      NsPhi
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: LsTheta,      LsPhi
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: NsThetaGlb,   NsPhiGlb
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: LsThetaGlb,   LsPhiGlb
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Incident power density
  !--------------------------------------------------------------------
  REAL(C_DOUBLE) :: Sinc
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! RCS
  !--------------------------------------------------------------------
  REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: RCS
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_RADAR_CROSS_SECTION
    !------------------------------------------------------------------
    !---- Compute Radar Cross Section
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Quadrature rules
    !------------------------------------------------------------------
    CALL QUADRATURES_TRIANGLE(quadOrder)
    !------------------------------------------------------------------
    !---- Allocations of arrays for RCS computation
    !------------------------------------------------------------------
    CALL RCS_ALLOCATIONS
    !------------------------------------------------------------------
    !---- Initialize exterior elements to the Tfsf surface
    !------------------------------------------------------------------
    CALL INITIALIZE_TFSF_ELEMENT
    !------------------------------------------------------------------
    !---- Compute surface currents at quadrature points
    !------------------------------------------------------------------
    DO iFace=1,Mesh%nFacTfsf
       !
       !---- Scattered field
       CALL COMPUTE_SCATTERED_FIELD_QUAD_PTS(iFace)
       !
       !--- Surface currents
       CALL COMPUTE_SURFACE_CURRENTS_QUAD_PTS(iFace)
    ENDDO
    !------------------------------------------------------------------
    !---- Compute radiation vectors
    !------------------------------------------------------------------
    CALL COMPUTE_RADIATION_VECTORS
    ! !------------------------------------------------------------------
    ! !---- Compute incident power density
    ! !------------------------------------------------------------------
    ! CALL COMPUTE_INCIDENT_POWER_DENSITY
    !------------------------------------------------------------------
    !---- Compute RCS
    !------------------------------------------------------------------
    CALL COMPUTE_RCS
    !------------------------------------------------------------------
    !---- Deallocations of arrays for RCS computation
    !------------------------------------------------------------------
    CALL RCS_DEALLOCATIONS
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RADAR_CROSS_SECTION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RCS_ALLOCATIONS
    !------------------------------------------------------------------
    !---- Allocations of arrays for RCS computation
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MESH_DATA
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace

    REAL(C_DOUBLE) :: dPhi, dTheta
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Dynamic allocations
    !------------------------------------------------------------------
    !
    !---- Array for exterior elements to the Tfsf surface
    ALLOCATE(RCSelt(Mesh%nFacTfsf))
    !
    !---- Scattered field
    ALLOCATE(ExSca(nDofSMax), EySca(nDofSMax), EzSca(nDofSMax))
    ALLOCATE(HxSca(nDofSMax), HySca(nDofSMax), HzSca(nDofSMax))
    !
    ALLOCATE(Esca(nq2d(quadOrder),3))
    ALLOCATE(Hsca(nq2d(quadOrder),3))
    !
    !---- Surface currents
    ALLOCATE(Js(Mesh%nFacTfsf), Ms(Mesh%nFacTfsf))
    DO iFace=1,Mesh%nFacTfsf
       Js(iFace)%ndof1 = nq2d(quadOrder)
       Js(iFace)%ndof2 = 3
       ALLOCATE(Js(iFace)%tab(Js(iFace)%ndof1,Js(iFace)%ndof2))
       !
       Ms(iFace)%ndof1 = nq2d(quadOrder)
       Ms(iFace)%ndof2 = 3
       ALLOCATE(Ms(iFace)%tab(Ms(iFace)%ndof1,Ms(iFace)%ndof2))
    ENDDO
    !
    !---- Radiation vectors
    ALLOCATE(NsTheta(nThetaRCS,nPhiRCS),    NsPhi(nThetaRCS,nPhiRCS))
    ALLOCATE(LsTheta(nThetaRCS,nPhiRCS),    LsPhi(nThetaRCS,nPhiRCS))
    ALLOCATE(NsThetaGlb(nThetaRCS,nPhiRCS), NsPhiGlb(nThetaRCS,nPhiRCS))
    ALLOCATE(LsThetaGlb(nThetaRCS,nPhiRCS), LsPhiGlb(nThetaRCS,nPhiRCS))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RCS_ALLOCATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE INITIALIZE_TFSF_ELEMENT
    !------------------------------------------------------------------
    !---- Initialization of the array of exterior elements of the Tfsf
    !---- surface
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: cellNum1, cellNum2, faceNum, iFace, log1, log2

    CHARACTER(LEN=2, KIND=C_CHAR) :: zone1, zone2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    DO iFace=1,Mesh%nFacTfsf
       RCSelt(iFace) = -1
    ENDDO
    !
    DO iFace=1,Mesh%nFacTfsf
       faceNum  = MeshFacTfsf(iFace)%faceNum
       cellNum1 = MeshFac(faceNum)%neighb(1)
       zone1    = MeshCell(cellNum1)%zone
       IF (MeshFac(faceNum)%isArtificial) THEN
          zone2 = 'TF'
       ELSE
          cellNum2 = MeshFac(faceNum)%neighb(2)
          zone2     = MeshCell(cellNum2)%zone
       ENDIF
       IF (zone1 == 'SF') RCSelt(iFace) = cellNum1
       IF (zone2 == 'SF') RCSelt(iFace) = cellNum2
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE INITIALIZE_TFSF_ELEMENT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_SCATTERED_FIELD_QUAD_PTS(iFace)
    !------------------------------------------------------------------
    !---- Compute scattered field at quadrature points
    !
    !---- IN
    !---- iFace
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE BASES_LAGRANGE
    USE MATRICES_STRUCTURES
    USE MESH_CELLS
    USE MESH_FACES
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE REFERENCE_CELLS
    USE SOURCE_STRUCTURES
    USE SPECIAL_FUNCTIONS
    USE MATH_PHYS_CONSTANTS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)                  :: iFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                              :: cellNum, cellNum1,       &
         cellNum2, faceNum, i, iDof, iDofPerm, iPerm, iPlan, j, k, nDofS,   &
         nDofV, neighbext, numLocal, orderTet

    REAL(C_DOUBLE)                              :: dJac, measure, threshold
    REAL(C_DOUBLE), DIMENSION(3)                :: normal
    REAL(C_DOUBLE), DIMENSION(3)                :: ridge1, ridge2
    REAL(C_DOUBLE), DIMENSION(3)                :: vRef1, vRef2, vRef3
    REAL(C_DOUBLE), DIMENSION(3)                :: vRefk, vFack
    REAL(C_DOUBLE), DIMENSION(3)                :: vTmp1, vTmp2, vTmp3
    REAL(C_DOUBLE), DIMENSION(4)                :: x, y, z
    REAL(C_DOUBLE), DIMENSION(3,3)              :: iMtx, Mtx
    REAL(C_DOUBLE), DIMENSION(3,3)              :: vFac
    REAL(C_DOUBLE), DIMENSION(3,4)              :: vTet
    REAL(C_DOUBLE), DIMENSION(:),  ALLOCATABLE  :: phi
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization
    !------------------------------------------------------------------
    Esca(:,:) = (0.d0, 0.d0)
    Hsca(:,:) = (0.d0, 0.d0)
    !------------------------------------------------------------------
    !---- Compute scattered field on Tfsf surface
    !------------------------------------------------------------------
    cellNum = RCSelt(iFace)
    IF (cellNum /= -1) THEN
       faceNum    = MeshFacTfsf(iFace)%faceNum
       cellNum1   = MeshFac(faceNum)%neighb(1)
       cellNum2   = MeshFac(faceNum)%neighb(2)
       IF (cellNum == cellNum1) neighbext = 1
       IF (cellNum == cellNum2) neighbext = 2
       orderTet   = hdgEM(cellNum)%order
       nDofS      = hdgEM(cellNum)%nDofS
       nDofV      = hdgEM(cellNum)%nDofV
       numLocal   = MeshFac(faceNum)%numLocal(neighbext)
       iPerm      = MeshCell(cellNum)%perm(neighbext,numLocal)
       !
       !---- Coordinates of the vertices of the tetrahedron
       vTet(:,1) = MeshVert(MeshCell(cellNum)%vertex(1))%coor(:)
       vTet(:,2) = MeshVert(MeshCell(cellNum)%vertex(2))%coor(:)
       vTet(:,3) = MeshVert(MeshCell(cellNum)%vertex(3))%coor(:)
       vTet(:,4) = MeshVert(MeshCell(cellNum)%vertex(4))%coor(:)
       !
       !---- Matrix of the affine transformation of the reference 
       !---- tetrahedron to the current tetrahedron
       DO i=1,3
          DO j=1,3
             Mtx(i,j) = vTet(i,j+1) - vTet(i,1)
          ENDDO
       ENDDO
       !
       !---- Inverse of the matrix Mtx for the transformation of the 
       !---- current tetrahedron to the reference tetrahedron
       threshold = 1.0e-15
       CALL INVERSE_3X3(Mtx,threshold,iMtx,dJac)
       !
       !---- Coordinates of the face in the reference tetrahedron
       vFac(:,1) = MeshVert(MeshFac(faceNum)%vertex(1))%coor(:)
       vFac(:,2) = MeshVert(MeshFac(faceNum)%vertex(2))%coor(:)
       vFac(:,3) = MeshVert(MeshFac(faceNum)%vertex(3))%coor(:)
       !
       vTmp1(:) = vFac(:,1) - vTet(:,1)
       vTmp2(:) = vFac(:,2) - vTet(:,1)
       vTmp3(:) = vFac(:,3) - vTet(:,1)
       CALL MATVECMUL(iMtx,vTmp1,3,3,vRef1)
       CALL MATVECMUL(iMtx,vTmp2,3,3,vRef2)
       CALL MATVECMUL(iMtx,vTmp3,3,3,vRef3)
       !
       !---- Orientation of the face on the reference tetrahedron
       ridge1(:) = vRef2(:) - vRef1(:)
       ridge2(:) = vRef3(:) - vRef1(:)
       !
       CALL CROSS_PROD(ridge1,ridge2,normal)
       normal = Tref_surf*normal
       CALL NORMALIZE_VECTOR(normal,measure)
       !
       IF ((ABS(normal(1)) < threshold) .AND. (ABS(normal(2)) < threshold)) THEN
          iPlan = 1 ! The face is in the plan <x,y>
       ELSEIF ((ABS(normal(1)) < threshold) .AND. (ABS(normal(3)) < threshold)) THEN
          iPlan = 2 ! The face is in the plan <x,z>
       ELSEIF ((ABS(normal(2)) < threshold ) .AND. (ABS(normal(3)) < threshold)) THEN
          iPlan = 3 ! The face is in the plan <y,z>
       ELSE
          iPlan = 4
       ENDIF
       !------------------------------------------------------------------
       !--- Compute scattered field at DoFs of the face
       !------------------------------------------------------------------
       !
       !---- Initialization
       ExSca(:) = (0.d0, 0.d0)
       EySca(:) = (0.d0, 0.d0)
       EzSca(:) = (0.d0, 0.d0)
       HxSca(:) = (0.d0, 0.d0)
       HySca(:) = (0.d0, 0.d0)
       HzSca(:) = (0.d0, 0.d0)
       !
       !--- Compute incident field at DoFs of the face (incident field
       !--- is stored in ExScx, EyScx, EzScx, HxScx, HyScx and HzScx)
       CALL FACE_SOL(faceNum, nDofS, nDofV, orderTet, omega)
       !
       !---- Scattered field at DoFs of the face (stored in ExSca, EySca,
       !---- EzSca, HxSca, HySca and HzSca
       DO iDof=1,hdgEM(cellNum)%nDofS
          iDofPerm    = hdgEM(cellNum)%iPermDof%ptr%tab(iDof,iPerm)
          ExSca(iDof) = hdgEM(cellNum)%E%vec(iDofPerm,1) - ExScx(iDof)
          EySca(iDof) = hdgEM(cellNum)%E%vec(iDofPerm,2) - EyScx(iDof)
          EzSca(iDof) = hdgEM(cellNum)%E%vec(iDofPerm,3) - EzScx(iDof)
          HxSca(iDof) = hdgEM(cellNum)%H%vec(iDofPerm,1) - HxScx(iDof)
          HySca(iDof) = hdgEM(cellNum)%H%vec(iDofPerm,2) - HyScx(iDof)
          HzSca(iDof) = hdgEM(cellNum)%H%vec(iDofPerm,3) - HzScx(iDof)
       ENDDO
       !------------------------------------------------------------------
       !---- Compute scattered field at quadrature points (scattered field 
       !---- is stored in Esca and Hsca
       !------------------------------------------------------------------
       DO k=1,nq2d(quadOrder)
          !
          !---- Quadrature pts
          x(1) = qx2d(k)
          y(1) = qy2d(k)
          z(1) = 0.d0
          x(2) = qx2d(k)
          y(2) = 0.d0
          z(2) = qy2d(k)
          x(3) = 0.d0
          y(3) = qx2d(k)
          z(3) = qy2d(k)
          x(4) = qx2d(k)
          y(4) = qy2d(k)
          z(4) = 1.d0 - qx2d(k) - qy2d(k)
          !
          !---- Quadrature pts on the face of the reference tetrahedron
          vRefk(1) = x(iPlan)
          vRefk(2) = y(iPlan)
          vRefk(3) = z(iPlan)
          !
          !---- Quadrature pts ont the current face
          CALL MATVECMUL(Mtx,vRefk,3,3,vFack)
          vFack(:) = vFack(:) + vTet(:,1)
          !
          CALL BASES_LAGRANGE_POLY_TET(vTet,vFack(1),vFack(2),vFack(3),orderTet,phi)
          !
          DO iDof=1,hdgEM(cellNum)%nDofS
             iDofPerm = hdgEM(cellNum)%iPermDof%ptr%tab(iDof,iPerm)
             Esca(k,1) = Esca(k,1) + phi(iDofPerm)*ExSca(iDof)
             Esca(k,2) = Esca(k,2) + phi(iDofPerm)*EySca(iDof)
             Esca(k,3) = Esca(k,3) + phi(iDofPerm)*EzSca(iDof)
             Hsca(k,1) = Hsca(k,1) + phi(iDofPerm)*HxSca(iDof)
             Hsca(k,2) = Hsca(k,2) + phi(iDofPerm)*HySca(iDof)
             Hsca(k,3) = Hsca(k,3) + phi(iDofPerm)*HzSca(iDof)
          ENDDO
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_SCATTERED_FIELD_QUAD_PTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_SURFACE_CURRENTS_QUAD_PTS(iFace)
    !------------------------------------------------------------------
    !---- Compute surface currents at quadrature points
    !
    !---- IN
    !---- iFace
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_FACES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)              :: iFace
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                          :: cellNum, cellNum1,      &
         cellNum2, faceNum, iDof, k, neighbext, numLocal

    REAL(C_DOUBLE),            DIMENSION(3) :: normal

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3) :: cmplxNormal
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3) :: EscaLoc, HscaLoc
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3) :: JsLoc, MsLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization
    !------------------------------------------------------------------
    Js(iFace)%tab(:,:) = (0.d0, 0.d0)
    Ms(iFace)%tab(:,:) = (0.d0, 0.d0)
    !------------------------------------------------------------------
    !---- Compute surface currents
    !------------------------------------------------------------------
    cellNum = RCSelt(iFace)
    IF (cellNum /= -1) THEN
       faceNum    = MeshFacTfsf(iFace)%faceNum
       cellNum1   = MeshFac(faceNum)%neighb(1)
       cellNum2   = MeshFac(faceNum)%neighb(2)
       IF (cellNum == cellNum1) neighbext = 1
       IF (cellNum == cellNum2) neighbext = 2
       numLocal       = MeshFac(faceNum)%numLocal(neighbext)
       normal(:)      = MeshCell(cellNum)%normal(:,numLocal)
       cmplxNormal(:) = normal(:)
       DO k=1,nq2d(quadOrder)
          HscaLoc(:) = Hsca(k,:)
          CALL CROSS_PROD(HscaLoc,cmplxNormal,JsLoc)
          Js(iFace)%tab(k,:) = JsLoc(:)
          !
          EscaLoc(:) = Esca(k,:)       
          CALL CROSS_PROD(cmplxNormal,EscaLoc,MsLoc)
          Ms(iFace)%tab(k,:) = MsLoc(:)
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_SURFACE_CURRENTS_QUAD_PTS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_RADIATION_VECTORS
    !------------------------------------------------------------------
    !---- Compute radiation vectors
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MATH_PHYS_CONSTANTS
    USE MATRICES_STRUCTURES
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    USE SOURCE_STRUCTURES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                              :: cellNum, cellNum1,  &
         cellNum2, faceNum, i, iFace, iPlan, iPhi, iTheta, j, k,       &
         neighbext, numLocal 

    REAL(C_DOUBLE)                              :: cosPhi, cosTheta,   &
         dJac, dPhi, dTheta, measure, Phi, sinPhi, sinTheta, surface,  &
         Theta, threshold, xk, yk, zk

    REAL(C_DOUBLE), DIMENSION(3)                :: ridge1, ridge2
    REAL(C_DOUBLE), DIMENSION(3)                :: normal     
    REAL(C_DOUBLE), DIMENSION(3)                :: vRefk, vFack
    REAL(C_DOUBLE), DIMENSION(3)                :: vRef1, vRef2, vRef3
    REAL(C_DOUBLE), DIMENSION(3)                :: vTmp1, vTmp2, vTmp3
    REAL(C_DOUBLE), DIMENSION(4)                :: x, y, z
    REAL(C_DOUBLE), DIMENSION(3,3)              :: iMtx, Mtx        
    REAL(C_DOUBLE), DIMENSION(3,3)              :: vFac
    REAL(C_DOUBLE), DIMENSION(3,4)              :: vTet
    
    COMPLEX(C_DOUBLE_COMPLEX)                   :: expk, LphiLoc,      &
         LthetaLoc, NphiLoc, NthetaLoc
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !--- Compute radiation vectors
    !------------------------------------------------------------------
    !
    !---- Initialization
    NsTheta(:,:) = (0.d0, 0.d0)
    NsPhi(:,:)   = (0.d0, 0.d0)
    LsTheta(:,:) = (0.d0, 0.d0)
    LsPhi(:,:)   = (0.d0, 0.d0)
    dTheta       = ABS(ThetaEndRCS - ThetaInitRCS)/REAL(nThetaRCS)
    dPhi         = ABS(PhiEndRCS - PhiInitRCS)/REAL(nPhiRCS)    
    !
    DO iTheta=1,nThetaRCS
       Theta    = ThetaInitRCS + REAL(iTheta-1)*dTheta
       CALL DEGREES_TO_RADIANS(Theta)
       cosTheta = COS(Theta)
       sinTheta = SIN(Theta)
       DO iPhi=1,nPhiRCS
          Phi    = PhiInitRCS + REAL(iPhi-1)*dPhi
          CALL DEGREES_TO_RADIANS(Phi)
          cosPhi = COS(Phi)
          sinPhi = SIN(Phi)
          do_Tfsf : DO iFace=1,Mesh%nFacTfsf
             cellNum    = RCSelt(iFace)
             IF (cellNum == -1) CYCLE do_Tfsf
             faceNum    = MeshFacTfsf(iFace)%faceNum
             cellNum1   = MeshFac(faceNum)%neighb(1)
             cellNum2   = MeshFac(faceNum)%neighb(2)
             IF (cellNum1 == cellNum) neighbext = 1
             IF (cellNum2 == cellNum) neighbext = 2
             numLocal   = MeshFac(faceNum)%numLocal(neighbext)
             surface    = MeshCell(cellNum)%facSurf(numLocal)
             !
             !---- Coordinates of the vertices of the tetrahedron
             vTet(:,1) = MeshVert(MeshCell(cellNum)%vertex(1))%coor(:)
             vTet(:,2) = MeshVert(MeshCell(cellNum)%vertex(2))%coor(:)
             vTet(:,3) = MeshVert(MeshCell(cellNum)%vertex(3))%coor(:)
             vTet(:,4) = MeshVert(MeshCell(cellNum)%vertex(4))%coor(:)
             !
             !
             !---- Matrix of the affine transformation of the reference 
             !---- tetrahedron to the current tetrahedron
             DO i=1,3
                DO j=1,3
                   Mtx(i,j) = vTet(i,j+1) - vTet(i,1)
                ENDDO
             ENDDO
             !
             !---- Inverse of the matrix Mtx for the transformation of the 
             !---- current tetrahedron to the reference tetrahedron
             threshold = 1.0e-15
             CALL INVERSE_3X3(Mtx,threshold,iMtx,dJac)
             !
             !---- Coordinates of the face in the reference tetrahedron
             vFac(:,1) = MeshVert(MeshFac(faceNum)%vertex(1))%coor(:)
             vFac(:,2) = MeshVert(MeshFac(faceNum)%vertex(2))%coor(:)
             vFac(:,3) = MeshVert(MeshFac(faceNum)%vertex(3))%coor(:)
             !
             vTmp1(:) = vFac(:,1) - vTet(:,1)
             vTmp2(:) = vFac(:,2) - vTet(:,1)
             vTmp3(:) = vFac(:,3) - vTet(:,1)
             CALL MATVECMUL(iMtx,vTmp1,3,3,vRef1)
             CALL MATVECMUL(iMtx,vTmp2,3,3,vRef2)
             CALL MATVECMUL(iMtx,vTmp3,3,3,vRef3)
             !
             !---- Orientation of the face on the reference tetrahedron
             ridge1(:) = vRef2(:) - vRef1(:)
             ridge2(:) = vRef3(:) - vRef1(:)
             !
             CALL CROSS_PROD(ridge1,ridge2,normal)
             normal = Tref_surf*normal
             CALL NORMALIZE_VECTOR(normal,measure)
             !
             IF ((ABS(normal(1)) < threshold) .AND. (ABS(normal(2)) < threshold)) THEN
                iPlan = 1 ! The face is in the plan <x,y>
             ELSEIF ((ABS(normal(1)) < threshold) .AND. (ABS(normal(3)) < threshold)) THEN
                iPlan = 2 ! The face is in the plan <x,z>
             ELSEIF ((ABS(normal(2)) < threshold ) .AND. (ABS(normal(3)) < threshold)) THEN
                iPlan = 3 ! The face is in the plan <y,z>
             ELSE
                iPlan = 4
             ENDIF
             !
             DO k=1,nq2d(quadOrder)
                !
                !---- Quadrature pts
                x(1) = qx2d(k)
                y(1) = qy2d(k)
                z(1) = 0.d0
                x(2) = qx2d(k)
                y(2) = 0.d0
                z(2) = qy2d(k)
                x(3) = 0.d0
                y(3) = qx2d(k)
                z(3) = qy2d(k)
                x(4) = qx2d(k)
                y(4) = qy2d(k)
                z(4) = 1.d0 - qx2d(k) - qy2d(k)
                !
                !---- Quadrature pts on the face of the reference tetrahedron
                vRefk(1) = x(iPlan)
                vRefk(2) = y(iPlan)
                vRefk(3) = z(iPlan)
                !
                !---- Quadrature pts ont the current face
                CALL MATVECMUL(Mtx,vRefk,3,3,vFack)
                vFack(:) = vFack(:) + vTet(:,1)
                !
                expk     = EXP(iunit*omega*(vFack(1)*sinTheta*cosPhi   &
                     + vFack(2)*sinTheta*sinPhi + vFack(3)*cosTheta))
                !
                NthetaLoc = (Js(iFace)%tab(k,1)*cosTheta*cosPhi        &
                     + Js(iface)%tab(k,2)*cosTheta*sinPhi              &
                     - Js(iFace)%tab(k,3)*sinTheta)*expk
                NphiLoc   = (-Js(iFace)%tab(k,1)*sinPhi                &
                     + Js(iface)%tab(k,2)*cosPhi)*expk
                LthetaLoc = (Ms(iFace)%tab(k,1)*cosTheta*cosPhi        &
                     + Ms(iFace)%tab(k,2)*cosTheta*sinPhi              &
                     - Ms(iFace)%tab(k,3)*sinTheta)*expk
                LphiLoc   = (-Ms(iFace)%tab(k,1)*sinPhi                &
                     + Ms(iFace)%tab(k,2)*cosPhi)*expk
                !
                NsTheta(iTheta,iPhi) = NsTheta(iTheta,iPhi)            &
                     + qw2d(k)*NthetaLoc*surface
                NsPhi(iTheta,iPhi)   = NsPhi(iTheta,iPhi)              &
                     + qw2d(k)*NphiLoc*surface
                LsTheta(iTheta,iPhi) = LsTheta(iTheta,iPhi)            &
                     + qw2d(k)*LthetaLoc*surface
                LsPhi(iTheta,iPhi)   = LsPhi(iTheta,iPhi)              &
                     + qw2d(k)*LphiLoc*surface
             ENDDO
          ENDDO do_Tfsf
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Gather
    !------------------------------------------------------------------
    NsThetaGlb(:,:) = (0.d0, 0.d0)
    NsPhiGlb(:,:)   = (0.d0, 0.d0)
    LsThetaGlb(:,:) = (0.d0, 0.d0)
    LsPhiGlb(:,:)   = (0.d0, 0.d0)
    CALL MPI_ALLREDUCE(NsTheta(:,:), NsThetaGlb(:,:), nThetaRCS*nPhiRCS, &
         MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError) 
    CALL MPI_ALLREDUCE(NsPhi(:,:),   NsPhiGlb(:,:),   nThetaRCS*nPhiRCS, &
         MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError) 
    CALL MPI_ALLREDUCE(LsTheta(:,:), LsThetaGlb(:,:), nThetaRCS*nPhiRCS, &
         MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError) 
    CALL MPI_ALLREDUCE(LsPhi(:,:),   LsPhiGlb(:,:),   nThetaRCS*nPhiRCS, &
         MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RADIATION_VECTORS

  ! !====================================================================
  ! !====================================================================
  ! !====================================================================

  ! SUBROUTINE COMPUTE_INCIDENT_POWER_DENSITY
  !   !------------------------------------------------------------------
  !   !---- Compute incident power density: Sinc
  !   !
  !   !---- IN
  !   !
  !   !---- OUT
  !   !
  !   !------------------------------------------------------------------
  !   !---- Modules
  !   !------------------------------------------------------------------
  !   USE MATH_PHYS_CONSTANTS
  !   USE MESH_CELLS
  !   USE MESH_DATA
  !   USE MESH_FACES
  !   USE NUMERICAL_TYPES
  !   USE PARALLEL_STRUCTURES
  !   USE SOURCE_STRUCTURES
  !   USE SPECIAL_FUNCTIONS
  !   !
  !   USE PARALLEL_STRUCTURES
  !   !------------------------------------------------------------------
  !   !---- Local variables
  !   !------------------------------------------------------------------
  !   INTEGER(C_INT)                                   :: cellNum,       &
  !        faceNum, iDof, iFac, nDofS, nDofV, numLoc, order, cellNum1, cellNum2, neighbext

  !   REAL(C_DOUBLE)                                   :: poyntingFlux,  &
  !        SincLoc, surface
  !   REAL(C_DOUBLE),            DIMENSION(3)          :: normal
  !   REAL(C_DOUBLE),            DIMENSION(nDofSMax)   :: poyDotN, poyFlux
  !   REAL(C_DOUBLE),            DIMENSION(nDofSMax,3) :: poynting

  !   COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(3)          :: crossProd,     &
  !        EincLoc, HincLoc
  !   !------------------------------------------------------------------
  !   !---- Testing for bad input
  !   !------------------------------------------------------------------
  !   !
  !   !------------------------------------------------------------------
  !   SincLoc = 0.d0
  !   do_Sinc : DO iFac=1,Mesh%nFacTfsf
  !      cellNum = RCSelt(iFac)
  !      IF (cellNum == -1) CYCLE do_Sinc
  !      faceNum   = MeshFacTfsf(iFac)%faceNum
  !      cellNum1  = MeshFac(faceNum)%neighb(1)
  !      cellNum2  = MeshFac(faceNum)%neighb(2)
  !      IF (cellNum1 == cellNum) neighbext = 1
  !      IF (cellNum2 == cellNum) neighbext = 2
  !      numLoc    = MeshFac(faceNum)%numLocal(neighbext)
  !      normal(:) = MeshCell(cellNum)%normal(:,numLoc)
  !      surface   = MeshCell(cellNum)%facSurf(numLoc)
  !      nDofS     = hdgEM(cellNum)%nDofS
  !      nDofV     = hdgEM(cellNum)%nDofV
  !      order     = hdgEM(cellNum)%order
  !      !
  !      !---- Compute poynting vector
  !      CALL FACE_SOL(faceNum, nDofS, nDofV, order, omega)
  !      DO iDof=1,nDofS
  !         EincLoc(1)       =       ExScx(iDof)
  !         EincLoc(2)       =       EyScx(iDof)
  !         EincLoc(3)       =       EzScx(iDof)
  !         HincLoc(1)       = CONJG(HxScx(iDof))/z0
  !         HincLoc(2)       = CONJG(HyScx(iDof))/z0
  !         HincLoc(3)       = CONJG(HzScx(iDof))/z0
  !         CALL CROSS_PROD(EincLoc,HincLoc,crossProd)
  !         poynting(iDof,:) = 0.5*REAL(crossProd(:))
  !         poyDotN(iDof)    = DOT_PRODUCT(normal(:),poynting(iDof,:))
  !      ENDDO
  !      !
  !      CALL MATVECMUL(hdgEM(cellNum)%srfMtxE(numLoc)%ptr%tab,              &
  !           poyDotN(1:nDofS),nDofS,nDofS,poyFlux(1:nDofS))
  !      SincLoc             = SincLoc + SUM(poyFlux)*surface
  !   ENDDO do_Sinc
  !   SincLoc = ABS(SincLoc)
  !   !
  !   CALL MPI_ALLREDUCE(SincLoc, Sinc, 1, MPI_DOUBLE_PRECISION, MPI_SUM, &
  !        defaultComm, mpiError)
  !   !------------------------------------------------------------------
  !   !---- Testing for errors
  !   !------------------------------------------------------------------
  !   !
  !   !------------------------------------------------------------------
  ! END SUBROUTINE COMPUTE_INCIDENT_POWER_DENSITY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_RCS
    !------------------------------------------------------------------
    !---- Compute RCS
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MATH_PHYS_CONSTANTS
    USE PARALLEL_STRUCTURES
    USE SOURCE_STRUCTURES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iPhi, iTheta, nUnit

    REAL(C_DOUBLE) :: coeff, dPhi, dTheta, Phi, Theta
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IF (idProc == 0) THEN
       ALLOCATE(RCS(nThetaRCS,nPhiRCS))
       coeff = (omega**2)/(16.d0*Pi)
       DO iTheta=1,nThetaRCS
          DO iPhi=1,nPhiRCS
             RCS(iTheta,iPhi) = 10.d0*LOG10(coeff                                  &
                  *(REAL(ABS(LsPhiGlb(iTheta,iPhi) + NsThetaGlb(iTheta,iPhi)))**2  &
                  + REAL(ABS(LsThetaGlb(iTheta,iPhi) - NsPhiGlb(iTheta,iPhi)))**2))
          ENDDO
       ENDDO
    ENDIF
    !
    IF (idProc == 0) THEN
       dTheta = ABS(ThetaEndRCS - ThetaInitRCS)/REAL(nThetaRCS)
       dPhi   = ABS(PhiEndRCS - PhiInitRCS)/REAL(nPhiRCS)    
       nUnit  = 1
       OPEN(UNIT=nUnit, FILE="RCS", STATUS="UNKNOWN")
       DO iTheta=1,nThetaRCS
          Theta = ThetaInitRCS + REAL(iTheta-1)*dTheta
          DO iPhi=1,nPhiRCS
             Phi = PhiInitRCS + REAL(iPhi-1)*dPhi
             WRITE(nUnit,'(3(e15.6,1x))') Theta, Phi, RCS(iTheta,iPhi)
          ENDDO
       ENDDO
       CLOSE(nUnit)
       DEALLOCATE(RCS)
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RCS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE RCS_DEALLOCATIONS
    !------------------------------------------------------------------
    !---- Deallocations of arrays for RCS computation
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iFace
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Dynamic allocations
    !------------------------------------------------------------------
    !
    !---- Array for exterior elements to the Tfsf surface
    DEALLOCATE(RCSelt)
    !
    !---- Scattered field
    DEALLOCATE(ExSca, EySca, EzSca)
    DEALLOCATE(HxSca, HySca, HzSca)
    !
    DEALLOCATE(Esca)
    DEALLOCATE(Hsca)
    !
    !---- Surface currents
    DO iFace=1,Mesh%nFacTfsf
       DEALLOCATE(Js(iFace)%tab)
       DEALLOCATE(Ms(iFace)%tab)
    ENDDO
    DEALLOCATE(Js, Ms)
    !
    !---- Radiation vectors
    DEALLOCATE(NsTheta,    NsPhi)
    DEALLOCATE(LsTheta,    LsPhi)
    DEALLOCATE(NsThetaGlb, NsPhiGlb)
    DEALLOCATE(LsThetaGlb, LsPhiGlb)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE RCS_DEALLOCATIONS

  !====================================================================
  !====================================================================
  !====================================================================
  
END MODULE RCS_STRUCTURES
