MODULE SCHWARZ_CORE

  !====================================================================
  ! Main hdgm combined with a Schwarz algorithm routines
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Values of Lambda on interfaces at iteration "n" and "n+1" in 
  ! Schwarz algorithm, usefull to compute residual in Schwarz algorithm
  !--------------------------------------------------------------------
  COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: LambdaIt
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Residual of Schwarz algorithm
  !--------------------------------------------------------------------
  REAL(C_DOUBLE) :: Res0, Residual
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_SCHWARZ_ALGORITHM
    !------------------------------------------------------------------
    !---- Compute Schwarz algorithm
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_CORE
    USE INPUT_DATA
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE SCHWARZ_MUMPS_STRUCTURES
    USE SCHWARZ_STRUCTURES
    USE TIMERS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)            :: facNum, iCell, iFac, iter

    REAL(C_DOUBLE)            :: precision

    TYPE(Timer)               :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of Schwarz data structures and Lambda field on 
    !---- interfaces
    !------------------------------------------------------------------
    CALL SCHWARZ_DATA_STRUCTURES_INIT
    CALL SCHWARZ_LAMBDA_INTERFACE_INIT
    !------------------------------------------------------------------
    !---- Initialization of MUMPS control parameters
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_INIT
    !------------------------------------------------------------------
    !---- Compute local index of faces for assembling matrix and RHS in 
    !---- MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_FACES_INDICES
    !------------------------------------------------------------------
    !---- Assembling matrix in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- Analysis of the matrix in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_ANALYSIS
    !------------------------------------------------------------------
    !---- Factorization of the matrix in MUMPS package
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_FACTORIZATION
    !------------------------------------------------------------------
    !---- Schwarz algorithm
    !------------------------------------------------------------------
    !                                                                        
    !---- Stopping criterion threshold in the iterative solver
    SELECT CASE(thresholdIter)
    CASE("max")
       precision      = (globalMesh%hMax)**(genOrderMax + 2)
    CASE("avg")
       precision      = (globalMesh%hAvg)**(genOrderMax + 2)
    CASE("min")
       precision      = (globalMesh%hMin)**(genOrderMax + 2)
    END SELECT
    IF (precision < 1.0e-15) precision = 1.0e-15
    !
    !---- Iterations
    CALL myTimer%Start()
    do_iter: DO iter=1,maxIter
       !
       !---- Save Lambda DoFs on interfaces at iteration "n"
       CALL SCHWARZ_LAMBDA_INTERFACE(1)
       !
       !---- Exchange EM field at interfaces
       CALL SCHWARZ_EM_EXCHANGE
       !
       !---- Update RHS
       CALL SCHWARZ_RHS_UPDATE
       CALL SCHWARZ_MUMPS_RHS_UPDATE
       !
       !---- Solve the local linear system
       CALL SCHWARZ_MUMPS_SOLVE
       !
       !---- Save Lambda DoFs on interfaces at iteration "n+1"
       CALL SCHWARZ_LAMBDA_INTERFACE(2)
       !
       !---- Compute residual
       CALL SCHWARZ_COMPUTE_RESIDUAL(iter)
       !
       !---- Print some informations
       IF (idWrite)  THEN
          IF (iter == 1) THEN
             WRITE(6, 10)
             WRITE(6, '(a)') " Schwarz algorithm         "
             WRITE(6, 10)
             WRITE(6, '(a)') " Iteration        Residual "
          ENDIF
          WRITE(6, '(i10,a5,e15.6)') iter, "  ", Residual
       ENDIF
       !
       !---- Testing if the precision is achieved
       IF (Residual < precision) EXIT do_iter
       !
       !---- Compute EM field for elements at the interfaces between 
       !---- neighboring sub-meshes
       DO iFac=1,Mesh%nFacArt
          facNum = MeshFacArt(iFac)%faceNum
          iCell  = MeshFac(facNum)%neighb(1)
          CALL LOCAL_SOLVER(iCell)
       ENDDO
    ENDDO do_iter
    !
    !---- Deallocate all data structures used in Schwarz algoritm
    CALL SCHWARZ_ALGORITHM_END
    !------------------------------------------------------------------
    !---- Print informations
    !------------------------------------------------------------------
    IF (idWrite) THEN
       WRITE(6, '(a)')            " "
       WRITE(6, '(a)')            " Solution quality for Lambda            "
       WRITE(6, '(a,e15.6)')      "                     Wanted precision : ", precision
       WRITE(6, '(a,e15.6)')      "                   Effective residual : ", Residual
       WRITE(6, '(a,i15)')        "                    Max. nb. of iters : ", maxIter
       WRITE(6, '(a,i15)')        "                         Nb. of iters : ", iter
       WRITE(6, '(a)')            " "
       WRITE(6, '(a)')            " Time spent in iterative step (seconds) "
       WRITE(6, '(a,e15.6)')      "                             CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e15.6)')      "                            Wall time : ", myTimer%SYSTEM_TIME()
       WRITE(6, 10)
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_SCHWARZ_ALGORITHM

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_EM_EXCHANGE
    !------------------------------------------------------------------
    !---- Exchange EM field DoFs on interfaces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    USE SCHWARZ_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: cellNum, faceNum, faceNumArt, id, iFac, ik,      &
         iNeighb, iPerm, k, l, nDoFs 
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Exchange EM field on interfaces
    !------------------------------------------------------------------
    DO iNeighb=1,nbNeighb
       id = 1
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum    = neighb(iNeighb)%faces(iFac)
          cellNum    = MeshFac(faceNum)%neighb(1)
          nDoFs      = hdgEM(cellNum)%nDofS
          l          = MeshFac(faceNum)%numLocal(1)
          iPerm      = MeshCell(cellNum)%perm(1,l)
          faceNumArt = indArt(faceNum)
          DO k=1,nDoFs
             ik             = hdgEM(cellNum)%iPermDof%ptr%tab(k,iPerm)
             EMsend(id)     = hdgEM(cellNum)%E%vec(ik,1)
             EMsend(id + 1) = hdgEM(cellNum)%E%vec(ik,2)
             EMsend(id + 2) = hdgEM(cellNum)%E%vec(ik,3)
             EMsend(id + 3) = hdgEM(cellNum)%H%vec(ik,1)
             EMsend(id + 4) = hdgEM(cellNum)%H%vec(ik,2)
             EMsend(id + 5) = hdgEM(cellNum)%H%vec(ik,3)
             id             = id + 6
          ENDDO  
       ENDDO
       !
       CALL MPI_SENDRECV(EMsend, 6*nDoFsComm(iNeighb), MPI_DOUBLE_COMPLEX, &
            neighb(iNeighb)%id, 1, EMrecv, 6*nDoFsComm(iNeighb),           &
            MPI_DOUBLE_COMPLEX, neighb(iNeighb)%id, 1, defaultComm,        &
            status, mpiError)
       !
       id = 1
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum    = neighb(iNeighb)%faces(iFac)
          faceNumArt = indArt(faceNum)
          DO k=1,EMex(faceNumArt)%nDoFs
             EMex(faceNumArt)%E%vec(k,1) = EMrecv(id)
             EMex(faceNumArt)%E%vec(k,2) = EMrecv(id + 1)
             EMex(faceNumArt)%E%vec(k,3) = EMrecv(id + 2)
             EMex(faceNumArt)%H%vec(k,1) = EMrecv(id + 3)
             EMex(faceNumArt)%H%vec(k,2) = EMrecv(id + 4)
             EMex(faceNumArt)%H%vec(k,3) = EMrecv(id + 5)
             id                          = id + 6
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_EM_EXCHANGE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_RHS_UPDATE
    !------------------------------------------------------------------
    !---- Update RHS on interfaces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE SCHWARZ_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, faceNumArt, i, iCell, iFac, j, k, nDofF, nDofS

    REAL(C_DOUBLE) :: aera

    COMPLEX(C_DOUBLE_COMPLEX) :: tmp_usum, tmp_wsum, zero
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: tmp_u, tmp_w
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Updating RHS on interfaces
    !------------------------------------------------------------------
    zero =(0.d0, 0.d0)
    !
    DO iFac=1,Mesh%nFacArt
       faceNum    = MeshFacArt(iFac)%faceNum
       aera       = MeshFac(faceNum)%measure
       nDofF      = Lambda(faceNum)%nDof
       iCell      = MeshFac(faceNum)%neighb(1)
       k          = MeshFac(faceNum)%numLocal(1)
       faceNumArt = indArt(faceNum)
       nDofS      = EMex(faceNumArt)%nDoFs
       ALLOCATE(tmp_u(nDofS), tmp_w(nDofS))
       DO i=1,nDofS
          tmp_u(i) = -(MeshCell(iCell)%normal(3,k)*MeshFac(faceNum)%axis(2,1) &
               - MeshCell(iCell)%normal(2,k)*MeshFac(faceNum)%axis(3,1))      &
               * EMex(faceNumArt)%E%vec(i,1)*aera                             &
               -(MeshCell(iCell)%normal(1,k)*MeshFac(faceNum)%axis(3,1)       &
               - MeshCell(iCell)%normal(3,k)*MeshFac(faceNum)%axis(1,1))      &
               * EMex(faceNumArt)%E%vec(i,2)*aera                             &
               -(MeshCell(iCell)%normal(2,k)*MeshFac(faceNum)%axis(1,1)       &
               - MeshCell(iCell)%normal(1,k)*MeshFac(faceNum)%axis(2,1))      &
               * EMex(faceNumArt)%E%vec(i,3)*aera                             &
               -(MeshFac(faceNum)%axis(1,1)*EMex(faceNumArt)%H%vec(i,1)       &
               + MeshFac(faceNum)%axis(2,1)*EMex(faceNumArt)%H%vec(i,2)       &
               + MeshFac(faceNum)%axis(3,1)*EMex(faceNumArt)%H%vec(i,3))      &
               * aera*hdgEM(iCell)%z
          tmp_w(i) = -(MeshCell(iCell)%normal(3,k)*MeshFac(faceNum)%axis(2,2) &
               - MeshCell(iCell)%normal(2,k)*MeshFac(faceNum)%axis(3,2))      &
               * EMex(faceNumArt)%E%vec(i,1)*aera                             &
               -(MeshCell(iCell)%normal(1,k)*MeshFac(faceNum)%axis(3,2)       &
               - MeshCell(iCell)%normal(3,k)*MeshFac(faceNum)%axis(1,2))      &
               * EMex(faceNumArt)%E%vec(i,2)*aera                             &
               -(MeshCell(iCell)%normal(2,k)*MeshFac(faceNum)%axis(1,2)       &
               - MeshCell(iCell)%normal(1,k)*MeshFac(faceNum)%axis(2,2))      &
               * EMex(faceNumArt)%E%vec(i,3)*aera                             &
               -(MeshFac(faceNum)%axis(1,2)*EMex(faceNumArt)%H%vec(i,1)       &
               + MeshFac(faceNum)%axis(2,2)*EMex(faceNumArt)%H%vec(i,2)       &
               + MeshFac(faceNum)%axis(3,2)*EMex(faceNumArt)%H%vec(i,3))      &
               * aera*hdgEM(iCell)%z
       ENDDO
       DO i=1,nDofF
          tmp_usum = zero
          tmp_wsum = zero
          DO j=1,nDofS
             tmp_usum = tmp_usum + EMex(faceNumArt)%srfMtx%ptr%tab(j,i)*tmp_u(j)
             tmp_wsum = tmp_wsum + EMex(faceNumArt)%srfMtx%ptr%tab(j,i)*tmp_w(j)
          ENDDO
          rhsSchArt(faceNumArt)%vec(i,1) = tmp_usum
          rhsSchArt(faceNumArt)%vec(i,2) = tmp_wsum
       ENDDO
       DEALLOCATE(tmp_u, tmp_w)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_RHS_UPDATE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_LAMBDA_INTERFACE_INIT
    !------------------------------------------------------------------
    !---- Initialization of LambdaIt on interfaces
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE SCHWARZ_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, iFac

    COMPLEX(C_DOUBLE_COMPLEX) :: zero
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of LambdaIt
    !------------------------------------------------------------------
    zero = (0.d0, 0.d0)
    ALLOCATE(LambdaIt(nDoFsInterface,2))
    LambdaIt(:,:) = zero
    DO iFac=1,Mesh%nFacArt
       faceNum                    = MeshFacArt(iFac)%faceNum
       Lambda(faceNum)%L%vec(:,:) = zero
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_LAMBDA_INTERFACE_INIT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_LAMBDA_INTERFACE(niter)
    !------------------------------------------------------------------
    !---- Save Lambda values on interfaces
    !
    !---- IN
    !     niter : available values are 1 or 2
    !             niter = 1  ==> Interface DoFs of Lambda at iteration "n"
    !             niter = 2  ==> Interface DoFs of Lambda at iteration "n+1"
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: niter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, i, iFac, k, nDof
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    IF ((niter /= 1) .AND. (niter /= 2)) THEN
       IF (idWrite) THEN
          WRITE(6, '(a)') " Bad input in Lambda interface routine for Schwarz algorithm     "
          WRITE(6, '(a)') " Available values of niter are 1 or 2                            "
          WRITE(6, '(a)') " niter = 1 => save DoFs of Lambda on interfaces at iteration n   "
          WRITE(6, '(a)') " niter = 2 => save DoFs of Lambda on interfaces at iteration n+1 "
       ENDIF
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Save Lambda values on interfaces
    !------------------------------------------------------------------
    k = 0
    DO iFac=1,Mesh%nFacArt
       faceNum = MeshFacArt(iFac)%faceNum
       nDof    = Lambda(faceNum)%nDof
       DO i=1,nDof
          LambdaIt(k+i,niter)      = Lambda(faceNum)%L%vec(i,1)
          LambdaIt(k+i+nDof,niter) = Lambda(faceNum)%L%vec(i,2)
       ENDDO
       k = k + 2*nDof
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_LAMBDA_INTERFACE   

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_COMPUTE_RESIDUAL(niter)
    !------------------------------------------------------------------
    !---- Compute residual
    !
    !---- IN
    !     niter : iteration number
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    USE SCHWARZ_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: niter
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: ResLoc, ResMax
    !------------------------------------------------------------------
    !---- Compute local residual
    !------------------------------------------------------------------
    ResLoc = 0.d0
    DO i=1,nDoFsInterface
       ResLoc = MAX(ResLoc, ABS(LambdaIt(i,2) - LambdaIt(i,1)))
    ENDDO
    !------------------------------------------------------------------
    !---- Compute maximal residual
    !------------------------------------------------------------------
    CALL MPI_ALLREDUCE(ResLoc, ResMax, 1, MPI_DOUBLE_PRECISION, MPI_MAX, &
         defaultComm, mpiError) 
    IF (niter == 1) Res0 = ResMax
    !------------------------------------------------------------------
    !---- Compute residual for iteration "niter"
    !------------------------------------------------------------------
    Residual = ResMax / Res0
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_COMPUTE_RESIDUAL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_ALGORITHM_END
    !------------------------------------------------------------------
    !---- Deallocate all data structures specific to Schwarz algorithm
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE SCHWARZ_MUMPS_STRUCTURES
    USE SCHWARZ_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Destroy all MUMPS instances
    !------------------------------------------------------------------
    CALL SCHWARZ_MUMPS_END
    !------------------------------------------------------------------
    !---- Deallocate all data strctures defined in SCHWARZ_structures 
    !---- module
    !------------------------------------------------------------------
    CALL SCHWARZ_DATA_STRUCTURES_END
    !------------------------------------------------------------------
    !---- Deallocate Lambda field on interfaces (defined in this module)
    !------------------------------------------------------------------
    DEALLOCATE(LambdaIt)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_ALGORITHM_END

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE SCHWARZ_CORE
