MODULE SCHWARZ_MUMPS_STRUCTURES

  !====================================================================
  ! Everything for using MUMPS in hdgm combined with a Schwarz algorithm
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  ! MUMPS structure
  !--------------------------------------------------------------------
  INCLUDE 'zmumps_struc.h'
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! MUMPS data structure
  !--------------------------------------------------------------------
  TYPE(ZMUMPS_STRUC) mumps_par
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Array for index of faces in MUMPS data structure
  !--------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: indLocMtx
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_INIT
    !------------------------------------------------------------------
    !---- Initialization of MUMPS control parameters and data structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iCell1, iCell2, iFac, j, nDof, facNum,        &
         nNeighb1, nNeighb2, nnz, ordMtx, sum1, sum2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Order of the matrix
    !------------------------------------------------------------------
    ordMtx = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum = MeshFacI(iFac)%faceNum
       ordMtx = ordMtx + 2*Lambda(facNum)%nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum = MeshFacPhy(iFac)%faceNum
       ordMtx = ordMtx + 2*Lambda(facNum)%nDof
    ENDDO
    !
    !---- Artificial boundary faces
    DO iFac=1,Mesh%nFacArt
       facNum = MeshFacArt(iFac)%faceNum
       ordMtx = ordMtx + 2*Lambda(facNum)%nDof
    ENDDO
    !------------------------------------------------------------------
    !---- Number of non-zero entries in the matrix
    !------------------------------------------------------------------
    nnz    = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum   = MeshFacI(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1) - nDof
       !
       iCell2   = MeshFac(facNum)%neighb(2)
       nNeighb2 = MeshCell(iCell2)%nNeighb
       sum2     = hdgEM(iCell2)%sumDofF(nNeighb2) - nDof
       !
       nnz      = nnz + 4*nDof*(nDof + sum1 + sum2)
    ENDDO
    !
    !---- Physical and artificial boundary faces
    DO iFac=1,Mesh%nFacB
       facNum   = MeshFacB(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1)
       !
       nnz      = nnz + 4*nDof*sum1
    ENDDO
    !------------------------------------------------------------------
    !---- Sets up MUMPS control parameters
    !------------------------------------------------------------------
    !
    !---- Definition of the MPI communicator for MUMPS
    mumps_par%COMM = MPI_COMM_SELF
    !
    !---- Non-symmetric matrix
    mumps_par%SYM = 0
    !
    mumps_par%PAR = 1
    !
    !---- Initialization of one instance of the package
    mumps_par%JOB = -1
    CALL ZMUMPS(mumps_par)
    !
    !---- Standard output
    mumps_par%ICNTL(1)  = -1
    mumps_par%ICNTL(2)  = -1
    mumps_par%ICNTL(3)  = -1
    mumps_par%ICNTL(4)  = -1
    !
    !---- Definition of the matrix, distributed assembled format
    mumps_par%ICNTL(5)  =  0
    mumps_par%ICNTL(18) =  0
    !
    !---- Pre-treatment of the matrix (permutation, ordering)
    mumps_par%ICNTL(6)  =  7
    !
    !---- Package used for ordering
    ! mumps_par%ICNTL(7)  = 3  ! SCOTCH
    ! mumps_par%ICNTL(7)  = 4  ! PORD
    ! mumps_par%ICNTL(7)  = 5  ! MeTiS
    ! mumps_par%ICNTL(7)  = 6  ! QAMD
    mumps_par%ICNTL(7)  =  7  ! Automatic choice
    !
    mumps_par%ICNTL(8)  = 77
    ! !
    ! !----Increase MAXIS (cf. doc MUMPS)
    ! mumps_par%ICNTL(14)  = 30
    !------------------------------------------------------------------
    !---- Characterization of th problem
    !------------------------------------------------------------------
    mumps_par%N  = ordMtx
    mumps_par%NZ = nnz
    !------------------------------------------------------------------
    !---- Allocations of data structure for the definition of the RHS 
    !---- and the matrix
    !------------------------------------------------------------------
    !
    !---- RHS
    ALLOCATE(mumps_par%RHS(mumps_par%N))
    !
    !---- Local matrix A
    ALLOCATE(mumps_par%IRN(mumps_par%NZ))
    ALLOCATE(mumps_par%JCN(mumps_par%NZ))
    ALLOCATE(mumps_par%A  (mumps_par%NZ))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_INIT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_FACES_INDICES
    !------------------------------------------------------------------ 
    !---- Compute local index of faces for assembling matrix and RHS 
    !---- in MUMPS package
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, iFac, preFac, preDof
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of indMtxFac for MUMPS package
    !
    !---- We first consider pure inner faces and physical boundary faces 
    !---- and finally artificial boundary faces in indMtxFac
    !
    !---- Variables in different loops:
    !     - facNum is the index of the current face in MeshFac structure
    !     - preFac is the index of the previous face in MeshFac structure
    !     - preDof is the number of DoFs of the previous face i.e. the 
    !       number DoFs of the face with index 'preFac' in MeshFac 
    !       structure
    !------------------------------------------------------------------
    !---- Dynamic allocation of indLocMtx
    !------------------------------------------------------------------
    ALLOCATE(indLocMtx(Mesh%nFac))
    !------------------------------------------------------------------
    !---- Pure inner faces, physical boundary faces and interfaces
    !------------------------------------------------------------------
    !
    !---- Inner faces
    indLocMtx(MeshFacI(1)%faceNum) = 0
    preFac = MeshFacI(1)%faceNum
    preDof = Lambda(preFac)%nDof
    DO iFac=2,Mesh%nFacI
       facNum            = MeshFacI(iFac)%faceNum
       indLocMtx(facNum) = indLocMtx(preFac) + 2*preDof
       !
       preFac = facNum
       preDof = Lambda(facNum)%nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum            = MeshFacPhy(iFac)%faceNum
       indLocMtx(facNum) = indLocMtx(preFac) + 2*preDof
       !
       preFac = facNum
       preDof = Lambda(facNum)%nDof
    ENDDO
    !
    !---- Interfaces i.e. artificial boundary faces
    DO iFac=1,Mesh%nFacArt
       facNum            = MeshFacArt(iFac)%faceNum
       indLocMtx(facNum) = indLocMtx(preFac) + 2*preDof
       !
       preFac = facNum
       preDof = Lambda(facNum)%nDof
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_FACES_INDICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- Assembling the matrix 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_MATRICES_STRUCTURES
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iCell, iCell1, iCell11, iCell12, iCell2,      &
         iCell21, iCell22, iFac, indCol1, indCol2, indLoc, indRow1,    &
         indRow2, j, jCell11, jCell12, jCell21, jCell22, k, k1, k2,    &
         kFac, nDof, nDofFk, facNum, nNeighb1, nNeighb2, sum1, sum2, sumk, iError

    COMPLEX(C_DOUBLE_COMPLEX) :: valMtx
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Assembling the matrix 
    !------------------------------------------------------------------
    indLoc = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum   = MeshFacI(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       k1       = MeshFac(facNum)%numLocal(1)
       sum1     = hdgEM(iCell1)%sumDofF(k1) - nDof
       !
       iCell2   = MeshFac(facNum)%neighb(2)
       nNeighb2 = MeshCell(iCell2)%nNeighb
       k2       = MeshFac(facNum)%numLocal(2)
       sum2     = hdgEM(iCell2)%sumDofF(k2) - nDof
       !
       DO i=1,nDof
          indRow1 = indLocMtx(facNum) + i
          indRow2 = indRow1 + nDof
          iCell11 = 2*sum1  + i
          iCell12 = iCell11 + nDof
          iCell21 = 2*sum2  + i
          iCell22 = iCell21 + nDof
          !
          !---- The diagonal blocks of Ke(iCell1) and Ke(iCell2)
          DO j=1,nDof
             indCol1 = indLocMtx(facNum) + j
             indCol2 = indCol1 + nDof
             jCell11 = 2*sum1  + j
             jCell12 = jCell11 + nDof
             jCell21 = 2*sum2  + j
             jCell22 = jCell21 + nDof
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell11,jCell11) + Ke(iCell2)%tab(iCell21,jCell21)
             mumps_par%IRN(indLoc) = indRow1
             mumps_par%JCN(indLoc) = indCol1
             mumps_par%A  (indLoc) = valMtx  
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell11,jCell12) + Ke(iCell2)%tab(iCell21,jCell22)
             mumps_par%IRN(indLoc) = indRow1
             mumps_par%JCN(indLoc) = indCol2
             mumps_par%A  (indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell12,jCell11) + Ke(iCell2)%tab(iCell22,jCell21)
             mumps_par%IRN(indLoc) = indRow2
             mumps_par%JCN(indLoc) = indCol1
             mumps_par%A  (indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell12,jCell12) + Ke(iCell2)%tab(iCell22,jCell22)
             mumps_par%IRN(indLoc) = indRow2
             mumps_par%JCN(indLoc) = indCol2
             mumps_par%A  (indLoc) = valMtx
          ENDDO
          !
          !---- The off-diagonal blocks of Ke(iCell1)
          DO k=1,nNeighb1
             IF (k /= k1) THEN
                kFac = MeshCell(iCell1)%facNum(k)
                nDofFk  = hdgEM(iCell1)%nDofF(k)
                sumk    = hdgEM(iCell1)%sumDofF(k) - nDofFk
                DO j=1,nDofFk
                   indCol1 = indLocMtx(kFac) + j
                   indCol2 = indCol1 + nDofFk
                   jCell11 = 2*sumk  + j
                   jCell12 = jCell11 + nDofFk
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell11,jCell11)
                   mumps_par%IRN(indLoc) = indRow1
                   mumps_par%JCN(indLoc) = indCol1
                   mumps_par%A  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell11,jCell12)
                   mumps_par%IRN(indLoc) = indRow1
                   mumps_par%JCN(indLoc) = indCol2
                   mumps_par%A  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell12,jCell11)
                   mumps_par%IRN(indLoc) = indRow2
                   mumps_par%JCN(indLoc) = indCol1
                   mumps_par%A  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell12,jCell12)
                   mumps_par%IRN(indLoc) = indRow2
                   mumps_par%JCN(indLoc) = indCol2
                   mumps_par%A  (indLoc) = valMtx
                ENDDO
             ENDIF
          ENDDO
          !
          !---- The off-diagonal blocks of Ke(iCell2)
          DO k=1,nNeighb2
             IF (k /= k2) THEN
                kFac = MeshCell(iCell2)%facNum(k)
                nDofFk  = hdgEM(iCell2)%nDofF(k)
                sumk    = hdgEM(iCell2)%sumDofF(k) - nDofFk
                DO j=1,nDofFk
                   indCol1 = indLocMtx(kFac) + j
                   indCol2 = indCol1 + nDofFk
                   jCell21 = 2*sumk  + j
                   jCell22 = jCell21 + nDofFk
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell21,jCell21)
                   mumps_par%IRN(indLoc) = indRow1
                   mumps_par%JCN(indLoc) = indCol1
                   mumps_par%A  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell21,jCell22)
                   mumps_par%IRN(indLoc) = indRow1
                   mumps_par%JCN(indLoc) = indCol2
                   mumps_par%A  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell22,jCell21)
                   mumps_par%IRN(indLoc) = indRow2
                   mumps_par%JCN(indLoc) = indCol1
                   mumps_par%A  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell22,jCell22)
                   mumps_par%IRN(indLoc) = indRow2
                   mumps_par%JCN(indLoc) = indCol2
                   mumps_par%A  (indLoc) = valMtx
                ENDDO
             ENDIF
          ENDDO
       ENDDO
    ENDDO
    !
    !---- Boundary faces
    DO iFac=1,Mesh%nFacB
       facNum   = MeshFacB(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       k1       = MeshFac(facNum)%numLocal(1)
       sum1     = hdgEM(iCell1)%sumDofF(k1) - nDof
       !
       DO i=1,nDof
          indRow1 = indLocMtx(facNum) + i
          indRow2 = indRow1 + nDof
          iCell11 = 2*sum1  + i
          iCell12 = iCell11 + nDof
          DO k=1,nNeighb1
             kFac = MeshCell(iCell1)%facNum(k)
             nDofFk  = hdgEM(iCell1)%nDofF(k)
             sumk    = hdgEM(iCell1)%sumDofF(k) - nDofFk
             DO j=1,nDofFk
                indCol1 = indLocMtx(kFac) + j
                indCol2 = indCol1 + nDofFk
                jCell11 = 2*sumk  + j
                jCell12 = jCell11 + nDofFk
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell11,jCell11)
                mumps_par%IRN(indLoc) = indRow1
                mumps_par%JCN(indLoc) = indCol1
                mumps_par%A  (indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell11,jCell12)
                mumps_par%IRN(indLoc) = indRow1
                mumps_par%JCN(indLoc) = indCol2
                mumps_par%A  (indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell12,jCell11)
                mumps_par%IRN(indLoc) = indRow2
                mumps_par%JCN(indLoc) = indCol1
                mumps_par%A  (indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell12,jCell12)
                mumps_par%IRN(indLoc) = indRow2
                mumps_par%JCN(indLoc) = indCol2
                mumps_par%A  (indLoc) = valMtx
             ENDDO
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Deallocate Ke 
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       DEALLOCATE(Ke(iCell)%tab)
    ENDDO
    DEALLOCATE(Ke)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_ASSEMBLING_MATRIX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_ANALYSIS
    !------------------------------------------------------------------
    !---- MUMPS analysis step 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    USE TIMERS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError 

    TYPE(Timer)    :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    
    !------------------------------------------------------------------
    !---- Analysis
    !------------------------------------------------------------------
    mumps_par%JOB = 1
    CALL myTimer%Start()
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------    
    iError = mumps_par%INFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in analysis step of MUMPS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Print MUMPS informations
    !------------------------------------------------------------------
    IF(idWrite) THEN
       WRITE(6, 10)
       WRITE(6, '(a)')       " MUMPS statistics                           "
       WRITE(6, 10)
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       " ** Analysis step statistics                "
       WRITE(6, '(a)')       "           Time spent on analysis (seconds) "
       WRITE(6, '(a,e12.5)') "                                 CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e12.5)') "                                Wall time : ", myTimer%SYSTEM_TIME()
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_ANALYSIS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_FACTORIZATION
    !------------------------------------------------------------------
    !---- MUMPS factorization step 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    USE TIMERS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError, MaxRAM

    TYPE(Timer)    :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    
    !------------------------------------------------------------------
    !---- Factorization
    !------------------------------------------------------------------
    mumps_par%JOB = 2
    CALL myTimer%Start()
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------    
    iError = mumps_par%INFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in factorization step of MUMPS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Maximal memory used for factorization
    !------------------------------------------------------------------    
    CALL MPI_ALLREDUCE(mumps_par%INFO(16), MaxRAM, 1, MPI_INT, MPI_MAX, defaultComm, mpiError)
    !------------------------------------------------------------------
    !---- Print MUMPS informations
    !------------------------------------------------------------------
    IF(idWrite) THEN
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       " ** Factorization step statistics           "
       WRITE(6, '(a)')       "      Memory effectively used during facto. "
       WRITE(6, '(a,i12)')   "                Most consuming proc. (MB) : ", MaxRAM
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       "     Time spent for factorization (seconds) "
       WRITE(6, '(a,e12.5)') "                                 CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e12.5)') "                                Wall time : ", myTimer%SYSTEM_TIME()
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_FACTORIZATION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_RHS_UPDATE
    !------------------------------------------------------------------
    !---- Updating RHS on absorbing boundary faces and interfaces
    !---- (for classical Schwarz algorithm)
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE SCHWARZ_TYPES
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, faceNumArt, i, iCell1, iCell2, iFac,    &
         indLocu, indLocw, iu1, iu2, iw1, iw2, k1, k2, nDoFs, sum1, sum2 

    COMPLEX(C_DOUBLE_COMPLEX) :: zero
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of RHS
    !------------------------------------------------------------------
    zero = (0.d0, 0.d0)
    mumps_par%RHS(:) = zero
    !------------------------------------------------------------------
    !---- RHS, absorbing boundary condition contribution
    !------------------------------------------------------------------
    IF (ALLOCATED(rhsAbs)) THEN
       DO iFac=1,Mesh%nFacAbs
          faceNum = MeshFacAbs(iFac)%faceNum
          nDoFs    = Lambda(faceNum)%nDof
          DO i=1,nDoFs
             indLocu                = indLocMtx(faceNum) + i
             indLocw                = indLocu + nDoFs
             mumps_par%RHS(indLocu) = rhsAbs(iFac)%vec(i,1)
             mumps_par%RHS(indLocw) = rhsAbs(iFac)%vec(i,2)
          ENDDO
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- RHS, current source term contribution
    !------------------------------------------------------------------
    IF (ALLOCATED(rhsJglb)) THEN
       !
       !---- Inner faces
       DO iFac=1,Mesh%nFacI
          faceNum = MeshFacI(iFac)%faceNum
          nDoFs   = Lambda(faceNum)%nDof
          !
          iCell1  = MeshFac(faceNum)%neighb(1)
          k1      = MeshFac(faceNum)%numLocal(1)
          sum1    = hdgEM(iCell1)%sumDofF(k1) - nDoFs
          !
          iCell2  = MeshFac(faceNum)%neighb(2)
          k2      = MeshFac(faceNum)%numLocal(2)
          sum2    = hdgEM(iCell2)%sumDofF(k2) - nDoFs
          !
          DO i=1,nDoFs
             indLocu           = indLocMtx(faceNum) + i
             indLocw           = indLocu + nDoFs
             iu1               = 2*sum1  + i
             iw1               = iu1 + nDoFs
             iu2               = 2*sum2  + i
             iw2               = iu2 + nDoFs
             mumps_par%RHS(indLocu) = mumps_par%RHS(indLocu)                     &
                  + rhsJglb(iCell1)%tab(iu1) + rhsJglb(iCell2)%tab(iu2)
             mumps_par%RHS(indLocw) = mumps_par%RHS(indLocw)                     &
                  + rhsJglb(iCell1)%tab(iw1) + rhsJglb(iCell2)%tab(iw2)
          ENDDO
       ENDDO
       !
       !---- Boundary faces and interfaces (i.e. physical and artificial faces)
       DO iFac=1,Mesh%nFacB
          faceNum = MeshFacB(iFac)%faceNum
          nDoFs   = Lambda(faceNum)%nDof
          !
          iCell1  = MeshFac(faceNum)%neighb(1)
          k1      = MeshFac(faceNum)%numLocal(1)
          sum1    = hdgEM(iCell1)%sumDofF(k1) - nDoFs
          !
          DO i=1,nDoFs
             indLocu           = indLocMtx(faceNum) + i
             indLocw           = indLocu + nDoFs
             iu1               = 2*sum1  + i
             iw1               = iu1 + nDoFs
             mumps_par%RHS(indLocu) = mumps_par%RHS(indLocu) + rhsJglb(iCell1)%tab(iu1)
             mumps_par%RHS(indLocw) = mumps_par%RHS(indLocw) + rhsJglb(iCell1)%tab(iw1)
          ENDDO
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- RHS, interfaces contributions
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFacArt
       faceNum    = MeshFacArt(iFac)%faceNum
       nDoFs      = Lambda(faceNum)%nDof
       faceNumArt = indArt(faceNum)
       DO i=1,nDoFs
          indLocu                = indLocMtx(faceNum) + i
          indLocw                = indLocu + nDoFs
          mumps_par%RHS(indLocu) = mumps_par%RHS(indLocu) + rhsSchArt(faceNumArt)%vec(i,1)
          mumps_par%RHS(indLocw) = mumps_par%RHS(indLocw) + rhsSchArt(faceNumArt)%vec(i,2)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_RHS_UPDATE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_RHS_UPDATE_INTERFACES
    !------------------------------------------------------------------
    !---- Updating RHS only on interfaces
    !---- (for Schwarz algorithm accelerated by a Krylov subspace method)
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE SCHWARZ_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, faceNumArt, i, iFac, indLocu, indLocw, nDof

    COMPLEX(C_DOUBLE_COMPLEX) :: zero
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of RHS
    !------------------------------------------------------------------
    zero = (0.d0, 0.d0)
    mumps_par%RHS(:) = zero
    !------------------------------------------------------------------
    !---- Update of RHS on interfaces
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFacArt
       facNum     = MeshFacArt(iFac)%faceNum
       nDof       = Lambda(facNum)%nDof
       faceNumArt = indArt(facNum)
       DO i=1,nDof
          indLocu                = indLocMtx(facNum) + i
          indLocw                = indLocu + nDof
          mumps_par%RHS(indLocu) = rhsSchArt(faceNumArt)%vec(i,1)
          mumps_par%RHS(indLocw) = rhsSchArt(faceNumArt)%vec(i,2)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_RHS_UPDATE_INTERFACES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_SOLVE
    !------------------------------------------------------------------
    !---- MUMPS solve step 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, i, iError, indLocu, indLocw, nDof
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Solve
    !------------------------------------------------------------------
    mumps_par%JOB = 3
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------    
    iError = mumps_par%INFO(1)
    IF (iError < 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in solve step of MUMPS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Distribution of the DoFs of the hybrid variable (RHS stores 
    !---- the solution)
    !------------------------------------------------------------------
    DO facNum=1,Mesh%nFac
       nDof = Lambda(facNum)%nDof
       DO i=1,nDof
          indLocu                   = indLocMtx(facNum) + i
          indLocw                   = indLocu + nDof
          Lambda(facNum)%L%vec(i,1) = mumps_par%RHS(indLocu)
          Lambda(facNum)%L%vec(i,2) = mumps_par%RHS(indLocw)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE SCHWARZ_MUMPS_SOLVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE SCHWARZ_MUMPS_END
    !------------------------------------------------------------------
    !---- MUMPS final step : deallocations 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !--- Final phase - Deallocations
    !------------------------------------------------------------------
    DEALLOCATE(mumps_par%RHS)
    DEALLOCATE(mumps_par%IRN)
    DEALLOCATE(mumps_par%JCN)
    DEALLOCATE(mumps_par%A  )
    DEALLOCATE(indLocMtx    )
    !------------------------------------------------------------------
    !---- Destroy the instance (deallocate internal data structures)
    !------------------------------------------------------------------
    mumps_par%JOB = -2
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------ 
  END SUBROUTINE SCHWARZ_MUMPS_END

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE SCHWARZ_MUMPS_STRUCTURES
