  SUBROUTINE BICGSTAB(l)
    !-------------------------------------------------------------------- 
    !---- BiCGStab(l) iterative method applied to the interface system 
    !---- in Schwarz algorithm accelerated by a Krylov subspace method
    !
    !---- IN
    !
    !---- OUT
    !
    !--------------------------------------------------------------------
    !---- Modules
    !--------------------------------------------------------------------
    USE INPUT_DATA
    USE ISO_C_BINDING
    USE HDG_CORE
    USE KRYLOV_CORE
    USE KRYLOV_STRUCTURES
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    USE SCHWARZ_CORE
    USE SCHWARZ_MUMPS_STRUCTURES
    USE SCHWARZ_STRUCTURES
    USE SCHWARZ_TYPES
    USE TIMERS
    !--------------------------------------------------------------------
    IMPLICIT NONE
    !--------------------------------------------------------------------
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN)  :: l
    !--------------------------------------------------------------------
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: bp, facNum, faceNumArt, i, ierr, iCell, id, iFac,  &
         iPerm, iter, j, ji, k, n, nDof, nmv, order, r, rr, u, xp, y0,   &
         y, yl, z1, zz

    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: iwork

    REAL(C_DOUBLE) :: delta, mxnrmx, mxnrmr, precision, rnrm0, rnrm

    COMPLEX(C_DOUBLE_COMPLEX) :: alpha, beta, hatgamma, kappa0, kappal,  &
         omega1, rho0, rho1, rho1l, sigm, sigml, varrho, zone, zzero

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:)  , ALLOCATABLE :: vecx    
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: rwork
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:,:), ALLOCATABLE :: work
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:)  , ALLOCATABLE :: zscal

    LOGICAL(C_BOOL) :: rcmp, xpdt

    TYPE(Timer) :: myTimer
    !--------------------------------------------------------------------
    !---- External function
    !--------------------------------------------------------------------
    REAL(C_DOUBLE), EXTERNAL :: MYZNRM

    COMPLEX*16, EXTERNAL :: ZDOTC
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
    !---- Dynamic allocations and initialization
    !--------------------------------------------------------------------
    zone  = (1.d0, 0.d0)
    zzero = (0.d0, 0.d0)
    !
    ALLOCATE(iwork(l+1)                 )
    ALLOCATE(vecx(6*nDofEMart)          )    
    ALLOCATE(rwork(l+1,3+2*(l+1))       )
    ALLOCATE(work(6*nDofEMart,3+2*(l+1)))
    ALLOCATE(zscal(l+1)                 )
    iwork(:)   = 0
    vecx(:)    = zzero
    rwork(:,:) = zzero
    work(:,:)  = zzero
    zscal(:)   = zzero
    !--------------------------------------------------------------------
    !---- Executable statements
    !--------------------------------------------------------------------     
    n     = 6*nDofEMart
    ierr  = 0
    delta = 1.0e-02
    rr    = 1
    r     = rr + 1
    u     = r + (l + 1)
    xp    = u + (l + 1)
    bp    = xp + 1
    z1    = 1
    zz    = z1 + (l + 1)
    y0    = zz + (l + 1)
    yl    = y0 + 1
    y     = yl + 1
    !--------------------------------------------------------------------
    !---- Initialize first residual
    !--------------------------------------------------------------------
    id = 1
    DO iFac=1,Mesh%nFacArt
       facNum = MeshFacArt(iFac)%faceNum
       iCell  = MeshFac(facNum)%neighb(1)
       order  = hdgEM(iCell)%order
       nDof   = hdgEM(iCell)%nDofS
       k      = MeshFac(facNum)%numLocal(1)
       iPerm  = MeshCell(iCell)%perm(1,k)
       DO j=1,nDof
          ji = hdgEM(iCell)%iPermDof%ptr%tab(j,iPerm)
          DO i=1,6
             work(id+i-1,r) = EMsave(iCell)%vec(ji,i)
          ENDDO
          id = id + 6
       ENDDO
    ENDDO
    !--------------------------------------------------------------------
    !---- Initialize iteration loop
    !--------------------------------------------------------------------
    iter      = 0
    !
    !---- Stopping criterion threshold in the iterative solver
    SELECT CASE(thresholdIter)
    CASE("max")
       precision      = (globalMesh%hMax)**(genOrderMax + 2)
    CASE("avg")
       precision      = (globalMesh%hAvg)**(genOrderMax + 2)
    CASE("min")
       precision      = (globalMesh%hMin)**(genOrderMax + 2)
    END SELECT
    IF (precision < 1.0e-15) precision = 1.0e-15
    !
    nmv       = 0
    CALL ZCOPY(n, work(1,r), 1, work(1,rr), 1)
    CALL ZCOPY(n, work(1,r), 1, work(1,bp), 1)
    CALL ZCOPY(n, vecx, 1, work(1,xp), 1)
    CALL ZLASET('n', n, 1, zzero, zzero, vecx, 1)
    rnrm0     = MYZNRM(n, work(1,r))
    rnrm      = rnrm0
    mxnrmx    = rnrm0
    mxnrmr    = rnrm0
    rcmp      = .FALSE.
    xpdt      = .FALSE.
    alpha     = zzero
    omega1    = zone
    sigm      = zone
    rho0      = zone
    !-------------------------------------------------------------------- 
    !---- Iterate
    !--------------------------------------------------------------------
    !
    !---- Print informations
    IF (idWrite) THEN
       WRITE(6, 10)
       WRITE(6, '(a,i0,a)') " BiCGStab(", l,") iterative method "
       WRITE(6, 10)
       WRITE(6, '(a)')      " Iteration        Residual         "
    ENDIF
    !
    CALL myTimer%Start()
    DO WHILE( (rnrm > precision*rnrm0) .AND. (iter < maxIter) )
       IF((iter == 0) .AND. (idWrite)) WRITE(6, '(i10,a5,e15.6)') iter, " ", rnrm/rnrm0
       !
       !---- The BiCG part
       rho0 = -omega1*rho0
       DO k=1,l
          rho1l = ZDOTC(n, work(1,rr), 1, work(1,r+k-1), 1)
          CALL MPI_ALLREDUCE(rho1l, rho1, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError)
          !
          !---- Testing for errors
          IF(rho0 == zzero) THEN
             WRITE(6, '(a)') " Breakdown of the BiCGStab(l) algorithm &
                  (try to enlarge parameter l to get rid of this) "
             CALL TILT
          ENDIF
          !
          beta = alpha*(rho1/rho0)
          rho0 = rho1
          DO j=0,k-1
             DO i=1,n
                work(i,u+j) = work(i,r+j) - beta*work(i,u+j) 
             ENDDO
          ENDDO
          id = 1
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             nDof       = hdgEM(iCell)%nDofS
             DO j=1,nDof
                DO i=1,6
                   EMold(faceNumArt)%vec(j,i) = work(id+i-1,u+k-1)
                ENDDO
                id = id + 6
             ENDDO
          ENDDO
          CALL KRYLOV_EM_EXCHANGE
          CALL SCHWARZ_RHS_UPDATE
          CALL SCHWARZ_MUMPS_RHS_UPDATE_INTERFACES
          CALL SCHWARZ_MUMPS_SOLVE
          DO iFac=1,Mesh%nFacArt
             facNum = MeshFacArt(iFac)%faceNum
             iCell  = MeshFac(facNum)%neighb(1)
             CALL LOCAL_SOLVER_WITHOUT_CURRENT_SRC(iCell)
          ENDDO
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             order      = hdgEM(iCell)%order
             nDof       = hdgEM(iCell)%nDofS
             i          = MeshFac(facNum)%numLocal(1)
             iPerm      = MeshCell(iCell)%perm(1,i)
             DO j=1,nDof
                ji                         = hdgEM(iCell)%iPermDof%ptr%tab(j,iPerm)
                EMnew(faceNumArt)%vec(j,1) = EMold(faceNumArt)%vec(j,1) - hdgEM(iCell)%E%vec(ji,1)
                EMnew(faceNumArt)%vec(j,2) = EMold(faceNumArt)%vec(j,2) - hdgEM(iCell)%E%vec(ji,2)
                EMnew(faceNumArt)%vec(j,3) = EMold(faceNumArt)%vec(j,3) - hdgEM(iCell)%E%vec(ji,3)
                EMnew(faceNumArt)%vec(j,4) = EMold(faceNumArt)%vec(j,4) - hdgEM(iCell)%H%vec(ji,1)
                EMnew(faceNumArt)%vec(j,5) = EMold(faceNumArt)%vec(j,5) - hdgEM(iCell)%H%vec(ji,2)
                EMnew(faceNumArt)%vec(j,6) = EMold(faceNumArt)%vec(j,6) - hdgEM(iCell)%H%vec(ji,3)
             ENDDO
          ENDDO
          id = 1
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             nDof       = hdgEM(iCell)%nDofS
             DO j=1,nDof
                DO i=1,6
                   work(id+i-1,u+k) = EMnew(faceNumArt)%vec(j,i)
                ENDDO
                id = id + 6
             ENDDO
          ENDDO
          nmv   = nmv + 1
          sigml = ZDOTC(n, work(:,rr), 1, work(:,u+k), 1)
          CALL MPI_ALLREDUCE(sigml, sigm, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError)
          !
          !---- Testing for errors
          IF( sigm == zzero ) THEN
             WRITE(6, '(a)') " Breakdown of the BiCGStab(l) algorithm &
                  (try to enlarge parameter l to get rid of this) "
             CALL TILT
          ENDIF
          !
          alpha = rho1/sigm
          CALL ZAXPY(n, alpha, work(1,u), 1, vecx, 1)
          DO j=0,k-1
             CALL ZAXPY(n, (-alpha), work(1,u+j+1), 1, work(1,r+j), 1)
          ENDDO
          id = 1
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             nDof       = hdgEM(iCell)%nDofS
             DO j=1,nDof
                DO i=1,6
                   EMold(faceNumArt)%vec(j,i) = work(id+i-1,r+k-1)
                ENDDO
                id = id +6
             ENDDO
          ENDDO
          CALL KRYLOV_EM_EXCHANGE
          CALL SCHWARZ_RHS_UPDATE
          CALL SCHWARZ_MUMPS_RHS_UPDATE_INTERFACES
          CALL SCHWARZ_MUMPS_SOLVE
          DO iFac=1,Mesh%nFacArt
             facNum = MeshFacArt(iFac)%faceNum
             iCell  = MeshFac(facNum)%neighb(1)
             CALL LOCAL_SOLVER_WITHOUT_CURRENT_SRC(iCell)
          ENDDO
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             order      = hdgEM(iCell)%order
             nDof       = hdgEM(iCell)%nDofS
             i          = MeshFac(facNum)%numLocal(1)
             iPerm      = MeshCell(iCell)%perm(1,i)
             DO j=1,nDof
                ji                   = hdgEM(iCell)%iPermDof%ptr%tab(j,iPerm)
                EMnew(faceNumArt)%vec(j,1) = EMold(faceNumArt)%vec(j,1) - hdgEM(iCell)%E%vec(ji,1)
                EMnew(faceNumArt)%vec(j,2) = EMold(faceNumArt)%vec(j,2) - hdgEM(iCell)%E%vec(ji,2)
                EMnew(faceNumArt)%vec(j,3) = EMold(faceNumArt)%vec(j,3) - hdgEM(iCell)%E%vec(ji,3)
                EMnew(faceNumArt)%vec(j,4) = EMold(faceNumArt)%vec(j,4) - hdgEM(iCell)%H%vec(ji,1)
                EMnew(faceNumArt)%vec(j,5) = EMold(faceNumArt)%vec(j,5) - hdgEM(iCell)%H%vec(ji,2)
                EMnew(faceNumArt)%vec(j,6) = EMold(faceNumArt)%vec(j,6) - hdgEM(iCell)%H%vec(ji,3)
             ENDDO
          ENDDO
          id = 1
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             nDof       = hdgEM(iCell)%nDofS
             DO j=1,nDof
                DO i=1,6
                   work(id+i-1,r+k) = EMnew(faceNumArt)%vec(j,i)
                ENDDO
                id = id + 6
             ENDDO
          ENDDO
          nmv  = nmv + 1
          rnrm = MYZNRM(n, work(1,r))
          mxnrmx = MAX(mxnrmx, rnrm)
          mxnrmr = MAX(mxnrmr, rnrm)
       ENDDO
       !
       !---- The convex polynomial part
       DO i=1,l+1
          CALL ZGEMV('c', n, l+1-(i-1), zone, work(1,r+i-1), n, &
               work(1,r+i-1), 1, zzero, rwork(i,z1+i-1), 1)
          CALL MPI_ALLREDUCE(rwork(i,z1+i-1), zscal, l+1-(i-1), &
               MPI_DOUBLE_COMPLEX, MPI_SUM, defaultComm, mpiError)
          DO j=1,l+1-(i-1)
             rwork(i+j-1,z1+i-1) = zscal(j)
          ENDDO
          IF( (l-(i-1)) /= 0 ) THEN
             CALL ZCOPY(l-(i-1), rwork(i+1,z1+i-1), 1, rwork(i,z1+i), l+1)
             CALL ZLACGV(l-(i-1), rwork(i,z1+i), l+1)
          ENDIF
       ENDDO
       CALL ZLACPY('a', l+1, l+1, rwork(1,z1), l+1, rwork(1,zz), l+1)
       CALL ZGETRF(l-1, l-1, rwork(2,zz+1), l+1, iwork, ierr)
       !
       rwork(1,y0) = -zone
       CALL ZCOPY(l-1, rwork(2,z1), 1, rwork(2,y0), 1)
       CALL ZGETRS('n', l-1, 1, rwork(2,zz+1), l+1, iwork, rwork(2,y0), l+1, ierr)
       rwork(l+1,y0) = zzero
       rwork(1,yl) = zzero
       CALL ZCOPY(l-1, rwork(2,z1+l), 1, rwork(2,yl), 1)
       CALL ZGETRS('n', l-1, 1, rwork(2,zz+1), l+1, iwork, rwork(2,yl), l+1, ierr)
       rwork(l+1,yl) = -zone
       !
       CALL ZHEMV('u', l+1, zone, rwork(1,z1), l+1, rwork(1,y0), 1,     &
            zzero, rwork(1,y), 1)
       kappa0 = SQRT(ZDOTC(l+1, rwork(:,y0), 1, rwork(:,y), 1))
       CALL ZHEMV('u', l+1, zone, rwork(1,z1), l+1, rwork(1,yl), 1,     &
            zzero, rwork(1,y), 1)
       kappal = SQRT(ZDOTC(l+1, rwork(:,yl), 1, rwork(:,y), 1))
       CALL ZHEMV('u', l+1, zone, rwork(1,z1), l+1, rwork(1,y0), 1,     &
            zzero, rwork(1,y), 1)
       varrho = ZDOTC(l+1, rwork(:,yl), 1, rwork(:,y), 1)/(kappa0*kappal)
       hatgamma = varrho/ABS(varrho)*MAX(ABS(varrho), 7.0d-01)*(kappa0/kappal)
       CALL ZAXPY(l+1, (-hatgamma), rwork(1,yl), 1, rwork(1,y0), 1)
       !
       omega1 = rwork(l+1,y0)
       CALL ZGEMV('n', n, l, (-zone), work(1,u+1), n, rwork(2,y0), 1,   & 
            zone, work(1,u), 1)
       CALL ZGEMV('n', n, l, zone, work(1,r), n, rwork(2,y0), 1,        & 
            zone, vecx, 1)
       CALL ZGEMV('n', n, l, (-zone), work(1,r+1), n, rwork(2,y0), 1,   & 
            zone, work(1,r), 1)
       CALL ZHEMV('u', l+1, zone, rwork(1,z1), l+1, rwork(1,y0), 1,     &
            zzero, rwork(1,y), 1)
       rnrm = SQRT(ZDOTC(l+1, rwork(1,y0), 1, rwork(1,y), 1))
       !
       mxnrmx = MAX(mxnrmx, rnrm)
       mxnrmr = MAX(mxnrmr, rnrm)
       xpdt = ((rnrm < delta*rnrm0) .AND. (rnrm0 < mxnrmx))
       rcmp = (((rnrm < delta*mxnrmr) .AND. (rnrm0 < mxnrmr)) .OR. xpdt)
       IF( rcmp ) THEN
          id = 1
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             nDof       = hdgEM(iCell)%nDofS
             DO j=1,nDof
                DO i=1,6
                   EMold(faceNumArt)%vec(j,i) = vecx(id+i-1)
                ENDDO
                id = id + 6
             ENDDO
          ENDDO
          CALL KRYLOV_EM_EXCHANGE
          CALL SCHWARZ_RHS_UPDATE
          CALL SCHWARZ_MUMPS_RHS_UPDATE_INTERFACES
          CALL SCHWARZ_MUMPS_SOLVE
          DO iFac=1,Mesh%nFacArt
             facNum = MeshFacArt(iFac)%faceNum
             iCell  = MeshFac(facNum)%neighb(1)
             CALL LOCAL_SOLVER_WITHOUT_CURRENT_SRC(iCell)
          ENDDO
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             order      = hdgEM(iCell)%order
             nDof       = hdgEM(iCell)%nDofS
             i          = MeshFac(facNum)%numLocal(1)
             iPerm      = MeshCell(iCell)%perm(1,i)
             DO j=1,nDof
                ji                   = hdgEM(iCell)%iPermDof%ptr%tab(j,iPerm)
                EMnew(faceNumArt)%vec(j,1) = EMold(faceNumArt)%vec(j,1) - hdgEM(iCell)%E%vec(ji,1)
                EMnew(faceNumArt)%vec(j,2) = EMold(faceNumArt)%vec(j,2) - hdgEM(iCell)%E%vec(ji,2)
                EMnew(faceNumArt)%vec(j,3) = EMold(faceNumArt)%vec(j,3) - hdgEM(iCell)%E%vec(ji,3)
                EMnew(faceNumArt)%vec(j,4) = EMold(faceNumArt)%vec(j,4) - hdgEM(iCell)%H%vec(ji,1)
                EMnew(faceNumArt)%vec(j,5) = EMold(faceNumArt)%vec(j,5) - hdgEM(iCell)%H%vec(ji,2)
                EMnew(faceNumArt)%vec(j,6) = EMold(faceNumArt)%vec(j,6) - hdgEM(iCell)%H%vec(ji,3)
             ENDDO
          ENDDO
          id = 1
          DO iFac=1,Mesh%nFacArt
             facNum     = MeshFacArt(iFac)%faceNum
             faceNumArt = indArt(facNum)
             iCell      = MeshFac(facNum)%neighb(1)
             nDof       = hdgEM(iCell)%nDofS
             DO j=1,nDof
                DO i=1,6
                   work(id+i-1,r) = EMnew(faceNumArt)%vec(j,i)
                ENDDO
                id = id + 6
             ENDDO
          ENDDO
          DO i=1,n
             work(i,r) = work(i,bp) - work(i,r)
          ENDDO
          mxnrmr = rnrm
          IF( xpdt ) THEN
             CALL ZAXPY(n, zone, vecx, 1, work(1,xp),1)
             CALL ZLASET('n', n, 1, zzero, zzero, vecx, 1)
             CALL ZCOPY(n, work(1,r), 1, work(1,bp), 1)
             mxnrmx = rnrm
          ENDIF
       ENDIF
       iter = iter + 1
       !
       !---- Print informations
       IF(idWrite) WRITE(6, '(i10,a5,e15.6)') iter, " ", rnrm/rnrm0
    ENDDO
    !------------------------------------------------------------------
    !---- Print informations
    !------------------------------------------------------------------
    IF (idWrite) THEN
       WRITE(6, '(a)')            " "
       WRITE(6, '(a)')            " Solution quality for Lambda            "
       WRITE(6, '(a,e15.6)')      "                     Wanted precision : ", precision
       WRITE(6, '(a,e15.6)')      "                   Effective residual : ", rnrm/rnrm0
       WRITE(6, '(a,i15)')        "                    Max. nb. of iters : ", maxIter
       WRITE(6, '(a,i15)')        "                         Nb. of iters : ", iter
       WRITE(6, '(a)')            " "
       WRITE(6, '(a)')            " Time spent in iterative step (seconds) "
       WRITE(6, '(a,e15.6)')      "                             CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e15.6)')      "                            Wall time : ", myTimer%SYSTEM_TIME()
       WRITE(6, 10)
    ENDIF
    !
    CALL ZAXPY(n, zone, work(1,xp), 1, vecx, 1)
    id = 1
    DO iFac=1,Mesh%nFacArt
       facNum     = MeshFacArt(iFac)%faceNum
       faceNumArt = indArt(facNum)
       iCell      = MeshFac(facNum)%neighb(1)
       nDof       = hdgEM(iCell)%nDofS
       DO j=1,nDof
          DO i=1,6
             EMold(faceNumArt)%vec(j,i) = vecx(id+i-1)
          ENDDO
          id = id +6
       ENDDO
    ENDDO
    CALL KRYLOV_EM_EXCHANGE
    CALL SCHWARZ_RHS_UPDATE
    CALL SCHWARZ_MUMPS_RHS_UPDATE_INTERFACES
    CALL SCHWARZ_MUMPS_SOLVE
    !--------------------------------------------------------------------
    !---- Deallocations
    !--------------------------------------------------------------------
    DEALLOCATE(iwork)
    DEALLOCATE(vecx)    
    DEALLOCATE(rwork)
    DEALLOCATE(work)
    DEALLOCATE(zscal)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
10  FORMAT(60('-'))
    !--------------------------------------------------------------------
  END SUBROUTINE BICGSTAB

  !====================================================================
  !====================================================================
  !====================================================================

  REAL(C_DOUBLE) FUNCTION MYZNRM(n,zx)

    !-------------------------------------------------------------------- 
    !---- Returns the euclidean norm of a vector
    !
    !---- IN
    !
    !---- OUT
    !
    !-------------------------------------------------------------------- 
    !---- Modules
    !--------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    !-------------------------------------------------------------------- 
    !---- I/O variables
    !--------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: n

    COMPLEX(C_DOUBLE_COMPLEX), INTENT(IN) :: zx(n)
    !-------------------------------------------------------------------- 
    !---- Local variables
    !--------------------------------------------------------------------
    INTEGER(C_INT) :: i

    REAL(C_DOUBLE) :: rnrml, rnrmg
    !-------------------------------------------------------------------- 
    rnrml = 0.d0
    DO i=1,n
       rnrml = rnrml + ABS(zx(i))**2
    ENDDO
    CALL MPI_ALLREDUCE(rnrml, rnrmg, 1, MPI_DOUBLE_PRECISION,MPI_SUM, defaultComm, mpiError)
    MYZNRM = SQRT(rnrmg)
    !--------------------------------------------------------------------
    !---- Testing for bad input
    !--------------------------------------------------------------------
    !
    !--------------------------------------------------------------------
  END FUNCTION MYZNRM
