MODULE GAUSSIAN

  !====================================================================
  ! Routines related to 3D Gaussian function  
  ! Gauss(x,y,z) = 
  ! A*EXP(-[(x-xs)^2/(2*sigmax^2)
  !        +(y-ys)^2/(2*sigmay^2)
  !        +(z-zs)^2/(2*sigmaz^2)])
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! GAUSSIAN DATA ('GAUSSIAN')
  !--------------------------------------------------------------------
  ! amplGauss    ==> Amplitude for Gaussian         :  A
  ! centerGauss  ==> Center of the Gaussian         : (xs,     ys,     zs)
  ! sigmaGauss   ==> Coefficient of the Gaussian    : (sigmax, sigmay, sigmaz)
  !--------------------------------------------------------------------
  REAL(C_DOUBLE)               :: amplGauss

  REAL(C_DOUBLE), DIMENSION(3) :: centerGauss, sigmaGauss
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! GAUSSIAN Function
  !--------------------------------------------------------------------
  ! GaussCell    ==> Gaussian function at DoFs of a cell
  ! GaussFace    ==> Gaussian function at DoFs of a face
  !--------------------------------------------------------------------
  REAL(C_DOUBLE), DIMENSION(:), ALLOCATABLE :: GaussCell, GaussFace
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE CELL_GAUSSIAN(iCell, nDofV, order)
    !------------------------------------------------------------------
    !---- Get values of the Gaussian at the DoFs of a cell
    !
    !---- IN
    !---- iCell : number of the cell
    !---- nDofV : number of DoFs in the cell
    !---- order : polynomial order in the cell
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell, nDofV, order
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iDoF
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute Gauss function
    !------------------------------------------------------------------
    IF(.NOT. ALLOCATED(GaussCell)) ALLOCATE(GaussCell(nDofVmax))
    GaussCell = 0.d0
    CALL GET_CELL_DOF(iCell, order, nDofV, dofXV(1:nDofV), dofYV(1:nDofV), dofZV(1:nDofV))
    !
    DO iDoF=1,nDofV
       CALL GAUSSIAN_FUNCTION(dofXV(iDoF), dofYV(iDoF), dofZV(iDoF), GaussCell(iDoF))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_GAUSSIAN 

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE FACE_GAUSSIAN(iFace, nDofS, nDofV, order)
    !------------------------------------------------------------------
    !---- Get values of the Gaussian at the DoFs of a face
    !
    !---- IN
    !---- iFace : number of the face
    !---- nDofS : number of DoFs in the face
    !---- nDofV : number of DoFs in the cell
    !---- order : polynomial order in the face
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_ELEMENTS_DOFS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iFace, nDofS, nDofV,order
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)             :: iDoF
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute Gauss function
    !------------------------------------------------------------------
    IF(.NOT. ALLOCATED(GaussFace)) ALLOCATE(GaussFace(nDofSmax))
    GaussFace = 0.d0
    CALL GET_FACE_DOF(iFace, order, nDofS, nDofV, dofXS(1:nDofS), dofYS(1:nDofS), dofZS(1:nDofS))
    !
    DO iDoF=1,nDofS
       CALL GAUSSIAN_FUNCTION(dofXS(iDoF), dofYS(iDoF), dofZS(iDoF), GaussFace(iDoF))
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE FACE_GAUSSIAN 

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE GAUSSIAN_FUNCTION(x, y, z, Gauss)
    !------------------------------------------------------------------
    !---- Value of the Gaussian at point (x, y, z)
    !
    !---- IN
    !---- x, y, z : coordinates for the Gaussian    
    !
    !---- OUT
    !---- Gauss   : value of the Gaussian at (x, y, z)
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN)  :: x, y, z

    REAL(C_DOUBLE), INTENT(OUT) :: Gauss
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Gauss function at (x, y, z)
    !------------------------------------------------------------------
    Gauss = amplGauss*EXP(-((centerGauss(1) - x)**2/(2*sigmaGauss(1)**2) &
         + (centerGauss(2) - y)**2/(2*sigmaGauss(2)**2)                  &
         + (centerGauss(3) - z)**2/(2*sigmaGauss(3)**2)))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE GAUSSIAN_FUNCTION 

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE GAUSSIAN
