MODULE INPUT_DATA

  !====================================================================
  ! Handles input data
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Reference data
  !--------------------------------------------------------------------
  ! dataFile  ==> Name of the data file
  ! dataUnit  ==> Unit for reading dataFile
  !--------------------------------------------------------------------
  CHARACTER(LEN=9, KIND=C_CHAR), PARAMETER :: dataFile = "dataHorse"

  INTEGER(C_INT), PARAMETER :: dataUnit = 1

  !--------------------------------------------------------------------
  ! Solver data
  !--------------------------------------------------------------------
  ! hdgMethod      ==> Methods currently available:
  !                    - '0' (Classical hdgm)
  !                    - '1' (hdgm combined to a Schwarz algo)
  !                    - '2' (hdgm combined to a Schwarz 
  !                    algo accelerated by a Krylov subspace method)
  ! solverPack     ==> Solver package used to solve the global linear system 
  !                    in terms of dofs of the hybrid variable in the hdgm
  !                    Packages currently available:
  !                    - "mumps"  (MUMPS, direct solver)
  !                    - "maphys" (MaPHyS, direct / iterative solver)
  ! BiCGStabLevel  ==> Level for BiCGStab method (usefull for hdgMethod = 2)
  ! freq           ==> frequency (Hz)
  ! ptsWavelength  ==> Minimal number of pts per wavelength
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: BiCGStabLevel, hdgMethod, maxIter, ptsWavelength
  
  REAL(C_DOUBLE) :: freq
  
  CHARACTER(LEN=8, KIND=C_CHAR) :: thresholdIter
  CHARACTER(LEN=8, KIND=C_CHAR) :: solverPack
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! RCS data (angles Phi and Theta)
  !--------------------------------------------------------------------
  ! PhiInitRCS     ==> Initial Phi angle for RCS computation (degree)
  ! PhiEndRCS      ==> Final   Phi angle for RCS computation (degree)
  ! nPhi           ==> # of Phi angles
  ! ThetaInitRCS   ==> Initial Theta angle for RCS computation (degree)
  ! ThetaEndRCS    ==> Final   Theta angle for RCS computation (degree)
  ! nTheta         ==> # of Theta angles
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: nPhiRCS, nThetaRCS
  
  REAL(C_DOUBLE) :: PhiEndRCS, PhiInitRCS, ThetaEndRCS, ThetaInitRCS
  !--------------------------------------------------------------------


CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_DATA
    !------------------------------------------------------------------
    !---- Read dataHorse file and check for incompatibilities in input data
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read data
    !------------------------------------------------------------------
    CALL READ_MESHNAME
    CALL READ_DATASOLVER
    CALL READ_DATASOURCE
    CALL READ_DATAOBSERVABLE
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_DATA

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_MESHNAME
    !------------------------------------------------------------------
    !---- This routine builds:
    !---- - The name attribute of the Mesh structure
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readChain
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read meshName
    !------------------------------------------------------------------
    OPEN(UNIT=dataUnit, FILE=dataFile, STATUS="UNKNOWN")
    !
    DO WHILE (INDEX(readChain,'#MeshName')==0)
       READ(dataUnit,'(a)') readChain 
    ENDDO
    !
    READ(dataUnit,*) Mesh%name
    !
    CLOSE(dataUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_MESHNAME

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_DATASOLVER
    !------------------------------------------------------------------
    !---- Read DataSolver in dataHorse and check for incompatibilities
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    USE MATH_PHYS_CONSTANTS
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readChain
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: tmpOrder
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read meshName
    !------------------------------------------------------------------
    OPEN(UNIT=dataUnit, FILE=dataFile, STATUS="UNKNOWN")
    !
    DO WHILE (INDEX(readChain,'#DataSolver')==0)
       READ(dataUnit,'(a)') readChain
    ENDDO
    !
    READ(dataUnit,*) hdgMethod
    READ(dataUnit,*) solverPack
    READ(dataUnit,*) polyBase
    READ(dataUnit,*) orderMin, orderMax
    READ(dataUnit,*) orderMinH, orderMaxH
    READ(dataUnit,*) ptsWavelength
    READ(dataUnit,*) thresholdIter
    READ(dataUnit,*) maxIter
    READ(dataUnit,*) BiCGStabLevel
    !
    CLOSE(dataUnit)
    !------------------------------------------------------------------
    !---- Create strings for orderMin and orderMax
    !------------------------------------------------------------------    
    CALL INT_TO_CHAR(orderMin, cOrderMin)
    CALL INT_TO_CHAR(orderMax, cOrderMAx)
    CALL INT_TO_CHAR(orderMinH, cOrderMinH)
    CALL INT_TO_CHAR(orderMaxH, cOrderMaxH)
    !------------------------------------------------------------------
    !---- Output + Testing for errors (existence of hdgMethod)
    !------------------------------------------------------------------
    !
    !---- Identification and existence of hdgMethod
    SELECT CASE(hdgMethod)
    CASE(0)
       IF (idWrite) THEN
          WRITE(6, 10) 
          WRITE(6, '(a)') " Classical HDGM "
          WRITE(6, 10)
       ENDIF
    CASE(1)
       IF (idWrite) THEN
          WRITE(6, 10)
          WRITE(6, '(a)') " HDGM combined to a Schwarz algorithm "
          WRITE(6, 10) 
       ENDIF
    CASE(2)
       IF (idWrite) THEN
          WRITE(6, 10)
          WRITE(6, '(a)') " HDGM combined to a Schwarz algorithm accelerated by a "
          WRITE(6, '(a)') " Krylov subspace method "
          WRITE(6, 10)
       ENDIF
    CASE DEFAULT
       IF (idWrite) THEN
          WRITE(6, 10)
          WRITE(6, '(a,i1)') " Unknown hdg method: hdgMethod = ", hdgMethod
          WRITE(6, '(a)')    " Available methods are:"
          WRITE(6, '(a)')    " - hdgMethod = 0 for classical hdgm "
          WRITE(6, '(a)')    " - hdgMethod = 1 for hdgm combined to a Schwarz &
               algorithm "
          WRITE(6, '(a)')    " - hdgMethod = 2 for hdgm combined to a Schwarz &
               algorithm accelerated by a Krylov subspace method "
       ENDIF
       CALL TILT
    END SELECT
    !
    !---- Testing bad input for solverPack
    SELECT CASE(TRIM(solverPack))
    CASE("mumps")
    CASE("maphys")
    CASE DEFAULT
       IF (idWrite) THEN
          WRITE(6, '(a)') "Unknown solver package: ", TRIM(solverPack)
          WRITE(6, '(a)') "Available solver packages are:                        "
          WRITE(6, '(a)') " - mumps  for direct solver MUMPS                     "
          WRITE(6, '(a)') " - maphys for hybrid (direct/iterative) solver MaPHyS "
       ENDIF
       CALL TILT
    END SELECT
    !
    !---- Testing bad input for polyBase
    SELECT CASE(polyBase)
    CASE("L")
    CASE DEFAULT
       IF (idWrite) THEN
          WRITE(6, '(a)') "Unknown polynomial basis: ", polyBase
          WRITE(6, '(a)') "Available basis are: L for Lagrange polynomial &
               basis"
          WRITE(6, '(a)') "Please refer to manual for more informations"
       ENDIF
       CALL TILT
    END SELECT
    !
    !---- Checking minimal and maximal orders
    IF (orderMax < orderMin) THEN
       IF (idWrite) THEN
          WRITE(6, '(a)')    "Maximal polynomial order must be higher than &
               minimal polynomial order"
          WRITE(6, '(a)')    "I will set that for you"
       ENDIF
       tmpOrder = orderMin
       orderMin = orderMax
       orderMax = tmpOrder
       IF (idWrite) THEN
          WRITE(6, '(a,i1)') "Minimal order order is now: ", orderMin
          WRITE(6, '(a,i1)') "Maximal order order is now: ", orderMax
       ENDIF
    ENDIF
    !
    IF (orderMax > 4) THEN
       IF (idWrite) THEN
          WRITE(6, '(a)')    "Polynomial orders are out of range"
          WRITE(6, '(a)')    "Minimal value is 0 and maximal value is 4"
          WRITE(6, '(a)')    "I will set that for you"
       ENDIF
       orderMax = 4
       IF (orderMin > 4) THEN
          orderMin = 4
       ENDIF
       IF (idWrite) THEN
          WRITE(6, '(a,i1)') "Minimal order order is now: ", orderMin
          WRITE(6, '(a,i1)') "Maximal order order is now: ", orderMax
       ENDIF
    ENDIF
    !------------------------------------------------------------------
10  FORMAT(60('='))
  END SUBROUTINE READ_DATASOLVER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_DATASOURCE
    !------------------------------------------------------------------
    !---- Read DataSource in dataHorse and check for incompatibilities
    !---- - The name attribute of the Mesh structure
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE GAUSSIAN
    USE MATH_PHYS_CONSTANTS
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE PLANE_WAVE
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readChain

    LOGICAL(C_BOOL) :: sourceError
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    sourceError = .FALSE.
    !------------------------------------------------------------------
    !---- Read meshName
    !------------------------------------------------------------------
    OPEN(UNIT=dataUnit, FILE=dataFile, STATUS="UNKNOWN")
    !
    DO WHILE (INDEX(readChain,'#DataSource')==0)
       READ(dataUnit,'(a)') readChain 
    ENDDO
    !
    READ(dataUnit,*) sourceType
    !
    !---- Read source-specific informations
    !---- Now we read only the source part that we need
    SELECT CASE(sourceType)
       !
       !---- Plane wave
    CASE("PW")
       !
       REWIND(dataUnit)
       DO WHILE (INDEX(readChain,'#PW')==0)
          READ(dataUnit,'(a)') readChain
       ENDDO
       !
       READ(dataUnit,*) thetaPW, phiPW
       CALL DEGREES_TO_RADIANS(thetaPW)
       CALL DEGREES_TO_RADIANS(phiPW)
       READ(dataUnit,*) ampliPW
       READ(dataUnit,*) freqMinPW
       freqMaxPW = freqMinPW
       freq      = freqMinPW
       READ(dataUnit,*) polarPW(:)
       READ(dataUnit,*) computeErrPW
       !
       !---- Gaussian current source term
    CASE("GA")
       !
       REWIND(dataUnit)
       DO WHILE (INDEX(readChain,'#GA')==0)
          READ(dataUnit,'(a)') readChain
       ENDDO
       !
       READ(dataUnit,*) amplGauss
       READ(dataUnit,*) freq
       READ(dataUnit,*) centerGauss(:)
       READ(dataUnit,*) sigmaGauss(:)
    CASE DEFAULT
       sourceError = .TRUE.
    END SELECT
    !
    CLOSE(dataUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    IF ((sourceError) .AND. idWrite) THEN
       WRITE(6, '(a)') "Unknown source type: ", sourceType
       WRITE(6, '(a)') "Available sources are: "
       WRITE(6, '(a)') "PW (plane wave)"
       WRITE(6, '(a)') "GA (gaussian)"
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE READ_DATASOURCE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_DATAOBSERVABLE
    !------------------------------------------------------------------
    !---- Read DataObservable in dataHorse
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    CHARACTER(LEN=1000, KIND=C_CHAR) :: readChain 
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read observables
    !------------------------------------------------------------------
    OPEN(UNIT=dataUnit, FILE=dataFile, STATUS="UNKNOWN")
    !
    !---- Enables observables
    DO WHILE (INDEX(readChain,'#DataObservable')==0)
       READ(dataUnit,'(a)') readChain 
    ENDDO
    READ(dataUnit,*) enableObs
    !
    !---- RCS (Radar Cross Section)    
    DO WHILE (INDEX(readChain,'#RadarCrossSection')==0)
       READ(dataUnit,'(a)') readChain  
    ENDDO
    READ(dataUnit,*) computeRCS
    READ(dataUnit,*) ThetaInitRcs, ThetaEndRCS, nThetaRCS
    READ(dataUnit,*) PhiInitRcs,   PhiEndRCS,   nPhiRCS
    !
    !----     
    DO WHILE (INDEX(readChain,'#3dFrequencyDomain')==0)
       READ(dataUnit,'(a)') readChain  
    ENDDO
    READ(dataUnit,*) freq3dView
    !
    CLOSE(dataUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE READ_DATAOBSERVABLE

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE INPUT_DATA
