MODULE MAPHYS_STRUCTURES

  !====================================================================
  ! Everything for using MaPHyS as global solver in terms of DoFs of 
  ! the hybrid variable in hdgm
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  USE ZMPH_maphys_mod
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! MaPHyS data structure
  !--------------------------------------------------------------------
  TYPE(ZMPH_maphys_t) :: mphs
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Array for local index in MaPHyS (matrix and rhs)
  !--------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: indMtxLoc
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! MaPHyS variables and arrays for distributed interface
  !--------------------------------------------------------------------
  INTEGER(C_INT) :: mynbvi, myndof, mysizeIntrf

  INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: myindexintrf
  INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: myindexVi
  INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: myinterface
  INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: myptrindexVi
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_MAPHYS_SOLVER
    !------------------------------------------------------------------
    !---- Compute MaPHyS solver in terms of DoFs of the hybrid variable 
    !---- in hdgm
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of MaPHyS control parameters and data structures
    !------------------------------------------------------------------
    CALL MAPHYS_INIT
    !------------------------------------------------------------------
    !---- MaPHyS data structures for the distributed interface
    !------------------------------------------------------------------
    CALL MAPHYS_DISTRIBUTED_INTERFACE
    !------------------------------------------------------------------
    !---- Assembling the local matrix in MaPHyS package
    !------------------------------------------------------------------
    CALL MAPHYS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- Assembling the local RHS in MaPHyS package
    !------------------------------------------------------------------
    CALL MAPHYS_ASSEMBLING_RHS
    !------------------------------------------------------------------
    !---- MaPHyS analysis step
    !------------------------------------------------------------------
    CALL MAPHYS_ANALYSIS
    !------------------------------------------------------------------
    !---- MaPHyS factorization step
    !------------------------------------------------------------------
    CALL MAPHYS_FACTORIZATION
    !------------------------------------------------------------------
    !---- MaPHyS preconditioning step
    !------------------------------------------------------------------
    CALL MAPHYS_PRECONDITIONING
    !------------------------------------------------------------------
    !---- MaPHyS solve step
    !------------------------------------------------------------------
    CALL MAPHYS_SOLVE
    !------------------------------------------------------------------
    !---- Destroy the instance (deallocate MaPHyS data structures)
    !------------------------------------------------------------------
    CALL MAPHYS_END
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_MAPHYS_SOLVER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_INIT
    !------------------------------------------------------------------
    !---- Initialization of MaPHyS control parameters and data structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, iCell1, iCell2, iFac, nDof, nNeighb1,    &
         nNeighb2, nnz, ordMtx, sum1, sum2

    REAL(C_DOUBLE) :: precision
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Order of the local matrix
    !------------------------------------------------------------------
    ordMtx = 0
    DO facNum=1,Mesh%nFac
       ordMtx = ordMtx + 2*Lambda(facNum)%nDof
    ENDDO
    !------------------------------------------------------------------
    !---- Number of non-zero entries in the local matrix
    !------------------------------------------------------------------
    nnz = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum   = MeshFacI(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1) - nDof
       !
       iCell2   = MeshFac(facNum)%neighb(2)
       nNeighb2 = MeshCell(iCell2)%nNeighb
       sum2     = hdgEM(iCell2)%sumDofF(nNeighb2) - nDof
       !
       nnz      = nnz + 4*nDof*(nDof + sum1 + sum2)
    ENDDO
    !
    !---- Boundary faces (physical and artificial)
    DO iFac=1,Mesh%nFacB
       facNum = MeshFacB(iFac)%faceNum
       nDof   = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1)
       !
       nnz = nnz + 4*nDof*sum1
    ENDDO
    !------------------------------------------------------------------
    !---- Sets up MaPHyS control parameters
    !------------------------------------------------------------------
    !
    !---- Control of parallelism
    mphs%COMM = defaultComm
    !
    !---- Initialize the instance
    mphs%JOB = -1
    CALL ZMPH_maphys_driver(mphs)
    !
    !---- Matrix type
    mphs%SYM = 0
    !
    !---- Control printing informations
    mphs%ICNTL(4)  = 5 !3 
    mphs%ICNTL(5)  = 1 !0 
    mphs%ICNTL(6)  = 2 !0 
    !
    !---- Control the sparse direct solver
    !---- 1        => MUMPS
    !---- 2        => PaSTiX
    !---- 3        => use multiple sparse direct solvers (see ICNTL(15), ICNTL(32))
    !---- Default  => the first available starting from index 1
    mphs%ICNTL(13) = 1
    !
    !---- Control the sparse direct solver for computing the preconditionner
    !---- 1        => MUMPS
    !---- 2        => PaSTiX
    !---- Default  => value of ICNTL(13)
    mphs%ICNTL(15) = 1
    !
    !---- Control iterative solver
    !---- 1        => GMRES
    !---- 2        => CG
    !---- 3        => CG if SPD, GMRES otherwise
    !---- Default  => 3
    mphs%ICNTL(20) = 3
    !
    !---- Control preconditionning technic for the Schur
    !---- 1        => dense facto of the assembled local Schur
    !---- 2        => sparse facto of the assembled local Schur 
    !                 (specify the threshold RCNTL(11))
    !---- 3        => facto of a sparse matrix 
    !                 (specify the threshold RCNTL(9))
    !---- 4        => no preconditionner
    !---- 5        => autodetect (order of preference 1 > 2 > 3 > 4)
    !---- Default  => 2
    mphs%ICNTL(21) = 5
    !
    mphs%ICNTL(22) = 3
    mphs%ICNTL(24) = maxIter
    mphs%ICNTL(25) = 0
    mphs%ICNTL(26) = maxIter
    mphs%ICNTL(27) = 0
    mphs%ICNTL(32) = 2
    !
    !---- Stopping criterion threshold in the iterative solver (default = 1.e-5)
    SELECT CASE(thresholdIter)
    CASE("max")
       precision      = (globalMesh%hMax)**(genOrderMax + 2)
    CASE("avg")
       precision      = (globalMesh%hAvg)**(genOrderMax + 2)
    CASE("min")
       precision      = (globalMesh%hMin)**(genOrderMax + 2)
    END SELECT
    IF (precision < 1.0e-15) precision = 1.0e-15 
    mphs%RCNTL(21) = precision
    !
    !---- Level of parallelism
    OPEN(UNIT=1, FILE='./dataMaPHyS', STATUS='UNKNOWN')
    READ(1, *) mphs%ICNTL(36) ! specify how to bind thread inside MaPHyS
    READ(1, *) mphs%ICNTL(37) ! specify the number of nodes
    READ(1, *) mphs%ICNTL(38) ! specify the number of cores per node
    READ(1, *) mphs%ICNTL(39) ! specify the number of threads per domain
    READ(1, *) mphs%ICNTL(40) ! specify the number of domains
    READ(1, *) mphs%ICNTL(42) ! control parameter that active the 2 levels of parallelism version
    CLOSE(1)
    !
    !---- Control the input system
    !---- 1        => centralized on the host
    !---- 2        => distributed
    !---- Default  => 1
    mphs%ICNTL(43) = 2
    !
    !---- Control the local output after analysis
    !---- This option is unactivated by default
    !---- 1        => 1 performs a dump of the local matrices, RHS and 
    !                 domain data after analysis (the files will be 
    !                 output in the working directory)
    ! mphs%ICNTL(45) = 1
    !------------------------------------------------------------------
    !---- Characterization of the local problem
    !------------------------------------------------------------------
    mphs%N   = ordMtx
    mphs%NNZ = nnz
    !------------------------------------------------------------------
    !---- Allocations of data structures for the definition of the local 
    !---- Matrix and the local RHS
    !------------------------------------------------------------------
    ALLOCATE(mphs%ROWS  (mphs%NNZ))
    ALLOCATE(mphs%COLS  (mphs%NNZ))
    ALLOCATE(mphs%VALUES(mphs%NNZ))
    ALLOCATE(mphs%RHS   (mphs%N  ))
    ALLOCATE(mphs%SOL   (mphs%N  ))
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_INIT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_DISTRIBUTED_INTERFACE
    !------------------------------------------------------------------
    !---- Initialization of MaPHyS data structures for the distributed 
    !---- interface + creating the local domain instances for MaPHyS 
    !---- distributed interface
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, i, id, iFac, ind, iNeighb, indLocal,     &
         indStart, j, k, nbFacNeighbMax,  nDof, preDof, preFac

    INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: indGlb
    INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: indRecv
    INTEGER(C_INT), DIMENSION(:),   ALLOCATABLE :: indSend
    INTEGER(C_INT), DIMENSION(:,:), ALLOCATABLE :: indArt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute MaPHyS data structure (distributed inteface) for inner 
    !---- and physical boundary faces
    !------------------------------------------------------------------
    ALLOCATE(indMtxLoc(Mesh%nFac))
    !
    !---- Inner faces
    indMtxLoc(MeshFacI(1)%faceNum) = 0
    preFac = MeshFacI(1)%faceNum
    preDof = Lambda(preFac)%nDof
    DO iFac=2,Mesh%nFacI
       facNum            = MeshFacI(iFac)%faceNum
       indMtxLoc(facNum) = indMtxLoc(preFac) + 2*preDof
       preFac            = facNum
       preDof            = Lambda(facNum)%nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum            = MeshFacPhy(iFac)%faceNum
       indMtxLoc(facNum) = indMtxLoc(preFac) + 2*preDof
       preFac            = facNum
       preDof            = Lambda(facNum)%nDof
    ENDDO
    !------------------------------------------------------------------
    !---- Compute MaPHyS data structure (distributed interface)  
    !---- + local and global indices of interfaces i.e. artificial 
    !---- boundary faces 
    !------------------------------------------------------------------
    ALLOCATE(indGlb(Mesh%nFac))
    DO i=1,Mesh%nFac
       indGlb(i) = 0
    ENDDO
    mynbvi = nbNeighb
    ALLOCATE(myptrindexVi(mynbvi + 1))
    ALLOCATE(myindexVi(mynbvi)        )
    DO iNeighb=1,mynbvi
       myindexVi(iNeighb) = neighb(iNeighb)%id
    ENDDO
    !
    myndof = 0
    DO facNum=1,Mesh%nFac
       myndof = myndof + 2*Lambda(facNum)%nDof
    ENDDO
    !
    mysizeIntrf = 0
    DO iFac=1,Mesh%nFacArt
       facNum      = MeshFacArt(iFac)%faceNum
       mysizeIntrf = mysizeIntrf + 2*Lambda(facNum)%nDof
    ENDDO
    ALLOCATE(myinterface(mysizeIntrf) )
    ALLOCATE(myindexintrf(mysizeIntrf))
    !
    indStart = 0
    DO iNeighb=1,nbNeighb
       DO iFac=1,neighb(iNeighb)%nbFac
          facNum            = neighb(iNeighb)%faces(iFac)
          indStart          = indStart + 2*Lambda(facNum)%nDof
          indMtxLoc(facNum) = myndof - mysizeIntrf + indStart - 2*Lambda(facNum)%nDof
       ENDDO
    ENDDO
    !
    nbFacNeighbMax = 0
    DO iNeighb=1,nbNeighb
       nbFacNeighbMax = MAX(nbFacNeighbMax,neighb(iNeighb)%nbFac)
    ENDDO
    ALLOCATE(indArt(nbNeighb,nbFacNeighbMax))
    ALLOCATE(indSend(nbFacNeighbMax)        )
    ALLOCATE(indRecv(nbFacNeighbMax)        )
    indLocal = 0
    DO iNeighb=1,nbNeighb
       IF (idProc < neighb(iNeighb)%id) THEN
          DO iFac=1,neighb(iNeighb)%nbFac
             facNum   = neighb(iNeighb)%faces(iFac)
             indLocal = indLocal + 2*Lambda(facNum)%nDof
          ENDDO
       ENDIF
    ENDDO
    CALL MPI_SCAN(indLocal, ind, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    indStart = ind - indLocal
    DO iNeighb=1,nbNeighb
       IF (idProc < neighb(iNeighb)%id) THEN
          id = 1
          DO iFac=1,neighb(iNeighb)%nbFac
             facNum               = neighb(iNeighb)%faces(iFac)
             indStart             = indStart + 2*Lambda(facNum)%nDof
             indArt(iNeighb,iFac) = indStart - 2*Lambda(facNum)%nDof
             indSend(id)          = indArt(iNeighb,iFac)
             indGlb(facNum)       = indArt(iNeighb,iFac)
             id                   = id + 1
          ENDDO
          CALL MPI_SEND(indSend, neighb(iNeighb)%nbFac, MPI_INTEGER,   &
               neighb(iNeighb)%id, 100, defaultComm, mpiError)
       ELSE
          CALL MPI_RECV(indRecv, neighb(iNeighb)%nbFac, MPI_INTEGER,   &
               neighb(iNeighb)%id, 100, defaultComm, status, mpiError)
          id = 1
          DO iFac=1,neighb(iNeighb)%nbFac
             facNum               = neighb(iNeighb)%faces(iFac)
             indArt(iNeighb,iFac) = indRecv(id)
             indGlb(facNum)       = indArt(iNeighb,iFac)
             id                   = id + 1
          ENDDO
       ENDIF
    ENDDO
    DEALLOCATE(indArt )
    DEALLOCATE(indSend)
    DEALLOCATE(indRecv)
    k = 1
    j = 1
    myptrindexVi(j) = 1
    DO iNeighb=1,nbNeighb
       DO iFac=1,neighb(iNeighb)%nbFac
          facNum = neighb(iNeighb)%faces(iFac)
          nDof   = Lambda(facNum)%nDof
          DO i=1,nDof
             myindexintrf(k)     = k
             myinterface(k)      = indGlb(facNum) + i
             myindexintrf(k + 1) = k + 1
             myinterface(k + 1)  = myinterface(k) + nDof
             k                   = k + 2
          ENDDO
       ENDDO
       j = j + 1
       myptrindexVi(j) = k
    ENDDO
    DEALLOCATE(indGlb)
    !------------------------------------------------------------------
    !---- Creating the local domain instances for MaPHyS distributed 
    !---- interface 
    !------------------------------------------------------------------
    CALL ZMPH_distributed_interface(mphs, myndof, mysizeIntrf, myinterface, &
         mynbvi, myindexVi, myptrindexVi, myindexintrf)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_DISTRIBUTED_INTERFACE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- Assembling the local matrix of the global linear system to 
    !---- solve in terms of DoFs of the hybrid variable
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_MATRICES_STRUCTURES
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iCell, iCell1, iCell11, iCell12, iCell2,      &
         iCell21, iCell22, iFac, indCol1, indCol2, indLoc, indRow1,    &
         indRow2, j, jCell11, jCell12, jCell21, jCell22, k, k1, k2,    &
         kFac, nDof, nDofFk, facNum, nNeighb1, nNeighb2, sum1, sum2,   &
         sumk, iError

    COMPLEX(C_DOUBLE_COMPLEX) :: valMtx
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Assembling the local matrix
    !------------------------------------------------------------------
    indLoc = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum   = MeshFacI(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       k1       = MeshFac(facNum)%numLocal(1)
       sum1     = hdgEM(iCell1)%sumDofF(k1) - nDof
       !
       iCell2   = MeshFac(facNum)%neighb(2)
       nNeighb2 = MeshCell(iCell2)%nNeighb
       k2       = MeshFac(facNum)%numLocal(2)
       sum2     = hdgEM(iCell2)%sumDofF(k2) - nDof
       !
       DO i=1,nDof
          indRow1 = indMtxLoc(facNum) + i
          indRow2 = indRow1 + nDof
          iCell11 = 2*sum1  + i
          iCell12 = iCell11 + nDof
          iCell21 = 2*sum2  + i
          iCell22 = iCell21 + nDof
          !
          !---- The diagonal blocks of Ke(iCell1) and Ke(iCell2)
          DO j=1,nDof
             indCol1 = indMtxLoc(facNum) + j
             indCol2 = indCol1 + nDof
             jCell11 = 2*sum1  + j
             jCell12 = jCell11 + nDof
             jCell21 = 2*sum2  + j
             jCell22 = jCell21 + nDof
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell11,jCell11) + Ke(iCell2)%tab(iCell21,jCell21)
             mphs%ROWS  (indLoc) = indRow1
             mphs%COLS  (indLoc) = indCol1
             mphs%VALUES(indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell11,jCell12) + Ke(iCell2)%tab(iCell21,jCell22)
             mphs%ROWS  (indLoc) = indRow1
             mphs%COLS  (indLoc) = indCol2
             mphs%VALUES(indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell12,jCell11) + Ke(iCell2)%tab(iCell22,jCell21)
             mphs%ROWS  (indLoc) = indRow2
             mphs%COLS  (indLoc) = indCol1
             mphs%VALUES(indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell12,jCell12) + Ke(iCell2)%tab(iCell22,jCell22)
             mphs%ROWS  (indLoc) = indRow2
             mphs%COLS  (indLoc) = indCol2
             mphs%VALUES(indLoc) = valMtx
          ENDDO
          !
          !---- The off-diagonal blocks of Ke(iCell1)
          DO k=1,nNeighb1
             IF (k /= k1) THEN
                kFac = MeshCell(iCell1)%facNum(k)
                nDofFk  = hdgEM(iCell1)%nDofF(k)
                sumk    = hdgEM(iCell1)%sumDofF(k) - nDofFk
                DO j=1,nDofFk
                   indCol1 = indMtxLoc(kFac) + j
                   indCol2 = indCol1 + nDofFk
                   jCell11 = 2*sumk  + j
                   jCell12 = jCell11 + nDofFk
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell11,jCell11)
                   mphs%ROWS  (indLoc) = indRow1
                   mphs%COLS  (indLoc) = indCol1
                   mphs%VALUES(indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell11,jCell12)
                   mphs%ROWS  (indLoc) = indRow1
                   mphs%COLS  (indLoc) = indCol2
                   mphs%VALUES(indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell12,jCell11)
                   mphs%ROWS  (indLoc) = indRow2
                   mphs%COLS  (indLoc) = indCol1
                   mphs%VALUES(indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell12,jCell12)
                   mphs%ROWS  (indLoc) = indRow2
                   mphs%COLS  (indLoc) = indCol2
                   mphs%VALUES(indLoc) = valMtx
                ENDDO
             ENDIF
          ENDDO
          !
          !---- The off-diagonal blocks of Ke(iCell2)
          DO k=1,nNeighb2
             IF (k /= k2) THEN
                kFac = MeshCell(iCell2)%facNum(k)
                nDofFk  = hdgEM(iCell2)%nDofF(k)
                sumk    = hdgEM(iCell2)%sumDofF(k) - nDofFk
                DO j=1,nDofFk
                   indCol1 = indMtxLoc(kFac) + j
                   indCol2 = indCol1 + nDofFk
                   jCell21 = 2*sumk  + j
                   jCell22 = jCell21 + nDofFk
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell21,jCell21)
                   mphs%ROWS  (indLoc) = indRow1
                   mphs%COLS  (indLoc) = indCol1
                   mphs%VALUES(indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell21,jCell22)
                   mphs%ROWS  (indLoc) = indRow1
                   mphs%COLS  (indLoc) = indCol2
                   mphs%VALUES(indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell22,jCell21)
                   mphs%ROWS  (indLoc) = indRow2
                   mphs%COLS  (indLoc) = indCol1
                   mphs%VALUES(indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell22,jCell22)
                   mphs%ROWS  (indLoc) = indRow2
                   mphs%COLS  (indLoc) = indCol2
                   mphs%VALUES(indLoc) = valMtx
                ENDDO
             ENDIF
          ENDDO
       ENDDO
    ENDDO
    !
    !---- Boundary faces
    DO iFac=1,Mesh%nFacB
       facNum   = MeshFacB(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       k1       = MeshFac(facNum)%numLocal(1)
       sum1     = hdgEM(iCell1)%sumDofF(k1) - nDof
       !
       DO i=1,nDof
          indRow1 = indMtxLoc(facNum) + i
          indRow2 = indRow1 + nDof
          iCell11 = 2*sum1  + i
          iCell12 = iCell11 + nDof
          DO k=1,nNeighb1
             kFac = MeshCell(iCell1)%facNum(k)
             nDofFk  = hdgEM(iCell1)%nDofF(k)
             sumk    = hdgEM(iCell1)%sumDofF(k) - nDofFk
             DO j=1,nDofFk
                indCol1 = indMtxLoc(kFac) + j
                indCol2 = indCol1 + nDofFk
                jCell11 = 2*sumk  + j
                jCell12 = jCell11 + nDofFk
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell11,jCell11)
                mphs%ROWS  (indLoc) = indRow1
                mphs%COLS  (indLoc) = indCol1
                mphs%VALUES(indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell11,jCell12)
                mphs%ROWS  (indLoc) = indRow1
                mphs%COLS  (indLoc) = indCol2
                mphs%VALUES(indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell12,jCell11)
                mphs%ROWS  (indLoc) = indRow2
                mphs%COLS  (indLoc) = indCol1
                mphs%VALUES(indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell12,jCell12)
                mphs%ROWS  (indLoc) = indRow2
                mphs%COLS  (indLoc) = indCol2
                mphs%VALUES(indLoc) = valMtx
             ENDDO
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Deallocate Ke 
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       DEALLOCATE(Ke(iCell)%tab)
    ENDDO
    DEALLOCATE(Ke)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_ASSEMBLING_MATRIX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_ASSEMBLING_RHS
    !------------------------------------------------------------------
    !---- Assembling the local RHS of the global linear system to 
    !---- solve in terms of DoFs of the hybrid variable
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, i, iCell1, iCell2, iFac, indLocu, indLocw, &
         iu1, iu2, iw1, iw2, k1, k2, nDoFs, sum1, sum2 
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Assembling the local RHS
    !------------------------------------------------------------------
    DO i=1,mphs%N
       mphs%RHS(i) = (0.d0, 0.d0)
    ENDDO
    !
    IF (ALLOCATED(rhsAbs)) THEN
    DO iFac=1,Mesh%nFacAbs
       faceNum = MeshFacAbs(iFac)%faceNum
       nDoFs   = Lambda(faceNum)%nDof
       DO i=1,nDoFs
          indLocu           = indMtxLoc(faceNum) + i
          indLocw           = indLocu + nDoFs
          mphs%RHS(indLocu) = mphs%RHS(indLocu) + rhsAbs(iFac)%vec(i,1)
          mphs%RHS(indLocw) = mphs%RHS(indLocw) + rhsAbs(iFac)%vec(i,2)
       ENDDO
    ENDDO
    ENDIF
    !
    IF (ALLOCATED(rhsJglb)) THEN
       !
       !---- Inner faces
       DO iFac=1,Mesh%nFacI
          faceNum = MeshFacI(iFac)%faceNum
          nDoFs   = Lambda(faceNum)%nDof
          !
          iCell1  = MeshFac(faceNum)%neighb(1)
          k1      = MeshFac(faceNum)%numLocal(1)
          sum1    = hdgEM(iCell1)%sumDofF(k1) - nDoFs
          !
          iCell2  = MeshFac(faceNum)%neighb(2)
          k2      = MeshFac(faceNum)%numLocal(2)
          sum2    = hdgEM(iCell2)%sumDofF(k2) - nDoFs
          !
          DO i=1,nDoFs
             indLocu           = indMtxLoc(faceNum) + i
             indLocw           = indLocu + nDoFs
             iu1               = 2*sum1  + i
             iw1               = iu1 + nDoFs
             iu2               = 2*sum2  + i
             iw2               = iu2 + nDoFs
             mphs%RHS(indLocu) = mphs%RHS(indLocu)                     &
                  + rhsJglb(iCell1)%tab(iu1) + rhsJglb(iCell2)%tab(iu2)
             mphs%RHS(indLocw) = mphs%RHS(indLocw)                     &
                  + rhsJglb(iCell1)%tab(iw1) + rhsJglb(iCell2)%tab(iw2)
          ENDDO
       ENDDO
       !
       !---- Boundary faces and interfaces (i.e. physical and artificial faces)
       DO iFac=1,Mesh%nFacB
          faceNum = MeshFacB(iFac)%faceNum
          nDoFs   = Lambda(faceNum)%nDof
          !
          iCell1  = MeshFac(faceNum)%neighb(1)
          k1      = MeshFac(faceNum)%numLocal(1)
          sum1    = hdgEM(iCell1)%sumDofF(k1) - nDoFs
          !
          DO i=1,nDoFs
             indLocu           = indMtxLoc(faceNum) + i
             indLocw           = indLocu + nDoFs
             iu1               = 2*sum1  + i
             iw1               = iu1 + nDoFs
             mphs%RHS(indLocu) = mphs%RHS(indLocu) + rhsJglb(iCell1)%tab(iu1)
             mphs%RHS(indLocw) = mphs%RHS(indLocw) + rhsJglb(iCell1)%tab(iw1)
          ENDDO
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_ASSEMBLING_RHS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_ANALYSIS
    !------------------------------------------------------------------
    !---- MaPHyS analysis step
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Analysis
    !------------------------------------------------------------------
    mphs%JOB = 1
    CALL ZMPH_maphys_driver(mphs)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    iError = mphs%IINFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in analysis step of MaPHyS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_ANALYSIS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_FACTORIZATION
    !------------------------------------------------------------------
    !---- MaPHyS factorization step
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Factorization
    !------------------------------------------------------------------
    mphs%JOB = 2
    CALL ZMPH_maphys_driver(mphs)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    iError = mphs%IINFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in factorization step of MaPHyS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_FACTORIZATION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_PRECONDITIONING
    !------------------------------------------------------------------
    !---- MaPHyS preconditioning step
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES    
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Preconditionning
    !------------------------------------------------------------------
    mphs%JOB = 3
    CALL ZMPH_maphys_driver(mphs)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    iError = mphs%IINFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in preconditioning step of MaPHyS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_PRECONDITIONING

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_SOLVE
    !------------------------------------------------------------------
    !---- MaPHyS solve step
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE PARALLEL_STRUCTURES    
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, i, iError, indLocu, indLocw, nDof
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Solve
    !------------------------------------------------------------------
    mphs%JOB = 4
    CALL ZMPH_maphys_driver(mphs)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    iError = mphs%IINFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in solve step of MaPHyS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Distribution of the DoFs of the hybrid variable
    !------------------------------------------------------------------
    DO facNum=1,Mesh%nFac
       nDof = Lambda(facNum)%nDof
       DO i=1,nDof
          indLocu = indMtxLoc(facNum) + i
          indLocw = indLocu + nDof
          Lambda(facNum)%L%vec(i,1) = mphs%SOL(indLocu)
          Lambda(facNum)%L%vec(i,2) = mphs%SOL(indLocw)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_SOLVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MAPHYS_END
    !------------------------------------------------------------------
    !---- MaPHyS final step : deallocations
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Final step - Deallocations
    !------------------------------------------------------------------
    DEALLOCATE(mphs%ROWS   )
    DEALLOCATE(mphs%COLS   )
    DEALLOCATE(mphs%VALUES )
    DEALLOCATE(mphs%SOL    )
    DEALLOCATE(indMtxLoc   )
    DEALLOCATE(myindexintrf)
    DEALLOCATE(myindexVi   )
    DEALLOCATE(myinterface )
    DEALLOCATE(myptrindexVi)
    !------------------------------------------------------------------
    !---- Destroy the instance (deallocate internal data structures)
    !------------------------------------------------------------------
    mphs%JOB = -2
    CALL ZMPH_maphys_driver(mphs)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MAPHYS_END

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MAPHYS_STRUCTURES
