MODULE MAXWELL_STRUCTURES

  !====================================================================
  ! All the necessary pre-processing on numerical elements and faces
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_MAXWELL
    !------------------------------------------------------------------
    !---- Evaluate local polynomial order and allocate solution structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE BASES
    USE INPUT_DATA
    USE MESH_DATA
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialize polynomial basis
    !------------------------------------------------------------------
    CALL BASES_INITIALIZATION(polyBase)
    !------------------------------------------------------------------
    !---- Allocate numerical structure
    !------------------------------------------------------------------
    ALLOCATE(hdgEM(Mesh%nCell))
    ALLOCATE(Lambda(Mesh%nFac))
    !------------------------------------------------------------------
    !---- Fill eps and mu arrays
    !------------------------------------------------------------------
    CALL DISPATCH_MATERIALS
    !------------------------------------------------------------------
    !---- Evaluate local polynomial orders over the mesh
    !------------------------------------------------------------------
    CALL EVALUATE_POLYNOMIAL_ORDER
    !------------------------------------------------------------------
    !---- Allocate E, H, Lambda fields and local vectors usefull for 
    !---- computations
    !------------------------------------------------------------------
    CALL ALLOCATE_FIELDS
    !------------------------------------------------------------------
    !---- Point to right matrices (EM and Lambda fields)
    !------------------------------------------------------------------
    CALL DISPATCH_EM_MATRICES
    !------------------------------------------------------------------
    !---- Point to right permutations
    !------------------------------------------------------------------
    CALL DISPATCH_PERMUTATIONS
    !------------------------------------------------------------------
    !---- Print numerical informations
    !------------------------------------------------------------------
    CALL PRINT_NUMERICAL_STATS
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_MAXWELL

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE EVALUATE_POLYNOMIAL_ORDER
    !------------------------------------------------------------------
    !---- Evaluate local polynomial order and complete related structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE BASES
    USE INPUT_DATA
    USE MATH_PHYS_CONSTANTS
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_ELEMENTS
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE SPECIAL_FUNCTIONS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: cellNum, EffOrderMax, EffOrderMin, faceNum,      &
         iCell, iCell1, iCell2, iFac, iNeighb, iVert, k,               &
         nbFacNeighbMax, nDof, nDofS, nDofV, neighbor, order, orderH,  &
         order1, order2, tmpSum

    INTEGER(C_INT), DIMENSION(4)                :: pts
    INTEGER(C_INT), DIMENSION(0:4)              :: genPFacesA(0:4), nPFacesA(0:4)
    INTEGER(C_INT), DIMENSION(0:4)              :: genQFacesA(0:4), nQFacesA(0:4)
    INTEGER(C_INT), DIMENSION(:)  , ALLOCATABLE :: recvDeg, sendDeg

    REAL(C_DOUBLE) :: distMax, qdistMax, wavelength

    REAL(C_DOUBLE), DIMENSION(6)                :: distance
    REAL(C_DOUBLE), DIMENSION(3,4)              :: coorVert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Fill information for tetrahedra and triangular faces (EM field)
    !---- Local distribution of interpolation orders (as function of 
    !---- the minimal number of points per wavelength)
    !------------------------------------------------------------------
    !
    !---- Number of pts per edge (function of the interpolation order)
    pts(1) = 2
    pts(2) = 3
    pts(3) = 4
    pts(4) = 5
    !
    EffOrderMin = 10
    EffOrderMax = -1
    DO iCell=1,Mesh%nCellTet
       cellNum           = MeshCellTet(iCell)%cellNum
       hdgEM(cellNum)%order =-1
       !
       !---- Compute the max. distance between vertices of the hexahedron
       distMax  =-1.0e+20
       DO iVert = 1,4
          coorVert(:,iVert) = MeshVert(MeshCell(cellNum)%vertex(iVert))%coor(:)
       ENDDO
       CALL COMPUTE_R3_DISTANCE(coorVert(:,1),coorVert(:,2),distance(1))
       CALL COMPUTE_R3_DISTANCE(coorVert(:,1),coorVert(:,3),distance(2))
       CALL COMPUTE_R3_DISTANCE(coorVert(:,1),coorVert(:,4),distance(3))
       CALL COMPUTE_R3_DISTANCE(coorVert(:,2),coorVert(:,3),distance(4))
       CALL COMPUTE_R3_DISTANCE(coorVert(:,2),coorVert(:,4),distance(5))
       CALL COMPUTE_R3_DISTANCE(coorVert(:,3),coorVert(:,4),distance(6))
       DO k=1,6
          distMax = MAX(distMax, distance(k))
       ENDDO
       qdistMax = 1.d0/distMax
       !
       !---- Evaluation of the tetrahedron interpolation order as function 
       !---- of the minimal nb of pts per wavelength
       wavelength = c0/(SQRT(hdgEM(cellNum)%eps*hdgEM(cellNum)%mu)*freq)
       do_Pk : DO k=orderMin,orderMax
          IF (pts(k)*wavelength*qdistMax >= ptsWaveLength) THEN
             hdgEM(cellNum)%order = k
             CALL ORDER_TO_NDOFV(hdgEM(cellNum)%order,"TET1",nDofV)
             hdgEM(cellNum)%nDofV = nDofV
             CALL ORDER_TO_NDOFS(hdgEM(cellNum)%order,"TRI1",nDofS)
             hdgEM(cellNum)%nDofS = nDofS
             EXIT do_Pk
          ENDIF
       ENDDO do_Pk
       IF (hdgEM(cellNum)%order == -1) THEN
          hdgEM(cellNum)%order = orderMax
          CALL ORDER_TO_NDOFV(hdgEM(cellNum)%order,"TET1",nDofV)
          hdgEM(cellNum)%nDofV = nDofV
          CALL ORDER_TO_NDOFS(hdgEM(cellNum)%order,"TRI1",nDofS)
          hdgEM(cellNum)%nDofS = nDofS
       ENDIF
       !
       !---- Update effective min. and max. orders
       EffOrderMin = MIN(hdgEM(cellNum)%order,EffOrderMin)
       EffOrderMax = MAX(hdgEM(cellNum)%order,EffOrderMax)
    ENDDO
    !
    !---- Effective min. and max. order for tetrahedra
    orderMin = EffOrderMin
    orderMax = EffOrderMax
    !------------------------------------------------------------------
    !---- Fill information for hexahedra and quadrangular faces (EM field)
    !---- Local distribution of interpolation orders (as function of 
    !---- the minimal number of points per wavelength)
    !------------------------------------------------------------------
    EffOrderMin = 10
    EffOrderMax = -1
    DO iCell=1,Mesh%nCellHex
       cellNum           = MeshCellHex(iCell)%cellNum
       hdgEM(cellNum)%order =-1
       !
       !---- Compute the max. distance between vertices of the hexahedron
       coorVert(:,1) = MeshVert(MeshCell(cellNum)%vertex(1))%coor(:)
       coorVert(:,2) = MeshVert(MeshCell(cellNum)%vertex(7))%coor(:)
       CALL COMPUTE_R3_DISTANCE(coorVert(:,1),coorVert(:,2), distMax)
       qdistMax      = 1.d0/distMax
       !
       !---- Evaluation of the hexahedron interpolation order as function 
       !---- of the minimal nb of pts per wavelength
       wavelength = c0/(SQRT(hdgEM(cellNum)%eps*hdgEM(cellNum)%mu)*freq)
       do_Qk : DO k=orderMinH,orderMaxH
          IF (pts(k)*wavelength*qdistMax >= ptsWaveLength) THEN
             hdgEM(cellNum)%order = k
             CALL ORDER_TO_NDOFV(hdgEM(cellNum)%order,"HEX1",nDofV)
             hdgEM(cellNum)%nDofV = nDofV
             CALL ORDER_TO_NDOFS(hdgEM(cellNum)%order,"QUA1",nDofS)
             hdgEM(cellNum)%nDofS = nDofS
             EXIT do_Qk
          ENDIF
       ENDDO do_Qk
       IF (hdgEM(cellNum)%order == -1) THEN
          hdgEM(cellNum)%order = orderMaxH
          CALL ORDER_TO_NDOFV(hdgEM(cellNum)%order,"HEX1",nDofV)
          hdgEM(cellNum)%nDofV = nDofV
          CALL ORDER_TO_NDOFS(hdgEM(cellNum)%order,"QUA1",nDofS)
          hdgEM(cellNum)%nDofS = nDofS
       ENDIF
       !
       !---- Update effective min. and max. orders
       EffOrderMin = MIN(hdgEM(cellNum)%order,EffOrderMin)
       EffOrderMax = MAX(hdgEM(cellNum)%order,EffOrderMax)
    ENDDO
    !
    !---- Effective min. and max. orders for hexahedra
    orderMinH = EffOrderMin
    orderMaxH = EffOrderMax
    !------------------------------------------------------------------
    !---- Minimal and maximal number of DoFs for volume and surface
    !------------------------------------------------------------------
    nDofVMin  = 1000
    nDofVMinH = 1000
    nDofSMin  = 1000
    nDofSMinH = 1000
    nDofVMax  = 0
    nDofVMaxH = 0
    nDofSMax  = 0
    nDofSMaxH = 0
    !
    IF (Mesh%nCellTet /= 0) THEN
       CALL ORDER_TO_NDOFV(orderMin,"TET1",nDofVMin)
       CALL ORDER_TO_NDOFV(orderMax,"TET1",nDofVMax)
       CALL ORDER_TO_NDOFS(orderMin,"TRI1",nDofSMin)
       CALL ORDER_TO_NDOFS(orderMax,"TRI1",nDofSMax)
    ENDIF
    !
    IF (Mesh%nCellHex /= 0) THEN
       CALL ORDER_TO_NDOFV(orderMinH,"HEX1",nDofVMinH)
       CALL ORDER_TO_NDOFV(orderMaxH,"HEX1",nDofVMaxH)
       CALL ORDER_TO_NDOFS(orderMinH,"QUA1",nDofSMinH)
       CALL ORDER_TO_NDOFS(orderMaxH,"QUA1",nDofSMaxH)
    ENDIF
    !
    nDofVmin = MIN(nDofVMin,nDofVMinH)
    nDofSmin = MIN(nDofSMin,nDofSMinH)
    nDofVmax = MAX(nDofVMax,nDofVMaxH)
    nDofSmax = MAX(nDofSMax,nDofSMaxH)
    !------------------------------------------------------------------
    !---- Fill information for faces (Lambda field)
    !------------------------------------------------------------------
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       faceNum = MeshFacI(iFac)%faceNum
       iCell1  = MeshFac(faceNum)%neighb(1)
       order1  = hdgEM(iCell1)%order
       iCell2  = MeshFac(faceNum)%neighb(2)
       order2  = hdgEM(iCell2)%order
       IF (MeshFac(faceNum)%isHybrid) THEN
          IF (MeshCell(iCell1)%cellType == "TET1") THEN
             Lambda(faceNum)%order = order1
          ELSE
             Lambda(faceNum)%order = order2
          ENDIF
       ELSE
          Lambda(faceNum)%order = MAX(order1,order2)
       ENDIF
       CALL ORDER_TO_NDOFS(Lambda(faceNum)%order,MeshFac(faceNum)%faceType,nDof)
       Lambda(faceNum)%nDof  = nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       faceNum               = MeshFacPhy(iFac)%faceNum
       iCell1                = MeshFac(faceNum)%neighb(1)
       order1                = hdgEM(iCell1)%order
       Lambda(faceNum)%order = order1
       CALL ORDER_TO_NDOFS(Lambda(faceNum)%order,MeshFac(faceNum)%faceType,nDof)
       Lambda(faceNum)%nDof  = nDof
    ENDDO
    !
    !---- Interfaces, i.e. artificial boundary faces
    nbFacNeighbMax = 0
    DO iNeighb=1,nbNeighb
       nbFacNeighbMax = MAX(nbFacNeighbMax,neighb(iNeighb)%nbFac)
    ENDDO
    !
    ALLOCATE(recvDeg(nbFacNeighbMax))
    ALLOCATE(sendDeg(nbFacNeighbMax))
    DO iNeighb=1,nbNeighb
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum       = neighb(iNeighb)%faces(iFac)
          iCell         = MeshFac(faceNum)%neighb(1)
          sendDeg(iFac) = hdgEM(iCell)%order
       ENDDO
       !
       !---- Send/Recv interpolation orders of the neighboring cells on
       !---- interfaces
       CALL MPI_SENDRECV(sendDeg, neighb(iNeighb)%nbFac, MPI_INTEGER,  &
            neighb(iNeighb)%id, 1, recvDeg, neighb(iNeighb)%nbFac,     &
            MPI_INTEGER, neighb(iNeighb)%id, 1, defaultComm, status,   &
            mpiError)
       !
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum = neighb(iNeighb)%faces(iFac)
          iCell   = MeshFac(faceNum)%neighb(1)
          order   = hdgEM(iCell)%order
          IF (MeshFac(faceNum)%isHybrid) THEN
             IF (MeshCell(iCell)%cellType == "TET1") THEN
                Lambda(faceNum)%order = order
             ELSE
                Lambda(faceNum)%order = recvDeg(iFac)
             ENDIF
          ELSE
             Lambda(faceNum)%order = MAX(order,recvDeg(iFac))
          ENDIF
          CALL ORDER_TO_NDOFS(Lambda(faceNum)%order,MeshFac(faceNum)%faceType,nDof)
          Lambda(faceNum)%nDof  = nDof
       ENDDO
    ENDDO
    DEALLOCATE(recvDeg)
    DEALLOCATE(sendDeg)
    !------------------------------------------------------------------
    !---- Fill information for faces (EM field), usefull for HDGM
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       tmpSum = 0
       ALLOCATE(hdgEM(iCell)%orderF(MeshCell(iCell)%nNeighb))
       ALLOCATE(hdgEM(iCell)%nDofF(MeshCell(iCell)%nNeighb))
       ALLOCATE(hdgEM(iCell)%sumDofF(MeshCell(iCell)%nNeighb))
       DO k=1,MeshCell(iCell)%nNeighb
          faceNum              = MeshCell(iCell)%facNum(k)
          hdgEM(iCell)%orderF(k)  = Lambda(faceNum)%order
          hdgEM(iCell)%nDofF(k)   = Lambda(faceNum)%nDof
          hdgEM(iCell)%sumDofF(k) = Lambda(faceNum)%nDof + tmpSum
          tmpSum               = hdgEM(iCell)%sumDofF(k)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Count the number of cells and faces of each order
    !------------------------------------------------------------------
    nPCells = 0
    DO iCell=1,Mesh%nCellTet
       order          = hdgEM(MeshCellTet(iCell)%cellNum)%order
       nPCells(order) = nPCells(order) + 1
    ENDDO
    DO iCell=1,Mesh%nCellHex
       order          = hdgEM(MeshCellHex(iCell)%cellNum)%order
       nQCells(order) = nQCells(order) + 1
    ENDDO
    !
    nPFacesA = 0
    nQFacesA = 0
    nPFaces  = 0
    nQFaces  = 0
    DO iFac=1,Mesh%nFacTri
       order              = Lambda(MeshFacTri(iFac)%faceNum)%order
       IF(MeshFac(MeshFacTri(iFac)%faceNum)%isArtificial) THEN
          nPFacesA(order) = nPFacesA(order) + 1
       ELSE
          nPFaces(order)  = nPFaces(order) + 1
       ENDIF
    ENDDO
    DO iFac=1,Mesh%nFacQuad
       order              = Lambda(MeshFacQuad(iFac)%faceNum)%order
       IF(MeshFac(MeshFacQuad(iFac)%faceNum)%isArtificial) THEN
          nQFacesA(order) = nQFacesA(order) + 1
       ELSE
          nQFaces(order)  = nQFaces(order) + 1
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Gather information from other procs
    !------------------------------------------------------------------
    IF (.NOT. parallelRun) THEN
       genOrderMin  = orderMin
       genOrderMax  = orderMax
       genOrderMinH = orderMinH
       genOrderMaxH = orderMaxH
       genPCells    = nPCells
       genPFaces    = nPFaces
       genQCells    = nQCells
       genQFaces    = nQFaces
    ELSE
       CALL GLOBAL_OP(orderMin,  MPI_MIN, genOrderMin)
       CALL GLOBAL_OP(orderMax,  MPI_MAX, genOrderMax)
       CALL GLOBAL_OP(orderMinH, MPI_MIN, genOrderMinH)
       CALL GLOBAL_OP(orderMaxH, MPI_MAX, genOrderMaxH)
       CALL GLOBAL_OP(nPCells,   MPI_SUM, genPCells)
       CALL GLOBAL_OP(nPFaces,   MPI_SUM, genPFaces)
       CALL GLOBAL_OP(nQCells,   MPI_SUM, genQCells)
       CALL GLOBAL_OP(nQFaces,   MPI_SUM, genQFaces)
       CALL GLOBAL_OP(nPFacesA,  MPI_SUM, genPFacesA)
       CALL GLOBAL_OP(nQFacesA,  MPI_SUM, genQFacesA)
       genPFacesA = genPFacesA / 2
       genQFacesA = genQFacesA / 2
       genPFaces  = genPFaces + genPFacesA
       genQFaces  = genQFaces + genQFacesA
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE EVALUATE_POLYNOMIAL_ORDER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE ALLOCATE_FIELDS
    !------------------------------------------------------------------
    !---- Allocate the physical fields
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE INPUT_DATA
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, iFac, nDofS, nDofV
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Allocate fields in EM structure
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       nDofV = hdgEM(iCell)%nDofV
       ALLOCATE(hdgEM(iCell)%E%vec(nDofV,3))
       ALLOCATE(hdgEM(iCell)%H%vec(nDofV,3))
       hdgEM(iCell)%E%vec(:,:) = (0.d0,0.d0)
       hdgEM(iCell)%H%vec(:,:) = (0.d0,0.d0)
    ENDDO
    !------------------------------------------------------------------
    !---- Allocate fields in Lambda structure
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFac
       nDofS = Lambda(iFac)%nDof
       ALLOCATE(Lambda(iFac)%L%vec(nDofS,2))
       Lambda(iFac)%L%vec(:,:) = (0.d0,0.d0)
    ENDDO
    !------------------------------------------------------------------
    !---- Allocate local vector usefull for computations
    !------------------------------------------------------------------
    ALLOCATE(dofXS(nDofSMax), dofYS(nDofSMax), dofZS(nDofSMax))
    ALLOCATE(ExScx(nDofSMax),   EyScx(nDofSMax),   EzScx(nDofSMax)  )
    ALLOCATE(HxScx(nDofSMax),   HyScx(nDofSMax),   HzScx(nDofSMax)  )
    !
    ALLOCATE(dofXV(nDofVMax), dofYV(nDofVMax), dofZV(nDofVMax))
    ALLOCATE(ExVcx(nDofVMax),   EyVcx(nDofVMax),   EzVcx(nDofVMax)  )
    ALLOCATE(HxVcx(nDofVMax),   HyVcx(nDofVMax),   HzVcx(nDofVMax)  )
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE ALLOCATE_FIELDS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DISPATCH_EM_MATRICES
    !------------------------------------------------------------------
    !---- Point to the right FE matrices
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE BASES_LAGRANGE
    USE MESH_CELLS
    USE MESH_FIND_FACES
    USE MESH_DATA
    USE MESH_VERTICES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: cellNum, faceNum, i, iCell, iFac, iHex, iNeighb, &
         iTet, iVert, j, k, nNeighb, order, orderF, orderHex

    REAL(C_DOUBLE)                 :: measureTet, measureSend
    REAL(C_DOUBLE), DIMENSION(12)  :: coorTetRecv, coorTetSend
    REAL(C_DOUBLE), DIMENSION(3,4) :: coorTet
    REAL(C_DOUBLE), DIMENSION(3,8) :: coorHex
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Point to the right FE matrices
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCellTet
       cellNum = MeshCellTet(iCell)%cellNum
       order   =  hdgEM(cellNum)%order
       !
       !---- Volumic matrices
       hdgEM(cellNum)%massMtx  => massTet(order)
       hdgEM(cellNum)%iMassMtx => iMassTet(order)
       hdgEM(cellNum)%xStfMtx  => rdxTet(order)
       hdgEM(cellNum)%yStfMtx  => rdyTet(order)
       hdgEM(cellNum)%zStfMtx  => rdzTet(order)
       !
       !---- Surfacic matrices
       nNeighb = MeshCell(cellNum)%nNeighb
       ALLOCATE(hdgEM(cellNum)%srfMtxE(nNeighb))
       ALLOCATE(hdgEM(cellNum)%srfMtxF(nNeighb))
       ALLOCATE(hdgEM(cellNum)%srfMtxG(nNeighb))
       !
       DO k=1,nNeighb
          orderF = hdgEM(cellNum)%orderF(k)
          hdgEM(cellNum)%srfMtxE(k)%ptr => mSurfTet(order,order)
          hdgEM(cellNum)%srfMtxF(k)%ptr => mSurfTet(order,orderF)
          hdgEM(cellNum)%srfMtxG(k)%ptr => mSurfTet(orderF,orderF)
       ENDDO
    ENDDO
    !
    DO iCell=1,Mesh%nCellHex
       cellNum = MeshCellHex(iCell)%cellNum
       order   =  hdgEM(cellNum)%order
       !
       !---- Volumic matrices
       hdgEM(cellNum)%massMtx  => massHex(order)
       hdgEM(cellNum)%iMassMtx => iMassHex(order)
       hdgEM(cellNum)%xStfMtx  => rdxHex(order)
       hdgEM(cellNum)%yStfMtx  => rdyHex(order)
       hdgEM(cellNum)%zStfMtx  => rdzHex(order)
       !
       !---- Surfacic matrices
       nNeighb = MeshCell(cellNum)%nNeighb
       ALLOCATE(hdgEM(cellNum)%srfMtxE(nNeighb))
       ALLOCATE(hdgEM(cellNum)%srfMtxF(nNeighb))
       ALLOCATE(hdgEM(cellNum)%srfMtxG(nNeighb))
       !
       do_nNeighb : DO k=1,nNeighb
          faceNum = MeshCell(cellNum)%facNum(k)
          orderF  = hdgEM(cellNum)%orderF(k)
          IF (MeshFac(faceNum)%isHybrid) THEN
             !
             !---- Surfacic matrices for hybrid faces (NOT ARTIFICIAL 
             !---- i.e. the hybrid face is not an interface)
             IF (MeshFac(faceNum)%isArtificial) CYCLE do_nNeighb
             iTet       = MeshFac(faceNum)%neighb(1)
             measureTet = MeshCell(iTet)%measure
             DO iVert=1,4
                coorTet(:,iVert) = MeshVert(MeshCell(iTet)%vertex(iVert))%coor(:)
             ENDDO
             iHex     = MeshFac(faceNum)%neighb(2)
             orderHex = hdgEM(iHex)%order
             DO iVert=1,8
                coorHex(:,iVert) = MeshVert(MeshCell(iHex)%vertex(iVert))%coor(:)
             ENDDO
             CALL FIND_DERIVED_FACE_IN_LIST_FROM_VERTICES(MeshFac(faceNum)%vertex,    &
                  MeshFacHyb, Mesh%nFacHyb, faceNum)
             CALL BASES_LAGRANGE_MATRICES_HYBRID_HDG_E(faceNum, orderHex, coorHex)
             CALL BASES_LAGRANGE_MATRICES_HYBRID_HDG_F(faceNum, orderHex, coorHex,    &
                  measureTet, coorTet, orderF)
             hdgEM(cellNum)%srfMtxE(k)%ptr => mSurfHyb(faceNum)
             hdgEM(cellNum)%srfMtxF(k)%ptr => mSurfHybF(faceNum)
             hdgEM(cellNum)%srfMtxG(k)%ptr => mSurfTet(orderF,orderF)
          ELSE
             !
             !---- Surfacic matrices for other faces (not hybrid)
             hdgEM(cellNum)%srfMtxE(k)%ptr => mSurfHex(order,order)
             hdgEM(cellNum)%srfMtxF(k)%ptr => mSurfHex(order,orderF)
             hdgEM(cellNum)%srfMtxG(k)%ptr => mSurfHex(orderF,orderF)
          ENDIF
       ENDDO do_nNeighb
    ENDDO
    !
    !---- Surfacic matrices: special treatment for artificial hybrid faces
    DO iNeighb=1,nbNeighb
       DO iFac=1,neighb(iNeighb)%nbFac
          faceNum = neighb(iNeighb)%faces(iFac)
          IF (MeshFac(faceNum)%isHybrid) THEN
             iCell = MeshFac(faceNum)%neighb(1)
             order = hdgEM(iCell)%order
             IF (MeshCell(iCell)%cellType == "TET1") THEN
                measureSend = MeshCell(iCell)%measure
                k           = 0
                DO iVert=1,4
                   DO i=1,3
                      coorTetSend(i+k) = MeshVert(MeshCell(iCell)%vertex(iVert))%coor(i)
                   ENDDO
                   k = k + 3
                ENDDO
                CALL MPI_SEND(measureSend,  1, MPI_DOUBLE_PRECISION, neighb(iNeighb)%id,&
                     100, defaultComm, mpiError)
                CALL MPI_SEND(coorTetSend, 12, MPI_DOUBLE_PRECISION, neighb(iNeighb)%id,&
                     200, defaultComm, mpiError)
             ELSE
                CALL MPI_RECV(measureTet,   1, MPI_DOUBLE_PRECISION, neighb(iNeighb)%id,&
                     100, defaultComm, status, mpiError)
                CALL MPI_RECV(coorTetRecv, 12, MPI_DOUBLE_PRECISION, neighb(iNeighb)%id,&
                     200, defaultComm, status, mpiError)
                k = 0
                DO iVert=1,4
                   DO i=1,3
                      coorTet(i,iVert) = coorTetRecv(i+k)
                   ENDDO
                   k = k + 3
                ENDDO
                orderHex = hdgEM(iCell)%order
                DO iVert=1,8
                   coorHex(:,iVert) = MeshVert(MeshCell(iCell)%vertex(iVert))%coor(:)
                ENDDO
                k      = MeshFac(faceNum)%numLocal(1)
                orderF = Lambda(faceNum)%order
                CALL FIND_DERIVED_FACE_IN_LIST_FROM_VERTICES(MeshFac(faceNum)%vertex,   &
                     MeshFacHyb, Mesh%nFacHyb, faceNum)
                CALL BASES_LAGRANGE_MATRICES_HYBRID_HDG_E(faceNum, orderHex, coorHex)
                CALL BASES_LAGRANGE_MATRICES_HYBRID_HDG_F(faceNum, orderHex, coorHex,   &
                     measureTet, coorTet, orderF)
                hdgEM(iCell)%srfMtxE(k)%ptr => mSurfHyb(faceNum)
                hdgEM(iCell)%srfMtxF(k)%ptr => mSurfHybF(faceNum)
                hdgEM(iCell)%srfMtxG(k)%ptr => mSurfTet(orderF,orderF)
             ENDIF          
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DISPATCH_EM_MATRICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DISPATCH_PERMUTATIONS
    !------------------------------------------------------------------
    !---- Point to the right permutations
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE NUMERICAL_TYPES
    USE REFERENCE_CELLS
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: cellNum, iCell, order
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Point to the right permutations
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCellTet
       cellNum = MeshCellTet(iCell)%cellNum
       order   = hdgEM(cellNum)%order
       hdgEM(cellNum)%iPermDof%ptr => iPermDofTet(order)
    ENDDO
    !
    DO iCell=1,Mesh%nCellHex
       cellNum = MeshCellHex(iCell)%cellNum
       order   = hdgEM(cellNum)%order
       hdgEM(cellNum)%iPermDof%ptr => iPermDofHex(order)
    ENDDO

    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DISPATCH_PERMUTATIONS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE DISPATCH_MATERIALS
    !------------------------------------------------------------------
    !---- Fill eps, mu, z, sigma, rho, tau 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_MATERIALS
    USE MESH_CELLS
    USE MESH_DATA
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iCell, log
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Dispatch materials
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       log = MeshCell(iCell)%log
       !
       IF (log > 0) THEN ! Actual material
          hdgEM(iCell)%eps   = hdgMat(log)%eps
          hdgEM(iCell)%mu    = hdgMat(log)%mu
          hdgEM(iCell)%z     = hdgMat(log)%z
          hdgEM(iCell)%sigma = hdgMat(log)%sigma
          hdgEM(iCell)%rho   = hdgMat(log)%rho
          hdgEM(iCell)%tau   = hdgMat(log)%tau
       ELSE              ! Vacuum
          hdgEM(iCell)%eps   = 1.d0
          hdgEM(iCell)%mu    = 1.d0
          hdgEM(iCell)%z     = 1.d0
          hdgEM(iCell)%sigma = 0.d0
          hdgEM(iCell)%rho   = 1.d0
          hdgEM(iCell)%tau   = 1.d0
       ENDIF
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE DISPATCH_MATERIALS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PRINT_NUMERICAL_STATS
    !------------------------------------------------------------------
    !---- Module
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Print statistics
    !------------------------------------------------------------------
    IF (idWrite) THEN
       WRITE(6, 10)
       WRITE(6, '(a)')       " Numerical statistics              "
       WRITE(6, 10)
       WRITE(6, '(a)')       "                                   "
       WRITE(6, '(a,i9)')    "       Min. order for tetrahedra : ", genOrderMin
       WRITE(6, '(a,i9)')    "       Max. order for tetrahedra : ", genOrderMax
       WRITE(6, '(a,i9)')    "       Min. order for hexahedra  : ", genOrderMinH
       WRITE(6, '(a,i9)')    "       Max. order for hexahedra  : ", genOrderMaxH
       WRITE(6, '(a)')       "   Number of tetrahedral cells of each order  "
       WRITE(6, '(a,i9)')    "                              P0 : ", genPCells(0)
       WRITE(6, '(a,i9)')    "                              P1 : ", genPCells(1)
       WRITE(6, '(a,i9)')    "                              P2 : ", genPCells(2)
       WRITE(6, '(a,i9)')    "                              P3 : ", genPCells(3)
       WRITE(6, '(a,i9)')    "                              P4 : ", genPCells(4)
       WRITE(6, '(a)')       "   Number of hexahedral cells of each order   "
       WRITE(6, '(a,i9)')    "                              Q0 : ", genQCells(0)
       WRITE(6, '(a,i9)')    "                              Q1 : ", genQCells(1)
       WRITE(6, '(a,i9)')    "                              Q2 : ", genQCells(2)
       WRITE(6, '(a,i9)')    "                              Q3 : ", genQCells(3)
       WRITE(6, '(a,i9)')    "                              Q4 : ", genQCells(4)
       WRITE(6, '(a)')       "   Number of triangular faces of each order   "
       WRITE(6, '(a,i9)')    "                              P0 : ", genPFaces(0)
       WRITE(6, '(a,i9)')    "                              P1 : ", genPFaces(1)
       WRITE(6, '(a,i9)')    "                              P2 : ", genPFaces(2)
       WRITE(6, '(a,i9)')    "                              P3 : ", genPFaces(3)
       WRITE(6, '(a,i9)')    "                              P4 : ", genPFaces(4)
       WRITE(6, '(a)')       "   Number of quadrangular faces of each order "
       WRITE(6, '(a,i9)')    "                              Q0 : ", genQFaces(0)
       WRITE(6, '(a,i9)')    "                              Q1 : ", genQFaces(1)
       WRITE(6, '(a,i9)')    "                              Q2 : ", genQFaces(2)
       WRITE(6, '(a,i9)')    "                              Q3 : ", genQFaces(3)
       WRITE(6, '(a,i9)')    "                              Q4 : ", genQFaces(4)
       WRITE(6, 10)
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE PRINT_NUMERICAL_STATS

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MAXWELL_STRUCTURES
