MODULE MUMPS_STRUCTURES

  !====================================================================
  ! Everything for using MUMPS as global solver in terms of DoFs of the 
  ! hybrid variable in hdgm
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  ! MUMPS structure
  !--------------------------------------------------------------------
  INCLUDE 'zmumps_struc.h'
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! MUMPS data structure
  !--------------------------------------------------------------------
  TYPE(ZMUMPS_STRUC) mumps_par
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Array for index of faces in MUMPS data structure
  !--------------------------------------------------------------------
  INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: indMtxFac
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_MUMPS_SOLVER
    !------------------------------------------------------------------
    !---- Compute MUMPS global solver in terms of DoFs of the hybrid 
    !---- variable in hdgm
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of MUMPS control parameters and data structures
    !------------------------------------------------------------------
    CALL MUMPS_INIT
    !------------------------------------------------------------------
    !---- Compute global index of faces for assembling matrix and RHS
    !------------------------------------------------------------------
    CALL MUMPS_FACES_INDICES
    !------------------------------------------------------------------
    !---- Assembling the matrix in MUMPS package
    !------------------------------------------------------------------
    CALL MUMPS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- MUMPS analysis step 
    !------------------------------------------------------------------
    CALL MUMPS_ANALYSIS
    !------------------------------------------------------------------
    !---- MUMPS factorization step 
    !------------------------------------------------------------------
    CALL MUMPS_FACTORIZATION
    !------------------------------------------------------------------
    !---- Assembling the RHS in MUMPS package
    !------------------------------------------------------------------
    CALL MUMPS_ASSEMBLING_RHS
    !------------------------------------------------------------------
    !---- MUMPS solve step 
    !------------------------------------------------------------------
    CALL MUMPS_SOLVE
    !------------------------------------------------------------------
    !---- Destroy the instance (deallocate MUMPS data structures)
    !------------------------------------------------------------------
    CALL MUMPS_END
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_MUMPS_SOLVER

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_INIT
    !------------------------------------------------------------------
    !---- Initialization of MUMPS control parameters and data structures
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iCell1, iCell2, iFac, j, nDof, facNum,        &
         nNeighb1, nNeighb2, nnz, nnzGlb, nnzLoc, nnzSparse, ordMtx,   &
         ordMtxGlb, sum1, sum2

    REAL(C_DOUBLE) :: ratio
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Order of the matrix
    !------------------------------------------------------------------
    ordMtx = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum = MeshFacI(iFac)%faceNum
       ordMtx = ordMtx + 2*Lambda(facNum)%nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum = MeshFacPhy(iFac)%faceNum
       ordMtx = ordMtx + 2*Lambda(facNum)%nDof
    ENDDO
    !
    !---- Artificial boundary faces
    DO iFac=1,Mesh%nFacArt
       facNum = MeshFacArt(iFac)%faceNum
       ordMtx = ordMtx + Lambda(facNum)%nDof
    ENDDO
    !
    CALL MPI_ALLREDUCE(ordMtx, ordMtxGlb, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    !------------------------------------------------------------------
    !---- Number of non-zero entries in the matrix
    !------------------------------------------------------------------
    nnzLoc = 0
    nnz    = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum   = MeshFacI(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1) - nDof
       !
       iCell2   = MeshFac(facNum)%neighb(2)
       nNeighb2 = MeshCell(iCell2)%nNeighb
       sum2     = hdgEM(iCell2)%sumDofF(nNeighb2) - nDof
       !
       nnzLoc   = nnzLoc + 4*nDof*(nDof + sum1 + sum2)
       nnz      = nnzLoc
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum   = MeshFacPhy(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1)
       !
       nnzLoc   = nnzLoc + 4*nDof*sum1
       nnz      = nnzLoc
    ENDDO
    !
    !---- Artificial boundary faces
    DO iFac=1,Mesh%nFacArt
       facNum   = MeshFacArt(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       sum1     = hdgEM(iCell1)%sumDofF(nNeighb1)
       !
       nnzLoc   = nnzLoc + 4*nDof*sum1
       nnz      = nnz    + 4*nDof*sum1 - 2*nDof**2
    ENDDO
    !
    CALL MPI_ALLREDUCE(nnz, nnzGlb, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    !------------------------------------------------------------------
    !---- Sets up MUMPS control parameters
    !------------------------------------------------------------------
    !
    !---- Definition of the MPI communicator for MUMPS
    mumps_par%COMM = defaultComm
    !
    !---- Non-symmetric matrix
    mumps_par%SYM = 0
    !
    mumps_par%PAR = 1
    !
    !---- Initialization of one instance of the package
    mumps_par%JOB = -1
    CALL ZMUMPS(mumps_par)
    !
    !---- Standard output
    mumps_par%ICNTL(1)  = -1
    mumps_par%ICNTL(2)  = -1
    mumps_par%ICNTL(3)  = -1
    mumps_par%ICNTL(4)  =  2
    !
    !---- Definition of the matrix, distributed assembled format
    mumps_par%ICNTL(5)  =  0
    mumps_par%ICNTL(18) =  3
    !
    !---- Pre-treatment of the matrix (permutation, ordering)
    mumps_par%ICNTL(6)  =  7
    !
    !---- Package used for ordering
    ! mumps_par%ICNTL(7)  = 3  ! SCOTCH
    ! mumps_par%ICNTL(7)  = 4  ! PORD
    ! mumps_par%ICNTL(7)  = 5  ! MeTiS
    ! mumps_par%ICNTL(7)  = 6  ! QAMD
    mumps_par%ICNTL(7)  =  7  ! Automatic choice
    !
    mumps_par%ICNTL(8)  = 77
    ! !
    ! !----Increase MAXIS (cf. doc MUMPS)
    ! mumps_par%ICNTL(14)  = 30
    !
    !---- Sparse right hand side
    mumps_par%ICNTL(20)  =  1
    !------------------------------------------------------------------
    !---- Characterization of th problem
    !------------------------------------------------------------------
    IF (idProc == 0) THEN
       mumps_par%N  = ordMtxGlb
       mumps_par%NZ = nnzGlb
       ratio        = ((DBLE(mumps_par%NZ)/DBLE(mumps_par%N))/DBLE(mumps_par%N))*100.d0
    ENDIF
    mumps_par%NZ_loc = nnzLoc
    !------------------------------------------------------------------
    !---- Allocations of data structure for the definition of the RHS 
    !---- and the local matrices
    !------------------------------------------------------------------
    !
    !---- Global RHS
    IF (idProc == 0) ALLOCATE(mumps_par%RHS(mumps_par%N))
    !
    !---- Local matrix A_loc
    ALLOCATE(mumps_par%IRN_loc(mumps_par%NZ_loc))
    ALLOCATE(mumps_par%JCN_loc(mumps_par%NZ_loc))
    ALLOCATE(mumps_par%A_loc  (mumps_par%NZ_loc))
    !
    !---- Sparse RHS
    nnz = 0
    DO iFac=1,Mesh%nFacAbs
       facNum = MeshFacAbs(iFac)%faceNum
       nnz    = nnz + Lambda(facNum)%nDof
    ENDDO
    CALL MPI_ALLREDUCE(nnz, nnzSparse, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    IF (idProc == 0) THEN
       mumps_par%NRHS        = 1
       mumps_par%NZ_RHS     = 2*nnzSparse
       !
       ALLOCATE(mumps_par%RHS_SPARSE (mumps_par%NZ_RHS))
       ALLOCATE(mumps_par%IRHS_SPARSE(mumps_par%NZ_RHS))
       ALLOCATE(mumps_par%IRHS_PTR   (mumps_par%NRHS+1))
       !
       mumps_par%IRHS_PTR(1) = 1
       mumps_par%IRHS_PTR(2) = 1 + mumps_par%NZ_RHS
    ENDIF
    !------------------------------------------------------------------
    !---- Print MUMPS informations
    !------------------------------------------------------------------
    IF (idWrite) THEN
       WRITE(6, 10)
       WRITE(6, '(a)')       " MUMPS statistics                           "
       WRITE(6, 10)
       WRITE(6, '(a,i12)')   "                             Matrix order : ", mumps_par%N
       WRITE(6, '(a,i12)')   "               Number of non-zero entries : ", mumps_par%NZ
       WRITE(6, '(a,e12.5)') "                 Fill-in ratio percentage : ", ratio
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_INIT

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_FACES_INDICES
    !------------------------------------------------------------------ 
    !---- Compute global index of faces for assembling matrix and RHS 
    !---- in MUMPS package
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, id, iFac, indStart, iNeighb, nbFacNeighbMax, &
         nDof, nDofGlb, nDofLoc, preFac, preDof
    INTEGER(C_INT), DIMENSION(:)              , ALLOCATABLE :: indSend, indRecv
    INTEGER(C_INT), DIMENSION(:,:)            , ALLOCATABLE :: indMtxArt
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Initialization of indMtxFac for MUMPS package
    !
    !---- We first consider pure inner faces and physical boundary faces 
    !---- and finally artificial boundary faces in indMtxFac
    !
    !---- Variables in different loops:
    !     - facNum is the index of the current face in MeshFac structure
    !     - preFac is the index of the previous face in MeshFac structure
    !     - preDof is the number of DoFs of the previous face i.e. the 
    !       number DoFs of the face with index 'preFac' in MeshFac 
    !       structure
    !------------------------------------------------------------------
    !---- Dynamic allocation of indMtxFac
    !------------------------------------------------------------------
    ALLOCATE(indMtxFac(Mesh%nFac))
    !------------------------------------------------------------------
    !---- Pure inner faces and Physical boundary faces
    !------------------------------------------------------------------
    nDof = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum = MeshFacI(iFac)%faceNum
       nDof = nDof + 2*Lambda(facNum)%nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum = MeshFacPhy(iFac)%faceNum
       nDof = nDof + 2*Lambda(facNum)%nDof
    ENDDO
    !
    CALL MPI_SCAN     (nDof, nDofLoc, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    CALL MPI_ALLREDUCE(nDof, nDofGlb, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    !
    !---- Inner faces
    indMtxFac(MeshFacI(1)%faceNum) = nDofLoc - nDof
    preFac = MeshFacI(1)%faceNum
    preDof = Lambda(preFac)%nDof
    DO iFac=2,Mesh%nFacI
       facNum            = MeshFacI(iFac)%faceNum
       indMtxFac(facNum) = indMtxFac(preFac) + 2*preDof
       !
       preFac = facNum
       preDof = Lambda(facNum)%nDof
    ENDDO
    !
    !---- Physical boundary faces
    DO iFac=1,Mesh%nFacPhy
       facNum           = MeshFacPhy(iFac)%faceNum
       indMtxFac(facNum) = indMtxFac(preFac) + 2*preDof
       !
       preFac = facNum
       preDof = Lambda(facNum)%nDof
    ENDDO
    !------------------------------------------------------------------
    !---- Artificial boundary faces 
    !------------------------------------------------------------------
    nbFacNeighbMax = 0
    DO iNeighb=1,nbNeighb
       nbFacNeighbMax = MAX(nbFacNeighbMax,neighb(iNeighb)%nbFac)
    ENDDO
    ALLOCATE(indMtxArt(nbNeighb,nbFacNeighbMax))
    ALLOCATE(indSend(nbFacNeighbMax))
    ALLOCATE(indRecv(nbFacNeighbMax))
    nDof = 0
    DO iNeighb=1,nbNeighb
       IF (idProc < neighb(iNeighb)%id) THEN
          DO iFac=1,neighb(iNeighb)%nbFac
             facNum = neighb(iNeighb)%faces(iFac)
             nDof   = nDof + 2*Lambda(facNum)%nDof
          ENDDO
       ENDIF
    ENDDO
    CALL MPI_SCAN(nDof, nDofLoc, 1, MPI_INT, MPI_SUM, defaultComm, mpiError)
    indStart = nDofLoc - nDof
    DO iNeighb=1,nbNeighb
       IF (idProc < neighb(iNeighb)%id) THEN
          id = 1
          DO iFac=1,neighb(iNeighb)%nbFac
             facNum                  = neighb(iNeighb)%faces(iFac)
             indStart                = indStart + 2*Lambda(facNum)%nDof
             indMtxArt(iNeighb,iFac) = indStart - 2*Lambda(facNum)%nDof
             indSend(id)             = indMtxArt(iNeighb,iFac)
             indMtxFac(facNum)       = indMtxArt(iNeighb,iFac) + nDofGlb 
             id                      = id + 1
          ENDDO
          CALL MPI_SEND(indSend, neighb(iNeighb)%nbFac, MPI_INTEGER,   &
               neighb(iNeighb)%id, 100, defaultComm, mpiError)
       ELSE
          CALL MPI_RECV(indRecv, neighb(iNeighb)%nbFac, MPI_INTEGER,   &
               neighb(iNeighb)%id, 100, defaultComm, status, mpiError)
          id = 1
          DO iFac=1,neighb(iNeighb)%nbFac
             facNum                  = neighb(iNeighb)%faces(iFac)
             indMtxArt(iNeighb,iFac) = indRecv(id)
             indMtxFac(facNum)       = indMtxArt(iNeighb,iFac) + nDofGlb
             id                      = id + 1
          ENDDO
       ENDIF
    ENDDO
    DEALLOCATE(indMtxArt, indSend, indRecv)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_FACES_INDICES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_ASSEMBLING_MATRIX
    !------------------------------------------------------------------
    !---- Assembling the matrix of the global linear system to solve in 
    !---- terms of DoFs of the hybrid variable
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE HDG_MATRICES_STRUCTURES
    USE MESH_DATA
    USE MESH_CELLS
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iCell, iCell1, iCell11, iCell12, iCell2,      &
         iCell21, iCell22, iFac, indCol1, indCol2, indLoc, indRow1,    &
         indRow2, j, jCell11, jCell12, jCell21, jCell22, k, k1, k2, kFac,    &
         nDof, nDofFk, facNum, nNeighb1, nNeighb2, sum1, sum2, sumk, iError

    COMPLEX(C_DOUBLE_COMPLEX) :: valMtx
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Assembling the matrix of the global linear to solve in terms 
    !---- of DoFs of the hybrid variable 
    !------------------------------------------------------------------
    indLoc = 0
    !
    !---- Inner faces
    DO iFac=1,Mesh%nFacI
       facNum   = MeshFacI(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       k1       = MeshFac(facNum)%numLocal(1)
       sum1     = hdgEM(iCell1)%sumDofF(k1) - nDof
       !
       iCell2   = MeshFac(facNum)%neighb(2)
       nNeighb2 = MeshCell(iCell2)%nNeighb
       k2       = MeshFac(facNum)%numLocal(2)
       sum2     = hdgEM(iCell2)%sumDofF(k2) - nDof
       !
       DO i=1,nDof
          indRow1 = indMtxFac(facNum) + i
          indRow2 = indRow1 + nDof
          iCell11 = 2*sum1  + i
          iCell12 = iCell11 + nDof
          iCell21 = 2*sum2  + i
          iCell22 = iCell21 + nDof
          !
          !---- The diagonal blocks of Ke(iCell1) and Ke(iCell2)
          DO j=1,nDof
             indCol1 = indMtxFac(facNum) + j
             indCol2 = indCol1 + nDof
             jCell11 = 2*sum1  + j
             jCell12 = jCell11 + nDof
             jCell21 = 2*sum2  + j
             jCell22 = jCell21 + nDof
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell11,jCell11) + Ke(iCell2)%tab(iCell21,jCell21)
             mumps_par%IRN_loc(indLoc) = indRow1
             mumps_par%JCN_loc(indLoc) = indCol1
             mumps_par%A_loc  (indLoc) = valMtx  
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell11,jCell12) + Ke(iCell2)%tab(iCell21,jCell22)
             mumps_par%IRN_loc(indLoc) = indRow1
             mumps_par%JCN_loc(indLoc) = indCol2
             mumps_par%A_loc  (indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell12,jCell11) + Ke(iCell2)%tab(iCell22,jCell21)
             mumps_par%IRN_loc(indLoc) = indRow2
             mumps_par%JCN_loc(indLoc) = indCol1
             mumps_par%A_loc  (indLoc) = valMtx
             !
             indLoc = indLoc + 1
             valMtx = Ke(iCell1)%tab(iCell12,jCell12) + Ke(iCell2)%tab(iCell22,jCell22)
             mumps_par%IRN_loc(indLoc) = indRow2
             mumps_par%JCN_loc(indLoc) = indCol2
             mumps_par%A_loc  (indLoc) = valMtx
          ENDDO
          !
          !---- The off-diagonal blocks of Ke(iCell1)
          DO k=1,nNeighb1
             IF (k /= k1) THEN
                kFac = MeshCell(iCell1)%facNum(k)
                nDofFk  = hdgEM(iCell1)%nDofF(k)
                sumk    = hdgEM(iCell1)%sumDofF(k) - nDofFk
                DO j=1,nDofFk
                   indCol1 = indMtxFac(kFac) + j
                   indCol2 = indCol1 + nDofFk
                   jCell11 = 2*sumk  + j
                   jCell12 = jCell11 + nDofFk
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell11,jCell11)
                   mumps_par%IRN_loc(indLoc) = indRow1
                   mumps_par%JCN_loc(indLoc) = indCol1
                   mumps_par%A_loc  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell11,jCell12)
                   mumps_par%IRN_loc(indLoc) = indRow1
                   mumps_par%JCN_loc(indLoc) = indCol2
                   mumps_par%A_loc  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell12,jCell11)
                   mumps_par%IRN_loc(indLoc) = indRow2
                   mumps_par%JCN_loc(indLoc) = indCol1
                   mumps_par%A_loc  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell1)%tab(iCell12,jCell12)
                   mumps_par%IRN_loc(indLoc) = indRow2
                   mumps_par%JCN_loc(indLoc) = indCol2
                   mumps_par%A_loc  (indLoc) = valMtx
                ENDDO
             ENDIF
          ENDDO
          !
          !---- The off-diagonal blocks of Ke(iCell2)
          DO k=1,nNeighb2
             IF (k /= k2) THEN
                kFac = MeshCell(iCell2)%facNum(k)
                nDofFk  = hdgEM(iCell2)%nDofF(k)
                sumk    = hdgEM(iCell2)%sumDofF(k) - nDofFk
                DO j=1,nDofFk
                   indCol1 = indMtxFac(kFac) + j
                   indCol2 = indCol1 + nDofFk
                   jCell21 = 2*sumk  + j
                   jCell22 = jCell21 + nDofFk
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell21,jCell21)
                   mumps_par%IRN_loc(indLoc) = indRow1
                   mumps_par%JCN_loc(indLoc) = indCol1
                   mumps_par%A_loc  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell21,jCell22)
                   mumps_par%IRN_loc(indLoc) = indRow1
                   mumps_par%JCN_loc(indLoc) = indCol2
                   mumps_par%A_loc  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell22,jCell21)
                   mumps_par%IRN_loc(indLoc) = indRow2
                   mumps_par%JCN_loc(indLoc) = indCol1
                   mumps_par%A_loc  (indLoc) = valMtx
                   !
                   indLoc = indLoc + 1
                   valMtx = Ke(iCell2)%tab(iCell22,jCell22)
                   mumps_par%IRN_loc(indLoc) = indRow2
                   mumps_par%JCN_loc(indLoc) = indCol2
                   mumps_par%A_loc  (indLoc) = valMtx
                ENDDO
             ENDIF
          ENDDO
       ENDDO
    ENDDO
    !
    !---- Boundary faces
    DO iFac=1,Mesh%nFacB
       facNum   = MeshFacB(iFac)%faceNum
       nDof     = Lambda(facNum)%nDof
       !
       iCell1   = MeshFac(facNum)%neighb(1)
       nNeighb1 = MeshCell(iCell1)%nNeighb
       k1       = MeshFac(facNum)%numLocal(1)
       sum1     = hdgEM(iCell1)%sumDofF(k1) - nDof
       !
       DO i=1,nDof
          indRow1 = indMtxFac(facNum) + i
          indRow2 = indRow1 + nDof
          iCell11 = 2*sum1  + i
          iCell12 = iCell11 + nDof
          DO k=1,nNeighb1
             kFac = MeshCell(iCell1)%facNum(k)
             nDofFk  = hdgEM(iCell1)%nDofF(k)
             sumk    = hdgEM(iCell1)%sumDofF(k) - nDofFk
             DO j=1,nDofFk
                indCol1 = indMtxFac(kFac) + j
                indCol2 = indCol1 + nDofFk
                jCell11 = 2*sumk  + j
                jCell12 = jCell11 + nDofFk
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell11,jCell11)
                mumps_par%IRN_loc(indLoc) = indRow1
                mumps_par%JCN_loc(indLoc) = indCol1
                mumps_par%A_loc  (indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell11,jCell12)
                mumps_par%IRN_loc(indLoc) = indRow1
                mumps_par%JCN_loc(indLoc) = indCol2
                mumps_par%A_loc  (indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell12,jCell11)
                mumps_par%IRN_loc(indLoc) = indRow2
                mumps_par%JCN_loc(indLoc) = indCol1
                mumps_par%A_loc  (indLoc) = valMtx
                !
                indLoc = indLoc + 1
                valMtx = Ke(iCell1)%tab(iCell12,jCell12)
                mumps_par%IRN_loc(indLoc) = indRow2
                mumps_par%JCN_loc(indLoc) = indCol2
                mumps_par%A_loc  (indLoc) = valMtx
             ENDDO
          ENDDO
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Deallocate Ke 
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       DEALLOCATE( Ke(iCell)%tab )
    ENDDO
    DEALLOCATE( Ke )
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_ASSEMBLING_MATRIX

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_ANALYSIS
    !------------------------------------------------------------------
    !---- MUMPS analysis step 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    USE TIMERS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError 

    TYPE(Timer)    :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    
    !------------------------------------------------------------------
    !---- Analysis
    !------------------------------------------------------------------
    mumps_par%JOB = 1
    CALL myTimer%Start()
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------    
    iError = mumps_par%INFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in analysis step of MUMPS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Print MUMPS informations
    !------------------------------------------------------------------
    IF(idWrite) THEN
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       " ** Analysis step statistics                "
       IF ( mumps_par%INFOG(7) == 0 ) THEN
          WRITE(6, '(a)')    "                        Ordering based on :          AMD"
       ELSEIF ( mumps_par%INFOG(7) == 2 ) THEN
          WRITE(6, '(a)')    "                        Ordering based on :          AMF"
       ELSEIF ( mumps_par%INFOG(7) == 3 ) THEN
          WRITE(6, '(a)')    "                        Ordering based on :       SCOTCH"
       ELSEIF ( mumps_par%INFOG(7) == 4 ) THEN
          WRITE(6, '(a)')    "                        Ordering based on :         PORD"
       ELSEIF ( mumps_par%INFOG(7) == 5 ) THEN
          WRITE(6, '(a)')    "                        Ordering based on :        METIS"
       ELSEIF ( mumps_par%INFOG(7) == 6 ) THEN
          WRITE(6, '(a)')    "                        Ordering based on :         QAMD"
       ENDIF
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       "        Estimated memory for running facto. "
       WRITE(6, '(a,i12)')   "                Most consuming proc. (MB) : ", mumps_par%INFOG(16)
       WRITE(6, '(a,i12)')   "                  Sum over all proc. (MB) : ", mumps_par%INFOG(17)
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       "           Time spent on analysis (seconds) "
       WRITE(6, '(a,e12.5)') "                                 CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e12.5)') "                                Wall time : ", myTimer%SYSTEM_TIME()
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_ANALYSIS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_FACTORIZATION
    !------------------------------------------------------------------
    !---- MUMPS factorization step 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    USE TIMERS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: iError 

    TYPE(Timer)    :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    
    !------------------------------------------------------------------
    !---- Factorization
    !------------------------------------------------------------------
    mumps_par%JOB = 2
    CALL myTimer%Start()
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------    
    iError = mumps_par%INFO(1)
    IF (iError /= 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in factorization step of MUMPS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Print MUMPS informations
    !------------------------------------------------------------------
    IF(idWrite) THEN
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       " ** Factorization step statistics           "
       WRITE(6, '(a,i12)')   "             Number of entries in factors : ", mumps_par%INFOG(29)
       WRITE(6, '(a)')       "      Memory effectively used during facto. "
       WRITE(6, '(a,i12)')   "                Most consuming proc. (MB) : ", mumps_par%INFOG(21)
       WRITE(6, '(a,i12)')   "                  Sum over all proc. (MB) : ", mumps_par%INFOG(22)
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       "     Time spent for factorization (seconds) "
       WRITE(6, '(a,e12.5)') "                                 CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e12.5)') "                                Wall time : ", myTimer%SYSTEM_TIME()
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_FACTORIZATION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_ASSEMBLING_RHS
    !------------------------------------------------------------------
    !---- Assembling the RHS of the global linear system to solve in 
    !---- terms of DoFs of the hybrid variable
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE SOURCE_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, i, id, idRecv, idSend, iFac, k, nDof,    &
         nDofMax, nDofRecv, nDofSend

    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: idAbs
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: nDofAbs
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: irhsSend
    INTEGER(C_INT), DIMENSION(:), ALLOCATABLE :: irhsRecv

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: rhsSend
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: rhsRecv
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Computation of idAbs and nDofAbs on the host 
    !
    !---- idAbs(1:nproc)   => id of processors associated to sub-meshes
    !                         with absorbing boundary faces     
    !                         idAbs(k) =  k if the sub-mesh associated to the
    !                                          kth processor contains
    !                                          absorbing boundary faces
    !                         idAbs(k) = -1 else
    !
    !---- nDofAbs(1:nproc) => number DoFs of the right-hand side 
    !                         associated to the sub-meshes corresponding 
    !                         to a processor
    !------------------------------------------------------------------
    IF (idProc /= 0) THEN
       IF (Mesh%nFacAbs /= 0) THEN
          idSend   = idProc
          nDofSend = 0
          DO iFac=1,Mesh%nFacAbs
             facNum   = MeshFacAbs(iFac)%faceNum
             nDofSend = nDofSend + 2*Lambda(facNum)%nDof
          ENDDO
       ELSE
          idSend = -1
          nDofSend = 0
       ENDIF
       CALL MPI_SEND(idSend,   1, MPI_INTEGER, 0, 100, defaultComm, mpiError)
       CALL MPI_SEND(nDofSend, 1, MPI_INTEGER, 0, 200, defaultComm, mpiError)
    ELSE
       ALLOCATE(idAbs(nbProc), nDofAbs(nbProc))
       IF (Mesh%nFacAbs /= 0) THEN
          idAbs(1) = 0
          nDof     = 0
          DO iFac=1,Mesh%nFacAbs
             facNum = MeshFacAbs(iFac)%faceNum
             nDof   = nDof + 2*Lambda(facNum)%nDof
          ENDDO
       ELSE
          idAbs(1) = -1
       ENDIF
       DO id=1,nbProc-1
          CALL MPI_RECV(idRecv,   1, MPI_INTEGER, id, 100, defaultComm, status, mpiError)
          CALL MPI_RECV(nDofRecv, 1, MPI_INTEGER, id, 200, defaultComm, status, mpiError)
          idAbs(id + 1)   = idRecv
          nDofAbs(id + 1) = nDofRecv
       ENDDO
    ENDIF
    !------------------------------------------------------------------
    !---- Assembling the RHS of the global linear to solve in terms of 
    !---- DoFs of the hybrid variable (assembled on the host) 
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFacAbs
       facNum = MeshFacAbs(iFac)%faceNum
       nDof   = nDof + 2*Lambda(facNum)%nDof
    ENDDO
    CALL MPI_ALLREDUCE(nDof, nDofMax, 1, MPI_INTEGER, MPI_MAX, defaultComm, mpiError)
    !
    IF (idProc /= 0) THEN
       IF (Mesh%nFacAbs /= 0) THEN
          ALLOCATE(irhsSend(nDofMax))
          ALLOCATE( rhsSend(nDofMax))
          k = 1
          DO iFac=1,Mesh%nFacAbs
             facNum = MeshFacAbs(iFac)%faceNum
             nDof   = Lambda(facNum)%nDof
             DO i=1,nDof
                irhsSend(k)     = indMtxFac(facNum) + i
                rhsSend(k)      = rhsAbs(iFac)%vec(i,1)
                irhsSend(k + 1) = indMtxFac(facNum) + i + nDof
                rhsSend(k + 1)  = rhsAbs(iFac)%vec(i,2)
                k               = k + 2
             ENDDO
          ENDDO
          CALL MPI_SEND(irhsSend, k-1, MPI_INTEGER,        0, 300, defaultComm, mpiError)
          CALL MPI_SEND(rhsSend , k-1, MPI_DOUBLE_COMPLEX, 0, 400, defaultComm, mpiError)
       ENDIF
    ELSE
       ALLOCATE(irhsRecv(nDofMax))
       ALLOCATE(rhsRecv(nDofMax))
       k = 1
       IF (Mesh%nFacAbs /= 0) THEN
          DO iFac=1,Mesh%nFacAbs
             facNum = MeshFacAbs(iFac)%faceNum
             nDof   = Lambda(facNum)%nDof
             DO i=1,nDof
                mumps_par%IRHS_SPARSE(k)     = indMtxFac(facNum) + i
                mumps_par%RHS_SPARSE (k)     = rhsAbs(iFac)%vec(i,1)
                mumps_par%IRHS_SPARSE(k + 1) = indMtxFac(facNum) + i + nDof
                mumps_par%RHS_SPARSE (k + 1) = rhsAbs(iFac)%vec(i,2)
                k                            = k + 2
             ENDDO
          ENDDO
       ENDIF
       do_id: DO id=1,nbProc-1
          IF (idAbs(id + 1) == -1) CYCLE do_id
          CALL MPI_RECV(irhsRecv, nDofAbs(id + 1), MPI_INTEGER,        & 
               id, 300, defaultComm, status, mpiError)
          CALL MPI_RECV(rhsRecv,  nDofAbs(id + 1), MPI_DOUBLE_COMPLEX, &
               id, 400, defaultComm, status, mpiError)
          DO i=1,nDofAbs(id + 1)
             mumps_par%IRHS_SPARSE(k) = irhsRecv(i)
             mumps_par%RHS_SPARSE(k)  = rhsRecv(i)
             k                        = k + 1
          ENDDO
       ENDDO do_id
    ENDIF
    !------------------------------------------------------------------
    !---- Deallocations
    !------------------------------------------------------------------
    IF (idProc == 0) THEN
       DEALLOCATE(idAbs   )
       DEALLOCATE(nDofAbs )
       DEALLOCATE(irhsRecv)
       DEALLOCATE(rhsRecv )
    ENDIF
    !
    IF ((idProc /= 0) .AND. (Mesh%nFacAbs /= 0)) THEN
       DEALLOCATE(irhsSend)
       DEALLOCATE(rhsSend )
    ENDIF
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_ASSEMBLING_RHS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_SOLVE
    !------------------------------------------------------------------
    !---- MUMPS solve step 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    USE TIMERS
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables and arrays
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iError, iFac, indLocu, indLocw, k, nDof

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: LambdaGlb

    TYPE(Timer)    :: myTimer
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Solve
    !------------------------------------------------------------------
    mumps_par%JOB = 3
    CALL myTimer%Start()
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    iError = mumps_par%INFO(1)
    IF (iError < 0) THEN
       IF (idWrite) WRITE(6, '(a,i0)') " Error in solve step of MUMPS : INFO(1) = ", iError
       CALL TILT
    ENDIF
    !------------------------------------------------------------------
    !---- Print MUMPS informations
    !------------------------------------------------------------------
    IF (idWrite) THEN
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       " ** Solve step statistics                  "
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       "        Memory used during solve step       "
       WRITE(6, '(a,i12)')   "                Most consuming proc. (MB) : ", mumps_par%INFOG(30)
       WRITE(6, '(a,i12)')   "                  Sum over all proc. (MB) : ", mumps_par%INFOG(31)
       WRITE(6, '(a)')       " "
       WRITE(6, '(a)')       "           Time spent on solve    (seconds) "
       WRITE(6, '(a,e12.5)') "                                 CPU time : ", myTimer%CPU_TIME()
       WRITE(6, '(a,e12.5)') "                                Wall time : ", myTimer%SYSTEM_TIME()
       WRITE(6, 10)
    ENDIF
    !------------------------------------------------------------------
    !---- Distribution of the DoFs of the hybrid variable on each 
    !---- processor (the RHS, stored on the host, stores the solution)
    !------------------------------------------------------------------
    ALLOCATE(LambdaGlb(mumps_par%N))
    IF (idProc == 0) THEN
       DO k=1,mumps_par%N
          LambdaGlb(k) = mumps_par%RHS(k)
       ENDDO
    ENDIF
    CALL MPI_BCAST(LambdaGlb, mumps_par%N, MPI_DOUBLE_COMPLEX, 0, defaultComm, mpiError)
    DO iFac=1,Mesh%nFac
       nDof = Lambda(iFac)%nDof
       DO i=1,nDof
          indLocu = indMtxFac(iFac) + i
          indLocw = indLocu + nDof
          Lambda(iFac)%L%vec(i,1) = LambdaGlb(indLocu)
          Lambda(iFac)%L%vec(i,2) = LambdaGlb(indLocw)
       ENDDO

    ENDDO
    DEALLOCATE(LambdaGlb)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
10  FORMAT(60('-'))
    !------------------------------------------------------------------
  END SUBROUTINE MUMPS_SOLVE

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MUMPS_END
    !------------------------------------------------------------------
    !---- MUMPS final step : deallocations 
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------

    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    IMPLICIT NONE
    !------------------------------------------------------------------
    !---- I/O variable
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !--- Final phase - Deallocations
    !------------------------------------------------------------------
    IF (idProc == 0) DEALLOCATE(mumps_par%RHS)
    !
    DEALLOCATE(mumps_par%IRN_loc)
    DEALLOCATE(mumps_par%JCN_loc)
    DEALLOCATE(mumps_par%A_loc  )
    DEALLOCATE(indMtxFac        )
    !------------------------------------------------------------------
    !---- Destroy the instance (deallocate internal data structures)
    !------------------------------------------------------------------
    mumps_par%JOB = -2
    CALL ZMUMPS(mumps_par)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------ 
  END SUBROUTINE MUMPS_END

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE MUMPS_STRUCTURES
