MODULE SOURCE_STRUCTURES

  !====================================================================
  ! Initialize all the necessary source structures
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  USE MATRICES_STRUCTURES
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Current source type
  !--------------------------------------------------------------------
  CHARACTER(LEN=2, KIND=C_CHAR) :: currentSrcType

  !--------------------------------------------------------------------
  ! Angular frequency of the source (rad.m^-1)
  !--------------------------------------------------------------------
  REAL(C_DOUBLE) :: omega
  
  !--------------------------------------------------------------------
  ! Procedure pointer to compute incident fields and to initialize fields
  !--------------------------------------------------------------------
  PROCEDURE(), POINTER :: COMPUTE_INCID_FIELDS    => NULL()
  PROCEDURE(), POINTER :: COMPUTE_RHS_ABS         => NULL()
  PROCEDURE(), POINTER :: COMPUTE_CURRENT_SRC     => NULL()
  PROCEDURE(), POINTER :: CELL_SOL                => NULL()
  PROCEDURE(), POINTER :: FACE_SOL                => NULL()
  PROCEDURE(), POINTER :: CELL_CURRENT_SRC        => NULL()
  PROCEDURE(), POINTER :: COMPUTE_RHS_CURRENT_SRC => NULL()
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Declaration of the incident field
  !--------------------------------------------------------------------
  TYPE(complex_vector_field), DIMENSION(:), ALLOCATABLE :: Einc, Hinc
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Declaration of the current source
  !--------------------------------------------------------------------
  TYPE(vector_real), DIMENSION(:), ALLOCATABLE :: Jsrc
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! RHS of the linear system to solve in terms of DoFs of the hybrid 
  ! variable
  !--------------------------------------------------------------------
  TYPE(complex_vector_field), DIMENSION(:), ALLOCATABLE :: rhsAbs
  TYPE(vector_complex),       DIMENSION(:), ALLOCATABLE :: rhsJloc
  TYPE(vector_complex),       DIMENSION(:), ALLOCATABLE :: rhsJglb
  !--------------------------------------------------------------------   

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE PREPROCESS_SOURCES
    !------------------------------------------------------------------
    !---- Allocate fields and handle the different initializations
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MATH_PHYS_CONSTANTS
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    USE PLANE_WAVE
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, iCell, iFac, nDofS, nDofV, nNeighb, sumDofF
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Call preprocessing for each case, and point CELL_SOL, 
    !---- FACE_SOL, CELL_CURRENT_SRC to the approriate routines
    !------------------------------------------------------------------
    !
    !---- Sort source type
    SELECT CASE(sourceType)
    CASE('PW')
       incidentType   = 'AB'
       currentSrcType = 'NO'
       CELL_SOL     => CELL_SOL_PLANE_WAVE_FREQ
       FACE_SOL     => FACE_SOL_PLANE_WAVE_FREQ
       CALL PREPROCESS_PLANE_WAVE
       omega        = omegaPW
    CASE('GA')
       incidentType      = 'NO'
       currentSrcType    = 'GA'
       omega             = 2*Pi*freq/c0
    END SELECT
    !------------------------------------------------------------------
    !---- Allocate incident and rhs abs fields 
    !---- Point INCID_FIELDS and RHS_ABS to the right routine
    !------------------------------------------------------------------
    SELECT CASE(incidentType)
    CASE('AB') !---- Use ABC faces for incident field
       ALLOCATE(Einc(Mesh%nFacAbs), Hinc(Mesh%nFacAbs))
       DO iFac=1,Mesh%nFacAbs
          faceNum = MeshFacAbs(iFac)%faceNum
          iCell   = MeshFac(faceNum)%neighb(1)
          nDofS   = hdgEM(iCell)%nDofS
          ALLOCATE(Einc(iFac)%vec(nDofS,3),Hinc(iFac)%vec(nDofS,3))
       ENDDO
       COMPUTE_INCID_FIELDS    => COMPUTE_INCID_ABS_FIELDS
       !
       ALLOCATE(rhsAbs(Mesh%nFacAbs))
       DO iFac=1,Mesh%nFacAbs
          faceNum = MeshFacAbs(iFac)%faceNum
          nDofS = Lambda(faceNum)%nDof
          ALLOCATE(rhsAbs(iFac)%vec(nDofS,2))
       ENDDO
       COMPUTE_RHS_ABS         => COMPUTE_RHS_ABS_CONTRIBUTION
       !
    CASE('NO') !---- No incident field
       COMPUTE_INCID_FIELDS => COMPUTE_INCID_DUMMY
       COMPUTE_RHS_ABS      => COMPUTE_RHS_ABS_DUMMY
    END SELECT
    !------------------------------------------------------------------
    !---- Allocate current src and rhs abs fields 
    !---- Point CURRENT_SRC_FIELDS and RHS_CURRENT_SRC to the right routine
    !------------------------------------------------------------------
    SELECT CASE(currentSrcType)
    CASE('GA') !---- Gaussian current src term
       ALLOCATE(Jsrc(Mesh%nCell),rhsJloc(Mesh%nCell),rhsJglb(Mesh%nCell))
       DO iCell=1,Mesh%nCell
          nDofV               = hdgEM(iCell)%nDofV
          Jsrc(iCell)%ndof    = nDofV
          rhsJloc(iCell)%ndof = 6*nDofV
          nNeighb             = MeshCell(iCell)%nNeighb
          sumDofF             = hdgEM(iCell)%sumDofF(nNeighb)
          rhsJglb(iCell)%ndof = 2*sumDofF
          ALLOCATE(Jsrc(iCell)%tab(nDofV))
          ALLOCATE(rhsJloc(iCell)%tab(6*nDofV))
          ALLOCATE(rhsJglb(iCell)%tab(2*sumDofF))
       ENDDO
       CELL_CURRENT_SRC        => CELL_CURRENT_SRC_GAUSSIAN
       COMPUTE_RHS_CURRENT_SRC => COMPUTE_RHS_CURRENT_SRC_GAUSSIAN
       !
       CALL CELL_CURRENT_SRC
    CASE('NO') !---- No current src term
       COMPUTE_RHS_CURRENT_SRC => COMPUTE_RHS_CURRENT_SRC_DUMMY
    END SELECT
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE PREPROCESS_SOURCES

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_INCID_ABS_FIELDS(omegaValue)
    !------------------------------------------------------------------
    !---- Compute incident field via Silver-Muller faces
    !
    !---- IN
    !---- omegaValue : angular frequency
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE INPUT_DATA
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN) :: omegaValue
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: faceNum, k, nDofS, nDofV, iCell, iFac, order
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute incident field
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFacAbs
       faceNum = MeshFacAbs(iFac)%faceNum
       iCell   = MeshFac(faceNum)%neighb(1)
       nDofS   = hdgEM(iCell)%nDofS
       nDofV   = hdgEM(iCell)%nDofV
       order   = hdgEM(iCell)%order
       CALL FACE_SOL(faceNum, nDofS, nDofV, order, omegaValue)
       !
       DO k=1,nDofS
          Einc(iFac)%vec(k,1) = ExScx(k)
          Einc(iFac)%vec(k,2) = EyScx(k)
          Einc(iFac)%vec(k,3) = EzScx(k)
          Hinc(iFac)%vec(k,1) = HxScx(k)
          Hinc(iFac)%vec(k,2) = HyScx(k)
          Hinc(iFac)%vec(k,3) = HzScx(k)
       ENDDO
       !
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_INCID_ABS_FIELDS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_INCID_DUMMY(omegaValue)
    !------------------------------------------------------------------
    !---- Dummy incident field routine
    !
    !---- IN
    !---- omegaValue : angular frequency
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    REAL(C_DOUBLE), INTENT(IN) :: omegaValue
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_INCID_DUMMY

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE COMPUTE_RHS_ABS_CONTRIBUTION
    !------------------------------------------------------------------
    !---- Compute absorbing boundary contribution (Silver-Muller) for RHS
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_CELLS
    USE MESH_DATA
    USE MESH_FACES
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: facNum, i, iCell, iFac, j, k, nDofF, nDofS

    REAL(C_DOUBLE) :: aera

    COMPLEX(C_DOUBLE_COMPLEX) :: tmp_usum, tmp_wsum, zero
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(:), ALLOCATABLE :: tmp_u, tmp_w
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute RHS
    !------------------------------------------------------------------
    !
    !---- Initialization
    zero = (0.d0, 0.d0)
    DO iFac=1,Mesh%nFacAbs
       rhsAbs(iFac)%vec(:,:) = zero
    ENDDO
    !
    !---- Compute incident fields
    CALL COMPUTE_INCID_FIELDS(omega)
    !
    !--- Fill rhsAbs
    !------------------------------------------------------------------
    DO iFac=1,Mesh%nFacAbs
       facNum = MeshFacAbs(iFac)%faceNum
       aera   = MeshFac(facNum)%measure
       nDofF  = Lambda(facNum)%nDof
       iCell  = MeshFac(facNum)%neighb(1) 
       k      = MeshFac(facNum)%numLocal(1)
       nDofS  = hdgEM(iCell)%nDofS
       !
       ALLOCATE( tmp_u(nDofS), tmp_w(nDofS) )
       !
       DO i=1,nDofS
          tmp_u(i) = -(MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(2,1) &
               - MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(3,1))      &
               * Einc(iFac)%vec(i,1)*aera                                    &
               - (MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(3,1)      &
               - MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(1,1))      &
               * Einc(iFac)%vec(i,2)*aera                                    &
               - (MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(1,1)      &
               - MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(2,1))      &
               * Einc(iFac)%vec(i,3)*aera                                    &
               - (MeshFac(facNum)%axis(1,1)*Hinc(iFac)%vec(i,1)              &
               + MeshFac(facNum)%axis(2,1)*Hinc(iFac)%vec(i,2)               &
               + MeshFac(facNum)%axis(3,1)*Hinc(iFac)%vec(i,3))*aera
          !
          tmp_w(i) = -(MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(2,2) &
               - MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(3,2))      &
               * Einc(iFac)%vec(i,1)*aera                                    &
               - (MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(3,2)      &
               - MeshCell(iCell)%normal(3,k)*MeshFac(facNum)%axis(1,2))      &
               * Einc(iFac)%vec(i,2)*aera                                    &
               - (MeshCell(iCell)%normal(2,k)*MeshFac(facNum)%axis(1,2)      &
               - MeshCell(iCell)%normal(1,k)*MeshFac(facNum)%axis(2,2))      &
               * Einc(iFac)%vec(i,3)*aera                                    &
               - (MeshFac(facNum)%axis(1,2)*Hinc(iFac)%vec(i,1)              &
               + MeshFac(facNum)%axis(2,2)*Hinc(iFac)%vec(i,2)               &
               + MeshFac(facNum)%axis(3,2)*Hinc(iFac)%vec(i,3))*aera
       ENDDO
       !
       DO i=1,nDofF
          tmp_usum = zero
          tmp_wsum = zero
          DO j=1,nDofS
             tmp_usum = tmp_usum + hdgEM(iCell)%srfMtxF(k)%ptr%tab(j,i)*tmp_u(j)
             tmp_wsum = tmp_wsum + hdgEM(iCell)%srfMtxF(k)%ptr%tab(j,i)*tmp_w(j)
          ENDDO
          rhsAbs(iFac)%vec(i,1) = tmp_usum
          rhsAbs(iFac)%vec(i,2) = tmp_wsum
       ENDDO
       DEALLOCATE(tmp_u, tmp_w)
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RHS_ABS_CONTRIBUTION

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_RHS_ABS_DUMMY
    !------------------------------------------------------------------
    !---- Dummy RHS for absorbing boundaries
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RHS_ABS_DUMMY
  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE CELL_CURRENT_SRC_GAUSSIAN
    !------------------------------------------------------------------
    !---- Compute Gaussian current source term
    !
    !---- IN
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE GAUSSIAN
    USE MESH_CELLS
    USE MESH_DATA
    USE NUMERICAL_TYPES
    USE PARALLEL_STRUCTURES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, iCell, j, k, nDofV, order
    
    REAL(C_DOUBLE) :: norm, normGlb

    REAL(C_DOUBLE), DIMENSION(nDofVmax) :: vec
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute source current term
    !------------------------------------------------------------------
    DO iCell=1,Mesh%nCell
       order = hdgEM(iCell)%order
       nDofV = hdgEM(iCell)%nDofV
       CALL CELL_GAUSSIAN(iCell, nDofV, order)
       DO i=1,nDofV
          Jsrc(iCell)%tab(i) = GaussCell(i)
       ENDDO
    ENDDO
    !
    !---- Normalization
    norm   = 0.d0
    vec(:) = 0.d0
    DO iCell=1,Mesh%nCell
       DO k=1,Jsrc(iCell)%ndof
          vec(k) = 0.d0
          DO j=1,Jsrc(iCell)%ndof
             vec(k) = vec(k) + hdgEM(iCell)%massMtx%tab(k,j)*Jsrc(iCell)%tab(j)
          ENDDO
          norm = norm + MeshCell(iCell)%measure*vec(k)
       ENDDO
    ENDDO
    CALL MPI_ALLREDUCE(norm, normGlb, 1, MPI_DOUBLE_PRECISION, MPI_SUM, &
         defaultComm, mpiError)
    DO iCell=1,Mesh%nCell
       DO k=1,Jsrc(iCell)%ndof
          Jsrc(iCell)%tab(k) = Jsrc(iCell)%tab(k)/normGlb
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE CELL_CURRENT_SRC_GAUSSIAN

  !====================================================================
  !====================================================================
  !====================================================================
  
  SUBROUTINE COMPUTE_RHS_CURRENT_SRC_GAUSSIAN(iCell, nDofV, invAe, sumDofF, Be)
    !------------------------------------------------------------------
    !---- Compute Gaussian current source term contribution for RHS
    !
    !---- IN
    !---- iCell   : element
    !---- nDofV   : number of DoFs in cell
    !---- invAe   : inverse matrix [Ae]^-1 of size [6*nDofV,6*nDofV]
    !---- sumDofF : sum of number of DoFs of faces (DoFs for Lambda) of cell
    !---- Be      : matrix Be of size [2*sumDofF,6*nDofV]
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE GAUSSIAN
    USE MATH_PHYS_CONSTANTS
    USE MATRICES_STRUCTURES
    USE MESH_CELLS
    USE NUMERICAL_TYPES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell, nDofV, sumDofF

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(6*nDofV,6*nDofV),   INTENT(IN) :: invAe
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(2*sumDofF,6*nDofV), INTENT(IN) :: Be
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT) :: i, j, k, order, sizeF, sizeV

    REAL(C_DOUBLE),            DIMENSION(6*nDofV,6*nDofV) :: Mtx

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(6*nDofV)         :: res
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(6*nDofV)         :: srcVec
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Compute source current contribution for local systems i.e.
    !---- [Ae]^-1*M*[Jex,Jey,Jez,0,0,0]^T
    !------------------------------------------------------------------
    !
    !---- Mass matrix: M
    Mtx = 0.d0
    DO i=1,nDofV
       DO j=1,nDofV
          Mtx(i+2*nDofV,j+2*nDofV) =                                   &
               hdgEM(iCell)%massMtx%tab(i,j)*MeshCell(iCell)%measure
       ENDDO
    ENDDO
    !
    !---- Current source term vector: [Jex,Jey,Jez,0,0,0]^T
    srcVec = (0.d0,0.d0)
    DO i=1,nDofV
       !!! WARNING -> we only consider Jez, i.e. [0,0,Jez,0,0,0]^T
       srcVec(i+2*nDofV) = Jsrc(iCell)%tab(i)*z0*pi
    ENDDO
    !
    !---- M*[Jex,Jey,Jez,0,0,0]^T, result is stored in res
    sizeV = 6*nDofV
    CALL MATVECMUL(Mtx, srcVec, sizeV, sizeV, res)
    !
    !---- [Ae]^-1*M*[Jex,Jey,Jez,0,0,0]^T, result is stored in rhsJloc(iCell)%tab
    rhsJloc(iCell)%tab = (0.d0,0.d0)
    DO i=1,sizeV
       DO j=1,sizeV
          rhsJloc(iCell)%tab(i) = rhsJloc(iCell)%tab(i) + invAe(i,j)*res(j)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Compute source current contribution for global systems i.e.
    !---- Be*[Ae]^-1*M*[Jex,Jey,Jez,0,0,0]^T
    !---- Result is stored in rhsJglb(iCell)%tab
    !------------------------------------------------------------------
    sizeF = 2*sumDofF
    rhsJglb(iCell)%tab = (0.d0,0.d0)
    DO i=1,sizeF
       DO j=1,sizeV
          rhsJglb(iCell)%tab(i) = rhsJglb(iCell)%tab(i) + Be(i,j)*rhsJloc(iCell)%tab(j)
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RHS_CURRENT_SRC_GAUSSIAN

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE COMPUTE_RHS_CURRENT_SRC_DUMMY(iCell, nDofV, invAe, sumDofF, Be)
    !------------------------------------------------------------------
    !---- Dummy Gaussian current source term contribution for RHS
    !
    !---- IN
    !---- iCell   : element
    !---- nDofV   : number of DoFs in cell
    !---- invAe   : inverse matrix [Ae]^-1 of size [6*nDofV,6*nDofV]
    !---- sumDofF : sum of number of DoFs of faces (DoFs for Lambda) of cell
    !---- Be      : matrix Be of size [2*sumDofF,6*nDofV]
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    INTEGER(C_INT), INTENT(IN) :: iCell, nDofV, sumDofF

    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(6*nDofV,6*nDofV),   INTENT(IN) :: invAe
    COMPLEX(C_DOUBLE_COMPLEX), DIMENSION(2*sumDofF,6*nDofV), INTENT(IN) :: Be
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE COMPUTE_RHS_CURRENT_SRC_DUMMY

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE SOURCE_STRUCTURES
