MODULE POINT_CLOUDS

  !====================================================================
  ! Module containing routines to handle point clouds structures and
  ! turn them into exploitable meshes
  !====================================================================

  !--------------------------------------------------------------------
  ! Shared modules
  !--------------------------------------------------------------------
  USE ISO_C_BINDING
  !--------------------------------------------------------------------
  IMPLICIT NONE
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! Data structure associated to a *squared* regularly spaced point cloud
  !
  ! dataFile      ==> a file name in which information is stored
  ! nPts1D        ==> number of points in each direction
  ! tag           ==> a tag that will be applied on the resulting faces
  ! offset        ==> a possible offset of z coordinates
  ! altitude(:,:) ==> altitude of each point
  !--------------------------------------------------------------------
  TYPE point_cloud
     CHARACTER(LEN=1000, KIND=C_CHAR)            :: dataFile

     INTEGER(C_INT)                              :: nPts1D, tag, subSize

     REAL(C_DOUBLE)                              :: offset, xyStep
     REAL(C_DOUBLE), DIMENSION(:,:), ALLOCATABLE :: altitude
  END TYPE point_cloud
  !--------------------------------------------------------------------

CONTAINS

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE READ_LAYER_DATA(layer)
    !------------------------------------------------------------------
    !---- Read a point cloud data and store it into a layer
    !
    !---- IN
    !---- layer    : pointCloud structure
    !
    !---- OUT
    !---- layer    : modified in the process
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE IO_FILES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    TYPE(point_cloud), INTENT(INOUT) :: layer
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)            :: nRows, nCols, iPtX, iPtY, nUnit

    REAL(C_DOUBLE)            :: altitude
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Read data
    !---- Offset is automatically applied
    !------------------------------------------------------------------
    CALL GET_SIZE_OF_FILE(TRIM(layer%dataFile), nRows, nCols)
    layer%nPts1D = FLOOR(SQRT(DBLE(nRows)))
    ALLOCATE(layer%altitude(layer%subSize,layer%subSize))
    !
    CALL GET_NEW_FILE_UNIT(nUnit)
    OPEN(UNIT=nUnit, FILE=TRIM(layer%dataFile))
    DO iPtX=1,layer%nPts1D
       DO iPtY=1,layer%nPts1D
          READ(nUnit, *) altitude
          IF ((iPtX <= layer%subSize) .AND. (iPtY <= layer%subSize)) &
               layer%altitude(iPtX,iPtY) = altitude + layer%offset
       ENDDO
    ENDDO
    CLOSE(nUnit)
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a,a)'   ) 'Finished reading data from layer ',TRIM(layer%dataFile)
    !------------------------------------------------------------------
  END SUBROUTINE READ_LAYER_DATA

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_VERTICES_FROM_LAYER_ARRAY(layers)
    !------------------------------------------------------------------
    !---- Build vertices for future mesh from layers
    !
    !---- IN
    !---- layers   : array of pointCloud structure
    !
    !---- OUT
    !---- layers   : modified in the process
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    TYPE(point_cloud), DIMENSION(:), INTENT(INOUT) :: layers
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    INTEGER(C_INT)               :: nVertices, iPtX, iPtY, iSurf, subSize

    REAL(C_DOUBLE), DIMENSION(3) :: vert
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build vertices
    !------------------------------------------------------------------
    DO iSurf=1,SIZE(layers)
       subSize = layers(iSurf)%subSize
       !
       DO iPtX=1,subSize
          DO iPtY=1,subSize
             vert(1) = iPtX*layers(iSurf)%xyStep
             vert(2) = iPtY*layers(iSurf)%xyStep
             vert(3) = layers(iSurf)%altitude(iPtX,iPtY)
             CALL ADD_VERTEX(vert)
          ENDDO
       ENDDO
       !
       WRITE(6,'(a,i0)') 'Finished vertices for layer ',iSurf
       !

    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a,a)'   ) 'Finished building vertices from layers'
    WRITE(6,'(a,i9,a)') 'I created ',Mesh%nVert,' vertices'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_VERTICES_FROM_LAYER_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE BUILD_FACES_FROM_LAYER_ARRAY(layers)
    !------------------------------------------------------------------
    !---- Build faces for future mesh from layers
    !
    !---- IN
    !---- layers   : array of pointCloud structure
    !
    !---- OUT
    !---- layers   : modified in the process
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MESH_DATA
    USE MESH_FACES
    USE MESH_CELLS
    USE MESH_VERTICES
    USE MESH_ELEMENTS
    !------------------------------------------------------------------
    !---- I/O stuff
    !------------------------------------------------------------------
    TYPE(point_cloud), DIMENSION(:), INTENT(INOUT) :: layers
    !------------------------------------------------------------------
    !---- Local stuff
    !------------------------------------------------------------------
    CHARACTER(LEN=4, KIND=C_CHAR)                  :: eltType
    
    INTEGER(C_INT)                                 :: iPtX, iPtY, &
         iSurf, subSize, nVertsPrevious, facNum
    INTEGER(C_INT)                                 :: boundaryTag = 4
    INTEGER(C_INT), DIMENSION(4)                   :: vertices
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Build faces
    !------------------------------------------------------------------
    eltType = 'QUA1'
    nVertsPrevious = 0
    !
    DO iSurf=1,SIZE(layers)
       subSize = layers(iSurf)%subSize
       !
       DO iPtX=1,subSize-1
          DO iPtY=1,subSize-1
             vertices(1) = (iPtX-1)*subSize + iPtY
             vertices(2) =   (iPtX)*subSize + iPtY
             vertices(3) =   (iPtX)*subSize + iPty + 1
             vertices(4) = (iPtX-1)*subSize + iPty + 1
             vertices(:) = vertices(:) + nVertsPrevious
             CALL ADD_FACE(vertices, layers(iSurf)%tag, eltType, facNum)
          ENDDO
       ENDDO
       !
       nVertsPrevious = nVertsPrevious + subSize**2
       WRITE(6,'(a,i0)') 'Finished faces for layer ',iSurf
    ENDDO
    !
    !---- Faces connecting layers
    nVertsPrevious = 0
    !
    DO iSurf=1,SIZE(layers)-1
       subSize = layers(iSurf)%subSize
       !
       !---- First face
       iPtY = 1
       DO iPtX=1,subSize-1
          vertices(1) = (iPtX-1)*subSize + iPtY
          vertices(2) =   (iPtX)*subSize + iPtY
          vertices(3) =   (iPtX)*subSize + iPty + subSize**2
          vertices(4) = (iPtX-1)*subSize + iPty + subSize**2
          vertices(:) = vertices(:) + nVertsPrevious
          CALL ADD_FACE(vertices, boundaryTag, eltType, facNum)
       ENDDO
       !
       !---- Second face
       iPtY = subSize
       DO iPtX=1,subSize-1
          vertices(1) = (iPtX-1)*subSize + iPtY
          vertices(2) =   (iPtX)*subSize + iPtY
          vertices(3) =   (iPtX)*subSize + iPty + subSize**2
          vertices(4) = (iPtX-1)*subSize + iPty + subSize**2
          vertices(:) = vertices(:) + nVertsPrevious
          CALL ADD_FACE(vertices, boundaryTag, eltType, facNum)
       ENDDO
       !
       !---- Third face
       iPtX = 1
       DO iPtY=1,subSize-1
          vertices(1) = iPtY
          vertices(2) = iPtY + 1
          vertices(3) = iPty + 1 + subSize**2
          vertices(4) = iPty     + subSize**2
          vertices(:) = vertices(:) + nVertsPrevious
          CALL ADD_FACE(vertices, boundaryTag, eltType, facNum)
       ENDDO
       !
       !---- Fourth face
       iPtX = subSize
       DO iPtY=1,subSize-1
          vertices(1) = (iPtX-1)*subSize + iPtY
          vertices(2) = (iPtX-1)*subSize + iPtY + 1
          vertices(3) = (iPtX-1)*subSize + iPty + 1 + subSize**2
          vertices(4) = (iPtX-1)*subSize + iPty     + subSize**2
          vertices(:) = vertices(:) + nVertsPrevious
          CALL ADD_FACE(vertices, boundaryTag, eltType, facNum)
       ENDDO
       !
       nVertsPrevious = nVertsPrevious + subSize**2
    ENDDO
    !------------------------------------------------------------------
    !---- Build derived faces
    !------------------------------------------------------------------
    CALL BUILD_DERIVED_FACES
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    WRITE(6,'(a,a)'   ) 'Finished building faces from layers'
    WRITE(6,'(a,i9,a)') 'I created ',Mesh%nFac,' faces'
    !------------------------------------------------------------------
  END SUBROUTINE BUILD_FACES_FROM_LAYER_ARRAY

  !====================================================================
  !====================================================================
  !====================================================================

  SUBROUTINE MULT_BY_TUKEY_WINDOW(layer, param)
    !------------------------------------------------------------------
    !---- Multiply our vertices by tukey window
    !
    !---- IN
    !---- layer : layer to smoothly
    !     r1     : parameter of the tukey window
    !     r2     : ...
    !
    !---- OUT
    !
    !------------------------------------------------------------------
    !---- Modules
    !------------------------------------------------------------------
    USE MATH_PHYS_CONSTANTS
    USE MESH_DATA
    USE MESH_VERTICES
    !------------------------------------------------------------------
    !---- I/O variables
    !------------------------------------------------------------------
    TYPE(point_cloud), INTENT(INOUT) :: layer

    REAL(C_DOUBLE), INTENT(IN)           :: param
    !------------------------------------------------------------------
    !---- Local variables
    !------------------------------------------------------------------
    INTEGER(C_INT)                :: iPtX, iPtY, subSize

    REAL(C_DOUBLE)              :: mean, coorMin(2), coorMax(2), r1, r2
    !------------------------------------------------------------------
    !---- Testing for bad input
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
    !---- Calcul the window size
    !------------------------------------------------------------------
    subSize = layer%subSize
    !
    coorMin(:) = 1*layer%xyStep
    coorMax(:)= subSize*layer%xyStep
    !
    r1 = coorMin(1)+ param*(coorMax(1)-coorMin(1)+1)
    r2 = coorMin(2)+ param*(coorMax(2)-coorMin(2)+1)
    !
    mean    = SUM(layer%altitude(:,:))/(MAX(1,subSize**2))
    !
    DO iPtX=1,subSize
       DO iPtY=1,subSize
          IF(iPtX*layer%xyStep <= r1/2.0) THEN
             layer%altitude(iPtX,iPtY)=1/2.0 *(1+COS(2.0*Pi/r1*(iPtX*layer%xyStep-r1/2.0)))*&
                  layer%altitude(iPtX,iPtY) + 1/2.0 *(1-COS(2.0*Pi/r1*(iPtX*layer%xyStep-r1/2.0)))*&
                  mean
             !
          ENDIF
          IF(iPtY*layer%xyStep <= r2/2.0) THEN
             layer%altitude(iPtX,iPtY)=1/2.0*(1+COS(2.0*Pi/r2*(iPtY*layer%xyStep-r2/2.0)))*&
                  layer%altitude(iPtX,iPtY) + 1/2.0*(1-COS(2.0*Pi/r2*(iPtY*layer%xyStep-r2/2.0)))*&
                  mean
             !
          ENDIF
          IF(iPtX*layer%xyStep >= coorMax(1)-r1/2.0) THEN
             layer%altitude(iPtX,iPtY)=1/2.0 *(1+COS(2.0*Pi/r1*(iPtX*layer%xyStep-coorMax(1)+r1/2.0)))*&
                  layer%altitude(iPtX,iPtY) + 1/2.0 *(1-COS(2.0*Pi/r1*(iPtX*layer%xyStep-coorMax(1)+r1/2.0)))*&
                  mean
             !
          ENDIF
          IF(iPtY*layer%xyStep >= coorMax(2)-r2/2.0) THEN
             layer%altitude(iPtX,iPtY)=1/2.0 *(1+COS(2.0*Pi/r2*(iPtY*layer%xyStep-coorMax(2)+r2/2.0)))*&
                  layer%altitude(iPtX,iPtY) + 1/2.0 *(1-COS(2.0*Pi/r2*(iPtY*layer%xyStep-coorMax(2)+r2/2.0)))*&
                  mean
             !
          ENDIF
       ENDDO
    ENDDO
    !------------------------------------------------------------------
    !---- Testing for errors
    !------------------------------------------------------------------
    !
    !------------------------------------------------------------------
  END SUBROUTINE MULT_BY_TUKEY_WINDOW

  !====================================================================
  !====================================================================
  !====================================================================

END MODULE POINT_CLOUDS
